rem Paints patches by successive comparisons with a master image.
@rem
@rem Updated:
@rem   22-August-2022 Upgraded for IM v7.
@rem

setlocal enabledelayedexpansion


rem ---------------------------------------------------------------
rem Check parameters.


rem For sampling another image, we have 2 inputs: one to get sub-image coords, the other to sample.
rem This is so a monochrome %ppMASTER% we can lookup a monochrome image but grab a colour sample.
rem More commonly they will be the same image.

rem Input files
set ppMASTER=mlf.png
if "%%"=="" (
  echo Requires ppMASTER
  exit /B 1
) else if not exist %ppMASTER% (
  echo Can't find file ppMASTER = %ppMASTER%
  exit /B 1
)

rem Sampled patches need ppSAMPLED.
rem They also needs ppSUBSRCH, which defaults to ppSAMPLED.
if "%ppSAMPLED%"=="" set ppSAMPLED=%ppMASTER%
if "%ppSUBSRCH%"=="" set ppSUBSRCH=%ppSAMPLED%

rem Output file
if "%ppOUTPUT%"=="" set ppOUTPUT=ppOut.png

rem If ppMARKED is given, creates an image marked with limits.

rem Temporary files.
rem
set fOnePix=%TEMP%\ppOnePix.png
set fOnePixPostProc=%TEMP%\ppOnePixPostProc.png
set fSmc=%TEMP%\ppSmc.png
set fMask=%TEMP%\ppMask.png
set COMP_RESULT_BAT=%TEMP%\ppComp.bat

rem Background is one of:
rem 0: constant colour ppBACK_COL
rem 1: %ppMASTER% overlaid with ppBACK_COL at 90% opacity
rem 2: constant colour, average of %ppMASTER%
rem 3: graduated colour using the corners of %ppMASTER%
rem 4: NYI sample
rem 5: NYI another image, so we can "fade" from one image to another
if "%ppBACK_TYPE%"=="" set /A ppBACK_TYPE=0
if "%ppBACK_COL%"=="" set ppBACK_COL=gray50

rem Patches are one of:
rem 0: squares of average required colour
rem 1: circles of average required colour
rem 2: squares made from sparse-color at corners
rem 3: circles made from squares made from sparse-color at corners
rem 4: square samples of sampled image
rem 5: circle samples of sampled image
rem 6: NYI squares of colour of closest single pixel taken from sampled image (ie a palette)
rem 7: NYI circles of colour of closest single pixel taken from sampled image (ie a palette)
if "%ppPATCH_TYPE%"=="" set ppPATCH_TYPE=0

rem Location of centre of interest (CoI), as fraction of width and height.
rem (0,0) is top-left.
rem Used for not placing small patches distant from centre.
rem Eg 0.5 for centre of image.
rem (Was 0.51,0.416)
if "%ppFracWiTest%"=="" set ppFracWiTest=0.5
if "%ppFracHtTest%"=="" set ppFracHtTest=0.5

rem Maximum radius, as fraction of shortest dimension.
if "%ppMAX_RAD_FACT%"=="" set ppMAX_RAD_FACT=1/2

rem Patches of all sizes will be drawn with centres within ppCoiLo, provided ppPatchNumFracPow=1.
rem No patches will be drawn outside ppCoiHi.
rem These are fractions of the semi-diagonal.
if "%ppCoiLo%"=="" set ppCoiLo=0.2
if "%ppCoiHi%"=="" set ppCoiHi=1.0

rem Raise PatchNumFrac to this power before trial.
rem Set to >1 to reduce number of smaller pathes drawn, or 0<ppPatchNumFracPow<1 to increase then.
if "%ppPatchNumFracPow%"=="" set ppPatchNumFracPow=1

rem When "-compare" is less than this, don't write the patch.
rem Zero = always change, up to 1 = never change.
if "%ppMaxPatchDiff%"=="" set ppMaxPatchDiff=0.2

rem Margins as fractions of patch radius. Can be negative (patches will overlap edges).
if "%ppMargFracL%"=="" set ppMargFracL=0
if "%ppMargFracT%"=="" set ppMargFracT=0
if "%ppMargFracR%"=="" set ppMargFracR=0
if "%ppMargFracB%"=="" set ppMargFracB=0

rem Gutter between patches as fractions of patch radius. Can be negative (ie patches overlap each other).
if "%ppGutterFracX%"=="" set ppGutterFracX=0
if "%ppGutterFracY%"=="" set ppGutterFracY=0

rem If ppPOST_PROC is given, post-processing of a patch before drawing it.
rem eg  set ppPOST_PROC=-alpha set -channel A -evaluate multiply 0.8

rem If non-zero, processing will stop at radius ppFLOOR_RAD (in pixels).

rem Multiplier from one radius to the next.
if "%ppRAD_MULT%"=="" set ppRAD_MULT=4/5

rem Whether to save each frame.
if "%ppSAVE_FRAMES%"=="" set ppSAVE_FRAMES=0


rem ---------------------------------------------------------------
rem Process parameters

if "%ppMASTER%"=="" (
  echo Requires ppMASTER
)

set OUT_EXT=%ppOUTPUT:*.=%

set /A SamplePatch=(ppPATCH_TYPE^^4)^|(ppPATCH_TYPE^^5)
echo SamplePatch=%SamplePatch%

FOR /F "usebackq" %%L IN (`%IMG7%magick identify ^
  -format "CoiDiff=%%[fx:%ppCoiHi%-%ppCoiLo%]" ^
  xc:`) DO set %%L

FOR /F "usebackq" %%L IN (`%IMG7%magick identify ^
  -format "Err=%%[fx:%CoiDiff%<=0]" ^
  xc:`) DO set %%L

if %Err%==1 (
  echo ppCoiLo >= ppCoiHi
  exit /B 1
)


FOR /F "usebackq" %%L ^
IN (`%IMG7%magick identify -format "Width=%%w\nHeight=%%h" %ppMASTER%`) ^
DO set %%L

FOR /F "usebackq" %%L ^
IN (`%IMG7%magick identify ^
  -format "WiTest=%%[fx:int(%ppFracWiTest%*%Width%)]\nHtTest=%%[fx:int(%ppFracHtTest%*%Height%)]" xc:`) ^
DO set %%L

set /A SemiWidth=%Width%/2
set /A SemiHeight=%Height%/2

set /A SemiDiag2=%SemiWidth%*%SemiWidth%+%SemiHeight%*%SemiHeight%
echo SemiDiag2=%SemiDiag2%

set /A MaxRad=%WIDTH%*%ppMAX_RAD_FACT%
set /A MinRad=1

if "%ppFLOOR_RAD%"=="" set /A ppFLOOR_RAD=%MinRad%
if /I %ppFLOOR_RAD% LSS %MinRad% set /A ppFLOOR_RAD=%MinRad%

set /A nPatches=0
set /A nPatchesAdded=0

set /A nTrials=0

set /A ppPATCH_TYPE2Or3=0
set /A ppPATCH_TYPE4Or5=0

set DO_MASK=
set SaveMasterCrop=0
set PatchTypePalette=0
set PatchTypeSample=0

rem Next are not /A
rem For squares use 2, 1.
rem For circles use 17/10, 17/20.
echo ppPATCH_TYPE=%ppPATCH_TYPE%
if %ppPATCH_TYPE%==0 (
  set BlockFrac=2
  set SemiBlockFrac=1
) else if %ppPATCH_TYPE%==1 (
  set BlockFrac=17/10
  set SemiBlockFrac=17/20
) else if %ppPATCH_TYPE%==2 (
  set BlockFrac=2
  set SemiBlockFrac=1
  set SaveMasterCrop=1
  set /A ppPATCH_TYPE2Or3=1
) else if %ppPATCH_TYPE%==3 (
  set BlockFrac=2
  set SemiBlockFrac=1
  set SaveMasterCrop=1
  set DO_MASK=%fMask% -gravity NorthWest -alpha off -compose CopyOpacity -composite
  set /A ppPATCH_TYPE2Or3=1
) else if %ppPATCH_TYPE%==4 (
  set BlockFrac=2
  set SemiBlockFrac=1
  set SaveMasterCrop=1
  set /A ppPATCH_TYPE4Or5=1
  set PatchTypeSample=1
) else if %ppPATCH_TYPE%==5 (
  set BlockFrac=17/10
  set SemiBlockFrac=17/20
  set SaveMasterCrop=1
  set DO_MASK=%fMask% -gravity NorthWest -alpha off -compose CopyOpacity -composite
  set /A ppPATCH_TYPE4Or5=1
  set PatchTypeSample=1
) else if %ppPATCH_TYPE%==6 (
  set BlockFrac=2
  set SemiBlockFrac=1
  set PatchTypePalette=1
) else if %ppPATCH_TYPE%==7 (
  set BlockFrac=17/10
  set SemiBlockFrac=17/20
  set DO_MASK=%fMask% -gravity NorthWest -alpha off -compose CopyOpacity -composite
  set PatchTypePalette=1
) else (
  echo Unknown ppPATCH_TYPE=%ppPATCH_TYPE%
  exit /B 1
)

rem For type=3 we should do comparison of small square (less than circle rad)
rem then extract larger square (= circle rad) for sparse-colour gradient.


echo SaveMasterCrop=%SaveMasterCrop%

if %SaveMasterCrop%==1 (
  set SMC=^^^( -clone 0 -write %fSmc% +delete ^^^)
) else (
  set SMC=
)

echo SMC=%SMC%

rem Write a sample
if not "%ppMARKED%"=="" (

  FOR /F "usebackq" %%L ^
IN (`%IMG7%magick identify ^
-format "RadLo=%%[fx:int(sqrt(%SemiDiag2%)*%ppCoiLo%+0.5)]\nRadHi=%%[fx:int(sqrt(%SemiDiag2%)*%ppCoiHi%+0.5)]" ^
xc:`) ^
DO set /A %%L

  %IMG7%magick ^
    %ppMASTER% ^
    -fill Red ^
    -draw "line 0,%HtTest% %Width%,%HtTest%" ^
    -draw "line %WiTest%,0 %WiTest%,%Height%" ^
    -fill None ^
    -stroke Green -draw "translate %WiTest%,%HtTest% circle 0,0 0,!RadLo!" ^
    -stroke Red -draw "translate %WiTest%,%HtTest% circle 0,0 0,!RadHi!" ^
    %ppMARKED%
)


rem ---------------------------------------------------------------
rem Do the processing.

rem Count the number of radiuses we will use.

set /A Rad=%MaxRad%
set /A nRads=0

:loopCount
@set /A nRads+=1
@set PrevRad=%Rad%
@set /A Rad=%Rad%*%ppRAD_MULT%
@if %PrevRad%==%Rad% set /A Rad-=1
@if %Rad% GEQ %MinRad% goto loopCount

echo nRads=%nRads%

set /A Rad=%MaxRad%

set /A RadNum=%nRads%


%IMG7%magick -size %Width%x%Height% xc:gray50 %ppOUTPUT%

if %ppBACK_TYPE%==0 (
  %IMG7%magick -size %Width%x%Height% xc:%ppBACK_COL% %ppOUTPUT%
) else if %ppBACK_TYPE%==1 (
  rem FIXME
  %IMG7%magick ^
    %ppMASTER% ^
    ^( -size %Width%x%Height% xc:%ppBACK_COL% ^
       -alpha set -channel A -evaluate multiply 0.9 ^
    ^) ^
    -composite ^
    %ppOUTPUT%
) else if %ppBACK_TYPE%==2 (
  %IMG7%magick ^
    %ppMASTER% ^
    -scale "1x1^!" ^
    -scale "%Width%x%Height%^!" ^
    %ppOUTPUT%
) else if %ppBACK_TYPE%==3 (
  set /A Wm2=%Width%-2
  set /A Hm2=%Height%-2

  %IMG7%magick ^
    %ppMASTER% ^
    -gravity Center ^
    -chop !Wm2!x!Hm2!+0+0 ^
    -gravity Center ^
    -resize "%Width%x%Height%^!" ^
    %ppOUTPUT%

) else (
  echo Unknown ppBACK_TYPE=%ppBACK_TYPE%
  exit /B 1
)


:loopRad
FOR /F "usebackq" %%L ^
IN (`%IMG7%magick identify -format "MargL=%%[fx:int(%ppMargFracL%*%Rad%)]\nMargT=%%[fx:int(%ppMargFracT%*%Rad%)]\nMargR=%%[fx:int(%ppMargFracR%*%Rad%)]\nMargB=%%[fx:int(%ppMargFracB%*%Rad%)]" xc:`) ^
DO set %%L

FOR /F "usebackq" %%L ^
IN (`%IMG7%magick identify -format "RadmHalf=%%[fx:%Rad%-0.5]\nGutterX=%%[fx:int(%ppGutterFracX%*%Rad%)]\nGutterY=%%[fx:int(%ppGutterFracY%*%Rad%)]" xc:`) ^
DO set %%L

set /A TwoRad=%Rad%+%Rad%

@rem set /A nW=^(%Width%-^(%MargL%^)-^(%MargR%^)^)/^(%TwoRad%+(%GutterX%)^)+1
@rem set /A nH=^(%Height%-^(%MargT%^)-^(%MargB%^)^)/^(%TwoRad%+(%GutterY%)^)+1

FOR /F "usebackq" %%L ^
IN (`%IMG7%magick identify -format "nW=%%[fx:int((%Width%-(%MargL%)-(%MargR%)+(%GutterX%))/(%TwoRad%+(%GutterX%)))]" xc:`) ^
DO set %%L

FOR /F "usebackq" %%L ^
IN (`%IMG7%magick identify -format "nH=%%[fx:int((%Height%-(%MargT%)-(%MargB%)+(%GutterY%))/(%TwoRad%+(%GutterY%)))]" xc:`) ^
DO set %%L

if %nW%==0 set nW=1
if %nH%==0 set nH=1

echo nW=%nW% nH=%nH%

rem When nW and nW would not be integer we centralise the patches.

set /A LeftCent=^(^(%Width%-^(%MargL%^)-^(%MargR%^)^)-%nW%*^(%TwoRad%+^(%GutterX%^)^)+(%GutterX%^)^)/2
set /A TopCent=^(^(%Height%-^(%MargT%^)-^(%MargB%^)^)-%nH%*^(%TwoRad%+^(%GutterY%^)^)+(%GutterY%^)^)/2
echo LeftCent=%LeftCent% TopCent=%TopCent%

if %MaxRad% GTR %MinRad% (
FOR /F "usebackq" %%L IN (`%IMG7%magick identify ^
-format "CircFrac=%%[fx:(%Rad%-%MinRad%)/(%MaxRad%-%MinRad%)]" ^
xc:`) DO set %%L

FOR /F "usebackq" %%L IN (`%IMG7%magick identify ^
-format "CircFrac2=%%[fx:!CircFrac!*!CircFrac!]" ^
xc:`) DO set %%L
) else (
  set CircFrac=1
  set CircFrac2=1
)

FOR /F "usebackq" %%L IN (`%IMG7%magick identify &
-format "PatchNumFrac=%%[fx:(%RadNum%/%nRads%)^%ppPatchNumFracPow%]" ^
xc:`) DO set %%L

echo CircFrac2=%CircFrac2%

echo PatchNumFrac=%PatchNumFrac%

echo Rad=%Rad% nW=%nW% nH=%nH%

set /A nPatches+=%nW%*%nH%

@rem set /A BlockW=%TwoRad%
@rem set /A BlockH=%TwoRad%

set /A BlockW=%Rad%*%BlockFrac%
set /A BlockH=%Rad%*%BlockFrac%

set /A SemiBlockW=%Rad%*%SemiBlockFrac%
set /A SemiBlockH=%Rad%*%SemiBlockFrac%

rem We might, at this level, resize FACE and REFERENCE so we just need one pixel from each.

@set /A ChopW=%TwoRad%-2
@set /A ChopH=%TwoRad%-2

@if %ChopW% LSS 0 set /A ChopW=0
@if %ChopH% LSS 0 set /A ChopH=0

if not "%DO_MASK%"=="" (
  %IMG7%magick ^
    -size %BlockW%x%BlockH% xc:#000 ^
    -fill #fff ^
    -draw "circle %SemiBlockW%,%SemiBlockH% %SemiBlockW%,0" ^
    %fMask%
)

set EndTrials=0
set TrialedPrevRow=0

echo off
for /L %%Y in (1, 1, %nH%) do (
  set TrialedThisRow=0

  if !EndTrials!==0 (
  set /A TOP=%TopCent%+%MargT%+^(%%Y-1^)*^(%TwoRad%^+%GutterY%^)
  set /A CentY=!TOP!+%Rad%
  set /A BlockTop=!CentY!-%SemiBlockH%
  set /A BlockBot=!BlockTop!+%BlockH%
  set /A dy=!CentY!-%HtTest%
  set /A dy2=!dy!*!dy!
  for /L %%X in (1, 1, %nW%) do (
    set /A LEFT=%LeftCent%+%MargL%+^(%%X-1^)*^(%TwoRad%+%GutterX%^)
    set /A CentX=!LEFT!+%Rad%
    set /A dx=!CentX!-%WiTest%
    set /A dx2=!dx!*!dx!
    set /A FromCent2=!dx2!+!dy2!
    @rem echo CircFrac2=%CircFrac2% FromCent2=!FromCent2! SemiDiag2=%SemiDiag2%

    rem echo PatchNumFrac=%PatchNumFrac%  FromCent2=!FromCent2!  SemiDiag2=%SemiDiag2%

    FOR /F "usebackq" %%L IN (`%IMG7%magick identify -format "TRIAL_IT=%%[fx:%PatchNumFrac%^>(sqrt(!FromCent2!/%SemiDiag2%)-%ppCoiLo%)/%CoiDiff%]" xc:`) DO set /A %%L

    rem echo TRIAL_IT=!TRIAL_IT!

    if !TRIAL_IT!==1 (
      set TrialedThisRow=1
      set /A nTrials+=1
      echo Rad=%Rad% X=%%X Y=%%Y LEFT=!LEFT! TOP=!TOP! SemiDiag2=%SemiDiag2% FromCent2=!FromCent2!
      set /A BlockLeft=!CentX!-%SemiBlockW%
      set /A BlockRight=!BlockLeft!+%BlockW%

      set ONE_CROP=-crop %BlockW%x%BlockH%+!BlockLeft!+!BlockTop! +repage
      echo !ONE_CROP!

      set PATCH_DIFF=0
      FOR /F "usebackq" %%L IN (`%IMG7%magick %ppMASTER% %ppOUTPUT% !ONE_CROP! -metric RMSE -compare -format "PATCH_DIFF=%%[distortion]" info:`) DO set %%L
      FOR /F "usebackq" %%L IN (`%IMG7%magick identify -format "CHANGE_IT=%%[fx:!PATCH_DIFF!^>%ppMaxPatchDiff%]" xc:`) DO set /A %%L

      @rem echo PATCH_DIFF=!PATCH_DIFF!

      if !CHANGE_IT!==1 (
        %IMG7%magick %ppMASTER% !ONE_CROP! %SMC% -scale 1x1 -write %fOnePix% %ppPOST_PROC% %fOnePixPostProc%

        if %ppPATCH_TYPE%==0 (
          %IMG7%magick %ppOUTPUT% -tile %fOnePixPostProc% -draw "rectangle !BlockLeft!,!BlockTop! !BlockRight!,!BlockBot!" %ppOUTPUT%
        ) else if %ppPATCH_TYPE%==1 (
          FOR /F "usebackq" %%L IN (`%IMG7%magick identify -format "CentXmHalf=%%[fx:!CentX!-0.5]\nCentYmHalf=%%[fx:!CentY!-0.5]" xc:`) DO set %%L
          %IMG7%magick %ppOUTPUT% -tile %fOnePixPostProc% -draw "circle !CentXmHalf!,!CentYmHalf! !CentXmHalf!,!TOP!" %ppOUTPUT%
        ) else if %ppPATCH_TYPE2Or3% NEQ 0 (

@rem If chop isn't possible (because crop ran over edge of picture), chop doesn't happen.

          FOR /F "usebackq" %%L ^
IN (`%IMG7%magick identify -format "smcW=%%[fx:w]\nsmcWm2=%%[fx:w-2]\nsmcH=%%[fx:h]\nsmcHm2=%%[fx:h-2]" %fSmc%`) ^
DO set %%L

          if !smcWm2! LSS 0 set !smcWm2!=0
          if !smcHm2! LSS 0 set !smcHm2!=0

@rem echo smcW=!smcW! smcWm2=!smcWm2! smcH=!smcH! smcHm2=!smcHm2!

@rem But position is then wrong if overlapped left or top edge

          if !BlockLeft! LSS 0 set BlockLeft=0
          if !BlockTop! LSS 0 set BlockTop=0

          %IMG7%magick ^
            %ppOUTPUT% ^
            ^( %fSmc% -gravity Center ^
               -chop !smcWm2!x!smcHm2!+0+0 ^
               -resize "!smcW!x!smcH!^!" ^
               %DO_MASK% ^
               %ppPOST_PROC% ^
            ^) ^
            -gravity NorthWest ^
            -geometry +!BlockLeft!+!BlockTop! ^
            -compose Over -composite ^
            %ppOUTPUT%
        ) else if %PatchTypeSample%==1 (
          call %PICTBAT%srchImg %ppSUBSRCH% %fSmc% %COMP_RESULT_BAT%
          call %COMP_RESULT_BAT%

          FOR /F "usebackq" %%L IN (`%IMG7%magick identify -format "CHANGE_IT=%%[fx:!COMP_FLT!^<!PATCH_DIFF!]" xc:`) DO set /A %%L

          if !CHANGE_IT!==1 %IMG7%magick ^
            %ppOUTPUT% ^
            ^( %ppSAMPLED% -crop !COMP_CROP! +repage %DO_MASK% %ppPOST_PROC% ^) ^
            -gravity NorthWest ^
            -geometry +!BlockLeft!+!BlockTop! ^
            -compose Over -composite ^
            %ppOUTPUT%
        ) else if %PatchTypePalette%==1 (

          FOR /F "tokens=1-4 usebackq delims=()@, " %%R ^
IN (`%IMG7%magick compare ^
    -similarity-threshold 0 -dissimilarity-threshold 1 -subimage-search ^
    -metric RMSE %ppSUBSRCH% %fOnePix% NULL: 2^>^&1`) ^
DO (
    set COMP_INT=%%R
    set COMP_FLT=%%S
    set COMP_XW=%%T
    set COMP_YW=%%U
  )

          echo Palette: !COMP_INT! !COMP_FLT! !COMP_XW! !COMP_YW!

          FOR /F "usebackq" %%L IN (`%IMG7%magick identify -format "CHANGE_IT=%%[fx:!COMP_FLT!^<!PATCH_DIFF!]" xc:`) DO set /A %%L

          if !CHANGE_IT!==1 %IMG7%magick ^
            %ppOUTPUT% ^
            ^( %ppSAMPLED% -crop 1x1+!COMP_XW!+!COMP_YW! +repage -scale %TwoRad%x%TwoRad% %DO_MASK% %ppPOST_PROC% ^) ^
            -gravity NorthWest ^
            -geometry +!BlockLeft!+!BlockTop! ^
            -compose Over -composite ^
            %ppOUTPUT%

        ) else (
          echo Unknown ppPATCH_TYPE=%ppPATCH_TYPE%
          exit /B 1
        )

        if !CHANGE_IT!==1 (
          set /A nPatchesAdded+=1
          echo nPatchesAdded: !nPatchesAdded!
          if %ppSAVE_FRAMES%==1 (
            set frNum=00000!nPatchesAdded!
            set frNum=!frNum:~-6!
            copy /y %ppOUTPUT% %ppSAVE_FRAMES_PREFIX%!frNum!.%OUT_EXT%>nul:
          )
        )
      )
    )
  )
  rem EndTrials = (TrialedPrevRow==1) && (TrialledThisRow==0)
  @rem if %TrialedPrevRow%==1 set EndTrials==1 if %TrialedThisRow%==0 set EndTrials==1
  if !TrialedPrevRow!==1 if !TrialedThisRow!==0 set EndTrials=1
  echo TrialedPrevRow = !TrialedPrevRow! TrialedThisRow = !TrialedThisRow! EndTrials = !EndTrials!
  set TrialedPrevRow=!TrialedThisRow!
  )
)
echo on

set /A RadNum-=1
echo RadNum=%RadNum%

set PrevRad=%Rad%
set /A Rad=%Rad%*%ppRAD_MULT%
if %PrevRad%==%Rad% set /A Rad-=1

echo nPatches: %nPatches%  nTrials: %nTrials%  nPatchesAdded: %nPatchesAdded%

if %Rad% GEQ %ppFLOOR_RAD% goto loopRad

echo PaintPatches finished: created %ppOUTPUT%

