#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

/*
  This program is copyright (c) 2023 Alan Gibson ("snibgo").

  Anyone is permitted to use or adapt any of the code for any purpose, including commercial use.

  Build:
    bash snibgo\buildc.sh snibgo\acwise
*/

typedef int BOOL;
#define TRUE 1
#define FALSE 0

typedef struct {
  double x, y;
  double dx, dy;
} CoordT;

/* The directions assume origin is at top-left,
   so "N" is the coord with lowest y coordinate, etc.
*/
typedef enum {
  dN, dNE, dE, dSE, dS, dSW, dW, dNW, dUndef
} DirectionT;

typedef struct {
  DirectionT dirn;
  const char * strDirn;
} strDirectionT;

strDirectionT strDirections[] = {
  {dN,  "N"},
  {dNE, "NE"},
  {dE,  "E"},
  {dSE, "SE"},
  {dS,  "S"},
  {dSW, "SW"},
  {dW,  "W"},
  {dNW, "NW"},
  {dUndef, "undefined"}
};

typedef enum {
  ofCoords, ofPolygon, ofSpokes, ofTriangles, ofCurve, ofWheel, ofUndef
} OutFmtT;

typedef struct {
  OutFmtT outFmt;
  const char * strOutFmt;
} strOutFmtT;

strOutFmtT strOutFmts[] = {
  {ofCoords,    "coords"},
  {ofPolygon,   "polygon"},
  {ofSpokes,     "spokes"},
  {ofTriangles, "triangles"},
  {ofCurve,     "curve"},
  {ofWheel,     "wheel"},
  {ofUndef,     "undefined"}
};

/* TODO: curved polygons.
*/

typedef struct {
  BOOL do_verbose, do_help;
  BOOL do_sort, do_reverse, do_centroid;
  BOOL JoinPaths;
  BOOL SVGformat;
  FILE * fh_data;
  FILE * fh_explain;
  char * infile;
  char * outfile;
  char * explainPrefix;
  OutFmtT OutFmt;
  CoordT Centre;
  DirectionT dirn;
  double TanTight;
  size_t MaxCoords;
  size_t NumCoords;
  CoordT * Coords;
  size_t StartAt;
  char * PathPrefix;
  char * PathSuffix;
} acwiseT;

#define IM_PREFIX "path \'"
#define IM_SUFFIX "\'"
#define SVG_PREFIX "<path d=\""
#define SVG_SUFFIX "\"/>"


static DirectionT DirnOfStr (const char * s)
{
  size_t NumDirns = sizeof (strDirections) / sizeof (strDirectionT);

  size_t i;
  for (i=0; i < NumDirns; i++) {
    if (strcasecmp (s, strDirections[i].strDirn) == 0) break;
  }

  if (i < NumDirns) return strDirections[i].dirn;
  return dUndef;
}

static OutFmtT OutFmtOfStr (const char * s)
{
  size_t NumOutFmts = sizeof (strOutFmts) / sizeof (strOutFmtT);

  size_t i;
  for (i=0; i < NumOutFmts; i++) {
    if (strcasecmp (s, strOutFmts[i].strOutFmt) == 0) break;
  }

  if (i < NumOutFmts) return strOutFmts[i].outFmt;
  return ofUndef;
}

static void usage (void)
{
  size_t NumStrs;
  size_t i;

  printf ("Usage: acwise [OPTION]...'\n");
  printf ("\n");
  printf ("  -i,   --infile string     input text file, or - for stdin\n");
  printf ("  -o,   --outfile string    output text file, or - for stdout\n");
  printf ("  -x,   --explain string    prefix for explanation text file, or - for stdout\n");
  printf ("  -n,   --nosort            don't sort\n");
  printf ("  -r,   --reverse           reverse the order\n");
  printf ("  -c,   --centre X,Y        calculate from given centre (floating-point numbers)\n");
  printf ("  -s,   --startat string    start at direction; one of:");
  NumStrs = sizeof (strDirections) / sizeof (strDirectionT);
  for (i=0; i < NumStrs-1; i++) {
    printf (" %s", strDirections[i].strDirn);
  }
  printf ("\n");

  printf ("  -t,   --tantight number   tightness of tangent for curve and wheel (typically 0.0 to about 0.5)\n");

  printf ("  -fmt, --format string     output format; one of:");
  NumStrs = sizeof (strOutFmts) / sizeof (strOutFmtT);
  for (i=0; i < NumStrs-1; i++) {
    printf (" %s", strOutFmts[i].strOutFmt);
  }
  printf ("\n");

  printf ("  -j,   --joinpaths         join paths into single path\n");
  printf ("  -svg, --svgformat         SVG format (instead of IM \"-draw path\")\n");
  printf ("  -f,   --file string       write verbose text to stderr or stdout\n");
  printf ("  -v,   --verbose           write text information\n");
  printf ("  -h,   --help              write this usage text\n");
  printf ("\n");

  printf ("If the origin is at top-left, so coordinates increase to the right and down,\n");
  printf ("then the sorted order will be anti-clockwise. Use \"--reverse\" to sort clockwise.\n");
  printf ("\n");
  printf ("If the origin is at bottom-left, so coordinates increase to the right and up,\n");
  printf ("then the sorted order will be clockwise. Use \"--reverse\" to sort anti-clockwise.\n");
  printf ("In addition, the meanings of north and south are inverted.\n");
}
\
/* t relative length of tangents (typically 0.0 to 1.0)
*/


/*===
static int LocaleCompare (const char * s1, const char * s2)
{
  return strcasecmp (s1, s2);
}
===*/

static BOOL IsArg (char * pa, const char * ShtOpt, const char * LongOpt)
{
  if ((strcasecmp(pa, ShtOpt)==0) || (strcasecmp(pa, LongOpt)==0))
    return TRUE;

  return FALSE;
}


static BOOL menu(
  const int argc,
  const char **argv,
  acwiseT * pacw
)
/* Returns MagickTrue if okay. */
{
  BOOL okay = TRUE;
  int i;

  pacw->do_verbose = pacw->do_help = FALSE;
  pacw->do_sort = TRUE;
  pacw->do_reverse = FALSE;
  pacw->do_centroid = TRUE;
  pacw->fh_data = stdout;
  pacw->infile = NULL;
  pacw->outfile = NULL;
  pacw->explainPrefix = NULL;
  pacw->fh_explain = NULL;
  pacw->OutFmt = ofCoords;
  pacw->Centre.x = 0;
  pacw->Centre.y = 0;
  pacw->NumCoords = 0;
  pacw->Coords = NULL;
  pacw->dirn = dUndef;
  pacw->TanTight = 0.25;
  pacw->JoinPaths = FALSE;
  pacw->SVGformat = FALSE;
  pacw->PathPrefix = (char *)IM_PREFIX;
  pacw->PathSuffix = (char *)IM_SUFFIX;

  for (i=1; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "-i", "--infile")) {
      i++;
      if (i < argc) pacw->infile = (char *)argv[i];
      else {
        fprintf (stderr, "acwise: --input needs a filename\n");
        okay = FALSE;
      }
    } else if (IsArg (pa, "-o", "--outfile")) {
      i++;
      if (i < argc) pacw->outfile = (char *)argv[i];
      else {
        fprintf (stderr, "acwise: --outfile needs a string\n");
        okay = FALSE;
      }
    } else if (IsArg (pa, "-x", "--explain")) {
      i++;
      if (i < argc) pacw->explainPrefix = (char *)argv[i];
      else {
        fprintf (stderr, "acwise: --explain needs a string\n");
        okay = FALSE;
      }
    } else if (IsArg (pa, "-n", "--nosort")) {
      pacw->do_sort = FALSE;
    } else if (IsArg (pa, "-r", "--reverse")) {
      pacw->do_reverse = TRUE;
    } else if (IsArg (pa, "-s", "--startat")) {
      i++;
      pacw->dirn = DirnOfStr (argv[i]);
      if (pacw->dirn == dUndef) {
        fprintf (stderr, "acwise: --startat is invalid\n");
        okay = FALSE;
      }
    } else if (IsArg (pa, "-c", "--centre")) {
      i++;
      sscanf (argv[i], "%lg , %lg", &pacw->Centre.x, &pacw->Centre.y);
      pacw->do_centroid = FALSE;
    } else if (IsArg (pa, "-t", "--tantight")) {
      i++;
      sscanf (argv[i], "%lg", &pacw->TanTight);
    } else if (IsArg (pa, "-fmt", "--format")) {
      i++;
      pacw->OutFmt = OutFmtOfStr (argv[i]);
      if (pacw->OutFmt == ofUndef) {
        fprintf (stderr, "acwise: --outfmt is invalid\n");
        okay = FALSE;
      }
    } else if (IsArg (pa, "-j", "--joinpaths")) {
      pacw->JoinPaths = TRUE;
    } else if (IsArg (pa, "-svg", "--svgformat")) {
      pacw->SVGformat = TRUE;
      pacw->PathPrefix = (char *)SVG_PREFIX;
      pacw->PathSuffix = (char *)SVG_SUFFIX;

    } else if (IsArg (pa, "-f", "--file")) {
      i++;
      if (strcasecmp (argv[i], "stdout")==0) pacw->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) pacw->fh_data = stderr;
      else okay = FALSE;
    } else if (IsArg (pa, "-h", "--help")) {
      pacw->do_help = TRUE;
    } else if (IsArg (pa, "-v", "--verbose")) {
      pacw->do_verbose = TRUE;
    } else {
      fprintf (stderr, "acwise: ERROR: unknown option XX [%s]\n", pa);
      okay = FALSE;
    }
  }

  if (pacw->do_verbose) {
    fprintf (stderr, "acwise options:");
    if (pacw->infile) fprintf (stderr, "  --infile %s", pacw->infile);
    if (pacw->outfile) fprintf (stderr, "  --outfile %s", pacw->outfile);
    if (pacw->explainPrefix) fprintf (stderr, "  --explain %s", pacw->explainPrefix);
    if (!pacw->do_sort) fprintf (stderr, "  --nosort");
    if (pacw->do_reverse) fprintf (stderr, "  --reverse");
    if (pacw->dirn != dUndef) fprintf (stderr, "  --startat %s", strDirections[pacw->dirn].strDirn);
    if (!pacw->do_centroid) fprintf (stderr, "  --centre %g,%g", pacw->Centre.x, pacw->Centre.y);
    fprintf (stderr, "  --format %s", strOutFmts[pacw->OutFmt].strOutFmt);
    if (pacw->JoinPaths) (stderr, "  --joinpaths");
    if (pacw->SVGformat) (stderr, "  --svgformat");
    if (pacw->OutFmt == ofCurve || pacw->OutFmt == ofWheel) fprintf (stderr, "  --tantight %lg", pacw->TanTight);
    if (pacw->fh_data == stdout) fprintf (stderr, "  --file stdout");
    if (pacw->fh_data == stderr) fprintf (stderr, "  --file stderr");
    if (pacw->do_verbose) fprintf (stderr, "  --verbose");
    if (pacw->do_help) fprintf (stderr, "  --help");
    fprintf (stderr, "\n\n");
  }

  if (!pacw->infile && !pacw->do_help) {
    fprintf (stderr, "Requires --infile\n");
    okay = FALSE;
  }

  if (!okay || pacw->do_help)
    usage ();

  return okay;
}


#define TableExtend 0.1
#define InitMaxCoords 50

static BOOL InitAcw (acwiseT * pacw)
{
  pacw->MaxCoords = InitMaxCoords;
  pacw->NumCoords = 0;
  pacw->Coords = (CoordT *) malloc (pacw->MaxCoords * sizeof(CoordT));
  if (!pacw->Coords) {
    fprintf (stderr, "InitAcw oom");
    return FALSE;
  }
  pacw->StartAt = 0;

  return TRUE;
}

static BOOL DeInitAcw (acwiseT * pacw)
{
  pacw->NumCoords = 0;
  pacw->MaxCoords = 0;
  if (pacw->Coords) {
    free (pacw->Coords);
    pacw->Coords = NULL;
  }

  return TRUE;
}

static BOOL ExtendCoords (acwiseT * pacw)
{
  pacw->MaxCoords = (size_t)(ceil ((double)pacw->MaxCoords * (1.0 + TableExtend)));
  pacw->Coords = (CoordT *) realloc (pacw->Coords, pacw->MaxCoords * sizeof(CoordT));
  if (!pacw->Coords) {
    fprintf (stderr, "ExtendCoords oom");
    return FALSE;
  }

  return TRUE;
}


static BOOL AddCoords (acwiseT * pacw, double x, double y)
{
  if (pacw->NumCoords == pacw->MaxCoords) {
    if (!ExtendCoords (pacw)) return FALSE;
  }

  {
    CoordT * pc = &pacw->Coords[pacw->NumCoords];
    pc->x = x;
    pc->y = y;
    pacw->NumCoords++;
  }

  return TRUE;
}


static BOOL ReadCoords (acwiseT * pacw)
{
  FILE * fh;
  double x, y;
  int n;

  if (strcmp (pacw->infile, "-") == 0) {
    fh = stdin;
  } else {
    fh = fopen (pacw->infile, "rt");
    if (!fh) {
      fprintf (stderr, "Failed to open %s\n", pacw->infile);
      return FALSE;
    }
  }

  n = fscanf (fh, "%lg , %lg", &x, &y);

  while (n != EOF) {
    AddCoords (pacw, x, y);
    n = fscanf (fh, "%lg , %lg", &x, &y);
  }

  if (strcmp (pacw->infile, "-") != 0) {
    fclose (fh);
  }

  if (pacw->do_verbose) {
    fprintf (pacw->fh_data, "NumCoords = %lu\n", pacw->NumCoords);
  }
  return TRUE;
}


static void CalcCentroid (acwiseT * pacw)
{
  size_t i;
  double cx=0, cy=0;
  for (i=0; i < pacw->NumCoords; i++) {
    CoordT * pc = &pacw->Coords[i];
    cx += pc->x;
    cy += pc->y;
  }
  cx /= (double)pacw->NumCoords;
  cy /= (double)pacw->NumCoords;
  if (pacw->do_verbose) {
    fprintf (pacw->fh_data, "Centroid = %g,%g\n", cx, cy);
  }
  pacw->Centre.x = cx;
  pacw->Centre.y = cy;
}

static void CalcDiffs (acwiseT * pacw)
{
  size_t i;
  for (i=0; i < pacw->NumCoords; i++) {
    CoordT * pc = &pacw->Coords[i];
    pc->dx = pc->x - pacw->Centre.x;
    pc->dy = pc->y - pacw->Centre.y;
  }
}

static int CompareCoords (const void *a, const void *b)
{
  CoordT * pcA = (CoordT *) a;
  CoordT * pcB = (CoordT *) b;
  /* Determinant is the signed area of the parallogram of the two vectors: centre to a, and centre to b.
  */
  double det = pcA->dx * pcB->dy - pcB->dx * pcA->dy;
  if (det < 0) return -1;
  if (det > 0) return +1;

  /* They are on the same radius.
  */
  {
    double d1 = pcA->dx * pcA->dx + pcA->dy * pcA->dy;
    double d2 = pcB->dx * pcB->dx + pcB->dy * pcB->dy;
    if (d1 > d2) return -1;
    if (d1 < d2) return +1;
  }
  return 0;
}

static int CompareCoordsReverse (const void *a, const void *b)
{
  return CompareCoords (b, a);
}

static void SortCoords (acwiseT * pacw)
{
  /* When (0,0) is top-left, this gives anticlockwise.
  */
  if (pacw->NumCoords <= 1) return;

  if (pacw->do_reverse) {
    qsort ((void *)pacw->Coords, pacw->NumCoords, sizeof (CoordT), CompareCoordsReverse);
  } else {
    qsort ((void *)pacw->Coords, pacw->NumCoords, sizeof (CoordT), CompareCoords);
  }
}

static void FindStartAt (acwiseT * pacw)
{
  size_t limNdx = 0;
  double val, limVal = 0;
  BOOL better;
  size_t i;

  if (pacw->dirn == dUndef) return;

  for (i=0; i < pacw->NumCoords; i++) {
    const CoordT * pc = &pacw->Coords[i];
    val = pc->y;
    better = val < limVal;
    switch (pacw->dirn) {
      case dN:
        val = pc->y;
        better = val < limVal;
        break;
      case dNE:
        val = pc->x - pc->y;
        better = val > limVal;
        break;
      case dE:
        val = pc->x;
        better = val > limVal;
        break;
      case dSE:
        val = pc->x + pc->y;
        better = val > limVal;
        break;
      case dS:
        val = pc->y;
        better = val > limVal;
        break;
      case dSW:
        val = pc->x - pc->y;
        better = val < limVal;
        break;
      case dW:
        val = pc->x;
        better = val < limVal;
        break;
      case dNW:
        val = pc->x + pc->y;
        better = val < limVal;
        break;
      case dUndef:
        val = 0;
        better = FALSE;
        break;
    }
    if (i==0 || better) {
      limVal = val;
      limNdx = i;
    }
  }
  pacw->StartAt = limNdx;
}

#define NL_EVERY 50

static void WrCoords (const acwiseT * pacw, FILE * fh)
{
  size_t i;
  for (i=0; i < pacw->NumCoords; i++) {
    const CoordT * pc = &pacw->Coords[(i + pacw->StartAt) % pacw->NumCoords];
    fprintf (fh, "%g,%g\n", pc->x, pc->y);
  }
}


static void WrPolyPath (const acwiseT * pacw, FILE * fh)
/* Write data as a polygon path.
*/
{
  size_t i;
  CoordT * pc;

  if (pacw->NumCoords < 2) return;

  pc = &pacw->Coords[pacw->StartAt];
  fprintf (fh, "%s M%g,%g L", pacw->PathPrefix, pc->x, pc->y);

  for (i=1; i < pacw->NumCoords; i++) {
    const CoordT * pc = &pacw->Coords[(i + pacw->StartAt) % pacw->NumCoords];
    fprintf (fh, "%g,%g ", pc->x, pc->y);
    if ((i+1) % NL_EVERY == 0) fprintf (fh, "\n");
  }
  fprintf (fh, "z%s\n", pacw->PathSuffix);
}

static void WrSpokesPath (const acwiseT * pacw, FILE * fh)
/* Write data as a Spokes path.
*/
{
  size_t i;

  if (pacw->NumCoords < 2) return;

  fprintf (fh, "%s", pacw->PathPrefix);

  for (i=0; i < pacw->NumCoords; i++) {
    const CoordT * pc = &pacw->Coords[(i + pacw->StartAt) % pacw->NumCoords];
    fprintf (fh, "M%g,%g L%g,%g\n", pacw->Centre.x, pacw->Centre.y, pc->x, pc->y);
    if ((i+1) % NL_EVERY == 0) fprintf (fh, "\n");
  }
  fprintf (fh, "z%s\n", pacw->PathSuffix);
}

static void WrTrianglesPath (const acwiseT * pacw, FILE * fh)
/* Write data as triangles path.
*/
{
  size_t i;

  if (pacw->NumCoords < 2) return;

  if (pacw->JoinPaths) fprintf (fh, "%s", pacw->PathPrefix);

  for (i=0; i < pacw->NumCoords; i++) {
    const CoordT * pc = &pacw->Coords[(i + pacw->StartAt) % pacw->NumCoords];
    const CoordT * pcNext = &pacw->Coords[(i + pacw->StartAt + 1) % pacw->NumCoords];
    if (!pacw->JoinPaths) fprintf (fh, "%s", pacw->PathPrefix);
    fprintf (fh, "M%g,%g L%g,%g %g,%g z", pacw->Centre.x, pacw->Centre.y, pc->x, pc->y, pcNext->x, pcNext->y);
    if (!pacw->JoinPaths) fprintf (fh, "%s", pacw->PathSuffix);
    fprintf (fh, "\n");
  }
  if (pacw->JoinPaths) fprintf (fh, "%s\n", pacw->PathSuffix);
}

static void WrCurvePath (const acwiseT * pacw, FILE * fh)
/* Write data as one closed curved path. (Cubic Bezier.)
*/
{
  size_t i;
  CoordT * pc;

  if (pacw->NumCoords < 3) return;

  pc = &pacw->Coords[pacw->StartAt];
  fprintf (fh, "%s M%g,%g\n", pacw->PathPrefix, pc->x, pc->y);

  /* This sets the tangent at each point parallel to the line between the adjacent points, 
     and each semi-tangent is a fraction of the length of that line.

     An alternative would be to make the tangent perpendicular to the line to the centre,
     and the length could be proprtional to that.

     In each iteration, create the curve between nM1 and n0.

     Possible enhancement: limit tangent lengths so they don't cross.
  */

  for (i=1; i <= pacw->NumCoords; i++) {
    size_t nM2 = (i + pacw->StartAt - 2 + pacw->NumCoords) % pacw->NumCoords;
    size_t nM1 = (nM2 +1) % pacw->NumCoords;
    size_t n0  = (nM1 +1) % pacw->NumCoords;
    size_t nP1 = (n0  +1) % pacw->NumCoords;
    const CoordT * pcM2 = &pacw->Coords[nM2];
    const CoordT * pcM1 = &pacw->Coords[nM1];
    const CoordT * pc = &pacw->Coords[n0];
    const CoordT * pcP1 = &pacw->Coords[nP1];
    const double dx0 = (pcM2->x - pc->x)   * pacw->TanTight;
    const double dy0 = (pcM2->y - pc->y)   * pacw->TanTight;
    const double dx1 = (pcP1->x - pcM1->x) * pacw->TanTight;
    const double dy1 = (pcP1->y - pcM1->y) * pacw->TanTight;
    CoordT cp1, cp2;
    cp1.x = pcM1->x - dx0;
    cp1.y = pcM1->y - dy0;
    cp2.x = pc->x - dx1;
    cp2.y = pc->y - dy1;
    fprintf (fh, "C%g,%g %g,%g %g,%g\n", cp1.x, cp1.y, cp2.x, cp2.y, pc->x, pc->y);
    if (pacw->fh_explain) {
      fprintf (pacw->fh_explain, "-draw \"path 'M%g,%g L%g,%g'\"\n", pcM1->x, pcM1->y, cp1.x, cp1.y);
      fprintf (pacw->fh_explain, "-draw \"path 'M%g,%g L%g,%g'\"\n", pc->x, pc->y, cp2.x, cp2.y);
    }
  }
  fprintf (fh, "z%s\n", pacw->PathSuffix);
}

static void WrWheelPath (const acwiseT * pacw, FILE * fh)
/* A curve with spokes. Each set of two spokes and adjoining curve is a closed path.
*/
{
  size_t i;

  if (pacw->NumCoords < 3) return;

  if (pacw->JoinPaths) fprintf (fh, "%s", pacw->PathPrefix);

  /* Possible enhancement: limit tangent lengths so they don't cross.
  */

  for (i=1; i <= pacw->NumCoords; i++) {
    size_t nM2 = (i + pacw->StartAt - 2 + pacw->NumCoords) % pacw->NumCoords;
    size_t nM1 = (nM2 +1) % pacw->NumCoords;
    size_t n0  = (nM1 +1) % pacw->NumCoords;
    size_t nP1 = (n0  +1) % pacw->NumCoords;
    const CoordT * pcM2 = &pacw->Coords[nM2];
    const CoordT * pcM1 = &pacw->Coords[nM1];
    const CoordT * pc = &pacw->Coords[n0];
    const CoordT * pcP1 = &pacw->Coords[nP1];
    const double dx0 = (pcM2->x - pc->x)   * pacw->TanTight;
    const double dy0 = (pcM2->y - pc->y)   * pacw->TanTight;
    const double dx1 = (pcP1->x - pcM1->x) * pacw->TanTight;
    const double dy1 = (pcP1->y - pcM1->y) * pacw->TanTight;
    CoordT cp1, cp2;
    cp1.x = pcM1->x - dx0;
    cp1.y = pcM1->y - dy0;
    cp2.x = pc->x - dx1;
    cp2.y = pc->y - dy1;
    if (!pacw->JoinPaths) fprintf (fh, "%s", pacw->PathPrefix);
    fprintf (fh, "M%g,%g L%g,%g C%g,%g %g,%g %g,%g L%g,%g z",
      pacw->Centre.x, pacw->Centre.y,
      pcM1->x, pcM1->y,
      cp1.x, cp1.y, cp2.x, cp2.y, pc->x, pc->y,
      pacw->Centre.x, pacw->Centre.y);
    if (!pacw->JoinPaths) fprintf (fh, "%s", pacw->PathSuffix);
    fprintf (fh, "\n");
    if (pacw->fh_explain) {
      fprintf (pacw->fh_explain, "-draw \"path 'M%g,%g L%g,%g'\"\n", pcM1->x, pcM1->y, cp1.x, cp1.y);
      fprintf (pacw->fh_explain, "-draw \"path 'M%g,%g L%g,%g'\"\n", pc->x, pc->y, cp2.x, cp2.y);
    }
  }
  if (pacw->JoinPaths) fprintf (fh, "%s\n", pacw->PathSuffix);
}

static void WrNumberPoints (const acwiseT * pacw, FILE * fh)
{
  size_t i;
  for (i=0; i < pacw->NumCoords; i++) {
    size_t thisNum = (i + pacw->StartAt) % pacw->NumCoords;
    CoordT * pc = &pacw->Coords[thisNum];
    /* FIXME: shld be out a bit from the centre */
    fprintf (fh, "-annotate %+g%+g %lu\n", pc->x, pc->y, i);
  }
  fprintf (fh, "-annotate %+g%+g C\n", pacw->Centre.x, pacw->Centre.y);
}

int main (const int argc, const char *argv[])
{
  acwiseT acw;

  if (!menu (argc, argv, &acw)) return 1;
  if (!acw.infile) return 0;

  if (!InitAcw (&acw)) return 1;

  ReadCoords (&acw);
  if (acw.do_centroid) CalcCentroid (&acw);
  CalcDiffs (&acw);
  if (acw.do_sort) SortCoords (&acw);
  FindStartAt (&acw);

#define explainSuffix "_explain.lis"
 
  if (acw.explainPrefix) {
    if (strcmp (acw.explainPrefix, "-") == 0) {
      acw.fh_explain = stdout;
    } else {
      char * sExplain;
      size_t len1 = strlen (acw.explainPrefix);
      size_t len2 = strlen (explainSuffix);
      sExplain = malloc ((len1+len2+1) * sizeof(char));
      if (!sExplain) return -1;
      strcpy (sExplain, acw.explainPrefix);
      strcpy (sExplain+len1, explainSuffix);
      acw.fh_explain = fopen (sExplain, "wt");
      if (!acw.fh_explain) {
        fprintf (stderr, "Failed to open %s\n", sExplain);
        return -1;
      }
      free (sExplain);
    }
    WrNumberPoints (&acw, acw.fh_explain);
  }

  if (acw.outfile) {
    FILE * fh;
    if (strcmp (acw.outfile, "-") == 0) {
      fh = stdout;
    } else {
      fh = fopen (acw.outfile, "wt");
      if (!fh) {
        fprintf (stderr, "Failed to open %s\n", acw.outfile);
        return -1;
      }
    }
    switch (acw.OutFmt) {
      case ofCoords:    WrCoords (&acw, fh); break;
      case ofPolygon:   WrPolyPath (&acw, fh); break;
      case ofSpokes:    WrSpokesPath (&acw, fh); break;
      case ofTriangles: WrTrianglesPath (&acw, fh); break;
      case ofCurve:     WrCurvePath (&acw, fh); break;
      case ofWheel:     WrWheelPath (&acw, fh); break;
      default:          /* No output */ break;
    }

    if (strcmp (acw.outfile, "-") != 0) {
      fclose (fh);
    }
  }

  if (acw.explainPrefix) {
    fprintf (acw.fh_explain, "-write %s_explain.png\n", acw.explainPrefix);
    fprintf (acw.fh_explain, "NULL:\n");
    if (strcmp (acw.explainPrefix, "-") != 0) {
      fclose (acw.fh_explain);
    }
  }

  if (!DeInitAcw (&acw)) return 1;

  return 0;
}

