/* Updated:
     6-April-2018 Corrected bug in verbose.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"

#include "aggrsrch.inc"

#define VERSION "aggrsrch v1.0  Copyright (c) 2017 Alan Gibson"

typedef struct {
  MagickBooleanType
    verbose,
    makeMask;

  double
    aggression,     // 0.0 < aggression <= 1.0.
    maskThreshold;  // typically 0.0 to 1.0.

} SubSrchT;


static void InitSubSrch (SubSrchT * pss)
{
  pss->makeMask = MagickFalse;
  pss->verbose = MagickFalse;
  pss->maskThreshold = 0.75;
  pss->aggression = 1.0;
}

static void DeInitSubSrch (SubSrchT * pss)
{
  ; // Nothing
}


static void usage (void)
{
  printf ("Usage: -process 'aggrsrch [OPTION]...'\n");
  printf ("Search with aggressive resizing.\n");
  printf ("\n");
  printf ("  a,  aggression number  0.0 to 1.0\n");
  printf ("  m,  makeMask           make mask from result\n");
  printf ("  t,  threshold number   0.0 to 1.0\n");
  printf ("  v,  verbose            write text information to stdout\n");
  printf ("      version            write version information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "aggrsrch: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  SubSrchT * pss
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "m", "makeMask")==MagickTrue) {
      pss->makeMask = MagickTrue;
    } else if (IsArg (pa, "a", "aggression")==MagickTrue) {
      NEXTARG;
      pss->aggression = atof(argv[i]);
    } else if (IsArg (pa, "t", "threshold")==MagickTrue) {
      NEXTARG;
      pss->maskThreshold = atof(argv[i]);
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pss->verbose = MagickTrue;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "aggrsrch: ERROR: unknown option\n");
      status = MagickFalse;
    }
  }

  if (pss->verbose) {
    fprintf (stderr, "aggrsrch options:");
    fprintf (stderr, "  aggression %g", pss->aggression);

    if (pss->makeMask) {
      fprintf (stderr, "  makeMask");
      fprintf (stderr, "  threshold %g", pss->maskThreshold);
    }
    if (pss->verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


// The next function takes an image list,
// may modify it, and returns a status.
//
static MagickBooleanType aggrsrch (
  Image **images,
  SubSrchT * pss,
  ExceptionInfo *exception)
{
  Image * main_img = *images;

  Image * sub_img = GetNextImageInList (main_img);
  if (!sub_img) return MagickFalse;


  if (sub_img->columns > main_img->columns ||
      sub_img->rows    > main_img->rows)
  {
    fprintf (stderr, "aggrsrch: subimage is larger than main\n");
    return MagickFalse;
  }

  Image * r = aggrSrchDiff (
    main_img, sub_img, pss->makeMask,
    pss->aggression, pss->maskThreshold, pss->verbose, exception);
  if (!r) return MagickFalse;

  // FIXME: This makes a mask, which is useful.
  // But how about finding the search coordinates?
  // For each of the sinks, crop the appropriate rectangle from main,
  // and search it for sub.

  DeleteImageFromList (&sub_img);
  DeleteImageFromList (&main_img);

  *images = GetFirstImageInList (r);

  return MagickTrue;
}



ModuleExport size_t aggrsrchImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  MagickBooleanType
    status;

  SubSrchT
    ss;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  InitSubSrch (&ss);

  status = menu (argc, argv, &ss);
  if (status == MagickFalse)
    return (-1);

  int ListLen = (int)GetImageListLength(*images);
  if (ListLen != 2) {
    fprintf (stderr, "aggrsrch: needs 2 images\n");
    return (-1);
  }

  if (!aggrsrch (images, &ss, exception)) return (-1);

  DeInitSubSrch (&ss);

  return(MagickImageFilterSignature);
}
