#ifndef AGGRSRCH_INC
#define AGGRSRCH_INC

#include "resetpage.inc"
#include "findsinks.inc"
#include "cropchk.inc"


static MagickBooleanType SetImgRMS (
  Image *image,
  MagickRealType *vLo,
  ssize_t *fndX,
  ssize_t *fndY,
  ExceptionInfo *exception
)
// Sets R,G and B channels to the RMS of R, G and B.
// Sets alpha opaque.
// Return the darkest coordinate.
// Returns whether okay.
{
  MagickBooleanType okay = MagickTrue;
  ssize_t y;

  CacheView * in_view = AcquireAuthenticCacheView (image, exception);

  *vLo = QuantumRange;
  ssize_t xLo = 0, yLo = 0;

  for (y = 0; y < image->rows; y++) {
    if (!okay) continue;

    VIEW_PIX_PTR *p = GetCacheViewAuthenticPixels(
      in_view,0,y,image->columns,1,exception);
    if (!p) {okay = MagickFalse; continue; }

    ssize_t x;
    MagickRealType r, g, b, v;
    for (x = 0; x < image->columns; x++) {

      r = GET_PIXEL_RED   (image, p);
      g = GET_PIXEL_GREEN (image, p);
      b = GET_PIXEL_BLUE  (image, p);

      v = sqrt ((r*r + g*g + b*b) / 3.0);

      if (*vLo > v) {
        *vLo = v;
        xLo = x;
        yLo = y;
      }

      SET_PIXEL_RED   (image, v, p);
      SET_PIXEL_GREEN (image, v, p);
      SET_PIXEL_BLUE  (image, v, p);
      SET_PIXEL_ALPHA (image, QuantumRange, p);

      p += Inc_ViewPixPtr (image);
    }
    if (SyncCacheViewAuthenticPixels(in_view,exception) == MagickFalse)
      okay = MagickFalse;
  }
  in_view  = DestroyCacheView (in_view);

  *vLo /= QuantumRange;
  *fndX = xLo;
  *fndY = yLo;

  return okay;
}


static Image * aggrSrchDiff (
  Image * main_img,
  Image * sub_img,
  MagickBooleanType MakeMask,
  double aggression,     // 0.0 < aggression <= 1.0.
  double maskThreshold,  // 0.0 to 1.0. >= 1.0 for no threshold.
  MagickBooleanType verbose,
  ExceptionInfo *exception)
{
  /* This should account for an expected minimum size.
     Eg if the sub_img appears in main_img at half the size,
     then we can't be so aggressive.
  */

  if (aggression <= 0) {
    fprintf (stderr, "aggrSrch Diff: aggression=%g\n", aggression);
    return NULL;
  }

  ssize_t mainW = main_img->columns;
  ssize_t mainH = main_img->rows;
  ssize_t newW = floor (mainW / (aggression * sub_img->columns ) + 0.5);
  ssize_t newH = floor (mainH / (aggression * sub_img->rows    ) + 0.5);

  if (verbose) fprintf (stderr,
    "aggrSrch Diff: main %lix%li  sub %lix%li  newMain %lix%li threshold %g\n",
    mainW, mainH,
    sub_img->columns, sub_img->rows,
    newW, newH,
    maskThreshold);

  if (newW < 1 || newH < 1) {
    fprintf (stderr, "aggrSrchDiff: sub_img too small\n");
  }

  Image * main_res = ScaleImage (main_img, newW, newH, exception);
  if (!main_res) return NULL;

  Image * sub_res1 = ScaleImage (sub_img, 1, 1, exception);
  if (!sub_res1) return NULL;

#if IMV6OR7==6
  SetImageAlphaChannel (sub_res1, DeactivateAlphaChannel);
#else
  SetImageAlphaChannel (sub_res1, DeactivateAlphaChannel, exception);
#endif

  Image * sub_res2 = ScaleImage (sub_res1, newW, newH, exception);
  if (!sub_res2) return NULL;
  sub_res1 = DestroyImage (sub_res1);

  if (!COMPOSITE(main_res, DifferenceCompositeOp, sub_res2, 0,0, exception))
    return NULL;
  sub_res2 = DestroyImage (sub_res2);

//#if IMV6OR7==6
//  if (!GrayscaleImage (main_res, RMSPixelIntensityMethod)) return NULL;
//
//  TransformImageColorspace (main_res, sRGBColorspace);
//
//  SetImageAlphaChannel (main_res, DeactivateAlphaChannel);
//#else
//  if (!GrayscaleImage (main_res, RMSPixelIntensityMethod, exception)) return NULL;
//
//  TransformImageColorspace (main_res, sRGBColorspace, exception);
//
//  SetImageAlphaChannel (main_res, DeactivateAlphaChannel, exception);
//#endif

  MagickRealType vLo;
  ssize_t fndX, fndY;
  if (! SetImgRMS (main_res, &vLo, &fndX, &fndY, exception)) return NULL;

  if (verbose) {
    fprintf (stderr,
      "aggrsrch Res: %g @ %li,%li\n",
      vLo, (long int)fndX, (long int)fndY);
  }

  if (MakeMask) {
    Image * main_lns = LightenNonSinks (main_res, verbose, exception);
    if (!main_lns) return NULL;
    DestroyImage (main_res);
    main_res = main_lns;
  }

  Image * main_res2 = RESIZEIMG (main_res,
    mainW, mainH, exception);

  if (!main_res2) return NULL;
  DestroyImage (main_res);
  main_res = main_res2;

  if (MakeMask) {

    // Crop to size for the mask, offset by half subimage dimensions.

    RectangleInfo fndRect;
    fndRect.width  = mainW - sub_img->columns + 1;
    fndRect.height = mainH - sub_img->rows + 1;

    if (fndRect.width < 1 || fndRect.height < 1) {
      fprintf (stderr, "aggrSrch Diff: zero or neg mask\n");
      return NULL;
    }

    fndRect.x = sub_img->columns / 2;
    fndRect.y = sub_img->rows / 2;

    if (verbose) fprintf (stderr, "aggrSrch Mask: crop to %lix%li+%li+%li\n",
      fndRect.width, fndRect.height, fndRect.x, fndRect.y);

    Image * main_res3 = CropCheck (main_res, &fndRect, exception);
    if (!main_res3) return NULL;
    ResetPage (main_res3);
    DestroyImage (main_res);
    main_res = main_res3;

    // Set red and green channels to zero.
    // Where blue is more than a threshold, set it to 100%.

    double limitBlue = maskThreshold * QuantumRange;

    int nRemoved = 0;
    CacheView * img_view = AcquireAuthenticCacheView (main_res, exception);
    MagickBooleanType okay = MagickTrue;
    ssize_t y;
    for (y = 0; y < main_res->rows; y++) {
      if (!okay) continue;

      VIEW_PIX_PTR *q = GetCacheViewAuthenticPixels(
        img_view,0,y,main_res->columns,1,exception);
      if (!q) {okay = MagickFalse; continue; }

      ssize_t x;
      for (x = 0; x < main_res->columns; x++) {
        SET_PIXEL_RED (main_res, 0, q);
        SET_PIXEL_GREEN (main_res, 0, q);
        double b = GET_PIXEL_BLUE (main_res, q);
        if (b > limitBlue) {
          SET_PIXEL_BLUE (main_res, QuantumRange, q);
          nRemoved++;
        }
        q += Inc_ViewPixPtr (main_res);
      }
      if (SyncCacheViewAuthenticPixels(img_view,exception) == MagickFalse)
        okay = MagickFalse;
    }
    if (!okay) return NULL;

    if (verbose) fprintf (stderr, "aggrSrch Mask: removed %g%%\n",
      nRemoved * 100.0 / (double)(main_res->columns * main_res->rows));

    img_view = DestroyCacheView (img_view);
  }

  return main_res;
}

#endif
