/* Updated:
     7-September-2017 for v7.
     3-April-2018 for v7.0.7-28
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "vsn_defines.h"


typedef struct {
  double
    scale,
    bias;
  MagickBooleanType
    do_verbose;
} arctanT;


static void usage (void)
{
  printf ("Usage: -process 'arctan2 [OPTION]...'\n");
  printf ("From two equal-size images, calculates the arctangent.\n");
  printf ("\n");
  printf ("  s, scale N          scale [1/2pi]\n");
  printf ("  b, bias N           bias [0.5]\n");
  printf ("  v, verbose          write text information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "baryc: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  arctanT * pat
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  pat->do_verbose = MagickFalse;

  pat->scale = 1 / (2.0 * M_PI);
  pat->bias = 0.5;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "s", "scale")==MagickTrue) {
      NEXTARG;
      pat->scale = atof(argv[i]);
    } else if (IsArg (pa, "b", "bias")==MagickTrue) {
      NEXTARG;
      pat->bias = atof(argv[i]);
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pat->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "arctan2: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (pat->do_verbose) {
    fprintf (stderr, "arctan2 options:");

    if (pat->do_verbose) fprintf (stderr, "  verbose");

    // FIXME: thoughout, %g should respect "-precision"
    fprintf (stderr, "  scale %g", pat->scale);
    fprintf (stderr, "  bias %g", pat->bias);

    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *arctan2(
  Image *image1,
  Image *image2,
  arctanT * pat,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  CacheView
    *in_view1,
    *in_view2,
    *out_view;

  ssize_t
    y;

  MagickBooleanType
    status;


  assert(image1 != (Image *) NULL);
  assert(image1->signature == MAGICK_CORE_SIG);

  if (image1->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image1->filename);

  if (image1->columns != image2->columns || image1->rows != image2->rows ) {
    fprintf (stderr, "arctan2: needs same-sized images");
  }

  if (SetNoPalette (image1, exception) == MagickFalse)
    return (Image *)NULL;

  if (SetNoPalette (image2, exception) == MagickFalse)
    return (Image *)NULL;

  // Make a clone of this image, same size but undefined pixel values:
  //
  new_image=CloneImage(image1, image1->columns, image1->rows, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  in_view1 = AcquireVirtualCacheView (image1,exception);
  in_view2 = AcquireVirtualCacheView (image2,exception);

  out_view = AcquireAuthenticCacheView (new_image,exception);

  status = MagickTrue;

#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) shared(status) \
    MAGICK_THREADS(new_image,new_image,new_image->rows,1)
#endif
  for (y=0; y < (ssize_t) new_image->rows; y++)
  {
    register VIEW_PIX_PTR
      *p1, *p2, *q;

    register ssize_t
      x;

    if (status == MagickFalse)
      continue;

    p1=GetCacheViewAuthenticPixels(in_view1,0,y,new_image->columns,1,exception);
    if (p1 == (const VIEW_PIX_PTR *) NULL)
    {
      status=MagickFalse;
      continue;
    }

    p2=GetCacheViewAuthenticPixels(in_view2,0,y,new_image->columns,1,exception);
    if (p2 == (const VIEW_PIX_PTR *) NULL)
    {
      status=MagickFalse;
      continue;
    }

    q=GetCacheViewAuthenticPixels(out_view,0,y,new_image->columns,1,exception);
    if (q == (const VIEW_PIX_PTR *) NULL)
    {
      status=MagickFalse;
      continue;
    }

    for (x=0; x < (ssize_t) new_image->columns; x++)
    {
      SET_PIXEL_RED   (new_image,
                       QuantumRange *
                         (atan2 (GET_PIXEL_RED(image1,p1),
                                 GET_PIXEL_RED(image2,p2)
                                ) * pat->scale + pat->bias),
                       q);

      SET_PIXEL_GREEN (new_image,
                       QuantumRange *
                         (atan2 (GET_PIXEL_GREEN(image1,p1),
                                 GET_PIXEL_GREEN(image2,p2)
                                ) * pat->scale + pat->bias),
                       q);

      SET_PIXEL_BLUE  (new_image,
                       QuantumRange *
                         (atan2 (GET_PIXEL_BLUE(image1,p1),
                                 GET_PIXEL_BLUE(image2,p2)
                                ) * pat->scale + pat->bias),
                       q);

      p1 += Inc_ViewPixPtr (image1);
      p2 += Inc_ViewPixPtr (image2);
      q  += Inc_ViewPixPtr (new_image);
    }
    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse)
      status=MagickFalse;
  }
  out_view = DestroyCacheView (out_view);
  in_view2 = DestroyCacheView (in_view2);
  in_view1 = DestroyCacheView (in_view1);

  if (status == MagickFalse) return NULL;

  return (new_image);
}



ModuleExport size_t arctan2Image(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image1,
    *image2,
    *new_image;

  arctanT
    at;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  // Replace each pair of images with atan2(u,v).
  // The images must be the same size.

  int ListLen = (int)GetImageListLength(*images);
  if (ListLen !=2) {
    fprintf (stderr, "arctan2 needs 2 images\n");
    return (-1);
  }

  MagickBooleanType status = menu (argc, argv, &at);
  if (status == MagickFalse)
    return (-1);

  image1 = (*images);
  image2 = GetNextImageInList (image1);

  new_image = arctan2 (image1, image2, &at, exception);
  if (new_image == (Image *) NULL)
    return(-1);

  DeleteImageFromList (&image2);

  ReplaceImageInList (&image1, new_image);
  // Replace messes up the images pointer. Make it good:
  *images = GetFirstImageInList (image1);

  return (MagickImageFilterSignature);
}
