/* Updated:
     3-April-2018 for v7.0.7-28
     5-April-2018 improved error-reporting;
       explicit channel copy from scal_img to new_img.
*/

#ifndef AVGCONCRINGS_INC
#define AVGCONCRINGS_INC

#include "rmsealpha.inc"
#include "resetpage.inc"
#include "trimone.inc"

typedef enum {
  crmSlow,
  crmQuick
} ConcRingsMethodT;

typedef struct {
  int
    verbose;

  double
    minScale,
    maxScale;

  int
    nScales;

  ConcRingsMethodT
    ConcRingsMethod;

  // Used internally:
  RmseAlphaT
    ra;

  // Calculated:
  double
    scaleFact;  // A measure of error.

  double
    nScalesM1;
} ConcRingsT;

static void InitConcRings (ConcRingsT * p)
{
  p->verbose = 0;
  p->minScale = 0.5;
  p->maxScale = 2.0;
  p->nScales = 20;
  p->ConcRingsMethod = crmQuick;
  p->scaleFact = 0;

  p->nScalesM1 = (double)(p->nScales-1);

  InitRmseAlpha (&p->ra);
}

static void ReInitConcRings (ConcRingsT * p)
{
  ReInitRmseAlpha (&p->ra);
}

static void DeInitConcRings (ConcRingsT * p)
{
  DeInitRmseAlpha (&p->ra);
}


static double inline ConcRingsScale (
  ConcRingsT * pcr,
  ssize_t y
)
// Returns the scale at a given y-coordinate.
{
  if (pcr->nScalesM1==0) return pcr->maxScale;

  return pcr->minScale *
         pow (pcr->maxScale/pcr->minScale, y / pcr->nScalesM1);
}


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *avgconcrings (
  Image *image,
  ConcRingsT * pcr,
  ExceptionInfo *exception)
{
  Image * cln = CloneImage (image, 0, 0, MagickTrue, exception);
  if (!cln) return NULL;

//#if IMV6OR7==6
//  SetPixelCacheVirtualMethod (cln, TransparentVirtualPixelMethod);
//  SetImageAlphaChannel (cln, ActivateAlphaChannel);
//#else
//  SetPixelCacheVirtualMethod (cln,
//    TransparentVirtualPixelMethod, exception);
//  SetImageAlphaChannel (cln, ActivateAlphaChannel, exception);
//#endif

  VIRT_NONE(cln, exception);

  MagickBooleanType bestfit = MagickTrue;
  double minusOne=-1.0;

  Image *tmp_img = DistortImage (cln, DePolarDistortion,
    1, &minusOne, bestfit, exception);
  if (!tmp_img) return NULL;

  if (pcr->verbose) fprintf (stderr,
    "avgconcrings: tmp_img %lix%li\n",
    tmp_img->columns, tmp_img->rows);

  ssize_t unscaledHeight = tmp_img->rows;
  tmp_img = DestroyImage (tmp_img);

  if (pcr->verbose) fprintf (stderr,
    "avgconcrings: %lix%li unscaled %li\n",
    cln->columns, cln->rows, unscaledHeight);


  pcr->nScalesM1 = (double)(pcr->nScales-1);

  if (pcr->nScalesM1 == 0) {
    pcr->scaleFact = 0;
  } else {
    pcr->scaleFact = pow (pcr->maxScale/pcr->minScale, 1.0 / pcr->nScalesM1);
  }

  // Make an image for output: unscaledHeight x nScales, transparent black.

  Image * new_img = CloneImage(cln, unscaledHeight, pcr->nScales,
    MagickTrue, exception);
  if (!new_img) return NULL;

  SetAllBlack (new_img, exception);
#if IMV6OR7==6
  SetImageAlphaChannel (new_img, TransparentAlphaChannel);
#else
  SetImageAlphaChannel (new_img, TransparentAlphaChannel, exception);
#endif

  double scaleArray[2];
  scaleArray[0] = 1.0;  // Scale factor.
  scaleArray[1] = 0.0;  // Rotation angle.

  Image *depol_max2_img = NULL;

  if (pcr->ConcRingsMethod == crmQuick) {

    // For the quick method:
    // Make a single-column image at the maximum scale,
    // and resize it down for the others.

    scaleArray[0] = pcr->maxScale;

    Image *max_img = DistortImage (cln, ScaleRotateTranslateDistortion,
      2, scaleArray, bestfit, exception);
    if (!max_img) {
      fprintf (stderr, "avgconcrings: DistortImage SRT max_img failed\n");
      return NULL;
    }
    ResetPage (max_img);
    if (!TrimOne (&max_img, exception)) return NULL;

    if (pcr->verbose) fprintf (stderr,
      "avgconcrings: max_img %lix%li\n",
      max_img->columns, max_img->rows);

    Image *depol_max_img = DistortImage (max_img, DePolarDistortion,
      1, &minusOne, bestfit, exception);
    if (!depol_max_img) return NULL;

    DestroyImage (max_img);


    // Scale to 1 column.

    depol_max2_img = ScaleImage (depol_max_img, 1, depol_max_img->rows, exception);
    if (!depol_max2_img) return NULL;

    if (pcr->verbose) fprintf (stderr,
      "avgconcrings: depol_max2_img %lix%li\n",
      depol_max2_img->columns, depol_max2_img->rows);

    DestroyImage (depol_max_img);
  }

  MagickBooleanType okay = MagickTrue;
  int i;
  for (i=0; i < pcr->nScales; i++) {
    if (!okay) continue;

    scaleArray[0] = ConcRingsScale (pcr, i);

    if (pcr->verbose > 1) fprintf (stderr,
      "avgconcrings: %i %g\n", i, scaleArray[0]);

    Image *scal_img = NULL;

    if (pcr->ConcRingsMethod == crmSlow) {
      Image *tmp_img = DistortImage (cln, ScaleRotateTranslateDistortion,
        2, scaleArray, bestfit, exception);
      if (!tmp_img) {okay = MagickFalse; continue;}
      ResetPage (tmp_img);
      if (!TrimOne (&tmp_img, exception))
        {okay = MagickFalse; continue;};

      if (pcr->verbose > 1) fprintf (stderr,
        "avgconcrings: tmp_img %lix%li\n",
        tmp_img->columns, tmp_img->rows);

      Image *depol_img = DistortImage (tmp_img, DePolarDistortion,
        1, &minusOne, bestfit, exception);
      if (!depol_img) {okay = MagickFalse; continue;}

      tmp_img = DestroyImage (tmp_img);

      if (pcr->verbose > 1) fprintf (stderr,
        "avgconcrings: depol_img %lix%li\n",
        depol_img->columns, depol_img->rows);

      scal_img = ScaleImage (depol_img, 1, depol_img->rows, exception);
      if (!scal_img) {okay = MagickFalse; continue;}

      depol_img = DestroyImage (depol_img);

    } else if (pcr->ConcRingsMethod == crmQuick) {
      // Alternative: make scal_img by scaling depol_max2_img down.
      scaleArray[0] /= pcr->maxScale;

#if IMV6OR7==6
      SetImageVirtualPixelMethod (depol_max2_img, EdgeVirtualPixelMethod);\
      //SetPixelCacheVirtualMethod (depol_max2_img, EdgeVirtualPixelMethod);
#else
      SetImageVirtualPixelMethod (depol_max2_img, EdgeVirtualPixelMethod, exception);\

      //SetPixelCacheVirtualMethod (depol_max2_img,
      //  EdgeVirtualPixelMethod, exception);
#endif

      scal_img = DistortImage (depol_max2_img, ScaleRotateTranslateDistortion,
        2, scaleArray, bestfit, exception);
      if (!scal_img) {
        fprintf (stderr, "avgconcrings DistortImage scal_img failed\n");
        okay = MagickFalse;
        continue;
      }
      ResetPage (scal_img);
      if (!TrimOne (&scal_img, exception)) {
        fprintf (stderr, "avgconcrings: TrimOne scal_img failed\n");
        okay = MagickFalse;
        continue;
      }
    }

    if (pcr->verbose > 1) fprintf (stderr,
      "avgconcrings: scal_img %lix%li\n",
      scal_img->columns, scal_img->rows);

    // Copy pixels from scal_img to new_img.

    CacheView * scal_view = AcquireVirtualCacheView (scal_img, exception);
    CacheView * new_view = AcquireAuthenticCacheView (new_img, exception);

    const VIEW_PIX_PTR *sp = GetCacheViewVirtualPixels (
      scal_view,0,0,1,scal_img->rows,exception);
    if (!sp) {okay = MagickFalse; continue;}

    VIEW_PIX_PTR *np = GetCacheViewAuthenticPixels (
      new_view,0,i,new_img->columns,1,exception);
    if (!np) {okay = MagickFalse; continue;}

    ssize_t maxN = new_img->columns;
    if (maxN > scal_img->rows) maxN = scal_img->rows;
    ssize_t n;
    for (n=0; n < maxN; n++) {
      //*np = *sp;
      SET_PIXEL_RED (new_img, GET_PIXEL_RED(scal_img, sp), np);
      SET_PIXEL_GREEN (new_img, GET_PIXEL_GREEN(scal_img, sp), np);
      SET_PIXEL_BLUE (new_img, GET_PIXEL_BLUE(scal_img, sp), np);
      SET_PIXEL_ALPHA (new_img, GET_PIXEL_ALPHA(scal_img, sp), np);
      sp += Inc_ViewPixPtr (scal_img);
      np += Inc_ViewPixPtr (new_img);
    }

    new_view  = DestroyCacheView (new_view);
    scal_view = DestroyCacheView (scal_view);

    scal_img = DestroyImage (scal_img);
  }

  if (!okay) {
    fprintf (stderr, "avgconcrings not okay\n");
    return NULL;
  }

  if (depol_max2_img) DestroyImage (depol_max2_img);

  cln = DestroyImage (cln);

  return new_img;
}


#endif // not AVGCONCRINGS_INC