/* Updated:
     14-December-2019 Moved 3x3 matrix functions to mat3x3.inc
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "vsn_defines.h"

#include "mat3x3.inc"

#define VERSION "barymap v1.0  Copyright (c) 2019 Alan Gibson"

typedef struct {
  double x;
  double y;
} coordsT;

typedef struct {
  double x;
  double y;
  double z;
} xyzCoordsT;

typedef struct {
  double b1;
  double b2;
  double b3;
} baryCoordsT;

typedef struct {
  coordsT primA;  // usually for red
  coordsT primB;  // usually for green
  coordsT primC;  // usually for blue

  // Calculated:
  double
    y23, x13, x32, y13, y31, x3, y3;
  double det;
} triangleT;

typedef enum {
  chRGB,
  chXYY,
  chXYZ
} channelsT;

typedef struct {
  char * name;
  double
    power,
    slope,
    limit,
    offset;

  double
    invPower,
    limit2,
    offsetP1;
} transT;

typedef struct {
  channelsT chans;
  triangleT triMain;
  coordsT   WP;     // white point
  double    gamma;  // -1 etc means sRGB etc.
  coordsT   XYZwp;  // white point for XYZ

  MagickBooleanType
    menuSetChan, menuSetPrim, menuSetWP, menuSetGam, menuSetXYZwp;

  // Calculated:
  //
  triangleT
    subTris[4];

  mat3x3T
    RGB2XYZ,
    XYZ2RGB,
    chromAdapTrans,
    RGB2XYZchr,
    XYZ2RGBchr;

  transT
    trans;
} colSpT;

typedef struct {
  double
    direction,
    width,
    factor;

  // Calculated:
  double
    halfWidth,
    minDir,
    maxDir;

} AngMultT;

#define MaxAngMults 10

typedef struct {
  colSpT colSpIn;
  colSpT colSpOut;

  int ndxCat;  // index into cats array

  int
    verbose,
    precision;

  double
    gain2,
    gain,
    bias,
    power;

  int nAngMults;
  AngMultT AngMults[MaxAngMults];

  MagickBooleanType
    clampXy,
    clampBary,
    ignoreWP,
    doTriangles,
    round16;

  FILE *
    fh_data;

  mat3x3T
    chromAdapTrans;

} barymT;


typedef struct {
  char * name;
  coordsT
    coordsR,
    coordsG,
    coordsB,
    coordsWP;
  double gamma;
} colSpNumsT;

// These are CIE 2 degrees
const colSpNumsT colSpNums[] = {
  {"sRGB",  // From https://en.wikipedia.org/wiki/SRGB,
            // also IM "-colorspace sRGB"
     {0.64, 0.33},
     {0.3, 0.6},
     {0.15, 0.06},
     {0.3127, 0.3290}, // D65
     -1
  },
  {"sRgbD50",  // From http://www.brucelindbloom.com/index.html?WorkingSpaceInfo.html
     {0.648431, 0.330856},
     {0.321152, 0.597871},
     {0.155886, 0.066044},
     {0.34567, 0.35850}, // D50
     -1
  },
  {"Rec709",  // From https://en.wikipedia.org/wiki/Rec._709
     {0.64, 0.33},
     {0.3, 0.6},
     {0.15, 0.06},
     {0.3127, 0.3290}, // D65
     2.4 // But not really, see https://en.wikipedia.org/wiki/Rec._709
  },
  {"P3D65", // From https://en.wikipedia.org/wiki/DCI-P3
     {0.680, 0.320},
     {0.265, 0.690},
     {0.150, 0.060},
     {0.3127, 0.3290}, // D65
     2.6
  },
  {"DisplayP3", // From https://en.wikipedia.org/wiki/DCI-P3
     {0.680, 0.320},
     {0.265, 0.690},
     {0.150, 0.060},
     {0.3127, 0.3290}, // D65
     -1
  },
  {"AdobeRGB", // From https://en.wikipedia.org/wiki/Adobe_RGB_color_space
     {0.6400, 0.3300},
     {0.2100, 0.7100},
     {0.1500, 0.0600},
     {0.3127, 0.3290}, // D65
     563/256  // = 2.19921875 exactly
  },
  {"Rec2020", // From https://en.wikipedia.org/wiki/Rec._2020
     {0.708, 0.292},
     {0.170, 0.797},
     {0.131, 0.046},
     {0.3127, 0.3290}, // D65
     2.4 // But not really, see https://en.wikipedia.org/wiki/Rec._2020#Transfer_characteristics
  },
  {"ProPhotoRGB", // From https://en.wikipedia.org/wiki/ProPhoto_RGB_color_space
     {0.7347, 0.2653},
     {0.1596, 0.8404},
     {0.0366, 0.0001},
     {0.3457, 0.3585},
     1.8 // Not really, see https://en.wikipedia.org/wiki/ProPhoto_RGB_color_space#Encoding_function
  },
  {"ACEScg", // From https://www.oscars.org/science-technology/aces/aces-documentation
     {0.713, 0.293},
     {0.165, 0.830},
     {0.128, 0.044},
     {0.32168, 0.33767}, // Close to D60, apparently.
     1
  },
  {"ACES2065-1", // From https://en.wikipedia.org/wiki/Academy_Color_Encoding_System
     {0.7347, 0.2653},
     {0.0000, 1.0000},
     {0.0001, -0.0770},
     {0.32168, 0.33767}, // Close to D60, apparently.
     1
  },
  {"xyY",
     {1.0, 0.0},
     {0.0, 1.0},
     {0.0, 0.0},
     {0.34567, 0.35850}, // Assumed to be D50
     1
  },
  {"XYZ",
     {1.0, 0.0},
     {0.0, 1.0},
     {0.0, 0.0},
     {0.34567, 0.35850}, // Assumed to be D50
     1
  }
};


typedef struct {
  char * name;
  coordsT
    coordsWP;
} wpCoordsT;

// White points,
// mostly from https://en.wikipedia.org/wiki/Standard_illuminant#White_point
//
// CIE 1931 2 degrees.
const wpCoordsT wpCoords [] = {
  {"A",    {0.44757, 0.40745}},
  {"D50",  {0.34567, 0.35850}},
  {"D60",  {0.32168, 0.33767}},  // from https://en.wikipedia.org/wiki/Academy_Color_Encoding_System
  {"D65",  {0.31271, 0.32902}},
  {"D65s", {0.3127,  0.3290}},
  {"D75",  {0.29902, 0.31485}},
  {"D100", {0.2824,  0.2898}},  // from Vendian
  {"D200", {0.2580,  0.2574}},  // from Vendian
  {"D300", {0.2516,  0.2481}},  // from Vendian
  {"D400", {0.2487,  0.2438}},  // from Vendian
  {"E",    {1/3.0,   1/3.0}}
};

// Vendian = http://www.vendian.org/mncharity/dir3/blackbody/UnstableURLs/bbr_color.html
// Vendian actually gives xy for 10000K etc, which is close to D100.

// "a" suffix from http://www.brucelindbloom.com/index.html?Eqn_ChromAdapt.html,
// citing "ASTM E308-01 except B which comes from Wyszecki & Stiles, p. 769".
// "i" suffix from ICC1v43_2010-12.
// These are values of X and Z (Y=1), not x and y.
const wpCoordsT wpCoordsXZ [] = {
  {"Aa",    {1.09850, 0.35585}},
  {"D50a",  {0.96422, 0.82521}},
  {"D55a",  {0.95682, 0.92149}},
  {"D65a",  {0.95047, 1.08883}},
  {"D75a",  {0.94972, 1.22638}},
  {"Ea",    {1,       1}},
  {"D50i",  {0.9642,  0.8249}}
};


//--------------------------------------------
//
// Colour conversion.
//

typedef struct {
  char * name;
  mat3x3T mat;
} catT;

// Ref: http://www.ivl.disco.unimib.it/download/bianco2010two-new.pdf
//
// http://www.brucelindbloom.com/index.html?Eqn_ChromAdapt.html

catT cats[] = {
  { "None",
     {1,0,0, 0,1,0, 0,0,1}
  },
  { "XYZ",
     {1,0,0, 0,1,0, 0,0,1}
  },
  { "VonKries",
     { 0.3897, 0.6890, -0.0787,
      -0.2298, 1.1834,  0.0464,
      0, 0, 1}
  },
  { "Bradford",
     { 0.8951, 0.2664, -0.1614,
      -0.7502, 1.7135,  0.0367,
       0.0389, -0.0685, 1.0296}
// Bianco gives -0.0686; Lindbloom gives -0.0685
  },
  { "Sharp",
     { 1.2694, 0.0988, -0.1706,
      -0.8364, 1.8006,  0.0357,
       0.0297, -0.0315, 1.0018}
  },
  { "CMCCAT2000",
     { 0.7982, 0.3389, -0.1371,
      -0.5918, 1.5512,  0.0406,
       0.0008, 0.239,   0.9753 }
  },
  { "CAT02",
     // See also http://www.rit-mcsl.org/fairchild/PDFs/PRO19.pdf eqn (7).
     { 0.7328, 0.4296, -0.1624,
      -0.7036, 1.6975,  0.0061,
       0.0030, 0.0136,  0.9834 }
  },
  { "BS",
     { 0.8752, 0.2787, -0.1539,
      -0.8904, 1.8709,  0.0195,
      -0.0061, 0.0162,  0.9899}
  },
  { "BS-PC",
     { 0.6489, 0.3915, -0.0404,
      -0.3775, 1.3055,  0.0720,
      -0.0271, 0.0888,  0.9383}
  },
  { "Fairchild",
     // From Spectral Sharpening of Color Sensors:  Diagonal Color Constancy and Beyond
     // https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4003926/
     { 0.8562, 0.3372, -0.1934,
      -0.8360, 1.8327,  0.0033,
       0.0357, -0.0469, 1.0112}
  }
};

#define dfltCat 3  // Bradford

typedef struct {
  char * name;
  double
    power,
    slope,
    limit,
    offset;
} trcT;

trcT trcs[] = {
  { "None",
     1, 1, 0, 0
  },
  { "sRGB",
     2.4, 12.92, 0.04045, 0.055
  },
  { "Rec709",
     1.0/0.45, 4.5, 0.018, 0.099
  },
  { "Rec2020",
     1.0/0.45, 4.5, 0.018053968510807, 0.09929682680944
  },
  { "ProPhoto", // also see http://www.color.org/ROMMRGB.pdf
     1.8, 16, 1.0/512.0, 0
  }
};


//--------------------------------------------

static MagickBooleanType ParseNamedPrims (
  barymT *pbary,
  const char * name,
  colSpT * pcs)
// return TRUE if okay.
{
  MagickBooleanType ok = MagickFalse;
  int i;
  for (i = 0; i < sizeof(colSpNums) / sizeof(colSpNumsT); i++) {
    const colSpNumsT * csn = &colSpNums[i];
    if (LocaleCompare(name, csn->name)==0) {
      if (pbary->verbose)
        fprintf (pbary->fh_data, "Setting colSp from %s\n", csn->name);
      pcs->triMain.primA.x = csn->coordsR.x;
      pcs->triMain.primA.y = csn->coordsR.y;
      pcs->triMain.primB.x = csn->coordsG.x;
      pcs->triMain.primB.y = csn->coordsG.y;
      pcs->triMain.primC.x = csn->coordsB.x;
      pcs->triMain.primC.y = csn->coordsB.y;
      pcs->WP.x = csn->coordsWP.x;
      pcs->WP.y = csn->coordsWP.y;
      pcs->XYZwp = pcs->WP;
      pcs->gamma = csn->gamma;
      ok = MagickTrue;
      break;
    }
  }

  return ok;
}

static MagickBooleanType ParseNamedWP (
  barymT *pbary,
  const char * name,
  coordsT * pwp)
// return TRUE if okay.
{
  MagickBooleanType ok = MagickFalse;
  int i;
  for (i = 0; i < sizeof(wpCoords) / sizeof(wpCoordsT); i++) {

    if (LocaleCompare(name, wpCoords[i].name)==0) {
      if (pbary->verbose)
        fprintf (pbary->fh_data, "Setting WP from %s\n", wpCoords[i].name);
      pwp->x = wpCoords[i].coordsWP.x;
      pwp->y = wpCoords[i].coordsWP.y;
      ok = MagickTrue;
      break;
    }
  }

  if (!ok) {
    for (i = 0; i < sizeof(wpCoordsXZ) / sizeof(wpCoordsT); i++) {

      if (LocaleCompare(name, wpCoordsXZ[i].name)==0) {
        if (pbary->verbose)
          fprintf (pbary->fh_data, "Setting WP from %s\n", wpCoords[i].name);
        pwp->x = wpCoordsXZ[i].coordsWP.x / (wpCoordsXZ[i].coordsWP.x+wpCoordsXZ[i].coordsWP.y+1);
        pwp->y = 1.0 / (wpCoordsXZ[i].coordsWP.x+wpCoordsXZ[i].coordsWP.y+1);
        ok = MagickTrue;
        break;
      }
    }
  }

  return ok;
}

static MagickBooleanType ParseNamedCat (
  barymT *pbary,
  const char * name)
// return TRUE if okay.
{
  MagickBooleanType ok = MagickFalse;
  int i;
  for (i = 0; i < sizeof(cats) / sizeof(catT); i++) {

    if (LocaleCompare(name, cats[i].name)==0) {
      pbary->ndxCat = i;
      ok = MagickTrue;
      break;
    }
  }

  return ok;
}

static MagickBooleanType ParseNamedTrc (
  barymT *pbary,
  const char * name,
  colSpT * pcs)
// return TRUE if okay.
{
//  fprintf (stderr, "ParseNamedTrc: [%s]\n", name);

  MagickBooleanType ok = MagickFalse;
  int i;
  for (i = 0; i < sizeof(trcs) / sizeof(trcT); i++) {
//    fprintf (stderr, "  [%s]\n", trcs[i].name);

    if (LocaleCompare(name, trcs[i].name)==0) {
      pcs->gamma = -i;
      ok = MagickTrue;
      break;
    }
  }

//  if (ok) fprintf (stderr, "ParseNamedTrc: [%s] found %g\n", name, pcs->gamma);
//  else fprintf (stderr, "ParseNamedTrc: [%s] not found\n", name);
  return ok;
}

static void ListColSp (FILE * fh)
{
  int i;
  for (i = 0; i < sizeof(colSpNums) / sizeof(colSpNumsT); i++) {
    fprintf (fh, "%s\n", colSpNums[i].name);
  }
}

static void ListWP (FILE * fh)
{
  int i;
  for (i = 0; i < sizeof(wpCoords) / sizeof(wpCoordsT); i++) {
    fprintf (fh, "%s\n", wpCoords[i].name);
  }
  for (i = 0; i < sizeof(wpCoordsXZ) / sizeof(wpCoordsT); i++) {
    fprintf (fh, "%s\n", wpCoordsXZ[i].name);
  }
}

static void ListTrans (FILE * fh)
{
  int i;
  for (i = 0; i < sizeof(trcs) / sizeof(trcT); i++) {
    fprintf (fh, "%s\n", trcs[i].name);
  }
}

static void ListCats (FILE * fh)
{
  int i;
  for (i = 0; i < sizeof(cats) / sizeof(catT); i++) {
    fprintf (fh, "%s\n", cats[i].name);
  }
}

static void WrCoords (barymT * pbary, coordsT * pc, FILE * fh)
{
  fprintf (fh, "%.*g,%.*g",
    pbary->precision, pc->x,
    pbary->precision, pc->y);
}

static void WrTri (barymT * pbary, triangleT * ptri, FILE * fh)
{
  WrCoords (pbary, &ptri->primA, fh);
  fprintf (fh, ",");
  WrCoords (pbary, &ptri->primB, fh);
  fprintf (fh, ",");
  WrCoords (pbary, &ptri->primC, fh);
}

static void WrColSp (barymT * pbary, FILE * fh, colSpT * cs, char * pref)
{
  fprintf (stderr, "  %sChannels ", pref);
  if (cs->menuSetChan) fprintf (fh, "* ");
  switch (cs->chans) {
    case chRGB: fprintf (stderr, "RGB"); break;
    case chXYY: fprintf (stderr, "xyY "); break;
    case chXYZ: fprintf (stderr, "XYZ "); break;
  }

  fprintf (fh, "\n  %sPrim ", pref);
  if (cs->menuSetPrim) fprintf (fh, "* ");
  WrTri (pbary, &cs->triMain, fh);

  fprintf (fh, "\n  %sWP ", pref);
  if (cs->menuSetWP) fprintf (fh, "* ");
  WrCoords (pbary, &cs->WP, fh);

  fprintf (fh, "\n  %sTransfer ", pref);
  if (cs->menuSetGam) fprintf (fh, "* ");
  if (cs->gamma >= 0) {
    fprintf (fh, "%.*g",
      pbary->precision, cs->gamma);
  } else {
    fprintf (fh, "%s",
      trcs[(int)floor(-cs->gamma)].name);

  }

  fprintf (fh, "\n  XYZwp ");
  if (cs->menuSetXYZwp) fprintf (fh, "* ");
  WrCoords (pbary, &cs->XYZwp, fh);

  fprintf (fh, "\n");
}

static void WrBarymap (barymT * pbary, FILE * fh)
{
  fprintf (fh, "In:\n");
  WrColSp (pbary, fh, &pbary->colSpIn, "in");

  fprintf (fh, "Out:\n");
  WrColSp (pbary, fh, &pbary->colSpOut, "out");
}

//--------------------------------------------
//
// Colour conversion functions.
//

typedef struct {
  double X;
  double Y;
  double Z;
} XYZT;

typedef struct {
  double R;
  double G;
  double B;
} RGBT;

static void xyYtoXYZ (
  double x,
  double y,
  double UY,
  XYZT * out
)
{
  // Convert xyY to xYZ.

  out->Y = UY;

  if (y==0) {
    out->X = out->Z = 0;
  } else {
    out->X = UY * x / y;
    out->Z = UY * (1 - x - y) / y;
  }
}

static void xyOneToXYZ (
  double x,
  double y,
  XYZT * out
)
{
  // Convert xyY to XYZ where Y==1.

  out->Y = 1;

  if (y==0) {
    out->X = out->Z = 0;
  } else {
    out->X = x / y;
    out->Z = (1 - x - y) / y;
  }
}

static double inline Ro16If (double v, MagickBooleanType DoRound)
{
  return (DoRound) ? round (v * 65536) / 65536 : v;
}

static void inline Round16 (double * v)
{
  *v = Ro16If (*v, MagickTrue);
}

static void inline Round16xy (barymT * pbary, coordsT * xy)
{
  if (pbary->round16) {
    Round16 (&xy->x);
    Round16 (&xy->y);
  }
}

static void inline Round16XYZ (barymT * pbary, XYZT * XYZ)
{
  if (pbary->round16) {
    Round16 (&XYZ->X);
    Round16 (&XYZ->Y);
    Round16 (&XYZ->Z);
  }
}

static void inline Round16mat3x3 (barymT * pbary, mat3x3T m)
{
  if (pbary->round16) {
    int i;
    for (i=0; i<9; i++) Round16 (&m[i]);
  }
}

static void WrWpNums (FILE * fh, barymT * pbary)
{
  int i;
  XYZT XYZ;
  fprintf (fh, "WpNums:\n");
  fprintf (fh, "  name,  x,y,z  X,Y,Z\n");

  for (i = 0; i < sizeof(wpCoords) / sizeof(wpCoordsT); i++) {
    const wpCoordsT * pwp = &wpCoords[i];
    fprintf (fh, "  %s", pwp->name);
    xyOneToXYZ (pwp->coordsWP.x, pwp->coordsWP.y, &XYZ);
    fprintf (fh, "  %.*g, %.*g, %.*g,  %.*g, %.*g, %.*g\n",
      pbary->precision, Ro16If (pwp->coordsWP.x, pbary->round16),
      pbary->precision, Ro16If (pwp->coordsWP.y, pbary->round16),
      pbary->precision, Ro16If (1.0 - pwp->coordsWP.x - pwp->coordsWP.y, pbary->round16),
      pbary->precision, Ro16If (XYZ.X, pbary->round16),
      pbary->precision, Ro16If (XYZ.Y, pbary->round16),
      pbary->precision, Ro16If (XYZ.Z, pbary->round16)
    );
  }

  for (i = 0; i < sizeof(wpCoordsXZ) / sizeof(wpCoordsT); i++) {
    const wpCoordsT * pwp = &wpCoordsXZ[i];
    xyzCoordsT xyzLo;
    double div = pwp->coordsWP.x + 1 + pwp->coordsWP.y;
    xyzLo.x = pwp->coordsWP.x / div;
    xyzLo.y = 1.0 / div;
    xyzLo.z = 1.0 - xyzLo.x - xyzLo.y;
    fprintf (fh, "  %s", pwp->name);
    xyOneToXYZ (xyzLo.x, xyzLo.y, &XYZ);
    fprintf (fh, "  %.*g, %.*g, %.*g,  %.*g, %.*g, %.*g\n",
      pbary->precision, Ro16If (xyzLo.x, pbary->round16),
      pbary->precision, Ro16If (xyzLo.y, pbary->round16),
      pbary->precision, Ro16If (xyzLo.z, pbary->round16),
      pbary->precision, Ro16If (XYZ.X, pbary->round16),
      pbary->precision, Ro16If (XYZ.Y, pbary->round16),
      pbary->precision, Ro16If (XYZ.Z, pbary->round16)
    );
  }
}

static void WrXYZ (barymT * pbary, const XYZT * in, char * title)
{
  fprintf (pbary->fh_data, "%s XYZ = %.*g, %.*g, %.*g\n",
    title,
    pbary->precision, in->X,
    pbary->precision, in->Y,
    pbary->precision, in->Z);
}

static void inline MultMatXYZ (const mat3x3T m, const XYZT *XYZ, XYZT * out)
// [out] = [m][XYZ]
{
  out->X = XYZ->X*m[0] + XYZ->Y*m[1] + XYZ->Z*m[2];
  out->Y = XYZ->X*m[3] + XYZ->Y*m[4] + XYZ->Z*m[5];
  out->Z = XYZ->X*m[6] + XYZ->Y*m[7] + XYZ->Z*m[8];
}

static void WrMat3x3Bary (barymT * pbary, const mat3x3T in, char * title)
{
  WrMat3x3 (pbary->fh_data, pbary->precision, in, title);
}

static void TstInvMat3x3 (barymT * pbary)
{
  mat3x3T out;

  // http://www.brucelindbloom.com/index.html?Eqn_chromAdap.html
  // Bradford MA:
  //
  double in[] = {
    0.8951000,  0.2664000, -0.1614000,
   -0.7502000,  1.7135000,  0.0367000,
    0.0389000, -0.0685000,  1.0296000};

  WrMat3x3Bary (pbary, in, "in:");

  IdentMat3x3 (out);

  if (!InvMat3x3 (in, out)) {
    fprintf (pbary->fh_data, "Not invertible.\n");
  }

  WrMat3x3Bary (pbary, out, "out");

  mat3x3T mult;
  MultMat3x3 (in, out, mult);
  WrMat3x3Bary (pbary, mult, "mult");
}

static MagickBooleanType ChromAdapInOut (
  barymT * pbary,
  coordsT * WPin,
  coordsT * WPout,
  mat3x3T matCa)
// From the in and out white points,
//   returns a 3x3 Chromatic Adaptation matrix.
// Also calculates unadapted and adapted primaries,
//   but does nothing with them.
// Ref http://www.brucelindbloom.com/index.html?Eqn_chromAdap.html
// Also https://ninedegreesbelow.com/photography/srgb-color-space-to-profile.html
// Also ICC spec, Annex E.
//
// Note: in ICC profiles, numbers are recorded as fixed-point
//   (not floating-point), with 16 bits for the fractional part,
//   about 4 decimal places.
{
  if (pbary->verbose >= 2)
    fprintf (pbary->fh_data, "ChromAdap: WP from %.*g, %.*g  to %.*g, %.*g\n",
      pbary->precision, WPin->x,
      pbary->precision, WPin->y,
      pbary->precision, WPout->x,
      pbary->precision, WPout->y);

  if ( (WPin->x != WPout->x) || (WPin->y != WPout->y)) {

    XYZT XYZsrc, XYZdst, coneSrc, coneDst;

    mat3x3T Ma, MaInv;
    CopyMat3x3 (cats[pbary->ndxCat].mat, Ma);

    if (!InvMat3x3 (Ma, MaInv)) {
      fprintf (pbary->fh_data, "Ma not invertible.\n");
      return MagickFalse;
    }

    if (pbary->verbose >= 2) {
      WrMat3x3Bary (pbary, Ma, "chromAdap: Ma");
      WrMat3x3Bary (pbary, MaInv, "chromAdap: MaInv:");
    }

    // For the white points, get XYZ from xy.
    xyOneToXYZ (WPin->x, WPin->y, &XYZsrc);
    xyOneToXYZ (WPout->x, WPout->y, &XYZdst);

    if (pbary->verbose >= 2) {
      WrXYZ (pbary, &XYZsrc, "src WP");
      WrXYZ (pbary, &XYZdst, "dst WP");
    }

    // Transform WP XYZs to cone response domains (rho, gamma, beta).
    MultMatXYZ (Ma, &XYZsrc, &coneSrc);
    MultMatXYZ (Ma, &XYZdst, &coneDst);

    if (pbary->verbose >= 2) {
      WrXYZ (pbary, &coneSrc, "cone src");
      WrXYZ (pbary, &coneDst, "cone dst");
    }

    mat3x3T crdMat;
    ZeroMat3x3 (crdMat);
    crdMat[0] = coneDst.X / coneSrc.X;
    crdMat[4] = coneDst.Y / coneSrc.Y;
    crdMat[8] = coneDst.Z / coneSrc.Z;

    mat3x3T tmp;
    MultMat3x3 (MaInv, crdMat, tmp);
    if (pbary->verbose >= 2) {
      WrMat3x3Bary (pbary, crdMat, "chromAdap: crdMat:");
      WrMat3x3Bary (pbary, tmp, "chromAdap: tmp:");
    }
    MultMat3x3 (tmp, Ma, matCa);

    Round16mat3x3 (pbary, matCa);
  } else {
    IdentMat3x3 (matCa);
  }

  if (pbary->verbose >= 2) {
    WrMat3x3Bary (pbary, matCa, "chromAdap: matCa:");
  }

  // Following Elle,
  // we calculate unadjusted and adjusted XYZ of primaries,
  // but we do nothing with these.

  XYZT rXYZ, gXYZ, bXYZ;

  xyOneToXYZ (pbary->colSpIn.triMain.primA.x, pbary->colSpIn.triMain.primA.y, &rXYZ);
  xyOneToXYZ (pbary->colSpIn.triMain.primB.x, pbary->colSpIn.triMain.primB.y, &gXYZ);
  xyOneToXYZ (pbary->colSpIn.triMain.primC.x, pbary->colSpIn.triMain.primC.y, &bXYZ);

  if (pbary->verbose >= 2) {
    WrXYZ (pbary, &rXYZ, "interim unadap r");
    WrXYZ (pbary, &gXYZ, "interim unadap g");
    WrXYZ (pbary, &bXYZ, "interim unadap b");
  }

  mat3x3T tmp;
  tmp[0] = rXYZ.X;
  tmp[1] = gXYZ.X;
  tmp[2] = bXYZ.X;

  tmp[3] = rXYZ.Y;
  tmp[4] = gXYZ.Y;
  tmp[5] = bXYZ.Y;

  tmp[6] = rXYZ.Z;
  tmp[7] = gXYZ.Z;
  tmp[8] = bXYZ.Z;

  if (pbary->verbose >= 2) {
    WrMat3x3Bary (pbary, tmp, "chromAdap: tmp:");
  }

  mat3x3T tmpInv;  // Elle's table 5c.

  if (!InvMat3x3 (tmp, tmpInv)) {
    fprintf (pbary->fh_data, "tmp not invertible.\n");
    return MagickFalse;
  }

  if (pbary->verbose >= 2) {
    WrMat3x3Bary (pbary, tmpInv, "chromAdap: tmpInv:");
  }

  // For the source white point, get XYZ from xy.
  XYZT XYZsrc;
  xyOneToXYZ (pbary->colSpIn.WP.x, pbary->colSpIn.WP.y, &XYZsrc);

  XYZT adjY; // Calculated unadapted Y for red, green and blue primaries
  MultMatXYZ (tmpInv, &XYZsrc, &adjY);

  xyYtoXYZ (pbary->colSpIn.triMain.primA.x, pbary->colSpIn.triMain.primA.y, adjY.X, &rXYZ);
  xyYtoXYZ (pbary->colSpIn.triMain.primB.x, pbary->colSpIn.triMain.primB.y, adjY.Y, &gXYZ);
  xyYtoXYZ (pbary->colSpIn.triMain.primC.x, pbary->colSpIn.triMain.primC.y, adjY.Z, &bXYZ);

  if (pbary->verbose >= 2) {
    WrXYZ (pbary, &rXYZ, "unadapted r");
    WrXYZ (pbary, &gXYZ, "unadapted g");
    WrXYZ (pbary, &bXYZ, "unadapted b");
  }

  XYZT rAdXYZ, gAdXYZ, bAdXYZ;
  MultMatXYZ (matCa, &rXYZ, &rAdXYZ);
  MultMatXYZ (matCa, &gXYZ, &gAdXYZ);
  MultMatXYZ (matCa, &bXYZ, &bAdXYZ);

  if (pbary->verbose >= 2) {
    WrXYZ (pbary, &rAdXYZ, "adapted r");
    WrXYZ (pbary, &gAdXYZ, "adapted g");
    WrXYZ (pbary, &bAdXYZ, "adapted b");
  }

  return MagickTrue;
}

static MagickBooleanType ChromAdap (barymT * pbary)
{
  if (pbary->verbose >= 2) fprintf (pbary->fh_data, "ChromAdap WP->WP\n");
  if (!ChromAdapInOut (
    pbary,
    &pbary->colSpIn.WP, &pbary->colSpOut.WP,
    pbary->chromAdapTrans)) return MagickFalse;

  if (pbary->verbose >= 2) fprintf (pbary->fh_data, "ChromAdap in WP->X\n");
  if (!ChromAdapInOut (
    pbary,
    &pbary->colSpIn.WP, &pbary->colSpIn.XYZwp,
    pbary->colSpIn.chromAdapTrans)) return MagickFalse;
  if (pbary->verbose >= 2) WrMat3x3Bary (pbary, pbary->colSpIn.chromAdapTrans, "ChromAdapTrans in ");

  if (pbary->verbose >= 2) fprintf (pbary->fh_data, "ChromAdap out X->WP\n");
  if (!ChromAdapInOut (
    pbary,
    &pbary->colSpOut.XYZwp, &pbary->colSpOut.WP,
    pbary->colSpOut.chromAdapTrans)) return MagickFalse;
  if (pbary->verbose >= 2) WrMat3x3Bary (pbary, pbary->colSpOut.chromAdapTrans, "ChromAdapTrans out ");

  // Multiply matrices to incorporate chromatic adaptation.

  if (pbary->verbose >= 2) {
    WrMat3x3Bary (pbary, pbary->colSpIn.chromAdapTrans, "RGB2XYZchr A");
    WrMat3x3Bary (pbary, pbary->colSpIn.RGB2XYZ, "RGB2XYZchr B");
  }
  MultMat3x3 (pbary->colSpIn.chromAdapTrans, pbary->colSpIn.RGB2XYZ, pbary->colSpIn.RGB2XYZchr);
  Round16mat3x3 (pbary, pbary->colSpIn.RGB2XYZchr);
  if (pbary->verbose >= 2) WrMat3x3Bary (pbary, pbary->colSpIn.RGB2XYZchr, "RGB2XYZchr");

  if (pbary->verbose >= 2) {
    WrMat3x3Bary (pbary, pbary->colSpOut.XYZ2RGB, "RGB2XYZchr C");
    WrMat3x3Bary (pbary, pbary->colSpOut.chromAdapTrans, "RGB2XYZchr D");
  }
  MultMat3x3 (pbary->colSpOut.XYZ2RGB, pbary->colSpOut.chromAdapTrans, pbary->colSpOut.XYZ2RGBchr);
  Round16mat3x3 (pbary, pbary->colSpIn.XYZ2RGBchr);
  if (pbary->verbose >= 2) WrMat3x3Bary (pbary, pbary->colSpOut.XYZ2RGBchr, "XYZ2RGBchr");

  return MagickTrue;
}

static MagickBooleanType RgbXyzMat (
  barymT * pbary,
  colSpT * colSp)
// Given a colorspace primaries and WP,
// Calculates the 3x3 matrices to convert between linear RGB and XYZ.
// Ref http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
{
  double Yr = 1.0;
  double Yg = 1.0;
  double Yb = 1.0;

  double Xr, Zr, Xg, Zg, Xb, Zb;

  if (colSp->triMain.primA.y==0) {
    Xr = Zr = 0;
  } else {
    Xr = colSp->triMain.primA.x / colSp->triMain.primA.y;
    Zr = (1 - colSp->triMain.primA.x - colSp->triMain.primA.y) /
              colSp->triMain.primA.y;
  }

  if (colSp->triMain.primB.y==0) {
    Xg = Zg = 0;
  } else {
    Xg = colSp->triMain.primB.x / colSp->triMain.primB.y;
    Zg = (1 - colSp->triMain.primB.x - colSp->triMain.primB.y) /
              colSp->triMain.primB.y;
  }

  if (colSp->triMain.primC.y==0) {
    Xb = Zb = 0;
  } else {
    Xb = colSp->triMain.primC.x / colSp->triMain.primC.y;
    Zb = (1 - colSp->triMain.primC.x - colSp->triMain.primC.y) /
              colSp->triMain.primC.y;
  }

  mat3x3T tmp, tmpInv;
  tmp[0] = Xr;
  tmp[1] = Xg;
  tmp[2] = Xb;
  tmp[3] = Yr;
  tmp[4] = Yg;
  tmp[5] = Yb;
  tmp[6] = Zr;
  tmp[7] = Zg;
  tmp[8] = Zb;

  if (pbary->verbose >= 2) {
    WrMat3x3Bary (pbary, tmp, "RgbXyzMat: tmp:");
  }

  if (!InvMat3x3 (tmp, tmpInv)) {
    fprintf (pbary->fh_data, "RgbXyzMat: tmp not invertible.\n");
    return MagickFalse;
  }

  if (pbary->verbose >= 2) {
    WrMat3x3Bary (pbary, tmpInv, "RgbXyzMat: tmpInv:");
  }

  // For the white point, get XYZ from xy.
  XYZT XYZsrc;
  xyOneToXYZ (colSp->WP.x, colSp->WP.y, &XYZsrc);

  if (pbary->verbose >= 2) {
    WrXYZ (pbary, &XYZsrc, "src WP");
  }

  XYZT S;
  MultMatXYZ (tmpInv, &XYZsrc, &S);

  if (pbary->verbose >= 2) {
    WrXYZ (pbary, &S, "S");
  }

  colSp->RGB2XYZ[0] = S.X * Xr;
  colSp->RGB2XYZ[1] = S.Y * Xg;
  colSp->RGB2XYZ[2] = S.Z * Xb;

  colSp->RGB2XYZ[3] = S.X * Yr;
  colSp->RGB2XYZ[4] = S.Y * Yg;
  colSp->RGB2XYZ[5] = S.Z * Yb;

  colSp->RGB2XYZ[6] = S.X * Zr;
  colSp->RGB2XYZ[7] = S.Y * Zg;
  colSp->RGB2XYZ[8] = S.Z * Zb;

  if (!InvMat3x3 (colSp->RGB2XYZ, colSp->XYZ2RGB)) {
    fprintf (pbary->fh_data, "RgbXyzMat: colSp->RGB2XYZ not invertible.\n");
    return MagickFalse;
  }

  Round16mat3x3 (pbary, colSp->RGB2XYZ);
  Round16mat3x3 (pbary, colSp->XYZ2RGB);
  // FIXME: or round before invert?

  if (pbary->verbose >= 2) {
    WrMat3x3Bary (pbary, colSp->RGB2XYZ, "RgbXyzMat: RGB2XYZ:");
    WrMat3x3Bary (pbary, colSp->XYZ2RGB, "RgbXyzMat: XYZ2RGB:");
  }

  return MagickTrue;
}


/* RGB to XYZ:
   Ref http://www.brucelindbloom.com/index.html?Eqn_RGB_to_XYZ.html

   inverse companding to linear RGB
   linear RGB to XYZ via 3x3 matrix
   if needed, apply chromatic adaptation transform to XYZ
*/

static double inline powchk (double v, double p)
{
  return (v <= 0) ? v : pow (v, p);
  // FIXME: Crude. When v<0, should use slope at zero.
}


static void ZeroTrans (transT * pt)
{
  pt->name = NULL;
  pt->power = 1;
  pt->slope = 1;
  pt->limit = 0;
  pt->offset = 0;

  pt->invPower = 1;
  pt->limit2 = 0;
  pt->offsetP1 = 1;
}

static void CalcTrans (barymT * pbary, transT * pt, int ndxTrc)
{
  trcT * ptf = &trcs[ndxTrc];

  pt->name = ptf->name;
  pt->power    = Ro16If (ptf->power, pbary->round16);
  pt->slope    = Ro16If (ptf->slope, pbary->round16);
  pt->limit    = Ro16If (ptf->limit, pbary->round16);
  pt->offset   = Ro16If (ptf->offset, pbary->round16);

  pt->invPower = Ro16If (1.0 / pt->power, pbary->round16);
  pt->limit2   = Ro16If (pt->limit / pt->slope, pbary->round16);
  pt->offsetP1 = Ro16If (pt->offset + 1.0, pbary->round16);

  if (pbary->verbose >= 2) {
    fprintf (pbary->fh_data, "%s %.*g %.*g %.*g %.*g  %.*g %.*g %.*g\n",
      pt->name,
      pbary->precision, pt->power,
      pbary->precision, pt->slope,
      pbary->precision, pt->limit,
      pbary->precision, pt->offset,
      pbary->precision, pt->invPower,
      pbary->precision, pt->limit2,
      pbary->precision, pt->offsetP1);
  }
}

static double inline TrcToLin (const transT * pt, double v)
{
  if (v < pt->limit) return v / pt->slope;

  return pow ((v + pt->offset) / pt->offsetP1, pt->power);
}

static double inline TrcFromLin (const transT * pt, double v)
{
  if (v < pt->limit2) return pt->slope * v;

  return pow (v, pt->invPower) * pt->offsetP1 - pt->offset;

}

static void inline Rgb2Xyz (const colSpT * colSp, RGBT * rgb, XYZT * XYZ)
{
  // From non-lin to lin.
  if (colSp->gamma != 1) {
    if (colSp->gamma > 0) {
      rgb->R = powchk(rgb->R, colSp->gamma);
      rgb->G = powchk(rgb->G, colSp->gamma);
      rgb->B = powchk(rgb->B, colSp->gamma);
    } else {
      rgb->R = TrcToLin(&colSp->trans, rgb->R);
      rgb->G = TrcToLin(&colSp->trans, rgb->G);
      rgb->B = TrcToLin(&colSp->trans, rgb->B);
    }
  }
  MultMatXYZ (colSp->RGB2XYZ, (XYZT *)rgb, XYZ);
}

/* XYZ to RGB:
   Ref http://www.brucelindbloom.com/index.html?Eqn_XYZ_to_RGB.html

   if needed, apply chromatic adaptation transform to XYZ
   XYZ to linear RGB via 3x3 matrix
   companding from linear RGB to non-linear RGB
*/

static void inline Xyz2Rgb (const colSpT * colSp, XYZT * XYZ, RGBT * rgb)
{
  MultMatXYZ (colSp->XYZ2RGB, XYZ, (XYZT *)rgb);

  // From lin to non-lin
  if (colSp->gamma != 1) {
    if (colSp->gamma > 0) {
      double g = 1.0 / colSp->gamma;
      rgb->R = powchk(rgb->R, g);
      rgb->G = powchk(rgb->G, g);
      rgb->B = powchk(rgb->B, g);
    } else {
      rgb->R = TrcFromLin(&colSp->trans, rgb->R);
      rgb->G = TrcFromLin(&colSp->trans, rgb->G);
      rgb->B = TrcFromLin(&colSp->trans, rgb->B);
    }
  }
}

static void inline ChromAdapXYZ (
  barymT * pbary,
  XYZT * xyz)
{
  XYZT tmp;
  MultMatXYZ (pbary->chromAdapTrans, xyz, &tmp);
  xyz->X = tmp.X;
  xyz->Y = tmp.Y;
  xyz->Z = tmp.Z;
}

static void inline xy2XYZ (
  coordsT * xy,
  XYZT * XYZ)
{
  if (xy->y == 0) {
    XYZ->X = 0;
    XYZ->Z = 0;
  } else {
    XYZ->X = XYZ->Y * xy->x / xy->y;
    XYZ->Z = XYZ->Y * (1 - xy->x - xy->y) / xy->y;
  }
}

static MagickBooleanType ImageHasChromaticity (Image * im)
{
  return (
    im->chromaticity.red_primary.x != 0 &&
    im->chromaticity.red_primary.y &&
    im->chromaticity.white_point.x &&
    im->chromaticity.white_point.y);
}

static void ColSpFromIm (
  Image * im,
  barymT * pbary)
/* Copy metadata from image _only_ where these were not set in menu.
*/
{
  if (!pbary->colSpIn.menuSetChan) {
    if (im->colorspace == XYZColorspace)
      pbary->colSpIn.chans = chXYZ;
    else if (im->colorspace == xyYColorspace)
      pbary->colSpIn.chans = chXYY;
    else pbary->colSpIn.chans = chRGB;
  }

  if (!pbary->colSpIn.menuSetPrim && ImageHasChromaticity (im)) {
    // These might not be set in im->chromaticity.
    pbary->colSpIn.triMain.primA.x = im->chromaticity.red_primary.x;
    pbary->colSpIn.triMain.primA.y = im->chromaticity.red_primary.y;
    pbary->colSpIn.triMain.primB.x = im->chromaticity.green_primary.x;
    pbary->colSpIn.triMain.primB.y = im->chromaticity.green_primary.y;
    pbary->colSpIn.triMain.primC.x = im->chromaticity.blue_primary.x;
    pbary->colSpIn.triMain.primC.y = im->chromaticity.blue_primary.y;

    if (!pbary->colSpIn.menuSetWP) {
      pbary->colSpIn.WP.x = im->chromaticity.white_point.x;
      pbary->colSpIn.WP.y = im->chromaticity.white_point.y;
    }

    if (!pbary->colSpIn.menuSetGam) {
      // FIXME: and gamma
    }
    // FIXME: and XYZwp
  }
}

static void ColSpInToOut (barymT * pbary)
/* Copy input metadata to output metadata _only_ where these were
   not set in menu.
*/
{
  if (!pbary->colSpOut.menuSetChan) {
    pbary->colSpOut.chans = pbary->colSpIn.chans;
  }
  if (!pbary->colSpOut.menuSetPrim) {
    pbary->colSpOut.triMain.primA.x = pbary->colSpIn.triMain.primA.x;
    pbary->colSpOut.triMain.primA.y = pbary->colSpIn.triMain.primA.y;
    pbary->colSpOut.triMain.primB.x = pbary->colSpIn.triMain.primB.x;
    pbary->colSpOut.triMain.primB.y = pbary->colSpIn.triMain.primB.y;
    pbary->colSpOut.triMain.primC.x = pbary->colSpIn.triMain.primC.x;
    pbary->colSpOut.triMain.primC.y = pbary->colSpIn.triMain.primC.y;

    if (!pbary->colSpOut.menuSetWP) {
      pbary->colSpOut.WP.x = pbary->colSpIn.WP.x;
      pbary->colSpOut.WP.y = pbary->colSpIn.WP.y;
    }
    if (!pbary->colSpOut.menuSetGam) {
      pbary->colSpOut.gamma = pbary->colSpIn.gamma;
    }
  }
}

static void ColSpToIm (
  barymT * pbary,
  Image * im)
{
  switch (pbary->colSpOut.chans) {
    case chRGB: im->colorspace = RGBColorspace; break;
    case chXYY: im->colorspace = xyYColorspace; break;
    case chXYZ: im->colorspace = XYZColorspace; break;
  }

  im->chromaticity.red_primary.x = pbary->colSpOut.triMain.primA.x;
  im->chromaticity.red_primary.y = pbary->colSpOut.triMain.primA.y;
  im->chromaticity.green_primary.x = pbary->colSpOut.triMain.primB.x;
  im->chromaticity.green_primary.y = pbary->colSpOut.triMain.primB.y;
  im->chromaticity.blue_primary.x = pbary->colSpOut.triMain.primC.x;
  im->chromaticity.blue_primary.y = pbary->colSpOut.triMain.primC.y;

  im->chromaticity.white_point.x = pbary->colSpOut.WP.x;
  im->chromaticity.white_point.y = pbary->colSpOut.WP.y;

  // FIXME: and gamma. Save negative values in attribute?
}


//--------------------------------------------


static void usage (void)
{
  printf ("Usage: -process 'barymap [OPTION]...'\n");
  printf ("process barycentric coordinates.\n");
  printf ("\n");
  printf ("  ic,  inChannels string   RGB or xyY or XYZ [xyY]\n");
  printf ("  ip,  inPrim 6_floats     xy coords of input primaries [sRGB]\n");
  printf ("  iw,  inWP 2_floats       xy coords of input white point\n");
  printf ("  it,  inTransfer number   name or gamma of input\n");
  printf ("  xw,  XYZwp 2_floats      xy coords of XYZ white point\n");
  printf ("  gn2, gain2 number        gain factor for squared distance from WP [0]\n");
  printf ("  gn,  gain number         gain factor for distance from WP [1]\n");
  printf ("  bs,  bias number         bias for distance from WP [0]\n");
  printf ("  pow, power number        power factor for distance from WP [1]\n");
  printf ("  am,  angleMult string    list of: direction,width,factor\n");
  printf ("  st,  skipTri             skip triangle processing\n");
  printf ("  ign, ignoreWP            ignore white point for triangulation\n");
  printf ("  clC, clampCartesian      clamp cartesian coordinates\n");
  printf ("  clB, clampBarycentric    clamp barycentric coordinates\n");
  printf ("  oc,  outChannels string  RGB or xyY or XYZ [xyY]\n");
  printf ("  op,  outPrim 6_floats    xy coords of output primaries [sRGB]\n");
  printf ("  ow,  outWP 2_floats      xy coords of output white point\n");
  printf ("  ot,  outTransfer number  name or gamma of output\n");
  printf ("  ca,  chromAdap string    name of Chromatic Adaptation Transform [Bradford]\n");
  printf ("  r16, round16             round some fractions to 16 binary places\n");
  printf ("  li,  list string         write list to stdout or stderr\n");
  printf ("  f,   file string         write to file stream stdout or stderr\n");
  printf ("  v,   verbose             write text information to stdout\n");
  printf ("  v2,  verbose2            write more text information\n");
  printf ("  v9,  verbose9            write voluminous text information\n");
  printf ("       version             write version information to stdout\n");

/*
  printf ("\nNamed primaries:\n");
  ListColSp (stdout);
  printf ("\nNamed illuminants:\n");
  ListWP (stdout);
  printf ("\nNamed transfers:\n");
  ListTrans (stdout);
  printf ("\nChromatic Adaptation Transforms:\n");
  ListCats (stdout);
  printf ("\n");
*/
}


static MagickBooleanType ParseCoord (const char * s, coordsT * pc)
{
  double d;
  int n;
  char * p = (char *)s;

  sscanf (p, "%lg%n", &d, &n);
  if (!n) return MagickFalse;
  p += n;
  if (*p != ',') return MagickFalse;
  p++;
  pc->x = d;

  sscanf (p, "%lg%n", &d, &n);
  if (!n) return MagickFalse;
  p += n;
  if (*p != '\0') return MagickFalse;
  pc->y = d;

  return MagickTrue;
}


static MagickBooleanType Parse3Coord (
  const char * s,
  coordsT * pc0,
  coordsT * pc1,
  coordsT * pc2)
{
  double d;
  int n;
  char * p = (char *)s;

  sscanf (p, "%lg%n", &d, &n);
  if (!n) return MagickFalse;
  p += n;
  if (*p != ',') return MagickFalse;
  p++;
  pc0->x = d;

  sscanf (p, "%lg%n", &d, &n);
  if (!n) return MagickFalse;
  p += n;
  if (*p != ',') return MagickFalse;
  p++;
  pc0->y = d;

  sscanf (p, "%lg%n", &d, &n);
  if (!n) return MagickFalse;
  p += n;
  if (*p != ',') return MagickFalse;
  p++;
  pc1->x = d;

  sscanf (p, "%lg%n", &d, &n);
  if (!n) return MagickFalse;
  p += n;
  if (*p != ',') return MagickFalse;
  p++;
  pc1->y = d;

  sscanf (p, "%lg%n", &d, &n);
  if (!n) return MagickFalse;
  p += n;
  if (*p != ',') return MagickFalse;
  p++;
  pc2->x = d;

  sscanf (p, "%lg%n", &d, &n);
  if (!n) return MagickFalse;
  p += n;
  if (*p != '\0') return MagickFalse;
  pc2->y = d;

  return MagickTrue;
}

static MagickBooleanType ParseGamma (
  barymT * pbary,
  const char * s,
  colSpT * pcs)
{
  ZeroTrans (&pcs->trans);

  int n;
  char * p = (char *)s;

  pcs->gamma = 0;

  sscanf (p, "%lg%n", &pcs->gamma, &n);
  if (!n) return MagickFalse;
  p += n;
  if (*p != '\0') return MagickFalse;

  if (pcs->gamma != 0) return MagickTrue;

  if (ParseNamedTrc (pbary, s, pcs)) {
    CalcTrans (pbary, &pcs->trans, -(int)floor(pcs->gamma));
    return MagickTrue;
  }

  return MagickFalse;
}

static MagickBooleanType Parse3Prims (
  barymT * pbary,
  const char * s,
  colSpT * pcs)
{
  if (Parse3Coord (s, &pcs->triMain.primA, &pcs->triMain.primB, &pcs->triMain.primC))
    return MagickTrue;

  if (ParseNamedPrims (pbary, s, pcs)) {
    ParseGamma (pbary, s, pcs);
    // (If not successful, no problem.)
    return MagickTrue;
  }

  return MagickFalse;
}

static MagickBooleanType ParseWPkelvin (
  barymT * pbary,
  const char * s,
  coordsT * pwp)
{
  // Calculate x,y from correlated colour temperature.

  int slen = strlen(s);
  if (slen < 2) return MagickFalse;
  if (toupper(s[slen-1]) != 'K') return MagickFalse;
  if (!isdigit((int)s[0])) return MagickFalse;

  char * p = (char *)s;

  double T = 0;

  int n = 0;
  sscanf (p, "%lg%n", &T, &n);
  if (!n) return MagickFalse;
  p += n+1;  // +1 to skip over terminal K.
  if (*p != '\0') return MagickFalse;

  // Adjust from nominal.
  // See https://vdocuments.mx/colorimetry-cie-colorimetry.html

  if (s[slen-1]=='k') T *= 1.4388/1.4380;

  // Formulae from https://en.wikipedia.org/wiki/Standard_illuminant#Illuminant_series_D
  // Also in http://www.brucelindbloom.com/index.html?Eqn_T_to_xy.html

  if (T < 4000 || T > 25000) return MagickFalse;

  if (T <= 7000) {
    pwp->x = 0.244063 + 0.09911e3/T + 2.9678e6/(T*T) - 4.6070e9 / (T*T*T);
  } else {
    pwp->x = 0.237040 + 0.24748e3/T + 1.9018e6/(T*T) - 2.0064e9 / (T*T*T);
  }
  pwp->y = -3.000*pwp->x*pwp->x + 2.870*pwp->x - 0.275;

  if (pbary->verbose >= 2) fprintf (pbary->fh_data, "ParseWPkelvin3 T=%g  xy=%g,%g\n", T, pwp->x, pwp->y);

  return MagickTrue;
}

static MagickBooleanType ParseWP (
  barymT * pbary,
  const char * s,
  coordsT * pwp)
{
  MagickBooleanType okay = ParseWPkelvin (pbary, s, pwp);

  if (!okay) okay = ParseCoord (s, pwp);

  if (!okay) okay = ParseNamedWP (pbary, s, pwp);

  if (okay) Round16xy (pbary, pwp);

  return okay;
}

static MagickBooleanType ParseAngleMult (barymT * pbary, const char * s)
{
  double d;
  int n;
  char * p = (char *)s;

  int i = 0;
  AngMultT *pam = &pbary->AngMults[0];
  MagickBooleanType Finished = MagickFalse;
  while (i < MaxAngMults && !Finished) {

    sscanf (p, "%lg%n", &d, &n);
    if (!n) return MagickFalse;
    p += n;
    if (*p != ',') {
      printf ("AngMult: expected ',' at %s\n", p);
      return MagickFalse;
    }
    p++;
    while (d >= 180.0) d -= 360.0;
    while (d < -180.0) d += 360.0;
    pam->direction = d * M_PI/180.0;

    sscanf (p, "%lg%n", &d, &n);
    if (!n) return MagickFalse;
    p += n;
    if (*p != ',') {
      printf ("AngMult: expected ',' at %s\n", p);
      return MagickFalse;
    }
    p++;
    pam->width = d * M_PI/180.0;
    pam->halfWidth = pam->width / 2.0;
    pam->minDir = pam->direction - pam->halfWidth;
    pam->maxDir = pam->direction + pam->halfWidth;

    sscanf (p, "%lg%n", &d, &n);
    if (!n) return MagickFalse;
    p += n;
    pam->factor = d;
    if (*p == '\0') Finished = MagickTrue;
    else {
      if (*p != ';') {
        printf ("AngMult: expected ';' at %s\n", p);
        return MagickFalse;
      }
      p++;
    }
    pam++;
    i++;
  }

  if (!Finished) {
    printf ("AngMult: too many\n");
  }
  pbary->nAngMults = i;
  return Finished;
}

static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "barymap: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  barymT * pbary
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status = MagickTrue;

  pbary->verbose = 0;
  pbary->fh_data = stderr;

  pbary->colSpIn.chans = chXYY;
  pbary->colSpOut.chans = chXYY;
  pbary->ndxCat = dfltCat;
  pbary->ignoreWP = MagickFalse;
  pbary->clampXy = MagickFalse;
  pbary->clampBary = MagickFalse;
  pbary->doTriangles = MagickTrue;
  pbary->round16 = MagickFalse;
  pbary->gain2 = 0.0;
  pbary->gain = 1.0;
  pbary->bias = 0.0;
  pbary->power = 1.0;
  pbary->nAngMults = 0;
  pbary->colSpIn.menuSetChan =
    pbary->colSpIn.menuSetPrim =
    pbary->colSpIn.menuSetWP =
    pbary->colSpIn.menuSetGam =
    pbary->colSpIn.menuSetXYZwp = MagickFalse;
  pbary->colSpOut.menuSetChan =
    pbary->colSpOut.menuSetPrim =
    pbary->colSpOut.menuSetWP =
    pbary->colSpOut.menuSetGam =
    pbary->colSpOut.menuSetXYZwp = MagickFalse;
  ParseNamedPrims (pbary, "sRGB", &pbary->colSpIn);
  ParseNamedPrims (pbary, "sRGB", &pbary->colSpOut);

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "ic", "inChannels")==MagickTrue) {
      NEXTARG;
      pbary->colSpIn.menuSetChan = MagickTrue;
      if (strcasecmp (argv[i], "rgb")==0) pbary->colSpIn.chans = chRGB;
      else if (strcasecmp (argv[i], "xyy")==0) pbary->colSpIn.chans = chXYY;
      else if (strcasecmp (argv[i], "xyz")==0) pbary->colSpIn.chans = chXYZ;
      else status = MagickFalse;
    } else if (IsArg (pa, "oc", "outChannels")==MagickTrue) {
      NEXTARG;
      pbary->colSpOut.menuSetChan = MagickTrue;
      if (strcasecmp (argv[i], "rgb")==0) pbary->colSpOut.chans = chRGB;
      else if (strcasecmp (argv[i], "xyy")==0) pbary->colSpOut.chans = chXYY;
      else if (strcasecmp (argv[i], "xyz")==0) pbary->colSpOut.chans = chXYZ;
      else status = MagickFalse;
    } else if (IsArg (pa, "st", "skipTri")==MagickTrue) {
      pbary->doTriangles = MagickFalse;
    } else if (IsArg (pa, "ign", "ignoreWP")==MagickTrue) {
      pbary->ignoreWP = MagickTrue;
    } else if (IsArg (pa, "clC", "clampCartesian")==MagickTrue) {
      pbary->clampXy = MagickTrue;
    } else if (IsArg (pa, "clB", "clampBarycentric")==MagickTrue) {
      pbary->clampBary = MagickTrue;
    } else if (IsArg (pa, "gn", "gain")==MagickTrue) {
      NEXTARG;
      pbary->gain = atof(argv[i]);
    } else if (IsArg (pa, "gn2", "gain2")==MagickTrue) {
      NEXTARG;
      pbary->gain2 = atof(argv[i]);
    } else if (IsArg (pa, "bs", "bias")==MagickTrue) {
      NEXTARG;
      pbary->bias = atof(argv[i]);
    } else if (IsArg (pa, "pow", "power")==MagickTrue) {
      NEXTARG;
      pbary->power = atof(argv[i]);
    } else if (IsArg (pa, "ip", "inPrim")==MagickTrue) {
      NEXTARG;
      pbary->colSpIn.menuSetPrim = MagickTrue;
      if (!Parse3Prims (pbary, argv[i], &pbary->colSpIn))
      {
        fprintf (stderr, "barymap: problem parsing inPrim\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "iw", "inWP")==MagickTrue) {
      NEXTARG;
      pbary->colSpIn.menuSetWP = MagickTrue;
      if (!ParseWP (pbary, argv[i], &pbary->colSpIn.WP)) {
        fprintf (stderr, "barymap: problem parsing inWP\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "xw", "XYZwp")==MagickTrue) {
      NEXTARG;
      pbary->colSpIn.menuSetXYZwp = MagickTrue;
      pbary->colSpOut.menuSetXYZwp = MagickTrue;
      if (!ParseWP (pbary, argv[i], &pbary->colSpIn.XYZwp)) {
        fprintf (stderr, "barymap: problem parsing XYZwp\n");
        status = MagickFalse;
      }
      else pbary->colSpOut.XYZwp = pbary->colSpIn.XYZwp;
    } else if (IsArg (pa, "it", "inTransfer")==MagickTrue) {
      NEXTARG;
      pbary->colSpIn.menuSetGam = MagickTrue;
      if (!ParseGamma (pbary, argv[i], &pbary->colSpIn)) {
        fprintf (stderr, "barymap: problem parsing inTransfer\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "op", "outPrim")==MagickTrue) {
      NEXTARG;
      pbary->colSpOut.menuSetPrim = MagickTrue;
      if (!Parse3Prims (pbary, argv[i], &pbary->colSpOut))
      {
        fprintf (stderr, "barymap: problem parsing outPrim\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "ow", "outWP")==MagickTrue) {
      NEXTARG;
      pbary->colSpOut.menuSetWP = MagickTrue;
      if (!ParseWP (pbary, argv[i], &pbary->colSpOut.WP)) {
        fprintf (stderr, "barymap: problem parsing outWP\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "ot", "outTransfer")==MagickTrue) {
      NEXTARG;
      pbary->colSpOut.menuSetGam = MagickTrue;
      if (!ParseGamma (pbary, argv[i], &pbary->colSpOut)) {
        fprintf (stderr, "barymap: problem parsing outTransfer\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "ca", "chromAdap")==MagickTrue) {
      NEXTARG;
      if (!ParseNamedCat (pbary, argv[i])) {
        fprintf (stderr, "barymap: problem parsing chromAdap\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "am", "angleMult")==MagickTrue) {
      NEXTARG;
      if (!ParseAngleMult (pbary, argv[i])) {
        fprintf (stderr, "barymap: problem parsing angleMult\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "r16", "round16")==MagickTrue) {
      pbary->round16 = MagickTrue;
    } else if (IsArg (pa, "li", "list")==MagickTrue) {
      NEXTARG;
      if (strcasecmp (argv[i], "list")==0) {
        fprintf (pbary->fh_data, "list\nprimaries\nilluminants\ntransfers\ncats\nwpnums\n");
      } else if (strcasecmp (argv[i], "primaries")==0) {
        ListColSp (pbary->fh_data);
      } else if (strcasecmp (argv[i], "illuminants")==0) {
        ListWP (pbary->fh_data);
      } else if (strcasecmp (argv[i], "transfers")==0) {
        ListTrans (pbary->fh_data);
      } else if (strcasecmp (argv[i], "cats")==0) {
        ListCats (pbary->fh_data);
      } else if (strcasecmp (argv[i], "wpnums")==0) {
        WrWpNums (pbary->fh_data, pbary);
      } else {
        fprintf (stderr, "barymap: ERROR: unknown option '%s' for list\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      NEXTARG;
      if (strcasecmp (argv[i], "stdout")==0) pbary->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) pbary->fh_data = stderr;
      else {
        fprintf (stderr, "barymap: ERROR: unknown option '%s' for file\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pbary->verbose = 1;
    } else if (IsArg (pa, "v2", "verbose2")==MagickTrue) {
      pbary->verbose = 2;
    } else if (IsArg (pa, "v9", "verbose9")==MagickTrue) {
      pbary->verbose = 9;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "barymap: ERROR: unknown option '%s'\n", pa);
      status = MagickFalse;
    }
  }

  if (!pbary->colSpIn.menuSetXYZwp)
    pbary->colSpIn.XYZwp = pbary->colSpIn.WP;
  if (!pbary->colSpOut.menuSetXYZwp)
    pbary->colSpOut.XYZwp = pbary->colSpIn.XYZwp;

  Round16xy (pbary, &pbary->colSpIn.WP);
  Round16xy (pbary, &pbary->colSpOut.WP);
  Round16xy (pbary, &pbary->colSpIn.XYZwp);
  Round16xy (pbary, &pbary->colSpOut.XYZwp);

  if (pbary->verbose) {
    fprintf (stderr, "barymap options:");

    if (pbary->colSpIn.menuSetChan) {
      fprintf (stderr, "  inChannels ");
      switch (pbary->colSpIn.chans) {
        case chRGB: fprintf (stderr, "RGB"); break;
        case chXYY: fprintf (stderr, "xyY "); break;
        case chXYZ: fprintf (stderr, "XYZ "); break;
      }
    }

    if (pbary->colSpIn.menuSetChan) {
      fprintf (stderr, "  outChannels ");
      switch (pbary->colSpOut.chans) {
        case chRGB: fprintf (stderr, "RGB"); break;
        case chXYY: fprintf (stderr, "xyY "); break;
        case chXYZ: fprintf (stderr, "XYZ "); break;
      }
    }

    if (!pbary->doTriangles) fprintf (stderr, "  skipTri");
    if (pbary->ignoreWP) fprintf (stderr, "  ignoreWP");
    if (pbary->clampXy) fprintf (stderr, "  clampCartesian");
    if (pbary->clampBary) fprintf (stderr, "  clampBarycentric");
    if (pbary->gain2 != 0) fprintf (stderr, "  gain2 %g", pbary->gain2);
    if (pbary->gain != 1) fprintf (stderr, "  gain %g", pbary->gain);
    if (pbary->bias != 0) fprintf (stderr, "  bias %g", pbary->bias);
    if (pbary->power != 1) fprintf (stderr, "  power %g", pbary->power);

    fprintf (stderr, "  chromAdap %s", cats[pbary->ndxCat].name);

    if (pbary->round16) fprintf (stderr, "  round16");
    if (pbary->verbose >= 2) fprintf (stderr, "  verbose%i", pbary->verbose);
    else if (pbary->verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");

    WrBarymap (pbary, stderr);  // FIXME: Prefer different words?
  }

  if (pbary->verbose) {
    fprintf (stderr, "nAngMults %i\n", pbary->nAngMults);
    int i;
    for (i=0; i < pbary->nAngMults; i++) {
      AngMultT *pam = &pbary->AngMults[i];
      fprintf (stderr, "%g %g %g %g %g\n",
        pam->direction, pam->minDir, pam->maxDir, pam->halfWidth, pam->factor);
    }
  }

  if (status && pbary->verbose) {
    // FIXME: temporary, for testing.
    if (pbary->verbose >= 2) {
      WrWpNums (pbary->fh_data, pbary);
      //TstInvMat3x3 (pbary);
    }
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}

static MagickBooleanType CalcTri (triangleT *ptri)
{
  ptri->y23 = ptri->primB.y - ptri->primC.y;
  ptri->x13 = ptri->primA.x - ptri->primC.x;
  ptri->x32 = ptri->primC.x - ptri->primB.x;
  ptri->y13 = ptri->primA.y - ptri->primC.y;
  ptri->y31 = -ptri->y13;
  ptri->x3 = ptri->primC.x;
  ptri->y3 = ptri->primC.y;

  ptri->det = ptri->y23 * ptri->x13 + ptri->x32 * ptri->y13;

  // FIXME: error if det is zero.
  if (fabs(ptri->det) < 1e-6) {
    printf ("ERROR: ptri->det is %g\n", ptri->det);
    return MagickFalse;
  }
  return MagickTrue;
}

static MagickBooleanType PopSubTrisInOut (colSpT * cs)
{
  MagickBooleanType okay = MagickTrue;

  if (!CalcTri (&cs->triMain)) okay = MagickFalse;
  cs->subTris[0] = cs->triMain;

  cs->subTris[1].primA = cs->WP;
  cs->subTris[1].primB = cs->triMain.primA;
  cs->subTris[1].primC = cs->triMain.primB;
  if (!CalcTri (&cs->subTris[1])) okay = MagickFalse;

  cs->subTris[2].primA = cs->WP;
  cs->subTris[2].primB = cs->triMain.primB;
  cs->subTris[2].primC = cs->triMain.primC;
  if (!CalcTri (&cs->subTris[2])) okay = MagickFalse;

  cs->subTris[3].primA = cs->WP;
  cs->subTris[3].primB = cs->triMain.primC;
  cs->subTris[3].primC = cs->triMain.primA;
  if (!CalcTri (&cs->subTris[3])) okay = MagickFalse;

  return okay;
}

static MagickBooleanType PopSubTris (barymT * pbary)
{
  MagickBooleanType okay = MagickTrue;

  if (!PopSubTrisInOut (&pbary->colSpIn)) okay = MagickFalse;
  if (!PopSubTrisInOut (&pbary->colSpOut)) okay = MagickFalse;

  return okay;
}

static void inline CartToBary (
  const coordsT * xy,
  baryCoordsT * cb,
  const triangleT * ptri)
{
  // Assumes det is not zero.
  double dx = xy->x - ptri->x3;
  double dy = xy->y - ptri->y3;
  cb->b1 = (ptri->y23 * dx + ptri->x32 * dy) / ptri->det;
  cb->b2 = (ptri->y31 * dx + ptri->x13 * dy) / ptri->det;
  cb->b3 = 1 - cb->b1 - cb->b2;
}

static void inline ClampCart (coordsT * xy)
{
  if      (xy->x < 0) xy->x = 0;
  else if (xy->x > 1) xy->x = 1;

  if      (xy->y < 0) xy->y = 0;
  else if (xy->y > 1) xy->y = 1;
}

static void inline ClampBary (baryCoordsT * cb)
{
  double div;
  if (cb->b1 < 0) {
    cb->b1 = 0;
    div = cb->b2 + cb->b3;
    cb->b2 /= div;
    cb->b3 /= div;
  }
  if (cb->b2 < 0) {
    cb->b2 = 0;
    div = cb->b1 + cb->b3;
    cb->b1 /= div;
    cb->b3 /= div;
  }
  if (cb->b3 < 0) {
    cb->b3 = 0;
    div = cb->b1 + cb->b2;
    cb->b1 /= div;
    cb->b2 /= div;
  }

  // FIXME: Next can never happen?
  if (cb->b1 > 1) cb->b1 = 1;
  if (cb->b2 > 1) cb->b2 = 1;
  if (cb->b3 > 1) cb->b3 = 1;
}

static void inline BaryToCart (
  const baryCoordsT * cb,
  coordsT * xy,
  const triangleT * ptri)
{
  xy->x = cb->b1 * ptri->primA.x + cb->b2 * ptri->primB.x + cb->b3 * ptri->primC.x;
  xy->y = cb->b1 * ptri->primA.y + cb->b2 * ptri->primB.y + cb->b3 * ptri->primC.y;
}

static MagickBooleanType inline IsInside (baryCoordsT * cb)
{
  return (cb->b1 >= 0) && (cb->b2 >= 0) && (cb->b3 >= 0);
}

static void WhichTri (
  barymT * pbary,
  colSpT * cs,
  coordsT * xy,
  baryCoordsT * cb,
  int * triNum)
{
  // Finds which triangle the cartesian coordinate is in.
  // Returns those barycentric coords in cb,
  // and triangle number in triNum.

  baryCoordsT cbs[4];

  if (pbary->ignoreWP) {
    CartToBary (xy, &cbs[0], &cs->subTris[0]);
    *triNum = 0;
  } else {
    *triNum = 0;  // Means we don't know yet.
    CartToBary (xy, &cbs[1], &cs->subTris[1]);
    if (IsInside (&cbs[1])) {
      *triNum = 1;
    } else {

      CartToBary (xy, &cbs[2], &cs->subTris[2]);
      if (IsInside (&cbs[2])) {
        *triNum = 2;
      } else {

        CartToBary (xy, &cbs[3], &cs->subTris[3]);
        if (IsInside (&cbs[3])) {
          *triNum = 3;
        } else {
          // Outside all triangles.
          // Fall back to main triangle.
          CartToBary (xy, &cbs[0], &cs->subTris[0]);
          *triNum = 0;
        }
      }
    }
  }
  *cb = cbs[*triNum];
}


// The next function takes an image, and returns it.
//
static Image * barymap (
  Image *image,
  barymT * pbary,
  ExceptionInfo *exception)
{
  // This function ignores alpha.

  // When in and out are xyY,
  // this function reads and writes the R and G channels.
  // leaving the B channel unchanged.

  MagickBooleanType
    status = MagickTrue;

  ColSpFromIm (image, pbary);
  ColSpInToOut (pbary);

  if (pbary->verbose >= 2) {
    fprintf (pbary->fh_data, "For this image:\n");
    WrBarymap (pbary, stderr);
  }

  ZeroTrans (&pbary->colSpIn.trans);
  ZeroTrans (&pbary->colSpOut.trans);
  if (pbary->colSpIn.gamma < 0) {
    CalcTrans (pbary, &pbary->colSpIn.trans, -(int)floor(pbary->colSpIn.gamma));
  }
  if (pbary->colSpOut.gamma < 0) {
    CalcTrans (pbary, &pbary->colSpOut.trans, -(int)floor(pbary->colSpOut.gamma));
  }

  if (!RgbXyzMat (pbary, &pbary->colSpIn)) {
    printf ("Problem in RgbXyzMat in\n");
    status = MagickFalse;
  }

  if (!RgbXyzMat (pbary, &pbary->colSpOut)) {
    printf ("Problem in RgbXyzMat out\n");
    status = MagickFalse;
  }

  if (!ChromAdap (pbary)) {
    printf ("Problem in ChromAdap\n");
    status = MagickFalse;
  }

  MagickBooleanType
    doGBP = (pbary->gain2 != 0.0) ||
            (pbary->gain  != 1.0) ||
            (pbary->bias  != 0.0) ||
            (pbary->power != 1.0);

  if (!PopSubTris (pbary)) {
    printf ("Problem in PopSubTris\n");
    status = MagickFalse;
  }

  CacheView * image_view = AcquireAuthenticCacheView (image,exception);

  if (pbary->verbose == 9)
    fprintf (pbary->fh_data, "Pixels:");

  ssize_t y;
#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) \
    shared(okay) \
    MAGICK_THREADS(image,image,image->rows,1)
#endif
  for (y = 0; y < image->rows; y++) {
    VIEW_PIX_PTR *p;

    if (status == MagickFalse)
      continue;

    p=GetCacheViewAuthenticPixels(image_view,0,y,image->columns,1,exception);
    if (!p) {
      status=MagickFalse;
      continue;
    }

    ssize_t x;
    RGBT rgb;
    double div;
    for (x = 0; x < image->columns; x++) {
      if (pbary->verbose == 9)
        fprintf (pbary->fh_data, "%li,%li:", x, y);

      coordsT xy = {0,0};
      XYZT XYZ = {0,0,0};

      switch (pbary->colSpIn.chans) {
        case chRGB:
          rgb.R = GET_PIXEL_RED   (image, p) / (MagickRealType)QuantumRange;
          rgb.G = GET_PIXEL_GREEN (image, p) / (MagickRealType)QuantumRange;
          rgb.B = GET_PIXEL_BLUE  (image, p) / (MagickRealType)QuantumRange;

          if (pbary->verbose == 9)
            fprintf (pbary->fh_data, "RGB: %.*g, %.*g, %.*g\n",
              pbary->precision, rgb.R,
              pbary->precision, rgb.G,
              pbary->precision, rgb.B);

          Rgb2Xyz (&pbary->colSpIn, &rgb, &XYZ);

          if (pbary->verbose == 9)
            fprintf (pbary->fh_data, "XYZ: %.*g, %.*g, %.*g\n",
              pbary->precision, XYZ.X,
              pbary->precision, XYZ.Y,
              pbary->precision, XYZ.Z);

          div = XYZ.X + XYZ.Y + XYZ.Z;
          if (div == 0) {
            xy.x = 0;
            xy.y = 0;
          } else {
            xy.x = XYZ.X / div;
            xy.y = XYZ.Y / div;
          }
          break;
        case chXYY:
          xy.x = GET_PIXEL_RED   (image, p) / (MagickRealType)QuantumRange;
          xy.y = GET_PIXEL_GREEN (image, p) / (MagickRealType)QuantumRange;
          XYZ.Y = GET_PIXEL_BLUE (image, p) / (MagickRealType)QuantumRange;
          break;
        case chXYZ:
          XYZ.X = GET_PIXEL_RED   (image, p) / (MagickRealType)QuantumRange;
          XYZ.Y = GET_PIXEL_GREEN (image, p) / (MagickRealType)QuantumRange;
          XYZ.Z = GET_PIXEL_BLUE (image, p) / (MagickRealType)QuantumRange;

          if (pbary->verbose == 9)
            fprintf (pbary->fh_data, "XYZ: %.*g, %.*g, %.*g\n",
              pbary->precision, XYZ.X,
              pbary->precision, XYZ.Y,
              pbary->precision, XYZ.Z);

          div = XYZ.X + XYZ.Y + XYZ.Z;
          if (div == 0) {
            xy.x = 0;
            xy.y = 0;
          } else {
            xy.x = XYZ.X / div;
            xy.y = XYZ.Y / div;
          }
          break;
      }

      if (pbary->clampXy) ClampCart (&xy);

      if (pbary->verbose == 9)
        fprintf (pbary->fh_data, "xy: %.*g, %.*g\n",
          pbary->precision, xy.x,
          pbary->precision, xy.y);

      if (doGBP) {
        double dx = xy.x - pbary->colSpIn.WP.x;
        double dy = xy.y - pbary->colSpIn.WP.y;
        double v = hypot (dx, dy);

        double vrat = pow (v * v * pbary->gain2
                             + v * pbary->gain
                             + pbary->bias,
                           pbary->power)
                      / v;

        xy.x = dx * vrat + pbary->colSpIn.WP.x;
        xy.y = dy * vrat + pbary->colSpIn.WP.y;

        if (pbary->clampXy) ClampCart (&xy);

        if (pbary->verbose == 9)
          fprintf (pbary->fh_data, "  xy: %.*g, %.*g\n",
            pbary->precision, xy.x,
            pbary->precision, xy.y);
      }

      if (pbary->nAngMults > 0) {
        double dx = xy.x - pbary->colSpIn.WP.x;
        double dy = xy.y - pbary->colSpIn.WP.y;
        double t = atan2 (dx, dy);
        int i;
        for (i = 0; i < pbary->nAngMults; i++) {
          AngMultT * pam = &pbary->AngMults[i];
          double f = fabs (t - pam->direction);

          if (f > pam->halfWidth) {
            f = fabs (t - M_TWOPI - pam->direction);
          }
          if (f < pam->halfWidth) {
            f = (f / pam->halfWidth);
            f = sin ((f-0.5) * M_PI);
            f = f * (1 - pam->factor) + pam->factor;
            xy.x = dx * f + pbary->colSpIn.WP.x;
            xy.y = dy * f + pbary->colSpIn.WP.y;

            if (pbary->clampXy) ClampCart (&xy);
          }
        }
        if (pbary->verbose == 9)
          fprintf (pbary->fh_data, "  xy: %.*g, %.*g\n",
            pbary->precision, xy.x,
            pbary->precision, xy.y);
      }

      if (pbary->doTriangles) {
        baryCoordsT cb;

        int triNum;
        WhichTri (pbary, &pbary->colSpIn, &xy, &cb, &triNum);
        if (pbary->clampBary) ClampBary (&cb);

        if (pbary->verbose == 9)
          fprintf (pbary->fh_data, "  t=%i  b: %.*g, %.*g, %.*g\n",
            triNum,
            pbary->precision, cb.b1,
            pbary->precision, cb.b2,
            pbary->precision, cb.b3);

        BaryToCart (&cb, &xy, &pbary->colSpOut.subTris[triNum]);
        if (pbary->clampXy) ClampCart (&xy);

        if (pbary->verbose == 9)
          fprintf (pbary->fh_data, "  xy: %.*g, %.*g\n",
            pbary->precision, xy.x,
            pbary->precision, xy.y);
      }

      switch (pbary->colSpOut.chans) {
        case chRGB:
          xy2XYZ (&xy, &XYZ);

          if (pbary->verbose == 9)
            fprintf (pbary->fh_data, "  XYZ: %.*g, %.*g, %.*g\n",
              pbary->precision, XYZ.X,
              pbary->precision, XYZ.Y,
              pbary->precision, XYZ.Z);

//          ChromAdapXYZ (pbary, &XYZ);

          if (pbary->verbose == 9)
            fprintf (pbary->fh_data, "  aca XYZ: %.*g, %.*g, %.*g\n",
              pbary->precision, XYZ.X,
              pbary->precision, XYZ.Y,
              pbary->precision, XYZ.Z);

          Xyz2Rgb (&pbary->colSpOut, &XYZ, &rgb);

          if (pbary->verbose == 9)
            fprintf (pbary->fh_data, "  RGB: %.*g, %.*g, %.*g\n",
              pbary->precision, rgb.R,
              pbary->precision, rgb.G,
              pbary->precision, rgb.B);

          SET_PIXEL_RED   (image, rgb.R * QuantumRange, p);
          SET_PIXEL_GREEN (image, rgb.G * QuantumRange, p);
          SET_PIXEL_BLUE  (image, rgb.B * QuantumRange, p);
          break;
        case chXYY:
          SET_PIXEL_RED   (image, xy.x * QuantumRange, p);
          SET_PIXEL_GREEN (image, xy.y * QuantumRange, p);
          SET_PIXEL_BLUE  (image, XYZ.Y * QuantumRange, p);
          break;
        case chXYZ:
          xy2XYZ (&xy, &XYZ);
          // We need CA, even if we stay in XYZ.
          // FIXME: this is the simple version, directly from in to out.
          ChromAdapXYZ (pbary, &XYZ);
          SET_PIXEL_RED   (image, XYZ.X * QuantumRange, p);
          SET_PIXEL_GREEN (image, XYZ.Y * QuantumRange, p);
          SET_PIXEL_BLUE  (image, XYZ.Z * QuantumRange, p);
          break;
      }

      p += Inc_ViewPixPtr (image);
    }
    if (!SyncCacheViewAuthenticPixels(image_view,exception)) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }
  }

  if (!status) return NULL;

  image_view = DestroyCacheView (image_view);

  ColSpToIm (pbary, image);

  return image;
}



ModuleExport size_t barymapImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  MagickBooleanType
    status;

  barymT
    pbary;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  pbary.precision = GetMagickPrecision();

  status = menu (argc, argv, &pbary);

  if (status == MagickFalse)
    return (-1);

  Image * image;
  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    Image * newImg = barymap (image, &pbary, exception);
    if (!newImg) return (-1);

    if (newImg != image) {
      ReplaceImageInList (&image, newImg);
      *images=GetFirstImageInList (newImg);
    }
  }

  return(MagickImageFilterSignature);
}
