/* Calculate the mean and standard deviation
   of the three colour channels,
   of an image or a VIEW_PIX_PTR.

   Updated:
     3-April-2018 for v7.0.7-28
*/


#ifndef CALCMNSD_INC
#define CALCMNSD_INC 1

typedef struct {
  // Returns values in range 0.0 to 1.0.
  double mnR;
  double mnG;
  double mnB;
  double sdR;
  double sdG;
  double sdB;
  MagickBooleanType isMeanSdDefined;
} MeanSdT;


static void WrMeanSd (MeanSdT * pms, FILE * fh)
{
  fprintf (fh, "isMeanSdDefined: %i\n", pms->isMeanSdDefined);
  if (pms->isMeanSdDefined) {
    fprintf (fh, "           mean: %g %g %g\n", pms->mnR, pms->mnG, pms->mnB);
    fprintf (fh, "             SD: %g %g %g\n", pms->sdR, pms->sdG, pms->sdB);
  }
}


static double inline sqrtEps (double v)
{
  return (v < 1e-10) ? 0 : sqrt(v);
}

static MagickBooleanType CalcMeanSdVP (
  const Image *image,
  const CacheView * in_view,
  ssize_t columns,
  ssize_t rows,
  ssize_t offsX,
  ssize_t offsY,
  MeanSdT * pms,
  ExceptionInfo *exception
)
{
  double sigR=0, sigG=0, sigB=0, sigA=0;
  double sigR2=0, sigG2=0, sigB2=0;
  MagickBooleanType okay = MagickTrue;

  ssize_t y;

#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) \
    MAGICK_THREADS(image,image,rows,1)
#endif

  for (y=0; y < rows; y++) {
    MagickBooleanType okayY = MagickTrue;
    const VIEW_PIX_PTR *p=GetCacheViewVirtualPixels (
      in_view,offsX,y+offsY,columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL) {
      okayY = MagickFalse;
      continue;
    }

    double sigRY=0, sigGY=0, sigBY=0, sigAY=0;
    double sigR2Y=0, sigG2Y=0, sigB2Y=0;

    ssize_t x;
    double r, g, b, a;
    for (x=0; x < columns; x++) {
      a = GET_PIXEL_ALPHA(image,p) / QuantumRange;
      r = a * GET_PIXEL_RED(image,p) / QuantumRange;
      g = a * GET_PIXEL_GREEN(image,p) / QuantumRange;
      b = a * GET_PIXEL_BLUE(image,p) / QuantumRange;
      sigRY += r;
      sigGY += g;
      sigBY += b;
      sigAY += a;
      sigR2Y += r*r;
      sigG2Y += g*g;
      sigB2Y += b*b;

      p += Inc_ViewPixPtr (image);
    }

#if defined(MAGICKCORE_OPENMP_SUPPORT)
#pragma omp atomic
    sigR += sigRY;
#pragma omp atomic
    sigG += sigGY;
#pragma omp atomic
    sigB += sigBY;
#pragma omp atomic
    sigA += sigAY;
#pragma omp atomic
    sigR2 += sigR2Y;
#pragma omp atomic
    sigG2 += sigG2Y;
#pragma omp atomic
    sigB2 += sigG2Y;
#pragma omp atomic
    okay &= okayY;
#else
    sigR += sigRY;
    sigG += sigGY;
    sigB += sigBY;
    sigA += sigAY;
    sigR2 += sigR2Y;
    sigG2 += sigG2Y;
    sigB2 += sigG2Y;
    okay &= okayY;
#endif
  }

  if (sigA == 0) {
    pms->isMeanSdDefined = MagickFalse;
    return MagickTrue;
  }

  pms->isMeanSdDefined = MagickTrue;
  pms->mnR = sigR / sigA;
  pms->mnG = sigG / sigA;
  pms->mnB = sigB / sigA;
  pms->sdR = sqrtEps (sigR2/sigA - (pms->mnR)*(pms->mnR));
  pms->sdG = sqrtEps (sigG2/sigA - (pms->mnG)*(pms->mnG));
  pms->sdB = sqrtEps (sigB2/sigA - (pms->mnB)*(pms->mnB));

  return MagickTrue;
}


static MagickBooleanType CalcMeanSdImg (

  const Image *image,
  MeanSdT * pms,
  ExceptionInfo *exception
)
// Calculate mean and standard deviation,
// accounting for alpha (eg ignoring pixels that are entirely transparent).
// Returns values in range 0.0 to 1.0.
// Returns false if major problem.
// Returns isMeanSdDefined = MagickFalse iff statistics are not defined
//   (image is entirely transparent).
{
  // mnR = sig(R) / sig(alpha)
  // sdR = sqrt ( sig(R^2) / sig(alpha) - mnR^2 )
  //
  // Likewise for G and B.

  CacheView *in_view = AcquireVirtualCacheView (image, exception);

  MagickBooleanType r = CalcMeanSdVP (
    image, in_view, image->columns, image->rows, 0, 0, pms, exception);

  in_view = DestroyCacheView (in_view);

  return r;
}

#endif
