/*
   Reference: http://im.snibgo.com/cols2mat.htm

   Updated:
     3-April-2018 for v7.0.7-28
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"
//#include "chklist.h"

#if IMV6OR7==7
#include <MagickCore/matrix-private.h>  // for GaussJordanElimination()
#endif

#include "cols2mat.inc"


#define VERSION "cols2mat v1.0  Copyright (c) 2017 Alan Gibson"

typedef struct {
  int
    precision;

  MagickBooleanType
    doTrans,
    weightAlpha,
    verbose;

  c2mTypeT
    c2mType;

  int
    polyDegree;

  double
    BottomLineWeight;

  FILE *
    fh_data;
} cols2matT;

static void usage (void)
{
  printf ("Usage: -process 'cols2mat [OPTION]...'\n");
  printf ("Finds the colour matrix (or polynomials) that transforms first image to second.\n");
  printf ("\n");
  printf ("  m,  method string       'Cross', 'NoCross', 'NoCrossPoly' or 'GainOnly'\n");
  printf ("  d,  degreePoly integer  degree for NoCrossPoly\n");
  printf ("  w,  weightLast number   weight for bottom line of image\n");
  printf ("  wa, weightAlpha         multiply weight by product of alphas\n");
  printf ("  x,  noTrans             don't replace images with transformation\n");
  printf ("  f,  file string         write to file stream stdout or stderr\n");
  printf ("  v,  verbose             write text information to stdout\n");
  printf ("      version             write version information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  // LocaleCompare is not case-sensitive,
  // so we use strcmp for the short option.

  if ((strcmp(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "cols2mat: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  cols2matT * psi
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status = MagickTrue;

  psi->verbose = MagickFalse;
  psi->weightAlpha = MagickFalse;
  psi->doTrans = MagickTrue;
  psi->fh_data = stderr;
  psi->c2mType = ctCross;
  psi->polyDegree = 2;
  psi->BottomLineWeight = 1.0;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];
    //printf ("Arg %i [%s]\n", i, pa);

    if (IsArg (pa, "x", "noTrans")==MagickTrue) {
      psi->doTrans = MagickFalse;
    } else if (IsArg (pa, "m", "method")==MagickTrue) {
      NEXTARG;
      if (strcasecmp (argv[i], "Cross")==0) psi->c2mType = ctCross;
      else if (strcasecmp (argv[i], "NoCross")==0) psi->c2mType = ctNoCross;
      else if (strcasecmp (argv[i], "NoCrossPoly")==0) psi->c2mType = ctNoCrossPoly;
      else if (strcasecmp (argv[i], "GainOnly")==0) psi->c2mType = ctGainOnly;
      else status = MagickFalse;
    } else if (IsArg (pa, "d", "degreePoly")==MagickTrue) {
      NEXTARG;
      psi->polyDegree = atoi(argv[i]);
    } else if (IsArg (pa, "w", "weightLast")==MagickTrue) {
      NEXTARG;
      psi->BottomLineWeight = atof(argv[i]);
    } else if (IsArg (pa, "wa", "weightAlpha")==MagickTrue) {
      psi->weightAlpha = MagickTrue;
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      NEXTARG;
      if (strcasecmp (argv[i], "stdout")==0) psi->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) psi->fh_data = stderr;
      else status = MagickFalse;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      psi->verbose = MagickTrue;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "cols2mat: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (psi->polyDegree < 0) {
    fprintf (stdout, "polyDegree is less than zero.\n");
    status = MagickFalse;
  }

  if (psi->verbose) {
    fprintf (stderr, "cols2mat options: ");

    if (!psi->doTrans) fprintf (stderr, " noTrans");

    fprintf (stderr, "  method ");
    switch (psi->c2mType) {
      case ctCross:
        fprintf (stderr, "Cross");
        break;
      case ctNoCross:
        fprintf (stderr, "NoCross");
        break;
      case ctNoCrossPoly:
        fprintf (stderr, "NoCrossPoly degreePoly %i", psi->polyDegree);
        break;
      case ctGainOnly:
        fprintf (stderr, "GainOnly");
        break;
    }

    fprintf (stderr, "  weightLast %g", psi->BottomLineWeight);
    if (psi->weightAlpha) fprintf (stderr, " weightAlpha");

    if (psi->fh_data == stdout) fprintf (stderr, " file stdout");

    if (psi->verbose)    fprintf (stderr, " verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


// The next function takes an image list,
// may modify it, and returns a status.
//
static MagickBooleanType cols2mat (
  Image **images,
  cols2matT * psi,
  ExceptionInfo *exception)
{
  if ((*images)->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",(*images)->filename);

  int ListLen = (int)GetImageListLength(*images);
  if (ListLen < 2 || ListLen > 3) {
    fprintf (stderr, "cols2mat needs 2 or 3 images\n");
    return MagickFalse;
  }

  Image * imgA = *images;
  Image * imgB = GetNextImageInList (imgA);
  assert(imgA->signature == MAGICK_CORE_SIG);
  assert(imgB->signature == MAGICK_CORE_SIG);
  Image * src_img = imgA;
  if (ListLen == 3) {
    src_img = GetNextImageInList (imgB);
    assert(src_img->signature == MAGICK_CORE_SIG);
  }

  if (psi->verbose) {
    fprintf (stderr, "cols2mat: imgA [%s] %ix%i depth is %i\n",
             imgA->filename,
             (int)imgA->columns, (int)imgA->rows,
             (int)imgA->depth);
    fprintf (stderr, "cols2mat: ListLen=%i\n", ListLen);
  }

  cols2matsT cols2mats;
  InitCols2mat (&cols2mats);
  cols2mats.verbose = psi->verbose;
  cols2mats.weightAlpha = psi->weightAlpha;
  cols2mats.c2mType = psi->c2mType;
  cols2mats.polyDegree = psi->polyDegree;
  cols2mats.BottomLineWeight = psi->BottomLineWeight;
  cols2mats.kernel = NULL;

  if (!cols2matGen (imgA, imgB, &cols2mats, psi->fh_data, exception))
    return MagickFalse;

  if (!cols2mats.solutionFound) {
    fprintf (stderr, "cols2mat: No solution\n");
    return MagickFalse;
  }

  if (psi->doTrans) {
    Image * new_img = NULL;

    switch (cols2mats.c2mType) {
      case ctNoCrossPoly:
        new_img = CloneImage (src_img, 0, 0, MagickTrue, exception);
        if (!new_img) return MagickFalse;

#if IMV6OR7==6
        if (!FunctionImageChannel(new_img, RedChannel,
              PolynomialFunction, psi->polyDegree+1,
              cols2mats.polyCoeff[0], exception)) return MagickFalse;

        if (!FunctionImageChannel(new_img, GreenChannel,
              PolynomialFunction, psi->polyDegree+1,
              cols2mats.polyCoeff[1], exception)) return MagickFalse;

        if (!FunctionImageChannel(new_img, BlueChannel,
              PolynomialFunction, psi->polyDegree+1,
              cols2mats.polyCoeff[2], exception)) return MagickFalse;
#else
        ChannelType channel_mask = SetImageChannelMask (new_img, RedChannel);

        if (!FunctionImage(new_img,
              PolynomialFunction, psi->polyDegree+1,
              cols2mats.polyCoeff[0], exception)) return MagickFalse;

        SetImageChannelMask (new_img, GreenChannel);

        if (!FunctionImage(new_img,
              PolynomialFunction, psi->polyDegree+1,
              cols2mats.polyCoeff[1], exception)) return MagickFalse;

        SetImageChannelMask (new_img, BlueChannel);

        if (!FunctionImage(new_img,
              PolynomialFunction, psi->polyDegree+1,
              cols2mats.polyCoeff[2], exception)) return MagickFalse;

        SetImageChannelMask(new_img,channel_mask);
#endif

        break;

      default:
        new_img = ColorMatrixImage (src_img, cols2mats.kernel, exception);
        if (!new_img) return MagickFalse;
    }

    DeleteImageFromList (&imgB);
    if (ListLen == 3) DeleteImageFromList (&src_img);

    ReplaceImageInList (images, new_img);
    // Replace messes up the images pointer. Make it good:
    *images = GetFirstImageInList (new_img);
  }

  if (cols2mats.kernel) {
    cols2mats.kernel = DestroyKernelInfo (cols2mats.kernel);
  }

  return MagickTrue;
}


ModuleExport size_t cols2matImage (
  Image **images,
  const int argc,
  const char **argv,
  ExceptionInfo *exception)
{
  cols2matT si;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  MagickBooleanType status = menu (argc, argv, &si);
  if (!status) return -1;

  si.precision = GetMagickPrecision();

  status = cols2mat (images, &si, exception);
  if (!status) return (-1);

  return(MagickImageFilterSignature);
}
