/* Updated:
     3-April-2018 for v7.0.7-28
*/

#ifndef COLS2MAT_INC
#define COLS2MAT_INC

#include "chklist.h"


typedef enum {
  ctCross,
  ctNoCross,
  ctNoCrossPoly,
  ctGainOnly
} c2mTypeT;

#define NUM_CH 3

typedef struct {
  MagickBooleanType
    weightAlpha,
    verbose,
    doWarn;

  c2mTypeT
    c2mType;

  int
    polyDegree;

  double
    **matrix,
    **vectors,
    *terms,
    *results;

  double
    *polyCoeff[NUM_CH];

  double
    BottomLineWeight;

  // Used internally
  MagickBooleanType
    imgAhasAlpha,
    imgBhasAlpha;

  // Returned results
  KernelInfo
    *kernel;

  MagickBooleanType
    solutionFound;
} cols2matsT;

// Option to add zero and 100% for matrix calculation?


static void InitCols2mat (
  cols2matsT * c2m)
{
  c2m->verbose = MagickFalse;
  c2m->weightAlpha = MagickFalse;
  c2m->doWarn = MagickTrue;
  c2m->c2mType = ctCross;
  c2m->polyDegree = 2;
  c2m->matrix = c2m->vectors = NULL;
  c2m->terms = c2m->results = NULL;
  c2m->kernel = NULL;
  c2m->BottomLineWeight = 1.0;
  int c;
  for (c = 0; c < NUM_CH; c++) c2m->polyCoeff[c] = NULL;
}


static MagickBooleanType AllocCols2mat (
  cols2matsT * c2m,
  int nTerms,
  int nResults
)
{
  c2m->terms   = (double *)AcquireMagickMemory (nTerms   * sizeof(double));
  if (!c2m->terms) return MagickFalse;

  c2m->results = (double *)AcquireMagickMemory (nResults * sizeof(double));
  if (!c2m->results) return MagickFalse;

  c2m->matrix = AcquireMagickMatrix (nTerms, nTerms);
  if (!c2m->matrix) return MagickFalse;

  c2m->vectors = AcquireMagickMatrix (nResults, nTerms);
  if (!c2m->vectors) return MagickFalse;

  return MagickTrue;
}


static void DeAllocCols2mat (
  cols2matsT * c2m,
  int nTerms,
  int nResults
)
{
  c2m->vectors = RelinquishMagickMatrix (c2m->vectors, nResults);
  c2m->matrix  = RelinquishMagickMatrix (c2m->matrix, nTerms);
  c2m->results = RelinquishMagickMemory (c2m->results);
  c2m->terms   = RelinquishMagickMemory (c2m->terms);
}


static void LeastSquaresAddTermsWt (
  double **matrix,
  double **vectors,
  const double *terms,
  const double *results,
  const size_t rank,
  const size_t number_vectors,
  double weight)
{
  register ssize_t
    i,
    j;

  for (j=0; j < (ssize_t) rank; j++)
  {
    for (i=0; i < (ssize_t) rank; i++)
      matrix[i][j] += weight*terms[i]*terms[j];

    for (i=0; i < (ssize_t) number_vectors; i++)
      vectors[i][j] += weight*results[i]*terms[j];
  }
}

static double inline AlphaWeight (
  cols2matsT *c2m,
  Image * imgA,
  Image * imgB,
  const VIEW_PIX_PTR *pA,
  const VIEW_PIX_PTR *pB
)
{
  double wt = 1.0;
  if (c2m->weightAlpha) {
    if (c2m->imgAhasAlpha) {
      wt = GET_PIXEL_ALPHA (imgA, pA) / QuantumRange;
    }
    if (c2m->imgBhasAlpha) {
      wt *= GET_PIXEL_ALPHA (imgB, pB) / QuantumRange;
    }
  }
  return wt;
}


static MagickBooleanType cols2matCross (
  Image * imgA,
  Image * imgB,
  cols2matsT *c2m,
  ExceptionInfo *exception
)
/* From two images, same size,
   creates color matrix with 36 elements
   that if applied to imgA would make it look like imgB.
   "Cross" -- includes parameter for cross-freed between channels.
*/
{
  c2m->solutionFound = MagickFalse;

  if (imgA->columns != imgB->columns || imgA->rows != imgB->rows) {
    fprintf (stderr, "cols2mat: images not the same size\n");
    return MagickFalse;
  }

#define nTermsC 4
#define nResultsC 3

  if (!AllocCols2mat (c2m, nTermsC, nResultsC)) { fprintf (stderr, "alloc fail\n"); return MagickFalse; }

  CacheView * imgA_view = AcquireVirtualCacheView (imgA, exception);
  CacheView * imgB_view = AcquireVirtualCacheView (imgB, exception);

  c2m->terms[3] = 1;

  ssize_t y;
  for (y=0; y < imgA->rows; y++) {
    const VIEW_PIX_PTR *pA = GetCacheViewVirtualPixels (
        imgA_view, 0, y, imgA->columns, 1, exception);
    if (!pA) return MagickFalse;

    const VIEW_PIX_PTR *pB = GetCacheViewVirtualPixels (
        imgB_view, 0, y, imgB->columns, 1, exception);
    if (!pB) return MagickFalse;

    double LnWeight = (y==imgA->rows-1) ? c2m->BottomLineWeight : 1.0;

    ssize_t x;
    for (x=0; x < imgA->columns; x++) {
      c2m->terms[0]   = GET_PIXEL_RED   (imgA, pA) / QuantumRange;
      c2m->terms[1]   = GET_PIXEL_GREEN (imgA, pA) / QuantumRange;
      c2m->terms[2]   = GET_PIXEL_BLUE  (imgA, pA) / QuantumRange;

      c2m->results[0] = GET_PIXEL_RED   (imgB, pB) / QuantumRange;
      c2m->results[1] = GET_PIXEL_GREEN (imgB, pB) / QuantumRange;
      c2m->results[2] = GET_PIXEL_BLUE  (imgB, pB) / QuantumRange;

      LeastSquaresAddTermsWt (
        c2m->matrix, c2m->vectors, c2m->terms, c2m->results,
        nTermsC, nResultsC, LnWeight * AlphaWeight (c2m,imgA,imgB,pA,pB));

      pA += Inc_ViewPixPtr (imgA);
      pB += Inc_ViewPixPtr (imgB);
    }
  }

  if (c2m->verbose > 1) {
    fprintf (stderr, "cols2mat: Matrix is:\n");
    int u, v;
    for (u = 0; u < nTermsC; u++) {
      for (v = 0; v < nTermsC; v++) {
        fprintf (stderr, "  %g", c2m->matrix[u][v]);
      }
      fprintf (stderr, "\n");
    }
  }

  if (!GaussJordanElimination(c2m->matrix,c2m->vectors,nTermsC,nResultsC) ) {
    fprintf (stderr, "No solution.\n");
  } else {
    if (c2m->verbose > 1) {
      int j, k;
      fprintf (stderr, "cols2mat: Vectors is:\n");
      for (j = 0; j < nResultsC; j++) {
        for (k = 0; k < nTermsC; k++) {
          fprintf (stderr, "  %g", c2m->vectors[j][k]);
        }
        fprintf (stderr, "\n");
      }
    }
    c2m->kernel->values[0]  = c2m->vectors[0][0];
    c2m->kernel->values[1]  = c2m->vectors[0][1];
    c2m->kernel->values[2]  = c2m->vectors[0][2];
    c2m->kernel->values[5]  = c2m->vectors[0][3];

    c2m->kernel->values[6]  = c2m->vectors[1][0];
    c2m->kernel->values[7]  = c2m->vectors[1][1];
    c2m->kernel->values[8]  = c2m->vectors[1][2];
    c2m->kernel->values[11] = c2m->vectors[1][3];

    c2m->kernel->values[12] = c2m->vectors[2][0];
    c2m->kernel->values[13] = c2m->vectors[2][1];
    c2m->kernel->values[14] = c2m->vectors[2][2];
    c2m->kernel->values[17] = c2m->vectors[2][3];

    c2m->solutionFound = MagickTrue;
  }

  imgB_view = DestroyCacheView (imgB_view);
  imgA_view = DestroyCacheView (imgA_view);

  DeAllocCols2mat (c2m, nTermsC, nResultsC);

#undef nTermsC
#undef nResultsC

//  if (c2m->verbose) {
//    ShowKernelInfo (c2m->kernel);
//  }

  return MagickTrue;
}


static MagickBooleanType cols2matNoCross (
  Image * imgA,
  Image * imgB,
  cols2matsT * c2m,
  ExceptionInfo *exception
)
/* From two images, same size,
   creates color matrix with 36 elements
   that if applied to imgA would make it look like imgB.
   "NoCross" -- excludes parameters for cross-freed between channels.
*/
{
  c2m->solutionFound = MagickFalse;

#define nTermsNC 2
#define nResultsNC 1

  if (!AllocCols2mat (c2m, nTermsNC, nResultsNC)) { fprintf (stderr, "alloc fail\n"); return MagickFalse; }

  CacheView * imgA_view = AcquireVirtualCacheView (imgA, exception);
  CacheView * imgB_view = AcquireVirtualCacheView (imgB, exception);

  ssize_t y;

  int nFound = 0;

  // Solve for red.
  //
  for (y=0; y < imgA->rows; y++) {
    const VIEW_PIX_PTR *pA = GetCacheViewVirtualPixels (
        imgA_view, 0, y, imgA->columns, 1, exception);
    if (!pA) return MagickFalse;

    const VIEW_PIX_PTR *pB = GetCacheViewVirtualPixels (
        imgB_view, 0, y, imgB->columns, 1, exception);
    if (!pB) return MagickFalse;

    double LnWeight = (y==imgA->rows-1) ? c2m->BottomLineWeight : 1.0;

    ssize_t x;
    for (x=0; x < imgA->columns; x++) {
      c2m->terms[0]   = GET_PIXEL_RED   (imgA, pA) / QuantumRange;
      c2m->terms[1]   = 1;
      c2m->results[0] = GET_PIXEL_RED   (imgB, pB) / QuantumRange;

      LeastSquaresAddTermsWt (
        c2m->matrix, c2m->vectors, c2m->terms, c2m->results,
        nTermsNC, nResultsNC, LnWeight * AlphaWeight (c2m,imgA,imgB,pA,pB));

      pA += Inc_ViewPixPtr (imgA);
      pB += Inc_ViewPixPtr (imgB);
    }
  }

  if (!GaussJordanElimination(c2m->matrix,c2m->vectors,nTermsNC,nResultsNC) ) {
    fprintf (stderr, "No solution R.\n");
  } else {
    c2m->kernel->values[0]  = c2m->vectors[0][0];
    c2m->kernel->values[5]  = c2m->vectors[0][1];
    nFound++;
  }

  DeAllocCols2mat (c2m, nTermsNC, nResultsNC);
  if (!AllocCols2mat (c2m, nTermsNC, nResultsNC)) { fprintf (stderr, "alloc fail\n"); return MagickFalse; }

  // Solve for green.
  //
  for (y=0; y < imgA->rows; y++) {
    const VIEW_PIX_PTR *pA = GetCacheViewVirtualPixels (
        imgA_view, 0, y, imgA->columns, 1, exception);
    if (!pA) return MagickFalse;

    const VIEW_PIX_PTR *pB = GetCacheViewVirtualPixels (
        imgB_view, 0, y, imgB->columns, 1, exception);
    if (!pB) return MagickFalse;

    double LnWeight = (y==imgA->rows-1) ? c2m->BottomLineWeight : 1.0;

    ssize_t x;
    for (x=0; x < imgA->columns; x++) {
      c2m->terms[0]   = GET_PIXEL_GREEN (imgA, pA) / QuantumRange;
      c2m->terms[1]   = 1;
      c2m->results[0] = GET_PIXEL_GREEN (imgB, pB) / QuantumRange;

      LeastSquaresAddTermsWt (
        c2m->matrix, c2m->vectors, c2m->terms, c2m->results,
        nTermsNC, nResultsNC, LnWeight * AlphaWeight (c2m,imgA,imgB,pA,pB));

      pA += Inc_ViewPixPtr (imgA);
      pB += Inc_ViewPixPtr (imgB);
    }
  }

  if (!GaussJordanElimination(c2m->matrix,c2m->vectors,nTermsNC,nResultsNC) ) {
    fprintf (stderr, "No solution G.\n");
  } else {
    c2m->kernel->values[7]  = c2m->vectors[0][0];
    c2m->kernel->values[11] = c2m->vectors[0][1];
    nFound++;
  }

  DeAllocCols2mat (c2m, nTermsNC, nResultsNC);
  if (!AllocCols2mat (c2m, nTermsNC, nResultsNC)) { fprintf (stderr, "alloc fail\n"); return MagickFalse; }

  // Solve for blue.
  //
  for (y=0; y < imgA->rows; y++) {
    const VIEW_PIX_PTR *pA = GetCacheViewVirtualPixels (
        imgA_view, 0, y, imgA->columns, 1, exception);
    if (!pA) return MagickFalse;

    const VIEW_PIX_PTR *pB = GetCacheViewVirtualPixels (
        imgB_view, 0, y, imgB->columns, 1, exception);
    if (!pB) return MagickFalse;

    double LnWeight = (y==imgA->rows-1) ? c2m->BottomLineWeight : 1.0;

    ssize_t x;
    for (x=0; x < imgA->columns; x++) {
      c2m->terms[0]   = GET_PIXEL_BLUE  (imgA, pA) / QuantumRange;
      c2m->terms[1]   = 1;
      c2m->results[0] = GET_PIXEL_BLUE  (imgB, pB) / QuantumRange;

      LeastSquaresAddTermsWt (
        c2m->matrix, c2m->vectors, c2m->terms, c2m->results,
        nTermsNC, nResultsNC, LnWeight * AlphaWeight (c2m,imgA,imgB,pA,pB));

      pA += Inc_ViewPixPtr (imgA);
      pB += Inc_ViewPixPtr (imgB);
    }
  }

  if (!GaussJordanElimination(c2m->matrix,c2m->vectors,nTermsNC,nResultsNC) ) {
    fprintf (stderr, "No solution B.\n");
  } else {
    c2m->kernel->values[14]  = c2m->vectors[0][0];
    c2m->kernel->values[17]  = c2m->vectors[0][1];
    nFound++;
  }

  c2m->solutionFound = (nFound ==3);

  imgB_view = DestroyCacheView (imgB_view);
  imgA_view = DestroyCacheView (imgA_view);

  DeAllocCols2mat (c2m, nTermsNC, nResultsNC);

#undef nTermsNC
#undef nResultsNC

//  if (c2m->verbose) {
//    ShowKernelInfo (c2m->kernel);
//  }

  return MagickTrue;
}


static MagickBooleanType cols2matGainOnly (
  Image * imgA,
  Image * imgB,
  cols2matsT * c2m,
  ExceptionInfo *exception
)
/* From two images, same size,
   creates color matrix with 36 elements
   that if applied to imgA would make it look like imgB.
   "GainOnly" -- no cross-feed, no bias.
*/
{
  c2m->solutionFound = MagickFalse;

#define nTermsGO 1
#define nResultsGO 1

  if (!AllocCols2mat (c2m, nTermsGO, nResultsGO)) { fprintf (stderr, "alloc fail\n"); return MagickFalse; }

  CacheView * imgA_view = AcquireVirtualCacheView (imgA, exception);
  CacheView * imgB_view = AcquireVirtualCacheView (imgB, exception);

  ssize_t y;

  int nFound = 0;

  // Solve for red.
  //
  for (y=0; y < imgA->rows; y++) {
    const VIEW_PIX_PTR *pA = GetCacheViewVirtualPixels (
        imgA_view, 0, y, imgA->columns, 1, exception);
    if (!pA) return MagickFalse;

    const VIEW_PIX_PTR *pB = GetCacheViewVirtualPixels (
        imgB_view, 0, y, imgB->columns, 1, exception);
    if (!pB) return MagickFalse;

    double LnWeight = (y==imgA->rows-1) ? c2m->BottomLineWeight : 1.0;

    ssize_t x;
    for (x=0; x < imgA->columns; x++) {
      c2m->terms[0]   = GET_PIXEL_RED   (imgA, pA) / QuantumRange;
      c2m->results[0] = GET_PIXEL_RED   (imgB, pB) / QuantumRange;

      LeastSquaresAddTermsWt (
        c2m->matrix, c2m->vectors, c2m->terms, c2m->results,
        nTermsGO, nResultsGO, LnWeight * AlphaWeight (c2m,imgA,imgB,pA,pB));

      pA += Inc_ViewPixPtr (imgA);
      pB += Inc_ViewPixPtr (imgB);
    }
  }

  if (!GaussJordanElimination(c2m->matrix,c2m->vectors,nTermsGO,nResultsGO) ) {
    fprintf (stderr, "No solution R.\n");
  } else {
    c2m->kernel->values[0]  = c2m->vectors[0][0];
    nFound++;
  }

  DeAllocCols2mat (c2m, nTermsGO, nResultsGO);
  if (!AllocCols2mat (c2m, nTermsGO, nResultsGO)) { fprintf (stderr, "alloc fail\n"); return MagickFalse; }

  // Solve for green.
  //
  for (y=0; y < imgA->rows; y++) {
    const VIEW_PIX_PTR *pA = GetCacheViewVirtualPixels (
        imgA_view, 0, y, imgA->columns, 1, exception);
    if (!pA) return MagickFalse;

    const VIEW_PIX_PTR *pB = GetCacheViewVirtualPixels (
        imgB_view, 0, y, imgB->columns, 1, exception);
    if (!pB) return MagickFalse;

    double LnWeight = (y==imgA->rows-1) ? c2m->BottomLineWeight : 1.0;

    ssize_t x;
    for (x=0; x < imgA->columns; x++) {
      c2m->terms[0]   = GET_PIXEL_GREEN (imgA, pA) / QuantumRange;
      c2m->results[0] = GET_PIXEL_GREEN (imgB, pB) / QuantumRange;

      LeastSquaresAddTermsWt (
        c2m->matrix, c2m->vectors, c2m->terms, c2m->results,
        nTermsGO, nResultsGO, LnWeight * AlphaWeight (c2m,imgA,imgB,pA,pB));

      pA += Inc_ViewPixPtr (imgA);
      pB += Inc_ViewPixPtr (imgB);
    }
  }

  if (!GaussJordanElimination(c2m->matrix,c2m->vectors,nTermsGO,nResultsGO) ) {
    fprintf (stderr, "No solution G.\n");
  } else {
    c2m->kernel->values[7]  = c2m->vectors[0][0];
    nFound++;
  }

  DeAllocCols2mat (c2m, nTermsGO, nResultsGO);
  if (!AllocCols2mat (c2m, nTermsGO, nResultsGO)) { fprintf (stderr, "alloc fail\n"); return MagickFalse; }

  // Solve for blue.
  //
  for (y=0; y < imgA->rows; y++) {
    const VIEW_PIX_PTR *pA = GetCacheViewVirtualPixels (
        imgA_view, 0, y, imgA->columns, 1, exception);
    if (!pA) return MagickFalse;

    const VIEW_PIX_PTR *pB = GetCacheViewVirtualPixels (
        imgB_view, 0, y, imgB->columns, 1, exception);
    if (!pB) return MagickFalse;

    double LnWeight = (y==imgA->rows-1) ? c2m->BottomLineWeight : 1.0;

    ssize_t x;
    for (x=0; x < imgA->columns; x++) {
      c2m->terms[0]   = GET_PIXEL_BLUE  (imgA, pA) / QuantumRange;
      c2m->results[0] = GET_PIXEL_BLUE  (imgB, pB) / QuantumRange;

      LeastSquaresAddTermsWt (
        c2m->matrix, c2m->vectors, c2m->terms, c2m->results,
        nTermsGO, nResultsGO, LnWeight * AlphaWeight (c2m,imgA,imgB,pA,pB));

      pA += Inc_ViewPixPtr (imgA);
      pB += Inc_ViewPixPtr (imgB);
    }
  }

  if (!GaussJordanElimination(c2m->matrix,c2m->vectors,nTermsGO,nResultsGO) ) {
    fprintf (stderr, "No solution B.\n");
  } else {
    c2m->kernel->values[14]  = c2m->vectors[0][0];
    nFound++;
  }

  c2m->solutionFound = (nFound ==3);


  imgB_view = DestroyCacheView (imgB_view);
  imgA_view = DestroyCacheView (imgA_view);

  DeAllocCols2mat (c2m, nTermsGO, nResultsGO);

#undef nTermsGO
#undef nResultsGO

//  if (c2m->verbose) {
//    ShowKernelInfo (c2m->kernel);
//  }

  return MagickTrue;
}


static MagickBooleanType cols2matNoCrossPoly (
  Image * imgA,
  Image * imgB,
  cols2matsT * c2m,
  ExceptionInfo *exception
)
{
  c2m->solutionFound = MagickFalse;

  int nTerms = c2m->polyDegree + 1;
  int nResults = 1;

  // Note: terms[0] is term with largest degree.

  int c;
  for (c = 0; c < NUM_CH; c++) {
    // FIXME: This is never Relinguished.
    c2m->polyCoeff[c] = (double *)AcquireMagickMemory (nTerms * sizeof(double));
    if (!c2m->polyCoeff[c]) return MagickFalse;
  }

  CacheView * imgA_view = AcquireVirtualCacheView (imgA, exception);
  CacheView * imgB_view = AcquireVirtualCacheView (imgB, exception);

  int i;
  ssize_t y;

  int nFound = 0;

  if (!AllocCols2mat (c2m, nTerms, nResults)) { fprintf (stderr, "alloc fail\n"); return MagickFalse; }

  // Solve for red.
  //
  for (y=0; y < imgA->rows; y++) {
    const VIEW_PIX_PTR *pA = GetCacheViewVirtualPixels (
        imgA_view, 0, y, imgA->columns, 1, exception);
    if (!pA) return MagickFalse;

    const VIEW_PIX_PTR *pB = GetCacheViewVirtualPixels (
        imgB_view, 0, y, imgB->columns, 1, exception);
    if (!pB) return MagickFalse;

    double LnWeight = (y==imgA->rows-1) ? c2m->BottomLineWeight : 1.0;

    ssize_t x;
    for (x=0; x < imgA->columns; x++) {
      double v = GET_PIXEL_RED (imgA, pA) / QuantumRange;
      c2m->terms[nTerms-1]   = 1;
      for (i = 1; i < nTerms; i++) {
        int j = nTerms-1-i;
        c2m->terms[j] = v * c2m->terms[j+1];
//        printf ("term %i is %g\n", j, c2m->terms[j]);
      }
      c2m->results[0] = GET_PIXEL_RED (imgB, pB) / QuantumRange;

      LeastSquaresAddTermsWt (
        c2m->matrix, c2m->vectors, c2m->terms, c2m->results,
        nTerms, nResults, LnWeight * AlphaWeight (c2m,imgA,imgB,pA,pB));

      pA += Inc_ViewPixPtr (imgA);
      pB += Inc_ViewPixPtr (imgB);
    }
  }

  for (i = 0; i < nTerms; i++) {
//    printf ("term %i is %g\n", i, c2m->terms[i]);
  }

  if (!GaussJordanElimination(c2m->matrix,c2m->vectors,nTerms,nResults) ) {
    fprintf (stderr, "No solution R.\n");
  } else {
    for (i = 0; i < nTerms; i++) {
//      printf ("term %i is %g; coeff is ", i, c2m->terms[i]);
//      printf ("  %g\n", c2m->vectors[0][i]);
      c2m->polyCoeff[0][i] = c2m->vectors[0][i];
    }
//    printf ("\n");
    nFound++;
  }

  DeAllocCols2mat (c2m, nTerms, nResults);
  if (!AllocCols2mat (c2m, nTerms, nResults)) { fprintf (stderr, "alloc fail\n"); return MagickFalse; }

  // Solve for green.
  //
  for (y=0; y < imgA->rows; y++) {
    const VIEW_PIX_PTR *pA = GetCacheViewVirtualPixels (
        imgA_view, 0, y, imgA->columns, 1, exception);
    if (!pA) return MagickFalse;

    const VIEW_PIX_PTR *pB = GetCacheViewVirtualPixels (
        imgB_view, 0, y, imgB->columns, 1, exception);
    if (!pB) return MagickFalse;

    double LnWeight = (y==imgA->rows-1) ? c2m->BottomLineWeight : 1.0;

    ssize_t x;
    for (x=0; x < imgA->columns; x++) {
      double v = GET_PIXEL_GREEN (imgA, pA) / QuantumRange;
      c2m->terms[nTerms-1]   = 1;
      for (i = 1; i < nTerms; i++) {
        int j = nTerms-1-i;
        c2m->terms[j] = v * c2m->terms[j+1];
//        printf ("term %i is %g\n", j, c2m->terms[j]);
      }
      c2m->results[0] = GET_PIXEL_GREEN (imgB, pB) / QuantumRange;

      LeastSquaresAddTermsWt (
        c2m->matrix, c2m->vectors, c2m->terms, c2m->results,
        nTerms, nResults, LnWeight * AlphaWeight (c2m,imgA,imgB,pA,pB));

      pA += Inc_ViewPixPtr (imgA);
      pB += Inc_ViewPixPtr (imgB);
    }
  }

  for (i = 0; i < nTerms; i++) {
//    printf ("term %i is %g\n", i, c2m->terms[i]);
  }

  if (!GaussJordanElimination(c2m->matrix,c2m->vectors,nTerms,nResults) ) {
    fprintf (stderr, "No solution G.\n");
  } else {
    for (i = 0; i < nTerms; i++) {
//      printf ("term %i is %g; coeff is ", i, c2m->terms[i]);
//      printf ("  %g\n", c2m->vectors[0][i]);
      c2m->polyCoeff[1][i] = c2m->vectors[0][i];
    }
//    printf ("\n");
    nFound++;
  }

  DeAllocCols2mat (c2m, nTerms, nResults);
  if (!AllocCols2mat (c2m, nTerms, nResults)) { fprintf (stderr, "alloc fail\n"); return MagickFalse; }

  // Solve for blue.
  //
  for (y=0; y < imgA->rows; y++) {
    const VIEW_PIX_PTR *pA = GetCacheViewVirtualPixels (
        imgA_view, 0, y, imgA->columns, 1, exception);
    if (!pA) return MagickFalse;

    const VIEW_PIX_PTR *pB = GetCacheViewVirtualPixels (
        imgB_view, 0, y, imgB->columns, 1, exception);
    if (!pB) return MagickFalse;

    double LnWeight = (y==imgA->rows-1) ? c2m->BottomLineWeight : 1.0;

    ssize_t x;
    for (x=0; x < imgA->columns; x++) {
      double v = GET_PIXEL_BLUE (imgA, pA) / QuantumRange;
      c2m->terms[nTerms-1]   = 1;
      for (i = 1; i < nTerms; i++) {
        int j = nTerms-1-i;
        c2m->terms[j] = v * c2m->terms[j+1];
//        printf ("term %i is %g\n", j, c2m->terms[j]);
      }
      c2m->results[0] = GET_PIXEL_BLUE (imgB, pB) / QuantumRange;

      LeastSquaresAddTermsWt (
        c2m->matrix, c2m->vectors, c2m->terms, c2m->results,
        nTerms, nResults, LnWeight * AlphaWeight (c2m,imgA,imgB,pA,pB));

      pA += Inc_ViewPixPtr (imgA);
      pB += Inc_ViewPixPtr (imgB);
    }
  }

//  for (i = 0; i < nTerms; i++) {
//    printf ("term %i is %g\n", i, c2m->terms[i]);
//  }

  if (!GaussJordanElimination(c2m->matrix,c2m->vectors,nTerms,nResults) ) {
    fprintf (stderr, "No solution B.\n");
  } else {
    for (i = 0; i < nTerms; i++) {
//      printf ("term %i is %g; coeff is ", i, c2m->terms[i]);
//      printf ("  %g\n", c2m->vectors[0][i]);
      c2m->polyCoeff[2][i] = c2m->vectors[0][i];
    }
//    printf ("\n");
    nFound++;
  }

  DeAllocCols2mat (c2m, nTerms, nResults);

  c2m->solutionFound = (nFound == NUM_CH);

  imgB_view = DestroyCacheView (imgB_view);
  imgA_view = DestroyCacheView (imgA_view);

  return MagickTrue;
}

static void WrKernelVals (cols2matsT * c2m, FILE * fh, char * prefix)
{
  int precision = GetMagickPrecision();

  fprintf (fh, "%s", prefix);
  int i;
  for (i = 0; i < c2m->kernel->height * c2m->kernel->width; i++) {
    if (i) fprintf (fh, ",");
    fprintf (fh, "%.*g", precision, c2m->kernel->values[i]);
  }
  fprintf (fh, "\n");

  if (c2m->doWarn) {
    char *chNames[] = {"Red", "Green", "Blue"};

    int c,k;
    i = 0;
    for (c = 0; c < 2; c++) {
      double sumPosRow = 0.0;
      double last = 0.0;
      for (k = 0; k < 6; k++) {
        if (c2m->kernel->values[i] > 0) sumPosRow += c2m->kernel->values[i];
        if (k==5) last = c2m->kernel->values[i];
        i++;
      }
      if (last < 0.0) {
        fprintf (fh, "Warning: may clip %s shadows\n", chNames[c]);
      }
      if (sumPosRow > 1.0) {
        fprintf (fh, "Warning: may clip %s highlights\n", chNames[c]);
      }
    }
  }
}

static void WrPolynomials (cols2matsT * c2m, FILE * fh)
{
  char *chNames[] = {"Red", "Green", "Blue"};

  int precision = GetMagickPrecision();

  int nTerms = c2m->polyDegree + 1;

  int c, i;
  for (c = 0; c < NUM_CH; c++) {
    double sumPoly = 0.0;
    fprintf (fh, "Poly%s=", chNames[c]);
    for (i = 0; i < nTerms; i++) {
      if (i) fprintf (fh, ",");
      fprintf (fh, "%.*g", precision, c2m->polyCoeff[c][i]);
      sumPoly += c2m->polyCoeff[c][i];
    }
    fprintf (fh, "\n");
    if (c2m->doWarn) {
      if (c2m->polyCoeff[c][nTerms-1] < 0.0) {
        fprintf (fh, "Warning: may clip %s shadows\n", chNames[c]);
      }
      if (sumPoly > 1.0) {
        fprintf (fh, "Warning: may clip %s highlights\n", chNames[c]);
      }
    }
  }
}


static MagickBooleanType cols2matGen (
  Image * imgA,
  Image * imgB,
  cols2matsT * c2m,
  FILE *fh_out,
  ExceptionInfo *exception
)
// This may acquire a kernel. Caller should destroy it.
{
  c2m->solutionFound = MagickFalse;

  if (imgA->columns != imgB->columns || imgA->rows != imgB->rows) {
    fprintf (stderr, "cols2mat: images not the same size\n");
    return MagickFalse;
  }

  // FIXME: Following is a kludge.
  // We merely want to ensure we have three colour channels.
#if IMV6OR7==6
  TransformImageColorspace (imgA, sRGBColorspace);
  TransformImageColorspace (imgB, sRGBColorspace);
#else
  TransformImageColorspace (imgA, sRGBColorspace, exception);
  TransformImageColorspace (imgB, sRGBColorspace, exception);
#endif

  c2m->kernel = ACQUIRE_KERNEL(
    "6x6:1,0,0,0,0,0,0,1,0,0,0,0,0,0,1,0,0,0,0,0,0,1,0,0,0,0,0,0,1,0,0,0,0,0,0,1",
     exception);
  if (!c2m->kernel) {
    fprintf (stderr, "AcquireKernelInfo failed\n");
    return MagickFalse;
  }

  c2m->imgAhasAlpha = IS_ALPHA_CH(imgA);
  c2m->imgBhasAlpha = IS_ALPHA_CH(imgB);

  MagickBooleanType okay = MagickFalse;

  switch (c2m->c2mType) {
    case ctCross:
      okay = cols2matCross (imgA, imgB, c2m, exception);
      break;
    case ctNoCross:
      okay = cols2matNoCross (imgA, imgB, c2m, exception);
      break;
    case ctNoCrossPoly:
      okay = cols2matNoCrossPoly (imgA, imgB, c2m, exception);
      break;
    case ctGainOnly:
      okay = cols2matGainOnly (imgA, imgB, c2m, exception);
      break;
  }

  if (okay && fh_out) {
    if (c2m->solutionFound) {
      switch (c2m->c2mType) {
        case ctNoCrossPoly:
          WrPolynomials (c2m, fh_out);
          break;

        default:
          WrKernelVals (c2m, fh_out, "c2matrix=");
      }
    } else {
      fprintf (fh_out, "cols2mat: no solution found\n");
    }
  }

  return okay;
}


static KernelInfo * MultKernels (KernelInfo *kA, KernelInfo *kB)
// From two equal-sized square kernels,
// create a new kernel, which caller should eventually destroy.
{
  int nElements = kA->width;

  if (nElements != kA->height
   || nElements != kB->height
   || nElements != kB->width)
  {
    return NULL;
  }

  KernelInfo * newK = CloneKernelInfo (kA);
  if (!newK) return NULL;

  int x, y, i;

  for (i = 0; i < nElements*nElements; i++) {
    newK->values[i] = 0;
  }

  for (y = 0; y < nElements; y++) {
    for (x = 0; x < nElements; x++) {
      int nNew = y * nElements + x;
      for (i = 0; i < nElements; i++) {
        int nA = y * nElements + i;
        int nB = i * nElements + x;
        newK->values[nNew] = kA->values[nA] * kB->values[nB];
      }
    }
  }

  return newK;
}


#endif