/* Compare windows. 

   Compare a window in sub_view with the same-size window in ref_view.
   Returns a score.

   First written: 30-Nov-2015.
     (Hived-off from fillholes.c)

   Updates:
     1-Dec-2015 No longer rejects when sub coords == ref coords.
     1-Feb-2016 For v7.
*/

static void inline instantPp (Image * image, PIX_INFO *p)
{
#if IMV6OR7==6
  GetMagickPixelPacket (image, p);
#else
  GetPixelInfo (image, p);
#endif

//  SET_PIXEL_RED   (image, 0, p);
//  SET_PIXEL_GREEN (image, 0, p);
//  SET_PIXEL_BLUE  (image, 0, p);
}


static MagickBooleanType IsSubTrans (
  CompWindT * pcw,
  ExceptionInfo *exception
)
// Returns true if all pixels in subimage window are entirely transparent.
// Retuns false if any have any opacity.
{
  ssize_t
    xy;

  const VIEW_PIX_PTR
    *subxy;

  int
    wiht = pcw->win_wi * pcw->win_ht;

  subxy = GetCacheViewVirtualPixels(pcw->sub_view,
    pcw->subX,pcw->subY,
    pcw->win_wi,pcw->win_ht,exception);

  if (subxy == (const VIEW_PIX_PTR *) NULL) {
    printf ("bad subxy");
    pcw->Reason = cwrGetCache;
    return MagickTrue;
  }

  for (xy=0; xy < wiht; xy++) {
    // FIXME: Bug: we should use subxy+xy*NumChannels.
    if (GET_PIXEL_ALPHA (pcw->sub_image,subxy) > 0) {
      pcw->Reason = cwrSubAllTrans;
      return MagickFalse;
    }
  }
  pcw->Reason = cwrSuccess;
  return MagickTrue;
}

static void inline ViewPixPtr2Info (Image * img, const VIEW_PIX_PTR * vpp, PIX_INFO * pi)
{
  pi->red   = GET_PIXEL_RED   (img, vpp);
  pi->green = GET_PIXEL_GREEN (img, vpp);
  pi->blue  = GET_PIXEL_BLUE  (img, vpp);
}

static double CompareWindow (
  CompWindT * pcw,
  ExceptionInfo *exception
)
// Compares subimage window with reference starting at given top-left of each.
// Returns positive number MSE; closer to zero is closer match.
// If sx,sy has same number or more transparent pixels as hx,hy,
//   or either window is entirely transparent,
//   returns very large number.
{
  ssize_t
    xy;

  const VIEW_PIX_PTR
    *refxy, *refxySave,
    *subxy, *subxySave;

  // FIXME: V7 oops. In v7, these aren't structures.
  PIX_INFO
    minSub,
    maxSub;

  double
    score = 0;

  int
    wiht = pcw->win_wi * pcw->win_ht,
    cntNonTrans = 0,
    refXmult;
    //subXmult;

//printf ("cw\n");

  //printf ("CompareWindow %li,%li  %li,%li  %lix%li \n",
  //  pcw->refX, pcw->refY,
  //  pcw->subX, pcw->subY,
  //  pcw->win_wi,pcw->win_ht);

  pcw->Reason = cwrSuccess;
  pcw->nCompares++;

  if (   pcw->refX==pcw->subX
      && pcw->refY==pcw->subY
      && !pcw->AllowEquCoord)
  {
    pcw->Reason = cwrEquCoord;
    return pcw->WorstCompare;
  }

//printf ("cw1\n");

  refxy = GetCacheViewVirtualPixels(pcw->ref_view,
    pcw->refX,pcw->refY,
    pcw->win_wi,pcw->win_ht,exception);
  if (refxy == (const VIEW_PIX_PTR *) NULL) {
    printf ("bad refxy\n");
    pcw->Reason = cwrGetCache;
    return pcw->WorstCompare;
  }

//printf ("cw1a\n");

#if IMV6OR7==6
  refXmult = 1;
//  subXmult = 1;
#else
  refXmult = GetPixelChannels (pcw->ref_image);
//  subXmult = GetPixelChannels (pcw->sub_image);
#endif

//printf ("cw1b\n");

  // If central pixel in ref window is transparent, forget about it.
  if (GET_PIXEL_ALPHA (pcw->ref_image,refxy + refXmult*(wiht/2)) <= 0) {
    pcw->Reason = cwrCentTrans;
    return pcw->WorstCompare;
  }

  subxy = GetCacheViewVirtualPixels(pcw->sub_view,
    pcw->subX,pcw->subY,
    pcw->win_wi,pcw->win_ht,exception);
  if (subxy == (const VIEW_PIX_PTR *) NULL) {
    printf ("bad subxy");
    pcw->Reason = cwrGetCache;
    return pcw->WorstCompare;
  }

//printf ("cw2\n");

  refxySave = refxy;
  subxySave = subxy;
  instantPp (pcw->sub_image, &minSub);
  instantPp (pcw->sub_image, &maxSub);

//printf ("cw3\n");

  MagickBooleanType InitMinMax = MagickFalse;
  for (xy=0; xy < wiht; xy++) {
    double refa = GET_PIXEL_ALPHA (pcw->ref_image,refxy);
    double suba = GET_PIXEL_ALPHA (pcw->sub_image,subxy);

    if (refa <= 0 && suba > 0) {
      pcw->Reason = cwrRefTrans;
      return pcw->WorstCompare;
    }

    if (refa > 0 && suba > 0) {
      double d = (GET_PIXEL_RED(pcw->ref_image,refxy)-GET_PIXEL_RED(pcw->sub_image,subxy)) / QuantumRange;
      score += d*d;
      d = (GET_PIXEL_GREEN(pcw->ref_image,refxy)-GET_PIXEL_GREEN(pcw->sub_image,subxy)) / QuantumRange;
      score += d*d;
      d = (GET_PIXEL_BLUE(pcw->ref_image,refxy)-GET_PIXEL_BLUE(pcw->sub_image,subxy)) / QuantumRange;
      score += d*d;
      cntNonTrans++;
    }

    if (pcw->HomChkOn && suba > 0) {
      if (!InitMinMax) {
        // FIXME: in v7, these aren't structures.
//        minSub = *subxy;
//        maxSub = *subxy;

        ViewPixPtr2Info (pcw->sub_image, subxy, &minSub);
        ViewPixPtr2Info (pcw->sub_image, subxy, &maxSub);

        InitMinMax = MagickTrue;
      } else {
        // FIXME: or could record limits of sxy instead of hxy?
        if (minSub.red > GET_PIXEL_RED(pcw->sub_image, subxy))
            minSub.red = GET_PIXEL_RED(pcw->sub_image, subxy);

//          SET_PIXEL_RED(pcw->sub_image,
//                        GET_PIXEL_RED(pcw->sub_image, subxy),
//                        &minSub);

        if (minSub.green > GET_PIXEL_GREEN(pcw->sub_image, subxy))
            minSub.green = GET_PIXEL_GREEN(pcw->sub_image, subxy);

//          SET_PIXEL_GREEN(pcw->sub_image,
//                          GET_PIXEL_GREEN(pcw->sub_image, subxy),
//                          &minSub);

        if (minSub.blue > GET_PIXEL_BLUE(pcw->sub_image, subxy))
            minSub.blue = GET_PIXEL_BLUE(pcw->sub_image, subxy);

//          SET_PIXEL_BLUE(pcw->sub_image,
//                         GET_PIXEL_BLUE(pcw->sub_image, subxy),
//                         &minSub);

        if (maxSub.red < GET_PIXEL_RED(pcw->sub_image, subxy))
            maxSub.red = GET_PIXEL_RED(pcw->sub_image, subxy);

//          SET_PIXEL_RED(pcw->sub_image,
//                        GET_PIXEL_RED(pcw->sub_image, subxy),
//                        &maxSub);

        if (maxSub.green < GET_PIXEL_GREEN(pcw->sub_image, subxy))
            maxSub.green = GET_PIXEL_GREEN(pcw->sub_image, subxy);

//          SET_PIXEL_GREEN(pcw->sub_image,
//                          GET_PIXEL_GREEN(pcw->sub_image, subxy),
//                          &maxSub);

        if (maxSub.blue < GET_PIXEL_BLUE(pcw->sub_image, subxy))
            maxSub.blue = GET_PIXEL_BLUE(pcw->sub_image, subxy);

//          SET_PIXEL_BLUE(pcw->sub_image,
//                         GET_PIXEL_BLUE(pcw->sub_image, subxy),
//                         &maxSub);
      }
    }

    refxy += Inc_ViewPixPtr (pcw->ref_image);
    subxy += Inc_ViewPixPtr (pcw->sub_image);
  }

//printf ("cw4\n");

  if (cntNonTrans == 0) {
    pcw->Reason = cwrNoOpaq;
    return pcw->WorstCompare;
  }

  score /= (cntNonTrans * 3);

  if (pcw->HomChkOn) {
    refxy = refxySave;
    subxy = subxySave;
    int outside = 0;

    // FIXME: Again, in v7 these aren't structures.
    PIX_INFO tppMin, tppMax;
    instantPp (pcw->sub_image, &tppMin);
    instantPp (pcw->sub_image, &tppMax);

    double hcp1 = pcw->HomChk + 1;

//    SET_PIXEL_RED   (pcw->sub_image, hcp1*GET_PIXEL_RED(pcw->sub_image, &minSub)
//                          - pcw->HomChk*GET_PIXEL_RED(pcw->sub_image, &maxSub), &tppMin);
//    SET_PIXEL_GREEN (pcw->sub_image, hcp1*GET_PIXEL_GREEN(pcw->sub_image, &minSub)
//                          - pcw->HomChk*GET_PIXEL_GREEN(pcw->sub_image, &maxSub), &tppMin);
//    SET_PIXEL_BLUE  (pcw->sub_image, hcp1*GET_PIXEL_BLUE(pcw->sub_image, &minSub)
//                          - pcw->HomChk*GET_PIXEL_BLUE(pcw->sub_image, &maxSub), &tppMin);

    tppMin.red   = hcp1*minSub.red   - pcw->HomChk*maxSub.red;
    tppMin.green = hcp1*minSub.green - pcw->HomChk*maxSub.green;
    tppMin.blue  = hcp1*minSub.blue  - pcw->HomChk*maxSub.blue;


//    SET_PIXEL_RED   (pcw->sub_image, hcp1*GET_PIXEL_RED(pcw->sub_image, &maxSub)
//                          - pcw->HomChk*GET_PIXEL_RED(pcw->sub_image, &minSub), &tppMax);
//    SET_PIXEL_GREEN (pcw->sub_image, hcp1*GET_PIXEL_GREEN(pcw->sub_image, &maxSub)
//                          - pcw->HomChk*GET_PIXEL_GREEN(pcw->sub_image, &minSub), &tppMax);
//    SET_PIXEL_BLUE  (pcw->sub_image, hcp1*GET_PIXEL_BLUE(pcw->sub_image, &maxSub)
//                          - pcw->HomChk*GET_PIXEL_BLUE(pcw->sub_image, &minSub), &tppMax);

    tppMax.red   = hcp1*maxSub.red   - pcw->HomChk*minSub.red;
    tppMax.green = hcp1*maxSub.green - pcw->HomChk*minSub.green;
    tppMax.blue  = hcp1*maxSub.blue  - pcw->HomChk*minSub.blue;


    for (xy=0; xy < wiht; xy++) {
      double refa = GET_PIXEL_ALPHA (pcw->ref_image, refxy);
      double suba = GET_PIXEL_ALPHA (pcw->sub_image, subxy);

      if (refa > 0 && suba <= 0) {
//        if      (GET_PIXEL_RED (pcw->ref_image, refxy)
//               < GET_PIXEL_RED (pcw->sub_image, &tppMin)) outside++;
//        else if (GET_PIXEL_RED (pcw->ref_image, refxy)
//               > GET_PIXEL_RED (pcw->sub_image, &tppMax)) outside++;
//
//        if      (GET_PIXEL_GREEN (pcw->ref_image, refxy)
//               < GET_PIXEL_GREEN (pcw->sub_image, &tppMin)) outside++;
//        else if (GET_PIXEL_GREEN (pcw->ref_image, refxy)
//               > GET_PIXEL_GREEN (pcw->sub_image, &tppMax)) outside++;
//
//        if      (GET_PIXEL_BLUE (pcw->ref_image, refxy)
//               < GET_PIXEL_BLUE (pcw->sub_image, &tppMin)) outside++;
//        else if (GET_PIXEL_BLUE (pcw->ref_image, refxy)
//               > GET_PIXEL_BLUE (pcw->sub_image, &tppMax)) outside++;

        if      (GET_PIXEL_RED (pcw->ref_image, refxy) < tppMin.red) outside++;
        else if (GET_PIXEL_RED (pcw->ref_image, refxy) > tppMax.red) outside++;

        if      (GET_PIXEL_GREEN (pcw->ref_image, refxy) < tppMin.green) outside++;
        else if (GET_PIXEL_GREEN (pcw->ref_image, refxy) > tppMax.green) outside++;

        if      (GET_PIXEL_BLUE (pcw->ref_image, refxy)  < tppMin.blue) outside++;
        else if (GET_PIXEL_BLUE (pcw->ref_image, refxy)  > tppMax.blue) outside++;

      }
      // FIXME: Clear bugs. We don't use suba. We don't increment pointer refxy.
    }

    if (outside > 0) {
      pcw->Reason = cwrOutside;
      return pcw->WorstCompare;
    }
    // FIXME? Seems a bit harsh. Maybe make score just a bit worse.
  }

//printf ("end cw\n");

  return score;
}
