#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"

/*
   Latest update:
     1-Feb-2016 for V7.
     20-July-2017 added regardalpha.
     3-April-2018 for v7.0.7-28
*/

typedef struct {
  MagickBooleanType
    do_regardalpha,
    do_decumul,
    do_norm,
    do_verbose;
} cumulhistoT;

static void usage (void)
{
  printf ("Usage: -process 'cumulhisto [OPTION]...'\n");
  printf ("Cumulate (or decumulate) a histogram image.\n");
  printf ("\n");
  printf ("  r, regardalpha      pre-multiply RGB values by alpha\n");
  printf ("  d, decumul          de-cumulate histogram\n");
  printf ("  n, norm             normalise output so maximum count is quantum\n");
  printf ("  v, verbose          write text information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  cumulhistoT * pch
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  pch->do_regardalpha =
    pch->do_decumul =
    pch->do_norm =
    pch->do_verbose = MagickFalse;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "r", "regardalpha")==MagickTrue) {
      pch->do_regardalpha = MagickTrue;
    } else if (IsArg (pa, "n", "norm")==MagickTrue) {
      pch->do_norm = MagickTrue;
    } else if (IsArg (pa, "d", "decumul")==MagickTrue) {
      pch->do_decumul = MagickTrue;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pch->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "cumulhisto: ERROR: unknown option\n");
      status = MagickFalse;
    }
  }

  if (pch->do_verbose) {
    fprintf (stderr, "cumulhisto options:  ");
    if (pch->do_regardalpha) fprintf (stderr, "regardalpha ");
    if (pch->do_decumul) fprintf (stderr, "decumul ");
    if (pch->do_norm)    fprintf (stderr, "norm ");
    if (pch->do_verbose) fprintf (stderr, "verbose ");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


#if defined(MAGICKCORE_HDRI_SUPPORT)
#define ADD_HALF
#else
#define ADD_HALF +0.5
#endif


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *cumulhist (
  const Image *image,
  cumulhistoT * pch,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  CacheView
    *image_view,
    *out_view;

  ssize_t
    y,
    inXmult;

  MagickBooleanType
    status;

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (pch->do_verbose) {
    fprintf (stderr, "cumulhisto: Input image [%s] depth is %i\n",
             image->filename, (int)image->depth);
  }

  // De-cumul needs a new image.
  // Without this option, we could over-write the existing image.

  new_image=CloneImage(image, 0, 0, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  // Just in case image had a palette.
  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  SetAllBlack (new_image, exception);

//#if IMV6OR7==6
//  inXmult = 1;
//  outXmult = 1;
//#else
//  inXmult = GetPixelChannels (image);
//  outXmult = GetPixelChannels (new_image);
//#endif

  inXmult = Inc_ViewPixPtr (image);

  status=MagickTrue;

  out_view=AcquireAuthenticCacheView(new_image,exception);

  image_view=AcquireVirtualCacheView(image,exception);

#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) shared(status) \
    MAGICK_THREADS(image,image,image->rows,1)
#endif
  for (y=0; y < (ssize_t) image->rows; y++)
  {
    const VIEW_PIX_PTR
      *pSave,
      *p;

    VIEW_PIX_PTR
      *q_out;

    ssize_t
      x;

    long double
      mult_fact,
      alpha;

    if (status == MagickFalse)
      continue;
    p=GetCacheViewVirtualPixels(image_view,0,y,image->columns,1,exception);
    pSave = p;
    if (p == (const VIEW_PIX_PTR *) NULL)
    {
      fprintf (stderr, "bad GetCacheViewVirtualPixels image_view\n");
      status=MagickFalse;
      continue;
    }

    q_out=GetCacheViewAuthenticPixels(out_view,0,y,new_image->columns,1,exception);
    if (q_out == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "cumulhisto: bad GetCacheViewAuthenticPixels out_view\n");
      status=MagickFalse;
    }

    mult_fact = 1.0;
    alpha = 1.0;

    // FIXME: we need first pass only if normalising.
    if (pch->do_decumul) {
      long double
        diff,
        max_diff;

      max_diff = 0;

      p += Inc_ViewPixPtr (image);
      for (x=1; x < (ssize_t) image->columns; x++)
      {
// FIXME: subtract xfactor.
        diff = GET_PIXEL_RED(image,p) - GET_PIXEL_RED(image,p-inXmult);
        if (max_diff < diff) max_diff = diff;
        diff = GET_PIXEL_GREEN(image,p) - GET_PIXEL_GREEN(image,p-inXmult);
        if (max_diff < diff) max_diff = diff;
        diff = GET_PIXEL_BLUE(image,p) - GET_PIXEL_BLUE(image,p-inXmult);
        if (max_diff < diff) max_diff = diff;
        p += Inc_ViewPixPtr (image);
      }
      if (pch->do_verbose) {
        fprintf (stderr, "  row=%i  max_diff=%g\n",
                  (int)y,
                  (double)max_diff);
      }
      if (pch->do_norm && max_diff != 0.0) {
        mult_fact = QuantumRange / max_diff;
        if (pch->do_verbose) {
          fprintf (stderr, "  mult_fact=%g\n",
                    (double)mult_fact);
        }
      }
    } else {
      // We are not decumulating.
      long double
        sum_red,
        sum_green,
        sum_blue,
        max_sum;

      sum_red = sum_green = sum_blue = 0.0;

      for (x=0; x < (ssize_t) image->columns; x++)
      {
        if (pch->do_regardalpha)
          alpha = GET_PIXEL_ALPHA(image,p) / (long double)QuantumRange;

        sum_red   += alpha * GET_PIXEL_RED(image,p);
        sum_green += alpha * GET_PIXEL_GREEN(image,p);
        sum_blue  += alpha * GET_PIXEL_BLUE(image,p);
        p += Inc_ViewPixPtr (image);
      }

      max_sum = sum_red;
      if (max_sum < sum_green) max_sum = sum_green;
      if (max_sum < sum_blue)  max_sum = sum_blue;

      if (pch->do_verbose) {
        fprintf (stderr, "  row=%i  sum_red=%g  sum_green=%g  sum_blue=%g\n",
                  (int)y,
                  (double)sum_red, (double)sum_green, (double)sum_blue);
      }

      if (pch->do_norm && max_sum != 0.0) {
        mult_fact = QuantumRange / max_sum;
      }

      if (pch->do_verbose) {
        fprintf (stderr, "  max_sum=%g  mult_fact=%g\n",
                  (double)max_sum, (double)mult_fact);
      }
    }

    p = pSave;
    if (pch->do_decumul) {
      SET_PIXEL_RED   (new_image, GET_PIXEL_RED(image,p)   * mult_fact ADD_HALF,q_out);
      SET_PIXEL_GREEN (new_image, GET_PIXEL_GREEN(image,p) * mult_fact ADD_HALF,q_out);
      SET_PIXEL_BLUE  (new_image, GET_PIXEL_BLUE(image,p)  * mult_fact ADD_HALF,q_out);
      p += Inc_ViewPixPtr (image);
      q_out += Inc_ViewPixPtr (new_image);

      for (x=1; x < (ssize_t) image->columns; x++)
      {
        SET_PIXEL_RED
          (new_image,
           (GET_PIXEL_RED(image,p)   - GET_PIXEL_RED(image,p-inXmult))   * mult_fact ADD_HALF,
           q_out);
        SET_PIXEL_GREEN
          (new_image,
           (GET_PIXEL_GREEN(image,p) - GET_PIXEL_GREEN(image,p-inXmult)) * mult_fact ADD_HALF,
           q_out);
        SET_PIXEL_BLUE
          (new_image,
           (GET_PIXEL_BLUE(image,p)  - GET_PIXEL_BLUE(image,p-inXmult))  * mult_fact ADD_HALF,
           q_out);

        p += Inc_ViewPixPtr (image);
        q_out += Inc_ViewPixPtr (new_image);
      }
    } else {
      long double
        cumul_red,
        cumul_green,
        cumul_blue,
        cumul_alpha,
        alp_q;

      cumul_red = cumul_green = cumul_blue = cumul_alpha = 0;
      for (x=0; x < (ssize_t) image->columns; x++)
      {
        if (pch->do_regardalpha) {
          alp_q = GET_PIXEL_ALPHA(image,p);
          cumul_alpha += alp_q;
          SET_PIXEL_ALPHA (new_image, cumul_alpha ADD_HALF, q_out);
          alpha = alp_q / (long double)QuantumRange;
        } else {
          SET_PIXEL_ALPHA (new_image, GET_PIXEL_ALPHA(image,p), q_out);
        }

        cumul_red += (alpha * GET_PIXEL_RED(image,p));
        SET_PIXEL_RED (new_image, cumul_red * mult_fact ADD_HALF, q_out);

        cumul_green += (alpha * GET_PIXEL_GREEN(image,p));
        SET_PIXEL_GREEN (new_image, cumul_green * mult_fact ADD_HALF, q_out);

        cumul_blue += (alpha * GET_PIXEL_BLUE(image,p));
        SET_PIXEL_BLUE (new_image, cumul_blue * mult_fact ADD_HALF, q_out);

        p += Inc_ViewPixPtr (image);
        q_out += Inc_ViewPixPtr (new_image);
      }
    }

    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }
  }

  image_view=DestroyCacheView(image_view);
  out_view=DestroyCacheView(out_view);

  return (new_image);
}


ModuleExport size_t cumulhistoImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  cumulhistoT
    cumul_histo;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &cumul_histo);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = cumulhist (image, &cumul_histo, exception);

    ReplaceImageInList(&image,new_image);
    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
