/* Written by: Alan Gibson, 18 October 2015.

   References:

     http://im.snibgo.com/darkpath.htm
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "vsn_defines.h"

// FUTURE: We could weight diagonal steps by sqrt(2).

typedef struct {
  int
    max_iter,
    precision;

  MagickBooleanType
    autoIter,
    cumOnly,
    edgeLeft,
    edgeBottom,
    edgeRight,
    endIsGiven,
    do_verbose,
    do_verbose2;

  ssize_t
    endX,
    endY;

} darkestmeanderT;

static void usage (void)
{
  printf ("Usage: -process 'darkestmeander [OPTION]...'\n");
  printf ("Finds darkest meander from top to bottom.\n");
  printf ("\n");
  printf ("  m, maxiter N        maximum number of iterations [10]\n");
  printf ("  a, autoiter         stop iterating when path is stable\n");
  printf ("  c, cumerr           result is cumulative instead of meander\n");
  printf ("  s, side str         str contains any of lbr for left, bottom and right\n");
  printf ("  e, end_at X,Y       end the path at (integer) coordinates instead of a side\n");
  printf ("  v, verbose          write text information to stdout\n");
  printf ("  v2, verbose2        write more text information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  darkestmeanderT * pch
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  pch->autoIter = pch->cumOnly = pch->do_verbose = pch->do_verbose2 = MagickFalse;
  pch->edgeLeft = pch->edgeRight = MagickFalse;
  pch->edgeBottom = MagickTrue;
  pch->endIsGiven = MagickFalse;
  pch->endX = pch->endY = 0;

  pch->max_iter = 10;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "c", "cumerr")==MagickTrue) {
      pch->cumOnly = MagickTrue;
    } else if (IsArg (pa, "m", "maxiter")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i]))
      {
        fprintf (stderr, "Non-numeric argument to 'maxiter'.\n");
        status = MagickFalse;
      }
      else pch->max_iter = atoi (argv[i]);
    } else if (IsArg (pa, "a", "autoiter")==MagickTrue) {
      pch->autoIter = MagickTrue;
    } else if (IsArg (pa, "e", "end_at")==MagickTrue) {
      pch->endIsGiven = MagickTrue;
      i++;
      int n = sscanf (argv[i],"%zd,%zd",&pch->endX,&pch->endY);
      if (n != 2) {
        fprintf (stderr, "'end_at' needs X,Y.\n");
        status = MagickFalse;
      }
      if (pch->endX < 0 || pch->endY < 0) {
        fprintf (stderr, "'end_at' needs positive X,Y.\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "s", "side")==MagickTrue) {
      pch->edgeLeft = pch->edgeBottom = pch->edgeRight = MagickFalse;
      i++;
      const char * p = argv[i];
      while (*p) {
        if (*p=='l' || *p=='L') pch->edgeLeft = MagickTrue;
        else if (*p=='b' || *p=='B') pch->edgeBottom = MagickTrue;
        else if (*p=='r' || *p=='R') pch->edgeRight = MagickTrue;
        else {
          fprintf (stderr, "Argument to 'side' should contain lbr only.\n");
          status = MagickFalse;
        }
        p++;
      }
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pch->do_verbose = MagickTrue;
    } else if (IsArg (pa, "v2", "verbose2")==MagickTrue) {
      pch->do_verbose2 = MagickTrue;
      pch->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "darkestmeander: ERROR: unknown option\n");
      status = MagickFalse;
    }
  }

  if (!pch->edgeLeft && !pch->edgeBottom && !pch->edgeRight) {
    fprintf (stderr, "side: must contain any of lbr\n");
    status = MagickFalse;
  }

  if (pch->do_verbose) {
    fprintf (stderr, "darkestmeander options:");
    if (pch->cumOnly)  fprintf (stderr, "  cumerr");
    fprintf (stderr, "  maxiter %i", pch->max_iter);
    if (pch->autoIter)  fprintf (stderr, "  autoiter");
    fprintf (stderr, "  side ");
    if (pch->edgeLeft) fprintf (stderr, "l");
    if (pch->edgeBottom) fprintf (stderr, "b");
    if (pch->edgeRight) fprintf (stderr, "r");
    if (pch->endIsGiven)  fprintf (stderr, "  end_at %li,%li", pch->endX, pch->endY);
    if (pch->do_verbose2) fprintf (stderr, "  verbose2");
    else if (pch->do_verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


static MagickBooleanType FindMinAtEdge (const Image *image,
  darkestmeanderT * pch,
  CacheView *out_view,
  ssize_t * patX,
  ssize_t * patY,
  ExceptionInfo *exception)
{
  // FIXME? If we have a span of equally good minimums, take the middle one.

  MagickBooleanType
    status = MagickTrue,
    FoundOne;

  ssize_t
    x, y,
    atYL, atXB, atYR,
    atX=0, atY=0;

  long double
    minVal,
    minValL, minValB, minValR;

  VIEW_PIX_PTR
    *q_out;

  FoundOne = MagickFalse;
  minVal = 999;

  if (pch->do_verbose2) printf ("FindMinAtEdge\n");

  if (pch->edgeBottom) {
    if (pch->do_verbose2)
      printf ("Find minimum value on the bottom row.\n");

    q_out=GetCacheViewAuthenticPixels(out_view,0,image->rows-1,image->columns,1,exception);
    if (q_out == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "darkestmeander: bad q_out\n");
      status=MagickFalse;
    }

    minValB = GET_PIXEL_RED(image,q_out);
    atXB = 0;

    for (x=0; x < (ssize_t) image->columns; x++)
    {
      long double
        thisVal;

      thisVal = GET_PIXEL_RED(image,q_out);
      if (minValB > thisVal) {
        minValB = thisVal;
        atXB = x;
      }

      q_out += Inc_ViewPixPtr (image);
    }

    if (FoundOne == MagickFalse) {
      FoundOne = MagickTrue;
      minVal = minValB;
      atX = atXB;
      atY = image->rows-1;
    }

    if (pch->do_verbose)
      printf ("Min value on bottom row %.*g at x=%i.\n",
       pch->precision, (double)minValB,
       (int)atXB);
  }

  if (pch->edgeLeft) {
    if (pch->do_verbose2) {
      printf ("Find minimum value on the left edge.\n");
    }

    q_out=GetCacheViewAuthenticPixels(out_view,0,0,1,image->rows,exception);
    if (q_out == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "darkestmeander: bad q_out\n");
      status=MagickFalse;
    }

    minValL = GET_PIXEL_RED(image,q_out);
    atYL = 0;

    for (y=0; y < (ssize_t) image->rows; y++)
    {
      long double
        thisVal;

      thisVal = GET_PIXEL_RED(image,q_out);
      if (minValL > thisVal) {
        minValL = thisVal;
        atYL = y;
      }

      q_out += Inc_ViewPixPtr (image);
    }

    if (FoundOne == MagickFalse || minVal > minValL) {
      FoundOne = MagickTrue;
      minVal = minValL;
      atX = 0;
      atY = atYL;
    }

    if (pch->do_verbose)
      printf ("Min value on left %.*g at y=%i.\n",
       pch->precision, (double)minValL,
       (int)atYL);
  }

  if (pch->edgeRight) {
    if (pch->do_verbose2) {
      printf ("Find minimum value on the right edge.\n");
    }

    q_out=GetCacheViewAuthenticPixels(out_view,image->columns-1,0,1,image->rows,exception);
    if (q_out == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "darkestmeander: bad q_out\n");
      status=MagickFalse;
    }

    minValR = GET_PIXEL_RED(image,q_out);
    atYR = 0;

    for (y=0; y < (ssize_t) image->rows; y++)
    {
      long double
        thisVal;

      thisVal = GET_PIXEL_RED(image,q_out);
      if (minValR > thisVal) {
        minValR = thisVal;
        atYR = y;
      }

      q_out += Inc_ViewPixPtr (image);
    }

    if (FoundOne == MagickFalse || minVal > minValR) {
      FoundOne = MagickTrue;
      minVal = minValR;
      atX = image->columns-1;
      atY = atYR;
    }

    if (pch->do_verbose)
      printf ("Min value on right %.*g at y=%i.\n",
       pch->precision, (double)minValR,
       (int)atYR);

  }

  *patX = atX;
  *patY = atY;

  if (FoundOne) {
    if (pch->do_verbose)
      printf ("Min value at %i,%i\n", (int)atX, (int)atY);
  } else {
    status = MagickFalse;
  }

  return (status);
}



static MagickBooleanType CalcPathCumul (
  const Image *image,
  const Image *new_image,
  darkestmeanderT * pch,
  CacheView *out_view,
  ExceptionInfo *exception)
{
  // Calculates cumulation, as for darkestpath, more or less.

  ssize_t
    y;

  MagickBooleanType
    status = MagickTrue;

  if (pch->do_verbose2) printf ("CalcPathCumul: Get minimum from adjacent pixels on previous row.\n");

  // FIXME: green and blue values in row 0 are junk?

  // In q_out, we use blue and green channels to record delta x,y of parent of each pixel.
  // Values in these channels are 0, 50% or 100% representing delta of -1, 0, +1.
  // This simplifies finding the path after q_out is populated and stable.

  for (y=1; y < (ssize_t) image->rows; y++)
  {
    ssize_t
      x;

    long double
      minVal,
      prevVal;

    VIEW_PIX_PTR
      *q_out;

    //printf ("%i ", (int)y);

    const VIEW_PIX_PTR
      *p;

    if (status == MagickFalse)
      continue;

    p=GetCacheViewVirtualPixels(out_view,0,y-1,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL)
    {
      fprintf (stderr, "bad GetCacheViewVirtualPixels out_view\n");
      status=MagickFalse;
      continue;
    }

    q_out=GetCacheViewAuthenticPixels(out_view,0,y,image->columns,1,exception);
    if (q_out == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "darkestmeander: bad GetCacheViewAuthenticPixels out_view\n");
      status=MagickFalse;
    }

    for (x=0; x < (ssize_t) image->columns; x++)
    {
      signed int pv = 0;
      // Get minimum from adjacent pixels on previous row.
      minVal = GET_PIXEL_RED(image,p);
      if (x > 0) {
        prevVal = GET_PIXEL_RED(image,p-1);
        if (minVal > prevVal) {
          minVal = prevVal;
          pv = -1;
        }
      }
      if (x < image->columns-1) {
        prevVal = GET_PIXEL_RED(image,p+1);
        if (minVal > prevVal) {
          minVal = prevVal;
          pv = +1;
        }
      }

      if (minVal==0) printf ("minval==0 at %i,%i\n", (int)x,(int)y);

      SET_PIXEL_RED   (new_image, GET_PIXEL_RED(new_image,q_out)+minVal, q_out);
      SET_PIXEL_GREEN (new_image, 0, q_out);
      SET_PIXEL_BLUE  (new_image, (pv+1)/2.0*QuantumRange, q_out);

      p += Inc_ViewPixPtr (image);
      q_out += Inc_ViewPixPtr (new_image);
    }

    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }
  }

  return (status);
}


static MagickBooleanType Calc4Cumul (
  const Image *image,
  const Image *new_image,
  darkestmeanderT * pch,
  CacheView *image_view,
  CacheView *out_view,
  int * pnChanged,
  ExceptionInfo *exception)
{
  // Calculates cumulation in four directions.
  // Returns the number of pixels updated.

  MagickBooleanType
    status = MagickTrue;

  VIEW_PIX_PTR
    *q_out;

  int nIter;

  *pnChanged = 0;

  for (nIter=1; nIter <= 4; nIter++) {
    ssize_t
      startU, endU, dU,
      startV, endV, dV,
      u, v,
      cacheWi, cacheHt;

    int nCh = 0;

    switch (nIter % 4) {
      case 0:
      if (pch->do_verbose2) printf ("Down ");
      startU = 1;
      endU = (ssize_t)image->rows;
      dU = +1;
      startV = 0;
      endV = (ssize_t)image->columns;
      dV = +1;
      cacheWi = (ssize_t)image->columns;
      cacheHt = 1;
      break;

      case 1:
      if (pch->do_verbose2) printf ("Right ");
      startU = 1;
      endU = (ssize_t)image->columns;
      dU = +1;
      startV = 0;
      endV = (ssize_t)image->rows;
      dV = +1;
      cacheWi = 1;
      cacheHt = (ssize_t)image->rows;
      break;

      case 2:
      if (pch->do_verbose2) printf ("Up ");
      startU = (ssize_t)image->rows-2;
      endU = -1;
      dU = -1;
      startV = (ssize_t)image->columns-1;
      endV = -1;
      dV = -1;
      cacheWi = (ssize_t)image->columns;
      cacheHt = 1;
      break;

      case 3:
      default:
      if (pch->do_verbose2) printf ("Left ");
      startU = (ssize_t)image->columns-2;
      endU = -1;
      dU = -1;
      startV = (ssize_t)image->rows-1;
      endV = -1;
      dV = -1;
      cacheWi = 1;
      cacheHt = (ssize_t)image->rows;
      break;
    }

    // u and v are proxies for y and x respectively.

    for (u=startU; u != endU; u+=dU) {
      ssize_t
        cacheT, cacheL,
        cacheTm1, cacheLm1;

      const VIEW_PIX_PTR
        *p, *qm1;

      switch (nIter % 4) {
        case 0:
        cacheT = u;
        cacheL = 0;
        cacheTm1 = u-1;
        cacheLm1 = 0;
        break;

        case 1:
        cacheT = 0;
        cacheL = u;
        cacheTm1 = 0;
        cacheLm1 = u-1;
        break;

        case 2:
        cacheT = u;
        cacheL = 0;
        cacheTm1 = u+1;
        cacheLm1 = 0;
        break;

        case 3:
        default:
        cacheT = 0;
        cacheL = u;
        cacheTm1 = 0;
        cacheLm1 = u+1;
        break;
      }

      p=GetCacheViewVirtualPixels(image_view,cacheL,cacheT,cacheWi,cacheHt,exception);
      if (p == (const VIEW_PIX_PTR *) NULL)
      {
        fprintf (stderr, "bad GetCacheViewVirtualPixels image_view\n");
        status=MagickFalse;
        continue;
      }

      qm1=GetCacheViewVirtualPixels(out_view,cacheLm1,cacheTm1,cacheWi,cacheHt,exception);
      if (qm1 == (const VIEW_PIX_PTR *) NULL)
      {
        fprintf (stderr, "bad GetCacheViewVirtualPixels qm1\n");
        status=MagickFalse;
        continue;
      }

      q_out=GetCacheViewAuthenticPixels(out_view,cacheL,cacheT,cacheWi,cacheHt,exception);
      if (q_out == (const VIEW_PIX_PTR *) NULL)
      {
        fprintf (stderr, "bad GetCacheViewAthenticPixels q_out\n");
        status=MagickFalse;
        continue;
      }

      for (v=startV; v != endV; v+=dV) {
        long double
          minVal,
          prevVal;

        signed int pu=0, pv=0;

        minVal = (GET_PIXEL_RED(image,q_out)-GET_PIXEL_RED(image,p)) * 0.99999;

        prevVal = GET_PIXEL_RED(image,qm1);
        if (minVal > prevVal) {
          minVal = prevVal;
          pu=-1;
          pv=0;
        }


// FIXME: for v7, -1 or +1 should be  Inc_ViewPixPtr(image);

        if (v > startV) {
          prevVal = GET_PIXEL_RED(image,qm1-Inc_ViewPixPtr(image));
          if (minVal > prevVal) {
            minVal = prevVal;
            pu=-1;
            pv=-1;
          }
        }

        if (v < endV-1) {
          prevVal = GET_PIXEL_RED(image,qm1+Inc_ViewPixPtr(image));
          if (minVal > prevVal) {
            minVal = prevVal;
            pu=-1;
            pv=+1;
          }
        }

        if (v > startV) {
          prevVal = GET_PIXEL_RED(image,q_out-Inc_ViewPixPtr(image));
          if (minVal > prevVal) {
            minVal = prevVal;
            pu=0;
            pv=-1;
          }
        }

        if (v < endV-1) {
          prevVal = GET_PIXEL_RED(image,q_out+Inc_ViewPixPtr(image));
          if (minVal > prevVal) {
            minVal = prevVal;
            pu=0;
            pv=+1;
          }
        }

        if (pu != 0 || pv != 0) {
          pu *= dU;
          pv *= dV;

          if (minVal >= 0 || GET_PIXEL_RED(image,p) >= 0) {

            SET_PIXEL_RED (new_image, GET_PIXEL_RED(image,p)+minVal, q_out);

            if (nIter % 2 == 1) {
              signed int t = pu;
              pu = pv;
              pv = t;
            }
            SET_PIXEL_GREEN (new_image, (pu+1)/2.0*QuantumRange, q_out);
            SET_PIXEL_BLUE  (new_image, (pv+1)/2.0*QuantumRange, q_out);
            nCh++;
          }
        }

        p += Inc_ViewPixPtr (image);
        qm1 += Inc_ViewPixPtr (image);
        q_out += Inc_ViewPixPtr (image);
      }

      if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse) {
        fprintf (stderr, "bad sync\n");
        status=MagickFalse;
      }
    }

    if (pch->do_verbose2) printf ("%i nCh=%i\n", nIter, nCh);
    *pnChanged += nCh;
  }

  if (pch->do_verbose2) printf ("Calc4Cumul pnChanged=%i\n", *pnChanged);

  return (status);
}


static MagickBooleanType PaintPath (
  const Image *current_path_image,
  const Image *new_path_image,
  darkestmeanderT * pch,
  CacheView *out_view,
  CacheView *path_view,
  ssize_t atX,
  ssize_t atY,
  MagickBooleanType *Cyclic,
  ExceptionInfo *exception)
{
  ssize_t
    y,
    x;

  VIEW_PIX_PTR
    *q_path;

  MagickBooleanType
    status = MagickTrue;

  *Cyclic = MagickFalse;

  if (pch->do_verbose2) printf ("PaintPath: Set path end\n");

  // Get just the one pixel we need.
  q_path=GetCacheViewAuthenticPixels(path_view,atX,atY,1,1,exception);
  if (q_path == (const VIEW_PIX_PTR *) NULL) {
    fprintf (stderr, "darkestmeander: bad GetCacheViewAuthenticPixels q_path\n");
    status=MagickFalse;
  }

  SET_PIXEL_RED   (new_path_image, QuantumRange, q_path);
  SET_PIXEL_GREEN (new_path_image, QuantumRange, q_path);
  SET_PIXEL_BLUE  (new_path_image, QuantumRange, q_path);

  if (SyncCacheViewAuthenticPixels(path_view,exception) == MagickFalse) {
    fprintf (stderr, "bad sync\n");
    status=MagickFalse;
  }


  if (pch->do_verbose2) printf ("PaintPath: Walk up the image, painting the path white.\n");

  // Walk back through the meander, painting the path white.
  // At each point, find the parent.

  x = atX;
  y = atY;
  while (y > 0 && *Cyclic == MagickFalse && status == MagickTrue) {
    const VIEW_PIX_PTR
      *p;

    signed int di, dj;

    //if (pch->do_verbose2) printf ("PaintPath: xy=%i,%i  ", (int)x, (int)y);

    p=GetCacheViewVirtualPixels(out_view,x,y,1,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "darkestmeander: bad GetCacheViewVirtualPixels p\n");
      status=MagickFalse;
    }
    di = (signed int)(GET_PIXEL_BLUE  (current_path_image, p) / QuantumRange * 2.0 + 0.5) -1;
    dj = (signed int)(GET_PIXEL_GREEN (current_path_image, p) / QuantumRange * 2.0 + 0.5) -1;

    x += di;
    y += dj;

    //if (pch->do_verbose2) printf (" dij=%i,%i W%i,%i\n", di, dj, (int)x, (int)y);

    // Get just the one pixel we need.
    q_path=GetCacheViewAuthenticPixels(path_view,x,y,1,1,exception);
    if (q_path == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "PaintPath: bad GetCacheViewAuthenticPixels path_view\n");
      status=MagickFalse;
    }

    if (GET_PIXEL_RED (new_path_image, q_path) > 0) {
      if (pch->do_verbose) fprintf (stderr, "PaintPath: Path not black at %i,%i: cyclic\n", (int)x, (int)y);
      y = 0;
      *Cyclic = MagickTrue;
    }

    SET_PIXEL_RED   (new_path_image, QuantumRange, q_path);
    SET_PIXEL_GREEN (new_path_image, QuantumRange, q_path);
    SET_PIXEL_BLUE  (new_path_image, QuantumRange, q_path);

    if (SyncCacheViewAuthenticPixels(path_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }

  }

  if (pch->do_verbose2) printf ("PaintPath: done\n");

  return (status);
}


static Image *TryPath(Image *current_path_image,
  darkestmeanderT * pch,
  ssize_t atX,
  ssize_t atY,
  CacheView *out_view,
  MagickBooleanType *IsEqual,
  ExceptionInfo *exception)
// Makes a new path, and compare it to current_path_image.
// Destroys current_path_image, and returns the new path image (or NULL if failure).
{
  Image
    *new_path_image;

  CacheView
    *new_path_view;

  MagickBooleanType
    Cyclic;

  if (pch->do_verbose2) printf ("TryPath\n");

  // Create a new black image for the path.

  // Make a clone of this image, same size but undefined pixel values:
  //
  new_path_image = CloneImage (
    current_path_image,
    current_path_image->columns, current_path_image->rows,
    MagickTrue, exception);
  if (new_path_image == (Image *) NULL)
    return(new_path_image);

  if (SetNoPalette (new_path_image, exception) == MagickFalse)
    return (Image *)NULL;

  SetAllBlack (new_path_image, exception);

  new_path_view=AcquireAuthenticCacheView(new_path_image,exception);

  if (PaintPath (
       current_path_image, new_path_image,
       pch, out_view, new_path_view, atX, atY, &Cyclic, exception
                )==MagickFalse)
  {
    return (Image *)NULL;
  }

  *IsEqual = (Cyclic == MagickFalse)
              &&
#if IMV6OR7==6
              IsImagesEqual (new_path_image, current_path_image);
#else
              IsImagesEqual (new_path_image, current_path_image, exception);
#endif

  new_path_view=DestroyCacheView(new_path_view);

  DestroyImageList(current_path_image);

  return (new_path_image);
}


static MagickBooleanType CalcCumul (
  const Image *image,
  const Image *new_image,
  Image **current_path_image,
  darkestmeanderT * pch,
  CacheView *image_view,
  CacheView *out_view,
  ExceptionInfo *exception)
{
  MagickBooleanType
    status = MagickTrue;

  ssize_t
    prevAtX=-1, prevAtY=-1;

  status = CalcPathCumul (image, new_image, pch, out_view, exception);
  if (status == MagickFalse) return MagickFalse;

  if (pch->do_verbose2) printf ("CalcCumul: Get meander minimums.\n");

  int nIterCyc = 0;
  int nChanged = 0;

  MagickBooleanType IsEqual = MagickFalse;

  do {
    status = Calc4Cumul (image, new_image, pch, image_view, out_view, &nChanged, exception);
    nIterCyc++;

    if (pch->autoIter) {
      ssize_t
        atX, atY;

      if (pch->endIsGiven) {
        atX = pch->endX;
        atY = pch->endY;
      } else {
        status = FindMinAtEdge (image, pch, out_view, &atX, &atY, exception);
        if (status == MagickFalse) return MagickFalse;
      }

      //status = FindMinAtEdge (image, pch, out_view, &atX, &atY, exception);
      //if (status == MagickFalse) return MagickFalse;

      if (prevAtX == atX && prevAtY == atY) {
        *current_path_image =
        TryPath(*current_path_image, pch, atX, atY, out_view, &IsEqual, exception);
      }

      if (pch->do_verbose)
        printf ("CalcCumul: nIterCyc=%i at %i,%i IsEqual=%i\n",
                nIterCyc, (int)atX, (int)atY, (int)IsEqual);

      prevAtX = atX;
      prevAtY = atY;
    }
  } while (   nChanged > 0
           && IsEqual==MagickFalse
           && (pch->max_iter == 0 || nIterCyc < pch->max_iter));

  if (pch->do_verbose) {
    printf ("darkestmeander: nIterCyc=%i\n", nIterCyc);
    if (nChanged == 0) {
      printf ("darkestmeander: completed\n");
    }
    else if (pch->max_iter > 0 && nIterCyc >= pch->max_iter) {
      printf ("darkestmeander: nIterCyc bust: %i\n", nIterCyc);
    }
    else if (IsEqual) {
      printf ("darkestmeander: path stable\n");
    }
  }

  return (status);
}


static MagickBooleanType EnsureNonZero (Image *image,
  darkestmeanderT * pch,
  ExceptionInfo *exception)
{
  CacheView
    *image_view;

  ssize_t
    y;

  MagickBooleanType
    status = MagickTrue;

#if defined(MAGICKCORE_HDRI_SUPPORT)
#define NON_ZERO_VALUE 1.0
#else
#define NON_ZERO_VALUE 1
#endif

  if (pch->do_verbose2) {
    printf ("EnsureNonZero: %g\n", NON_ZERO_VALUE);
  }

  image_view=AcquireAuthenticCacheView(image,exception);

  for (y=0; y < (ssize_t) image->rows; y++)
  {
    ssize_t
      x;

    VIEW_PIX_PTR
      *p;

    if (status == MagickFalse)
      continue;

    p=GetCacheViewAuthenticPixels(image_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "EnsureNonZero: bad GetCacheViewAuthenticPixels p\n");
      status=MagickFalse;
    }

    for (x=0; x < (ssize_t) image->columns; x++)
    {
      if (GET_PIXEL_RED(image,p) == 0)
        SET_PIXEL_RED(image, NON_ZERO_VALUE, p);
      p += Inc_ViewPixPtr (image);
    }

    if (SyncCacheViewAuthenticPixels(image_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }
  }

  image_view=DestroyCacheView(image_view);

  return (status);
}


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *darkestmeander(Image *image,
  darkestmeanderT * pch,
  ExceptionInfo *exception)
{
  Image
    *new_image,
    *path_image;

  CacheView
    *image_view,
    *out_view;


  ssize_t
    atX, atY;

  MagickBooleanType
    status = MagickTrue;


  pch->precision = GetMagickPrecision();

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (!Has3Channels (image)) {
    fprintf (stderr, "Image has less than three channels.\n");
    return (Image *)NULL;
  }

  // No good, because it knocks out negative values.
  // We must change _only_ where Red is exactly zero.
  // status = EvaluateImageChannel(image, RedChannel, MaxEvaluateOperator, 1, exception);
  // if (status==MagickFalse) return (Image *) NULL;


  EnsureNonZero (image, pch, exception);

  //const ChannelType channel,const MagickEvaluateOperator op,const double value,
  //ExceptionInfo *exception)

  // Copy the current image.

  new_image=CloneImage(image, 0, 0, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  //GetMagickVIEW_PIX_PTR(new_image, &mppBlack);
  //(void)SetImageColor(new_image, &mppBlack);

  status=MagickTrue;

  out_view=AcquireAuthenticCacheView(new_image,exception);

  image_view=AcquireVirtualCacheView(image,exception);


  if (pch->do_verbose2) printf ("Create a new black image for the path.\n");

  // Create a new black image for the path.

  // Make a clone of this image, same size but undefined pixel values:
  //
  path_image=CloneImage(image, image->columns, image->rows, MagickTrue, exception);
  if (path_image == (Image *) NULL)
    return(path_image);

  if (SetNoPalette (path_image, exception) == MagickFalse)
    return (Image *)NULL;

  SetAllBlack (path_image, exception);


  status = CalcCumul (image, new_image, &path_image, pch, image_view, out_view, exception);
  if (status == MagickFalse) return (Image *)NULL;

  if (pch->cumOnly) {

    if (pch->do_verbose2) printf ("End.\n");

    image_view=DestroyCacheView(image_view);
    out_view=DestroyCacheView(out_view);

    DestroyImageList(path_image);
    return (new_image);

  } else {

    CacheView
      *path_view;

    MagickBooleanType
      Cyclic;

    if (pch->endIsGiven) {
      atX = pch->endX;
      atY = pch->endY;
    } else {
      status = FindMinAtEdge (image, pch, out_view, &atX, &atY, exception);
      if (status == MagickFalse) return (Image *)NULL;
    }

    // FIXME: maybe we have a good path already.

    SetAllBlack (path_image, exception);

    path_view=AcquireAuthenticCacheView(path_image,exception);

    status = PaintPath (image, path_image, pch, out_view, path_view, atX, atY, &Cyclic, exception);
    if (status == MagickFalse) return (Image *)NULL;

    if (Cyclic == MagickTrue) {
      fprintf (stderr, "darkestmeander: Path is cyclic at %i,%i\n", (int)atX, (int)atY);
    }

    path_view=DestroyCacheView(path_view);
    image_view=DestroyCacheView(image_view);
    out_view=DestroyCacheView(out_view);

    DestroyImageList(new_image);
    return (path_image);
  }
}


ModuleExport size_t darkestmeanderImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  darkestmeanderT
    dm;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

#if defined(MAGICKCORE_HDRI_SUPPORT)
#else
  fprintf (stderr, "No HDRI. darkestmeander may not work correctly.");
#endif

  status = menu (argc, argv, &dm);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    if (dm.endIsGiven) {
      if (dm.endX >= image->columns || dm.endY >= image->rows) {
        fprintf (stderr, "Given end is outside image.\n");
        return (-1);
      }
    }

    new_image = darkestmeander (image, &dm, exception);
    if (new_image == (Image *)NULL) return -1;

    ReplaceImageInList(&image,new_image);
    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
