/* Written by: Alan Gibson, 18 October 2015.

   References:

     http://graphics.cs.cmu.edu/people/efros/research/quilting/quilting.pdf
     Image Quilting for Texture Synthesis and Transfer</a>,
     Alexei A. Efros and William T. Freeman, 2001.

     http://im.snibgo.com/darkpath.htm

  Updated:
    21-May-2016
      For v7.
      To use intensity instead of merely red channel.
      Fix offset-1 bug in lowest value of bottom row.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"

// FUTURE: We could weight diagonal steps by sqrt(2).

typedef struct {
  MagickBooleanType
    do_cumerr,
    do_verbose;
} darkestPathT;

static void usage (void)
{
  printf ("Usage: -process 'darkestpath [OPTION]...'\n");
  printf ("Finds darkest path from top to bottom.\n");
  printf ("\n");
  printf ("  c, cumerr           result is cumulative error instead of path\n");
  printf ("  v, verbose          write text information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  darkestPathT * pch
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  pch->do_cumerr = pch->do_verbose = MagickFalse;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "c", "cumerr")==MagickTrue) {
      pch->do_cumerr = MagickTrue;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pch->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "darkestpath: ERROR: unknown option\n");
      status = MagickFalse;
    }
  }

  if (pch->do_verbose) {
    fprintf (stderr, "darkestpath options:  ");
    if (pch->do_cumerr)  fprintf (stderr, "cumerr ");
    if (pch->do_verbose) fprintf (stderr, "verbose ");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


#if defined(MAGICKCORE_HDRI_SUPPORT)
#define ADD_HALF
#else
#define ADD_HALF +0.5
#endif


static void inline SetPixIntensity (
  const Image *image,
  double v,
  VIEW_PIX_PTR *q_out)
{
  if (Has3Channels (image)) {
    SET_PIXEL_RED   (image, v, q_out);
    SET_PIXEL_GREEN (image, v, q_out);
    SET_PIXEL_BLUE  (image, v, q_out);
  } else {
    SET_PIXEL_RED   (image, v, q_out);
  }
}


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *darkestpath (const Image *image,
  darkestPathT * pch,
  ExceptionInfo *exception)
{
  Image
    *new_image,
    *path_image;

  CacheView
    *out_view,
    *path_view;

  ssize_t
    y,
    x,
    atX;

  MagickBooleanType
    status;

  VIEW_PIX_PTR
    *q_out,
    *q_path;

  long double
    minVal,
    prevVal;

  int
    precision;


  precision = GetMagickPrecision();

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  // Copy the current image.

  new_image=CloneImage(image, 0, 0, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  //if (Has3Channels (new_image) == MagickFalse) {
  //  fprintf (stderr, "Image has less than 3 channels.\n");
  //  return (Image *)NULL;
  //}

  status=MagickTrue;

  out_view=AcquireAuthenticCacheView(new_image,exception);

  if (pch->do_verbose) printf ("Get minimum from adjacent pixels on previous row.\n");

  for (y=1; y < (ssize_t) image->rows; y++)
  {
    //printf ("%i ", (int)y);

    const VIEW_PIX_PTR
      *p;

    if (status == MagickFalse)
      continue;
    p=GetCacheViewVirtualPixels(out_view,0,y-1,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL)
    {
      fprintf (stderr, "bad GetCacheViewVirtualPixels out_view\n");
      status=MagickFalse;
      continue;
    }

    q_out=GetCacheViewAuthenticPixels(out_view,0,y,new_image->columns,1,exception);
    if (q_out == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "darkestpath: bad GetCacheViewAuthenticPixels out_view\n");
      status=MagickFalse;
    }

// FIXME: for v7, -1 and +1 are wrong

    for (x=0; x < (ssize_t) image->columns; x++)
    {
      // Get minimum from adjacent pixels on previous row.
      //minVal = GET_PIXEL_RED(image,p);
      minVal = GetPixelIntensity(new_image, p);
      if (x > 0) {
        //prevVal = GET_PIXEL_RED(new_image,p-1);
        prevVal = GetPixelIntensity(new_image, p - Inc_ViewPixPtr(new_image));

        if (minVal > prevVal) {
          minVal = prevVal;
        }
      }
      if (x < image->columns-1) {
        //prevVal = GET_PIXEL_RED(new_image,p+1);
        prevVal = GetPixelIntensity(new_image, p + Inc_ViewPixPtr(new_image));
        if (minVal > prevVal) {
          minVal = prevVal;
        }
      }
      //SET_PIXEL_RED (new_image, GET_PIXEL_RED(new_image,q_out)+minVal, q_out);
      double intens = GetPixelIntensity(new_image,q_out)+minVal;
      SetPixIntensity (new_image, intens, q_out);

      p += Inc_ViewPixPtr (new_image);
      q_out += Inc_ViewPixPtr (new_image);
    }

    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }
  }

  // Find minimum value on the bottom row.
  // FIXME? If we have a span of equally good minimums, take the middle one.

  if (pch->do_verbose) printf ("Find minimum value on the bottom row.\n");

  q_out=GetCacheViewAuthenticPixels(out_view,0,image->rows-1,new_image->columns,1,exception);
  if (q_out == (const VIEW_PIX_PTR *) NULL) {
    fprintf (stderr, "darkestpath: bad GetCacheViewAuthenticPixels out_view\n");
    status=MagickFalse;
  }

  //minVal = GET_PIXEL_RED(new_image,q_out);
  minVal = GetPixelIntensity(new_image, q_out);
  atX = 0;

  for (x=0; x < (ssize_t) image->columns; x++)
  {
    long double
      thisVal;

    //thisVal = GET_PIXEL_RED(new_image,q_out);
    thisVal = GetPixelIntensity(new_image, q_out);
    if (minVal > thisVal) {
      minVal = thisVal;
      atX = x;
    }

    q_out += Inc_ViewPixPtr (new_image);
  }

  if (pch->do_verbose)
    printf ("Min value on bottom row %.*g at x=%i.\n",
     precision, (double)minVal,
     (int)atX);

  if (pch->do_verbose) printf ("Create a new black image for the path.\n");

  // Create a new black image for the path.

  // Make a clone of this image, same size but undefined pixel values:
  //
  path_image=CloneImage(image, image->columns, image->rows, MagickTrue, exception);
  if (path_image == (Image *) NULL)
    return(path_image);

  if (SetNoPalette (path_image, exception) == MagickFalse)
    return (Image *)NULL;

  SetAllBlack (path_image, exception);

  path_view=AcquireAuthenticCacheView(path_image,exception);



  if (pch->do_verbose) printf ("Set path end.\n");

  // Get just the one pixel we need.
  q_path=GetCacheViewAuthenticPixels(path_view,atX,new_image->rows-1,1,1,exception);
  if (q_path == (const VIEW_PIX_PTR *) NULL) {
    fprintf (stderr, "darkestpath: bad GetCacheViewAuthenticPixels path_view\n");
    status=MagickFalse;
  }

  //SET_PIXEL_RED   (path_image, QuantumRange, q_path);
  //SET_PIXEL_GREEN (path_image, QuantumRange, q_path);
  //SET_PIXEL_BLUE  (path_image, QuantumRange, q_path);
  SetPixIntensity (path_image, QuantumRange, q_path);

  if (SyncCacheViewAuthenticPixels(path_view,exception) == MagickFalse) {
    fprintf (stderr, "bad sync\n");
    status=MagickFalse;
  }


  if (pch->do_verbose) printf ("Walk up the image, painting the path white.\n");

  // Walk up the image, painting the path white.

  for (y=1; y < (ssize_t) image->rows; y++)
  {
    signed int
      dx;

    q_out=GetCacheViewAuthenticPixels(out_view,0,new_image->rows-y-1,new_image->columns,1,exception);
    if (q_out == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "darkestpath: bad GetCacheViewAuthenticPixels out_view\n");
      status=MagickFalse;
    }

// FIXME: for v7, +atX-1 etc are wrong

    // Get position of minimum.
    //minVal = GET_PIXEL_RED(image,q_out+atX);
    minVal = GetPixelIntensity(new_image, q_out + (atX)*Inc_ViewPixPtr(new_image));
    dx = 0;
    if (atX > 0) {
      //prevVal = GET_PIXEL_RED(image,q_out+atX-1);
      prevVal = GetPixelIntensity(new_image, q_out + (atX-1)*Inc_ViewPixPtr(new_image));
      if (minVal > prevVal) {
        minVal = prevVal;
        dx = -1;
      }
    }
    if (atX < image->columns-1) {
      //prevVal = GET_PIXEL_RED(image,q_out+atX+1);
      prevVal = GetPixelIntensity(new_image, q_out + (atX+1)*Inc_ViewPixPtr(new_image));
      if (minVal > prevVal) {
        minVal = prevVal;
        dx = +1;
      }
    }

    // FIXME: get just the one pixel we need?
    q_path=GetCacheViewAuthenticPixels(path_view,0,new_image->rows-y-1,new_image->columns,1,exception);
    if (q_path == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "darkestpath: bad GetCacheViewAuthenticPixels path_view\n");
      status=MagickFalse;
    }

    atX += dx;

    //if (pch->do_verbose) printf ("%i ", (int)atX);

//    SetPixelRed   (q_path+atX, QuantumRange);
//    SetPixelGreen (q_path+atX, QuantumRange);
//    SetPixelBlue  (q_path+atX, QuantumRange);
    //SET_PIXEL_RED   (path_image, QuantumRange, q_path+atX*Inc_ViewPixPtr(path_image));
    //SET_PIXEL_GREEN (path_image, QuantumRange, q_path+atX*Inc_ViewPixPtr(path_image));
    //SET_PIXEL_BLUE  (path_image, QuantumRange, q_path+atX*Inc_ViewPixPtr(path_image));
    SetPixIntensity (path_image, QuantumRange, q_path+atX*Inc_ViewPixPtr(path_image));

    if (SyncCacheViewAuthenticPixels(path_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }
  }

  if (pch->do_verbose) printf ("End.\n");

  path_view=DestroyCacheView(path_view);
  out_view=DestroyCacheView(out_view);

  if (pch->do_cumerr) {
    DestroyImageList(path_image);
    return (new_image);
  } else {
    DestroyImageList(new_image);
    return (path_image);
  }
}


ModuleExport size_t darkestpathImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  darkestPathT
    cumul_histo;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &cumul_histo);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = darkestpath (image, &cumul_histo, exception);
    if (new_image == (Image *) NULL) return (-1);

    ReplaceImageInList(&image,new_image);
    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
