/* Written by: Alan Gibson, 18 October 2015.

   References:
     https://en.wikipedia.org/wiki/Dijkstra's_algorithm
     Wikipedia: Dijkstra's algorithm.

     http://im.snibgo.com/darkpath.htm

  Updated:
    9-May-2016 added "data" option.
    15-October-2016 fixed incorrect warning that start or end exceeds threshold.
    16-December-2016 Added "print" option.
    14-January-2017 Added coordList stuff.
    23-August-2017 Moved user-coordinate stuff to external file.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <float.h>

#ifndef IMV6OR7
#include "vsn_defines.h"
#endif

#include "chklist.h"

#define VERIFY 0

#include "usercoord.inc"



//---------------------------------------------------------------------------
// Coordinate list data structures.

typedef struct {
  double x;
  double y;
} CoordT;

typedef struct {
  int listInitSize;
  int listSize;
  int numUsed;
  CoordT * coords;
} CoordListT;

//---------------------------------------------------------------------------
// Data structures.

typedef double dppValueT;

// Neighbours are numbered 1 to 9 except 5.
// Up-left is 1, up-right is 3, etc.
#define NUM_NEIGHB 9

typedef struct {
  Quantum
    value;

  dppValueT
    tentativeDistance;

  MagickBooleanType
    visited;

  int
    parent;
} NodeT;

#define INFINITE_DIST DBL_MAX

typedef struct {
  ssize_t
    x,
    y;
} pqNodeT;

typedef struct {
  MagickBooleanType
    stopAtEnd,
    wrData,
    suppressImage,
    printPix,
    do_verbose,
    do_warnings,
    hasVisitedThreshold,
    foundEnd;

  dppValueT
    visitedThreshold;

  UserCoordXyT
    start,
    end;

  ssize_t
    width,
    height;

  FILE *
    fh_data;

  int
    precision;

  dppValueT
    neighbourWeights[NUM_NEIGHB+1];

  NodeT
    **nodes;

  ssize_t
    pqNumNodes;

  ssize_t
    pqMaxNodes;

  pqNodeT
    *pqNodes;

  CoordListT
    *coordList; // If not null, list will be added to.
} darkestPntPntT;

static void InitDarkestPntPnt (darkestPntPntT * dpp)
{
  dpp->do_verbose = MagickFalse;
  dpp->do_warnings = MagickTrue;
  dpp->hasVisitedThreshold = MagickFalse;

  dpp->start.x.Pix = dpp->start.y.Pix = 0;
  dpp->start.x.cs = dpp->start.y.cs = csPix;
  dpp->end.x.Pix = dpp->end.y.Pix = 0;
  dpp->end.x.cs = dpp->end.y.cs = csPix;
  dpp->wrData = MagickFalse;
  dpp->suppressImage = MagickFalse;
  dpp->stopAtEnd = MagickTrue;
  dpp->printPix = MagickFalse;
  dpp->fh_data = stderr;

  dpp->coordList = NULL;

  int i;
  for (i=0; i <= NUM_NEIGHB; i++) dpp->neighbourWeights[i] = 1;

  dpp->neighbourWeights[1]
    = dpp->neighbourWeights[3]
    = dpp->neighbourWeights[7]
    = dpp->neighbourWeights[9] = sqrt (2.0);
}

//---------------------------------------------------------------------------
// Coordinate list functions.

static CoordListT * CreateCoordList (int initSize)
// If oom (or initSize is <= zero), returns NULL.
{
  if (initSize <= 0) return NULL;

  CoordListT * pcl = (CoordListT *)malloc(sizeof(CoordListT));
  if (pcl==NULL) return NULL;
  pcl->listInitSize = initSize;
  pcl->listSize = initSize;
  pcl->numUsed = 0;
  pcl->coords = (CoordT *)malloc(pcl->listSize*sizeof(CoordT));
  if (pcl->coords==NULL) {
    free (pcl);
    return NULL;
  }
  return pcl;
}

static CoordListT * DestroyCoordList (CoordListT * pcl)
{
  if (pcl) {
    if (pcl->coords) free (pcl->coords);
    free (pcl);
  }
  return NULL;
}

static MagickBooleanType AddToCoordList (CoordListT * pcl, double x, double y)
// If list full, expands it.
// If oom, returns false.
{
  if (!pcl) return MagickFalse;

  if (pcl->numUsed > pcl->listSize) {
    printf ("AddToCoordList: bug\n");
    return MagickFalse;
  }

  if (pcl->numUsed == pcl->listSize) {
    // Expand it.
    pcl->listSize += pcl->listInitSize;
    CoordT * tmp = (CoordT *)realloc(pcl->coords, pcl->listSize*sizeof(CoordT));
    if (tmp==NULL) {
      free (pcl->coords);
      free (pcl);
      return MagickFalse;
    }
    pcl->coords = tmp;
  }

  pcl->coords[pcl->numUsed].x = x;
  pcl->coords[pcl->numUsed].y = y;
  pcl->numUsed++;
  return MagickTrue;
}

static MagickBooleanType AddToCoordListIf (CoordListT * pcl, double x, double y)
// If list empty, or new coords different to last entry,
// adds to list.
{
  if (!pcl) return MagickFalse;

  if (pcl->numUsed == 0
   || pcl->coords[pcl->numUsed-1].x != x
   || pcl->coords[pcl->numUsed-1].y != y)
  {
    return AddToCoordList (pcl, x, y);
  }
  printf ("f");
  return MagickTrue;
}

static void InvertCoordList (CoordListT * pcl, int nStart)
{
  if (!pcl) return;

  printf ("InvertCoordList %i\n", pcl->numUsed);
  int i;
  int q = pcl->numUsed-1;
  double tx, ty;
  for (i = nStart; i < q; i++) {
    tx = pcl->coords[i].x;
    ty = pcl->coords[i].y;
    pcl->coords[i].x = pcl->coords[q].x;
    pcl->coords[i].y = pcl->coords[q].y;
    pcl->coords[q].x = tx;
    pcl->coords[q].y = ty;
    q--;
  }
}

static void EmptyCoordList (CoordListT * pcl)
{
  if (!pcl) return;

  pcl->numUsed = 0;
}

static void WrCoords (CoordListT * pcl, FILE * fhOut)
{
  if (!pcl) return;

  int i;
  for (i = 0; i < pcl->numUsed; i++) {
    fprintf (fhOut, "%g,%g\n",
      (double)pcl->coords[i].x, (double)pcl->coords[i].y);
  }
}

static void DumpCoordList (CoordListT * pcl, FILE * fhOut)
{
  if (!pcl) return;

  fprintf (fhOut, "Init=%i  Size=%i numUsed=%i\n",
    pcl->listInitSize, pcl->listSize, pcl->numUsed);
  WrCoords (pcl, fhOut);
}

//---------------------------------------------------------------------------
// Priority queue functions.

#define LCHILD(x) 2 * x + 1
#define RCHILD(x) 2 * x + 2
#define PARENT(x) (x - 1) / 2

static dppValueT inline pqDataOfNode (darkestPntPntT * dpp, ssize_t NodeNum)
{
  ssize_t x = dpp->pqNodes[NodeNum].x;
  ssize_t y = dpp->pqNodes[NodeNum].y;
  return dpp->nodes[y][x].tentativeDistance;
}


#if VERIFY==1

static void pqVerify (darkestPntPntT * dpp)
{
  ssize_t
    i;

  //if (dpp->do_verbose) printf ("pqVerify pqNumNodes=%li\n", dpp->pqNumNodes);

  for (i=0; i < dpp->pqNumNodes; i++) {
    ssize_t lch = LCHILD(i);
    ssize_t rch = RCHILD(i);

    MagickBooleanType HasL = lch < dpp->pqNumNodes;
    MagickBooleanType HasR = rch < dpp->pqNumNodes;

    if (HasL && pqDataOfNode (dpp, lch) < pqDataOfNode (dpp, i))
      printf ("** Bad %li left\n", i);

    if (HasR && pqDataOfNode (dpp, rch) < pqDataOfNode (dpp, i))
      printf ("** Bad %li right\n", i);
  }
}

static void pqDumpOrder (darkestPntPntT * dpp, ssize_t NodeNum)
{
  //printf("%d ", hp->elem[i].data) ;

  ssize_t x = dpp->pqNodes[NodeNum].x;
  ssize_t y = dpp->pqNodes[NodeNum].y;
  dppValueT v = dpp->nodes[y][x].tentativeDistance;

  printf ("  %li  %li,%li %.*g\n", NodeNum, x, y, dpp->precision, v);

  ssize_t lch = LCHILD(NodeNum);
  ssize_t rch = RCHILD(NodeNum);

  MagickBooleanType
    LFirst = MagickTrue,
    HasL = lch < dpp->pqNumNodes,
    HasR = rch < dpp->pqNumNodes;

  if (HasL && HasR) {
    LFirst = pqDataOfNode (dpp, lch) < pqDataOfNode (dpp, rch);
  }

  if(HasL && LFirst) {
    pqDumpOrder (dpp, lch) ;
  }
  if(HasR) {
    pqDumpOrder (dpp, rch) ;
  }
  if(HasL && !LFirst) {
    pqDumpOrder (dpp, lch) ;
  }
}

static void pqDump (darkestPntPntT * dpp)
{
  ssize_t
    i;

  printf ("pqNumNodes=%li\n", dpp->pqNumNodes);

  for (i=0; i < dpp->pqNumNodes; i++) {
    pqNodeT * pqn = &dpp->pqNodes[i];
    NodeT * pn = &(dpp->nodes[pqn->y][pqn->x]);
    printf ("  %li  %li,%li %.*g\n",
      i, pqn->x, pqn->y, dpp->precision, pn->tentativeDistance);
  }

  //printf ("In order:\n");
  //pqDumpOrder (dpp, 0);

  pqVerify (dpp);
}

#endif

static void inline pqSwapNodes (darkestPntPntT * dpp, ssize_t n1, ssize_t n2)
{
  ssize_t t = dpp->pqNodes[n1].x;
  dpp->pqNodes[n1].x = dpp->pqNodes[n2].x;
  dpp->pqNodes[n2].x = t;

  t = dpp->pqNodes[n1].y;
  dpp->pqNodes[n1].y = dpp->pqNodes[n2].y;
  dpp->pqNodes[n2].y = t;
}

static void pqBalanceNode (darkestPntPntT * dpp, ssize_t NodeNum)
// "Bubble-down".
// Finds the smallest of this node and its children.
// If the node isn't the smallest of the three,
//   swaps data with that child and recursively balances that child.
{
  ssize_t smallest = NodeNum;
  ssize_t lch = LCHILD(NodeNum);
  ssize_t rch = RCHILD(NodeNum);

  dppValueT SmData = pqDataOfNode (dpp, NodeNum);

  if (lch < dpp->pqNumNodes) {
    dppValueT ldist = pqDataOfNode (dpp, lch);
    if (ldist < SmData) {
      SmData = ldist;
      smallest = lch;
    }
  }

  if (rch < dpp->pqNumNodes) {
    dppValueT rdist = pqDataOfNode (dpp, rch);
    if (rdist < SmData) {
      SmData = rdist;
      smallest = rch;
    }
  }

  if (smallest != NodeNum) {
    pqSwapNodes (dpp, smallest, NodeNum);
    pqBalanceNode (dpp, smallest);
  }
}

static void pqInsertNew (darkestPntPntT * dpp, ssize_t x, ssize_t y)
{
  //printf ("pqIns %li,%li ", x, y);

  if (dpp->pqNumNodes >= dpp->pqMaxNodes) {
    fprintf (stderr, "pq bust\n");
  }

  dppValueT newDist = dpp->nodes[y][x].tentativeDistance;

  ssize_t n = dpp->pqNumNodes;

  // "Bubble up".
  // If this data is less than node's parent,
  //   drop parent data into this node
  //   and consider putting new data into that parent.
  while (n && newDist < pqDataOfNode (dpp, PARENT(n))) {
    dpp->pqNodes[n] = dpp->pqNodes[PARENT(n)];
    n = PARENT(n);
  }
  dpp->pqNodes[n].x = x;
  dpp->pqNodes[n].y = y;

  dpp->pqNumNodes++;

#if VERIFY==1
  pqVerify (dpp);
#endif
}

static ssize_t pqFindNode (
  darkestPntPntT * dpp,
  ssize_t x, ssize_t y,
  dppValueT val, ssize_t NodeStart)
// Returns -1 if not present.
// Note: recursive.
{
  if (dpp->pqNumNodes==0) return -1;

  pqNodeT * pqn = &dpp->pqNodes[NodeStart];

  if (pqn->x == x && pqn->y == y) return NodeStart;

  if (pqDataOfNode (dpp, NodeStart) > val) return -1;

  ssize_t lch = LCHILD(NodeStart);
  if (lch < dpp->pqNumNodes) {
    ssize_t nch = pqFindNode (dpp, x, y, val, lch);
    if (nch >= 0) return nch;
  }

  ssize_t rch = RCHILD(NodeStart);
  if (rch < dpp->pqNumNodes) {
    ssize_t nch = pqFindNode (dpp, x, y, val, rch);
    if (nch >= 0) return nch;
  }

  return -1;
}

#if VERIFY==1

static void pqFindMin (darkestPntPntT * dpp, ssize_t *x, ssize_t *y)
{
  if (!dpp->pqNumNodes) fprintf (stderr, "** Bad pqfm: empty");

  pqNodeT * pqn = &dpp->pqNodes[0];
  *x = pqn->x;
  *y = pqn->y;
}

#endif


static int pqRemoveMin (darkestPntPntT * dpp, ssize_t *x, ssize_t *y)
// Returns 1 if okay, or 0 if no data.
{
  if (!dpp->pqNumNodes) {
    //fprintf (stderr, "** Bad pqrm: empty");
    return 0;
  }

  pqNodeT * pqn = &dpp->pqNodes[0];
  *x = pqn->x;
  *y = pqn->y;

  // Put largest into root, and rebalance.

  dpp->pqNodes[0] = dpp->pqNodes[--dpp->pqNumNodes];

  pqBalanceNode (dpp, 0);

  return 1;
}


static void inline UpdateDist (
  darkestPntPntT * dpp,
  ssize_t nodeNum,
  ssize_t x, ssize_t y,
  //dppValueT oldVal,
  dppValueT newVal)
// Updates the tentative distance of a pixel.
{
  //printf ("UpdDist %li,%li: ", x, y);

/*==
  assert(newVal <= oldVal);

  ssize_t n = pqFindNode (dpp, x, y, oldVal, 0);

  if (n < 0) {
    fprintf (stderr, "** BUG: UpdateDist can't find %li,%li\n", x, y);
    return;
  }
==*/

  ssize_t n = nodeNum;

  dpp->nodes[y][x].tentativeDistance = newVal;

  // "Bubble up".
  // If this data is less than node's parent,
  //   swap with parent and iterate.
  while (n && newVal < pqDataOfNode (dpp, PARENT(n))) {
    pqSwapNodes (dpp, n, PARENT(n));
    n = PARENT(n);
  }

  // Bubble down from the top.
  // FIXME: do we need this?
//  pqBalanceNode (dpp, 0);

#if VERIFY==1
  pqVerify (dpp);
#endif
}


//---------------------------------------------------------------------------




static void usage (void)
{
  printf ("Usage: -process 'darkestpntpnt [OPTION]...'\n");
  printf ("Finds darkest path between two points.\n");
  printf ("\n");
  printf ("  s, start_at X,Y         start the path at (integer) coordinates\n");
  printf ("  e, end_at X,Y           end the path at (integer) coordinates\n");
  printf ("  t, threshold_visited N  where initial values above N, don't visit\n");
  printf ("  n, no_end               don't stop processing at path end\n");
  printf ("  d, data                 write data image instead of path image\n");
  printf ("  p, print                write each visited coordinate to stderr\n");
  printf ("  f, file string          write to file stream stdout or stderr\n");
  printf ("  v, verbose              write text information to stdout\n");
  printf ("\n");
  printf ("sizeof(NodeT)=%i\n", (int)sizeof(NodeT));
  printf ("sizeof(pqNodeT)=%i\n", (int)sizeof(pqNodeT));
  printf ("INFINITE_DIST=%g\n", INFINITE_DIST);
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  darkestPntPntT * dpp
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  InitDarkestPntPnt (dpp);

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "s", "start_at")==MagickTrue) {
      i++;

      int r = ParseXy (argv[i], &dpp->start);
      if (!r) {
        printf ("Failed to parse start [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "e", "end_at")==MagickTrue) {
      i++;
      int r = ParseXy (argv[i], &dpp->end);
      if (!r) {
        printf ("Failed to parse end [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "t", "threshold_visited")==MagickTrue) {
      dpp->hasVisitedThreshold = MagickTrue;
      i++;
      dpp->visitedThreshold = atof (argv[i]);
    } else if (IsArg (pa, "n", "no_end")==MagickTrue) {
      dpp->stopAtEnd = MagickFalse;
    } else if (IsArg (pa, "d", "data")==MagickTrue) {
      dpp->wrData = MagickTrue;
    } else if (IsArg (pa, "p", "print")==MagickTrue) {
      dpp->printPix = MagickTrue;
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      i++;
      if (strcasecmp (argv[i], "stdout")==0) dpp->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) dpp->fh_data = stderr;
      else status = MagickFalse;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      dpp->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "darkestpntpnt: ERROR: unknown option\n");
      status = MagickFalse;
    }
  }

  if (dpp->do_verbose) {
    fprintf (stderr, "darkestpntpnt options: ");
    fprintf (stderr, "  start_at ");
    WrUserCoord (&dpp->start);
    fprintf (stderr, "  end_at ");
    WrUserCoord (&dpp->end);
    if (dpp->hasVisitedThreshold)
      fprintf (stderr, "  threshold_visited %.*g",
        dpp->precision, dpp->visitedThreshold);
    if (dpp->wrData) fprintf (stderr, "  data");
    if (dpp->printPix) fprintf (stderr, "  print");
    if (dpp->fh_data == stdout) fprintf (stderr, " file stdout");
    if (!dpp->stopAtEnd) fprintf (stderr, "  no_end");
    if (dpp->do_verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


static MagickBooleanType initialise(const Image *image,
  darkestPntPntT * dpp,
  ExceptionInfo *exception
)
{
  ssize_t
    x, y;

  MagickBooleanType
    status = MagickTrue;

  CacheView
    *image_view;

  dpp->width = image->columns;
  dpp->height = image->rows;

  // Allocate memory

  //if (dpp->do_verbose) printf ("Alloc %ix%i\n", (int)dpp->width, (int)dpp->height);

  dpp->pqMaxNodes = dpp->height * dpp->width;
  dpp->pqNodes = (pqNodeT *) AcquireQuantumMemory(dpp->pqMaxNodes, sizeof(pqNodeT));
  if (dpp->pqNodes == (pqNodeT *) NULL) {
    return MagickFalse;
  }

  dpp->pqNumNodes = 0;

  dpp->nodes = (NodeT **) AcquireQuantumMemory(dpp->height, sizeof(*dpp->nodes));
  if (dpp->nodes == (NodeT **) NULL) {
    RelinquishMagickMemory(dpp->pqNodes);
    return MagickFalse;
  }
  for (y = 0; y < dpp->height; y++) {
    dpp->nodes[y] = (NodeT *) AcquireQuantumMemory(dpp->width, sizeof(**dpp->nodes));
    if (dpp->nodes[y] == (NodeT *) NULL) break;
  }
  if (y < dpp->height) {
    for (y--; y >= 0; y--) {
      if (dpp->nodes[y] != (NodeT *) NULL)
        dpp->nodes[y] = (NodeT *) RelinquishMagickMemory(dpp->nodes[y]);
    }
    dpp->nodes = (NodeT **) RelinquishMagickMemory(dpp->nodes);
    RelinquishMagickMemory(dpp->pqNodes);
    return MagickFalse;
  }

  // Populate values

  //if (dpp->do_verbose) printf ("Populate\n");

  image_view = AcquireVirtualCacheView(image,exception);

  for (y = 0; y < dpp->height; y++) {
    VIEW_PIX_PTR const
      *p;

    if (status == MagickFalse)
      continue;

    p=GetCacheViewVirtualPixels(image_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL)
      status=MagickFalse;

    for (x = 0; x < dpp->width; x++) {
      //printf ("Pop %i,%i\n", (int)x, (int)y);

      NodeT * pn = &(dpp->nodes[y][x]);
      pn->value = GetPixelIntensity (image, p) / (dppValueT)QuantumRange;
      pn->tentativeDistance = INFINITE_DIST;
      pn->visited = MagickFalse;
      if (dpp->hasVisitedThreshold) {
        if (pn->value > dpp->visitedThreshold) {
          pn->visited = MagickTrue;
        }
      }
      pn->parent = 0;
      p += Inc_ViewPixPtr (image);
    }
  }

  image_view=DestroyCacheView(image_view);

  NodeT * pn = &dpp->nodes[dpp->start.y.Pix][dpp->start.x.Pix];
  pn->tentativeDistance = 0;
  if (dpp->hasVisitedThreshold && pn->value > dpp->visitedThreshold) {
    if (dpp->do_warnings) fprintf (stderr,
      "Problem: start pixel has value (%.*g) above threshold.\n",
      dpp->precision, pn->value);
  }

  pn = &dpp->nodes[dpp->end.y.Pix][dpp->end.x.Pix];
  if (dpp->hasVisitedThreshold && pn->value > dpp->visitedThreshold) {
    if (dpp->do_warnings) fprintf (stderr,
      "Problem: end pixel has value (%.*g) above threshold.\n",
      dpp->precision, pn->value);
  }

  return (status);
}


static void deIinitialise(
  darkestPntPntT * dpp
)
{
  ssize_t
    y;

  for (y = 0; y < dpp->height; y++) {
    dpp->nodes[y] = (NodeT *) RelinquishMagickMemory(dpp->nodes[y]);
  }
  dpp->nodes = (NodeT **) RelinquishMagickMemory(dpp->nodes);

  RelinquishMagickMemory(dpp->pqNodes);
}


#if VERIFY==1

static void DumpNodes (
  darkestPntPntT * dpp
)
{
  ssize_t
    x, y;

  printf ("DumpNodes %lix%li\n", dpp->width, dpp->height);

  for (y = 0; y < dpp->height; y++) {
    for (x = 0; x < dpp->width; x++) {
      NodeT * pn = &dpp->nodes[y][x];

      printf ("%li,%li %.*g %.*g %s %i\n",
        x, y,
        dpp->precision, (double)pn->value,
        dpp->precision, (double)pn->tentativeDistance,
        pn->visited? "V": "nv",
        pn->parent);
    }
  }
}

#endif

static MagickBooleanType NearestUnvNode(
  darkestPntPntT * dpp,
  ssize_t * nx,
  ssize_t * ny
)
// Returns whether okay.
// If none, returns FALSE.
{
  // This needs to be fast.

/*===
//
// This is how it would be done without a priority queue.
// Much simpler, but very much slower.
//
//  dppValueT
//    nearestDist;
//
//  ssize_t
//    x, y;
//
//  nearestDist = INFINITE_DIST;
//
//  // This could be parallelised.
//
//  for (y = 0; y < dpp->height; y++) {
//    for (x = 0; x < dpp->width; x++) {
//      NodeT * pn = &dpp->nodes[y][x];
//      if (!pn->visited) {
//        if (nearestDist > pn->tentativeDistance) {
//          nearestDist = pn->tentativeDistance;
//          *nx = x;
//          *ny = y;
//          *allDone = MagickFalse;
//        }
//      }
//    }
//  }
//  printf ("nun %li,%li ad=%i\n", *nx, *ny, (int)*allDone);
//
===*/

  ssize_t pqX=0, pqY=0;
  if ( pqRemoveMin (dpp, &pqX, &pqY) == 0) return MagickFalse;

//  *allDone = (r == 1) ? MagickFalse : MagickTrue;

  //if (*nx != pqX || *ny != pqY) {
  //  printf ("NearestUnvNode: %li %li %li %li\n", *nx, pqX, *ny, pqY);
  //}

  *nx = pqX;
  *ny = pqY;

  return MagickTrue;
}


static void inline ProcessNeighbour (
  darkestPntPntT * dpp,
  dppValueT thisDist,
  int nbx,
  int nby,
  int ParentNum)
{
  NodeT * pnb = &dpp->nodes[nby][nbx];

  if (!pnb->visited) {
    dppValueT newDist = thisDist + pnb->value * dpp->neighbourWeights[ParentNum];
    if (pnb->tentativeDistance > newDist) {
      ssize_t pqNode = -1;

      if (pnb->tentativeDistance < INFINITE_DIST) {
        pqNode = pqFindNode (dpp, nbx, nby, pnb->tentativeDistance, 0);
      }

      if (pqNode == -1) {
        pnb->tentativeDistance = newDist;
        pqInsertNew (dpp, nbx, nby);
      } else {
        UpdateDist (dpp, pqNode, nbx, nby, /* pnb->tentativeDistance, */ newDist);
      }

      pnb->parent = ParentNum;
    }
  }
}


static MagickBooleanType VisitNodes (
  darkestPntPntT * dpp
)
{
  MagickBooleanType
    finished = MagickFalse,
    DoCoordList = MagickFalse;

  ssize_t
    nx, ny;

  int nVisited = 0;

  nx = dpp->start.x.Pix;
  ny = dpp->start.y.Pix;

  if (dpp->printPix) {
    fprintf (dpp->fh_data, "%li,%li\n", nx, ny);
  }

  dpp->foundEnd = MagickFalse;

  if (dpp->coordList != NULL && dpp->wrData == MagickTrue) {
    DoCoordList = MagickTrue;

    // Record this start pixel.
    // FIXME: but not if out of threshold?

    AddToCoordList (dpp->coordList, nx, ny);
  }

  while (finished == MagickFalse) {

    //printf ("vn %li,%li\n", nx, ny);

    NodeT * pn = &dpp->nodes[ny][nx];

    // Process the 8 neighbours.
    if (ny > 0) {
      if (nx > 0) {
        ProcessNeighbour (dpp, pn->tentativeDistance, nx-1, ny-1, 1);
      }
      ProcessNeighbour (dpp, pn->tentativeDistance, nx, ny-1, 2);
      if (nx < dpp->width-1) {
        ProcessNeighbour (dpp, pn->tentativeDistance, nx+1, ny-1, 3);
      }
    }
    if (nx > 0) {
      ProcessNeighbour (dpp, pn->tentativeDistance, nx-1, ny, 4);
    }

    if (nx < dpp->width-1) {
      ProcessNeighbour (dpp, pn->tentativeDistance, nx+1, ny, 6);
    }

    if (ny < dpp->height-1) {
      if (nx > 0) {
        ProcessNeighbour (dpp, pn->tentativeDistance, nx-1, ny+1, 7);
      }
      ProcessNeighbour (dpp, pn->tentativeDistance, nx, ny+1, 8);

      if (nx < dpp->width-1) {
        ProcessNeighbour (dpp, pn->tentativeDistance, nx+1, ny+1, 9);
      }
    }

    //printf ("vn2 %li,%li\n", nx, ny);
    if ( pn->visited == MagickTrue
      && (   nx != dpp->start.x.Pix
          || ny != dpp->start.y.Pix ))
    {
      fprintf (stderr, "Bug: already visited %li,%li\n", nx, ny);
      finished = MagickTrue;

#if VERIFY==1
      DumpNodes (dpp);
      pqVerify (dpp);
      pqDump (dpp);
#endif

    }
    pn->visited = MagickTrue;

    if (dpp->do_verbose && ((++nVisited % 100000) == 0)) {
      printf (".\n");
    }

    if (nx == dpp->end.x.Pix && ny == dpp->end.y.Pix) {
      dpp->foundEnd = MagickTrue;
      if (dpp->stopAtEnd) {
        finished = MagickTrue;
      }
    }

    if (finished == MagickFalse) {
      if (NearestUnvNode (dpp, &nx, &ny)) {
        if (dpp->printPix && !finished) {
          fprintf (dpp->fh_data, "%li,%li\n", nx, ny);
        }
        if (DoCoordList && !finished) {
          AddToCoordList (dpp->coordList, nx, ny);
        }

      } else {
        finished = MagickTrue;
      }
    } else if (dpp->printPix) {
      fprintf (dpp->fh_data, "%li,%li\n", nx, ny);

      if (DoCoordList) {
        AddToCoordList (dpp->coordList, nx, ny);
      }
    }
  }

  if (dpp->do_verbose) fprintf (stderr, "nVisited: %i\n", nVisited);

  return (MagickTrue);
}

static Image * writePath(const Image *image,
  darkestPntPntT * dpp,
  ExceptionInfo *exception
)
{
  Image *
    path_image;

  CacheView
    *path_view;

  VIEW_PIX_PTR
    *q_path;

  MagickBooleanType
    status = MagickTrue;

  ssize_t
    x, y;

  //if (dpp->do_verbose) printf ("writePath: WH=%li,%li\n", image->columns, image->rows);

  path_image=CloneImage(image, image->columns, image->rows, MagickTrue, exception);
  if (path_image == (Image *) NULL)
    return(path_image);

  if (SetNoPalette (path_image, exception) == MagickFalse)
    return (Image *)NULL;

  SetAllBlack (path_image, exception);

  path_view=AcquireAuthenticCacheView(path_image,exception);

  x = dpp->end.x.Pix;
  y = dpp->end.y.Pix;

  ssize_t nInPath = 0;

  dppValueT totalValue = 0.0;

  MagickBooleanType DoCoordList;
  int clNumStart = 0;
  if (dpp->coordList==NULL) {
    DoCoordList = MagickFalse;
  } else {
    DoCoordList = MagickTrue;
    clNumStart = dpp->coordList->numUsed;
  }

  if (dpp->foundEnd == MagickTrue) {

    if (dpp->do_verbose) {
      NodeT * pn = &dpp->nodes[y][x];

      fprintf (stderr, "maxDist: %.*g\n",
        dpp->precision, pn->tentativeDistance * QuantumRange);
      fprintf (stderr, "maxDistPc: %.*g\n",
        dpp->precision, pn->tentativeDistance * 100.0);
    }

    do {
      //printf ("xy=%li,%li\n", x, y);

      // Get just the one pixel we need.
      q_path=GetCacheViewAuthenticPixels(path_view,x,y,1,1,exception);
      if (q_path == (const VIEW_PIX_PTR *) NULL) {
        fprintf (stderr, "darkestpntpng: bad GetCacheViewAuthenticPixels q_path\n");
        status=MagickFalse;
        break;
      }

      if (GET_PIXEL_RED (path_image, q_path) != 0) {
        fprintf (stderr, "Bug: cycle\n");
        status=MagickFalse;
        break;
      }

      SET_PIXEL_RED   (path_image, QuantumRange, q_path);
      SET_PIXEL_GREEN (path_image, QuantumRange, q_path);
      SET_PIXEL_BLUE  (path_image, QuantumRange, q_path);

      if (SyncCacheViewAuthenticPixels(path_view,exception) == MagickFalse) {
        fprintf (stderr, "bad sync\n");
        status=MagickFalse;
        break;
      }

      if (DoCoordList) {
        if (!AddToCoordList (dpp->coordList, x, y)) {
          status=MagickFalse;
        }
      }

      if (x == dpp->start.x.Pix && y == dpp->start.y.Pix) break;

      NodeT * pn = &dpp->nodes[y][x];

      nInPath++;
      totalValue += pn->value;

      switch (pn->parent) {
        case 1: y+=1; x+=1; break;
        case 2: y+=1;       break;
        case 3: y+=1; x-=1; break;
        case 4:       x+=1; break;
        case 6:       x-=1; break;
        case 7: y-=1; x+=1; break;
        case 8: y-=1;       break;
        case 9: y-=1; x-=1; break;
        default: 
          if (x != dpp->start.x.Pix || y != dpp->start.y.Pix) {
            fprintf (stderr, "bad parent\n");
            status=MagickFalse;
          }
      }
    } while (status==MagickTrue);

  } else {
    if (dpp->do_warnings) fprintf (stderr, "Problem: end is not on path.\n");
  }

  if (DoCoordList) {
    // FIXME: Oops, we can do this only if it was empty.
    InvertCoordList (dpp->coordList, clNumStart);
  }

  path_view=DestroyCacheView(path_view);

  if (dpp->do_verbose)
    fprintf (stderr, "nInPath: %li\nTotalValue: %.*g\nAvgValue: %.*g\n",
      nInPath,
       dpp->precision, totalValue,
       dpp->precision, totalValue / nInPath);

  return path_image;
}


static Image * writeData (const Image *image,
  darkestPntPntT * dpp,
  ExceptionInfo *exception
)
{
  // Copy values to red channel;
  // copy distances to green channel;
  // copy parent number to blue channel.

  Image *
    data_image;

  ssize_t
    x, y;

  MagickBooleanType
    status = MagickTrue;

  CacheView
    *data_view;

  data_image=CloneImage(image, image->columns, image->rows, MagickTrue, exception);
  if (data_image == (Image *) NULL)
    return(data_image);

  if (SetNoPalette (data_image, exception) == MagickFalse)
    return (Image *)NULL;

  data_view = AcquireAuthenticCacheView(data_image,exception);

  ssize_t nInPath = 0;
  dppValueT maxDist = -1;

  // FIXME: parallelise

  for (y = 0; y < dpp->height; y++) {
    VIEW_PIX_PTR
      *p;

    if (status == MagickFalse)
      continue;

    p=GetCacheViewAuthenticPixels(data_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL)
      status=MagickFalse;

    for (x = 0; x < dpp->width; x++) {
      NodeT * pn = &(dpp->nodes[y][x]);

      SET_PIXEL_RED   (data_image, pn->value * QuantumRange, p);

      if (pn->tentativeDistance == INFINITE_DIST) {
        SET_PIXEL_GREEN (data_image, -1, p);
      } else {
        dppValueT dist = pn->tentativeDistance * QuantumRange;
        if (maxDist < dist) maxDist = dist;
        SET_PIXEL_GREEN (data_image, dist, p);
        nInPath++;
      }

      SET_PIXEL_BLUE  (data_image, pn->parent, p);

      p += Inc_ViewPixPtr (data_image);
    }

    if (SyncCacheViewAuthenticPixels(data_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
      break;
    }
  }

  data_view=DestroyCacheView(data_view);

  if (dpp->do_verbose) {
    fprintf (stderr, "nInPath: %li\n", nInPath);

    if (maxDist > -1) {
      fprintf (stderr, "maxDist: %.*g\n",
        dpp->precision, maxDist);
      fprintf (stderr, "maxDistPc: %.*g\n",
        dpp->precision, maxDist * 100.0 / QuantumRange);
    }
  }

  return (data_image);
}


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *darkestpntpnt(Image *image,
  darkestPntPntT * dpp,
  ExceptionInfo *exception)
{
  Image
    *out_image;

  MagickBooleanType
    status;


  dpp->precision = GetMagickPrecision();

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  status = initialise(image, dpp, exception);
  if (status==MagickFalse) return (Image *) NULL;

  //if (dpp->do_verbose) printf ("DumpNodes.\n");
  //DumpNodes (dpp);

  //if (dpp->do_verbose) printf ("VisitNodes.\n");
  VisitNodes (dpp);

#if VERIFY==1
  //if (dpp->do_verbose) printf ("Verify pq.\n");
  pqVerify (dpp);
#endif

  //if (dpp->do_verbose) printf ("DumpNodes again.\n");
  //DumpNodes (dpp);

  //pqDump (dpp);

  if (dpp->suppressImage) {
    if (dpp->do_verbose) printf ("Image suppressed.\n");
    out_image = NULL;
  } else {
    if (dpp->wrData) {
      if (dpp->do_verbose) printf ("writeData.\n");
      out_image = writeData (image, dpp, exception);
    } else {
      if (dpp->do_verbose) printf ("writePath.\n");
      out_image = writePath (image, dpp, exception);
    }
  }

  deIinitialise(dpp);

  return (out_image);
}


ModuleExport size_t darkestpntpntImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  darkestPntPntT
    dpp;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  InitDarkestPntPnt (&dpp);

  status = menu (argc, argv, &dpp);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    int r;

    r = ResolveUserCoords (&dpp.start, image->columns, image->rows, 1.0);
    if (!r) return -1;

    r = ResolveUserCoords (&dpp.end, image->columns, image->rows, 1.0);
    if (!r) return -1;

    new_image = darkestpntpnt (image, &dpp, exception);

    ReplaceImageInList(&image,new_image);
    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
