/*
    Reference:
      http://im.snibgo.com/customim.htm
      http://im.snibgo.com/integim.htm

    Latest update:
      24-July-2017 Fixed bug on top line and left column.
                   (x1 and y1 can now be -1.)
      3-April-2018 for v7.0.7-28
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"

typedef enum {
  aaAuto,
  aaDisregard,
  aaRegard
} alphaActionT;

typedef struct {
  FILE *
    fh_data;

  char
    *sWindow;

  MagickBooleanType
    do_regardalpha,
    do_verbose,
    dontDivide;

  alphaActionT
    alphaAction;

  ssize_t
    offsetx,
    offsety,
    winWidth,
    winHeight;
} deintegimT;

static void usage (void)
{
  printf ("Usage: -process 'deintegim [OPTION]...'\n");
  printf ("De-integrate an HDRI integral image (summed area table).\n");
  printf ("\n");
  printf ("  w, window string    window size, WxH\n");
  printf ("  pd, postdiv         'yes' or 'no' or 'auto'\n");
  printf ("  dd, dontdivide      don't divide colour channels\n");
  printf ("  ox, offsetX int     offset for window centre in X-direction\n");
  printf ("  oy, offsetY int     offset for window centre in Y-direction\n");
  printf ("  f, file string      Write verbose text to stderr or stdout\n");
  printf ("  v, verbose          write text information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  deintegimT * pch
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  pch->fh_data = stderr;
  pch->sWindow = NULL;
  pch->alphaAction = aaRegard;

  pch->do_regardalpha =
    pch->do_verbose = MagickFalse;
  pch->dontDivide = MagickFalse;
  pch->offsetx = 0;
  pch->offsety = 0;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "w", "window")==MagickTrue) {
      i++;
      pch->sWindow = (char *)argv[i];
    } else if (IsArg (pa, "pd", "postdiv")==MagickTrue) {
      i++;
      if (strcasecmp (argv[i], "auto")==0) pch->alphaAction = aaAuto;
      else if (strcasecmp (argv[i], "yes")==0) pch->alphaAction = aaRegard;
      else if (strcasecmp (argv[i], "no")==0) pch->alphaAction = aaDisregard;
      else status = MagickFalse;
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      i++;
      if (strcasecmp (argv[i], "stdout")==0) pch->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) pch->fh_data = stderr;
      else status = MagickFalse;
    } else if (IsArg (pa, "ox", "offsetX")==MagickTrue) {
      i++;
      pch->offsetx = atoi(argv[i]);
    } else if (IsArg (pa, "oy", "offsetY")==MagickTrue) {
      i++;
      pch->offsety = atoi(argv[i]);
    } else if (IsArg (pa, "dd", "dontdivide")==MagickTrue) {
      pch->dontDivide = MagickTrue;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pch->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "deintegim: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (pch->do_verbose) {
    fprintf (stderr, "deintegim options:");
    if (pch->sWindow) fprintf (stderr, "  window %s", pch->sWindow);
    if (pch->offsetx) fprintf (stderr, "  offsetX %li", pch->offsetx);
    if (pch->offsety) fprintf (stderr, "  offsetY %li", pch->offsety);
    if (pch->alphaAction != aaAuto) {
      fprintf (stderr, "  postdiv ");
      if (pch->alphaAction==aaRegard) fprintf (stderr, "yes");
      else if (pch->alphaAction==aaDisregard) fprintf (stderr, "no");
      else fprintf (stderr, "??");
    }
    if (pch->fh_data == stdout) fprintf (stderr, "  file stdout");
    if (pch->dontDivide) fprintf (stderr, "  dontdivide");
    if (pch->do_verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}



static MagickBooleanType GetNumber (
  char ** p,
  double *pValue,
  ssize_t dim)
// Returns whether valid.
{
  int len;
  if (sscanf (*p, "%lf%n", pValue, &len) != 1) return MagickFalse;
  *p += len;
  char c = **p;
  switch (c) {
    case 'c':
    case 'C':
    case '%':
      *pValue = floor (*pValue/100.0 * dim + 0.5);
      if (*pValue < 1) *pValue = 1;
      (*p)++;
      break;
    case 'p':
    case 'P':
      *pValue = floor (*pValue * dim + 0.5);
      if (*pValue < 1) *pValue = 1;
      (*p)++;
      break;
    case '\0':
    case 'x':
    case 'X':
      /* nothing */ ;
      break;
    default:
      printf ("Bad terminator? [%c]\n", c);
  }
  return MagickTrue;
}


static MagickBooleanType ParseWindow (
  deintegimT * pch,
  ssize_t imgWidth,
  ssize_t imgHeight
)
// Expected format: WxH
// where W and H are numbers optionally followed by '%' or 'c' or 'p'.
// Returns whether okay.
{
  double w=0, h=0;

  char * p = pch->sWindow;

  if (!*p) return MagickFalse;

  if (!GetNumber (&p, &w, imgWidth)) {
    fprintf (stderr, "deintegim error: Window needs two dimensions.\n");
    return MagickFalse;
  }

  if (*p!='x') {
    fprintf (stderr, "deintegim error: Window needs 'x' betweem two dimensions.\n");
    return MagickFalse;
  }

  p++;
  if (!GetNumber (&p, &h, imgHeight)) {
    fprintf (stderr, "deintegim error: Window needs two dimensions.\n");
    return MagickFalse;
  }

  if (w < 1 || h < 1) {
    fprintf (stderr, "deintegim error: Window dimensions must be >=1.\n");
    return MagickFalse;
  }

  if (*p) {
    fprintf (stderr, "deintegim error: Window dimensions have extra characters [%s].\n", p);
    return MagickFalse;
  }

  pch->winWidth = w;
  pch->winHeight = h;

  return MagickTrue;
}


#if defined(MAGICKCORE_HDRI_SUPPORT)
#define ADD_HALF
#else
#define ADD_HALF +0.5
#endif


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image * deintegim2 (
  Image **images,
  deintegimT * pch,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  CacheView
    *imageA_view,
    *imageB_view,
    *out_view;

  ssize_t
    y,
    inBmult,
    outXmult;

  Image * imageA = *images;

  MagickBooleanType
    IsAlphaOn = IS_ALPHA_CH(imageA),
    status;

  if (imageA->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",imageA->filename);

  assert(imageA != (Image *) NULL);
  assert(imageA->signature == MAGICK_CORE_SIG);
  if (imageA->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",imageA->filename);

  Image * imageB = GetNextImageInList (imageA);
  if (!imageB) {
    fprintf (stderr, "deintegim2: needs two images, the same size\n");
    return MagickFalse;
  }

  if (imageA->columns != imageB->columns ||
      imageA->rows    != imageB->rows)
  {
    fprintf (stderr, "deintegim2: the two images must be the same size\n");
    return MagickFalse;
  }

  if (SetNoPalette (imageA, exception) == MagickFalse)
    return MagickFalse;

  if (SetNoPalette (imageB, exception) == MagickFalse)
    return MagickFalse;

  //const ssize_t incA = Inc_ViewPixPtr (imageA);
  //const ssize_t incB = Inc_ViewPixPtr (imageB);

  if (pch->do_verbose) {
    fprintf (pch->fh_data,
             "deintegim2: Input image [%s] depth is %i IsAlpha %s\n",
             imageA->filename,
             (int)imageA->depth,
             IsAlphaOn ? "ON":"OFF"
            );
  }

  pch->winWidth = pch->winHeight = 1;

  if (pch->sWindow) {
    if (! ParseWindow (pch, imageA->columns, imageA->rows))
      return (Image *)NULL;
  }

  if (pch->do_verbose)
    fprintf (pch->fh_data, "deintegim Window: %lix%li\n",
      pch->winWidth, pch->winHeight);

  ssize_t winSemiHt = (pch->winHeight+1) / 2;
  ssize_t winSemiWi = (pch->winWidth+1) / 2;

  switch (pch->alphaAction) {
    case aaAuto:
      pch->do_regardalpha = IsAlphaOn;
      if (pch->do_verbose)
        fprintf (pch->fh_data, "deintegim2: alphaaction auto: regardalpha %s\n",
          pch->do_regardalpha ? "TRUE" : "FALSE");
      break;
    case aaRegard:
      pch->do_regardalpha = MagickTrue;
      break;
    case aaDisregard:
      pch->do_regardalpha = MagickFalse;
      break;
  }

  MagickBooleanType UseAlpha = (IsAlphaOn && pch->do_regardalpha);

  // Note: We can't do this in-place,
  // because an output pixel generally depends on a _later_ input pixel.

  new_image=CloneImage(imageA, imageA->columns, imageA->rows, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  // Just in case image had a palette.
  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  //inXmult = Inc_ViewPixPtr (imageA);
  inBmult = Inc_ViewPixPtr (imageB);
  outXmult = Inc_ViewPixPtr (new_image);

  status=MagickTrue;

  imageA_view=AcquireVirtualCacheView(imageA,exception);
  imageB_view=AcquireVirtualCacheView(imageB,exception);

  out_view=AcquireAuthenticCacheView(new_image,exception);

#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) shared(status) \
    MAGICK_THREADS(new_image,new_image,new_image->rows,1)
#endif
  for (y=0; y < (ssize_t) new_image->rows; y++)
  {
    const VIEW_PIX_PTR
      *pB;

    VIEW_PIX_PTR
      *q_out;

    ssize_t
      y1, y2,
      x;

    long double
      alpha;

    if (status==MagickFalse) continue;

/*===
    p=GetCacheViewVirtualPixels(imageA_view,0,y2,imageA->columns,1,exception);
    if (!p) {
      fprintf (stderr, "deintegim2: bad GetCacheViewVirtualPixels imageA_view\n");
      status=MagickFalse;
    }

    if (y1 >= 0) {
      p_prev=GetCacheViewVirtualPixels(imageA_view,0,y1,imageA->columns,1,exception);
      if (!p_prev) {
        fprintf (stderr, "deintegim2: bad GetCacheViewAuthenticPixels p_prev\n");
        status=MagickFalse;
      }
    } else {
      p_prev = NULL;
    }
===*/

    pB=GetCacheViewVirtualPixels(imageB_view,0,y,imageB->columns,1,exception);
    if (!pB) {
      fprintf (stderr, "deintegim2: bad GetCacheViewVirtualPixels imageB_view\n");
      status=MagickFalse;
    }

    q_out=GetCacheViewAuthenticPixels(out_view,0,y,new_image->columns,1,exception);
    if (!q_out) {
      fprintf (stderr, "deintegim2: bad GetCacheViewAuthenticPixels out_view\n");
      status=MagickFalse;
    }

    alpha = 1.0;

    if (status==MagickFalse) continue;

    long double
      left_r = 0, up_r = 0, upleft_r = 0,
      left_g = 0, up_g = 0, upleft_g = 0,
      left_b = 0, up_b = 0, upleft_b = 0,
      left_a = 0, up_a = 0, upleft_a = 0;

    for (x=0; x < (ssize_t) new_image->columns; x++)
    {
      double factX = GET_PIXEL_RED(imageB, pB) / (long double)QuantumRange;
      double factY = GET_PIXEL_GREEN(imageB, pB) / (long double)QuantumRange;

      ssize_t winW = floor ((pch->winWidth-1) * factX + 1.5);
      winSemiWi = floor (winW/2.0 + 0.5);

      ssize_t x1 = x - winSemiWi + pch->offsetx;
      ssize_t x2 = x1 + winW;

      if (x1 < 0) x1 = -1;
      if (x2 > imageA->columns-1) x2 = imageA->columns-1;
      ssize_t winWi = x2 - x1;
      if (winWi==0) winWi = 1;

      ssize_t winH = floor ((pch->winHeight-1) * factY + 1.5);
      winSemiHt = floor (winH/2.0 + 0.5);

      y1 = y - winSemiHt + pch->offsety;
      y2 = y1 + winH;

      if (y1 < 0) y1 = -1;
      if (y2 > imageA->rows-1) y2 = imageA->rows-1;
      ssize_t winHt = y2 - y1;
      if (winHt==0) winHt = 1;

      int winArea = winHt * winWi;

      if (x1 >= 0) {
        //const VIEW_PIX_PTR * pv = p+x1*inXmult;
        const VIEW_PIX_PTR * pv = GetCacheViewVirtualPixels(imageA_view,x1,y2,1,1,exception);
        if (!pv) {
          fprintf (stderr, "deintegim2: bad GetCacheViewVirtualPixels imageA_view\n");
          status=MagickFalse;
        }
        left_r = GET_PIXEL_RED(imageA, pv);
        left_g = GET_PIXEL_GREEN(imageA, pv);
        left_b = GET_PIXEL_BLUE(imageA, pv);
        left_a = GET_PIXEL_ALPHA(imageA, pv);
      }

      if (y1 >= 0) {
        //const VIEW_PIX_PTR * pv = p_prev+x2*inXmult;
        const VIEW_PIX_PTR * pv = GetCacheViewVirtualPixels(imageA_view,x2,y1,1,1,exception);
        if (!pv) {
          fprintf (stderr, "deintegim2: bad GetCacheViewVirtualPixels imageA_view\n");
          status=MagickFalse;
        }
        up_r = GET_PIXEL_RED(imageA, pv);
        up_g = GET_PIXEL_GREEN(imageA, pv);
        up_b = GET_PIXEL_BLUE(imageA, pv);
        up_a = GET_PIXEL_ALPHA(imageA, pv);
        if (x1 >= 0) {
          //const VIEW_PIX_PTR * pv = p_prev + x1*inXmult;
          const VIEW_PIX_PTR * pv = GetCacheViewVirtualPixels(imageA_view,x1,y1,1,1,exception);
          if (!pv) {
            fprintf (stderr, "deintegim2: bad GetCacheViewVirtualPixels imageA_view\n");
            status=MagickFalse;
          }
          upleft_r = GET_PIXEL_RED(imageA, pv);
          upleft_g = GET_PIXEL_GREEN(imageA, pv);
          upleft_b = GET_PIXEL_BLUE(imageA, pv);
          upleft_a = GET_PIXEL_ALPHA(imageA, pv);
        }
      }

      // point to bottom-right of window
      //const VIEW_PIX_PTR * pbr = p + x2*inXmult;
      const VIEW_PIX_PTR * pbr = GetCacheViewVirtualPixels(imageA_view,x2,y2,1,1,exception);
      if (!pbr) {
        fprintf (stderr, "deintegim2: bad GetCacheViewVirtualPixels imageA_view\n");
        status=MagickFalse;
      }

      if (IsAlphaOn) {
        // If alpha isn't on, image alpha values are 100%.
        alpha = GET_PIXEL_ALPHA(imageA,pbr) - left_a - up_a + upleft_a;

        SET_PIXEL_ALPHA (new_image, alpha / winArea ADD_HALF, q_out);
      }

      if (UseAlpha && alpha != 0) {
        alpha /= (long double)QuantumRange;
      } else {
        alpha = winArea;
      }

      if (pch->dontDivide) alpha = 1.0; // temp hack

      if (alpha == 0) {
        SET_PIXEL_RED (new_image, 0, q_out);
        SET_PIXEL_GREEN (new_image, 0, q_out);
        SET_PIXEL_BLUE (new_image, 0, q_out);
      } else {
        SET_PIXEL_RED (new_image,
          (GET_PIXEL_RED(imageA,pbr) - left_r - up_r + upleft_r)/alpha ADD_HALF,
          q_out);

        SET_PIXEL_GREEN (new_image,
          (GET_PIXEL_GREEN(imageA,pbr) - left_g - up_g + upleft_g)/alpha ADD_HALF,
          q_out);

        SET_PIXEL_BLUE (new_image,
          (GET_PIXEL_BLUE(imageA,pbr) - left_b - up_b + upleft_b)/alpha ADD_HALF,
          q_out);
      }

      pB += inBmult;
      q_out += outXmult;
    }

    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }
  }

  imageA_view=DestroyCacheView(imageA_view);
  imageB_view=DestroyCacheView(imageB_view);
  out_view=DestroyCacheView(out_view);

  DeleteImageFromList (&imageB);

  return (new_image);
}


ModuleExport size_t deintegim2Image(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  MagickBooleanType
    status;

  deintegimT
    di;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &di);
  if (status == MagickFalse)
    return (-1);

  int ListLen = (int)GetImageListLength(*images);
  if (ListLen != 2) {
    fprintf (stderr, "deintegim2: needs 2 images\n");
    return (-1);
  }

  Image * new_image = deintegim2 (images, &di, exception);
  if (!new_image) return(-1);

  ReplaceImageInList (images, new_image);
  *images=GetFirstImageInList(new_image);

  return(MagickImageFilterSignature);
}
