/* Updated:
     5-April-2018 for v7.0.7-28, and correct intensity bug.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <ctype.h>
#include "vsn_defines.h"

#include "findsinks.inc"


typedef struct {
  MagickBooleanType
    verbose,
    findPeaks;
} findsinksT;

#define VERSION "findsinks v1.0  Copyright (c) 2017 Alan Gibson"

static void usage (void)
{
  printf ("Usage: -process 'findsinks [OPTION]...'\n");
  printf ("Find sinks or peaks.\n");
  printf ("\n");
  printf ("  p, findPeaks            find peaks instead of sinks\n");
  printf ("  v, verbose              write text information to stdout\n");
  printf ("     version              write version information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  // LocaleCompare is not case-sensitive,
  // so we use strcmp for the short option.

  if ((strcmp(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "findsinks: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  findsinksT * pfs
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status = MagickTrue;

  pfs->findPeaks = MagickFalse;
  pfs->verbose = MagickFalse;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "p", "findPeaks")==MagickTrue) {
      pfs->findPeaks = MagickTrue;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pfs->verbose = MagickTrue;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "findsinks: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (pfs->verbose) {
    fprintf (stderr, "findsinks options: ");
    if (pfs->findPeaks)  fprintf (stderr, " findPeaks");
    if (pfs->verbose)    fprintf (stderr, " verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}



static MagickBooleanType SetImgGray (
  Image *image,
  ExceptionInfo *exception)
// Sets R,G and B channels to the pixel intensity.
// Does not change alpha.
// Returns whether okay.
{
  MagickBooleanType okay = MagickTrue;
  ssize_t y;

  CacheView * in_view = AcquireAuthenticCacheView (image, exception);

  for (y = 0; y < image->rows; y++) {
    if (!okay) continue;

    VIEW_PIX_PTR *p = GetCacheViewAuthenticPixels(
      in_view,0,y,image->columns,1,exception);
    if (!p) {okay = MagickFalse; continue; }

    ssize_t x;
    for (x = 0; x < image->columns; x++) {

      MagickRealType v = GetPixelIntensity (image, p);

      SET_PIXEL_RED   (image, v, p);
      SET_PIXEL_GREEN (image, v, p);
      SET_PIXEL_BLUE  (image, v, p);

      p += Inc_ViewPixPtr (image);
    }
    if (SyncCacheViewAuthenticPixels(in_view,exception) == MagickFalse)
      okay = MagickFalse;
  }
  in_view  = DestroyCacheView (in_view);

  return okay;
}


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *findsinks (
  Image *image,
  findsinksT * pfs,
  ExceptionInfo *exception)
{
  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (pfs->verbose) {
    fprintf (stderr, "findsinks: Input image [%s] %ix%i depth is %i\n",
             image->filename,
             (int)image->columns, (int)image->rows,
             (int)image->depth);
  }

  // Clone, make grayscale. Perhaps negate.

  Image * new_img = CloneImage (image, 0, 0, MagickTrue, exception);
  if (!new_img) return NULL;

//#if IMV6OR7==6
//  if (!GrayscaleImage (new_img, MSPixelIntensityMethod)) return NULL;
//#else
//  if (!GrayscaleImage (new_img, MSPixelIntensityMethod, exception)) return NULL;
//#endif

  SetImgGray (new_img, exception);

  Quantum outNotFnd;
  if (pfs->findPeaks) {
#if IMV6OR7==6
    if (!NegateImage (new_img, MagickFalse)) return NULL;
#else
    ChannelType channel_mask = SetImageChannelMask (
      new_img,
      RedChannel | GreenChannel | BlueChannel);

    if (!NegateImage (new_img, MagickFalse, exception)) return NULL;

    SetImageChannelMask (new_img, channel_mask);
#endif
    outNotFnd = 0;
  } else {
    outNotFnd = QuantumRange;
  }

  Image * new_img2 = LightenNonSinks (new_img, pfs->verbose, exception);
  if (!new_img2) return NULL;

  DestroyImage (new_img);

  // Where new_img2 has blue at QuantumRange (or zero),
  //   make image pixel white (or black).
  // Put another way: 
  //   where new_img2 has blue at not QuantumRange (or not zero),
  //   copy colours from image.

  CacheView * in_view = AcquireVirtualCacheView (image, exception);
  CacheView * out_view = AcquireAuthenticCacheView (new_img2, exception);

  MagickBooleanType okay = MagickTrue;
  ssize_t y;

  for (y = 0; y < image->rows; y++) {
    if (!okay) continue;

    VIEW_PIX_PTR *q = GetCacheViewAuthenticPixels(
      out_view,0,y,image->columns,1,exception);
    if (!q) {okay = MagickFalse; continue; }

    const VIEW_PIX_PTR *pIn = GetCacheViewVirtualPixels(
      in_view,0,y,image->columns,1,exception);
    if (!pIn) {okay = MagickFalse; continue; }

    ssize_t x;
    for (x = 0; x < image->columns; x++) {

      if ( GET_PIXEL_BLUE (new_img2, q) == QuantumRange) {
        SET_PIXEL_RED   (new_img2, outNotFnd, q);
        SET_PIXEL_GREEN (new_img2, outNotFnd, q);
        SET_PIXEL_BLUE  (new_img2, outNotFnd, q);
      } else {
        SET_PIXEL_RED   (new_img2, GET_PIXEL_RED(image,pIn), q);
        SET_PIXEL_GREEN (new_img2, GET_PIXEL_GREEN(image,pIn), q);
        SET_PIXEL_BLUE  (new_img2, GET_PIXEL_BLUE(image,pIn), q);
      }

      SET_PIXEL_ALPHA (new_img2, GET_PIXEL_ALPHA(image,pIn), q);

      q += Inc_ViewPixPtr (new_img2);
      pIn += Inc_ViewPixPtr (image);
    }
    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse)
      okay = MagickFalse;
  }
  out_view = DestroyCacheView (out_view);
  in_view  = DestroyCacheView (in_view);

  return (new_img2);
}


ModuleExport size_t findsinksImage (
  Image **images,
  const int argc,
  const char **argv,
  ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  findsinksT s3d;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &s3d);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = findsinks (image, &s3d, exception);

    ReplaceImageInList(&image,new_image);
    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
