/* Updated:
     3-April-2018 for v7.0.7-28
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <math.h>
#include "vsn_defines.h"

#if IMV6OR7==7
#include "MagickCore/gem-private.h"
#endif

#include "chklist.h"


static void usage (void)
{
  printf ("Usage: -process 'geodist [OPTION]...'\n");
  printf ("Distort image.\n");
  printf ("\n");
  printf ("  s, style N     style N: 0 to 3\n");
}

static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  int *pstyle)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  *pstyle = 0;

  status = MagickTrue;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];
    //printf ("Arg %i [%s]\n", i, pa);

    if (IsArg (pa, "s", "style")==MagickTrue) {
      i++;
      *pstyle = atoi (argv[i]);
    } else {
      fprintf (stderr, "ERROR\n");
      status = MagickFalse;
    }
  }
  if (status == MagickFalse)
    usage ();

  return (status);
}


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *GeoDImage(const Image *image,
  int style,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  long double
    twoPi;

  CacheView
    *image_view,
    *out_view;

  ssize_t
    y;

  MagickBooleanType
    status;

  double
    dx,
    dy;

  assert(image != (Image *) NULL);
  assert(image->signature == MAGICK_CORE_SIG);
  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  twoPi = 2 * M_PI;

  dx = image->columns * 0.10;
  dy = image->rows    * 0.10;

  // Clone, and we must initialise.
  new_image=CloneImage(image, 0, 0, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  out_view = AcquireAuthenticCacheView(new_image,exception);

  image_view = AcquireVirtualCacheView(image,exception);

  status = MagickTrue;

#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) shared(status) \
    MAGICK_THREADS(image,image,image->rows,1)
#endif
  for (y=0; y < (ssize_t) image->rows; y++)
  {
    VIEW_PIX_PTR
      *q_out;

    register ssize_t
      x;

    if (status == MagickFalse)
      continue;

    q_out=GetCacheViewAuthenticPixels(out_view,0,y,new_image->columns,1,exception);
    if (q_out == (const VIEW_PIX_PTR *) NULL)
      status=MagickFalse;

    for (x=0; x < (ssize_t) image->columns; x++)
    {
      PIX_INFO
        src_pixel;

      switch (style)
      {
        case 0:
        default:
        {
          break;
        }
        case 1:
        {
          double
            hue, saturation, brightness;

// For v7, use:
// static MagickBooleanType sRGBTransformImage(Image *image,
//  const ColorspaceType colorspace,ExceptionInfo *exception)

/*++
          {
  double r=(double) GET_PIXEL_RED(image,q_out);
  double g=(double) GET_PIXEL_GREEN(image,q_out);
  double b=(double) GET_PIXEL_BLUE(image,q_out);
  double min = r < g ? r : g;
  if (b < min) min = b;
  double max = r > g ? r : g;
  if (b > max) max = b;
  if (max > 0.0) {
    delta = max-min;
    saturation = delta/max;
    brightness = QuantumScale*max;
    if (delta == 0.0)
      return;
    if (r == max)
      hue = (g-b)/delta;
    else {
      if (g == max) hue = 2.0+(b-r)/delta;
      else          hue = 4.0+(r-g)/delta;
    }
    hue /= 6.0;
    if (hue < 0.0) hue += 1.0;
  } else {
    saturation = 0.0;
    brightness = 0.0;
    hue = 0.0;
  }
          }
++*/

          ConvertRGBToHSB(
              GET_PIXEL_RED(image,q_out), GET_PIXEL_GREEN(image,q_out), GET_PIXEL_BLUE(image,q_out),
              &hue, &saturation, &brightness);

          // hsb are 0.0 to 1.0

          dx = (saturation - 0.5) * image->columns;
          dy = (brightness - 0.5) * image->rows;

          break;
        }
        case 2:
        {
          double
            hue, saturation, brightness, rads;

          ConvertRGBToHSB(
              GET_PIXEL_RED(image,q_out), GET_PIXEL_GREEN(image,q_out), GET_PIXEL_BLUE(image,q_out),
              &hue, &saturation, &brightness);

          rads = twoPi * hue;
          dx = sin(rads) * brightness * image->columns;
          dy = sin(rads) * brightness * image->rows;

          break;
        }
      }

#if IMV6OR7==6
      InterpolateMagickPixelPacket (
        image, image_view,
        image->interpolate,
        x + dx, y + dy, &src_pixel, exception);
#else
     InterpolatePixelInfo (
        image, image_view,
        image->interpolate,
        x + dx, y + dy, &src_pixel, exception);
#endif

      SET_PIXEL_RED   (new_image, src_pixel.red,   q_out);
      SET_PIXEL_GREEN (new_image, src_pixel.green, q_out);
      SET_PIXEL_BLUE  (new_image, src_pixel.blue,  q_out);
#if IMV6OR7==6
      SET_PIXEL_ALPHA (new_image, QuantumRange - src_pixel.opacity, q_out);
#else
      SET_PIXEL_ALPHA (new_image, src_pixel.alpha, q_out);
#endif

      q_out += Inc_ViewPixPtr (new_image);
    }
    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse)
      status=MagickFalse;
  }

  image_view=DestroyCacheView(image_view);
  out_view=DestroyCacheView(out_view);

  //chklist("end GeoDImage new_image",&new_image);

  return (new_image);
}


ModuleExport size_t geodistImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  int
    style;

  MagickBooleanType
    Error;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  Error = !menu (argc, argv, &style);

  if (Error == MagickTrue) {
    return (-1);
  }

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = GeoDImage (image, style, exception);
    if (new_image == (Image *) NULL)
      return(-1);

    //chklist("after GeoDImage: &new_image",&new_image);
    //chklist("after GeoDImage: images",images);
    //chklist("after GeoDImage: images",images);

    ReplaceImageInList(&image,new_image);
    //chkentry("after replace: &image",&image);
    *images=GetFirstImageInList(image);
    //chklist("after replace: images",images);
  }

  //chklist("end geodistImage: images",images);

  return(MagickImageFilterSignature);
}
