#ifndef GOLDSECTSRCH_INC
#define GOLDSECTSRCH_INC


#include <stdio.h>
#include <stdlib.h>
#include <math.h>


/*
  Ref: https://en.wikipedia.org/wiki/Golden_section_search

       +------+------+-----+
       x0     x1     x2    x3
       <--a--><--b--><--c-->

Required:

  b+c   a   a+b+c
  --- = - = -----
   a    b    b+c

Which implies:
  b^2 + bc = a^2
  c = (a^2-b^2) / b

  ab + ac = ab + b^2 + bc
  ac = b^2 + bc
  c(a-b) = b^2
  c = b^2 / (a-b)

So:
  (a^2-b^2) / b = b^2 / (a-b)

  (a^2-b^2)*(a-b) = b^3

  a^3-a.b^2 -a^2.b + b^3 = b^3
  a^3-a.b^2 -a^2.b = 0
  a^2-b^2 -a.b = 0
  a^2 -a.b -b^2 = 0

      +b +-sqrt (b^2 + 4.1.b^2)
  a = -------------------------
                  2

a = b * (1 +- sqrt(5)) / 2

a/b = phi
a = b.phi

But
  c = b^2 / (a-b)
so:
  c = b^2 / (b(phi-1))
  c = b / (phi-1)
but phi-1 = 1/phi so
  c = b.phi
*/



typedef double getYcbT (double x, void * pg, int * okay);

typedef struct {
  double xLo;
  double xHi;
  double epsX;
  double epsY;
  getYcbT *getYcb; // This function will be called to calculate y.
  int verbose;

  // Returns results:
  double fndX;
  double xPlusOrMinus;
  double calcY;
  int dodgy;
  int nIter;
} goldSectSrchT;

// FIXME: perhaps also return if less than xLo or greater than xHi.


/* This is a framework for searching for a minimum
   within a 1D inverse-unimodal function,
   such as finding the resize that makes an image best match another image.
*/

static void InitGoldSectSrch (goldSectSrchT *goldSectSrch)
{
  goldSectSrch->xLo  = 0.5;
  goldSectSrch->xHi  = 1.5;
  goldSectSrch->epsX = 1e-2;
  goldSectSrch->epsY = 0;
  goldSectSrch->dodgy = 0;
  goldSectSrch->nIter = 0;
  goldSectSrch->getYcb = NULL;
  goldSectSrch->verbose = 0;
}


static int GoldSectSrch (
  goldSectSrchT *goldSectSrch,
  void * pData)
// This finds x, xLo <= x <= xHi, such that getY(x) is a minimum.
// Returns 1 if okay, otherwise 0.
{
  int okay = 1;

  goldSectSrch->dodgy = 0;

  double x0 = goldSectSrch->xLo;
  double x3 = goldSectSrch->xHi;
  double x1, x2;

  double origX0 = x0;
  double origX3 = x3;

  double phi = (1.0 + sqrt(5.0))/2.0;

  double y0 = goldSectSrch->getYcb (x0, pData, &okay);
  if (!okay) return 0;
  double y3 = goldSectSrch->getYcb (x3, pData, &okay);
  if (!okay) return 0;

  x1 = x3 - (x3-x0) / phi;
  x2 = x0 + (x3-x0) / phi;

  if (goldSectSrch->verbose > 1) {
    fprintf (stderr, "x1=%g x2=%g phi=%g phi=%g\n",
      x1, x2,
      (x3-x1)/(x1-x0),
      (x3-x0)/(x3-x1));
  }

  goldSectSrch->dodgy = 1;

  double y1 = goldSectSrch->getYcb (x1, pData, &okay);
  if (!okay) return 0;
  double y2 = goldSectSrch->getYcb (x2, pData, &okay);
  if (!okay) return 0;

  if (y0 < y1 && y1 < y2 && y2 < y3) {
    if (goldSectSrch->verbose) {
      fprintf (stderr, "x: %g %g %g %g\n", x0, x1, x2, x3);
      fprintf (stderr, "y: %g %g %g %g\n", y0, y1, y2, y3);
      fprintf (stderr, "gss: Solution possibly < %g\n", x0);
    }
    return 0;
  }

  if (y0 > y1 && y1 > y2 && y2 > y3) {
    if (goldSectSrch->verbose) {
      fprintf (stderr, "x: %g %g %g %g\n", x0, x1, x2, x3);
      fprintf (stderr, "y: %g %g %g %g\n", y0, y1, y2, y3);
      fprintf (stderr, "gss: Solution possibly > %g\n", x3);
    }
    return 0;
  }

  if (y1 > y0 || y2 > y3) {
    if (goldSectSrch->verbose) {
      fprintf (stderr, "x: %g %g %g %g\n", x0, x1, x2, x3);
      fprintf (stderr, "y: %g %g %g %g\n", y0, y1, y2, y3);
      fprintf (stderr, "gss: dodgy initial y1 or y2\n");
    }
    return 0;
  }

  int cnt=0;

  goldSectSrch->dodgy = 0;

  for (;;) {

    if (fabs (x3-x0) <= goldSectSrch->epsX) break;
    if (fabs (y3-y0) <= goldSectSrch->epsY) break;

    if (y1 > y0 && y1 > y3) {
      if (goldSectSrch->verbose) {
        fprintf (stderr, "x: %g %g %g %g\n", x0, x1, x2, x3);
        fprintf (stderr, "y: %g %g %g %g\n", y0, y1, y2, y3);
        fprintf (stderr, "gss: dodgy y1\n");
      }
      goldSectSrch->dodgy = 1;
      break;
    }

    if (y2 > y0 && y2 > y3) {
      if (goldSectSrch->verbose) {
        fprintf (stderr, "x: %g %g %g %g\n", x0, x1, x2, x3);
        fprintf (stderr, "y: %g %g %g %g\n", y0, y1, y2, y3);
        fprintf (stderr, "gss: dodgy y2\n");
      }
      goldSectSrch->dodgy = 1;
      break;
    }

    if (cnt++ > 100) break;

    if (goldSectSrch->verbose > 1) {
      fprintf (stderr, "%i x: %g %g %g %g\n", cnt, x0, x1, x2, x3);
      fprintf (stderr, "y: %g %g %g %g\n", y0, y1, y2, y3);
    }

    if (y1 < y2) {
      x3 = x2;
      y3 = y2;
      x2 = x1;
      y2 = y1;
      x1 = x3 - (x3-x0) / phi;
      y1 = goldSectSrch->getYcb (x1, pData, &okay);
    } else {
      x0 = x1;
      y0 = y1;
      x1 = x2;
      y1 = y2;
      x2 = x0 + (x3-x0) / phi;
      y2 = goldSectSrch->getYcb (x2, pData, &okay);
    }
    if (!okay) return 0;
  }

  if (goldSectSrch->verbose) {
    fprintf (stderr, "\ngss: cnt=%i\nx: %g %g %g %g\n",
      cnt, x0, x1, x2, x3);
    fprintf (stderr, "y: %g %g %g %g\n", y0, y1, y2, y3);
  }

  if (origX0 == x0) {
    fprintf (stderr, "gss: dodgy bottom\n");
    goldSectSrch->dodgy = 1;
  }
  if (origX3 == x3) {
    fprintf (stderr, "gss: dodgy top\n");
    goldSectSrch->dodgy = 1;
  }

  goldSectSrch->fndX  = (x0+x3) / 2.0;
  goldSectSrch->xPlusOrMinus = (x3-x0) / 2.0;
  goldSectSrch->calcY = goldSectSrch->getYcb (goldSectSrch->fndX, pData, &okay);
  goldSectSrch->nIter = cnt;

  return 1;
}


#endif
