/*
    Reference: im.snibgo.com/growcut.htm

    Updated:
      3-April-2018 for v7.0.7-28.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "vsn_defines.h"

#define CH_R 1
#define CH_G 2
#define CH_B 4


#define FIRST_LINE_SHORTCUT 1

#if MAGICKCORE_QUANTUM_DEPTH==8
#define FIRST_LINE_SHORTCUT 0
#endif


typedef enum {
  a_scan_merge = 0,
  a_grow_cut = 1
} algorithmT;

typedef enum {
  direct = 0,
  clust_mean = 1
} comp_methodT;

typedef struct {
  MagickBooleanType
    do_verbose,
    do_verbose2,
    do_write_frames;

  int
    max_iter;

  float
    threshold, // like fuzz, eg 20%. Could come from "-fuzz".
    sq_threshold; // from 0.0 to 1.0

  unsigned int
    channels; // flags

  int
    num_chan;

  int
    precision,
    n_connected;

  int
    neighbour_offset[8];

  //ssize_t
  //  current_y;

  char
    write_filename[MaxTextExtent];

  // Input images
  Image
    *input_image,
    *swipe_image,
    *divide_image;

  // Output images
  Image
    *cluster_of, // R channel is cluster number of this pixel; B and G unused.
    *swipe_of;

  CacheView
    *input_view,
    *swipe_view,
    *divide_view;

  //const VIEW_PIX_PTR
  //  *input_pp3,
  //  *divide_pp;
} growcutT;


/*
We could output the unique colours of the cluster averages,
possible filtered by minimum size or the N largest clusters.
*/

static void usage (void)
{
  printf ("Usage: -process 'growcut [OPTION]...'\n");
  printf ("Segments an image.\n");
  printf ("\n");
  printf ("  i,  max_iterations N    maximum number of iterations (0=no limit)\n");
  printf ("  n,  n_connected N       connected (4 or 8)\n");
  printf ("  c,  channel string      which channels to use in comparisons\n");
  printf ("  v,  verbose             write text information to stdout\n");
  printf ("  v2, verbose2            write more text information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  growcutT * ssm
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  ssm->max_iter = 0;
  ssm->n_connected = 8;
  ssm->channels = (CH_R | CH_G | CH_B);
  ssm->do_write_frames = MagickFalse;
  ssm->write_filename[0] = '\0';
  ssm->do_verbose = ssm->do_verbose2 = MagickFalse;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "i", "max_iterations")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i]))
      {
        fprintf (stderr, "Non-numeric argument to 'max_iterations'.\n");
        status = MagickFalse;
      }
      else ssm->max_iter = (comp_methodT)atoi(argv[i]);
    } else if (IsArg (pa, "n", "n_connected")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i]))
      {
        fprintf (stderr, "Non-numeric argument to 'n_connected'.\n");
        status = MagickFalse;
      }
      else ssm->n_connected = (comp_methodT)atoi(argv[i]);
    } else if (IsArg (pa, "c", "channel")==MagickTrue) {
      i++;
      ssm->channels = 0;
      const char * p = argv[i];
      while (*p) {
        switch (toupper ((int)*p)) {
          case 'R':
            ssm->channels |= CH_R;
            break;
          case 'G':
            ssm->channels |= CH_G;
            break;
          case 'B':
            ssm->channels |= CH_B;
            break;
          case 'L':
            ssm->channels |= CH_R;
            break;
          case 'A':
            ssm->channels |= CH_G;
            break;
          default:
            fprintf (stderr, "Invalid 'channels' [%s]\n", argv[i]);
            status = MagickFalse;
        }
        p++;
      }
    } else if (IsArg (pa, "w", "write")==MagickTrue) {
      ssm->do_write_frames = MagickTrue;
      i++;
      CopyMagickString (ssm->write_filename, argv[i], MaxTextExtent);
      if (!*ssm->write_filename) {
        fprintf (stderr, "Invalid 'write' [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      ssm->do_verbose = MagickTrue;
    } else if (IsArg (pa, "v2", "verbose2")==MagickTrue) {
      ssm->do_verbose = MagickTrue;
      ssm->do_verbose2 = MagickTrue;
    } else {
      fprintf (stderr, "growcut: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  ssm->num_chan = 0;
  if (ssm->channels & CH_R) ssm->num_chan++;
  if (ssm->channels & CH_G) ssm->num_chan++;
  if (ssm->channels & CH_B) ssm->num_chan++;
  if (ssm->channels == 0) {
    fprintf (stderr, "No channels\n");
    status = MagickFalse;
  }

  if (ssm->do_verbose) {
    fprintf (stderr, "growcut options: ");

    if (ssm->max_iter) fprintf (stderr, " max_iterations %i", ssm->max_iter);

    if (ssm->n_connected != 8) fprintf (stderr, " n_connected %i", ssm->n_connected);

    if (ssm->channels != (CH_R | CH_G | CH_B)) {
      fprintf (stderr, " channels ");
      if (ssm->channels & CH_R) fprintf (stderr, "R");
      if (ssm->channels & CH_G) fprintf (stderr, "G");
      if (ssm->channels & CH_B) fprintf (stderr, "B");
    }

    if (ssm->do_verbose2) fprintf (stderr, " verbose2");
    else {
      if (ssm->do_verbose) fprintf (stderr, " verbose");
    }
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


static MagickBooleanType InitSsm (
  Image **images,
  growcutT *pssm,
  ExceptionInfo *exception)
{
  int
   numImages,
   i;


  pssm->precision = GetMagickPrecision();

  pssm->threshold = 0;
  pssm->num_chan = 3;
  pssm->input_image = (Image *)NULL;
  pssm->swipe_image = (Image *)NULL;
  pssm->divide_image = (Image *)NULL;

  numImages = (int)GetImageListLength(*images);

  if (pssm->do_verbose) {
    fprintf (stderr, "  numImages %i\n", numImages);
  }

  if (numImages < 2) {
    fprintf (stderr, "GrowCut needs 2 or 3 images\n");
    return -1;
  }

  pssm->input_image = GetFirstImageInList(*images);

  if (pssm->input_image == (Image *)NULL) {
    fprintf (stderr, "Bad first image\n");
    return -1;
  }

  if (numImages >= 2) {
    pssm->swipe_image = GetNextImageInList(pssm->input_image);
    if (pssm->swipe_image == (Image *)NULL) {
      fprintf (stderr, "Bad swipe image\n");
      return -1;
    }
    if (pssm->swipe_image->columns != pssm->input_image->columns ||
        pssm->swipe_image->rows    != pssm->input_image->rows) {
      fprintf (stderr, "Swipe image dimensions don't match input image\n");
      return -1;
    }
  }

  if (numImages >= 3) {
    pssm->divide_image = GetNextImageInList(pssm->swipe_image);
    if (pssm->divide_image == (Image *)NULL) {
      fprintf (stderr, "Bad divide image\n");
      return -1;
    }
    if (pssm->divide_image->columns != pssm->input_image->columns ||
        pssm->divide_image->rows    != pssm->input_image->rows) {
      fprintf (stderr, "Divide image dimensions don't match input image\n");
      return -1;
    }
  }

  pssm->threshold = pssm->input_image->fuzz / QuantumRange;
  pssm->sq_threshold = pssm->threshold * pssm->threshold;

  if (pssm->do_verbose) {
    fprintf (stderr, "  threshold %.*g\n", pssm->precision, pssm->threshold);
    //fprintf (stderr, "sq_threshold %.*g\n", pssm->precision, pssm->sq_threshold);
  }

  // First four are 4-connected:
  // Neighbours are numbered:
  //   4 0 5
  //   2 . 3
  //   6 1 7

  {
    int n_x[] = {0,0,-1,+1, -1,+1,-1,+1};
    int n_y[] = {-1,+1,0,0, -1,-1,+1,+1};

    for (i = 0; i < 8; i++) {
      pssm->neighbour_offset[i] = 
        (n_y[i] + 1) * (pssm->input_image->columns + 2) + (n_x[i] + 1);

      //printf ("offset[%i]=%i\n", i, pssm->neighbour_offset[i]);
    }
  }

  return MagickTrue;
}



/*
   In the mask,
   a _black_ pixel prevents any "<= threshold" test at that pixel from succeeding.
*/

static void OpenViews (growcutT *pssm, ExceptionInfo *exception)
{
  pssm->input_view=AcquireVirtualCacheView (pssm->input_image, exception);
  if (pssm->swipe_image)  pssm->swipe_view=AcquireVirtualCacheView (pssm->swipe_image, exception);
  if (pssm->divide_image) pssm->divide_view=AcquireVirtualCacheView (pssm->divide_image, exception);
}

static void CloseViews (growcutT *pssm)
{
  if (pssm->divide_image) pssm->divide_view = DestroyCacheView (pssm->divide_view);
  if (pssm->swipe_image)  pssm->swipe_view = DestroyCacheView (pssm->swipe_view);
  pssm->input_view = DestroyCacheView (pssm->input_view);
}

static MagickBooleanType SyncViews (growcutT *pssm, ExceptionInfo *exception)
{
  return MagickTrue;
}

static MagickBooleanType SetPixPack (growcutT *pssm, ExceptionInfo *exception)
{
  //if (pssm->divide_image) {
  //  pssm->divide_pp=GetCacheViewVirtualPixels(pssm->divide_view,0,0,pssm->divide_image->columns,1,exception);
  //  if (pssm->divide_pp == (const VIEW_PIX_PTR *) NULL) return MagickFalse;
  //}
  return MagickTrue;
}

static MagickBooleanType SetThisRow (growcutT *pssm, ssize_t y, ExceptionInfo *exception)
{
  //pssm->input_pp3=GetCacheViewVirtualPixels(pssm->input_view,-1,y-1,pssm->input_image->columns+2,3,exception);
  //if (pssm->input_pp3 == (const VIEW_PIX_PTR *) NULL) return MagickFalse;

  //pssm->current_y = y;

  return MagickTrue;
}

static inline double PixPacketDiff (
  growcutT *pssm,
  const VIEW_PIX_PTR *p0,
  const VIEW_PIX_PTR *p1)
// Returns the difference, range 0.0 to 1.0.
{
  double
    d,
    diff;

  diff = 0;
  if (pssm->channels & CH_R) {
    d = (GET_PIXEL_RED(pssm->input_image, p1)   - GET_PIXEL_RED(pssm->input_image, p0)) / QuantumRange;
    diff = d * d;
  }
  if (pssm->channels & CH_G) {
    d = (GET_PIXEL_GREEN(pssm->input_image, p1) - GET_PIXEL_GREEN(pssm->input_image, p0)) / QuantumRange;
    diff += d * d;
  }
  if (pssm->channels & CH_B) {
    d = (GET_PIXEL_BLUE(pssm->input_image, p1)  - GET_PIXEL_BLUE(pssm->input_image, p0)) / QuantumRange;
    diff += d * d;
  }

  return sqrt (diff / pssm->num_chan);
}



/*---
static void chkpix (const char * msg, const VIEW_PIX_PTR *p)
{
  if (GetPixelRed (p) < 0)   printf ("%s Red < 0 ", msg);
  if (GetPixelGreen (p) < 0) printf ("%s Green < 0 ", msg);
  if (GetPixelBlue (p) < 0)  printf ("%s Blue < 0 ", msg);
}
---*/



static MagickStatusType WriteGcFrame (
  growcutT *pssm,
  Image * img,
  int frame_num,
  ExceptionInfo *exception)
{
  ImageInfo
    *ii;

  MagickStatusType
    status;

  Image
    *copy_img;

  ii = AcquireImageInfo ();

  copy_img = CloneImage(img, 0, 0, MagickTrue, exception);
  if (copy_img == (Image *) NULL)
    return(MagickFalse); // FIXME: raise error

  copy_img->scene = frame_num;

  CopyMagickString (copy_img->filename, pssm->write_filename, MaxTextExtent);

#if IMV6OR7==6
  status = WriteImage (ii, copy_img);
#else
  status = WriteImage (ii, copy_img, exception);
#endif

  DestroyImageList(copy_img);

  ii = DestroyImageInfo (ii);

  return status;
}


static inline double StrengthOf (const Image * image, const VIEW_PIX_PTR * p)
{
  double s = GET_PIXEL_ALPHA (image, p) / (double)QuantumRange;

  return (s <= 0) ? 0 : s;
}

/*---
static inline MagickBooleanType IsRgbaDiff(const VIEW_PIX_PTR * p0, const VIEW_PIX_PTR * p1)
{
  return (
    GetPixelRed (p0) != GET_PIXEL_RED(image,p1) ||
    GetPixelGreen (p0) != GET_PIXEL_GREEN(image,p1) ||
    GetPixelBlue (p0) != GET_PIXEL_BLUE(image,p1)
  );
}
---*/

static MagickBooleanType GrowCutIter (
  growcutT *pssm,
  Image **copy_swipe,
  ExceptionInfo *exception)
// Does one iteration of GrowCut.
// Replaces copy_swipe with next version of image.
// Returns whether a change was made.
{
  MagickBooleanType
    changed;

  int
    num_ch,
    cols_plus_3;

  Image
    *clone_swipe_out_image,
    *new_list;

  CacheView
    *in_view,
    *clone_swipe_out_view;

  ssize_t
    y;

  MagickBooleanType
    status = MagickTrue;


  in_view = AcquireVirtualCacheView(*copy_swipe,exception);

  // Clone the swipe image, same size, copied pixels:
  //
  clone_swipe_out_image=CloneImage(*copy_swipe, 0, 0, MagickTrue, exception);
  if (clone_swipe_out_image == (Image *) NULL)
    return(MagickFalse); // FIXME: raise error

  clone_swipe_out_view = AcquireAuthenticCacheView(clone_swipe_out_image,exception);


  changed = MagickFalse;

  num_ch = 0;

  cols_plus_3 = pssm->input_image->columns + 3;

#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) shared(status) \
    MAGICK_THREADS(clone_swipe_out_image,clone_swipe_out_image,clone_swipe_out_image->rows,1)
#endif
  for (y=0; y < (ssize_t) pssm->input_image->rows; y++)
  {
    ssize_t
      x;

    int
      i;

    double
      strength_this,
      strength_neighbour,
      diff_str,
      prev_diff;

    const VIEW_PIX_PTR
      *pp_this_swipe,
      *pp_neighbour_swipe,
      *pp_this_pix,
      *pp_neighbour;

    const VIEW_PIX_PTR
      *pp_in3;

    VIEW_PIX_PTR
      *q_clone_swipe,
      *qq_out;

    const VIEW_PIX_PTR
      *input_pp3;

    if (!status) continue;
    //status = MagickTrue;

    if (!SetThisRow (pssm, y, exception)) {
      status = MagickFalse;
      continue;
      //return MagickFalse;
    }

    pp_in3 = GetCacheViewVirtualPixels(in_view,-1,y-1,(*copy_swipe)->columns+2,3,exception);
    if (!pp_in3) {
      status=MagickFalse;
      continue;
    }

    q_clone_swipe = GetCacheViewAuthenticPixels(clone_swipe_out_view,0,y,clone_swipe_out_image->columns,1,exception);
    if (!q_clone_swipe) {
      status=MagickFalse;
      continue;
    }

    input_pp3=GetCacheViewVirtualPixels(pssm->input_view,-1,y-1,pssm->input_image->columns+2,3,exception);
    if (!input_pp3) {
      status=MagickFalse;
      continue;
    }

    for (x=0; x < (ssize_t) pssm->input_image->columns; x++)
    {
      pp_this_swipe = pp_in3 + cols_plus_3 + x;
      strength_this = StrengthOf (*copy_swipe, pp_this_swipe);

      if (strength_this < 1.0) { // If pixel can't be changed, don't bother checking.
        pp_this_pix = input_pp3 + cols_plus_3 + x;
        qq_out = q_clone_swipe + x;
        prev_diff = 0;
        for (i = 0; i < pssm->n_connected; i++) {
          pp_neighbour_swipe = pp_in3 + pssm->neighbour_offset[i] + x;
          strength_neighbour = StrengthOf (*copy_swipe, pp_neighbour_swipe);

          // This test isn't needed, but improves performance x2.
          if (strength_neighbour > 0) {
            pp_neighbour = input_pp3 + pssm->neighbour_offset[i] + x;
            //diff = 1 - PixPacketDiff (pssm, pp_this_pix, pp_neighbour);
            diff_str = (1 - PixPacketDiff (pssm, pp_this_pix, pp_neighbour))
                     * strength_neighbour;
            if (    prev_diff < diff_str
                 && strength_this + 1.0e-7 < diff_str
               )
            {
              prev_diff = diff_str;
              SET_PIXEL_RED   (clone_swipe_out_image, GET_PIXEL_RED   (*copy_swipe, pp_neighbour_swipe), qq_out);
              SET_PIXEL_GREEN (clone_swipe_out_image, GET_PIXEL_GREEN (*copy_swipe, pp_neighbour_swipe), qq_out);
              SET_PIXEL_BLUE  (clone_swipe_out_image, GET_PIXEL_BLUE  (*copy_swipe, pp_neighbour_swipe), qq_out);
              SET_PIXEL_ALPHA (clone_swipe_out_image, diff_str * QuantumRange, qq_out); // FIXME: PLUS_HALF
              num_ch++;
              changed = MagickTrue;
            }
          }
        }
      }
    }
    if (!SyncCacheViewAuthenticPixels(clone_swipe_out_view,exception))
      status=MagickFalse;

    if (!SyncViews (pssm, exception)) {
      status=MagickFalse;
      continue;
    }

    //if (!status) break;
  }

  if (!status) return MagickFalse;

  clone_swipe_out_view=DestroyCacheView(clone_swipe_out_view);
  in_view=DestroyCacheView(in_view);

  DestroyImageList(*copy_swipe);

  new_list = NewImageList();
  AppendImageToList(&new_list, clone_swipe_out_image);

  *copy_swipe = GetFirstImageInList(new_list);

  if (pssm->do_verbose2) {
    fprintf (stderr, "  num_ch = %i\n", num_ch);
  }

  return changed;
}

static Image * GrowCut (
  growcutT *pssm,
  ExceptionInfo *exception)
{
  int
    iter_num;

  iter_num = 0;

  Image
    *copy_swipe;

  // Clone the swipe image, same size, copied pixels:
  //
  copy_swipe=CloneImage(pssm->swipe_image, 0, 0, MagickTrue, exception);
  if (copy_swipe == (Image *) NULL)
    return(MagickFalse); // FIXME: raise error

  if (SetNoPalette (copy_swipe, exception) == MagickFalse)
    return (MagickFalse);

  while (GrowCutIter (pssm, &copy_swipe, exception)) {
    if (pssm->do_write_frames)
      if (!WriteGcFrame (pssm, copy_swipe, iter_num, exception)) break;

    iter_num++;
    if (pssm->max_iter && iter_num >= pssm->max_iter) break;
  }

  if (pssm->do_verbose) {
    fprintf (stderr, "iter_num=%i\n", iter_num);
  }

  return copy_swipe;
}


ModuleExport size_t growcutImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *copy_image,
    *new_list;

  growcutT
    ssm;

  MagickBooleanType
    status;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &ssm);
  if (status == MagickFalse)
    return (-1);

  status = InitSsm (images, &ssm, exception);

  if (status != MagickTrue) {
    fprintf (stderr, "InitSsm failed\n");
    return (-1);
  }

  if (ssm.do_verbose) fprintf (stderr, "OpenViews...\n");
  OpenViews      (&ssm, exception);
  if (ssm.do_verbose) fprintf (stderr, "SetPixPack...\n");
  if (!SetPixPack     (&ssm, exception)) {
    fprintf (stderr, "SetPixPack failed");
    return (-1);
  }

  if (ssm.do_verbose) fprintf (stderr, "BasicAlgorithm...\n");

  copy_image = GrowCut (&ssm, exception);

  if (copy_image == (Image *)NULL) {
    fprintf (stderr, "BasicAlgorithm failed\n");
    return (-1);
  }

  CloseViews (&ssm);

  DestroyImageList(*images);

  new_list = NewImageList();

  AppendImageToList(&new_list, copy_image);

  *images = GetFirstImageInList(new_list);

  if (!status) return (-1);

  return(MagickImageFilterSignature);
}
