#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <MagickWand/MagickWand.h>

// Inspired from a forum post: https://github.com/ImageMagick/ImageMagick/issues/5898
//
// c:\cygwin64\home\Alan\imagemagick-7.1.0-19\ImageMagick\im-win-forfx\ImageMagick
// bash snibgo\buildwand.sh snibgo\imserv
// %IM7DEV%imserv

typedef struct {
  ExceptionInfo
    *exception;

  ImageInfo
    *image_info;

  MagickBooleanType
    exitOnFileEof,
    exitOnImErr,
    exitNow;

  int
    verbose;

  char
    *sTerminate;

  FILE *
    fh_data;
} imservT;

#define VERSION "imserv v1.0  Copyright (c) 2023 Alan Gibson"

static void usage (void)
{
  printf ("Usage: imserv {OPTION}...'\n");
  printf ("Reads IM commands from stdin and executes them.\n");
  printf ("\n");
  printf ("  -feof,  --exitOnFileEof string  whether a file eof forces program exit\n");
  printf ("  -imerr, --exitOnImErr string    whether an IM error forces program exit\n");
  printf ("  -t,     --terminate string      when program receives this string, it will terminate\n");
  printf ("  -f,     --file string           write verbose text to stderr or stdout\n");
  printf ("  -v,     --verbose               write text information to stderr\n");
  printf ("  -v2,    --verbose2              write more text information to stderr\n");
  printf ("          --version               write version information to stdout\n");
  printf ("  -h,     --help                  write this help to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  // LocaleCompare is not case-sensitive,
  // so we use strcmp for the short option.

  if ((strcmp(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "imserv: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }


static MagickBooleanType menu (
  const int argc,
  const char **argv,
  imservT * pims
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status = MagickTrue;

  pims->exitOnFileEof = MagickTrue;
  pims->exitOnImErr = MagickFalse;
  pims->verbose = 0;
  pims->fh_data = stderr;
  pims->sTerminate = "EOF";
  pims->exitNow = MagickFalse;

  for (i=1; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "-feof", "--exitOnFileEof")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "true")==0) pims->exitOnFileEof = MagickTrue;
      else if (LocaleCompare(argv[i], "false")==0) pims->exitOnFileEof = MagickFalse;
      else {
        fprintf (stderr, "imserv: ERROR: option %s needs true or false\n", pa);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "-imerr", "--exitOnImErr")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "true")==0) pims->exitOnImErr = MagickTrue;
      else if (LocaleCompare(argv[i], "false")==0) pims->exitOnImErr = MagickFalse;
      else {
        fprintf (stderr, "imserv: ERROR: option %s needs true or false\n", pa);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "-t", "--terminate")==MagickTrue) {
      NEXTARG;
      pims->sTerminate = (char *)argv[i];
    } else if (IsArg (pa, "-f", "--file")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare (argv[i], "stdout")==0) pims->fh_data = stdout;
      else if (LocaleCompare (argv[i], "stderr")==0) pims->fh_data = stderr;
      else {
        fprintf (stderr, "imserv: ERROR: option %s needs stdout or stderr\n", pa);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "-v", "--verbose")==MagickTrue) {
      pims->verbose = 1;
    } else if (IsArg (pa, "-v2", "--verbose2")==MagickTrue) {
      pims->verbose = 2;
    } else if (IsArg (pa, "-version", "--version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
      pims->exitNow = MagickTrue;
    } else if (IsArg (pa, "-h", "--help")==MagickTrue) {
      usage ();
      pims->exitNow = MagickTrue;
    } else {
      fprintf (stderr, "imserv: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (status == MagickFalse) {
    usage ();
    pims->exitNow = MagickTrue;
  }

  return (status);
}


static MagickBooleanType ExecIm (imservT * pims, int argc, char **argv)
{
  MagickBooleanType
    status;

  status = MagickImageCommand (pims->image_info, argc, argv, NULL, pims->exception);

  if (status == MagickFalse) {
    fprintf (pims->fh_data, "MagickImageCommand failed\n");
    MagickError (pims->exception->severity, pims->exception->reason, pims->exception->description);
    status = MagickFalse;
  }

  if (pims->exception->severity != UndefinedException) {
    fprintf (pims->fh_data, "MagickImageCommand exception\n");
    CatchException (pims->exception);
    status = MagickFalse;
  }

  if (pims->verbose) fprintf (pims->fh_data, "IM command %s\n", status ? "succeeded" : "failed");

  return status;
}


static inline void StripTrailNewline (char ** Cmd, ssize_t *CmdLen)
// Strip trailing \n, then \r
{
  if (*(*Cmd + *CmdLen-1) == '\n') {
    *(*Cmd + *CmdLen - 1) = '\0';
    (*CmdLen)--;
  }
  if (*(*Cmd + *CmdLen-1) == '\r') {
    *(*Cmd + *CmdLen - 1) = '\0';
    (*CmdLen)--;
  }
}


static MagickBooleanType GetCommand (imservT * pims, char ** Cmd, ssize_t *CmdLen)
// Allocates Cmd; needs to be freed somewhere.
// Returns whether okay.
// Return MagickFalse if either oom, or file eof.
{
  MagickBooleanType okay = MagickTrue;
  *Cmd = NULL;
  *CmdLen = 0;
  size_t max_line_len = 0;
  ssize_t inlen = getline (Cmd, &max_line_len, stdin);

  if (inlen < 0) {
    fprintf (pims->fh_data, "getline failed; perhaps eof\n");
    if (pims->exitOnFileEof) okay = MagickFalse;
  }

  StripTrailNewline (Cmd, &inlen);

  // While inbuf ends with line-continuation character,
  // get another line and append them.

  #define cContinueLine '\\'

  char * p = *Cmd + inlen - 1;
  while (*p == cContinueLine) {
    char * NextLine = NULL;
    size_t NextLineLen = 0;
    ssize_t NewLen = getline (&NextLine, &NextLineLen, stdin);

    if (NewLen < 0) {
      fprintf (pims->fh_data, "next getline failed; perhaps EOF\n");
      if (pims->exitOnFileEof) okay = MagickFalse;
      break;
    }

    StripTrailNewline (&NextLine, &NewLen);

    inlen += NewLen - 1;
    *Cmd = realloc (*Cmd, (size_t)(inlen + 2));
    if (! *Cmd) {
      fprintf (pims->fh_data, "realloc failed\n");
      okay = MagickFalse;
      break;
    }
    memcpy (p, NextLine, (size_t)(NewLen+1));
    free (NextLine);
    p = *Cmd + inlen - 1;
  }

  *CmdLen = (okay) ? inlen : 0;

  return okay;
}


static MagickBooleanType ProcessCommands (imservT * pims)
// Returns only when file eof, or EOF received, or error.
{
  MagickBooleanType okay = MagickTrue;

  char ** argv;
  int argc;

  char * inbuf = NULL;
  ssize_t inlen;


  while (okay) {
    okay = GetCommand (pims, &inbuf, &inlen);
    if (!okay) break;

    if (*pims->sTerminate && (strcmp (inbuf, pims->sTerminate)==0)) {
      if (pims->verbose) fprintf (pims->fh_data, "Terminate received\n");
      okay = MagickFalse;
      break;
    }

    if (pims->verbose > 1) fprintf (pims->fh_data, "inlen=%li [%s]\n", inlen, inbuf);

    if (inlen) {
      argv = StringToArgv (inbuf, &argc);
      if (!argv) {
        fprintf (pims->fh_data, "StringToArgv failed\n");
        okay = MagickFalse;
        break;
      }

      MagickBooleanType ImOkay = ExecIm (pims, argc, argv);

      if (pims->exitOnImErr && !ImOkay) okay = MagickFalse;

      ssize_t j;
      for (j=0; j < (ssize_t) argc; j++) {
        if (pims->verbose > 1) fprintf (pims->fh_data, "  [%s]", argv[j]);
        argv[j] = DestroyString(argv[j]);
      }
      if (pims->verbose > 1) fprintf (pims->fh_data, "\n");

      argv=(char **) RelinquishMagickMemory (argv);
    }

    if (inbuf) free (inbuf);
    inbuf = NULL;
  }

  if (inbuf) free (inbuf);

  return okay;
}


int main (int argc, const char **argv)
{
  imservT ims;

  MagickBooleanType status = menu (argc, argv, &ims);
  if (status == MagickFalse) return (-1);
  if (ims.exitNow) return 0;

  MagickWandGenesis ();

  ims.exception = AcquireExceptionInfo ();
  ims.image_info = AcquireImageInfo ();

  MagickBooleanType okay = ProcessCommands (&ims);

  ims.image_info = DestroyImageInfo (ims.image_info);
  ims.exception = DestroyExceptionInfo (ims.exception);

  MagickWandTerminus ();

  return (okay ? 0 : 1);
}