/*
    Reference:
      http://im.snibgo.com/customim.htm
      http://im.snibgo.com/integim.htm

    Latest update:
      24-July-2017
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"

/*
   Latest update:
*/

typedef enum {
  aaAuto,
  aaDisregard,
  aaRegard
} alphaActionT;

typedef struct {
  FILE *
    fh_data;

  MagickBooleanType
    do_regardalpha,
    do_verbose;

  alphaActionT
    alphaAction;

} integimT;

static void usage (void)
{
  printf ("Usage: -process 'integim [OPTION]...'\n");
  printf ("Create an HDRI integral image (summed area table).\n");
  printf ("\n");
  printf ("  pm, premult         'yes' or 'no' or 'auto'\n");
  printf ("  f, file string      Write verbose text to stderr or stdout\n");
  printf ("  v, verbose          write text information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  integimT * pch
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  pch->fh_data = stderr;
  pch->do_regardalpha =
    pch->do_verbose = MagickFalse;
  pch->alphaAction = aaRegard;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "pm", "premult")==MagickTrue) {
      i++;
      if (strcasecmp (argv[i], "auto")==0) pch->alphaAction = aaAuto;
      else if (strcasecmp (argv[i], "yes")==0) pch->alphaAction = aaRegard;
      else if (strcasecmp (argv[i], "no")==0) pch->alphaAction = aaDisregard;
      else status = MagickFalse;
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      i++;
      if (strcasecmp (argv[i], "stdout")==0) pch->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) pch->fh_data = stderr;
      else status = MagickFalse;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pch->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "integim: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (pch->do_verbose) {
    fprintf (stderr, "integim options:");
//    if (pch->do_regardalpha) fprintf (stderr, "  regardalpha");
    if (pch->alphaAction != aaAuto) {
      fprintf (stderr, "  premult ");
      if (pch->alphaAction==aaRegard) fprintf (stderr, "yes");
      else if (pch->alphaAction==aaDisregard) fprintf (stderr, "no");
      else fprintf (stderr, "??");
    }
    if (pch->fh_data == stdout) fprintf (stderr, "  file stdout");
    if (pch->do_verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


#if defined(MAGICKCORE_HDRI_SUPPORT)
#define ADD_HALF
#else
#define ADD_HALF +0.5
#endif


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image * integim (
  const Image *image,
  integimT * pch,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  CacheView
    *image_view,
    *out_view;

  ssize_t
    y,
    inXmult,
    outXmult;

  MagickBooleanType
    IsAlphaOn = IS_ALPHA_CH(image),
    status;

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (pch->do_verbose) {
    fprintf (pch->fh_data,
             "integim: Input image [%s] depth is %i IsAlpha %s\n",
             image->filename,
             (int)image->depth,
             IsAlphaOn ? "ON":"OFF"
            );
  }

  switch (pch->alphaAction) {
    case aaAuto:
      pch->do_regardalpha = IsAlphaOn;
      if (pch->do_verbose)
        fprintf (pch->fh_data, "integim: alphaaction auto: regardalpha %s\n",
          pch->do_regardalpha ? "TRUE" : "FALSE");
      break;
    case aaRegard:
      pch->do_regardalpha = MagickTrue;
      break;
    case aaDisregard:
      pch->do_regardalpha = MagickFalse;
      break;
  }

  MagickBooleanType UseAlpha = (IsAlphaOn && pch->do_regardalpha);

  // FIXME: Can we do this in-place?

  new_image=CloneImage(image, image->columns, image->rows, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  // Just in case image had a palette.
  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  inXmult = Inc_ViewPixPtr (image);
  outXmult = Inc_ViewPixPtr (new_image);

  status=MagickTrue;

  out_view=AcquireAuthenticCacheView(new_image,exception);

  image_view=AcquireVirtualCacheView(image,exception);

  // We can't easily parallelise
  // because each output row needs the previous row.

  for (y=0; y < (ssize_t) image->rows; y++)
  {
    const VIEW_PIX_PTR
      *p,
      *q_prev = NULL;

    VIEW_PIX_PTR
      *q_out;

    ssize_t
      x;

    long double
      alpha;

    if (status==MagickFalse) continue;

    p=GetCacheViewVirtualPixels(image_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL)
    {
      fprintf (stderr, "bad GetCacheViewVirtualPixels image_view\n");
      status=MagickFalse;
    }

    q_out=GetCacheViewAuthenticPixels(out_view,0,y,new_image->columns,1,exception);
    if (q_out == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "integim: bad GetCacheViewAuthenticPixels out_view\n");
      status=MagickFalse;
    }

    if (y > 0) {
      q_prev=GetCacheViewVirtualPixels(out_view,0,y-1,new_image->columns,1,exception);
      if (q_prev == (const VIEW_PIX_PTR *) NULL) {
        fprintf (stderr, "integim: bad GetCacheViewAuthenticPixels q_prev\n");
        status=MagickFalse;
      }
    }

    alpha = 1.0;

    if (status==MagickFalse) continue;

    long double
      left_r = 0, up_r = 0, upleft_r = 0,
      left_g = 0, up_g = 0, upleft_g = 0,
      left_b = 0, up_b = 0, upleft_b = 0,
      left_a = 0, up_a = 0, upleft_a = 0;

    for (x=0; x < (ssize_t) image->columns; x++)
    {

      if (x > 0) {
        left_r = GET_PIXEL_RED(new_image, q_out-outXmult);
        left_g = GET_PIXEL_GREEN(new_image, q_out-outXmult);
        left_b = GET_PIXEL_BLUE(new_image, q_out-outXmult);
        left_a = GET_PIXEL_ALPHA(new_image, q_out-outXmult);
      }

      if (y > 0) {
        up_r = GET_PIXEL_RED(new_image, q_prev);
        up_g = GET_PIXEL_GREEN(new_image, q_prev);
        up_b = GET_PIXEL_BLUE(new_image, q_prev);
        up_a = GET_PIXEL_ALPHA(new_image, q_prev);
        if (x > 0) {
          upleft_r = GET_PIXEL_RED(new_image, q_prev-outXmult);
          upleft_g = GET_PIXEL_GREEN(new_image, q_prev-outXmult);
          upleft_b = GET_PIXEL_BLUE(new_image, q_prev-outXmult);
          upleft_a = GET_PIXEL_ALPHA(new_image, q_prev-outXmult);
        }
      }

      if (UseAlpha) {
        alpha = GET_PIXEL_ALPHA(image,p) / (long double)QuantumRange;
      }

      SET_PIXEL_RED (new_image,
        alpha * GET_PIXEL_RED(image,p) + left_r + up_r - upleft_r ADD_HALF,
        q_out);

      SET_PIXEL_GREEN (new_image,
        alpha * GET_PIXEL_GREEN(image,p) + left_g + up_g - upleft_g ADD_HALF,
        q_out);

      SET_PIXEL_BLUE (new_image,
        alpha * GET_PIXEL_BLUE(image,p) + left_b + up_b - upleft_b ADD_HALF,
        q_out);

      if (IsAlphaOn) {
        SET_PIXEL_ALPHA (new_image,
          GET_PIXEL_ALPHA(image,p) + left_a + up_a - upleft_a ADD_HALF,
          q_out);
      }

      p += inXmult;
      q_out += outXmult;
      q_prev += outXmult;
    }

    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }
  }

  image_view=DestroyCacheView(image_view);
  out_view=DestroyCacheView(out_view);

  return (new_image);
}


ModuleExport size_t integimImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  integimT
    cumul_histo;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &cumul_histo);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = integim (image, &cumul_histo, exception);
    if (!new_image) return(-1);

    ReplaceImageInList(&image,new_image);
    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
