/* Updated:
     3-April-2018 for v7.0.7-28
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"


typedef enum {
  mtAbsolute,
  mtRelative
} MapTypeT;

typedef struct {
  MagickBooleanType
    do_verbose;
  MapTypeT
    MapType;
} invDispMapT;

static void usage (void)
{
  printf ("Usage: -process 'invdispmap [OPTION]...'\n");
  printf ("Invert a displacement map.\n");
  printf ("\n");
  printf ("  t, type STRING      STRING is Relative or Absolute\n");
  printf ("  v, verbose          write text information to stdout\n");

  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  invDispMapT * idm
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  idm->do_verbose = MagickFalse;
  idm->MapType = mtAbsolute;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "t", "type")==MagickTrue) {
      i++;
      char * pa = (char *)argv[i];
      if (LocaleCompare(pa, "Absolute")==0) idm->MapType = mtAbsolute;
      else if (LocaleCompare(pa, "Relative")==0) idm->MapType = mtRelative;
      else {
        fprintf (stderr, "invdispmap: ERROR: unknown map type\n");
        status = MagickFalse;
      }

    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      idm->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "invdispmap: ERROR: unknown option\n");
      status = MagickFalse;
    }
  }

  if (idm->do_verbose) {
    fprintf (stderr, "invdispmap options:  ");
    fprintf (stderr, "type ");
    fprintf (stderr, idm->MapType==mtAbsolute?"Absolute ":"Relative ");

    if (idm->do_verbose) fprintf (stderr, "verbose ");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


#if defined(MAGICKCORE_HDRI_SUPPORT)
#define ADD_HALF
#else
#define ADD_HALF +0.5
#endif


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *invDispMap(const Image *image,
  invDispMapT * idm,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  CacheView
    *image_view,
    *out_view;

  ssize_t
    y;

  MagickBooleanType
    status;

  PIX_INFO
    mppBlack;

  long double
    QoverCols,
    QoverRows,
    relHalf;

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (idm->do_verbose) {
    fprintf (stderr, "invdispmap: Input image [%s] depth is %i\n",
             image->filename, (int)image->depth);
  }

  new_image=CloneImage(image, 0, 0, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  // Just in case image had a palette.
  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

#if IMV6OR7==6
  GetMagickPixelPacket (new_image, &mppBlack);
  mppBlack.matte=MagickTrue;
  mppBlack.opacity=(MagickRealType) TransparentOpacity;
  SetImageColor (new_image, &mppBlack);
#else
  GetPixelInfo (new_image, &mppBlack);
  mppBlack.alpha = 0;
  SetImageColor (new_image, &mppBlack, exception);
#endif

  status=MagickTrue;

  image_view=AcquireVirtualCacheView(image,exception);

  out_view=AcquireAuthenticCacheView(new_image,exception);

  QoverCols = QuantumRange / (long double)(image->columns-1);
  QoverRows = QuantumRange / (long double)(image->rows-1);

  if (idm->MapType == mtRelative) {
    relHalf = QuantumRange / 2.0;
  } else {
    relHalf = 0.0;
  }

  // Unlike most processes,
  // this reads pixels sequentially
  // and writes them randomly.

// Two threads might want to update the same row,
// so don't allow multiple threads.
// #if defined(MAGICKCORE_OPENMP_SUPPORT)
//  #pragma omp parallel for schedule(static,4) shared(status)
//    MAGICK_THREADS(image,image,image->rows,1)
// #endif

  for (y=0; y < (ssize_t) image->rows; y++)
  {
    const VIEW_PIX_PTR
      *p;

    VIEW_PIX_PTR
      *q_out;

    ssize_t
      x, red, grn, xa, ya;

    long double
      x_norm, y_norm;

    if (status == MagickFalse)
      continue;
    p=GetCacheViewVirtualPixels(image_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL)
    {
      fprintf (stderr, "bad GetCacheViewVirtualPixels image_view\n");
      status=MagickFalse;
      continue;
    }

    for (x=0; x < (ssize_t) image->columns; x++)
    {
      if (GET_PIXEL_ALPHA (image, p) > 0) {
        red = floor((GET_PIXEL_RED (image, p)   - relHalf) / QoverCols + 0.5);
        grn = floor((GET_PIXEL_GREEN (image, p) - relHalf) / QoverRows + 0.5);

        if (idm->MapType == mtRelative) {
          red += x;
          grn += y;
          xa = x - red;
          ya = y - grn;
        } else {
          xa = x;
          ya = y;
        }

        if (red >= 0 && red < image->columns
         && grn >= 0 && grn < image->rows)
        {
          q_out=GetCacheViewAuthenticPixels(out_view,red,grn,1,1,exception);
          if (q_out == (const VIEW_PIX_PTR *) NULL) {
            fprintf (stderr, "invdispmap: bad GetCacheViewAuthenticPixels out_view (%lu %lu)\n", red, grn);
            status=MagickFalse;
          }

          x_norm  = xa * QoverCols + relHalf ADD_HALF;
          y_norm  = ya * QoverRows + relHalf ADD_HALF;

          SET_PIXEL_RED   (new_image, x_norm,       q_out);
          SET_PIXEL_GREEN (new_image, y_norm,       q_out);
          SET_PIXEL_BLUE  (new_image, 0,            q_out);
          SET_PIXEL_ALPHA (new_image, QuantumRange, q_out);

          if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse) {
            fprintf (stderr, "bad sync\n");
            status=MagickFalse;
          }
        } else {
          fprintf (stderr, "ignored: red=%li grn=%li\n", red, grn);
        }
      }

      p += Inc_ViewPixPtr (image);
    }
  }

  image_view=DestroyCacheView(image_view);
  out_view=DestroyCacheView(out_view);

  return (new_image);
}


ModuleExport size_t invdispmapImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  invDispMapT
    idm;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &idm);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = invDispMap (image, &idm, exception);

    ReplaceImageInList(&image,new_image);
    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
