/*
   Reference: http://im.snibgo.com/kcluster.htm

   Last update: 9-August-2017.
     18-August-2017 added SetNoPalette ()
     21-August-2017 removed 'j' option; 'k' can now be range.
     7-September-2017 for v7.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include "vsn_defines.h"

#define VERSION "kcluster v1.0  Copyright (c) 2017 Alan Gibson"

typedef enum {
  mNull,
  mMeans,
  mFuzzy,
  mHarmonic
} methodT;

typedef enum {
  iColors,
  iForgy,
  iRandPart,
  iSdLine,
  iFromList
} initT;

typedef struct {
  double r;
  double g;
  double b;
} normaliseT;

typedef struct {
  Quantum red;
  Quantum green;
  Quantum blue;
  Quantum alpha;
  double  count;
} clusterT;

typedef struct {
  int
    verbose;

  methodT
    method;

  initT
    init;

  int
    precision,
    jump0,
    jump1,
    kCols,
    maxIter;

  MagickBooleanType
    hasJump,
    hasSetY,
    regardAlpha,
    sdNorm,
    dither,
    debug;

  normaliseT
    normalise;

  double
    fuzzyR,
    harmonicP,
    tolerance,
    jumpY;

  double
    powParam1,
    powParam2,
    powParam3;

  FILE *
    fh_data;

  RandomInfo
    *random_info;

  ssize_t
    *nearClust;

  clusterT
    *clusters;

  char
    *frameName;

  int
    frameNum;

  double
    *membDenoms;
} kclusterT;



static void usage (void)
{
  printf ("Usage: -process 'kcluster [OPTION]...'\n");
  printf ("Apply a k-clustering method.\n");
  printf ("\n");
  printf ("  m, method string        'Null', 'Means', 'Fuzzy' or 'Harmonic'\n");
  printf ("  k, kCols int[-int]      number [range] of clusters to find\n");
//  printf ("  j, jump integer,integer find k\n");
  printf ("  s, sdNorm               normalise channel SD\n");
  printf ("  o, dither               dither output\n");
  printf ("  a, regardAlpha          process opacity\n");
  printf ("  i, initialize string    'Colors', 'Forgy', 'RandPart', 'SdLine' or 'FromList'\n");
  printf ("  x, maxIter integer      maximum number of iterations\n");
  printf ("  y, jumpY number         y for jump power\n");
  printf ("  r, fuzzyR number        r for fuzzy method\n");
  printf ("  p, harmonicP number     p for harmonic method\n");
  printf ("  t, tolerance number     tolerance\n");
  printf ("  w, write filename       write iterations eg frame_%%06d.png\n");

  printf ("  d, debug                write debugging information to stderr\n");
  printf ("  f, file string          write verbose to file stream stdout or stderr\n");
  printf ("  v, verbose              write text information\n");
  printf ("  v2, verbose2            write more text information\n");
  printf ("     version              write version information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  // LocaleCompare is not case-sensitive,
  // so we use strcmp for the short option.

  if ((strcmp(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static void InitNormalise (
  kclusterT * pkc
)
{
  pkc->normalise.r = 1.0;
  pkc->normalise.g = 1.0;
  pkc->normalise.b = 1.0;
}


static MagickBooleanType GetInteger (
  char ** p,
  int    *pValue)
// Returns whether valid.
{
  int len;
  if (sscanf (*p, "%i%n", pValue, &len) != 1) return MagickFalse;
  *p += len;
  char c = **p;
  switch (c) {
    case '\0':
    case ',':
      /* nothing */ ;
      break;
    default:
      printf ("Bad terminator? [%c]\n", c);
  }
  return MagickTrue;
}

static MagickBooleanType ParseTwoJumps (
  kclusterT * pkc,
  char * p
)
// Expected format: A,B
// where A and B are numbers.
// Returns whether okay.
{
  int v0=0, v1=0;

  //char * p = pch->sWindow;

  if (!*p) return MagickFalse;

  if (!GetInteger (&p, &v0)) {
    fprintf (stderr, "kcluster error: needs two integers.\n");
    return MagickFalse;
  }

  if (*p!=',') {
    fprintf (stderr, "kcluster error: needs 'x' betweem two integers.\n");
    return MagickFalse;
  }

  p++;
  if (!GetInteger (&p, &v1)) {
    fprintf (stderr, "kcluster error: needs two integers.\n");
    return MagickFalse;
  }

  if (v0 < 1 || v1 < v0) {
    fprintf (stderr, "kcluster error: integers must be >=1, ascending.\n");
    return MagickFalse;
  }

  if (*p) {
    fprintf (stderr, "kcluster error: extra characters [%s].\n", p);
    return MagickFalse;
  }

  pkc->jump0 = v0;
  pkc->jump1 = v1;

  return MagickTrue;
}


static MagickBooleanType ParseRange (
  kclusterT * pkc,
  char * p,
  int * valA,
  int * valB,
  int * numVals
)
// Expected format: A or A-B where A and B are integers.
// Returns whether okay.
{
  *numVals = 0;
  int len;
  if (sscanf (p, "%i%n", valA, &len) != 1) return MagickFalse;
  p += len;
  *numVals = 1;

  if (*p == '-') {
    p++;

    if (sscanf (p, "%i%n", valB, &len) != 1) return MagickFalse;
    p += len;
    *numVals = 2;
  }
  if (*p) {
    fprintf (stderr, "kcluster ParseRange error: extra characters [%s].\n", p);
    return MagickFalse;
  }
  return MagickTrue;
}




#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "kcluster: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  kclusterT * pkc
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status = MagickTrue;

  char ** pargv = (char **)argv;

  pkc->verbose = 0;
  pkc->init = iColors;
  pkc->regardAlpha = MagickFalse;
  pkc->debug = MagickFalse;
  pkc->sdNorm = MagickFalse;
  pkc->dither = MagickFalse;
  pkc->method = mMeans;
  pkc->hasJump = MagickFalse;
  pkc->hasSetY = MagickFalse;
  pkc->jumpY = 0;
  pkc->jump0 = 1;
  pkc->jump1 = 10;
  pkc->kCols = 10;
  pkc->maxIter = 100;
  pkc->fuzzyR = 1.5;
  pkc->harmonicP = 3.5;
  pkc->tolerance = 0.01;
  pkc->fh_data = stderr;
  pkc->random_info = NULL;
  pkc->frameName = NULL;
  pkc->frameNum = 0;
  pkc->clusters = NULL;
  pkc->membDenoms = NULL;
  InitNormalise (pkc);

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "m", "method")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "means")==0) pkc->method = mMeans;
      else if (LocaleCompare(argv[i], "fuzzy")==0) pkc->method = mFuzzy;
      else if (LocaleCompare(argv[i], "harmonic")==0) pkc->method = mHarmonic;
      else if (LocaleCompare(argv[i], "null")==0) pkc->method = mNull;
      else {
        fprintf (stderr, "kcluster: ERROR: unknown method [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "k", "kCols")==MagickTrue) {
      NEXTARG;
      pkc->hasJump = MagickFalse;
      pkc->kCols = atoi(pargv[i]);

      int valA, valB, numVals;
      if (ParseRange (pkc, pargv[i], &valA, &valB, &numVals)) {
        if (numVals==1) {
          pkc->hasJump = MagickFalse;
          pkc->kCols = valA;
        } else if (numVals==2) {
          pkc->hasJump = MagickTrue;
          pkc->jump0 = valA;
          pkc->jump1 = valB;
        } else {
          fprintf (stderr, "kcluster: ERROR: bad kCols [%s]\n", argv[i]);
          status = MagickFalse;
        }
      } else {
        fprintf (stderr, "kcluster: ERROR: bad kCols [%s]\n", argv[i]);
        status = MagickFalse;
      }

    } else if (IsArg (pa, "j", "jump")==MagickTrue) {
      NEXTARG;
      pkc->hasJump = MagickTrue;
      if (!ParseTwoJumps (pkc, pargv[i])) status = MagickFalse;
    } else if (IsArg (pa, "i", "initialize")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "colors")==0) pkc->init = iColors;
      else if (LocaleCompare(argv[i], "colours")==0) pkc->init = iColors;
      else if (LocaleCompare(argv[i], "forgy")==0) pkc->init = iForgy;
      else if (LocaleCompare(argv[i], "randPart")==0) pkc->init = iRandPart;
      else if (LocaleCompare(argv[i], "sdLine")==0) pkc->init = iSdLine;
      else if (LocaleCompare(argv[i], "fromList")==0) pkc->init = iFromList;
      else {
        fprintf (stderr, "kcluster: ERROR: unknown initialize [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "x", "maxIter")==MagickTrue) {
      NEXTARG;
      pkc->maxIter = atoi(pargv[i]);
    } else if (IsArg (pa, "y", "jumpY")==MagickTrue) {
      NEXTARG;
      pkc->hasSetY = MagickTrue;
      pkc->jumpY = atof(pargv[i]);
    } else if (IsArg (pa, "r", "fuzzyR")==MagickTrue) {
      NEXTARG;
      pkc->fuzzyR = atof(pargv[i]);
    } else if (IsArg (pa, "p", "harmonicP")==MagickTrue) {
      NEXTARG;
      pkc->harmonicP = atof(pargv[i]);
    } else if (IsArg (pa, "t", "tolerance")==MagickTrue) {
      NEXTARG;
      pkc->tolerance = atof(pargv[i]);
    } else if (IsArg (pa, "s", "sdNorm")==MagickTrue) {
      pkc->sdNorm = MagickTrue;
    } else if (IsArg (pa, "o", "dither")==MagickTrue) {
      pkc->dither = MagickTrue;
    } else if (IsArg (pa, "a", "regardAlpha")==MagickTrue) {
      pkc->regardAlpha = MagickTrue;
    } else if (IsArg (pa, "w", "write")==MagickTrue) {
      NEXTARG;
      pkc->frameName = pargv[i];
    } else if (IsArg (pa, "d", "debug")==MagickTrue) {
      pkc->debug = MagickTrue;
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      NEXTARG;
      if (strcasecmp (argv[i], "stdout")==0) pkc->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) pkc->fh_data = stderr;
      else status = MagickFalse;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pkc->verbose = 1;
    } else if (IsArg (pa, "v2", "verbose2")==MagickTrue) {
      pkc->verbose = 2;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "kcluster: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (pkc->dither && pkc->method != mFuzzy) {
    fprintf (stderr, "kcluster: WARNING dither is available only for fuzzy.");
    pkc->dither = MagickFalse;
  }

  if (pkc->verbose) {
    fprintf (stderr, "kcluster options: ");

    fprintf (stderr, "  method ");
    switch (pkc->method) {
      case mNull:
        fprintf (stderr, "Null");
        break;
      case mMeans:
        fprintf (stderr, "Means");
        break;
      case mFuzzy:
        fprintf (stderr, "Fuzzy");
        fprintf (stderr, "  fuzzyR %.*g", pkc->precision, pkc->fuzzyR);
        break;
      case mHarmonic:
        fprintf (stderr, "Harmonic");
        fprintf (stderr, "  harmonicP %.*g", pkc->precision, pkc->harmonicP);
        break;
    }
    if (pkc->hasJump) {
      fprintf (stderr, "  kCols %i-%i", pkc->jump0, pkc->jump1);
      if (pkc->hasSetY) fprintf (stderr, "  jumpY %.*g", pkc->precision, pkc->jumpY);
    }

    if (pkc->init != iFromList && !pkc->hasJump)
      fprintf (stderr, "  kCols %i", pkc->kCols);

    fprintf (stderr, "  initialize ");
    switch (pkc->init) {
      case iColors:
        fprintf (stderr, "Colors");
        break;
      case iForgy:
        fprintf (stderr, "Forgy");
        break;
      case iRandPart:
        fprintf (stderr, "RandPart");
        break;
      case iSdLine:
        fprintf (stderr, "SdLine");
        break;
      case iFromList:
        fprintf (stderr, "FromList");
        break;
    }
    fprintf (stderr, "  tolerance %.*g", pkc->precision, pkc->tolerance);
    fprintf (stderr, "  maxIter %i", pkc->maxIter);

    if (pkc->sdNorm)          fprintf (stderr, "  sdNorm");
    if (pkc->dither)          fprintf (stderr, "  dither");
    if (pkc->regardAlpha)     fprintf (stderr, "  regardAlpha");

    if (pkc->frameName)       fprintf (stderr, "  write %s", pkc->frameName);

    if (pkc->debug)           fprintf (stderr, "  debug");
    if (pkc->fh_data == stdout) fprintf (stderr, " file stdout");
    if (pkc->verbose==1)      fprintf (stderr, "  verbose");
    else if (pkc->verbose==2) fprintf (stderr, "  verbose2");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}




static void InitRand (kclusterT * pkc)
{
  pkc->random_info=AcquireRandomInfo();

  // There seems to be a problem: the first few values show coherency,
  // so skip over them.

  int i;
  for (i=0; i < 20; i++) {
    GetPseudoRandomValue(pkc->random_info);
  }
}

static void DeInitRand (kclusterT * pkc)
{
  if (pkc->random_info)
    pkc->random_info=DestroyRandomInfo(pkc->random_info);
}

static void DumpImage (
  kclusterT * pkc,
  Image *image,
  ExceptionInfo *exception)
{
  CacheView *in_view = AcquireVirtualCacheView (image, exception);

  ssize_t x, y;
  for (y=0; y < (ssize_t) image->rows; y++) {
    const VIEW_PIX_PTR *p=GetCacheViewVirtualPixels(in_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL) return;

    for (x=0; x < (ssize_t) image->columns; x++) {
      fprintf (pkc->fh_data, "%li,%li: %g %g %g  %g\n",
        x, y,
        GET_PIXEL_RED(image,p),
        GET_PIXEL_GREEN(image,p),
        GET_PIXEL_BLUE(image,p),
        GET_PIXEL_ALPHA(image,p));

      p += Inc_ViewPixPtr (image);
    }
  }

  in_view = DestroyCacheView (in_view);
}

static void getRandomPixel (
  kclusterT * pkc,
  Image *image,
  Image *forgy_img,
  CacheView *in_view,
  VIEW_PIX_PTR * pf,  // points to pixel in forgy_img
  ExceptionInfo *exception
)
{
  // If we are regarding alpha,
  // try to get non-transparent colour.

  int i;
  for (i=0; i < 10; i++) {
    ssize_t x = image->columns * GetPseudoRandomValue(pkc->random_info);
    ssize_t y = image->rows * GetPseudoRandomValue(pkc->random_info);
    const VIEW_PIX_PTR *pi = GetCacheViewVirtualPixels(in_view,x,y,1,1,exception);

    SET_PIXEL_RED   (forgy_img, GET_PIXEL_RED(image,pi), pf);
    SET_PIXEL_GREEN (forgy_img, GET_PIXEL_GREEN(image,pi), pf);
    SET_PIXEL_BLUE  (forgy_img, GET_PIXEL_BLUE(image,pi), pf);
    double a = GET_PIXEL_ALPHA(image,pi);
    if (a < 0) a = 0;
    SET_PIXEL_ALPHA (forgy_img, a, pf);

    if (!pkc->regardAlpha) break;
    if (a > 0) break;
  }
}

static Image * InitialiseForgy (
  kclusterT * pkc,
  Image *image,
  ExceptionInfo *exception
)
{
  if (!pkc->random_info) InitRand (pkc);

  Image *forgy_img = CloneImage (image, pkc->kCols,1, MagickFalse, exception);
  if (!forgy_img) return NULL;

  CacheView *in_view = AcquireVirtualCacheView (image, exception);
  CacheView *forgy_view = AcquireAuthenticCacheView (forgy_img, exception);

  VIEW_PIX_PTR * pf = GetCacheViewAuthenticPixels (
    forgy_view,0,0,forgy_img->columns,1,exception);

  int i;
  for (i=0; i < pkc->kCols; i++) {
    getRandomPixel (pkc, image, forgy_img, in_view, pf, exception);

    pf += Inc_ViewPixPtr (forgy_img);
  }
  if (!SyncCacheViewAuthenticPixels(forgy_view,exception))
    return NULL;

  forgy_view = DestroyCacheView (forgy_view);
  in_view    = DestroyCacheView (in_view);

  // FIXME: then ensure they are unique.

  return forgy_img;
}

static Image * MakeUnique (
  kclusterT * pkc,
  Image *image,
  Image *sml_image,
  ExceptionInfo *exception
)
// Returns image of unique colours, or NULL if failure.
{
  Image *uniq_img = UniqueImageColors (sml_image, exception);
  if (!uniq_img) return NULL;

  if (pkc->debug) {
    fprintf (stderr, "AfterUnique:\n");
    DumpImage (pkc, uniq_img, exception);
  }

  if (uniq_img->columns < pkc->kCols) {
    if (pkc->verbose) {
      fprintf (pkc->fh_data, "kcluster: wanted %i but found %li unique\n",
        pkc->kCols, uniq_img->columns);
    }
    ssize_t skipOver = uniq_img->columns;

    uniq_img->gravity = WestGravity;
    // Background?
#if IMV6OR7==6
    SetImageExtent (uniq_img, pkc->kCols, 1);
#else
    SetImageExtent (uniq_img, pkc->kCols, 1, exception);
#endif
    if (pkc->verbose) {
      fprintf (pkc->fh_data, "kcluster: -- now have %li\n", uniq_img->columns);
    }

    if (!pkc->random_info) InitRand (pkc);

    CacheView *in_view = AcquireVirtualCacheView (image, exception);
    CacheView *uniq_view = AcquireAuthenticCacheView (uniq_img, exception);

    VIEW_PIX_PTR * pu = GetCacheViewAuthenticPixels (
      uniq_view, 0, 0, uniq_img->columns, 1, exception);
    pu += skipOver * Inc_ViewPixPtr (uniq_img);
    int i;
    for (i = skipOver; i < uniq_img->columns; i++) {
      getRandomPixel (pkc, image, uniq_img, in_view, pu, exception);
      pu += Inc_ViewPixPtr (uniq_img);
    }
    if (!SyncCacheViewAuthenticPixels(uniq_view,exception))
      return NULL;
    uniq_view = DestroyCacheView (uniq_view);
    in_view = DestroyCacheView (in_view);
  }

  if (pkc->debug) {
    fprintf (stderr, "After padding:\n");
    DumpImage (pkc, uniq_img, exception);
  }

  if (pkc->regardAlpha) {
    // If we have caught any fully-transparent colours,
    // try for better ones.

    if (!pkc->random_info) InitRand (pkc);

    CacheView *in_view = AcquireVirtualCacheView (image, exception);
    CacheView *uniq_view = AcquireAuthenticCacheView (uniq_img, exception);

    VIEW_PIX_PTR * pu = GetCacheViewAuthenticPixels(uniq_view,0,0,uniq_img->columns,1,exception);
    int i;
    for (i=0; i < uniq_img->columns; i++) {
      if (GET_PIXEL_ALPHA(uniq_img,pu) <= 0) {
        getRandomPixel (pkc, image, uniq_img, in_view, pu, exception);
      }
      pu += Inc_ViewPixPtr (uniq_img);
    }
    if (!SyncCacheViewAuthenticPixels(uniq_view,exception))
      return NULL;
    uniq_view = DestroyCacheView (uniq_view);
    in_view = DestroyCacheView (in_view);
  }
  return uniq_img;
}

static Image * InitialiseColors (
  kclusterT * pkc,
  Image *image,
  ExceptionInfo *exception
)
// Returns image of unique colours, or NULL if failure.
{
  Image *quant_img = CloneImage (image, 0,0, MagickFalse, exception);
  if (!quant_img) return NULL;

  QuantizeInfo *qi = AcquireQuantizeInfo (NULL);
  qi->number_colors = pkc->kCols;
  qi->dither_method = NoDitherMethod;

#if IMV6OR7==6
  qi->dither = MagickFalse;
  if (!QuantizeImage (qi, quant_img)) return NULL;
#else
  if (!QuantizeImage (qi, quant_img, exception)) return NULL;
#endif

  qi = DestroyQuantizeInfo (qi);

  Image * uniq_img = MakeUnique (pkc, image, quant_img, exception);

  Image * uniq_img2 = MakeUnique (pkc, image, uniq_img, exception);

  // FIXME: free images

  return uniq_img2;
}

static Image * InitialiseRandomPartition (
  kclusterT * pkc,
  Image *image,
  ExceptionInfo *exception
)
// Returns image of unique colours, or NULL if failure.
{
  // Assign each image pixel to a random partition.
  // Set each partition to the average of the pixels (each weighted by alpha).

  if (!pkc->random_info) InitRand (pkc);

  Image *rp_img = CloneImage (image, pkc->kCols,1, MagickFalse, exception);
  if (!rp_img) return NULL;

  if (!pkc->clusters) {
    fprintf (stderr, "kcluster: BUG: irp no clusters");
    return NULL;
  }

  CacheView *in_view = AcquireVirtualCacheView (image, exception);
  CacheView *rp_view = AcquireAuthenticCacheView (rp_img, exception);

  ssize_t x, y, qx;

  for (qx = 0; qx < rp_img->columns; qx++) {
    clusterT * pClust = &pkc->clusters[qx];
    pClust->red =
      pClust->green =
      pClust->blue =
      pClust->alpha =
      pClust->count = 0;
  }

  double r, g, b, a;

  for (y=0; y < (ssize_t) image->rows; y++) {
    const VIEW_PIX_PTR *p=GetCacheViewVirtualPixels(in_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL) return MagickFalse;

    for (x=0; x < (ssize_t) image->columns; x++) {
      a = GET_PIXEL_ALPHA(image,p) / (double)QuantumRange;
      r = a * GET_PIXEL_RED(image,p) / (double)QuantumRange;
      g = a * GET_PIXEL_GREEN(image,p) / (double)QuantumRange;
      b = a * GET_PIXEL_BLUE(image,p) / (double)QuantumRange;

      qx = rp_img->columns * GetPseudoRandomValue(pkc->random_info);

      clusterT * pClust = &pkc->clusters[qx];

      pClust->red += r;
      pClust->green += g;
      pClust->blue += b;
      pClust->alpha += a;
      pClust->count++;

      p += Inc_ViewPixPtr (image);
    }
  }

  // We may have clusters with count=0,
  //   no image pixels have been assigned to it.

  VIEW_PIX_PTR *rp=GetCacheViewAuthenticPixels(rp_view,0,0,rp_img->columns,1,exception);
  for (qx = 0; qx < rp_img->columns; qx++) {
    clusterT * pClust = &pkc->clusters[qx];

    if (pClust->count==0) {
      SET_PIXEL_ALPHA (rp_img, QuantumRange, rp);
      SET_PIXEL_RED (rp_img, 0, rp);
      SET_PIXEL_GREEN (rp_img, 0, rp);
      SET_PIXEL_BLUE (rp_img, 0, rp);
    } else {
      SET_PIXEL_ALPHA (rp_img, QuantumRange * pClust->alpha / pClust->count, rp);
      if (pClust->alpha == 0) pClust->alpha = 1.0;
      SET_PIXEL_RED   (rp_img, QuantumRange * pClust->red / pClust->alpha, rp);
      SET_PIXEL_GREEN (rp_img, QuantumRange * pClust->green / pClust->alpha, rp);
      SET_PIXEL_BLUE  (rp_img, QuantumRange * pClust->blue / pClust->alpha, rp);
    }
    rp += Inc_ViewPixPtr (rp_img);
  }

  if (!SyncCacheViewAuthenticPixels(rp_view,exception))
    return NULL;

  rp_view = DestroyCacheView (rp_view);
  in_view = DestroyCacheView (in_view);

  Image * uniq_img = MakeUnique (pkc, image, rp_img, exception);

  Image * uniq_img2 = MakeUnique (pkc, image, uniq_img, exception);

  // FIXME: free images

  return uniq_img2;
}



static MagickBooleanType CalcMeanSd (
  kclusterT * pkc,
  Image *image,
  double *mnR,
  double *mnG,
  double *mnB,
  double *sdR,
  double *sdG,
  double *sdB,
  MagickBooleanType *isMeanSdDefined,
  ExceptionInfo *exception
)
// Calculate mean and standard deviation,
// accounting for alpha (eg ignoring pixels that are entirely transparent).
// Returns values in range 0.0 to 1.0.
// Returns false if major problem.
// Returns isMeanSdDefined = MagickFalse iff statistics are not defined
//   (image is entirely transparent).
{
  // mnR = sig(R) / sig(alpha)
  // sdR = sqrt ( sig(R^2) / sig(alpha) - mnR^2 )
  //
  // Likewise for G and B.

  CacheView *in_view = AcquireVirtualCacheView (image, exception);

  double r, g, b, a;
  double sigR=0, sigG=0, sigB=0, sigA=0;
  double sigR2=0, sigG2=0, sigB2=0;

  ssize_t x, y;
  for (y=0; y < (ssize_t) image->rows; y++) {
    const VIEW_PIX_PTR *p=GetCacheViewVirtualPixels(in_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL) return MagickFalse;

    for (x=0; x < (ssize_t) image->columns; x++) {
      a = GET_PIXEL_ALPHA(image,p) / QuantumRange;
      r = a * GET_PIXEL_RED(image,p) / QuantumRange;
      g = a * GET_PIXEL_GREEN(image,p) / QuantumRange;
      b = a * GET_PIXEL_BLUE(image,p) / QuantumRange;
      sigR += r;
      sigG += g;
      sigB += b;
      sigA += a;
      sigR2 += r*r;
      sigG2 += g*g;
      sigB2 += b*b;

      p += Inc_ViewPixPtr (image);
    }
  }

  in_view = DestroyCacheView (in_view);

  if (sigA == 0) {
    *isMeanSdDefined = MagickFalse;
    return MagickTrue;
  }

  *isMeanSdDefined = MagickTrue;
  *mnR = sigR / sigA;
  *mnG = sigG / sigA;
  *mnB = sigB / sigA;
  *sdR = sqrt (sigR2/sigA - (*mnR)*(*mnR));
  *sdG = sqrt (sigG2/sigA - (*mnG)*(*mnG));
  *sdB = sqrt (sigB2/sigA - (*mnB)*(*mnB));

  return MagickTrue;
}

static Image * InitialiseSdLine (
  kclusterT * pkc,
  Image *image,
  ExceptionInfo *exception
)
// Returns image of unique colours, or NULL if failure.
{
  double mnR=0, mnG=0, mnB=0;
  double sdR=0, sdG=0, sdB=0;

  MagickBooleanType isMeanSdDefined;

  if (!CalcMeanSd (pkc, image,
          &mnR, &mnG, &mnB, &sdR, &sdG, &sdB,
          &isMeanSdDefined, exception))
    return MagickFalse;

  if (sdR==0 && sdG==0 && sdB==0) {
    fprintf (stderr, "kcluster: initialize SdLine with flat-colour image: only one cluster.");
    pkc->kCols = 1;
  }

  double loR = mnR - sdR;
  double loG = mnG - sdG;
  double loB = mnB - sdB;

  double semiR = sdR / (double)pkc->kCols;
  double semiG = sdG / (double)pkc->kCols;
  double semiB = sdB / (double)pkc->kCols;

  if (pkc->debug) fprintf (stderr,
    "InitialiseSdLine: mnR=%g sdR=%g loR=%g semiR=%g\n", 
    mnR, sdR, loR, semiR);

  Image *sd_img = CloneImage (image, pkc->kCols,1, MagickFalse, exception);
  if (!sd_img) return NULL;

  CacheView *sd_view = AcquireAuthenticCacheView (sd_img, exception);

  VIEW_PIX_PTR *q=GetCacheViewAuthenticPixels(sd_view,0,0,sd_img->columns,1,exception);

  int qx;
  for (qx = 0; qx < sd_img->columns; qx++) {
    SET_PIXEL_ALPHA (sd_img, QuantumRange, q);

    if (pkc->debug) fprintf (stderr,
      "InitialiseSdLine: qx=%i R=%g\n", 
      qx, loR + semiR*(2*qx+1));

    SET_PIXEL_RED   (sd_img, QuantumRange * (loR + semiR*(2*qx+1)), q);
    SET_PIXEL_GREEN (sd_img, QuantumRange * (loG + semiG*(2*qx+1)), q);
    SET_PIXEL_BLUE  (sd_img, QuantumRange * (loB + semiB*(2*qx+1)), q);

    q += Inc_ViewPixPtr (sd_img);
 }

  if (!SyncCacheViewAuthenticPixels(sd_view,exception))
    return NULL;

  sd_view = DestroyCacheView (sd_view);

  return sd_img;
}

static Image * InitialiseFromList (
  kclusterT * pkc,
  Image *image,
  ExceptionInfo *exception
)
// Returns image of unique colours, or NULL if failure.
{
  Image *last_img = GetLastImageInList(image);

  if (last_img == NULL || last_img == image) {
    fprintf (stderr, "kcluster: FromList: can't find last\n");
    return NULL;
  }

  if (pkc->verbose) {
    fprintf (pkc->fh_data, "kcluster: FromList [%s] %ix%i depth %i\n",
             last_img->filename,
             (int)last_img->columns, (int)last_img->rows,
             (int)last_img->depth);
  }

  Image *uniq_img = UniqueImageColors (last_img, exception);
  if (!uniq_img) return NULL;

  // FIXME: transparency?

  pkc->kCols = uniq_img->columns;

  if (pkc->verbose) fprintf (pkc->fh_data, "kcluster: FromList k=%i\n", pkc->kCols);

  return uniq_img;
}


static MagickBooleanType CalcNormalise (
  kclusterT * pkc,
  Image *image,
  ExceptionInfo *exception
)
{
  InitNormalise (pkc);

  double mnR=0, mnG=0, mnB=0;
  double sdR=0, sdG=0, sdB=0;

  MagickBooleanType isMeanSdDefined;

  if (!CalcMeanSd (pkc, image,
          &mnR, &mnG, &mnB, &sdR, &sdG, &sdB,
          &isMeanSdDefined, exception))
    return MagickFalse;

  if (pkc->debug) {
    fprintf (pkc->fh_data, "kcluster: mnsd defined: %i\n", isMeanSdDefined);

    fprintf (pkc->fh_data, "kcluster: mn: %.*g,%.*g,%.*g sd: %.*g,%.*g,%.*g\n",
      pkc->precision, mnR,
      pkc->precision, mnG,
      pkc->precision, mnB,
      pkc->precision, sdR,
      pkc->precision, sdG,
      pkc->precision, sdB);
  }

  if (!isMeanSdDefined) return MagickFalse;

  // FIXME: Can this also set Y for jump???

  double sdMin = sdR;
  if (sdMin > sdG) sdMin = sdG;
  if (sdMin > sdB) sdMin = sdB;

  double sdMax = sdR;
  if (sdMax < sdG) sdMax = sdG;
  if (sdMax < sdB) sdMax = sdB;

  if (sdMin==sdMax) return MagickTrue;

  if (sdR+sdG+sdB==sdMax) {
    if (pkc->verbose) fprintf (pkc->fh_data, "kcluster: only one channel\n");
    return MagickTrue;
  }

  if (sdMin==0) {
    if (pkc->verbose) fprintf (pkc->fh_data, "kcluster: only two channels\n");

    if (sdMax==sdR) {
      pkc->normalise.r = ((sdG==0) ? sdB : sdG) / sdR;
    } else if (sdMax==sdG) {
      pkc->normalise.g = ((sdR==0) ? sdB : sdR) / sdG;
    } else if (sdMax==sdB) {
      pkc->normalise.b = ((sdR==0) ? sdG : sdR) / sdB;
    }

  } else {
    if (pkc->verbose) fprintf (pkc->fh_data, "kcluster: three channels\n");

    if (sdMin==sdR) {
      pkc->normalise.g = sdR / sdG;
      pkc->normalise.b = sdR / sdB;
    } else if (sdMin==sdG) {
      pkc->normalise.r = sdG / sdR;
      pkc->normalise.b = sdG / sdB;
    } else if (sdMin==sdB) {
      pkc->normalise.r = sdB / sdR;
      pkc->normalise.g = sdB / sdG;
    }
  }

  if (pkc->normalise.r == 0) pkc->normalise.r = 1.0;
  if (pkc->normalise.g == 0) pkc->normalise.g = 1.0;
  if (pkc->normalise.b == 0) pkc->normalise.b = 1.0;

  if (pkc->verbose)
    fprintf (pkc->fh_data, "kcluster: normalise %.*g,%.*g,%.*g\n",
      pkc->precision, pkc->normalise.r,
      pkc->precision, pkc->normalise.g,
      pkc->precision, pkc->normalise.b);

  return MagickTrue;
}

static MagickStatusType WriteFrame (
  kclusterT * pkc,
  Image * img,
  ExceptionInfo *exception)
{
  ImageInfo
    *ii;

  MagickStatusType
    status;

  Image
    *copy_img;

  ii = AcquireImageInfo ();

  copy_img = CloneImage(img, 0, 0, MagickTrue, exception);
  if (copy_img == (Image *) NULL)
    return MagickFalse;

  copy_img->scene = pkc->frameNum++;

  CopyMagickString (copy_img->filename, pkc->frameName, MaxTextExtent);

#if IMV6OR7==6
  status = WriteImage (ii, copy_img);
#else
  status = WriteImage (ii, copy_img, exception);
#endif

  DestroyImageList(copy_img);

  ii = DestroyImageInfo (ii);

  return status;
}


static MagickBooleanType ApplyNormalise (
  kclusterT * pkc,
  Image *image,
  ExceptionInfo *exception,
  MagickBooleanType IsForwards
)
{
  double mr, mg, mb;
  if (IsForwards) {
    mr = pkc->normalise.r;
    mg = pkc->normalise.g;
    mb = pkc->normalise.b;
  } else {
    mr = 1.0 / pkc->normalise.r;
    mg = 1.0 / pkc->normalise.g;
    mb = 1.0 / pkc->normalise.b;
  }

  CacheView *in_view = AcquireAuthenticCacheView (image, exception);
  ssize_t x, y;
  for (y=0; y < (ssize_t) image->rows; y++) {
    VIEW_PIX_PTR *p=GetCacheViewAuthenticPixels(in_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL) return MagickFalse;

    for (x=0; x < (ssize_t) image->columns; x++) {
      SET_PIXEL_RED (image, mr * GET_PIXEL_RED(image,p), p);
      SET_PIXEL_GREEN (image, mg * GET_PIXEL_GREEN(image,p), p);
      SET_PIXEL_BLUE (image, mb * GET_PIXEL_BLUE(image,p), p);

      p += Inc_ViewPixPtr (image);
    }
    if (!SyncCacheViewAuthenticPixels(in_view,exception))
      return MagickFalse;
  }
  in_view = DestroyCacheView (in_view);

  return MagickTrue;
}

static double inline colDistRel (
  Image *image0,
  Image *image1,
  const VIEW_PIX_PTR *p0,
  const VIEW_PIX_PTR *p1
)
// Returns relative distance between colours.
{
  double r = GET_PIXEL_RED(image0,p0)   - GET_PIXEL_RED(image1,p1);
  double g = GET_PIXEL_GREEN(image0,p0) - GET_PIXEL_GREEN(image1,p1);
  double b = GET_PIXEL_BLUE(image0,p0)  - GET_PIXEL_BLUE(image1,p1);

  return (r*r + g*g + b*b);
}


static double inline colDist (
  Image *image0,
  Image *image1,
  const VIEW_PIX_PTR *p0,
  const VIEW_PIX_PTR *p1
)
// Returns distance between colours, scale 0.0 to 1.0.
{
  return sqrt (colDistRel(image0, image1, p0, p1) / 3.0) / QuantumRange;
}


static double inline powDist (
  Image *image0,
  Image *image1,
  const VIEW_PIX_PTR *p0,
  const VIEW_PIX_PTR *p1,
  double powParam
)
{
  double r = colDist (image0, image1, p0, p1);
  if (r==0) return 0;
  return pow (r, powParam);
}

static double inline sumPowDist (
  Image *image,
  Image *uniq_img,
  const VIEW_PIX_PTR *pImg,  // an image pixel
  const VIEW_PIX_PTR *pUniq, // first entry in uniq_img
  double powParam
)
{
  double r = 0.0, c;
  ssize_t qx;
  const VIEW_PIX_PTR *qu = pUniq;
  for (qx=0; qx < uniq_img->columns; qx++) {
    c = colDist (image, uniq_img, pImg, qu);
    if (c > 0) r += pow (c, powParam);
    qu += Inc_ViewPixPtr (uniq_img);
  }
  return r;
}

static double inline fuzzyMembership (
  kclusterT * pkc,
  Image *image,
  Image *uniq_img,
  const VIEW_PIX_PTR *pImg,  // an image pixel
  const VIEW_PIX_PTR *pUniq, // first entry in uniq_img
  ssize_t qx,                // index into uniq_img
  double denom
)
{
  double numer = powDist (image, uniq_img,
                   pImg, pUniq + qx * Inc_ViewPixPtr (uniq_img),
                   pkc->powParam1);

  if (numer == denom) return 1.0;

  if (pkc->debug) {
    if (isnan(numer)) fprintf (stderr, "numer is nan ");
    if (isnan(denom)) fprintf (stderr, "denom is nan ");

    if (isnan (numer / denom)) {
      fprintf (stderr, "fuzm %g,%g  qx=%li  ", numer, denom, qx);
    }

    if (numer > denom) fprintf (stderr, "fm %g/%g  ", numer, denom);
  }

  return numer / denom;
}


static double inline harmonicMembership (
  kclusterT * pkc,
  Image *image,
  Image *uniq_img,
  const VIEW_PIX_PTR *pImg,  // an image pixel
  const VIEW_PIX_PTR *pUniq, // first entry in uniq_img
  ssize_t qx,                // index into uniq_img
  double denom
)
{
  double numer = powDist (image, uniq_img,
                   pImg, pUniq + qx * Inc_ViewPixPtr (uniq_img),
                   pkc->powParam2);
  //double denom = sumPowDist (image, uniq_img, pImg, pUniq, pkc->powParam2);

  if (numer == denom) return 1.0;

  return numer / denom;
}


static double inline harmonicWeight (
  kclusterT * pkc,
  Image *image,
  Image *uniq_img,
  const VIEW_PIX_PTR *pImg, // an image pixel
  const VIEW_PIX_PTR *pUniq // first entry in uniq_img
)
{
  double numer = sumPowDist (image, uniq_img, pImg, pUniq, pkc->powParam2);
  double denom = sumPowDist (image, uniq_img, pImg, pUniq, pkc->powParam3);
  denom = denom * denom;

  if (numer == denom) return 1.0;

  if (pkc->debug) {
    if (isnan(numer)) fprintf (stderr, "numer is nan ");
    if (isnan(denom)) fprintf (stderr, "denom is nan ");

    if (isnan (numer / denom)) {
      fprintf (stderr, "harm w %g,%g  ", numer, denom);
    }
  }

  return numer / denom;
}

static void fuzzyDither (
  kclusterT * pkc,
  Image *image,
  Image *uniq_img,
  ssize_t x,
  ssize_t y,
  VIEW_PIX_PTR *pImg,        // an image pixel
  const VIEW_PIX_PTR *pUniq  // first entry in uniq_img
)
{
  double rand = GetPseudoRandomValue(pkc->random_info);
  double cumMemb = 0;

  if (!pkc->membDenoms)
    fprintf (stderr, "null membDenoms\n");

  double denom = pkc->membDenoms[y*image->columns+x];

  ssize_t qx;
  for (qx=0; qx < uniq_img->columns; qx++) {

    // stop when cumulative membership >= rand
    cumMemb += fuzzyMembership (pkc, image, uniq_img, pImg, pUniq, qx, denom);

    if (cumMemb >= rand) {
      const VIEW_PIX_PTR *qu = pUniq + qx * Inc_ViewPixPtr (uniq_img);
      SET_PIXEL_RED   (image, GET_PIXEL_RED   (uniq_img, qu), pImg);
      SET_PIXEL_GREEN (image, GET_PIXEL_GREEN (uniq_img, qu), pImg);
      SET_PIXEL_BLUE  (image, GET_PIXEL_BLUE  (uniq_img, qu), pImg);
      break;
    }
  }
}


static MagickBooleanType setMembDenoms (
  kclusterT * pkc,
  Image *image,
  Image *uniq_img,
  VIEW_PIX_PTR *pUniq,
  CacheView *in_view,
  double powParam, // pkc->powParam1
  ExceptionInfo *exception)
{
  if (!pkc->membDenoms) return MagickFalse;

  ssize_t x, y;

  for (y=0; y < (ssize_t) image->rows; y++) {
    const VIEW_PIX_PTR *p=GetCacheViewVirtualPixels(in_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL) return MagickFalse;

    for (x=0; x < (ssize_t) image->columns; x++) {
      pkc->membDenoms[y*image->columns+x] = sumPowDist (image, uniq_img, p, pUniq, powParam);
      p += Inc_ViewPixPtr (image);
    }
  }
  return MagickTrue;
}



static MagickBooleanType updateImage (
  kclusterT * pkc,
  Image *image,
  Image *uniq_img,
  VIEW_PIX_PTR *pUniq,
  ExceptionInfo *exception
)
// Returns whether okay.
{
  // Populate image with the updated cluster colours.

  assert(image != (Image *) NULL);
  assert(image->signature == MAGICK_CORE_SIG);

  assert (pkc->clusters);

  CacheView *upd_view = AcquireAuthenticCacheView (image, exception);

  if (pkc->dither) {
    if (!setMembDenoms (
          pkc, image, uniq_img, pUniq, upd_view, pkc->powParam1,
          exception)) return MagickFalse;
  }

  ssize_t x, y, qx;
  for (y=0; y < (ssize_t) image->rows; y++) {
    VIEW_PIX_PTR *p=GetCacheViewAuthenticPixels(upd_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL)
      return MagickFalse;

    for (x=0; x < (ssize_t) image->columns; x++) {
      if (pkc->dither) {
        fuzzyDither (pkc, image, uniq_img, x, y, p, pUniq);
      } else {
        qx = pkc->nearClust[y*image->columns+x];
        clusterT * pClust = &pkc->clusters[qx];

        SET_PIXEL_RED (image, pClust->red, p);
        SET_PIXEL_GREEN (image, pClust->green, p);
        SET_PIXEL_BLUE (image, pClust->blue, p);
      }
      p += Inc_ViewPixPtr (image);
    }
    if (!SyncCacheViewAuthenticPixels(upd_view,exception))
        return MagickFalse;
  }

  upd_view = DestroyCacheView (upd_view);

  return MagickTrue;
}



// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *kclusterOneK (
  Image *image,
  kclusterT * pkc,
  ExceptionInfo *exception)
//
// This may adjust pkc->kCols downwards.
//
{
  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  ssize_t nPixels = image->columns * image->rows;

  // Beware: if FromList, we don't know kCol yet.

  if (pkc->kCols < 1) pkc->kCols = 1;

  if (pkc->kCols > nPixels) pkc->kCols = nPixels;

  if (pkc->verbose) {
    fprintf (pkc->fh_data, "kcluster: Input image [%s] %ix%i depth %i; nPixels %li  k=%i\n",
             image->filename,
             (int)image->columns, (int)image->rows,
             (int)image->depth,
             (long int)nPixels,
             pkc->kCols);
  }

  pkc->nearClust = (ssize_t *) AcquireQuantumMemory (nPixels, sizeof(ssize_t));
  if (!pkc->nearClust) {
    fprintf (stderr, "kcluster: oom2\n");
    return (Image *)NULL;
  }

  // If FromList, we don't know kCol yet.
  // But RandPart needs clusters.

  pkc->clusters = NULL;

  Image *uniq_img = NULL;
  switch (pkc->init) {
    case iColors:
      default:
      uniq_img = InitialiseColors (pkc, image, exception);
      break;
    case iForgy:
      uniq_img = InitialiseForgy (pkc, image, exception);
      break;
    case iRandPart:
      pkc->clusters = (clusterT *) AcquireQuantumMemory (pkc->kCols, sizeof(clusterT));
      if (!pkc->clusters) {
        fprintf (stderr, "kcluster: oom2\n");
        return (Image *)NULL;
      }
      uniq_img = InitialiseRandomPartition (pkc, image, exception);
      break;
    case iSdLine:
      uniq_img = InitialiseSdLine (pkc, image, exception);
      break;
    case iFromList:
      uniq_img = InitialiseFromList (pkc, image, exception);

      if (pkc->kCols > nPixels) pkc->kCols = nPixels;
      break;
  }

  if (!uniq_img) {
    fprintf (stderr, "kcluster: Initialise failed\n");
    return NULL;
  }

  if (!pkc->clusters) {
    pkc->clusters = (clusterT *) AcquireQuantumMemory (pkc->kCols, sizeof(clusterT));
    if (!pkc->clusters) {
      fprintf (stderr, "kcluster: oom2\n");
      return (Image *)NULL;
    }
  }

  if (pkc->debug) DumpImage (pkc, uniq_img, exception);

  if (pkc->verbose)
    fprintf (pkc->fh_data, "kcluster: %li uniq_img\n", uniq_img->columns);

  CacheView *uniq_view = AcquireAuthenticCacheView (uniq_img,exception);

  VIEW_PIX_PTR *q_uniq = GetCacheViewAuthenticPixels(uniq_view,0,0,uniq_img->columns,1,exception);
  if (!q_uniq) return NULL;

  CacheView *in_view = AcquireVirtualCacheView (image, exception);

  pkc->powParam1 = -2 / (pkc->fuzzyR - 1);
  pkc->powParam2 = -pkc->harmonicP - 2;
  pkc->powParam3 = -pkc->harmonicP;

  pkc->membDenoms = (double *) AcquireQuantumMemory (nPixels, sizeof(double));
  if (!pkc->membDenoms) {
    fprintf (stderr, "kcluster: oom membDen\n");
    return (Image *)NULL;
  }

  double opacity = 1.0;

  // Note that an iteration is NOT guaranteed to be better than the previous one.

  int iter;
  for (iter=0; iter < 9999; iter++) {

    ssize_t x, y;
    ssize_t qx;

    // Loop through x, y, uniq, to populate nearClust.

    for (y=0; y < (ssize_t) image->rows; y++) {
      const VIEW_PIX_PTR *p=GetCacheViewVirtualPixels(in_view,0,y,image->columns,1,exception);
      if (p == (const VIEW_PIX_PTR *) NULL) return (Image *)NULL;

      for (x=0; x < (ssize_t) image->columns; x++) {
        double minDist=0;
        VIEW_PIX_PTR *qu = q_uniq;
        for (qx=0; qx < uniq_img->columns; qx++) {
          double dist = colDistRel (image, uniq_img, p, qu);
          if (qx==0 || minDist > dist) {
            minDist = dist;
            pkc->nearClust[y*image->columns+x] = qx;
          }
          qu += Inc_ViewPixPtr (uniq_img);
        }
        p += Inc_ViewPixPtr (image);
      }
    }

    if (pkc->method == mMeans) {
      // Zero cluster accumulators.

      for (qx=0; qx < uniq_img->columns; qx++) {
        clusterT * pClust = &pkc->clusters[qx];
        pClust->red = pClust->green = pClust->blue = 0.0;
        pClust->count = 0;
      }

      // Loop though y,x; add colour to appropriate cluster accumulator.

      for (y=0; y < (ssize_t) image->rows; y++) {
        const VIEW_PIX_PTR *p=GetCacheViewVirtualPixels(in_view,0,y,image->columns,1,exception);
        if (p == (const VIEW_PIX_PTR *) NULL) return (Image *)NULL;

        for (x=0; x < (ssize_t) image->columns; x++) {
          ssize_t qx = pkc->nearClust[y*image->columns+x];
          clusterT * pClust = &pkc->clusters[qx];
          if (pkc->regardAlpha)
            opacity = GET_PIXEL_ALPHA (image, p) / QuantumRange;
          pClust->red   += opacity * GET_PIXEL_RED (image, p);
          pClust->green += opacity * GET_PIXEL_GREEN (image, p);
          pClust->blue  += opacity * GET_PIXEL_BLUE (image, p);
          pClust->count += opacity;

          p += Inc_ViewPixPtr (image);
        }
      }

      // Calc the mean colour per cluster.

      for (qx=0; qx < uniq_img->columns; qx++) {
        clusterT * pClust = &pkc->clusters[qx];
        if (pClust->count) {
          pClust->red   /= pClust->count;
          pClust->green /= pClust->count;
          pClust->blue  /= pClust->count;
        }
      }
    } else if (pkc->method == mFuzzy) {
      if (!setMembDenoms (
            pkc, image, uniq_img, q_uniq, in_view, pkc->powParam1,
            exception)) return NULL;

      for (qx=0; qx < uniq_img->columns; qx++) {
        double sumM = 0.0, sumMr = 0.0, sumMg = 0.0, sumMb = 0.0;
        for (y=0; y < (ssize_t) image->rows; y++) {
          const VIEW_PIX_PTR *p=GetCacheViewVirtualPixels(in_view,0,y,image->columns,1,exception);
          if (p == (const VIEW_PIX_PTR *) NULL) return (Image *)NULL;

          for (x=0; x < (ssize_t) image->columns; x++) {
            double m = fuzzyMembership (pkc,
                         image, uniq_img, p, q_uniq, qx,
                         pkc->membDenoms[y*image->columns+x]);

            if (isnan(m)) fprintf (stderr, "m is nan  xy=%li,%li  ", x, y);

            if (pkc->regardAlpha) {
              opacity = GET_PIXEL_ALPHA (image, p) / QuantumRange;
              m *= opacity;
            }

            sumMr += m * GET_PIXEL_RED (image, p);
            sumMg += m * GET_PIXEL_GREEN (image, p);
            sumMb += m * GET_PIXEL_BLUE (image, p);
            sumM  += m;

            p += Inc_ViewPixPtr (image);
          }
        }
        if (pkc->debug) fprintf (pkc->fh_data, "fuzzy %li %g, %g,%g,%g\n", 
          qx,
          sumM, sumMr, sumMg, sumMb);

        clusterT * pClust = &pkc->clusters[qx];

        if (sumM==0) {
          pClust->red = pClust->green = pClust->blue = 0;
        } else {
          pClust->red   = sumMr / sumM;
          pClust->green = sumMg / sumM;
          pClust->blue  = sumMb / sumM;
        }
      }

    } else if (pkc->method == mHarmonic) {

      double *weights = (double *) AcquireQuantumMemory (nPixels, sizeof(double));
      if (!weights) {
        fprintf (stderr, "kcluster: oomwt\n");
        return (Image *)NULL;
      }
      for (y=0; y < (ssize_t) image->rows; y++) {
        const VIEW_PIX_PTR *p=GetCacheViewVirtualPixels(in_view,0,y,image->columns,1,exception);
        if (p == (const VIEW_PIX_PTR *) NULL) return (Image *)NULL;

        for (x=0; x < (ssize_t) image->columns; x++) {
          weights[y*image->columns+x] = harmonicWeight (pkc, image, uniq_img, p, q_uniq);
          pkc->membDenoms[y*image->columns+x] = sumPowDist (image, uniq_img, p, q_uniq, pkc->powParam2);
          p += Inc_ViewPixPtr (image);
        }
      }

      for (qx=0; qx < uniq_img->columns; qx++) {
        double sumM = 0.0, sumMr = 0.0, sumMg = 0.0, sumMb = 0.0;
        for (y=0; y < (ssize_t) image->rows; y++) {
          const VIEW_PIX_PTR *p=GetCacheViewVirtualPixels(in_view,0,y,image->columns,1,exception);
          if (p == (const VIEW_PIX_PTR *) NULL) return (Image *)NULL;

          for (x=0; x < (ssize_t) image->columns; x++) {
            double m = harmonicMembership (
              pkc, image, uniq_img, p, q_uniq, qx,
              pkc->membDenoms[y*image->columns+x]);
            if (isnan(m)) fprintf (stderr, "m is nan ");

            double w = weights[y*image->columns+x];
            if (isnan(w)) fprintf (stderr, "w is nan ");

            if (pkc->regardAlpha) {
              opacity = GET_PIXEL_ALPHA (image, p) / QuantumRange;
              m *= opacity;
            }
            double mw = m * w;
            sumMr += mw * GET_PIXEL_RED (image, p);
            sumMg += mw * GET_PIXEL_GREEN (image, p);
            sumMb += mw * GET_PIXEL_BLUE (image, p);
            sumM  += mw;

            p += Inc_ViewPixPtr (image);
          }
        }
        clusterT * pClust = &pkc->clusters[qx];

        if (sumM==0) {
          pClust->red = pClust->green = pClust->blue = 0;
        } else {
          pClust->red   = sumMr / sumM;
          pClust->green = sumMg / sumM;
          pClust->blue  = sumMb / sumM;
        }
      }
      weights = RelinquishMagickMemory (weights);

    } else if (pkc->method == mNull) {

      VIEW_PIX_PTR *qu = q_uniq;
      for (qx=0; qx < uniq_img->columns; qx++) {
        clusterT * pClust = &pkc->clusters[qx];
        pClust->red   = GET_PIXEL_RED (uniq_img, qu);
        pClust->green = GET_PIXEL_GREEN (uniq_img, qu);
        pClust->blue  = GET_PIXEL_BLUE (uniq_img, qu);
        qu += Inc_ViewPixPtr (uniq_img);
      }

    } // mNull

    // Calc how much we have changed;
    // update the cluster colours.

    VIEW_PIX_PTR *qu = q_uniq;
    double diff, maxDiff = 0;

    for (qx=0; qx < uniq_img->columns; qx++) {
      clusterT * pClust = &pkc->clusters[qx];
      diff = fabs (pClust->red - GET_PIXEL_RED (uniq_img, qu));
      if (maxDiff < diff) maxDiff = diff;
      diff = fabs (pClust->green - GET_PIXEL_GREEN (uniq_img, qu));
      if (maxDiff < diff) maxDiff = diff;
      diff = fabs (pClust->blue - GET_PIXEL_BLUE (uniq_img, qu));
      if (maxDiff < diff) maxDiff = diff;

      SET_PIXEL_RED   (uniq_img, pClust->red, qu);
      SET_PIXEL_GREEN (uniq_img, pClust->green, qu);
      SET_PIXEL_BLUE  (uniq_img, pClust->blue, qu);

      qu += Inc_ViewPixPtr (uniq_img);
    }

    if (!SyncCacheViewAuthenticPixels(uniq_view,exception))
        return NULL;

    maxDiff /= QuantumRange;

    if (pkc->verbose > 1) {
      fprintf (pkc->fh_data,
        "kcluster: iteration %i  maxDiff %.*g\n",
        iter, pkc->precision, maxDiff);
    }

    if (pkc->debug) DumpImage (pkc, uniq_img, exception);


    if (pkc->frameName) {
      // For performance: don't need to create and destroy every time.
      Image * frame_img = CloneImage(image, 0, 0, MagickTrue, exception);
      if (frame_img == (Image *) NULL)
        return MagickFalse;

      if (!updateImage (pkc, frame_img, uniq_img, q_uniq, exception)) return NULL;
      WriteFrame (pkc, frame_img, exception);
      DestroyImage (frame_img);
    }


    if (maxDiff <= pkc->tolerance && (iter > 2 || pkc->init != iRandPart))
      break;

    if (pkc->maxIter && iter >= pkc->maxIter-1) break;

  } // end iteration loop


  // Populate image with the updated cluster colours.

  if (!updateImage (pkc, image, uniq_img, q_uniq, exception)) return NULL;

  if (pkc->frameName) WriteFrame (pkc, image, exception);


  if (pkc->membDenoms) 
    pkc->membDenoms = RelinquishMagickMemory (pkc->membDenoms);

  in_view = DestroyCacheView (in_view);

  uniq_view = DestroyCacheView (uniq_view);



  if (pkc->verbose) {
    fprintf (pkc->fh_data, "kcluster: %i iterations\n", iter+1);
  }

  DestroyImage (uniq_img);

  pkc->clusters  = RelinquishMagickMemory (pkc->clusters);
  pkc->nearClust  = RelinquishMagickMemory (pkc->nearClust);

  return (image);
}


static MagickBooleanType CompareRmseAlpha (
  Image *image1,
  Image *image2,
  double * score,
  ExceptionInfo *exception)
// Returns RMSE score (aka "distortion") of two equal-size images,
// taking alpha into account.
{
  CacheView *in1_view = AcquireVirtualCacheView (image1, exception);
  CacheView *in2_view = AcquireVirtualCacheView (image2, exception);

  if ((image1->columns != image2->columns) || (image1->rows != image2->rows)) {
    fprintf (stderr, "CompareRmseAlpha: images sizes should match");
    return MagickFalse;
  }

  double sigScore=0, sigmAlpha=0;

  ssize_t y;
  for (y=0; y < (ssize_t) image1->rows; y++) {
    const VIEW_PIX_PTR *p1=GetCacheViewVirtualPixels(in1_view,0,y,image1->columns,1,exception);
    if (p1 == (const VIEW_PIX_PTR *) NULL) return MagickFalse;
    const VIEW_PIX_PTR *p2=GetCacheViewVirtualPixels(in2_view,0,y,image2->columns,1,exception);
    if (p2 == (const VIEW_PIX_PTR *) NULL) return MagickFalse;

    ssize_t x;
    double a1, a2;
    for (x=0; x < (ssize_t) image1->columns; x++) {

      a1 = GET_PIXEL_ALPHA(image1, p1) / QuantumRange;
      a2 = GET_PIXEL_ALPHA(image2, p2) / QuantumRange;

      double mAlpha = a1 * a2;

      double dRed   =
         (GET_PIXEL_RED(image1,p1)
        - GET_PIXEL_RED(image2,p2)) / QuantumRange;
      double dGreen =
         (GET_PIXEL_GREEN(image1,p1)
        - GET_PIXEL_GREEN(image2,p2)) / QuantumRange;
      double dBlue  =
         (GET_PIXEL_BLUE(image1,p1)
        - GET_PIXEL_BLUE(image2,p2)) / QuantumRange;

      if (mAlpha > 0) {
        sigScore += mAlpha * (dRed*dRed + dGreen*dGreen + dBlue*dBlue);

        sigmAlpha += mAlpha;
      }

      p1 += Inc_ViewPixPtr (image1);
      p2 += Inc_ViewPixPtr (image2);
    }
  }

  in2_view = DestroyCacheView (in2_view);
  in1_view = DestroyCacheView (in1_view);

  if (sigmAlpha==0) {
    // Every location has one or other pixel fully transparent.
    *score = 0.0;
  } else {
    *score = sqrt (sigScore / sigmAlpha / 3.0);
  }

  return MagickTrue;
}


static Image *kcluster (
  Image *image,
  kclusterT * pkc,
  ExceptionInfo *exception)
{
  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (! SetNoPalette (image, exception))
    return NULL;

  if (pkc->sdNorm) {
    if (!CalcNormalise (pkc, image, exception)) {
      fprintf (stderr, "kcluster: CalcNormalise failed\n");
      return (Image *)NULL;
    }

    if (!ApplyNormalise (pkc, image, exception, MagickTrue)) {
      fprintf (stderr, "kcluster: ApplyNormalise failed\n");
      return (Image *)NULL;
    }
  }

  if (!pkc->hasJump) {
    image = kclusterOneK (image, pkc, exception);
  } else {

    if (pkc->verbose) {
      fprintf (pkc->fh_data, "kcluster: jump %i to %i\n", pkc->jump0, pkc->jump1);
    }

    double jumpY;   // The transform power
    if (pkc->hasSetY) {
      jumpY = pkc->jumpY;
    } else {
      double jumpP;
#if IMV6OR7==6
      jumpP = IsGrayImage (image, exception) ? 1 : 3; // Number of dimensions.
#else
      jumpP = IsImageGray (image) ? 1 : 3; // Number of dimensions.
#endif
      jumpY = jumpP / 2.0;
    }
    if (pkc->verbose) {
      fprintf (pkc->fh_data, "kcluster: jumpY = %.*g\n",
        pkc->precision, jumpY);
    }

    double trans, J, maxJ = 0;
    int kAtMax = 0;
    double transPrev = 0.0;

    // if jump0 > 1,
    // start at jump0-1 but don't count first result for maxJ.

    MagickBooleanType ignoreThis = MagickFalse;
    int initK = pkc->jump0;
    if (initK > 1) {
      ignoreThis = MagickTrue;
      initK--;
    }

    int k;
    for (k = initK; k <= pkc->jump1; k++) {

      // clone it, cluster it, compare it

      pkc->kCols = k;

      Image * jump_img = CloneImage(image, 0, 0, MagickTrue, exception);
      if (jump_img == (Image *) NULL)
        return MagickFalse;

      jump_img = kclusterOneK (jump_img, pkc, exception);

      double dist, distArg;
      // FIXME: following ignores alpha.
      if (!GetImageDistortion (
       image, jump_img, RootMeanSquaredErrorMetric, &dist, exception))
      {
        return MagickFalse;
      }

      if (!CompareRmseAlpha (
       image, jump_img, &distArg, exception))
      {
        return MagickFalse;
      }

      trans = (dist > 0) ? pow (dist, -jumpY) : 0;
      J = trans - transPrev;

      if (!ignoreThis) {
        if (maxJ < J) {
          maxJ = J;
          kAtMax = pkc->kCols;
        }

        if (pkc->verbose) {
          fprintf (pkc->fh_data, "kcluster: k=%i  d=%.*g  dArg=%.*g  trans=%.*g  J=%.*g\n",
            pkc->kCols,
            pkc->precision, dist,
            pkc->precision, distArg,
            pkc->precision, trans,
            pkc->precision, J);
        }
      }

      transPrev = trans;
      ignoreThis = MagickFalse;

      DestroyImage (jump_img);
    }

    if (pkc->verbose) {
      fprintf (pkc->fh_data, "kcluster: maxJ=%.*g  at k=%i\n",
        pkc->precision, maxJ,
        kAtMax);
    }

    if (kAtMax == 0) kAtMax = 1;

    // FIXME: write this to attribute?

    pkc->kCols = kAtMax;
    image = kclusterOneK (image, pkc, exception);
  }

  if (pkc->sdNorm) {
    if (!ApplyNormalise (pkc, image, exception, MagickFalse)) {
      fprintf (stderr, "kcluster: ApplyNormalise failed\n");
      return MagickFalse;
    }
  }

  return image;
}



ModuleExport size_t kclusterImage (
  Image **images,
  const int argc,
  const char **argv,
  ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  kclusterT pkc;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  pkc.precision = GetMagickPrecision();

  status = menu (argc, argv, &pkc);
  if (status == MagickFalse)
    return (-1);

  if (!pkc.random_info) InitRand (&pkc);

  if (pkc.init == iFromList) {
    int ListLen = (int)GetImageListLength(*images);
    if (ListLen < 2) {
      fprintf (stderr, "kcluster: initialize FromList needs at least 2 images\n");
      return (-1);
    }
  }

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    Image * next = GetNextImageInList(image);
    if (next == NULL && pkc.init == iFromList) {
      //fprintf (stderr, "discarding map image\n");
      DeleteImageFromList (&image);
    } else {
      new_image = kcluster (image, &pkc, exception);
      if (!new_image) return (-1);

      if (new_image != image) {
        ReplaceImageInList(&image,new_image);
      }
      *images=GetFirstImageInList(image);
    }
  }

  DeInitRand (&pkc);

  return(MagickImageFilterSignature);
}
