#define DEBUG_MATCH 0

static void SetDefaultMatch (MatchT * pm)
{
  pm->ref_rows = pm->ref_columns = 0;
  pm->matchRadX = pm->matchRadY = 1;
  pm->AutoLs = MagickTrue;
  pm->Refine = MagickTrue;
  pm->SearchToEdges = MagickTrue;
  pm->simThreshold = pm->simThresholdSq = 0.0;
  pm->dissimThreshold = pm->dissimThresholdSq = 0.0;
  pm->autoCorrectThreshold = pm->autoCorrectThresholdSq = 0.0;
  pm->DissimOn = MagickFalse;
  pm->searchWhat = swEntire;
  pm->LimSrchRad = 0;
  pm->LimSrchRad_IsPc = MagickFalse;
  pm->limSrchRadX = pm->limSrchRadY = 0;
  pm->SetAutoLimit = pm->UseAutoLimit = MagickFalse;
  pm->RandSearchesF = pm->SkipNumF = 0;
  pm->RandSearchesI = pm->SkipNumI = 0;
  pm->RandSearches_IsPc = pm->SkipNum_IsPc = MagickFalse;
  pm->DoSubTransTest = MagickFalse;
  pm->limLeft = pm->limTop = pm->limRight = pm->limBot = 0;
  pm->nLimitCnt = 0;

  pm->bestX = pm->bestY = 0;
  pm->bestScore = 0;
}

static void CalcMatchRange (
  ssize_t ref_x,
  ssize_t ref_y,
  ssize_t limX,
  ssize_t limY,
  MatchT * pm,
  RangeT * rng)
{
  if (pm->SearchToEdges) {
    rng->i0 = -pm->matchRadY;
    rng->i1 = pm->ref_rows - pm->matchRadY;
    rng->j0 = -pm->matchRadX;
    rng->j1 = pm->ref_columns - pm->matchRadX;
  } else {
    rng->i0 = 0;
    rng->i1 = pm->ref_rows - 2*pm->matchRadY - 1;
    rng->j0 = 0;
    rng->j1 = pm->ref_columns - 2*pm->matchRadX - 1;
  }

  if (limX) {
    ssize_t
      limj0 = ref_x - limX - pm->matchRadX,
      limj1 = ref_x + limX - pm->matchRadX;

    if (rng->j0 < limj0) rng->j0 = limj0;
    if (rng->j1 > limj1) rng->j1 = limj1;
  }

  if (limY) {
    ssize_t
      limi0 = ref_y - limY - pm->matchRadY,
      limi1 = ref_y + limY - pm->matchRadY;

    if (rng->i0 < limi0) rng->i0 = limi0;
    if (rng->i1 > limi1) rng->i1 = limi1;
  }

  //if (pfh->debug==MagickTrue) printf ("CalcMatchRange ji A %li-%li %li-%li\n", rng->j0, rng->j1, rng->i0, rng->i1);

  if (pm->UseAutoLimit) {
    ssize_t
      j0s = rng->j0,
      j1s = rng->j1,
      i0s = rng->i0,
      i1s = rng->i1;

    if (rng->j0 < pm->limLeft)    rng->j0 = pm->limLeft;
    if (rng->i0 < pm->limTop)     rng->i0 = pm->limTop;
    if (rng->j1 > pm->limRight+1) rng->j1 = pm->limRight+1;
    if (rng->i1 > pm->limBot+1)   rng->i1 = pm->limBot+1;

    if (rng->j1 <= rng->j0 || rng->i1 <= rng->i0) {
      // AutoLimit would reduce search space to nothing, so don't do it.

      printf ("AutoLimit %li %li %li %li\n", rng->j0, rng->j1, rng->i0, rng->i1);
      rng->j0 = j0s;
      rng->j1 = j1s;
      rng->i0 = i0s;
      rng->i1 = i1s;
    }
  }

}

static double Match (
  ssize_t x,
  ssize_t y,
  ssize_t ref_x,
  ssize_t ref_y,
  MatchT * pm,
  CompWindT * cmpwin,
  RandomInfo *restrict random_info,
  ExceptionInfo *exception)
// x,y is central pixel of window to be matched.
//
// ref_x, ref_y is guess for central pixel in reference.
//   Relevant only when search radius is limited.
//
// Always updates pm->bestX, pm->bestY and pm->bestScore,
//   even if bestScore is lousy.

{
  RangeT
    rng;

  ssize_t
    besti=0, bestj=0;

  if (pm->DoSubTransTest && IsSubTrans (cmpwin, exception)) {
    pm->bestX = 0;
    pm->bestY = 0;
    pm->bestScore = cmpwin->WorstCompare;
    return pm->bestScore;
  }

  CalcMatchRange (ref_x, ref_y, pm->limSrchRadX, pm->limSrchRadY, pm, &rng);

  double bestScore = cmpwin->WorstCompare;
  double v;

#if DEBUG_MATCH==1
  printf ("Match ji B %li to %li, %li to %li\n", rng.j0, rng.j1, rng.i0, rng.i1);
#endif

  cmpwin->subX = x - pm->matchRadX;
  cmpwin->subY = y - pm->matchRadY;

  switch (pm->searchWhat) {
    case swEntire:
    default: {
      for (cmpwin->refY = rng.i0; cmpwin->refY < rng.i1; cmpwin->refY++) {
        for (cmpwin->refX = rng.j0; cmpwin->refX < rng.j1; cmpwin->refX++) {
          v = CompareWindow (cmpwin, exception);
          if (bestScore > v) {
            bestScore = v;
            besti = cmpwin->refY;
            bestj = cmpwin->refX;
            if (bestScore <= pm->simThresholdSq) break;
          }
        }
        if (bestScore <= pm->simThresholdSq) break;
      }

      break;
    }

    case swRandom: {
      int
        defaultN,
        nRand,
        n;

      defaultN = rng.i1-rng.i0+1;
      if (defaultN > rng.j1-rng.j0+1) defaultN = rng.j1-rng.j0+1;

      if (pm->RandSearches_IsPc) {
        nRand = defaultN * pm->RandSearchesF/100.0;
      } else {
        nRand = (pm->RandSearchesI == 0) ? defaultN : pm->RandSearchesI;
      }

      if (nRand < 1) nRand = 1;

      for (n = 0; n < nRand; n++) {
        cmpwin->refY = rng.i0 + (rng.i1-rng.i0)*GetPseudoRandomValue(random_info);
        cmpwin->refX = rng.j0 + (rng.j1-rng.j0)*GetPseudoRandomValue(random_info);

        v = CompareWindow (cmpwin, exception);
        if (bestScore > v) {
          bestScore = v;
          besti = cmpwin->refY;
          bestj = cmpwin->refX;
          if (bestScore <= pm->simThresholdSq) break;
        }
      }

      // FIXME: just RadY?
      if (pm->Refine && bestScore > pm->simThresholdSq && pm->matchRadY > 1) {
        // Refine the search.
        if (rng.i0 < besti - pm->matchRadY) rng.i0 = besti - pm->matchRadY;
        if (rng.i1 > besti + pm->matchRadY) rng.i1 = besti + pm->matchRadY;
        if (rng.j0 < bestj - pm->matchRadX) rng.j0 = bestj - pm->matchRadX;
        if (rng.j1 > bestj + pm->matchRadX) rng.j1 = bestj + pm->matchRadX;

        for (cmpwin->refY = rng.i0; cmpwin->refY < rng.i1; cmpwin->refY++) {
          for (cmpwin->refX = rng.j0; cmpwin->refX < rng.j1; cmpwin->refX++) {
            v = CompareWindow (cmpwin, exception);
            if (bestScore > v) {
              bestScore = v;
              besti = cmpwin->refY;
              bestj = cmpwin->refX;
              if (bestScore <= pm->simThresholdSq) break;
            }
          }
          if (bestScore <= pm->simThresholdSq) break;
        }
      }
      break;
    }
    case swSkip: {
      int defaultN = pm->matchRadX;  // FIXME: x or y or ?
      int nSkip;

      if (pm->SkipNum_IsPc) {
        nSkip = defaultN * pm->SkipNumF/100.0;
      } else {
        nSkip = (pm->SkipNumI == 0) ? defaultN : pm->SkipNumI;
      }

      if (nSkip < 1) nSkip = 1;
//printf ("skip:\n");
      for (cmpwin->refY = rng.i0; cmpwin->refY < rng.i1; cmpwin->refY += nSkip) {
        for (cmpwin->refX = rng.j0; cmpwin->refX < rng.j1; cmpwin->refX += nSkip) {
          v = CompareWindow (cmpwin, exception);
          if (bestScore > v) {
            bestScore = v;
            besti = cmpwin->refY;
            bestj = cmpwin->refX;
            if (bestScore <= pm->simThresholdSq) break;
          }
        }
        if (bestScore <= pm->simThresholdSq) break;
      }

      if (pm->Refine && bestScore > pm->simThresholdSq && nSkip > 1) {
        // Refine the search.
//printf ("refine skip:\n");
        if (rng.i0 < besti-nSkip+1) rng.i0 = besti-nSkip+1;
        if (rng.i1 > besti+nSkip  ) rng.i1 = besti+nSkip;
        if (rng.j0 < bestj-nSkip+1) rng.j0 = bestj-nSkip+1;
        if (rng.j1 > bestj+nSkip  ) rng.j1 = bestj+nSkip;

        for (cmpwin->refY = rng.i0; cmpwin->refY < rng.i1; cmpwin->refY++) {
          for (cmpwin->refX = rng.j0; cmpwin->refX < rng.j1; cmpwin->refX++) {
            v = CompareWindow (cmpwin, exception);
            if (bestScore > v) {
              bestScore = v;
              besti = cmpwin->refY;
              bestj = cmpwin->refX;
              if (bestScore <= pm->simThresholdSq) break;
            }
          }
          if (bestScore <= pm->simThresholdSq) break;
        }
      }
      break;
    }
  }

  if (bestScore < pm->dissimThresholdSq) {
    if (pm->SetAutoLimit) {
      if (pm->limLeft  > bestj) pm->limLeft  = bestj;
      if (pm->limTop   > besti) pm->limTop   = besti;
      if (pm->limRight < bestj) pm->limRight = bestj;
      if (pm->limBot   < besti) pm->limBot   = besti;
    }
  }


  // Experiment: auto-correction
  if (pm->autoCorrectThresholdSq > 0 && bestScore > pm->autoCorrectThresholdSq) {

    if (pm->SearchToEdges) {
      rng.i0 = -pm->matchRadY;
      rng.i1 = pm->ref_rows - pm->matchRadY;
      rng.j0 = -pm->matchRadX;
      rng.j1 = pm->ref_columns - pm->matchRadX;
    } else {
      rng.i0 = 0;
      rng.i1 = pm->ref_rows - 2*pm->matchRadY - 1;
      rng.j0 = 0;
      rng.j1 = pm->ref_columns - 2*pm->matchRadX - 1;
    }

    //MagickBooleanType HasCorrected = MagickFalse;

    for (cmpwin->refY = rng.i0; cmpwin->refY < rng.i1; cmpwin->refY++) {
      for (cmpwin->refX = rng.j0; cmpwin->refX < rng.j1; cmpwin->refX++) {
        v = CompareWindow (cmpwin, exception);
        if (bestScore > v) {
          bestScore = v;
          besti = cmpwin->refY;
          bestj = cmpwin->refX;
          //HasCorrected = MagickTrue;
          if (bestScore <= pm->simThresholdSq) break;
        }
      }
      if (bestScore <= pm->simThresholdSq) break;
    }

    //if (HasCorrected) printf ("hasc %g ", bestScore);
  }

  pm->bestX = bestj + pm->matchRadX;
  pm->bestY = besti + pm->matchRadY;
  pm->bestScore = bestScore;

#if DEBUG_MATCH==1
  printf ("Match best %li %li %g\n", pm->bestX, pm->bestY, pm->bestScore);
#endif

  return bestScore;
}

