#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include "vsn_defines.h"

/* Updated:
     15-August-2017 Permit 'c' suffix meaning "percent".
*/

typedef struct {
  size_t
    width;
  long double
    mean,
    std_dev,
    skew;
  int
    skew_meth;
  MagickBooleanType
    mean_IsPc,
    std_dev_IsPc,
    skew_IsPc,
    rev_skew,
    do_zeroize,
    do_cumul,
    do_norm,
    do_verbose;
} mkgaussT;

static void usage (void)
{
  printf ("Usage: -process 'mkgauss [OPTION]...'\n");
  printf ("Make a Gaussian clut.\n");
  printf ("\n");
  printf ("  w,  width N              output width in pixels\n");
  printf ("  m,  mean N               mean\n");
  printf ("  sd, standarddeviation N  standard deviation\n");
  printf ("  k,  skew N               skew to place peak way from mean\n");
  printf ("  st, skewtype N           skew type: 1 or 2\n");
  printf ("  z,  zeroize              subtract minimum value\n");
  printf ("  c,  cumul                cumulate the values\n");
  printf ("  n,  norm                 normalise output so maximum count is quantum\n");
  printf ("  v,  verbose              write text information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static inline MagickBooleanType EndsPc (const char *s)
{
  char ch = *(s+strlen(s)-1);
  return ((ch == '%') || (ch == 'c'));
}

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  mkgaussT * pmg
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  pmg->width = 256;
  pmg->mean = 50;
  pmg->mean_IsPc = MagickTrue;

  pmg->std_dev = 30;
  pmg->std_dev_IsPc = MagickTrue;

  pmg->skew = 0;
  pmg->skew_IsPc = MagickTrue;
  pmg->skew_meth = 1;

  pmg->do_zeroize = pmg->do_cumul = pmg->do_norm = pmg->do_verbose = MagickFalse;
  pmg->rev_skew = MagickFalse;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "w", "width")==MagickTrue) {
      i++;
      if (i >= argc) {
        fprintf (stderr, "mkgauss: 'width' needs an argument\n");
        status = MagickFalse;
        break;
      }
      pmg->width = atoi(argv[i]);
      if (pmg->width < 2) {
        fprintf (stderr, "mkgauss: invalid 'width' [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "m", "mean")==MagickTrue) {
      i++;
      if (i >= argc) {
        fprintf (stderr, "mkgauss: 'mean' needs an argument\n");
        status = MagickFalse;
        break;
      }
      pmg->mean = atof(argv[i]);
      pmg->mean_IsPc = EndsPc (argv[i]);
    } else if (IsArg (pa, "sd", "standarddeviation")==MagickTrue) {
      i++;
      if (i >= argc) {
        fprintf (stderr, "mkgauss: 'standarddeviation' needs an argument\n");
        status = MagickFalse;
        break;
      }
      pmg->std_dev = atof(argv[i]);
      pmg->std_dev_IsPc = EndsPc (argv[i]);
      if (pmg->std_dev <= 0) {
        fprintf (stderr, "mkgauss: invalid 'standarddeviation' [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "k", "skew")==MagickTrue) {
      i++;
      if (i >= argc) {
        fprintf (stderr, "mkgauss: 'skew' needs an argument\n");
        status = MagickFalse;
        break;
      }
      pmg->skew = atof(argv[i]);
      pmg->skew_IsPc = EndsPc (argv[i]);
      pmg->rev_skew = (pmg->skew > 0);
    } else if (IsArg (pa, "sm", "skewmethod")==MagickTrue) {
      i++;
      if (i >= argc) {
        fprintf (stderr, "mkgauss: 'skewmethod' needs an argument\n");
        status = MagickFalse;
        break;
      }
      pmg->skew_meth = atoi(argv[i]);
    } else if (IsArg (pa, "z", "zeroize")==MagickTrue) {
      pmg->do_zeroize = MagickTrue;
    } else if (IsArg (pa, "n", "norm")==MagickTrue) {
      pmg->do_norm = MagickTrue;
    } else if (IsArg (pa, "c", "cumul")==MagickTrue) {
      pmg->do_cumul = MagickTrue;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pmg->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "mkgauss: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (pmg->do_verbose) {
    fprintf (stderr, "mkgauss options:  ");
    fprintf (stderr, "  width %lu", pmg->width);
    fprintf (stderr, "  mean %g", (double)pmg->mean);
    if (pmg->mean_IsPc == MagickTrue) fprintf (stderr, "%%");

    fprintf (stderr, "  standarddeviation %g", (double)pmg->std_dev);
    if (pmg->std_dev_IsPc == MagickTrue) fprintf (stderr, "%%");

    fprintf (stderr, "  skew %g", (double)pmg->skew);
    if (pmg->skew_IsPc == MagickTrue) fprintf (stderr, "%%");
    fprintf (stderr, "  skewmethod %i", pmg->skew_meth);

    if (pmg->do_cumul)   fprintf (stderr, "  cumul");
    if (pmg->do_norm)    fprintf (stderr, "  norm");
    if (pmg->do_verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


#if defined(MAGICKCORE_HDRI_SUPPORT)
#define ADD_HALF
#else
#define ADD_HALF +0.5
#endif


// The next function corresponds in style to functions in transform.c
// It takes one image (but ignores it, aside from CloneImage),
// and returns an image.
//
static Image *make_gauss(const Image *image,
  mkgaussT * pmg,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  CacheView
    *out_view;

  ssize_t
    y;

  int
    precision;

  MagickBooleanType
    status;

  long double
    fmean,
    fstd_dev,
    fskew,
    fpow,
    gnd_fact,
    two_sdsq,
    slopeA,
    slopeB;

  precision = GetMagickPrecision();

  if (pmg->mean_IsPc == MagickTrue) {
    fmean = pmg->mean / 100.0;
  } else {
    fmean = pmg->mean / pmg->width;
  }

  if (pmg->std_dev_IsPc == MagickTrue) {
    fstd_dev = pmg->std_dev / 100.0;
  } else {
    fstd_dev = pmg->std_dev / pmg->width;
  }

  gnd_fact = 1.0 / (fstd_dev * sqrt (2 * M_PI));
  two_sdsq = 2 * fstd_dev * fstd_dev;

  if (pmg->skew_IsPc == MagickTrue) {
    fskew = fmean + (pmg->rev_skew ? -pmg->skew : pmg->skew) / 100.0;
  } else {
    fskew = fmean + (pmg->rev_skew ? -pmg->skew : pmg->skew) / pmg->width;
  }

  fpow = 1.0;
  if (fskew != 0.0) {
    long double logskew = log(fskew);
    if (logskew == 0.0) {
      fprintf (stderr, "logskew is zero");
      return (Image *) NULL;
    }
    fpow = log(fmean) / logskew;
  }

  if (fskew==0 || fskew==1) {
    slopeA = slopeB = 0;
  } else {
    slopeA = 0.5 / fskew;
    slopeB = 0.5 / (1 - fskew);
  }

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (pmg->do_verbose) {
    fprintf (stderr, "mkgauss: Input image [%s] depth is %i\n",
      image->filename, (int)image->depth);

    fprintf (stderr, "  mean %.*g  sd %.*g  skew %.*g  pow %.*g  slopeA %.*g  slopeB %.*g\n",
      precision, (double)fmean, precision, (double)fstd_dev,
      precision, (double)fskew, precision, (double)fpow,
      precision, (double)slopeA, precision, (double)slopeB);
  }

  new_image=CloneImage(image, pmg->width, 1, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  SetAllBlack (new_image, exception);

  status=MagickTrue;

  out_view=AcquireAuthenticCacheView(new_image,exception);

  for (y=0; y < (ssize_t) new_image->rows; y++)
  {
    VIEW_PIX_PTR
      *q_out;

    ssize_t
      x;

    long double
      xf,
      value,
      min_value,
      max_value,
      sum_value,
      delta_v,
      mult_fact,
      pix_value;

    if (status == MagickFalse)
      continue;

    q_out=GetCacheViewAuthenticPixels(out_view,0,y,new_image->columns,1,exception);
    if (q_out == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "mkgauss: bad GetCacheView out_view\n");
      status=MagickFalse;
    }

    max_value = 0.0;
    min_value = QuantumRange + 1.0;
    sum_value = 0.0;

    mult_fact = 1.0;

    for (x=0; x < (ssize_t) new_image->columns; x++)
    {
      xf = x / (long double)(pmg->width-1);
      //if (fpow != 1.0) xf = pow (xf, fpow);
      //value = gnd_fact * exp (-(xf-fmean)*(xf-fmean)/two_sdsq);

      //if (fpow != 1.0) xf = pow (fabs(xf-fmean), fpow);
      //else xf -= fmean;

      if (pmg->rev_skew) xf = 1.0 - xf;

      switch (pmg->skew_meth) {
        case 1: default:
          if (fpow != 1.0) xf = pow (xf, fpow);
          value = gnd_fact * exp (-(xf-fmean)*(xf-fmean)/two_sdsq);
          break;
        case 2:
          xf = (xf < fskew) ? xf * slopeA : (1-xf) * slopeB;
          value = gnd_fact * exp (-(xf-fmean)*(xf-fmean)/two_sdsq);
          break;
      }

      if (max_value < value) max_value = value;
      if (min_value > value) min_value = value;
      sum_value += value;

      //FormatLocaleFile (stdout, "xf=%.*g value=%.*g\n",
      //  precision, (double)xf, precision, (double)value);
    }

    if (pmg->do_verbose) {
      FormatLocaleFile (stderr,
        "  min_value %.*g, max_value %.*g, sum_value %.*g\n",
        precision, (double)min_value,
        precision, (double)max_value,
        precision, (double)sum_value);
    }

    if (pmg->do_cumul && pmg->do_zeroize) {
      sum_value -= min_value * pmg->width;

      if (pmg->do_verbose) {
        FormatLocaleFile (stderr, "  revised sum_value %.*g\n",
          precision, (double)sum_value);
      }
    } else if (pmg->do_zeroize) {
      max_value -= min_value;
      if (pmg->do_verbose) {
        FormatLocaleFile (stderr, "  revised max_value %.*g\n",
          precision, (double)max_value);
      }
    }

    delta_v = pmg->do_zeroize ? min_value : 0.0;

    mult_fact = QuantumRange;

    if (pmg->do_norm && pmg->do_cumul) {
      mult_fact = QuantumRange / sum_value;
    } else if (pmg->do_norm) {
      mult_fact = QuantumRange / max_value;
    } else if (pmg->do_cumul) {
      mult_fact = QuantumRange / (double)pmg->width;
    }

    if (pmg->do_verbose) {
      FormatLocaleFile (stderr, "  mult_fact %.*g\n",
        precision, (double)mult_fact);
    }

    sum_value = 0.0;
    for (x=0; x < (ssize_t) new_image->columns; x++)
    {
      xf = x / (long double)(pmg->width-1);

      //if (fpow != 1.0) xf = pow (xf, fpow);
      //value = gnd_fact * exp (-(xf-fmean)*(xf-fmean)/two_sdsq) - delta_v;

      if (pmg->rev_skew) xf = 1.0 - xf;

//      if (fpow != 1.0) xf = pow (xf, fpow);
//      value = gnd_fact * exp (-(xf-fmean)*(xf-fmean)/two_sdsq) - delta_v;

//--      xf = (xf < fskew) ? xf * slopeA : (1-xf) * slopeB;
//--      value = gnd_fact * exp (-(xf-fmean)*(xf-fmean)/two_sdsq) - delta_v;

      switch (pmg->skew_meth) {
        case 1: default:
          if (fpow != 1.0) xf = pow (xf, fpow);
          value = gnd_fact * exp (-(xf-fmean)*(xf-fmean)/two_sdsq) - delta_v;
          break;
        case 2:
          xf = (xf < fskew) ? xf * slopeA : (1-xf) * slopeB;
          value = gnd_fact * exp (-(xf-fmean)*(xf-fmean)/two_sdsq) - delta_v;
          break;
      }

      if (pmg->do_cumul) {
        sum_value += value;
        pix_value = sum_value * mult_fact ADD_HALF ;
      } else {
        pix_value = value * mult_fact ADD_HALF ;
      }

      //FormatLocaleFile (stdout,
      //  "xf=%.*g value=%.*g sum_value=%.*g pix_value=%.*g\n",
      //  precision, (double)xf,
      //  precision, (double)value,
      //  precision, (double)sum_value,
      //  precision, (double)pix_value);

      SET_PIXEL_RED   (new_image, pix_value, q_out);
      SET_PIXEL_GREEN (new_image, pix_value, q_out);
      SET_PIXEL_BLUE  (new_image, pix_value, q_out);

      q_out += Inc_ViewPixPtr (new_image);
    }

    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }
  }

  out_view=DestroyCacheView(out_view);

  return (new_image);
}


ModuleExport size_t mkgaussImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  mkgaussT
    mkgauss;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &mkgauss);
  if (status == MagickFalse)
    return (-1);

  // Point to the last image in the list
  //
  image=GetLastImageInList(*images);
  if (image == (Image *) NULL) {
    fprintf (stderr, "mkgauss: no images in list\n");
    return (-1);
  }

  new_image = make_gauss(image, &mkgauss, exception);
  if (new_image == (Image *) NULL)
    return(-1);

  // Add the new image to the end of the list:
  AppendImageToList(images,new_image);

  return(MagickImageFilterSignature);
}
