#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <math.h>
#include <ctype.h>
#include "vsn_defines.h"

// Latest update:
//    27 September 2015
//    1 February 2016: for v7.
//    3 April 2018 for v7.0.7-28
//    9 April 2018 added log option.
//    14 March 2020 Limit processing to 1, 2 or 3 channels.

typedef struct {
  int
    cap_numbuckets;
  int
    mult;
  MagickBooleanType
    do_alpha,
    do_log,
    do_cumul,
    do_norm,
    do_verbose;
} mkhistoT;

/* Maybe see transform.c TransformImages and TransformImage.
*/

#define DEFAULT_CAP 65536

static void usage (void)
{
  printf ("Usage: -process 'mkhisto [OPTION]...'\n");
  printf ("Make a histogram image.\n");
  printf ("\n");
  printf ("  b, capnumbuckets N  limit number of buckets to N\n");
  printf ("  m, multiply N       multiply counts by N\n");
  printf ("  r, regardalpha      pre-multiply count by alpha\n");
  printf ("  l, log              use logs of counts\n");
  printf ("  c, cumul            make cumulative histogram\n");
  printf ("  n, norm             normalise output so maximum count is quantum\n");
  printf ("  v, verbose          write text information to stdout\n");
  printf ("\n");
  printf ("Default capnumbuckets is %i.\n", DEFAULT_CAP);
  printf ("Default multiply is 1.\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  mkhistoT * mkh
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  mkh->cap_numbuckets = DEFAULT_CAP;
  mkh->mult = 1;
  mkh->do_norm =
    mkh->do_alpha =
    mkh->do_cumul =
    mkh->do_verbose =
    mkh->do_log =
      MagickFalse;

  //printf ("mkhisto argc=%i\n", argc);
  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];
    //printf ("Arg %i [%s]\n", i, pa);

    if (IsArg (pa, "n", "norm")==MagickTrue) {
      mkh->do_norm = MagickTrue;
    } else if (IsArg (pa, "r", "regardalpha")==MagickTrue) {
      mkh->do_alpha = MagickTrue;
    } else if (IsArg (pa, "l", "log")==MagickTrue) {
      mkh->do_log = MagickTrue;
    } else if (IsArg (pa, "c", "cumul")==MagickTrue) {
      mkh->do_cumul = MagickTrue;
    } else if (IsArg (pa, "b", "capnumbuckets")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i]))
      {
        fprintf (stderr, "Non-numeric argument to 'capnumbuckets'.\n");
        status = MagickFalse;
      }
      else mkh->cap_numbuckets = atoi (argv[i]);
    } else if (IsArg (pa, "m", "multiply")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i]))
      {
        fprintf (stderr, "Non-numeric argument to 'multiply'.\n");
        status = MagickFalse;
      }
      else mkh->mult = atoi (argv[i]);
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      mkh->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "mkhisto: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }
  if (mkh->cap_numbuckets < 1)
    status = MagickFalse;

  if (mkh->do_verbose) {
    fprintf (stderr,
     "mkhisto options: capnumbuckets %i  multiply %i  ",
      mkh->cap_numbuckets, mkh->mult);
    if (mkh->do_alpha)   fprintf (stderr, "regardalpha ");
    if (mkh->do_log)     fprintf (stderr, "log ");
    if (mkh->do_cumul)   fprintf (stderr, "cumul ");
    if (mkh->do_norm)    fprintf (stderr, "norm ");
    if (mkh->do_verbose) fprintf (stderr, "verbose ");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


static MagickBooleanType MultiplyColor (
  const Image *new_image,
  CacheView *out_view,
  long double mult_fact,
  ExceptionInfo *exception)
/* For all pixels, multiply RGB channels by a factor.
*/
{
  ssize_t
    y;

  MagickBooleanType
    status = MagickTrue;

  int numChannels = NumColChannels (new_image);

  if (mult_fact != 1.0) {
#if defined(MAGICKCORE_OPENMP_SUPPORT)
    #pragma omp parallel for schedule(static,4) shared(status)\
      MAGICK_THREADS(new_image,new_image,new_image->rows,1)
#endif

    for (y=0; y < (ssize_t) new_image->rows; y++)
    {
      register VIEW_PIX_PTR
        *q;

      register ssize_t
        x;

      if (status == MagickFalse)
        continue;

      q=GetCacheViewAuthenticPixels(out_view,0,y,new_image->columns,1,exception);
      if (q == (const VIEW_PIX_PTR *) NULL)
        status=MagickFalse;

      for (x=0; x < (ssize_t) new_image->columns; x++)
      {
#if defined(MAGICKCORE_HDRI_SUPPORT)
        SET_PIXEL_RED (new_image,
          GET_PIXEL_RED (new_image, q) * mult_fact,
          q);
        if (numChannels >= 2) SET_PIXEL_GREEN (new_image,
          GET_PIXEL_GREEN (new_image, q) * mult_fact,
          q);
        if (numChannels >= 3) SET_PIXEL_BLUE (new_image,
          GET_PIXEL_BLUE (new_image, q) * mult_fact,
          q);
#else
        SET_PIXEL_RED (new_image,
          GET_PIXEL_RED (new_image, q) * mult_fact + 0.5,
          q);
        if (numChannels >= 2) SET_PIXEL_GREEN (new_image,
          GET_PIXEL_GREEN (new_image, q) * mult_fact + 0.5,
          q);
        if (numChannels >= 3) SET_PIXEL_BLUE (new_image,
          GET_PIXEL_BLUE (new_image, q) * mult_fact + 0.5,
          q);
#endif
        q += Inc_ViewPixPtr (new_image);
      }
      if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse)
        status=MagickFalse;
    }
  }
  return (status);
}


static MagickBooleanType CumulateColor (
  const Image *image,
  CacheView *out_view,
  long double mult_fact,
  ExceptionInfo *exception)
/* For all pixels, cumulate RGB channels.
   The accumulation is across the entire image, not each row separately.
*/
{
  ssize_t
    y;

  MagickBooleanType
    status = MagickTrue;

  int numChannels = NumColChannels (image);

  long double
    cumul_red, cumul_green, cumul_blue;

  cumul_red = cumul_green = cumul_blue = 0;

// I suppose multi-threading wouldn't work here.
//
//#if defined(MAGICKCORE_OPENMP_SUPPORT)
//    #pragma omp parallel for schedule(static,4) shared(status)
//      MAGICK_THREADS(image,image,image->rows,1)
//#endif

  for (y=0; y < (ssize_t) image->rows; y++)
  {
    register VIEW_PIX_PTR
      *q;

    register ssize_t
      x;

    if (status == MagickFalse)
      continue;

    q=GetCacheViewAuthenticPixels(out_view,0,y,image->columns,1,exception);
    if (q == (const VIEW_PIX_PTR *) NULL)
      status=MagickFalse;

    for (x=0; x < (ssize_t) image->columns; x++)
    {
      cumul_red   += GET_PIXEL_RED   (image, q);
      cumul_green += GET_PIXEL_GREEN (image, q);
      cumul_blue  += GET_PIXEL_BLUE  (image, q);

      SET_PIXEL_RED   (image, mult_fact * cumul_red,   q);
      if (numChannels >= 2) SET_PIXEL_GREEN (image, mult_fact * cumul_green, q);
      if (numChannels >= 3) SET_PIXEL_BLUE  (image, mult_fact * cumul_blue,  q);

      q += Inc_ViewPixPtr (image);
    }
    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse)
      status=MagickFalse;
  }
  return (status);
}



#if defined(MAGICKCORE_HDRI_SUPPORT)
#define ADD_HALF
#else
#define ADD_HALF +0.5
#endif


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *mkhImage(const Image *image,
  mkhistoT * pmkh,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  CacheView
    *image_view,
    *out_view;

  ssize_t
    y,
    NumBuckets,
    outXmult;

  MagickBooleanType
    status;

  VIEW_PIX_PTR
    *q_out;

  MagickRealType
    BucketSize;

  long double
    Increment,
    value,
    min_value,
    max_value,
    sum_red,
    sum_green,
    sum_blue;

  int
    precision;

  int numChannels = NumColChannels (image);

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (pmkh->do_verbose) {
    fprintf (stderr, "mkhisto: Input image [%s] depth is %i\n",
             image->filename, (int)image->depth);
  }

  precision = GetMagickPrecision();

  NumBuckets = exp2 (image->depth);
  if (NumBuckets > pmkh->cap_numbuckets) NumBuckets = pmkh->cap_numbuckets;
  Increment = pmkh->mult;

  BucketSize = exp2 (MAGICKCORE_QUANTUM_DEPTH) /(MagickRealType)NumBuckets;

  if (pmkh->do_verbose) {
    fprintf (stderr, "  NumBuckets %i  BucketSize %.*g\n",
             (int)NumBuckets, precision, (double)BucketSize);
  }

  new_image=CloneImage(image, NumBuckets, 1, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  // FIXME: set new_image depth to Quantum?
  // new_image->depth=MAGICKCORE_QUANTUM_DEPTH;

  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  SetAllBlack (new_image, exception);

#if IMV6OR7==6
  outXmult = 1;
#else
  outXmult = GetPixelChannels (new_image);
#endif

  out_view=AcquireAuthenticCacheView(new_image,exception);

  status=MagickTrue;

  q_out=GetCacheViewAuthenticPixels(out_view,0,0,new_image->columns,1,exception);
  if (q_out == (const VIEW_PIX_PTR *) NULL) {
    fprintf (stderr, "mkhisto: bad GetCacheViewAuthenticPixels out_view\n");
    status=MagickFalse;
  }

  image_view=AcquireVirtualCacheView(image,exception);

  sum_red = sum_green = sum_blue = 0.0;
  max_value = -QuantumRange;
  min_value = QuantumRange + 1.0;

/* I'm unsure about the multi-threading.
   What if two threads want to update the same histogram bucket,
   and sum_red etc?
*/

//#if defined(MAGICKCORE_OPENMP_SUPPORT)
//  #pragma omp parallel for schedule(static,4) shared(status)
//    MAGICK_THREADS(image,image,image->rows,1)
//#endif
  for (y=0; y < (ssize_t) image->rows; y++)
  {
    register const VIEW_PIX_PTR
      *p;

    ssize_t
      x,
      out_x;

    if (status == MagickFalse)
      continue;
    p=GetCacheViewVirtualPixels(image_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL)
    {
      fprintf (stderr, "bad GetCacheViewAuthenticPixels image_view\n");
      status=MagickFalse;
      continue;
    }

    for (x=0; x < (ssize_t) image->columns; x++)
    {

      if (pmkh->do_alpha) {
        Increment = 
          (pmkh->mult * GET_PIXEL_ALPHA(image,p) / (long double)QuantumRange) ADD_HALF ;
      }

      out_x = (GET_PIXEL_RED(image,p) / BucketSize);
      if (out_x >= NumBuckets) out_x = NumBuckets-1;
      else if (out_x < 0) out_x = 0;
      out_x *= outXmult;
      value = GET_PIXEL_RED(image,q_out+out_x)+Increment;
      if (min_value > value) min_value = value;
      if (max_value < value) max_value = value;
      sum_red += Increment;
      SET_PIXEL_RED (new_image, value, q_out+out_x);

      if (numChannels >= 2) {
        out_x = (GET_PIXEL_GREEN(image,p) / BucketSize);
        if (out_x >= NumBuckets) out_x = NumBuckets-1;
        else if (out_x < 0) out_x = 0;
        out_x *= outXmult;
        value = GET_PIXEL_GREEN(image,q_out+out_x)+Increment;
        if (min_value > value) min_value = value;
        if (max_value < value) max_value = value;
        sum_green += Increment;
        SET_PIXEL_GREEN (new_image, value, q_out+out_x);
      }

      if (numChannels >= 3) {
        out_x = (GET_PIXEL_BLUE(image,p) / BucketSize);
        if (out_x >= NumBuckets) out_x = NumBuckets-1;
        else if (out_x < 0) out_x = 0;
        out_x *= outXmult;
        value = GET_PIXEL_BLUE(image,q_out+out_x)+Increment;
        if (min_value > value) min_value = value;
        if (max_value < value) max_value = value;
        sum_blue += Increment;
        SET_PIXEL_BLUE (new_image, value, q_out+out_x);
      }

      p += Inc_ViewPixPtr (image);
    }

    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }
  }

  if (pmkh->do_log) {
    ssize_t
      x;

    max_value = -QuantumRange;
    min_value = QuantumRange + 1.0;

    VIEW_PIX_PTR *ql = q_out;

    for (x=0; x < (ssize_t) new_image->columns; x++)
    {
      value = GET_PIXEL_RED(new_image, ql);
      value = (value <= 1) ? 0 : log(value);
      if (min_value > value) min_value = value;
      if (max_value < value) max_value = value;
      SET_PIXEL_RED (new_image, value, ql);

      if (numChannels >= 2)  {
        value = GET_PIXEL_GREEN(new_image, ql);
        value = (value <= 1) ? 0 : log(value);
        if (min_value > value) min_value = value;
        if (max_value < value) max_value = value;
        SET_PIXEL_GREEN (new_image, value, ql);
      }

      if (numChannels >= 3) {
        value = GET_PIXEL_BLUE(new_image, ql);
        value = (value <= 1) ? 0 : log(value);
        if (min_value > value) min_value = value;
        if (max_value < value) max_value = value;
        SET_PIXEL_BLUE (new_image, value, ql);
      }

      ql += Inc_ViewPixPtr (new_image);
    }
    sum_red   = (sum_red   <= 1) ? 0 : log(sum_red);
    sum_green = (sum_green <= 1) ? 0 : log(sum_green);
    sum_blue  = (sum_blue  <= 1) ? 0 : log(sum_blue);

    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse) {
      fprintf (stderr, "bad sync\n");
      status=MagickFalse;
    }
  }

  if (pmkh->do_verbose) {
    fprintf (stderr, "  counts: min_value %.*g, max_value %.*g\n",
            precision, (double)min_value,
            precision, (double)max_value);
    fprintf (stderr, "  sum_red %.*g, sum_green %.*g, sum_blue %.*g\n",
            precision, (double)sum_red,
            precision, (double)sum_green,
            precision, (double)sum_blue);
  }

#if !defined(MAGICKCORE_HDRI_SUPPORT)
  if (max_value >= (long double)QuantumRange) {
    fprintf (stderr, "Warning: max_value >= QuantumRange (%.*g >=  %.*g)\n",
             (double)max_value, (double)QuantumRange);
  }
#endif

  // FIXME: but we don't care about min_value??

  // As required: cumulate, normalise or both.
  //
  if (pmkh->do_cumul==MagickTrue && pmkh->do_norm==MagickTrue) {
    long double
      max_sum,
      mult_fact;

    max_sum = sum_red;
    if (max_sum < sum_green) max_sum = sum_green;
    if (max_sum < sum_blue)  max_sum = sum_blue;

    if (max_sum != 0.0) {
      mult_fact = QuantumRange / max_sum;
      if (pmkh->do_verbose) {
        fprintf (stderr, "  Cumulating and normalising...\n");
        fprintf (stderr, "    max_sum=%.*g mult_fact=%.*g\n",
                precision, (double)max_sum,
                precision, (double)mult_fact);
      }
      status = CumulateColor (new_image, out_view, mult_fact, exception);
    }
  } else if (pmkh->do_norm==MagickTrue) {
    long double
      mult_fact;

    if (pmkh->do_verbose) {
      fprintf (stderr, "  Normalising...\n");
    }
    if (max_value != 0.0) {
      mult_fact = QuantumRange / max_value;
      if (pmkh->do_verbose) {
        fprintf (stderr, "    max_value=%.*g  mult_fact=%.*g\n",
                precision, (double)max_value,
                precision, (double)mult_fact);
      }
      MultiplyColor (new_image, out_view, mult_fact, exception);
    }
  } else if (pmkh->do_cumul==MagickTrue) {
    if (pmkh->do_verbose) {
      fprintf (stderr, "  Cumulating...\n");
    }
    status = CumulateColor (new_image, out_view, 1.0, exception);
  }

  image_view=DestroyCacheView(image_view);
  out_view=DestroyCacheView(out_view);

  return (new_image);
}


ModuleExport size_t mkhistoImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  mkhistoT
    mkh;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &mkh);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = mkhImage (image, &mkh, exception);

    ReplaceImageInList(&image,new_image);
    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
