rem Given image %1 is black mainland around white "sea",
rem finds floating-point coordinates (ncX,ncY) of point on the mainland
rem    nearest to given coordinate (%2,%3).
rem If the given coord is in the white "sea", the nearest point will be on the black "coast".
@rem
@rem   %2, %3 coords of central point  [default center of %1]
@rem        Each coord may be suffixed with "%", for percentage of width or height.
@rem        Default: centre of image.
@rem   %4 proportion towards coast point (0 to 100) [100]
@rem   %5 output debugging image [default NULL:, no output]
@rem
@rem Also uses:
@rem   ncSUP_SAMP Factor for supersampling [default 1].
@rem   ncPREFIX prefix for temporary files
@rem   ncSEA_COL sea color: black, white or auto. Default white.
@rem     auto will choose automatically, which takes time.
@rem   nc?? automatic choice for negate
@rem   nc?? if 1, assumes %1 is already supersampled and unrolled.
@rem
@rem Returns data about nearest point:
@rem   ncRAD   distance from given coord. 0 means given coord wasn't sea.
@rem   ncTHETA bearing from given coord to found point, anti-clockwise from north.
@rem   ncCST_X } x,y coord of found point.
@rem   ncCST_Y }
@rem
@rem Updated:
@rem   16-July-2016
@rem   24-August-2022 for IM v7.
@rem


@if "%1"=="" findstr /B "rem @rem" %~f0 & exit /B 1

@setlocal enabledelayedexpansion

@call echoOffSave

call %PICTBAT%setInOut %1 nc


set CENT_X=%2
set CENT_Y=%3

if "%CENT_X%"=="." set CENT_X=
if "%CENT_Y%"=="." set CENT_Y=

set BLANK_CENT=0
if "%CENT_X%"=="" if "%CENT_Y%"=="" set BLANK_CENT=1

if "%CENT_X%"=="" set CENT_X=50%%
if "%CENT_Y%"=="" set CENT_Y=50%%

set X_SUFFIX=%CENT_X:~-1%
if "%X_SUFFIX%"=="%%" (
  set nX=%CENT_X:~0,-1%
  set X_MULT=^(w-1^)/100
) else (
  set nX=%CENT_X%
  set X_MULT=1
)

set Y_SUFFIX=%CENT_Y:~-1%
if "%Y_SUFFIX%"=="%%" (
  set nY=%CENT_Y:~0,-1%
  set Y_MULT=^(h-1^)/100
) else (
  set nY=%CENT_Y%
  set Y_MULT=1
)

set PROP=%4
if "%PROP%"=="." set PROP=
if "%PROP%"=="" set PROP=100

set OUTFILE=%5
if "%OUTFILE%"=="." set OUTFILE=
if "%OUTFILE%"=="" set OUTFILE=NULL:

set SUP_SAMP=%ncSUP_SAMP%
if "%SUP_SAMP%"=="." set SUP_SAMP=
if "%SUP_SAMP%"=="" set SUP_SAMP=1

for /F "usebackq" %%L in (`%IMG7%magick identify ^
  -format "WW=%%w\nHH=%%h\nnX=%%[fx:%nX%*%X_MULT%]\nnY=%%[fx:%nY%*%Y_MULT%]"
  %INFILE%`) do set %%L

for /F "usebackq" %%L in (`%IMG7%magick identify ^
  -format "W_SS=%%[fx:%WW%*%SUP_SAMP%]\nH_SS=%%[fx:%HH%*%SUP_SAMP%]\nnXS=%%[fx:%nX%*%SUP_SAMP%]\nnYS=%%[fx:%nY%*%SUP_SAMP%]"
  xc:`) do set %%L

rem echo %0: WW=%WW%  HH=%HH%  W_SS=%W_SS%  H_SS=%H_SS%  nX=%nX%  nY=%nY%  nXS=%nXS%  nYS=%nYS%

if %BLANK_CENT%==1 (
  set sUNROLL=-distort depolar "-1,0"
) else (
  set sUNROLL=-distort depolar "-1,0,%nXS%,%nYS%"
)

if %SUP_SAMP%==1 (
  set sSUP_UP=
) else (
  set sSUP_UP=-resize "%W_SS%x%H_SS%^^^!"
)

if "%ncSEA_COL%"=="" set ncSEA_COL=white

if "%ncSEA_COL%"=="white" (
  set sNEG=-negate
) else if "%ncSEA_COL%"=="black" (
  set sNEG=
) else if "%ncSEA_COL%"=="auto" (
  for /F "usebackq" %%L in (`%IMG7%magick identify ^
    -format "doNEG=%%[fx:p{0,0}.intensity>0.5?1:0]" ^
    %INFILE%`) do set %%L

  if "!doNeg!"=="1" (
    set sNEG=-negate
  ) else (
    set sNEG=
  )
) else (
  echo %0: bad ncSEA_COL "%ncSEA_COL%"
  exit /B 1
)

for /F "usebackq tokens=1-3 delims=:, " %%W in (`%IM7DEV%magick ^
  %INFILE% ^
  %sNEG% ^
  %sSUP_UP% ^
  -virtual-pixel Black ^
  %sUNROLL% ^
  -threshold 50%% ^
  -process onewhite ^
  NULL: 2^>^&1`) do (
  set MAGIC=%%W
  set X=%%X
  set Y=%%Y
)

if not "%magic%"=="onewhite" exit /B 1

for /F "usebackq" %%L in (`%IMG7%magick identify ^
  -format "dx=%%[fx:%nXS%>%W_SS%/2?%nXS%:%W_SS%-%nXS%]\ndy=%%[fx:%nYS%>%H_SS%/2?%nYS%:%H_SS%-%nYS%]"
  %INFILE%`) do set %%L

for /F "usebackq" %%L in (`%IMG7%magick identify ^
  -format "RAD=%%[fx:%Y%/%SUP_SAMP%*hypot(%dx%,%dy%)/%H_SS%]\nTHETA=%%[fx:%X%*2*pi/%W_SS%]"
  xc:`) do set %%L

for /F "usebackq" %%L in (`%IMG7%magick identify ^
  -format "CST_X=%%[fx:%nX%-%RAD%*sin(%THETA%)]\nCST_Y=%%[fx:%nY%-%RAD%*cos(%THETA%)]"
  xc:`) do set %%L

rem echo %0: RAD=%RAD% THETA=%THETA% CST_X=%CST_X% CST_Y=%CST_Y%

if /I not "%OUTFILE%"=="NULL:" (

  if "%sNEG%"=="" (
    set DBG_COL=yellow
  ) else (
    set DBG_COL=red
  )

  %IMG7%magick ^
    %INFILE% ^
    -fill None -stroke !DBG_COL! ^
    -draw "line   %nX%,%nY% %CST_X%,%CST_Y%" ^
    -draw "circle %nX%,%nY% %CST_X%,%CST_Y%" ^
    -alpha off ^
    %OUTFILE%
)

call echoRestore

@endlocal & set ncOUTFILE=%OUTFILE%& set ncRAD=%RAD%& set ncTHETA=%THETA%& set ncCST_CX=%nX%& set ncCST_CY=%nY%& set ncCST_X=%CST_X%& set ncCST_Y=%CST_Y%
