#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"


typedef enum {
  ctPixel,
  ctPercent,
  ctProportion
} CoordTypeT;

typedef struct {
  double 
    incx, incy;

  CoordTypeT
    ctx, cty;

  MagickBooleanType
    do_verbose;
} nearestWhiteT;


static void usage (void)
{
  printf ("Usage: -process 'nearestwhite [OPTION]...'\n");
  printf ("Finds the white pixel nearest to the centre.\n");
  printf ("\n");
  printf ("  cx N                     centre x-coordinate\n");
  printf ("  cy N                     centre y-coordinate\n");
  printf ("  v,  verbose              write text information\n");
  printf ("\n");
}

static CoordTypeT CoordType (const char * s)
{
  char c = *(s+strlen(s)-1);

  if (c=='%' || c == 'c' || c == 'C') return ctPercent;
  if (c=='p' || c == 'P') return ctProportion;
  return ctPixel;
}

static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu (
  const int argc,
  const char **argv,
  nearestWhiteT * nw
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  nw->incx = nw->incy = 50;
  nw->ctx = nw->cty = ctPercent;

  nw->do_verbose = MagickFalse;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "cx", "cx")==MagickTrue) {
      i++;
      if (i >= argc) {
        fprintf (stderr, "nearestwhite: 'cx' needs an argument\n");
        status = MagickFalse;
        break;
      }
      nw->incx = atof(argv[i]);
      nw->ctx = CoordType (argv[i]);

    } else if (IsArg (pa, "cy", "cy")==MagickTrue) {
      i++;
      if (i >= argc) {
        fprintf (stderr, "nearestwhite: 'cy' needs an argument\n");
        status = MagickFalse;
        break;
      }
      nw->incy = atof(argv[i]);
      nw->cty = CoordType (argv[i]);

    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      nw->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "nearestwhite: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (nw->do_verbose) {
    fprintf (stderr, "nearestwhite options:");
    fprintf (stderr, "  cx %g", nw->incx);
    if (nw->ctx == ctPercent) fprintf (stderr, "%%");
    else if (nw->ctx == ctProportion) fprintf (stderr, "p");
    fprintf (stderr, "  cy %g", nw->incy);
    if (nw->cty == ctPercent) fprintf (stderr, "%%");
    else if (nw->cty == ctProportion) fprintf (stderr, "p");

    if (nw->do_verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}




// The next function corresponds in style to functions in enhance.c
// It takes one image, and returns a status.
//
// Note: Image *image is no longer const, for SetImageProperty().
//
static MagickBooleanType nearestwhite(Image *image,
  nearestWhiteT * nw,
  ExceptionInfo *exception)
{
  MagickBooleanType
    status,
    found;

  CacheView
    *image_view;

  ssize_t
    y,
    nearX, nearY;

  double cx=0, cy=0, nearDistSq = 0;

  char
    text[MaxTextExtent];

  assert(image != (Image *) NULL);
  assert(image->signature == MAGICK_CORE_SIG);
  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  int precision = GetMagickPrecision();

  image_view = AcquireVirtualCacheView(image,exception);

  status = MagickTrue;
  found = MagickFalse;

  switch (nw->ctx) {
    case ctPixel:      cx = nw->incx; break;
    case ctPercent:    cx = nw->incx * (image->columns - 1) / 100.0; break;
    case ctProportion: cx = nw->incx * (image->columns - 1); break;
    default: status=MagickFalse;
  }

  switch (nw->cty) {
    case ctPixel:      cy = nw->incy; break;
    case ctPercent:    cy = nw->incy * (image->rows - 1) / 100.0; break;
    case ctProportion: cy = nw->incy * (image->rows - 1); break;
    default: status=MagickFalse;
  }

  if (nw->do_verbose) {
    fprintf (stderr, "nearestwhite.cx = %.*g\n", precision, cx);
    fprintf (stderr, "nearestwhite.cy = %.*g\n", precision, cy);

    //fprintf (stderr, "(fuzz = %.*g)\n", precision, image->fuzz);
  }

#define LIMIT (QuantumRange - image->fuzz)

  for (y=0; y < (ssize_t) image->rows; y++)
  {
    VIEW_PIX_PTR const
      *p;

    register ssize_t
      x;

    double
      dx, dy;

    if (status == MagickFalse)
      continue;

    p=GetCacheViewAuthenticPixels(image_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL)
      status=MagickFalse;

    for (x=0; x < (ssize_t) image->columns; x++)
    {
      if (  GET_PIXEL_RED(image,p)   >= LIMIT
         && GET_PIXEL_GREEN(image,p) >= LIMIT
         && GET_PIXEL_BLUE(image,p)  >= LIMIT)
      {
        if (found) {
          dx = cx - x;
          dy = cy - y;
          double DistSq = dx*dx + dy*dy;
          if (nearDistSq > DistSq) {
            nearDistSq = DistSq;
            nearX = x;
            nearY = y;
          }
        } else {
          found = MagickTrue;
          dx = cx - x;
          dy = cy - y;
          nearDistSq = dx*dx + dy*dy;
          nearX = x;
          nearY = y;
        }
      }

      p += Inc_ViewPixPtr (image);
    }
  }

  if (found == MagickFalse) {
    fprintf (stderr, "nearestwhite: none\n");
    DeleteImageProperty (image, "filter:nearestwhite"); // FIXME: Or set to "none"?
  } else {
    sprintf (text, "%lu,%lu", nearX, nearY);
    fprintf (stderr, "nearestwhite: %s\n", text);

#if IMV6OR7==6
    SetImageProperty (image, "filter:nearestwhite", text);
#else
    SetImageProperty (image, "filter:nearestwhite", text, exception);
#endif
  }
  return MagickTrue;
}


ModuleExport size_t nearestwhiteImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image;

  MagickBooleanType
    status;

  nearestWhiteT nw;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &nw);

  if (status == MagickTrue) {
    for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
    {
      status = nearestwhite(image, &nw, exception);

      if (status == MagickFalse)
        continue;
    }
  }

  assert (status == MagickTrue);

  return(MagickImageFilterSignature);
}
