#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include "vsn_defines.h"

#if IMV6OR7==6
#include <magick/MagickCore.h>
#elif IMV6OR7==7
#include <MagickCore/MagickCore.h>
#else
#error IMV6OR7 defined but not valid
#endif

#ifdef _OPENMP
#include <omp.h>
#define OPENMP_SUPPORT 1
#endif



typedef struct {
  int  version;
  int  verbose;
  char * infile;
  char * outfile;
  int  numToFind;
  char * sRadius;
} ProgOptT;


static void version (void)
{
  printf ("nlightest v1.0 (c) 2017 Alan Gibson 5-May-2017\n");
  printf ("%s\n", GetMagickVersion(NULL));

  printf ("Depth Q%i ", MAGICKCORE_QUANTUM_DEPTH);

#if defined(MAGICKCORE_HDRI_SUPPORT)
  printf ("HDRI\n");
#else
  printf ("integer\n");
#endif

#ifdef _OPENMP
  printf ("Compiled by an OpenMP-compliant implementation.\n");
#else
  printf ("No OpenMP.\n");
#endif

#if defined(OPENMP_SUPPORT)
  printf ("OpenMP support.\n");
#else
  printf ("No OpenMP support.\n");
#endif

  exit (0);
}

static void usage (void)
{
  printf ("Writes coordinates of (n) lightest pixels to stdout.\n\n");

  printf ("Usage: nlightest {options}\n");
  printf ("  -h  --help       write usage data and exit\n");
  printf ("  -v  --verbose    write text information about the processing to stderr\n");
  printf ("  -i  --infile     filename for input\n");
  printf ("  -o  --outfile    filename for output\n");
  printf ("  -n  --numToFind  number to find [default 10]\n");
  printf ("  -r  --radius     radius to be ignored; can be suffixed %% or c or p\n");
  printf ("      --version    write version data and exit\n");
  exit (0);
}

static void WrOptions (ProgOptT *po)
{
  fprintf (stderr, "Options:\n");
  if (po->verbose) fprintf (stderr, "  --verbose\n");
  if (po->infile)  fprintf (stderr, "  --infile %s\n", po->infile);
  if (po->outfile) fprintf (stderr, "  --outfile %s\n", po->outfile);
  fprintf (stderr, "  --numToFind %i\n", po->numToFind);
  fprintf (stderr, "  --radius %s\n", po->sRadius);
}

static void GetOptions (int argc, char **argv, ProgOptT * po)
// If there is a problem, this doesn't return.
{
  struct option long_options[] = {
    /* Second field is one of: no_argument, required_argument, optional_argument. */
    /* Third field is 0 or pointer to int. */
    /* Fourth field is:
         an integer, if third is a pointer to int;
         a character index, otherwise.
    */

    /* These options set a flag. */
    {"version", no_argument,       &po->version, 1},
    {"verbose", no_argument,       &po->verbose, 1},
    {"brief",   no_argument,       &po->verbose, 0},
    /* These options don't set a flag.
       We distinguish them by their indices. */
    {"help",    no_argument,       0, 'h'},
    {"infile",  required_argument, 0, 'i'},
    {"outfile", required_argument, 0, 'o'},
    {"numToFind", required_argument, 0, 'n'},
    {"radius", required_argument, 0, 'r'},
    {0, 0, 0, 0}
  };

  char * short_options = "hvi:o:n:r:";

  MagickBooleanType
    okay = MagickTrue;

  po->version = po->verbose = 0;
  po->infile = po->outfile = NULL;
  po->numToFind = 10;
  po->sRadius = "";

  int c;

  while (1)
  {
    int option_index = 0;

    c = getopt_long (argc, argv, short_options,
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1) break;

//    printf ("char [%c]\n", c);

    switch (c)
    {
      case 0:
//        printf ("long option with option_index `%i'\n", option_index);
        /* If this option sets a flag, do nothing else now. */
        if (long_options[option_index].flag != 0)
          break;
        //printf ("option %s", long_options[option_index].name);
        if (optarg)
          printf (" with arg %s", optarg);
        printf ("\n");
        break;

      case 'h':
        usage ();
        break;

      case 'v':
        po->verbose = 1;
        break;

      case 'i':
        po->infile = optarg;
        break;

      case 'o':
        po->outfile = optarg;
        break;

      case 'n':
        po->numToFind = atoi(optarg);
        break;

      case 'r':
        po->sRadius = optarg;
        break;

      case '?':
        /* getopt_long already printed an error message. */
        break;

      default:
        abort ();
    }
  }

  if (po->version) {
    version ();
  }

  /* Print any remaining command line arguments (not options). */
  if (optind < argc)
  {
    printf ("non-option ARGV-elements: ");
    while (optind < argc)
      printf ("%s ", argv[optind++]);
    putchar ('\n');
    okay = MagickFalse;
  }

  if (po->verbose) {
    WrOptions (po);
  }

  if (!okay) exit (-1);
}


// ThrowWandException() doesn't return.

#define ThrowWandException(wand) \
{ \
  char \
    *description; \
 \
  ExceptionType \
    severity; \
 \
  description=MagickGetException(wand,&severity); \
  fprintf(stderr,"%s %s %lu %s\n",GetMagickModule(),description); \
  description=(char *) MagickRelinquishMemory(description); \
  exit(-1); \
}

#define snibw_threads(source,destination,chunk,expression) \
  num_threads((expression) == 0 ? 1 : \
    (((chunk) > (32*GetMagickResourceLimit(ThreadResource))) && \
     (GetImagePixelCacheType(source) != DiskCache)) && \
     (GetImagePixelCacheType(destination) != DiskCache) ? \
      GetMagickResourceLimit(ThreadResource) : \
      GetMagickResourceLimit(ThreadResource) < 2 ? 1 : 2)

#define snibw_threadsXX(source,destination,chunk,expression) \
  num_threads(GetMagickResourceLimit(ThreadResource))



typedef struct {
  ssize_t
    xAtMax,
    yAtMax;

  Quantum maxVal;
} ThreadT;


static MagickBooleanType FindLightest (ProgOptT *po)
{
  if (!po->infile) return MagickFalse;

  ExceptionInfo *exception = AcquireExceptionInfo ();

  ImageInfo *image_info = CloneImageInfo((ImageInfo *) NULL);
  (void) strcpy(image_info->filename,po->infile);
  Image * image = ReadImage(image_info,exception);
  if (!image) return MagickFalse;

  if (po->verbose) printf ("ThreadResource %llu\n", GetMagickResourceLimit(ThreadResource));

  if (!TransformImageColorspace(image, GRAYColorspace, exception))
    return MagickFalse;

  if (po->verbose) fprintf (stderr, "numToFind=%i\n", po->numToFind);

  ssize_t radius = 0;

  if (po->sRadius) {
    ssize_t minDim = image->columns;
    if (minDim > image->rows) {
      minDim = image->rows;
    }

    char * chLast;
    double val = InterpretLocaleValue (po->sRadius, &chLast);

    if (*chLast == '%') *chLast = 'c';

    if (*chLast == 'c') {
      radius = floor (val * minDim / 100.0 + 0.5);
    } else if (*chLast == 'p') {
      radius = floor (val * minDim + 0.5);
    } else {
      radius = floor (val + 0.5);
    }
    if (po->verbose) fprintf (stderr, "val=%g radius=%li\n", val, radius);
  }

  if (po->verbose) fprintf (stderr, "radius=%li\n", radius);

  DrawInfo *draw_info = AcquireDrawInfo();
  // Default fill colour is black.

  ssize_t
    y;

  MagickBooleanType okay = MagickTrue;

  int nFound = 0;

  int maxThr = GetMagickResourceLimit(ThreadResource);
  if (po->verbose) printf ("maxThr=%i\n", maxThr);
  ThreadT * threads = (ThreadT *)malloc (maxThr * sizeof(ThreadT));

  CacheView *image_view = AcquireVirtualCacheView (image,exception);;

  do {
    int i;
    for (i = 0; i < maxThr; i++) {
      threads[i].xAtMax = 0;
      threads[i].yAtMax = 0;
      threads[i].maxVal = 0;
    }

#if defined(OPENMP_SUPPORT)
    #pragma omp parallel for \
      schedule(static) \
      snibw_threads ( \
        image, \
        image, \
        image->rows,1)
#endif

    for (y=0; y < (ssize_t) image->rows; y++)
    {
      int thrNum = omp_get_thread_num();

      const VIEW_PIX_PTR *p = GetCacheViewVirtualPixels (
        image_view,0,y,image->columns,1,exception);
      if (!p) {
        okay = MagickFalse;
        continue;
      }

      ThreadT * pt = &threads[thrNum];
      ssize_t x;
      for (x=0; x < image->columns; x++)
      {
        double r = GET_PIXEL_RED (image, p);
        if (pt->maxVal < r) {
          pt->maxVal = r;
          pt->xAtMax = x;
          pt->yAtMax = y;
          //printf ("  y=%li  %g %li,%li\n", y, r, x, y);
        }
        p += Inc_ViewPixPtr (image);
      }
    }  // end y loop

    if (!okay) break;

    ssize_t
      yAtMax=0, xAtMax=0;

    Quantum maxVal = 0;

    for (i = 0; i < maxThr; i++) {
      // printf ("thr %i: %g @ %li,%li\n", i, threads[i].maxVal, threads[i].xAtMax, threads[i].yAtMax);
      if (maxVal < threads[i].maxVal) {
        maxVal = threads[i].maxVal;
        xAtMax = threads[i].xAtMax;
        yAtMax = threads[i].yAtMax;
      }
    }

    if (maxVal==0) break;

    if (po->verbose) fprintf (stderr,
      "max val=%g @ %li,%li\n", maxVal, xAtMax, yAtMax);
    printf ("%g @ %li,%li\n", maxVal/(double)QuantumRange, xAtMax, yAtMax);

    char primitive[MaxTextExtent];

    if (radius < 1) {
      FormatLocaleString(primitive,MaxTextExtent,
        "point %li,%li", xAtMax, yAtMax);
    } else {
      FormatLocaleString(primitive,MaxTextExtent,
        "circle %li,%li %li,%li", xAtMax, yAtMax, xAtMax, (long int)(yAtMax+radius+0.5));
    }
    CloneString(&draw_info->primitive,primitive);

#if IMV6OR7==6
    if (!DrawImage (image, draw_info)) break;
#else
    if (!DrawImage (image, draw_info, exception)) break;
#endif

    nFound++;

  } while (po->numToFind==0 || nFound < po->numToFind);

  if (po->verbose) fprintf (stderr, "Number found: %i\n", nFound);


  free (threads);

  if (po->outfile) {
    strcpy (image->filename, po->outfile);
    if (!WriteImage (image_info, image, exception)) {
      fprintf (stderr, "WriteImage [%s] failed.\n", po->outfile);
    }
  }

  image_view = DestroyCacheView (image_view);
  image = DestroyImage (image);
  image_info = DestroyImageInfo (image_info);
  exception = DestroyExceptionInfo (exception);

  return MagickTrue;
}


int main (int argc,char **argv)
{
  int ret = 0;

  ProgOptT ProgOpt;

  MagickCoreGenesis (*argv, MagickTrue);

  GetOptions (argc, argv, &ProgOpt);

  int i;
  for (i=0; i < 10; i++) {
  if (!FindLightest (&ProgOpt)) {
    fprintf (stderr, "FindLightest failure\n");
    ret = 1;
  }
  }

  MagickCoreTerminus ();

  return (ret);
}
