/*
    Ref: http://www.openmp.org/wp-content/uploads/omp-hands-on-SC08.pdf

    Build with:
      gcc -fopenmp -o omptest.exe omptest.c
*/

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <omp.h>


static void WrOmpData (void)
{
  printf ("OMP data:\n");
  printf ("  thread_limit %i\n", omp_get_thread_limit());
  printf ("  max_threads %i\n", omp_get_max_threads());
  printf ("  num_threads %i\n", omp_get_num_threads());
  printf ("  thread_num %i\n", omp_get_thread_num());
  printf ("  num_procs %i\n", omp_get_num_procs());
  printf ("  in_parallel %i\n", omp_in_parallel());
  printf ("  dynamic %i\n", omp_get_dynamic());
  printf ("  nested %i\n", omp_get_nested());
  printf ("  wtime %g\n", omp_get_wtime());
  printf ("  wtick %g\n", omp_get_wtick());
}

static void Test1 (void)
{
  printf ("------------------------------------------\n");
  printf ("Test1: omp parallel\n");

  #pragma omp parallel
  {
    printf ("  Hello from thread num %i of %i\n",
            omp_get_thread_num(),
            omp_get_num_threads()
           );

#pragma omp critical
    WrOmpData ();

    printf ("  Goodbye from thread num %i of %i\n",
            omp_get_thread_num(),
            omp_get_num_threads()
           );
  }
}


static void Test2 (int n)
{
  printf ("------------------------------------------\n");
  printf ("Test2: omp parallel num_threads(%i)\n", n);

  #pragma omp parallel num_threads(n)
  {
    printf ("  Thread num %i of %i\n",
            omp_get_thread_num(),
            omp_get_num_threads()
           );
    printf ("  Thread num %i end.\n",
            omp_get_thread_num()
           );
  }

}


static void Test3 (int n)
{
  printf ("------------------------------------------\n");
  printf ("Test3: Sum a series of numbers, zero to %i\n", n-1);
  printf ("  Total should be %i, but it will be wrong.\n", n * (n-1) / 2);

  int i, total=0;

#pragma omp parallel for
  for (i = 0; i < n; i++) {
    int tmp = total;
    printf (" %i ", omp_get_thread_num());
    double s = sin(i);
    total = tmp + i;
  }

  printf ("\ntotal: %i\n", total);
}


static void Test3b (int n)
{
  printf ("------------------------------------------\n");
  printf ("Test3b: Sum a series of numbers, zero to %i\n", n-1);
  printf ("  Total should be %i, but it could be wrong.\n", n * (n-1) / 2);

  int i, total=0;

#pragma omp parallel for
  for (i = 0; i < n; i++) {
    printf (" %i ", omp_get_thread_num());
    double s = sin(i);
    total = total + i;
  }

  printf ("\ntotal: %i\n", total);
}


static void Test3c (int n)
{
  printf ("------------------------------------------\n");
  printf ("Test3c: Sum a series of numbers, zero to %i\n", n-1);
  printf ("  Total should be %i, and should be correct.\n", n * (n-1) / 2);

  int i, total=0;

#pragma omp parallel for
  for (i = 0; i < n; i++) {
    printf (" %i ", omp_get_thread_num());
    double s = sin(i);
#pragma omp atomic
    total = total + i;
  }

  printf ("\ntotal: %i\n", total);
}


static void Test4 (int *pa, int width, int height)
{
  printf ("------------------------------------------\n");
  printf ("Test4: Initialise array\n");

  int y, x, t;
  for (y = 0; y < height; y++) {
    for (x = 0; x < width; x++) {
      pa[y*width+x] = x;
      t = 99;
    }
  }

}


int main (void)
{
  printf ("Hello wizards!\n");

# ifdef _OPENMP
  printf("Compiled by an OpenMP-compliant implementation.\n");
# endif

  WrOmpData ();

  Test1();

  printf ("Sequential part\n");

  Test2(4);

  Test2(12);

  int width = 70;
  int height = 46;

  int *my_array = (int *)malloc (width * height * sizeof(int));

  Test3 (100);

  Test3b (100);

  Test3c (100);

  Test4 (my_array, width, height);

  free (my_array);

  WrOmpData ();

  return 0;
}