/* Updated:

     11-December-2018 corrected typos in options.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <ctype.h>
#include <float.h>
#include "vsn_defines.h"

typedef enum {
  omLinear,
  omPower
} oogMethodT;

typedef struct {
  // Settings:
  MagickBooleanType
    verbose,
    doForceMin,
    doForceMax;

  oogMethodT
    oogMethod;

  double
    loLimit,
    hiLimit,
    forceMin,
    forceMax;

  int
    chanNum, // -1 for all channels together.
    precision;

  // Used internally:
  MagickBooleanType
    doLo,
    doHi;

  double
    min, max,
    a, b, c, d,
    A0, B0,
    A1, B1;
} oogboxT;

#define VERSION "oogbox v1.0  Copyright (c) 2017 Alan Gibson"

static void usage (void)
{
  printf ("Usage: -process 'oogbox [OPTION]...'\n");
  printf ("Puts out-of-gamut values back in the box.\n");
  printf ("\n");
  printf ("  m, method string        Linear or Power [Power]\n");
  printf ("  l, loLimit number       low limit for shadow processing [0.1]\n");
  printf ("  h, hiLimit number       high limit for highlight processing [0.9]\n");
  printf ("  fmn, forceMin number    force minimum value (override calculation)\n");
  printf ("  fmx, forceMax number    force maximum value (override calculation)\n");
  printf ("  c, channel integer      channel number, or -1 for all [-1]\n");
  printf ("  v, verbose              write text information to stdout\n");
  printf ("     version              write version information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  // LocaleCompare is not case-sensitive,
  // so we use strcmp for the short option.

  if ((strcmp(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "oogbox: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  oogboxT * pob
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status = MagickTrue;

  pob->oogMethod = omPower;
  pob->loLimit = 0.1;
  pob->hiLimit = 0.9;
  pob->chanNum = -1;
  pob->doLo = pob->doHi = MagickFalse;
  pob->doForceMin = pob->doForceMax = MagickFalse;
  pob->forceMin = 0.0;
  pob->forceMax = 1.0;
  pob->a = pob->b = pob->c = pob->d = 0;
  pob->A0 = pob->B0 = 0;
  pob->A1 = pob->B1 = 0;

  pob->verbose = MagickFalse;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "m", "method")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "Linear")==0) pob->oogMethod = omLinear;
      else if (LocaleCompare(argv[i], "Power")==0) pob->oogMethod = omPower;
      else {
        fprintf (stderr, "oogbox: ERROR: unknown method [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "l", "loLimit")==MagickTrue) {
      NEXTARG;
      pob->loLimit = atof(argv[i]);
    } else if (IsArg (pa, "h", "hiLimit")==MagickTrue) {
      NEXTARG;
      pob->hiLimit = atof(argv[i]);
    } else if (IsArg (pa, "fmn", "forceMin")==MagickTrue) {
      NEXTARG;
      pob->forceMin = atof(argv[i]);
      pob->doForceMin = MagickTrue;
    } else if (IsArg (pa, "fmx", "forceMax")==MagickTrue) {
      NEXTARG;
      pob->forceMax = atof(argv[i]);
      pob->doForceMax = MagickTrue;
    } else if (IsArg (pa, "c", "channel")==MagickTrue) {
      NEXTARG;
      pob->chanNum = atoi(argv[i]);

    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pob->verbose = MagickTrue;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "oogbox: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (pob->verbose) {
    fprintf (stderr, "oogbox options: ");
    fprintf (stderr, " method ");
    switch (pob->oogMethod) {
      case omLinear: fprintf (stderr, "Linear"); break;
      case omPower:  fprintf (stderr, "Power");  break;
    }
    fprintf (stderr, " loLimit %.*g", pob->precision, pob->loLimit);
    fprintf (stderr, " hiLimit %.*g", pob->precision, pob->hiLimit);
    if (pob->doForceMin) fprintf (stderr, " forceMin %.*g", pob->precision, pob->forceMin);
    if (pob->doForceMax) fprintf (stderr, " forceMax %.*g", pob->precision, pob->forceMax);
    fprintf (stderr, " channel %i", pob->chanNum);
    if (pob->verbose)    fprintf (stderr, " verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


static double inline CalcLow (
    oogboxT * pob,
    double v)
{
  switch (pob->oogMethod) {
    case omLinear: return pob->a*v + pob->b;
    case omPower: {
      double dv = v - pob->min;
      if (dv < 0) return v;
      return pob->A0 * pow (v - pob->min, pob->B0);
    }
  }
  return 0; // Keep compiler happy.
}

static double inline CalcHigh (
    oogboxT * pob,
    double v)
{
  switch (pob->oogMethod) {
    case omLinear: return pob->c*v + pob->d;
    case omPower:  {
      double dv = pob->max - v;
      if (dv < 0) return v;
      return 1 - (pob->A1 * pow (dv, pob->B1));
    }
  }
  return 0; // Keep compiler happy.
}

// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *oogbox (
  Image *image,
  oogboxT * pob,
  ExceptionInfo *exception)
{
  // This ignores the alpha channel.

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (pob->verbose) {
    fprintf (stderr, "oogbox: Input image [%s] %ix%i depth is %i\n",
             image->filename,
             (int)image->columns, (int)image->rows,
             (int)image->depth);
  }

  pob->doLo = (pob->loLimit > 0.0);
  pob->doHi = (pob->hiLimit < 1.0);


  // Ensure we have non-palette, 3 channels.

  Image * new_img = CloneImage (image, 0, 0, MagickTrue, exception);
  if (!new_img) return NULL;

  if (!SetNoPalette (new_img, exception)) return NULL;

  // Find min and max.
  CacheView * in_view = AcquireVirtualCacheView (new_img, exception);

  MagickBooleanType okay = MagickTrue;
  ssize_t y;

  pob->min = DBL_MAX;
  pob->max = DBL_MIN;

  MagickBooleanType DoR = (pob->chanNum == -1) || (pob->chanNum == 0);
  MagickBooleanType DoG = (pob->chanNum == -1) || (pob->chanNum == 1);
  MagickBooleanType DoB = (pob->chanNum == -1) || (pob->chanNum == 2);

  for (y = 0; y < new_img->rows; y++) {
    if (!okay) continue;

    const VIEW_PIX_PTR *pIn = GetCacheViewVirtualPixels(
      in_view,0,y,new_img->columns,1,exception);
    if (!pIn) {okay = MagickFalse; continue; }

    ssize_t x;
    for (x = 0; x < image->columns; x++) {

      if (DoR) {
        double v = GET_PIXEL_RED   (new_img, pIn);
        if (pob->min > v) pob->min = v;
        if (pob->max < v) pob->max = v;
      }
      if (DoG) {
        double v = GET_PIXEL_GREEN (new_img, pIn);
        if (pob->min > v) pob->min = v;
        if (pob->max < v) pob->max = v;
      }
      if (DoB) {
        double v = GET_PIXEL_BLUE  (new_img, pIn);
        if (pob->min > v) pob->min = v;
        if (pob->max < v) pob->max = v;
      }

      pIn += Inc_ViewPixPtr (image);
    }
  }

  in_view  = DestroyCacheView (in_view);

  pob->min /= QuantumRange;
  pob->max /= QuantumRange;

  if (pob->verbose) {
    fprintf (stderr, "oogbox: calculated range %.*g to %.*g\n",
      pob->precision, pob->min,
      pob->precision, pob->max);
  }

  if (pob->doForceMin) pob->min = pob->forceMin;
  if (pob->doForceMax) pob->max = pob->forceMax;

  // What do we need to do?

  pob->doLo &= (pob->min < 0.0);
  pob->doHi &= (pob->max > 1.0);

  if (pob->verbose) {
    fprintf (stderr, "oogbox: min=%.*g  max=%.*g  doLo=%i doH=%i\n",
      pob->precision, pob->min,
      pob->precision, pob->max,
      pob->doLo,
      pob->doHi);
  }

  if (!pob->doLo && !pob->doHi) return new_img;

  // Calculate the parameters.

  switch (pob->oogMethod) {
    case omLinear: {
      if (pob->doLo) {
        pob->a = -pob->loLimit / (pob->min - pob->loLimit);
        pob->b = pob->loLimit * (pob->min) / (pob->min - pob->loLimit);
      }
      if (pob->doHi) {
        pob->c = (1 - pob->hiLimit) / (pob->max - pob->hiLimit);
        pob->d = pob->hiLimit * (pob->max - 1) / (pob->max - pob->hiLimit);
      }

      if (pob->verbose) {
        fprintf (stderr, "oogbox: Linear a=%.*g  b=%.*g  c=%.*g  d=%.*g\n",
          pob->precision, pob->a,
          pob->precision, pob->b,
          pob->precision, pob->c,
          pob->precision, pob->d);
      }
      break;
    }
    case omPower: {
      if (pob->doLo) {
        pob->B0 = (pob->loLimit - pob->min) / (pob->loLimit);
        pob->A0 = pob->loLimit / pow(pob->loLimit - pob->min,pob->B0);
      }
      if (pob->doHi) {
        pob->B1 = (pob->max - pob->hiLimit) / (1 - pob->hiLimit);
        pob->A1 = (1 - pob->hiLimit) / pow(pob->max - pob->hiLimit,pob->B1);
      }

      if (pob->verbose) {
        fprintf (stderr, "oogbox: Power A0=%.*g  B0=%.*g  A1=%.*g  B1=%.*g\n",
          pob->precision, pob->A0,
          pob->precision, pob->B0,
          pob->precision, pob->A1,
          pob->precision, pob->B1);
      }
      break;
    }
  }


  // Calculate the new values.

  // Linear transformation:
  //   y=a*x+b at shadows
  //   y=c*x+d at highlights
  //
  // Power transformation:
  //   y=A0*x+B0 at shadows
  //   y=A1*x+B1 at highlights

  CacheView * upd_view = AcquireAuthenticCacheView (new_img, exception);

  for (y = 0; y < new_img->rows; y++) {
    if (!okay) continue;

    VIEW_PIX_PTR *pIn = GetCacheViewAuthenticPixels(
      upd_view,0,y,new_img->columns,1,exception);
    if (!pIn) {okay = MagickFalse; continue; }

    ssize_t x;
    for (x = 0; x < image->columns; x++) {

      if (pob->doLo) {
        if (DoR) {
          double v = GET_PIXEL_RED   (new_img, pIn) / QuantumRange;
          if (v < pob->loLimit) {
            SET_PIXEL_RED   (new_img, QuantumRange * CalcLow (pob, v), pIn);
          }
        }
        if (DoG) {
          double v = GET_PIXEL_GREEN (new_img, pIn) / QuantumRange;
          if (v < pob->loLimit) {
            SET_PIXEL_GREEN (new_img, QuantumRange * CalcLow (pob, v), pIn);
          }
        }
        if (DoB) {
          double v = GET_PIXEL_BLUE  (new_img, pIn) / QuantumRange;
          if (v < pob->loLimit) {
            SET_PIXEL_BLUE  (new_img, QuantumRange * CalcLow (pob, v), pIn);
          }
        }
      }

      if (pob->doHi) {
        if (DoR) {
          double v = GET_PIXEL_RED   (new_img, pIn) / QuantumRange;
          if (v > pob->hiLimit) {
            SET_PIXEL_RED   (new_img, QuantumRange * CalcHigh (pob, v), pIn);
          }
        }
        if (DoG) {
          double v = GET_PIXEL_GREEN (new_img, pIn) / QuantumRange;
          if (v > pob->hiLimit) {
            SET_PIXEL_GREEN (new_img, QuantumRange * CalcHigh (pob, v), pIn);
          }
        }
        if (DoB) {
          double v = GET_PIXEL_BLUE  (new_img, pIn) / QuantumRange;
          if (v > pob->hiLimit) {
            SET_PIXEL_BLUE  (new_img, QuantumRange * CalcHigh (pob, v), pIn);
          }
        }
      }

      pIn += Inc_ViewPixPtr (image);
    }
    if (SyncCacheViewAuthenticPixels(upd_view,exception) == MagickFalse)
      okay = MagickFalse;
  }

  upd_view  = DestroyCacheView (upd_view);

  return (new_img);
}


ModuleExport size_t oogboxImage (
  Image **images,
  const int argc,
  const char **argv,
  ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  oogboxT ob;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  ob.precision = GetMagickPrecision();

  status = menu (argc, argv, &ob);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = oogbox (image, &ob, exception);
    if (!new_image) return (-1);

    ReplaceImageInList(&image,new_image);
    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
