/*
   Reference: http://im.snibgo.com/paintpatches.htm

   Updated:
     3-April-2018 for v7.0.7-28
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <ctype.h>
#include "vsn_defines.h"
#include "chklist.h"

#include "usercoord.inc"
#include "rmsealpha.inc"
#include "resetpage.inc"

#define VERSION "paintpatches v1.0  Copyright (c) 2017 Alan Gibson"

// FIXME: composites shouldn't clamp.


typedef enum {
  cfUndef,
  cfAvg,
  cfCent,
  cfGradient,
  cfPalette,
  cfSample
} colFromT;

typedef enum {
  psRect,
  psEllipse,
  psMin
} patchShapeT;


typedef struct {
  int
    precision,
    verbose;

  Image
    *master,
    *canvas,
    *saliency, // grayscale: white=important, black=unimportant
    *samples,
    *subsrch;

  int
    patchWi,
    patchHt,
    maxIter;

  UserCoordXyT
    patchDims;

// patch width and height, min and max

  double
    patchOpacity, // 0.0 to 1.0.
    multSal,
    wrngThresh,
    adjLcComp,
    adjLcPat,
    featherEdges;

  colFromT
    colFrom;

  patchShapeT
    patchShape;

  MagickBooleanType
    debug,
    setRectCanv,
    hotSpot,
    doMebc;

  char
    *frameName;

  ssize_t
    wrngX,
    wrngY;

  double
    wrongness;

  int
    frameNum;

  RmseAlphaT
    rmseAlpha;
} paintpatchesT;

static void usage (void)
{
  printf ("Usage: -process 'paintpatches [OPTION]...'\n");
  printf ("Paint an image with patches.\n");
  printf ("\n");
  printf ("  c,  colFrom string       'Avg', 'Cent', 'Gradient', 'Palette' or 'Sample'\n");
  printf ("  ps, patchShape string    'Rect', 'Ellipse' or 'Min'\n");
  printf ("  pd, patchDims W,H        maximum patch dimensions\n");
  printf ("  o,  opacity number       0.0 to 1.0\n");
  printf ("  ms, multSal number       multiplier for saliency 0.0 to 1.0\n");
  printf ("  wt, wrngThresh number    wrongness threshold 0.0 to 1.0\n");
  printf ("  ac, adjLcComp number     adjust lightness & contrast for comparisons\n");
  printf ("  ap, adjLcPat number      adjust lightness & contrast of patches\n");
  printf ("  fe, feather number       feather patch edges (blur sigma)\n");
  printf ("  x,  maxIter integer      maximum iterations\n");
  printf ("  m,  mebc                 do minimum error boundary cuts\n");
  printf ("  rc, rectCanv             set rectangle to canvas\n");
  printf ("  hs, hotSpot              hotspot only (for ps Min)\n");
  printf ("  w,  write filename       write iterations eg frame_%%06d.png\n");
  printf ("  d,  debug                write debugging information\n");
  printf ("  v,  verbose              write text information\n");
  printf ("  v2, verbose2             write more text information\n");
  printf ("      version              write version information to stdout\n");
  printf ("\n");
}
// FIXME: option for centre of interest: generates saliency map.

static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  // LocaleCompare is not case-sensitive,
  // so we use strcmp for the short option.

  if ((strcmp(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}

#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "paintpatches: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  paintpatchesT * ppp
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status = MagickTrue;

  ppp->master = ppp->canvas = ppp->saliency
    = ppp->samples = ppp->subsrch = NULL;

  ppp->verbose = 0;
  ppp->debug = MagickFalse;
  ppp->patchOpacity = 1.0;
  ppp->multSal = 0.5;
  ppp->wrngThresh = 0.1;
  ppp->adjLcComp = 0.0;
  ppp->adjLcPat = 0.0;
  ppp->featherEdges = 0.0;
  ppp->colFrom = cfUndef;
  ppp->patchShape = psRect;
  ppp->doMebc = MagickFalse;
  ppp->setRectCanv = MagickFalse;
  ppp->hotSpot = MagickFalse;
  ppp->patchWi = 50;
  ppp->patchHt = 50;
  ppp->maxIter = 1000;
  ppp->frameName = NULL;
  ppp->frameNum = 0;
  if (!ParseXy ("40c,40c", &ppp->patchDims)) return MagickFalse;
    // FIXME: If patchDims not specified, default according to mebc etc?

  char ** pargv = (char **)argv;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "c", "colFrom")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "avg")==0) ppp->colFrom = cfAvg;
      else if (LocaleCompare(argv[i], "cent")==0) ppp->colFrom = cfCent;
      else if (LocaleCompare(argv[i], "gradient")==0) ppp->colFrom = cfGradient;
      else if (LocaleCompare(argv[i], "palette")==0) ppp->colFrom = cfPalette;
      else if (LocaleCompare(argv[i], "sample")==0) ppp->colFrom = cfSample;
      else {
        fprintf (stderr, "kcluster: ERROR: unknown colFrom [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "ps", "patchShape")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "rect")==0) ppp->patchShape = psRect;
      else if (LocaleCompare(argv[i], "ellipse")==0) ppp->patchShape = psEllipse;
      else if (LocaleCompare(argv[i], "min")==0) ppp->patchShape = psMin;
      else {
        fprintf (stderr, "kcluster: ERROR: unknown patchShape [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "o", "opacity")==MagickTrue) {
      NEXTARG;
      ppp->patchOpacity = atof (argv[i]);
    } else if (IsArg (pa, "ms", "multSal")==MagickTrue) {
      NEXTARG;
      ppp->multSal = atof (argv[i]);
    } else if (IsArg (pa, "wt", "wrngThresh")==MagickTrue) {
      NEXTARG;
      ppp->wrngThresh = atof (argv[i]);
    } else if (IsArg (pa, "ac", "adjLcComp")==MagickTrue) {
      NEXTARG;
      ppp->adjLcComp = atof (argv[i]);
    } else if (IsArg (pa, "ap", "adjLcPat")==MagickTrue) {
      NEXTARG;
      ppp->adjLcPat = atof (argv[i]);
    } else if (IsArg (pa, "fe", "feather")==MagickTrue) {
      NEXTARG;
      ppp->featherEdges = atof (argv[i]);
    } else if (IsArg (pa, "pd", "patchdims")==MagickTrue) {
      NEXTARG;
      if (!ParseXy (argv[i], &ppp->patchDims)) {
        fprintf (stderr, "kcluster: ERROR: bad XY [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "x", "maxiter")==MagickTrue) {
      NEXTARG;
      ppp->maxIter = atoi (argv[i]);
    } else if (IsArg (pa, "m", "mebc")==MagickTrue) {
      ppp->doMebc = MagickTrue;
    } else if (IsArg (pa, "rc", "rectCanv")==MagickTrue) {
      ppp->setRectCanv = MagickTrue;
    } else if (IsArg (pa, "hs", "hotSpot")==MagickTrue) {
      ppp->hotSpot = MagickTrue;
    } else if (IsArg (pa, "w", "write")==MagickTrue) {
      NEXTARG;
      ppp->frameName = pargv[i];
    } else if (IsArg (pa, "d", "debug")==MagickTrue) {
      ppp->debug = MagickTrue;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      ppp->verbose = 1;
    } else if (IsArg (pa, "v2", "verbose2")==MagickTrue) {
      ppp->verbose = 2;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "paintpatches: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (ppp->verbose) {
    fprintf (stderr, "paintpatches options: ");
    if (ppp->verbose==1)    fprintf (stderr, " verbose");
    if (ppp->verbose==2)    fprintf (stderr, " verbose2");

    fprintf (stderr, "  colFrom ");
    switch (ppp->colFrom) {
      case cfUndef:
        fprintf (stderr, "Undefined");
        break;
      case cfAvg:
        fprintf (stderr, "Avg");
        break;
      case cfCent:
        fprintf (stderr, "Cent");
        break;
      case cfGradient:
        fprintf (stderr, "Gradient");
        break;
      case cfPalette:
        fprintf (stderr, "Palette");
        break;
      case cfSample:
        fprintf (stderr, "Sample");
        break;
    }
    fprintf (stderr, "  patchShape ");
    switch (ppp->patchShape) {
      case psRect:
        fprintf (stderr, "Rect");
        break;
      case psEllipse:
        fprintf (stderr, "Ellipse");
        break;
      case psMin:
        fprintf (stderr, "Min");
        break;
    }

    fprintf (stderr, "  patchDims ");
    WrUserCoord (&ppp->patchDims);

    fprintf (stderr, "  opacity %g", ppp->patchOpacity);
    fprintf (stderr, "  multSal %g", ppp->multSal);
    fprintf (stderr, "  wrngThresh %g", ppp->wrngThresh);
    fprintf (stderr, "  adjLcComp %g", ppp->adjLcComp);
    fprintf (stderr, "  adjLcPat %g", ppp->adjLcPat);
    fprintf (stderr, "  maxIter %i", ppp->maxIter);
    fprintf (stderr, "  featherEdges %g", ppp->featherEdges);

    if (ppp->doMebc) fprintf (stderr, "  mebc ");
    if (ppp->setRectCanv) fprintf (stderr, "  rectCanv ");
    if (ppp->hotSpot) fprintf (stderr, "  hotSpot ");
    if (ppp->frameName) fprintf (stderr, "  write %s", ppp->frameName);

    if (ppp->debug) fprintf (stderr, "  debug ");

    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


static MagickBooleanType SameSize (Image * i0, Image * i1)
{
  if (!i1) return MagickFalse;
  return ((i0->columns == i1->columns) && (i0->rows == i1->rows));
}

static void WrImage (char * title, Image * img)
{
  fprintf (stderr, "%s: ", title);

  if (img) {
    if (img->filename)
      fprintf (stderr, "[%s] ", img->filename);

    fprintf (stderr, "%lix%li\n", img->columns, img->rows);
    chkentry ("WrImage", &img);
  } else {
    fprintf (stderr, "none\n");
  }
}


static MagickStatusType WriteFrame (
  paintpatchesT * ppp,
  Image * img,
  ExceptionInfo *exception)
{
  if (!ppp->frameName) return MagickTrue;

  ImageInfo *ii = AcquireImageInfo ();

  Image * copy_img = CloneImage (img, 0, 0, MagickTrue, exception);
  if (!copy_img) return MagickFalse;

  copy_img->scene = ppp->frameNum++;

  CopyMagickString (copy_img->filename, ppp->frameName, MaxTextExtent);

  MagickBooleanType okay = WRITEIMAGE(ii, copy_img, exception);

  DestroyImageList (copy_img);

  ii = DestroyImageInfo (ii);

  return okay;
}


static MagickBooleanType MostWrong (
  paintpatchesT * ppp,
  ExceptionInfo *exception)

// FIXME: Or least wrong first? To get the broad-brush stuff first?
{
  Image * wrng_img = CloneImage (ppp->master, 0, 0, MagickTrue, exception);
  if (!wrng_img) {
    fprintf (stderr, "MostWrong: clone failed\n");
    return MagickFalse;
  }

  // FIXME: If we have adjusted patches,
  // we should do likewise for the comparison.

  if (!COMPOSITE(wrng_img, DifferenceCompositeOp, ppp->canvas, 0,0, exception))
  {
    fprintf (stderr, "MostWrong: composite failed\n");
    return MagickFalse;
  }

// Blurring removes outliers, but it's expensive.
//
//  Image * wrng_img2 = BlurImage (wrng_img, 0, 3, exception);
//  if (!wrng_img2) return MagickFalse;
//  ReplaceImageInList (&wrng_img, wrng_img2);

  if (ppp->verbose > 1) WrImage ("wrong", wrng_img);

  MagickBooleanType
    status;

  CacheView * wrng_view = AcquireVirtualCacheView (wrng_img, exception);
  CacheView * slnc_view = AcquireVirtualCacheView (ppp->saliency, exception);

  status = MagickTrue;

  double val;
  ppp->wrongness = 99e9; // for most wrong last.
  ppp->wrongness = 0;    // for most wrong first.
  ppp->wrngX = ppp->wrngY = 0;

  ssize_t y;
  for (y=0; y < (ssize_t) wrng_img->rows; y++)
  {
    ssize_t x;

    if (status == MagickFalse) continue;

    VIEW_PIX_PTR const *wp = GetCacheViewVirtualPixels(
      wrng_view,0,y,wrng_img->columns,1,exception);
    if (!wp) status=MagickFalse;

    VIEW_PIX_PTR const *sp = GetCacheViewVirtualPixels(
      slnc_view,0,y,ppp->saliency->columns,1,exception);
    if (!sp) status=MagickFalse;

    for (x=0; x < (ssize_t) wrng_img->columns; x++) {
      val = GetPixelIntensity (wrng_img, wp)
            * GET_PIXEL_RED (ppp->saliency, sp) / (double)QuantumRange;

      // '<' for most wrong first
      if (ppp->wrongness < val) {
        ppp->wrongness = val;
        ppp->wrngX = x;
        ppp->wrngY = y;
        if (val < 0) {
          fprintf (stderr, "val<0 %li,%li %g %g\n",
            x, y,
            (double)GetPixelIntensity (wrng_img, wp),
            (double)GET_PIXEL_RED (ppp->saliency, sp));
        }
      }
      wp += Inc_ViewPixPtr (wrng_img);
      sp += Inc_ViewPixPtr (ppp->saliency);
    }
  }

  ppp->wrongness /= QuantumRange;

  // FIXME: for ordinary images, how can wrongness be negative?
  if (ppp->wrongness < 0) ppp->wrongness = 0;
  if (ppp->wrongness > 1) ppp->wrongness = 1;

  slnc_view = DestroyCacheView (slnc_view);
  wrng_view = DestroyCacheView (wrng_view);

  wrng_img = DestroyImage (wrng_img);

  return MagickTrue;
}


static double GetSalientG (
  paintpatchesT * ppp,
  ssize_t x,
  ssize_t y,
  ExceptionInfo *exception)
{
  CacheView * slnc_view = AcquireVirtualCacheView (ppp->saliency, exception);

  VIEW_PIX_PTR const *sp = GetCacheViewVirtualPixels(
    slnc_view,x,y,1,1,exception);
  if (!sp) return 0;

  //double val = GetPixelIntensity (ppp->saliency, sp) / (double)QuantumRange;
  double val = GET_PIXEL_GREEN (ppp->saliency, sp) / (double)QuantumRange;

  slnc_view = DestroyCacheView (slnc_view);

  // FIXME: For ordinary images, how can this be negative?
  if (val < 0) val = 0;
  if (val > 1) val = 1;

  if (ppp->verbose > 1) {
    fprintf (stderr, "GetSalientG %li,%li %g\n",
      x, y,
      val);
  }

  return val;
}

static MagickBooleanType GetRect (
  paintpatchesT *ppp,
  RectangleInfo *prect,
  ExceptionInfo *exception)
// Gets a rectangle that includes ppp->wrngX and ppp->wrngY.
// Rectangle may straddle canvas edges.
{
  double slntFact = 1.0 - GetSalientG (ppp, ppp->wrngX, ppp->wrngY, exception);

  //if (slntFact==0) return MagickFalse; // FIXME: so white saliency does nothing?!?!?

  // FIXME: we could vary the width and height randomly.

  prect->x = floor (ppp->wrngX - slntFact * ppp->patchWi/2.0 + 0.5);
  prect->y = floor (ppp->wrngY - slntFact * ppp->patchHt/2.0 + 0.5);
  prect->width  = floor (slntFact * ppp->patchWi + 0.5);
  prect->height = floor (slntFact * ppp->patchHt + 0.5);
  if (prect->width  < 1) prect->width  = 1;
  if (prect->height < 1) prect->height = 1;

  if (ppp->verbose > 1) {
    fprintf (stderr, "GetRect %lix%li+%li+%li  slntFact=%g\n",
      prect->width, prect->height, prect->x, prect->y, slntFact);
  }

  return MagickTrue;
}


static MagickBooleanType SetImgAlpha (
  paintpatchesT * ppp,
  Image * img,
  ExceptionInfo *exception)
{
  if (ppp->patchOpacity == 1.0) return MagickTrue;

  CacheView * img_view = AcquireAuthenticCacheView (img, exception);

  // FIXME: enable alpha for the image?

  double alph = ppp->patchOpacity * QuantumRange;

  MagickBooleanType okay = MagickTrue;

  ssize_t y;
#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) \
    shared(okay) \
    MAGICK_THREADS(img,img,img->rows,1)
#endif
  for (y = 0; y < img->rows; y++) {
    ssize_t x;

    if (!okay) continue;

    VIEW_PIX_PTR *sp = GetCacheViewAuthenticPixels(
      img_view,0,y,img->columns,1,exception);
    if (!sp) okay = MagickFalse;

    for (x = 0; x < img->columns; x++) {
      SET_PIXEL_ALPHA (img, alph, sp);
      sp += Inc_ViewPixPtr (img);
    }
    if (!SyncCacheViewAuthenticPixels(img_view,exception)) {
      okay = MagickFalse;
    }
  }

  if (!okay) return MagickFalse;

  img_view = DestroyCacheView (img_view);

  return MagickTrue;
}

static MagickBooleanType EllipseOnly (
  paintpatchesT * ppp,
  Image * pat_img,
  ExceptionInfo *exception)
// Makes pixels outside an ellipse transparent.
{
  CacheView * img_view = AcquireAuthenticCacheView (pat_img, exception);

  // FIXME: enable alpha for the image?

  // Quick and simple anti-aliasing.

  if (pat_img->columns <= 2 || pat_img->rows <= 2) return MagickTrue;

  // FIXME: following isn't symmetrical.

  double a = pat_img->columns / 2.0;
  double b = pat_img->rows / 2.0;

  double a2 = a*a;
  double b2 = b*b;
  double a2b2 = a2*b2;

  MagickBooleanType okay = MagickTrue;

  ssize_t y;
#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) \
    shared(okay) \
    MAGICK_THREADS(pat_img,pat_img,pat_img->rows,1)
#endif
  for (y = 0; y < pat_img->rows; y++) {
    ssize_t x;

    VIEW_PIX_PTR *sp = GetCacheViewAuthenticPixels(
      img_view,0,y,pat_img->columns,1,exception);
    if (!sp) okay = MagickFalse;

    if (!okay) continue;

    double a2y2   = a2 * (y-b) * (y-b);
    double a2y2m1 = a2 * (y-b-0.5) * (y-b-0.5);
    double a2y2p1 = a2 * (y-b+0.5) * (y-b+0.5);

    for (x = 0; x < pat_img->columns; x++) {
      double b2x2   = b2 * (x-a) * (x-a);
      double b2x2m1 = b2 * (x-a-0.5) * (x-a-0.5);
      double b2x2p1 = b2 * (x-a+0.5) * (x-a+0.5);
      int nIn = 0;

      if (b2x2m1 + a2y2m1 < a2b2) nIn++;
      if (b2x2   + a2y2m1 < a2b2) nIn++;
      if (b2x2p1 + a2y2m1 < a2b2) nIn++;

      if (b2x2m1 + a2y2   < a2b2) nIn++;
      if (b2x2   + a2y2   < a2b2) nIn++;
      if (b2x2p1 + a2y2   < a2b2) nIn++;

      if (b2x2m1 + a2y2p1 < a2b2) nIn++;
      if (b2x2   + a2y2p1 < a2b2) nIn++;
      if (b2x2p1 + a2y2p1 < a2b2) nIn++;

      if (nIn < 9) {
        SET_PIXEL_ALPHA (pat_img,
                         nIn / 9.0 * GET_PIXEL_ALPHA(pat_img,sp),
                         sp);
      }
      sp += Inc_ViewPixPtr (pat_img);
    }
    if (SyncCacheViewAuthenticPixels(img_view,exception) == MagickFalse)
      okay = MagickFalse;
  }

  if (!okay) return MagickFalse;

  img_view = DestroyCacheView (img_view);

  return MagickTrue;
}


static Image * SearchSamples (
  paintpatchesT * ppp,
  Image *pat_image,
  ExceptionInfo *exception)
// Search samples image for best match to pat_image.
// Creates and returns an image, which caller should destroy.
{
  // Note: InitRmseAlpha/DeInit are at a higher level.

  Image * srch = ppp->subsrch;
  if (!srch) srch = ppp->samples;

  ppp->rmseAlpha.adjustMeanSd = ppp->adjLcComp;
  ppp->rmseAlpha.saveInpScales = MagickTrue;
  ppp->rmseAlpha.do_verbose = (ppp->verbose > 1);

  if (!subRmseAlphaMS (&ppp->rmseAlpha, srch, pat_image, exception)) {
    fprintf (stderr, "SearchSamples: subRmseAlphaMS failed\n");
    return NULL;
  }
  if (ppp->verbose > 1) {
    fprintf (stderr, "SearchSamples: found %g at %li,%li\n",
      ppp->rmseAlpha.score, ppp->rmseAlpha.solnX, ppp->rmseAlpha.solnY);
  }

  RectangleInfo fndRect;
  fndRect.x = ppp->rmseAlpha.solnX;
  fndRect.y = ppp->rmseAlpha.solnY;
  fndRect.width  = pat_image->columns;
  fndRect.height = pat_image->rows;

  // FIXME: Option to get larger patch, same size as canvas.

  if (ppp->verbose > 1) {
    fprintf (stderr, "fndRect %lix%li+%li+%li  canvas %lix%li\n",
      fndRect.width, fndRect.height, fndRect.x, fndRect.y,
      ppp->canvas->columns, ppp->canvas->rows);
  }

  if (ppp->setRectCanv) {
    ssize_t dx = ppp->canvas->columns - fndRect.width;
    ssize_t dy = ppp->canvas->rows - fndRect.height;
    fndRect.x -= dx/2;
    fndRect.y -= dy/2;
    if (fndRect.x < 0) fndRect.x = 0;
    if (fndRect.y < 0) fndRect.y = 0;
    fndRect.width = ppp->canvas->columns;
    fndRect.height = ppp->canvas->rows;
    if (fndRect.x + fndRect.width > ppp->samples->columns) fndRect.x = ppp->samples->columns - fndRect.width;
    if (fndRect.y + fndRect.height > ppp->samples->rows) fndRect.y = ppp->samples->rows - fndRect.height;
  }

  if (ppp->verbose > 1) {
    fprintf (stderr, "fndRect %lix%li+%li+%li\n",
      fndRect.width, fndRect.height, fndRect.x, fndRect.y);
  }

  Image * fnd_img = CropImage (ppp->samples, &fndRect, exception);
  if (!fnd_img) {
    fprintf (stderr, "SearchSamples crop failed\n");
    return NULL;
  }

  return fnd_img;
}


static void ClipPatchRect (
  paintpatchesT * ppp,
  RectangleInfo * patchRect,
  RectangleInfo * clippedRect,
  int *dx,
  int *dy)
{
  *clippedRect = *patchRect;

  if (clippedRect->x < 0) {
    *dx = -clippedRect->x;
    clippedRect->width += clippedRect->x;
    clippedRect->x = 0;
  } else {
    *dx = 0;
  }

  if (clippedRect->y < 0) {
    *dy = -clippedRect->y;
    clippedRect->height += clippedRect->y;
    clippedRect->y = 0;
  } else {
    *dy = 0;
  }

  if (clippedRect->width + clippedRect->x > ppp->canvas->columns)
    clippedRect->width = ppp->canvas->columns - clippedRect->x;
  if (clippedRect->height + clippedRect->y > ppp->canvas->rows)
    clippedRect->height = ppp->canvas->rows - clippedRect->y;
}

static MagickBooleanType AdjLcFndToPat (
  paintpatchesT * ppp,
  Image * fnd_pat,
  Image * pat_img,
  ExceptionInfo *exception)
// Adjust lightness and contrast of found patch to match pat_img.
{
  MeanSdT msFnd, msPat;

  if (!CalcMeanSdImg (fnd_pat, &msFnd, exception)) return MagickFalse;
  if (!CalcMeanSdImg (pat_img, &msPat, exception)) return MagickFalse;

  GainBiasT gb;
  MeanSdToGainBias (ppp->adjLcPat, &msFnd, &msPat, &gb);

  // FIXME: apply the gain and bias to fnd_pat.

  CacheView * img_view = AcquireAuthenticCacheView (fnd_pat, exception);

  MagickBooleanType okay = MagickTrue;

  ssize_t y;
#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) \
    shared(okay) \
    MAGICK_THREADS(fnd_pat,pat_img,fnd_pat->rows,1)
#endif
  for (y = 0; y < fnd_pat->rows; y++) {
    ssize_t x;

    VIEW_PIX_PTR *sp = GetCacheViewAuthenticPixels(
      img_view,0,y,fnd_pat->columns,1,exception);
    if (!sp) okay = MagickFalse;

    if (!okay) continue;

    for (x = 0; x < fnd_pat->columns; x++) {
      SET_PIXEL_RED   (fnd_pat,
                       GET_PIXEL_RED(pat_img,sp)*gb.gainR+gb.biasR,
                       sp);
      SET_PIXEL_GREEN (fnd_pat,
                       GET_PIXEL_GREEN(pat_img,sp)*gb.gainG+gb.biasG,
                       sp);
      SET_PIXEL_BLUE  (fnd_pat,
                       GET_PIXEL_BLUE(pat_img,sp)*gb.gainB+gb.biasB,
                       sp);
      sp += Inc_ViewPixPtr (fnd_pat);
    }
    if (SyncCacheViewAuthenticPixels(img_view,exception) == MagickFalse)
      okay = MagickFalse;
  }

  if (!okay) return MagickFalse;

  img_view = DestroyCacheView (img_view);

  return MagickTrue;
}


static void SetRectCanv (
  paintpatchesT * ppp,
  RectangleInfo * rect
)
{
  rect->x = rect->y = 0;
  rect->width  = ppp->canvas->columns;
  rect->height = ppp->canvas->rows;
}

static Image *GetPatchImg (
  paintpatchesT * ppp,
  RectangleInfo * patchRect,
  ExceptionInfo *exception)
// Creates and returns an image, which caller should destroy.
{
  if (ppp->verbose > 1)
    fprintf (stderr, "GetPatchImg: %lix%li+%li+%li\n",
      patchRect->width, patchRect->height, patchRect->x, patchRect->y);

  Image * pat_img = CropImage (ppp->master, patchRect, exception);
  if (!pat_img) {
    fprintf (stderr, "GetPatchImg: crop failed\n");
    return NULL;
  }
  ResetPage (pat_img);
  // Caution: this may be smaller than requested, when at edge of ppp->master.

  if (ppp->colFrom == cfAvg) {
    Image * pat_img2 = ScaleImage (pat_img, 1, 1, exception);
    if (!pat_img2) return NULL;
    ReplaceImageInList (&pat_img, pat_img2);

    if (ppp->setRectCanv) SetRectCanv (ppp, patchRect);
    pat_img2 = SampleImage (pat_img, patchRect->width,
      patchRect->height, exception);
    if (!pat_img2) return NULL;
    ReplaceImageInList (&pat_img, pat_img2);
  } else if (ppp->colFrom == cfCent) {
    RectangleInfo rect;
    rect.x = patchRect->width / 2;
    rect.y = patchRect->height / 2;
    rect.width  = 1;
    rect.height = 1;
    // If patch overlaps edge, we may not have the centre point.
    if (rect.x < 0) rect.x = 0;
    if (rect.y < 0) rect.y = 0;
    if (rect.x >= pat_img->columns) rect.x = pat_img->columns-1;
    if (rect.y >= pat_img->rows)    rect.y = pat_img->rows-1;

    Image * pat_img2 = CropImage (pat_img, &rect, exception);
    if (!pat_img2) {
      fprintf (stderr, "GetPatchImg: pat_img2 fail: %lix%li+%li+%li\n",
        rect.width, rect.height, rect.x, rect.y);
      return NULL;
    }
    ResetPage (pat_img2);
    ReplaceImageInList (&pat_img, pat_img2);

    if (ppp->setRectCanv) SetRectCanv (ppp, patchRect);
    pat_img2 = SampleImage (pat_img, patchRect->width,
      patchRect->height, exception);
    if (!pat_img2) return NULL;
    ReplaceImageInList (&pat_img, pat_img2);
  } else if (ppp->colFrom == cfSample) {

    if (ppp->setRectCanv) SetRectCanv (ppp, patchRect);
    Image * fnd_pat = SearchSamples (ppp, pat_img, exception);
    if (!fnd_pat) {
      return NULL;
    }

    // FIXME: better place for next?
    if (ppp->adjLcPat > 0.0) {
      if (!AdjLcFndToPat (ppp, fnd_pat, pat_img, exception)) return NULL;
    }

    ReplaceImageInList (&pat_img, fnd_pat);
    // What if the image is smaller than requested?
    // Answer: extent.
    if (pat_img->columns < patchRect->width
     || pat_img->rows < patchRect->height)
    {
      RectangleInfo extRect;
      int dx, dy;
      ClipPatchRect (ppp, patchRect, &extRect, &dx, &dy);
      // Ignore calculated extRect; we want only dx and dy.
      extRect.width  = patchRect->width;
      extRect.height = patchRect->height;
      extRect.x = -dx;
      extRect.y = -dy;
      Image * pat_img2 = ExtentImage (pat_img, &extRect, exception);
      if (!pat_img2) return NULL;
      ReplaceImageInList (&pat_img, pat_img2);
    }
  } else {

    // FIXME: For now, get gradient from master.
    // Gradient: chop to get the 4 corners, and resize.

    // If too small to chop and resize, don't.

    if (patchRect->width > 2 && patchRect->height > 2) {
      RectangleInfo chpRect;
      chpRect.x = chpRect.y = 1;
      chpRect.width  = pat_img->columns  - 2;
      chpRect.height = pat_img->rows - 2;

      Image * pat_img2 = ChopImage (pat_img, &chpRect, exception);
      if (!pat_img2) return NULL;
      ReplaceImageInList (&pat_img, pat_img2);

      if (pat_img->columns != 2 || pat_img->rows != 2) {
        fprintf (stderr, "GetPatchImg: after chop %lix%li\n",
          pat_img->columns, pat_img->rows);
      }

      if (ppp->setRectCanv) SetRectCanv (ppp, patchRect);
#if IMV6OR7==6
      pat_img2 = ResizeImage (pat_img, patchRect->width,
        patchRect->height, UndefinedFilter, 1.0, exception);
#else
      pat_img2 = ResizeImage (pat_img, patchRect->width,
        patchRect->height, UndefinedFilter, exception);
#endif
      if (!pat_img2) return NULL;
      ReplaceImageInList (&pat_img, pat_img2);
    }
  }

  // FIXME: possibly adjust to match pat_img.
  if (ppp->adjLcPat > 0.0) {

    // Bugger, we've already replaced it.
    //Image * pat_img2 = AdjLcFndToPat (ppp, ??, ??, exception);
  }

#if IMV6OR7==6
  SetImageAlphaChannel (pat_img, SetAlphaChannel);
#else
  SetImageAlphaChannel (pat_img, SetAlphaChannel, exception);
#endif

  if (!SetImgAlpha (ppp, pat_img, exception)) return NULL;

  if (ppp->patchShape == psEllipse) {
    if (!EllipseOnly (ppp, pat_img, exception)) return NULL;
  }

  return pat_img;
}


static MagickBooleanType PaintPatchImg (
  paintpatchesT * ppp,
  Image * dest_img,
  RectangleInfo * patchRect,
  Image * pat_img,
  ExceptionInfo *exception)
{
  if (!COMPOSITE(dest_img,
         OverCompositeOp, pat_img, patchRect->x, patchRect->y, exception))
  {
    fprintf (stderr, "PaintPatch: composite failed\n");
    return MagickFalse;
  }
  return MagickTrue;
}

static MagickBooleanType PaintPatchCanvas (
  paintpatchesT * ppp,
  RectangleInfo * patchRect,
  Image * pat_img,
  ExceptionInfo *exception)
{
  return PaintPatchImg (ppp, ppp->canvas, patchRect, pat_img, exception);
}


static MagickBooleanType SetOneNonSalient (
  paintpatchesT * ppp,
  ssize_t x,
  ssize_t y,
  ExceptionInfo *exception)
{
  // FIXME: This could draw the patch shape, in black, on the saliency image.

  CacheView * slnc_view = AcquireAuthenticCacheView (ppp->saliency, exception);

  VIEW_PIX_PTR *sp = GetCacheViewAuthenticPixels(
    slnc_view,x,y,1,1,exception);
  if (!sp) return MagickFalse;

  SET_PIXEL_RED (ppp->saliency, 0, sp);
//  SET_PIXEL_GREEN (ppp->saliency, 0, sp);
//  SET_PIXEL_BLUE (ppp->saliency, 0, sp);

  if (SyncCacheViewAuthenticPixels(slnc_view,exception) == MagickFalse)
    return MagickFalse;

  slnc_view = DestroyCacheView (slnc_view);

  return MagickTrue;
}

static MagickBooleanType MultOneSalient (
  paintpatchesT * ppp,
  ssize_t x,
  ssize_t y,
  double fact,
  ExceptionInfo *exception)
{
  CacheView * slnc_view = AcquireAuthenticCacheView (ppp->saliency, exception);

  VIEW_PIX_PTR *sp = GetCacheViewAuthenticPixels(
    slnc_view,x,y,1,1,exception);
  if (!sp) return MagickFalse;

  SET_PIXEL_RED (ppp->saliency, fact * GET_PIXEL_RED(ppp->saliency,sp), sp);
//  SET_PIXEL_GREEN (ppp->saliency, fact * GET_PIXEL_GREEN(ppp->saliency,sp), sp);
//  SET_PIXEL_BLUE (ppp->saliency, fact * GET_PIXEL_BLUE(ppp->saliency,sp), sp);

  if (SyncCacheViewAuthenticPixels(slnc_view,exception) == MagickFalse)
    return MagickFalse;

  slnc_view = DestroyCacheView (slnc_view);

  return MagickTrue;
}

static MagickBooleanType MultSalientPatch (
  paintpatchesT * ppp,
  RectangleInfo * patchRect,
  Image *pat_img,
  double factor,
  ExceptionInfo *exception)
// Where patch is not totally transparent, multiplies saliency by factor.
{
  CacheView * slnc_view = AcquireAuthenticCacheView (ppp->saliency, exception);
  CacheView * pat_view = AcquireVirtualCacheView (pat_img, exception);

  if (ppp->verbose > 1)
    fprintf (stderr, "MultSalientPatch: rect %lix%li+%li+%li x%g\n",
      patchRect->width, patchRect->height, patchRect->x, patchRect->y,
      factor);

  RectangleInfo clippedRect;
  int dx, dy;
  ClipPatchRect (ppp, patchRect, &clippedRect, &dx, &dy);

  if (ppp->verbose > 1)
    fprintf (stderr, "MultSalientPatch: clpd %lix%li+%li+%li dx=%i dy=%i\n",
      clippedRect.width, clippedRect.height, clippedRect.x, clippedRect.y,
      dx, dy);

  MagickBooleanType okay = MagickTrue;

  ssize_t y;
#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) \
    shared(okay) \
    MAGICK_THREADS(pat_img,pat_img,clippedRect.height,1)
#endif
  for (y = clippedRect.y; y < clippedRect.y + clippedRect.height; y++) {
    VIEW_PIX_PTR *sp = GetCacheViewAuthenticPixels(
      slnc_view,clippedRect.x,y,clippedRect.width,1,exception);
    if (!sp) okay = MagickFalse;

    const VIEW_PIX_PTR *pp = GetCacheViewVirtualPixels(
      pat_view,dx,y-clippedRect.y+dy,clippedRect.width,1,exception);
    if (!pp) okay = MagickFalse;


    const double opacThresh = ppp->patchOpacity * 0.5;
    // Patch alpha may have been reduced.
    ssize_t x;
    for (x = 0; x < clippedRect.width; x++) {
      if (GET_PIXEL_ALPHA (pat_img, pp) > opacThresh) {
        SET_PIXEL_RED (ppp->saliency,
                       factor * GET_PIXEL_RED(ppp->saliency,sp),
                       sp);
      }
      sp += Inc_ViewPixPtr (ppp->saliency);
      pp += Inc_ViewPixPtr (pat_img);
    }
    if (SyncCacheViewAuthenticPixels(slnc_view,exception) == MagickFalse)
      okay = MagickFalse;
  }

  if (!okay) return MagickFalse;

  pat_view  = DestroyCacheView (pat_view);
  slnc_view = DestroyCacheView (slnc_view);

  return MagickTrue;
}


static double inline RelColDiff (
  Image * img0,
  Image * img1,
  const VIEW_PIX_PTR *vp0,
  const VIEW_PIX_PTR *vp1)
{
  double dr = GET_PIXEL_RED(img0,vp0)   - GET_PIXEL_RED(img1,vp1);
  double dg = GET_PIXEL_GREEN(img0,vp0) - GET_PIXEL_GREEN(img1,vp1);
  double db = GET_PIXEL_BLUE(img0,vp0)  - GET_PIXEL_BLUE(img1,vp1);

  return (dr*dr + dg+dg + db*db);
}

static double inline SumSq (
  Image * img,
  const VIEW_PIX_PTR *vp)
{
  double dr = GET_PIXEL_RED(img,vp);
  double dg = GET_PIXEL_GREEN(img,vp);
  double db = GET_PIXEL_BLUE(img,vp);

  return (dr*dr + dg+dg + db*db);
}


static MagickBooleanType NearestWhite (
  paintpatchesT * ppp,
  Image * img,
  ssize_t cx,
  ssize_t cy,
  ssize_t *fndX,
  ssize_t *fndY,
  MagickBooleanType * isFound,
  ExceptionInfo *exception)
// Returns the nearest white pixel to cx,cy.
{
  CacheView * img_view = AcquireVirtualCacheView (img, exception);

#define LIMIT (QuantumRange - 1e-6)
//#define EPS 1e-6

  MagickBooleanType okay = MagickTrue, found = MagickFalse;
  double nearDistSq = 0;

  ssize_t y, nearX=-1, nearY=-1;

#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) \
    shared(okay) \
    MAGICK_THREADS(img,img,img->rows,1)
#endif
  for (y = 0; y < img->rows; y++) {
    if (!okay) continue;

    double
      dx, dy;

    const VIEW_PIX_PTR *p = GetCacheViewVirtualPixels(
      img_view,0,y,img->columns,1,exception);
    if (!p) okay = MagickFalse;
    ssize_t x;
    for (x = 0; x < img->columns; x++) {

      if (  GET_PIXEL_RED(img,p)   >= LIMIT
         && GET_PIXEL_GREEN(img,p) >= LIMIT
         && GET_PIXEL_BLUE(img,p)  >= LIMIT)
      {
        if (found) {
          dx = cx - x;
          dy = cy - y;
          double DistSq = dx*dx + dy*dy;
          if (nearDistSq > DistSq) {
            nearDistSq = DistSq;
            nearX = x;
            nearY = y;
          }
        } else {
          found = MagickTrue;
          dx = cx - x;
          dy = cy - y;
          nearDistSq = dx*dx + dy*dy;
          nearX = x;
          nearY = y;
        }
      }

      p  += Inc_ViewPixPtr (img);
    }
  }

  img_view = DestroyCacheView (img_view);

  *fndX = nearX;
  *fndY = nearY;
  *isFound = found;

  if (ppp->debug) {
    fprintf (stderr, "NearestWhite: %li,%li => %li,%li\n",
      cx, cy, *fndX, *fndY);
  }

  return okay;
}

static MagickBooleanType OnlyHotspot (
  paintpatchesT * ppp,
  Image * opaq_img,
  ExceptionInfo *exception)
// opaq_img represents required opacity, white = opaque.
// This sets all pixels black, except those connected to wrngX,wrngY.
// If no pixels were white, returns a black image (and MagickTrue).
{
  // Poss flood-fill at hotspot. But where is the hotspot?
  // If we are doing rectCanv, the hotspot is wrngX,wrngY.

  ssize_t hsX, hsY;
  MagickBooleanType isFound;
  if (! NearestWhite (ppp,
          opaq_img, ppp->wrngX, ppp->wrngY,
          &hsX, &hsY, &isFound, exception)) return MagickFalse;

  if (!isFound) return MagickTrue;

//fprintf (stderr, "OnlyHotspot 1 %lix%li @ %li,%li => %li,%li\n",
//  opaq_img->columns, opaq_img->rows,
//  ppp->wrngX, ppp->wrngY, hsX, hsY);

  DrawInfo *draw_info = AcquireDrawInfo();

  PIX_INFO target; // white

#if IMV6OR7==6
  QueryColorDatabase ("gray50", &draw_info->fill, exception);

  GetMagickPixelPacket (opaq_img, &target);
  QueryMagickColor ("white", &target, exception);

  if (!FloodfillPaintImage (
    opaq_img, DefaultChannels, draw_info,
    &target, hsX, hsY, MagickFalse)) return MagickFalse;
#else
  QueryColorCompliance ("gray50", AllCompliance, &draw_info->fill, exception);

  GetPixelInfo (opaq_img, &target);
  QueryColorCompliance ("white", AllCompliance, &target, exception);

  if (!FloodfillPaintImage (
    opaq_img, draw_info,
    &target, hsX, hsY, MagickFalse, exception)) return MagickFalse;
#endif

  draw_info = DestroyDrawInfo(draw_info);

//  if (ppp->debug) {
//    WriteFrame (ppp, opaq_img, exception);
//  }

  // Now change white to black, and gray to white.

  CacheView * opaq_view = AcquireAuthenticCacheView (opaq_img, exception);

  VIEW_PIX_PTR vpBlack, vpWhite;
  SET_PIXEL_RED   (opaq_img, 0, &vpBlack);
  SET_PIXEL_GREEN (opaq_img, 0, &vpBlack);
  SET_PIXEL_BLUE  (opaq_img, 0, &vpBlack);
  SET_PIXEL_RED   (opaq_img, QuantumRange, &vpWhite);
  SET_PIXEL_GREEN (opaq_img, QuantumRange, &vpWhite);
  SET_PIXEL_BLUE  (opaq_img, QuantumRange, &vpWhite);

  const double NearWhite = QuantumRange * 0.9;
  const double NearBlack = QuantumRange * 0.1;

  MagickBooleanType okay = MagickTrue;
  ssize_t y;
#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) \
    shared(okay) \
    MAGICK_THREADS(opaq_img,opaq_img,opaq_img->rows,1)
#endif
  for (y = 0; y < opaq_img->rows; y++) {
    if (!okay) continue;

    VIEW_PIX_PTR *pp = GetCacheViewAuthenticPixels(
      opaq_view,0,y,opaq_img->columns,1,exception);
    if (!pp) okay = MagickFalse;

    double val;
    ssize_t x;
    for (x = 0; x < opaq_img->columns; x++) {
      val = GetPixelIntensity (opaq_img, pp);
      if (val > NearWhite) {
        *pp = vpBlack;
      } else if (val > NearBlack) {
        *pp = vpWhite;
      }

      pp  += Inc_ViewPixPtr (opaq_img);
    }

    if (SyncCacheViewAuthenticPixels(opaq_view,exception) == MagickFalse)
      okay = MagickFalse;
  }

  opaq_view = DestroyCacheView (opaq_view);

  return okay;
}

static MagickBooleanType TransNoImp (
  paintpatchesT * ppp,
  RectangleInfo * patchRect,
  Image *pat_img,
  MagickBooleanType * anyOpaque,
  ExceptionInfo *exception)
// Where patch would not improve the canvas, make patch pixels transparent.
// Optionally also: make transparent all pixels
//   except those that would be flood-filled from the hotspot.
// 
{
  if (ppp->verbose > 1) {
    fprintf (stderr, "TransNoImp\n");
    fprintf (stderr, "TransNoImp: rect %lix%li+%li+%li\n",
      patchRect->width, patchRect->height, patchRect->x, patchRect->y);
  }

  // Paste the patch over a trial copy of the canvas.

  Image * trial_canv = CloneImage(ppp->canvas, 0, 0, MagickTrue, exception);
  if (!trial_canv) return MagickFalse;

  if (!PaintPatchImg (ppp, trial_canv, patchRect, pat_img, exception))
    return MagickFalse;

  // Crop master, trial and canvas.
  // Find the differences:
  //   crp_trial := abs (master - trial)
  //   crp_canv  := abs (master - canvas)
  // Possibly blur.
  // The smallest tells us which we should keep.
  // So where crp_trial > crp_canv, make patch transparent.

  RectangleInfo clippedRect;
  int dx, dy;
  ClipPatchRect (ppp, patchRect, &clippedRect, &dx, &dy);

  Image * crp_mast = CropImage (ppp->master, &clippedRect, exception);
  if (!crp_mast) return MagickFalse;
  ResetPage (crp_mast);

  Image * crp_trial = CropImage (trial_canv, &clippedRect, exception);
  if (!crp_trial) return MagickFalse;
  ResetPage (crp_trial);

  Image * crp_canv = CropImage (ppp->canvas, &clippedRect, exception);
  if (!crp_canv) return MagickFalse;
  ResetPage (crp_canv);

  if (!COMPOSITE(crp_trial, DifferenceCompositeOp, crp_mast, 0,0, exception))
  {
    fprintf (stderr, "TransNoImp: composite1 failed\n");
    return MagickFalse;
  }

  if (!COMPOSITE(crp_canv, DifferenceCompositeOp, crp_mast, 0,0, exception))
  {
    fprintf (stderr, "TransNoImp: composite2 failed\n");
    return MagickFalse;
  }

  if (ppp->verbose > 1) {
    fprintf (stderr, "TransNoImp3: rect %lix%li+%li+%li\n",
      clippedRect.width, clippedRect.height, clippedRect.x, clippedRect.y);
  }

  Image * blr_img2 = BlurImage (crp_trial, 0, 3, exception);
  if (!blr_img2) return MagickFalse;
  ReplaceImageInList (&crp_trial, blr_img2);

  blr_img2 = BlurImage (crp_canv, 0, 3, exception);
  if (!blr_img2) return MagickFalse;
  ReplaceImageInList (&crp_canv, blr_img2);

  if (ppp->debug) {
    WrImage ("trial", crp_trial);
    WrImage (" canv", crp_canv);
    WrImage ("  pat", pat_img);
    WriteFrame (ppp, crp_trial, exception);
    WriteFrame (ppp, crp_canv, exception);
  }

  // Write opacity as black or white to a cloned image opaq_img,
  // possibly flood-fill etc,
  // then "-CopyOpacity" that to the patch.

  Image * opaq_img = CloneImage (
      pat_img, pat_img->columns, pat_img->rows, MagickTrue, exception);
  if (!opaq_img) return MagickFalse;
  SetAllOneCol (opaq_img, "White", exception);

  CacheView * crp_trial_view = AcquireVirtualCacheView (crp_trial, exception);
  CacheView * crp_canv_view = AcquireVirtualCacheView (crp_canv, exception);
  CacheView * opaq_view = AcquireAuthenticCacheView (opaq_img, exception);

  if (ppp->verbose > 1) fprintf (stderr, "TransNoImp4 dx=%i dy=%i\n", dx, dy);

  *anyOpaque = MagickFalse;
  MagickBooleanType okay = MagickTrue;
  ssize_t y;
#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) \
    shared(okay) \
    MAGICK_THREADS(crp_trial,crp_canv,clippedRect.height,1)
#endif
  for (y = 0; y < clippedRect.height; y++) {
    if (!okay) continue;

    const VIEW_PIX_PTR *ctp = GetCacheViewVirtualPixels(
      crp_trial_view,0,y,clippedRect.width,1,exception);
    if (!ctp) okay = MagickFalse;

    const VIEW_PIX_PTR *ccp = GetCacheViewVirtualPixels(
      crp_canv_view,0,y,clippedRect.width,1,exception);
    if (!ccp) okay = MagickFalse;

    VIEW_PIX_PTR *pp = GetCacheViewAuthenticPixels(
      opaq_view,dx,y+dy,clippedRect.width,1,exception);
    if (!pp) okay = MagickFalse;

    ssize_t x;
    for (x = 0; x < clippedRect.width; x++) {
      // FIXME: We could use a threshold here.
      if (SumSq (crp_trial, ctp) >= SumSq (crp_canv, ccp)) {
        SET_PIXEL_RED   (opaq_img, 0, pp);
        SET_PIXEL_GREEN (opaq_img, 0, pp);
        SET_PIXEL_BLUE  (opaq_img, 0, pp);
      } else {
        *anyOpaque = MagickTrue;
      }

      ctp += Inc_ViewPixPtr (crp_trial);
      ccp += Inc_ViewPixPtr (crp_canv);
      pp  += Inc_ViewPixPtr (opaq_img);
    }
    if (SyncCacheViewAuthenticPixels(opaq_view,exception) == MagickFalse)
      okay = MagickFalse;
  }

  if (!okay) return MagickFalse;

  if (ppp->verbose > 1) fprintf (stderr, "TransNoImp5\n");

  opaq_view      = DestroyCacheView (opaq_view);
  crp_canv_view  = DestroyCacheView (crp_canv_view);
  crp_trial_view = DestroyCacheView (crp_trial_view);

  if (ppp->debug) {
    WriteFrame (ppp, opaq_img, exception);
  }

  if (anyOpaque) {

    if (ppp->hotSpot) {
      if (!OnlyHotspot (ppp, opaq_img, exception)) {
        fprintf (stderr, "TransNoImp: OnlyHotspot failed\n");
        return MagickFalse;
      }
    }

    if (ppp->featherEdges > 0) {
      Image * blr_img2 = BlurImage (opaq_img,
        4*ppp->featherEdges, ppp->featherEdges, exception);
      if (!blr_img2) return MagickFalse;
      ReplaceImageInList (&opaq_img, blr_img2);
    }

    if (ppp->debug) {
      WriteFrame (ppp, opaq_img, exception);
    }

#if IMV6OR7==6
    SetImageAlphaChannel (opaq_img, DeactivateAlphaChannel);
#else
    SetImageAlphaChannel (opaq_img, DeactivateAlphaChannel, exception);
#endif

    if (!COMPOSITE(pat_img, COPY_OPACITY, opaq_img, 0,0, exception))
    {
      fprintf (stderr, "TransNoImp: composite CO failed\n");
      return MagickFalse;
    }

    if (ppp->debug) {
      WriteFrame (ppp, pat_img, exception);
    }
  }

  opaq_img   = DestroyImage (opaq_img);
  crp_canv   = DestroyImage (crp_canv);
  crp_trial  = DestroyImage (crp_trial);
  crp_mast   = DestroyImage (crp_mast);
  trial_canv = DestroyImage (trial_canv);

  return MagickTrue;
}


//#if IMV6OR7==6
//#  define CLAMP(img,exc) ClampImage(img)
//#else
//#  define CLAMP(img,exc) ClampImage(img,exc)
//#endif


// The next function is similar in style to functions in transform.c
// It takes one to five images, and returns one image.
//
static Image *paintpatches (
  Image *image,
  paintpatchesT * ppp,
  ExceptionInfo *exception)
{
  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (ppp->verbose) {
    fprintf (stderr, "paintpatches: Input image [%s] %ix%i depth is %i\n",
             image->filename,
             (int)image->columns, (int)image->rows,
             (int)image->depth);
  }

  // Check the images.

  ppp->master = ppp->canvas = ppp->saliency
    = ppp->samples = ppp->subsrch = NULL;

  ppp->master = image;
  if (!ppp->master) return NULL;
  CLAMP (ppp->master, exception);

  Image *i2=NULL, *i3=NULL, *i4=NULL;

  i2 = GetNextImageInList(ppp->master);
  if (i2) i3 = GetNextImageInList(i2);
  if (i3) i4 = GetNextImageInList(i3);

  if (SameSize (ppp->master, i2)) {
    ppp->canvas = i2;
    if (SameSize (ppp->canvas, i3)) {
      ppp->saliency = i3;
      ppp->samples = i4;
    } else {
      ppp->samples = i3;
    }
  } else {
    ppp->samples = i2;
  }
  if (ppp->saliency) {
    CLAMP (ppp->saliency, exception);
  }
  if (ppp->samples) {
    CLAMP (ppp->samples, exception);
    ppp->subsrch = GetNextImageInList(ppp->samples);
  }
  if (ppp->subsrch) {
    CLAMP (ppp->subsrch, exception);
  }

  if (ppp->verbose) {
    WrImage ("master", ppp->master);
    WrImage ("canvas", ppp->canvas);
    WrImage ("saliency", ppp->saliency);
    WrImage ("samples", ppp->samples);
    WrImage ("subsrch", ppp->subsrch);
  }

  if (ppp->master->rows < 2) {
    fprintf (stderr, "Master < 2 rows.\n");
    return NULL;
  }

  if (ppp->samples) {
    if ((ppp->samples->columns <= ppp->master->columns) ||
        (ppp->samples->rows <= ppp->master->rows))
    {
      fprintf (stderr, "Samples smaller than master.\n");
      return NULL;
    }
  }

  if (ppp->samples && ppp->subsrch) {
    if ((ppp->samples->columns != ppp->subsrch->columns) ||
        (ppp->samples->rows != ppp->subsrch->rows))
    {
      fprintf (stderr, "Samples and subsrch are different sizes.\n");
      return NULL;
    }
  }

  if (!SetNoPalette (ppp->master, exception))
    return (Image *)NULL;

  if (ppp->canvas) {
    if (!SetNoPalette (ppp->canvas, exception)) return (Image *)NULL;
  } else {
    if (ppp->verbose) fprintf (stderr, "Creating canvas\n");

    ppp->canvas = CloneImage (
      ppp->master, ppp->master->columns, ppp->master->rows, MagickTrue, exception);
    if (!ppp->canvas) return NULL;
    SetAllBlack (ppp->canvas, exception);
  }

  if (ppp->verbose) WrImage ("canvas", ppp->canvas);

  if (ppp->saliency) {
    if (!SetNoPalette (ppp->saliency, exception)) return (Image *)NULL;
  } else {
    if (ppp->verbose) fprintf (stderr, "Creating saliency\n");

    ppp->saliency = CloneImage (
      ppp->master, ppp->master->columns, ppp->master->rows, MagickTrue, exception);
    if (!ppp->saliency) return NULL;
    SetAllOneCol (ppp->saliency, "gray(50%)", exception);
  }

  if (ppp->verbose) WrImage ("saliency", ppp->saliency);
  // FIXME: negate red channel?

  if (ppp->colFrom == cfUndef && ppp->samples != NULL)
    ppp->colFrom = cfSample;

  if (ppp->colFrom == cfSample && !ppp->samples) {
    fprintf (stderr, "colFrom Samples but no samples image.\n");
    return (Image *)NULL;
  }

  if (!ResolveUserDims (&ppp->patchDims, ppp->master->columns, ppp->master->rows, 0.0))
    return NULL;
  ppp->patchWi = ppp->patchDims.x.Pix;
  ppp->patchHt = ppp->patchDims.y.Pix;

  if (ppp->verbose)
    fprintf (stderr, "patches dims: %ix%i\n", ppp->patchWi, ppp->patchHt);

  InitRmseAlpha (&ppp->rmseAlpha);

  ssize_t prev_wrngX = -1, prev_wrngY = -1;
  int cnt_same = 0;
  int iter;
  for (iter=0; iter < ppp->maxIter; iter++) {

    if (!MostWrong (ppp, exception)) {
      fprintf (stderr, "MostWrong failed\n");
      return NULL;
    }

    if (ppp->verbose) {
      fprintf (stderr, "iter=%i wrng: %g @ %li,%li\n",
        iter, ppp->wrongness, ppp->wrngX, ppp->wrngY);
    }

    if (ppp->wrongness < ppp->wrngThresh) break;

    RectangleInfo patchRect;

    if (GetRect (ppp, &patchRect, exception)) {

      Image * pat_img = GetPatchImg (ppp, &patchRect, exception);
      if (!pat_img) return NULL;

      if (ppp->verbose > 1) WrImage ("Patch", pat_img);

      MagickBooleanType anyOpaque = MagickTrue;

      if (ppp->patchShape == psMin) {
        if (!TransNoImp (ppp, &patchRect, pat_img, &anyOpaque, exception)) {
          fprintf (stderr, "TransNoImp failed\n");
          return NULL;
        }
      }

      if (anyOpaque) {
        if (!PaintPatchCanvas (ppp, &patchRect, pat_img, exception))
          return NULL;

        if (!MultSalientPatch (ppp, &patchRect, pat_img, ppp->multSal, exception))
          return NULL;
      }

      pat_img = DestroyImage (pat_img);

      if (ppp->frameName && anyOpaque) {
        // For performance: don't need to create and destroy every time.
        Image * frame_img = CloneImage(ppp->canvas, 0, 0, MagickTrue, exception);
        if (frame_img == (Image *) NULL)
          return MagickFalse;

        WriteFrame (ppp, frame_img, exception);
        DestroyImage (frame_img);
      }

      cnt_same = 0;

      if (!MultOneSalient (ppp, ppp->wrngX, ppp->wrngY, 0.25, exception)) {
        fprintf (stderr, "MultOneSalient failed\n");
        return NULL;
      }
    }

    if (ppp->wrngX == prev_wrngX && ppp->wrngY == prev_wrngY) {
      cnt_same++;
      if (cnt_same > 3) {
        cnt_same = 0;

        if (!SetOneNonSalient (ppp, ppp->wrngX, ppp->wrngY, exception)) {
          fprintf (stderr, "SetOneSalient failed\n");
          return NULL;
        }
      }
    } else {
      prev_wrngX = ppp->wrngX;
      prev_wrngY = ppp->wrngY;
      cnt_same = 0;
    }
    if (ppp->verbose > 2) fprintf (stderr, "loop\n");
  }

  if (ppp->verbose) {
    fprintf (stderr, "Finished paintpatches iter=%i\n", iter);
    if (ppp->frameNum)
      fprintf (stderr, " Created %i frames\n", ppp->frameNum);
  }

  chkentry ("canvas1", &ppp->canvas);

  DeInitRmseAlpha (&ppp->rmseAlpha);

  Image * new_img = CloneImage (ppp->canvas, 0, 0, MagickTrue, exception);
  if (!new_img) {
    fprintf (stderr, "new_img: clone failed\n");
    return NULL;
  }

  // Destroy all the images except for master.

  Image * nxt_img = GetNextImageInList(ppp->master);
  while (nxt_img) {
    fprintf (stderr, "DeleteImageFromList: ");
    WrImage ("  difl", nxt_img);
    DeleteImageFromList (&nxt_img);
    nxt_img = GetNextImageInList(ppp->master);
  }

  chkentry ("new_img1", &new_img);
  return (new_img);
}


ModuleExport size_t paintpatchesImage (
  Image **images,
  const int argc,
  const char **argv,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  MagickBooleanType
    status;

  paintpatchesT ppp;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &ppp);
  if (status == MagickFalse)
    return (-1);

  ppp.precision = GetMagickPrecision();

  chklist ("main0", images);
  new_image = paintpatches (*images, &ppp, exception);
  if (!new_image) {
    fprintf (stderr, "No new_image\n");
    return -1;
  }
  chkentry ("main1 newimage", &new_image);
  chkentry ("main1 image", images);

  chkentry ("main1 bef repl", images);

  ReplaceImageInList (images, new_image);

  chkentry ("main1 aft repl", images);
  chkentry ("main1 after repl", &new_image);

  return(MagickImageFilterSignature);
}
