/*
    Created 30-Nov-2015

    Updated:
      3-April-2018 for v7.0.7-28
*/


/* Takes two images: reference and source, any size, possibly with transparency.

   Returns image same size as source,
   with red and green channels as relative or absolute displacement map,
   and blue channel as RMSE score.

   The map points to pixels in the reference that centre on windows that match source window.

   Output pixels corresponding to fully-transparent source pixels have
   null displacement and zero score.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "vsn_defines.h"

#include "chklist.h"

#include "compwind.h"
#include "match.h"

// If you want search windows clamped to the bounds of the reference image,
// as suggested in the PatchMatch paper, define CLAMP_SEARCH_WINDOW as 1.
//
// If you want my modification that avoids generating random offsets
// that would exceed the bounds,
// define CLAMP_SEARCH_WINDOW as 0.
//
#define CLAMP_SEARCH_WINDOW 0

typedef double pmValueT;

typedef enum {
  psToMatch, // Find a match for this pixel.
  psTrans,   // Pixel is transparent; don't try to match it.
  psMatched, // A good-enough match has been found.
} pixStateT;

// Valid state transitions:


typedef enum {
  dtAbsolute,
  dtRelative
} DispTypeT;

typedef struct {
  pixStateT
    State;

  ssize_t
    ref_x,
    ref_y;

  pmValueT
    score; // MSE (_not_ RMSE)

} PixT;


typedef struct {
  CompWindT
    CompWind;

  MatchT
    Match;

  MagickBooleanType
    do_verbose,
    AutoRepeat,
    WindowRad_IsPc,
    tanc,
    do_search,
    do_part,
    do_sweep,
    debug;

  double
    WindowRad;

  int
    WindowRadPix,
    MaxPartIter;

  DispTypeT
    dispType;

  char
    write_filename[MaxTextExtent];

  /* Calculated. */
  int
    sqDim;

  ssize_t
    refWidth,
    refHeight,
    srcWidth,
    srcHeight;

  PixT
    ** pix;

  RandomInfo
    *restrict random_info;

  Image
    *src_image;

  // Also: relative or absolute

} PixMatchT;


#include "compwind.inc"
#include "match.inc"

#include "writeframe.inc"




static void usage (void)
{
  printf ("Usage: -process 'pixmatch [OPTION]...'\n");
  printf ("Matches pixels from source in reference.\n");
  printf ("\n");
  printf ("  wr,  window_radius N        radius of search window, >= 0\n");
  printf ("  lsr, limit_search_radius N  limit radius to search, >= 0\n");
  printf ("                                default = 0 = no limit\n");
  printf ("  st,  similarity_threshold N\n");
  printf ("                              stop searching when RMSE <= N (eg 0.01)\n");
  printf ("                                default 0\n");
  printf ("  hc,  hom_chk X              homogeneity check, X is off or a small number\n");
  printf ("                                default 0.1\n");
  printf ("  e,   search_to_edges X      search for matches to image edges, on or off\n");
  printf ("                                default on\n");

  printf ("  s,   search X               X=none or entire or random or skip\n");
  printf ("                                default entire\n");
  printf ("  rs,  rand_searches N        number of random searches (eg 100)\n");
  printf ("                                default 0\n");
  printf ("  sn,  skip_num N             number of searches to skip in each direction\n");
  printf ("                                (eg 10)\n");
  printf ("                                default 0\n");
  printf ("  ref, refine X               whether to refine random and skip searches,\n");
  printf ("                                on or off\n");
  printf ("                                default on\n");

  printf ("  part, propagate_and_random_trial X\n");
  printf ("                              on or off\n");
  printf ("                                default on\n");
  printf ("  mpi,  max_part_iter N       maximum number of PART iterations, >= 1\n");
  printf ("                                0 = no maximum\n");
  printf ("                                default 10\n");
  printf ("  tanc, terminate_after_no_change X\n");
  printf ("                              on or off\n");
  printf ("                                default on\n");

  printf ("  swp, sweep X                on or off\n");
  printf ("                                default off\n");

  printf ("  dpt, displacement_type X    displacement type for input and output,\n");
  printf ("                                absolute or relative\n");
  printf ("                                default absolute\n");

  printf ("  ac,  auto_correct N         when RMSE score > N, attempt to correct (eg 0.02)\n");
  printf ("                                default 0 = no auto correction\n");

  //printf ("  a,   auto_repeat            if pixels changed but any unfilled, repeat\n");
  //printf ("  w,   write filename         write frames to files\n");
  printf ("  v,   verbose                write text information to stdout\n");
  printf ("\n"); 
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static inline MagickBooleanType EndsPc (const char *s)
{
  char c = *(s+strlen(s)-1);
  if (c == '%' || c == 'c')
    return MagickTrue;
  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  PixMatchT * pmh
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  pmh->do_verbose = MagickFalse;
  pmh->WindowRad = 1.0;
  pmh->WindowRad_IsPc = MagickFalse;
  pmh->WindowRadPix = 1;
  pmh->debug = MagickFalse;
  pmh->AutoRepeat = MagickFalse;
  pmh->MaxPartIter = 10;
  pmh->do_search = pmh->do_part = pmh->tanc = MagickTrue;
  pmh->do_sweep = MagickFalse;
  pmh->dispType = dtAbsolute;
  pmh->write_filename[0] = '\0';

  pmh->CompWind.HomChkOn = MagickTrue;
  pmh->CompWind.HomChk = 0.1;
  pmh->CompWind.nCompares = 0;

  SetDefaultMatch (&pmh->Match);
  pmh->Match.DoSubTransTest = MagickTrue;
  pmh->Match.AutoLs = MagickFalse;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "wr", "window_radius")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i]))
      {
        fprintf (stderr, "Non-numeric argument to 'window_radius'.\n");
        status = MagickFalse;
      }
      else {
        pmh->WindowRad = atof(argv[i]);
        pmh->WindowRad_IsPc = EndsPc (argv[i]);
        if (!pmh->WindowRad_IsPc)
          pmh->WindowRadPix = pmh->WindowRad + 0.5;
      }
      if (pmh->WindowRad < 0) {
        fprintf (stderr, "Bad 'window_radius' value.\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "lsr", "limit_search_radius")==MagickTrue) {
      // FIXME: also allow percentage of smaller image dimension.
      i++;
      if (!isdigit ((int)*argv[i])) {
        fprintf (stderr, "Non-numeric argument to 'limit_search_radius'.\n");
        status = MagickFalse;
      } else {
        pmh->Match.LimSrchRad = atof(argv[i]);
        pmh->Match.LimSrchRad_IsPc = EndsPc (argv[i]);
        if (!pmh->Match.LimSrchRad_IsPc)
          pmh->Match.limSrchRadX = pmh->Match.limSrchRadY = pmh->Match.LimSrchRad + 0.5;
      }
      if (pmh->Match.LimSrchRad < 0) {
        fprintf (stderr, "Bad 'limit_search_radius' value.\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "ref", "refine")==MagickTrue) {
      i++;
      if (LocaleCompare(argv[i], "on")==0)
        pmh->Match.Refine = MagickTrue;
      else if (LocaleCompare(argv[i], "off")==0)
        pmh->Match.Refine = MagickFalse;
      else {
        fprintf (stderr, "Invalid 'refine' [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "dpt", "displacement_type")==MagickTrue) {
      i++;
      if (LocaleCompare(argv[i], "absolute")==0)
        pmh->dispType = dtAbsolute;
      else if (LocaleCompare(argv[i], "relative")==0)
        pmh->dispType = dtRelative;
      else {
        fprintf (stderr, "Invalid 'displacement_type' [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "s", "search")==MagickTrue) {
      i++;
      if (LocaleCompare(argv[i], "entire")==0) {
        pmh->Match.searchWhat = swEntire;
        pmh->do_search = MagickTrue;
      } else if (LocaleCompare(argv[i], "random")==0) {
        pmh->Match.searchWhat = swRandom;
        pmh->do_search = MagickTrue;
      } else if (LocaleCompare(argv[i], "skip")==0) {
        pmh->Match.searchWhat = swSkip;
        pmh->do_search = MagickTrue;
      } else if (LocaleCompare(argv[i], "none")==0) {
        pmh->Match.searchWhat = swEntire;
        pmh->do_search = MagickFalse;
      } else {
        fprintf (stderr, "Invalid 'search' [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "rs", "rand_searches")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i])) {
        fprintf (stderr, "Non-numeric argument to 'rand_searches'.\n");
        status = MagickFalse;
      } else {
        pmh->Match.RandSearchesF = atof (argv[i]);
        pmh->Match.RandSearches_IsPc = EndsPc (argv[i]);
        if (!pmh->Match.RandSearches_IsPc)
          pmh->Match.RandSearchesI = pmh->Match.RandSearchesF + 0.5;
      }
      if (pmh->Match.RandSearchesF <= 0) {
        fprintf (stderr, "Bad 'rand_searches' value.\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "sn", "skip_num")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i])) {
        fprintf (stderr, "Non-numeric argument to 'skip_num'.\n");
        status = MagickFalse;
      } else {
        pmh->Match.SkipNumF = atof (argv[i]);
        pmh->Match.SkipNum_IsPc = EndsPc (argv[i]);
        if (!pmh->Match.SkipNum_IsPc)
          pmh->Match.SkipNumI = pmh->Match.SkipNumF + 0.5;
      }
      if (pmh->Match.SkipNumF <= 0) {
        fprintf (stderr, "Bad 'skip_num' value.\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "hc", "hom_chk")==MagickTrue) {
      i++;
      if (LocaleCompare(argv[i], "off")==0) {
        pmh->CompWind.HomChkOn = MagickFalse;
      } else {
        pmh->CompWind.HomChkOn = MagickTrue;
        if (!isdigit ((int)*argv[i])) {
          fprintf (stderr, "'hom_chk' argument must be number or 'off'.\n");
          status = MagickFalse;
        } else {
          pmh->CompWind.HomChk = atof (argv[i]);
        }
      }
    } else if (IsArg (pa, "part", "propagate_and_random_trial")==MagickTrue) {
      i++;
      if (LocaleCompare(argv[i], "off")==0) {
        pmh->do_part = MagickFalse;
      } else if (LocaleCompare(argv[i], "on")==0) {
        pmh->do_part = MagickTrue;
      } else {
        fprintf (stderr, "'part' argument must be 'on' or 'off'.\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "swp", "sweep")==MagickTrue) {
      i++;
      if (LocaleCompare(argv[i], "off")==0) {
        pmh->do_sweep = MagickFalse;
      } else if (LocaleCompare(argv[i], "on")==0) {
        pmh->do_sweep = MagickTrue;
      } else {
        fprintf (stderr, "'sweep' argument must be 'on' or 'off'.\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "tanc", "terminate_after_no_change")==MagickTrue) {
      i++;
      if (LocaleCompare(argv[i], "off")==0) {
        pmh->tanc = MagickFalse;
      } else if (LocaleCompare(argv[i], "on")==0) {
        pmh->tanc = MagickTrue;
      } else {
        fprintf (stderr, "'tanc' argument must be 'on' or 'off'.\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "e", "search_to_edges")==MagickTrue) {
      i++;
      if (LocaleCompare(argv[i], "on")==0)
        pmh->Match.SearchToEdges = MagickTrue;
      else if (LocaleCompare(argv[i], "off")==0)
        pmh->Match.SearchToEdges = MagickFalse;
      else {
        fprintf (stderr, "Invalid 'search_to_edges' [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "a", "auto_repeat")==MagickTrue) {
      pmh->AutoRepeat = MagickTrue;
    } else if (IsArg (pa, "st", "similarity_threshold")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i])) {
        fprintf (stderr, "Non-numeric argument to 'similarity_threshold'.\n");
        status = MagickFalse;
      } else {
        pmh->Match.simThreshold = atof (argv[i]);
      }
    } else if (IsArg (pa, "ac", "auto_correct")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i])) {
        fprintf (stderr, "Non-numeric argument to 'auto_correct'.\n");
        status = MagickFalse;
      } else {
        pmh->Match.autoCorrectThreshold = atof (argv[i]);
      }
    } else if (IsArg (pa, "mpi", "max_part_iter")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i])) {
        fprintf (stderr, "Non-numeric argument to 'max_part_iter'.\n");
        status = MagickFalse;
      } else {
        pmh->MaxPartIter = atoi (argv[i]);
      }
    } else if (IsArg (pa, "w", "write")==MagickTrue) {
      i++;
      CopyMagickString (pmh->write_filename, argv[i], MaxTextExtent);
      if (!*pmh->write_filename) {
        fprintf (stderr, "Invalid 'write' [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pmh->do_verbose = MagickTrue;
    } else if (IsArg (pa, "d", "debug")==MagickTrue) {
      pmh->debug = MagickTrue;
    } else {
      fprintf (stderr, "pixmatch: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  pmh->Match.simThresholdSq = pmh->Match.simThreshold * pmh->Match.simThreshold;

  pmh->Match.autoCorrectThresholdSq = pmh->Match.autoCorrectThreshold * pmh->Match.autoCorrectThreshold;

  if (pmh->do_verbose) {
    fprintf (stderr, "pixmatch options:\n");

    fprintf (stderr, "  window_radius %g", pmh->WindowRad);
    if (pmh->WindowRad_IsPc) fprintf (stderr, "%c", '%');

    fprintf (stderr, "  limit_search_radius %g", pmh->Match.LimSrchRad);
    if (pmh->Match.LimSrchRad_IsPc) fprintf (stderr, "%c", '%');

    fprintf (stderr, "  similarity_threshold %g", pmh->Match.simThreshold);

    fprintf (stderr, "  hom_chk ");
    if (pmh->CompWind.HomChkOn) {
      fprintf (stderr, "%g", pmh->CompWind.HomChk);
    } else {
      fprintf (stderr, "off");
    }

    if (!pmh->Match.SearchToEdges) fprintf (stderr, "  search_to_edges off");


    fprintf (stderr, "\n  search ");
    if (pmh->do_search == MagickFalse) {
      fprintf (stderr, "none");
    } else {
      switch (pmh->Match.searchWhat) {
        case swEntire: fprintf (stderr, "entire"); break;
        case swSkip:
          fprintf (stderr, "skip  skip_num %g", pmh->Match.SkipNumF);
          if (pmh->Match.SkipNum_IsPc) fprintf (stderr, "%c", '%');
          break;
        case swRandom:
          fprintf (stderr, "random  rand_searches %g", pmh->Match.RandSearchesF);
          if (pmh->Match.RandSearches_IsPc) fprintf (stderr, "%c", '%');
          break;
        default: fprintf (stderr, "??");
      }
    }

    if (pmh->do_search == MagickTrue) {
      fprintf (stderr, "  refine %s", pmh->Match.Refine ? "on" : "off");

      if (pmh->Match.autoCorrectThreshold > 0)
        fprintf (stderr, "  auto_correct %g", pmh->Match.autoCorrectThreshold);
    }


    fprintf (stderr, "\n  propagate_and_random_trial %s", pmh->do_part ? "on" : "off");

    if (pmh->do_part) {
      fprintf (stderr, "  max_part_iter %i", pmh->MaxPartIter);

      fprintf (stderr, "  tanc %s", pmh->tanc ? "on" : "off");
    }

    fprintf (stderr, "\n  sweep %s", pmh->do_sweep ? "on" : "off");


    fprintf (stderr, "\n  displacement_type ");
    switch (pmh->dispType) {
      case dtAbsolute: fprintf (stderr, "absolute"); break;
      case dtRelative: fprintf (stderr, "relative"); break;
      default: fprintf (stderr, " ??");
    }

    if (pmh->AutoRepeat) fprintf (stderr, "  auto_repeat");

    if (pmh->debug) fprintf (stderr, "  debug");

    if (pmh->do_verbose) fprintf (stderr, "  verbose");

    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


static void InitRand (PixMatchT * pmh)
{
  pmh->random_info=AcquireRandomInfo();

  // There seems to be a problem: the first few values show coherency,
  // so skip over them.

  int i;
  for (i=0; i < 20; i++) {
    GetPseudoRandomValue(pmh->random_info);
  }
}

static void DeInitRand (PixMatchT * pmh)
{
  pmh->random_info=DestroyRandomInfo(pmh->random_info);
}

static MagickBooleanType Initialise (const Image *image,
  PixMatchT * pmh,
  ExceptionInfo *exception
)
{
  ssize_t
    x, y;

  MagickBooleanType
    status = MagickTrue;

  //CacheView
  //  *image_view;

  // Allocate memory

  pmh->pix = (PixT **) AcquireQuantumMemory(pmh->srcHeight, sizeof(*pmh->pix));
  if (pmh->pix == (PixT **) NULL) {
    return MagickFalse;
  }
  for (y = 0; y < pmh->srcHeight; y++) {
    pmh->pix[y] = (PixT *) AcquireQuantumMemory(pmh->srcWidth, sizeof(**pmh->pix));
    if (pmh->pix[y] == (PixT *) NULL) break;
  }
  if (y < pmh->srcHeight) {
    for (y--; y >= 0; y--) {
      if (pmh->pix[y] != (PixT *) NULL)
        pmh->pix[y] = (PixT *) RelinquishMagickMemory(pmh->pix[y]);
    }
    pmh->pix = (PixT **) RelinquishMagickMemory(pmh->pix);
    return MagickFalse;
  }

  // Populate values

//#if defined(MAGICKCORE_OPENMP_SUPPORT)
//    #pragma omp parallel for schedule(static,4) shared(status)
//      MAGICK_THREADS(image,image,image->rows,1)
//#endif

  for (y = 0; y < pmh->srcHeight; y++) {
    VIEW_PIX_PTR const
      *p;

    if (status == MagickFalse)
      continue;

    p=GetCacheViewVirtualPixels(pmh->CompWind.sub_view,0,y,image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL)
      status=MagickFalse;

    for (x = 0; x < pmh->srcWidth; x++) {

      PixT * pn = &(pmh->pix[y][x]);

      pn->State = (GET_PIXEL_ALPHA (image, p) <= 0) ? psTrans: psToMatch;
      pn->ref_x = GetPseudoRandomValue(pmh->random_info) * pmh->refWidth;
      pn->ref_y = GetPseudoRandomValue(pmh->random_info) * pmh->refHeight;
      // fprintf (stderr, "%li,%li ", pn->ref_x, pn->ref_y);
      pn->score = 123;
assert (pn->ref_x >= 0 && pn->ref_x < pmh->refWidth && pn->ref_y >= 0 && pn->ref_y < pmh->refHeight);

      p += Inc_ViewPixPtr (image);
    }
  }
  return (status);
}

static void deInitialise (PixMatchT * pmh)
{
  ssize_t
    y;

  for (y = 0; y < pmh->srcHeight; y++) {
    pmh->pix[y] = (PixT *) RelinquishMagickMemory(pmh->pix[y]);
  }
  pmh->pix = (PixT **) RelinquishMagickMemory(pmh->pix);
}


static void CalcScores (PixMatchT * pmh,
  ExceptionInfo *exception)
{
  ssize_t
    y;

  if (pmh->debug) fprintf (stderr, "CalcScores\n");

  for (y = 0; y < pmh->srcHeight; y++) {
    ssize_t
      x;

    for (x = 0; x < pmh->srcWidth; x++) {
      PixT * pn = &(pmh->pix[y][x]);
      if (pn->State == psToMatch) {
assert (pn->ref_x >= 0 && pn->ref_x < pmh->refWidth && pn->ref_y >= 0 && pn->ref_y < pmh->refHeight);
        pmh->CompWind.refX = pn->ref_x - pmh->WindowRadPix;
        pmh->CompWind.refY = pn->ref_y - pmh->WindowRadPix;
        pmh->CompWind.subX = x - pmh->WindowRadPix;
        pmh->CompWind.subY = y - pmh->WindowRadPix;

        pmValueT v = CompareWindow (&pmh->CompWind, exception);
        if (pmh->debug) fprintf (stderr, "%li,%li v=%g ", x, y, v);
        pn->score = v;
        if (v < pmh->Match.simThresholdSq) pn->State = psMatched;
      }
    }
    if (pmh->debug) fprintf (stderr, "\n");
  }
}


static MagickBooleanType inline WithinRefX (PixMatchT * pmh, ssize_t x)
{
  return (x >= 0 && x < pmh->refWidth);
}

static MagickBooleanType inline WithinRefY (PixMatchT * pmh, ssize_t y)
{
  return (y >= 0 && y < pmh->refHeight);
}

static MagickBooleanType inline WithinSrcX (PixMatchT * pmh, ssize_t x)
{
  return (x >= 0 && x < pmh->srcWidth);
}

static MagickBooleanType inline WithinSrcY (PixMatchT * pmh, ssize_t y)
{
  return (y >= 0 && y < pmh->srcHeight);
}

static MagickBooleanType WhichProp (PixMatchT * pmh,
  ssize_t x,
  ssize_t y,
  int dxy,
  int * dx,
  int * dy)
// Enter with dxy=-1 for upper/left,
// or dxy=+1 for lower/right.
// If either has score lower than this,
// returns MagickTrue and sets dx, dy.
// Otherwise returns MagickFalse.
// (Could also do diagonal.))
{
  PixT *pn = &(pmh->pix[y][x]);
  pmValueT v = pn->score;

assert (pn->ref_x >= 0 && pn->ref_x < pmh->refWidth && pn->ref_y >= 0 && pn->ref_y < pmh->refHeight);
  PixT *pno;

  *dx = *dy = 0;

  if (pmh->debug) fprintf (stderr, "WhichProp %li,%li, %i\n", x, y, dxy);

  if (WithinSrcY(pmh, y+dxy)) {
    pno = &(pmh->pix[y+dxy][x]);
    if (v > pno->score && WithinRefY (pmh, pno->ref_y - dxy) )
    {
      *dy = dxy;
      v = pno->score;
    }
  }
  if (WithinSrcX(pmh, x+dxy)) {
    pno = &(pmh->pix[y][x+dxy]);
    if (v > pno->score && WithinRefX (pmh, pno->ref_x - dxy) )
    {
      *dy = 0;
      *dx = dxy;
      v = pno->score;
    }
  }

  if (pmh->debug) fprintf (stderr, "  dx,dy=%i,%i, v=%g\n", *dx, *dy, v);

  return (*dx != 0 || *dy != 0);
}

static int Propagate (PixMatchT * pmh,
  ssize_t x,
  ssize_t y,
  int dxy,
  ExceptionInfo *exception)
// Returns 1 if result changed; otherwise 0.
{
  int dx, dy;

  PixT * pn = &(pmh->pix[y][x]);
assert (pn->ref_x >= 0 && pn->ref_x < pmh->refWidth && pn->ref_y >= 0 && pn->ref_y < pmh->refHeight);

  int nChanges = 0;

  if (pmh->debug) fprintf (stderr, "prop %li,%li: v=%g ", x, y, pn->score);
  if (WhichProp (pmh, x, y, dxy, &dx, &dy)) {
    PixT * pno = &(pmh->pix[y+dy][x+dx]);
    ssize_t candX = pno->ref_x - dx;
    ssize_t candY = pno->ref_y - dy;
    if (pmh->debug) fprintf (stderr, "Cand %li,%li ", candX, candY);

    // Now try match.

    pmh->CompWind.subX = x - pmh->WindowRadPix;
    pmh->CompWind.subY = y - pmh->WindowRadPix;

    ssize_t limX = pmh->Match.limSrchRadX;
    ssize_t limY = pmh->Match.limSrchRadY;
    if (limX == 0) limX = 3;
    if (limY == 0) limY = 3;
    RangeT rng;

    CalcMatchRange (candX, candY, limX, limY, &pmh->Match, &rng);

    for (pmh->CompWind.refY = rng.i0; pmh->CompWind.refY < rng.i1; pmh->CompWind.refY++) {
      for (pmh->CompWind.refX = rng.j0; pmh->CompWind.refX < rng.j1; pmh->CompWind.refX++) {

        pmValueT v = CompareWindow (&pmh->CompWind, exception);
        if (pmh->debug) fprintf (stderr, "v=%g ", v);
        if (pn->score > v) {
          if (pmh->debug) fprintf (stderr, "better ");
          pn->score = v;
          pn->ref_x = pmh->CompWind.refX + pmh->WindowRadPix;
          pn->ref_y = pmh->CompWind.refY + pmh->WindowRadPix;
assert (pn->ref_x >= 0 && pn->ref_x < pmh->refWidth && pn->ref_y >= 0 && pn->ref_y < pmh->refHeight);
          if (v < pmh->Match.simThresholdSq) {
            pn->State = psMatched;
            break;
          }
          nChanges = 1;
        }
        if (pn->State == psMatched) break;
      }
    }
  }
  if (pmh->debug) fprintf (stderr, "\n");
  return nChanges;
}


static int RandTrial (PixMatchT * pmh,
  ssize_t x,
  ssize_t y,
  ExceptionInfo *exception)
// Returns 1 if result changed; otherwise 0.
{
  #define ALPHA (0.5)

  pmValueT fact = 1.0;

  ssize_t maxRad = pmh->refWidth;
  if (maxRad < pmh->refHeight) maxRad = pmh->refHeight;

  pmValueT rad = maxRad * fact;

  pmh->CompWind.subX = x - pmh->WindowRadPix;
  pmh->CompWind.subY = y - pmh->WindowRadPix;

  PixT * pn = &(pmh->pix[y][x]);

  int changes = 0;

  do {
    ssize_t tx, ty;

#if CLAMP_SEARCH_WINDOW==1
    tx = pn->ref_x + rad * (GetPseudoRandomValue(pmh->random_info)*2-1);
    ty = pn->ref_y + rad * (GetPseudoRandomValue(pmh->random_info)*2-1);

    if (tx < 0) tx = 0;
    if (ty < 0) ty = 0;
    if (tx >= pmh->refWidth)  tx = pmh->refWidth-1;
    if (ty >= pmh->refHeight) ty = pmh->refHeight-1;
#else
    do {
      tx = pn->ref_x + rad * (GetPseudoRandomValue(pmh->random_info)*2-1);
    } while (tx < 0 || tx >= pmh->refWidth);

    do {
      ty = pn->ref_y + rad * (GetPseudoRandomValue(pmh->random_info)*2-1);
    } while (ty < 0 || ty >= pmh->refHeight);
#endif

    if (pmh->debug) fprintf (stderr, "trial %li,%li ", tx, ty);

    pmh->CompWind.refX = tx - pmh->WindowRadPix;
    pmh->CompWind.refY = ty - pmh->WindowRadPix;

    pmValueT v = CompareWindow (&pmh->CompWind, exception);
    //if (pmh->debug) fprintf (stderr, "v=%g ", v);
    if (pn->score > v) {
      if (pmh->debug) fprintf (stderr, "better %li,%li,%g=>%li,%li,%g ", x, y, pn->score, tx, ty, v);
      pn->score = v;
      pn->ref_x = tx;
      pn->ref_y = ty;

assert (pn->ref_x >= 0 && pn->ref_x < pmh->refWidth && pn->ref_y >= 0 && pn->ref_y < pmh->refHeight);

      if (v < pmh->Match.simThresholdSq) pn->State = psMatched;
      changes = 1;
    }

    fact *= ALPHA;
    rad = maxRad * fact;
  } while (rad > 1.0 && pn->State != psMatched);

  return (changes);
}

static int OnePart (PixMatchT * pmh,
  ExceptionInfo *exception)
// Returns the number of changes made.
{
  ssize_t
    y;

  int nChanges = 0;
  int nToMatch = 0;

  for (y = 0; y < pmh->srcHeight; y++) {
    ssize_t
      x;

    for (x = 0; x < pmh->srcWidth; x++) {
      PixT * pn = &(pmh->pix[y][x]);
      if (pn->State == psToMatch)
        nChanges += Propagate (pmh, x, y, -1, exception);
      if (pn->State == psToMatch)
        nChanges += RandTrial (pmh, x, y, exception);
    }
  }

  for (y = pmh->srcHeight-1; y >= 0; y--) {
    ssize_t
      x;

    for (x = pmh->srcWidth-1; x >= 0; x--) {
      PixT * pn = &(pmh->pix[y][x]);
      if (pn->State == psToMatch)
        nChanges += Propagate (pmh, x, y, +1, exception);
      if (pn->State == psToMatch)
        nChanges += RandTrial (pmh, x, y, exception);
      if (pn->State == psToMatch) nToMatch++;
    }
  }

  if (pmh->do_verbose) fprintf (stderr, "OnePart: nChanges=%i nToMatch=%i\n", nChanges, nToMatch);

  return nChanges;
}


static int Sweep (PixMatchT * pmh,
  ExceptionInfo *exception)
// Returns the number of changes made.
{
  ssize_t
    y;

  RangeT
    rng;

  CalcMatchRange (0, 0, 0, 0, &pmh->Match, &rng);

  int nChanges = 0;

  ssize_t
    besti=0, bestj=0;

  double v;

  CompWindT * cmpwin = &pmh->CompWind;

  for (y = 0; y < pmh->srcHeight; y++) {
    ssize_t
      x;

    for (x = 0; x < pmh->srcWidth; x++) {
      PixT * pn = &(pmh->pix[y][x]);
      if (pn->State == psToMatch) {

        double bestScore = pn->score;

        cmpwin->subX = x - pmh->Match.matchRadX;
        cmpwin->subY = y - pmh->Match.matchRadY;

        for (cmpwin->refY = rng.i0; cmpwin->refY < rng.i1; cmpwin->refY++) {
          for (cmpwin->refX = rng.j0; cmpwin->refX < rng.j1; cmpwin->refX++) {
            v = CompareWindow (cmpwin, exception);
            if (bestScore > v) {
              bestScore = v;
              besti = cmpwin->refY;
              bestj = cmpwin->refX;
              if (bestScore <= pmh->Match.simThresholdSq) break;
            }
          }
          if (bestScore <= pmh->Match.simThresholdSq) break;
        }

        if (pn->score > bestScore) {
          pn->ref_x = bestj + pmh->Match.matchRadX;
          pn->ref_y = besti + pmh->Match.matchRadY;
          pn->score = bestScore;
          nChanges++;
        }
      }

    }  // for x
  } // for y

  if (pmh->do_verbose) fprintf (stderr, "Sweep: nChanges=%i\n", nChanges);

  return nChanges;
}


static Image *MkDispImage (
  PixMatchT * pmh,
  ExceptionInfo *exception)
// Make an image from the pixel data.
{
  Image
    *new_image;

  MagickBooleanType
    status = MagickTrue;

  CacheView
    *new_view;

  ssize_t
    y;

  const long double
    factX = QuantumRange / (long double)(pmh->refWidth-1),
    factY = QuantumRange / (long double)(pmh->refHeight-1);

  if (pmh->debug) fprintf (stderr, "MkDispImage\n");

  // Clone the image, same size, copied pixels:
  //
  new_image=CloneImage(pmh->src_image, 0, 0, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  new_view = AcquireAuthenticCacheView (new_image, exception);

  chkentry("mdi",&new_image);

#if defined(MAGICKCORE_OPENMP_SUPPORT)
    #pragma omp parallel for schedule(static,4) shared(status) \
      MAGICK_THREADS(new_image,new_image,new_image->rows,1)
#endif
  for (y = 0; y < pmh->srcHeight; y++) {
    VIEW_PIX_PTR
      *p;

    ssize_t
      x;

    p = GetCacheViewAuthenticPixels(new_view,0,y,new_image->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL) status = MagickFalse;

    if (status == MagickFalse)
      continue;

    for (x = 0; x < pmh->srcWidth; x++) {

      // FIXME: plus half if not HDRI.
      PixT * pn = &(pmh->pix[y][x]);
assert (pn->ref_x >= 0 && pn->ref_x < pmh->refWidth && pn->ref_y >= 0 && pn->ref_y < pmh->refHeight);
      if (pmh->dispType == dtAbsolute) {
        SET_PIXEL_RED   (new_image, pn->ref_x * factX, p);
        SET_PIXEL_GREEN (new_image, pn->ref_y * factY, p);
      } else {
        SET_PIXEL_RED   (new_image, ((pn->ref_x - x) / (pmValueT)(2*pmh->refWidth) + 0.5) * QuantumRange, p);
        SET_PIXEL_GREEN (new_image, ((pn->ref_y - y) / (pmValueT)(2*pmh->refHeight) + 0.5) * QuantumRange, p);
      }
      SET_PIXEL_BLUE  (new_image, sqrt(pn->score) * QuantumRange, p);
      SET_PIXEL_ALPHA (new_image, (pn->score > 4) ? 0 : QuantumRange, p);

      p += Inc_ViewPixPtr (new_image);
    }
  }

  if (SyncCacheViewAuthenticPixels (new_view,exception) == MagickFalse)
    return (Image *)NULL;

  new_view=DestroyCacheView (new_view);

  chkentry("mdi2",&new_image);

  return (new_image);
}


static MagickBooleanType ReadPrevResult (
  PixMatchT * pmh,
  Image * prevResult,
  ExceptionInfo *exception)
{
  CacheView
    *view;

  MagickBooleanType
    status = MagickTrue;

  ssize_t
    y;

  const long double
    factX = QuantumRange / (long double)(pmh->refWidth-1),
    factY = QuantumRange / (long double)(pmh->refHeight-1);

  if (pmh->do_verbose) fprintf (stderr, "ReadPrevResult\n");

  if (SetNoPalette (prevResult, exception) == MagickFalse)
    return MagickFalse;

  if (prevResult->rows != pmh->srcHeight || prevResult->columns != pmh->srcWidth) {
    fprintf (stderr, "Sizes of source and previous result do not match.\n");
    return MagickFalse;
  }

  view = AcquireVirtualCacheView (prevResult, exception);

#if defined(MAGICKCORE_OPENMP_SUPPORT)
    #pragma omp parallel for schedule(static,4) shared(status) \
      MAGICK_THREADS(prevResult,prevResult,prevResult->rows,1)
#endif
  for (y = 0; y < pmh->srcHeight; y++) {
    const VIEW_PIX_PTR
      *p;

    ssize_t
      x;

    p = GetCacheViewVirtualPixels(view,0,y,prevResult->columns,1,exception);
    if (p == (const VIEW_PIX_PTR *) NULL) status = MagickFalse;

    if (status == MagickFalse)
      continue;

    for (x = 0; x < pmh->srcWidth; x++) {
      PixT * pn = &(pmh->pix[y][x]);

      if (pmh->dispType == dtAbsolute) {
        pn->ref_x = GET_PIXEL_RED   (prevResult, p) / factX;
        pn->ref_y = GET_PIXEL_GREEN (prevResult, p) / factY;
      } else {
        pn->ref_x = (GET_PIXEL_RED   (prevResult, p) / QuantumRange - 0.5)
                    * 2*pmh->refWidth;
        pn->ref_y = (GET_PIXEL_GREEN (prevResult, p) / QuantumRange - 0.5)
                    * 2*pmh->refHeight;
      }

      // The read image may have been resized, maybe with ringing,
      // possibly causing values to go out of range,
      // so we need to clamp.

      if (pn->ref_x < 0) pn->ref_x = 0;
      else if (pn->ref_x >= pmh->refWidth) pn->ref_x = pmh->refWidth-1;

      if (pn->ref_y < 0) pn->ref_y = 0;
      else if (pn->ref_y >= pmh->refHeight) pn->ref_y = pmh->refHeight-1;

assert (pn->ref_x >= 0 && pn->ref_x < pmh->refWidth && pn->ref_y >= 0 && pn->ref_y < pmh->refHeight);

      // The score may have been upscaled, so might not be accurate.

      pn->score = GET_PIXEL_BLUE (prevResult, p) / QuantumRange;
      pn->score *= pn->score;
      pn->State = psToMatch;   // Until we know better.
      p += Inc_ViewPixPtr (prevResult);
    }
  }

  view=DestroyCacheView (view);

  return MagickTrue;
}

static void IterParts (PixMatchT * pmh,
  ExceptionInfo *exception)
// Returns the number of changes made.
{
  int nChanges = 0;
  int nIter = 0;

  for (;;) {
    nChanges = OnePart (pmh, exception);

    if (*pmh->write_filename) {
      Image *tmp_image = MkDispImage (pmh, exception);
      WriteFrame (pmh->write_filename, tmp_image, nIter, exception);
      DestroyImage (tmp_image);
    }

    if (nChanges == 0 && pmh->tanc) break;
    nIter++;
    if (pmh->MaxPartIter > 0 && nIter >= pmh->MaxPartIter) break;
  }

  if (pmh->do_verbose) fprintf (stderr, "IterParts: nIter=%i\n", nIter);

  // FIXME: Also terminate if all pixels are state=matched
}


#if 0
static void DumpPixels (PixMatchT * pmh)
{
  ssize_t
    x, y;

  fprintf (stderr, "DumpPixels\n");

  for (y = 0; y < pmh->srcHeight; y++) {
    for (x = 0; x < pmh->srcWidth; x++) {
      PixT * pn = &(pmh->pix[y][x]);
      char cState;
      switch (pn->State) {
        case psToMatch: cState='m'; break;
        case psTrans:   cState='t'; break;
        case psMatched: cState='d'; break;
        default: cState = '?';
      }
      fprintf (stderr, "%li,%li %c %li,%li %g\n",
        x, y, cState,pn->ref_x, pn->ref_y, pn->score );
    }
  }
}
#endif


static void SearchPixels (
  PixMatchT * pmh,
  MagickBooleanType UsePrev,
  ExceptionInfo *exception)
{
  ssize_t
    y;

  //printf ("SearchPixels %i\n", (int)UsePrev);

  ssize_t
    nChanges = 0;

  if (UsePrev) {
    for (y = 0; y < pmh->srcHeight; y++) {
      ssize_t
        x;

      for (x = 0; x < pmh->srcWidth; x++) {
        PixT * pn = &(pmh->pix[y][x]);

        if (pn->State != psMatched) {
          // We could attempt propagation here.

          Match (x, y, pn->ref_x, pn->ref_y,
            &pmh->Match, &pmh->CompWind, pmh->random_info, exception);

          if (pn->score > pmh->Match.bestScore) {
            pn->ref_x = pmh->Match.bestX;
            pn->ref_y = pmh->Match.bestY;
assert (pn->ref_x >= 0 && pn->ref_x < pmh->refWidth && pn->ref_y >= 0 && pn->ref_y < pmh->refHeight);
            pn->score = pmh->Match.bestScore;
            nChanges++;
            if (pn->score < pmh->Match.simThresholdSq) pn->State = psMatched;
          }
        }
      }
    }
  } else {
    for (y = 0; y < pmh->srcHeight; y++) {
      ssize_t
        x;

      for (x = 0; x < pmh->srcWidth; x++) {
        PixT * pn = &(pmh->pix[y][x]);
        if (pn->State != psMatched) {
          Match (x, y, x, y,
            &pmh->Match, &pmh->CompWind, pmh->random_info, exception);
          pn->ref_x = pmh->Match.bestX;
          pn->ref_y = pmh->Match.bestY;
assert (pn->ref_x >= 0 && pn->ref_x < pmh->refWidth && pn->ref_y >= 0 && pn->ref_y < pmh->refHeight);
          pn->score = pmh->Match.bestScore;
          nChanges++;
          if (pn->score < pmh->Match.simThresholdSq) pn->State = psMatched;
        }
      }
    }
  }
  if (pmh->do_verbose) fprintf (stderr, "Search: nChanges=%li\n", nChanges);
}


// The next function takes two or three images, and returns one image.
//
static Image *pixmatch (
  PixMatchT * pmh,
  Image *ref_image,
  Image *src_image,
  Image *prev_image,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  MagickBooleanType
    UsePrev = MagickFalse;

  if (ref_image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",ref_image->filename);

  pmh->refWidth  = ref_image->columns;
  pmh->refHeight = ref_image->rows;
  pmh->srcWidth  = src_image->columns;
  pmh->srcHeight = src_image->rows;

  if (pmh->do_verbose) {
    fprintf (stderr, "---------------------------------------\npixmatch: ref %lix%li  src %lix%li\n",
      pmh->refWidth, pmh->refHeight, pmh->srcWidth, pmh->srcHeight);
  }

  pmh->CompWind.ref_image = ref_image;
  pmh->CompWind.sub_image = src_image;
  pmh->CompWind.ref_view = AcquireVirtualCacheView (ref_image,exception);
  pmh->CompWind.sub_view = AcquireVirtualCacheView (src_image,exception);

  pmh->src_image = src_image;

  if (Initialise (src_image, pmh, exception) == MagickFalse)
    return (Image *) NULL;

  if (prev_image != (Image *) NULL) {
    MagickBooleanType r = ReadPrevResult (pmh, prev_image, exception);
    if (r == MagickFalse) return (Image *) NULL;
    UsePrev = MagickTrue;
  }

  pmh->CompWind.win_wi = pmh->WindowRadPix*2+1;
  pmh->CompWind.win_ht = pmh->WindowRadPix*2+1;
  pmh->CompWind.WorstCompare = 9e11;
  pmh->CompWind.AllowEquCoord = MagickTrue;

  pmh->Match.ref_rows = ref_image->rows;
  pmh->Match.ref_columns = ref_image->columns;
  pmh->Match.matchRadX = pmh->WindowRadPix;
  pmh->Match.matchRadY = pmh->WindowRadPix;

  pmh->Match.AutoLs = MagickFalse;
  pmh->Match.DissimOn = MagickFalse;

  ssize_t nTotalCompares = 0;

  pmh->CompWind.nCompares = 0;
  CalcScores (pmh, exception);
  if (pmh->do_verbose) fprintf (stderr, "CalcScores: nCompares=%li\n", pmh->CompWind.nCompares);
  nTotalCompares += pmh->CompWind.nCompares;

  if (pmh->do_search) {
    pmh->CompWind.nCompares = 0;
    SearchPixels (pmh, UsePrev, exception);
    if (pmh->do_verbose) fprintf (stderr, "Search: nCompares=%li\n", pmh->CompWind.nCompares);
    nTotalCompares += pmh->CompWind.nCompares;
  }

  //if (pmh->debug) DumpPixels (pmh);

  if (pmh->do_part) {
    pmh->CompWind.nCompares = 0;
    IterParts (pmh, exception);
    if (pmh->do_verbose) fprintf (stderr, "IterParts: nCompares=%li\n", pmh->CompWind.nCompares);
    nTotalCompares += pmh->CompWind.nCompares;
  }

  if (pmh->do_sweep) {
    pmh->CompWind.nCompares = 0;
    Sweep (pmh, exception);
    if (pmh->do_verbose) fprintf (stderr, "Sweep: nCompares=%li\n", pmh->CompWind.nCompares);
    nTotalCompares += pmh->CompWind.nCompares;
  }

  if (pmh->do_verbose) fprintf (stderr, "nTotalCompares=%li\n", nTotalCompares);

  pmh->CompWind.sub_view=DestroyCacheView (pmh->CompWind.sub_view);
  pmh->CompWind.ref_view=DestroyCacheView (pmh->CompWind.ref_view);

  new_image = MkDispImage (pmh, exception);
  if (new_image == (Image *) NULL) {
    fprintf (stderr, "MkDispImage failed\n");
    return (new_image);
  }

  if (pmh->debug) fprintf (stderr, "deInitialise\n");
  deInitialise (pmh);

  chkentry("be",&new_image);

  return (new_image);
}


ModuleExport size_t pixmatchImage (
  Image **images,
  const int argc,
  const char **argv,
  ExceptionInfo *exception)
{
  Image
    *image,
    *ref_image,
    *src_image,
    *prev_image,
    *new_image;

  MagickBooleanType
    status;

  PixMatchT
    pm;

  int
    ListLen;

  (void) argc;
  (void) argv;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &pm);
  if (status == MagickFalse)
    return (-1);

  InitRand (&pm);

  chklist("chkL A", images);

  ListLen = (int)GetImageListLength(*images);
  if (ListLen < 2 || ListLen > 3) {
    fprintf (stderr, "pixmatch needs 2 or 3 images\n");
    return (-1);
  }

  image=(*images);
  ref_image = image;
  image=GetNextImageInList(image);
  src_image = image;
  if (ListLen == 3) {
    image=GetNextImageInList(image);
    prev_image = image;
  } else {
    prev_image = (Image *)NULL;
  }

  new_image = pixmatch(&pm, ref_image, src_image, prev_image, exception);
  if (new_image == (Image *) NULL) {
    fprintf (stderr, "pixmatch failed.\n");
    status = MagickFalse;
  }

  DeleteImageFromList (&src_image);
  if (ListLen == 3) {
    DeleteImageFromList (&prev_image);
  }
  image = ref_image;

  if (status == MagickTrue) {
    ReplaceImageInList(&image,new_image);
  }

  chklist("after rem", &image);

  *images=GetFirstImageInList(image);
  chklist("after rem list", images);

  DeInitRand (&pm);

  if (status == MagickFalse) return -1;

  return(MagickImageFilterSignature);
}
