/* Updated:
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <float.h>
#include "vsn_defines.h"

#define VERSION "plotrg v1.0  Copyright (c) 2018 Alan Gibson"

#define countT double

typedef struct {

  int countsDim;  // width and height

  int precision;

  MagickBooleanType
    verbose,
    regardAlpha,
    do_norm,
    do_norm_sum,
    calcOnly;

  FILE *
    fh_data;

  countT
    *counts,
    minCount,
    maxCount,
    numOutside;

  MagickRealType
    minVal,
    maxVal;

} plotrgT;


static void usage (void)
{
  printf ("Usage: -process 'plotrg [OPTION]...'\n");
  printf ("plot histogram of Red and Green channels.\n");
  printf ("\n");
  printf ("  d,  dim integer      width and height of square output\n");
  printf ("  ra, regardAlpha      use alpha for increment\n");
  printf ("  n,  norm             make max count 100%% of quantum\n");
  printf ("  ns, normSum          make sum of counts 100%% of quantum\n");
  printf ("  co, calcOnly         calculate only; don't replace image\n");
  printf ("  f,  file string      write to file stream stdout or stderr\n");
  printf ("  v,  verbose          write text information to stdout\n");
  printf ("      version          write version information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "plotrg: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  plotrgT * pprg
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status = MagickTrue;

  pprg->countsDim = 256;
  pprg->verbose = MagickFalse;
  pprg->regardAlpha = MagickFalse;
  pprg->do_norm = MagickFalse;
  pprg->do_norm_sum = MagickFalse;
  pprg->calcOnly = MagickFalse;
  pprg->fh_data = stderr;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "d", "dim")==MagickTrue) {
      NEXTARG;
      pprg->countsDim = atoi(argv[i]);
    } else if (IsArg (pa, "ra", "regardAlpha")==MagickTrue) {
      pprg->regardAlpha = MagickTrue;
    } else if (IsArg (pa, "n", "norm")==MagickTrue) {
      pprg->do_norm = MagickTrue;
      pprg->do_norm_sum = MagickFalse;
    } else if (IsArg (pa, "ns", "normSum")==MagickTrue) {
      pprg->do_norm = MagickFalse;
      pprg->do_norm_sum = MagickTrue;
    } else if (IsArg (pa, "co", "calcConly")==MagickTrue) {
      pprg->calcOnly = MagickTrue;
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      NEXTARG;
      if (strcasecmp (argv[i], "stdout")==0) pprg->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) pprg->fh_data = stderr;
      else status = MagickFalse;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pprg->verbose = MagickTrue;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "plotrg: ERROR: unknown option\n");
      status = MagickFalse;
    }
  }

  if (pprg->countsDim < 1) {
    fprintf (stderr, "dim must be at least one\n");
    status = MagickFalse;
  }

  if (pprg->verbose) {
    fprintf (stderr, "plotrg options:");
    fprintf (stderr, "  dim %i", pprg->countsDim);
    if (pprg->regardAlpha) fprintf (stderr, "  regardAlpha");
    if (pprg->verbose) fprintf (stderr, "  verbose");
    if (pprg->do_norm) fprintf (stderr, "  normalise");
    if (pprg->do_norm_sum) fprintf (stderr, "  normaliseSum");
    if (pprg->calcOnly) fprintf (stderr, "  calcOnly");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


// The next function takes an image,
// and returns it or a new image.
//
static Image * plotrg (
  Image *image,
  plotrgT * pprg,
  ExceptionInfo *exception)
{
  MagickBooleanType
    status = MagickTrue;

  int nArray = pprg->countsDim*pprg->countsDim;

  if (nArray < 1) return NULL;

  pprg->counts = (countT *)malloc(nArray*sizeof(countT));
  if (!pprg->counts) {
    fprintf (stderr, "Array allocation failed.\n");
    return NULL;
  }

  int i;
  countT * pcnt = pprg->counts;
  for (i=0; i < nArray; i++) *(pcnt++) = 0;

  CacheView * image_view = AcquireVirtualCacheView (image,exception);

  MagickRealType
    minR = DBL_MAX, minG = DBL_MAX,
    maxR = -DBL_MAX, maxG = -DBL_MAX;

  MagickRealType
    fact = pprg->countsDim / (MagickRealType)(QuantumRange+1);

  double numOOG = 0;
  double totalCnt = 0;

  ssize_t y;
  for (y = 0; y < image->rows; y++) {
    VIEW_PIX_PTR const
      *p;

    if (status == MagickFalse)
      continue;

    p=GetCacheViewVirtualPixels(image_view,0,y,image->columns,1,exception);
    if (!p) {
      status=MagickFalse;
      continue;
    }

    MagickRealType r, g;

    MagickRealType alph = 1.0;

    ssize_t x;
    for (x = 0; x < image->columns; x++) {
      r = GET_PIXEL_RED   (image, p);
      g = GET_PIXEL_GREEN (image, p);

      if (pprg->regardAlpha)
        alph = GET_PIXEL_ALPHA (image, p) / (MagickRealType)QuantumRange;

      if (minR > r) minR = r;
      if (minG > g) minG = g;
      if (maxR < r) maxR = r;
      if (maxG < g) maxG = g;

      int buckR = floor (fact * r);
      int buckG = floor (fact * g);

//      printf ("%i %i  ", buckR, buckG);

      if (buckR < 0 || buckG < 0 ||
          buckR >= pprg->countsDim || buckG >= pprg->countsDim)
      {
        numOOG += alph;
      } else {
        pprg->counts[buckG * pprg->countsDim + buckR] += alph;
        totalCnt += alph;
      }

      p += Inc_ViewPixPtr (image);
    }
  }

  if (!status) return NULL;

  image_view = DestroyCacheView (image_view);

  if (pprg->verbose) {
    fprintf (pprg->fh_data,
      "totalCnt=%.*g\n",
      pprg->precision, totalCnt);

    fprintf (pprg->fh_data,
      "minR=%.*g\nmaxR=%.*g\n",
      pprg->precision, minR / (double)QuantumRange,
      pprg->precision, maxR / (double)QuantumRange);

    fprintf (pprg->fh_data,
      "minG=%.*g\nmaxG=%.*g\n",
      pprg->precision, minG / (double)QuantumRange,
      pprg->precision, maxG / (double)QuantumRange);

    fprintf (pprg->fh_data,
      "minVal=%.*g\nmaxVal=%.*g\n",
      pprg->precision,
        ((minR<minG) ? minR : minG) / (double)QuantumRange,
      pprg->precision,
        ((maxR>maxG) ? maxR : maxG) / (double)QuantumRange);

    fprintf (pprg->fh_data,
      "numOOG=%.*g\n", pprg->precision, numOOG);
  }

  countT minCnt, maxCnt;
  minCnt = maxCnt = pprg->counts[0];
  pcnt = pprg->counts;
  for (i=0; i < nArray; i++) {
    if (minCnt > *pcnt) minCnt = *pcnt;
    if (maxCnt < *pcnt) maxCnt = *pcnt;
    pcnt++;
  }

  if (pprg->verbose) {
    fprintf (pprg->fh_data,
      "minCnt=%.*g\nmaxCnt=%.*g\n",
      pprg->precision, minCnt,
      pprg->precision, maxCnt);
  }

  if (pprg->do_norm) {
    pcnt = pprg->counts;
    MagickRealType fact = QuantumRange / (MagickRealType)maxCnt;
    for (i=0; i < nArray; i++) {
      *pcnt = *pcnt * fact;
      pcnt++;
    }
  } else if (pprg->do_norm_sum) {
    pcnt = pprg->counts;
    MagickRealType fact = QuantumRange / (MagickRealType)totalCnt;
    for (i=0; i < nArray; i++) {
      *pcnt = *pcnt * fact;
      pcnt++;
    }
  }


  Image * newImg = NULL;

  if (!pprg->calcOnly) {
    newImg = CloneImage (image, pprg->countsDim, pprg->countsDim, MagickTrue, exception);
    if (!newImg) return NULL;

    CacheView * out_view = AcquireAuthenticCacheView (newImg, exception);

    for (y = 0; y < newImg->rows; y++) {
      if (status == MagickFalse)
        continue;

      VIEW_PIX_PTR *q = GetCacheViewAuthenticPixels(
        out_view, 0, y, newImg->columns, 1, exception);
      if (!q) {
        status=MagickFalse;
        continue;
      }

      pcnt = &pprg->counts[y * pprg->countsDim];
      ssize_t x;
      for (x = 0; x < newImg->columns; x++) {
        SET_PIXEL_RED (newImg, *pcnt, q);
        SET_PIXEL_GREEN (newImg, *pcnt, q);
        SET_PIXEL_BLUE (newImg, *pcnt, q);
        SET_PIXEL_ALPHA (newImg, QuantumRange, q);
        q += Inc_ViewPixPtr (newImg);
        pcnt++;
      }

      if (!SyncCacheViewAuthenticPixels(out_view,exception)) {
        fprintf (stderr, "bad sync\n");
        status=MagickFalse;
      }
    }
    out_view = DestroyCacheView (out_view);
  }
  if (!status) return NULL;

  free (pprg->counts);

  return (newImg) ? newImg : image;
}



ModuleExport size_t plotrgImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  MagickBooleanType
    status;

  plotrgT
    prg;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  prg.precision = GetMagickPrecision();

  status = menu (argc, argv, &prg);
  if (status == MagickFalse)
    return (-1);

  Image * image;
  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    Image * newImg = plotrg (image, &prg, exception);
    if (!newImg) return (-1);

    if (newImg != image) {
      ReplaceImageInList (&image, newImg);
      *images=GetFirstImageInList (newImg);
    }
  }

  return(MagickImageFilterSignature);
}
