#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
    Extract JPEG image files from a JPEG MPF (Multi-Picture Format) file.

    Non-nested SOI/EOI pairs.

    If the input has only one SOI/EOI pair, it is copied to a single output,
    stripping any APP2 MPF segments if present in the input.

    FIXME: for HDR gain maps, we could also usefully remove the XMP tag.
*/

/*
  bash snibgo\buildcore.sh snibgo\rdjpgmpf

  tests:
  %IM7DEV%rdjpgmpf c:\web\im\DSCF0615.MPO
  %IM7DEV%rdjpgmpf c:\web\im\motorcycles.MPO
  %IM7DEV%rdjpgmpf c:\web\im\120377_CyprusProtarasRegion_W3D.MPO
  %IM7DEV%rdjpgmpf c:web\im\\GainMap\Gain_Map_Sample_Photos\samples_jpeg\01.jpg

  References:
     https://en.wikipedia.org/wiki/JPEG
     https://www.w3.org/Graphics/JPEG/jfif3.pdf
     https://www.w3.org/Graphics/JPEG/itu-t81.pdf Table B.1
     https://www.cipa.jp/std/documents/e/DC-X008-Translation-2019-E.pdf  Section 4.7.1
     https://www.cipa.jp/std/documents/download_e.html?CIPA_DC-007-2021_E

*/


#ifndef BOOL
#define BOOL int
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif



#define MARKER_PREFIX '\xFF'

#define SOI_MARKER '\xD8'

// Markers from 0xC0 to 0xDF
typedef struct {
  char * Symbol;
  char * Description;
  BOOL NeedLen;
} MarkerT;

MarkerT Markers[] = {
  { "SOF0",  "Baseline DCT",                    1},
  { "SOF1",  "Extended sequential DCT",         1},
  { "SOF2",  "Progressive DCT",                 1},
  { "SOF3",  "Lossless (sequential)",           1},
  { "DHT",   "Define Huffman table(s)",         1},
  { "SOF5",  "Differential sequential DCT",     1},
  { "SOF6",  "Differential progressive DCT",    1},
  { "SOF7",  "Differential lossless (sequential)", 1},
  { "JPG",   "Reserved for JPEG extensions",    1},
  { "SOF9",  "Extended sequential DCT",         1},
  { "SOF10", "Progressive DCT",                 1},
  { "SOF11", "Lossless (sequential)",           1},
  { "DAC",   "Define arithmetic coding conditioning(s)", 1},
  { "SOF13", "Differential sequential DCT",     1},
  { "SOF14", "Differential progressive DCT",    1},
  { "SOF15", "Differential lossless (sequential)", 1},
  { "RST0",  "Restart with modulo 8 count 0",   0},
  { "RST1",  "Restart with modulo 8 count 1",   0},
  { "RST2",  "Restart with modulo 8 count 2",   0},
  { "RST3",  "Restart with modulo 8 count 3",   0},
  { "RST4",  "Restart with modulo 8 count 4",   0},
  { "RST5",  "Restart with modulo 8 count 5",   0},
  { "RST6",  "Restart with modulo 8 count 6",   0},
  { "RST7",  "Restart with modulo 8 count 7",   0},
  { "SOI",   "Start of image",                  0}, 
  { "EOI",   "End of image",                    0}, 
  { "SOS",   "Start of scan",                   1},
  { "DQT",   "Define quantization table(s)",    1},
  { "DNL",   "Define number of lines",          1},
  { "DRI",   "Define restart interval",         1},
  { "DHP",   "Define hierarchical progression", 1},
  { "EXP",   "Expand reference component(s)",   1} 
};

char * TagTypes[] = {
  "",
  "Byte", "Ascii", "Short", "Long", "Rational", "??", "Undefined", "???", "Slong", "Srational"
};
// "Rational" and "Srational" need 8 bytes, so the "ValueOffset" is an offset.
// The 8 bytes are a pair of 4-byte longs, numerator and denominator.
// Each long is either unsigned (for "Rational") or signed (for "Srational").

typedef struct {
  uint16_t tag;
  uint16_t tagType;
  uint32_t tagCount;
  uint32_t tagValueOffset; // This is either a value, or an offset.
} IFDfieldT;

/* IFD = Image File Directory.
*/

typedef struct {
  uint32_t IndImgAttribute;
  uint32_t IndImgSize;
  uint32_t IndImgDataOffset;
  uint16_t DepImg1EntNum;
  uint16_t DepImg2EntNum;
} MPentryT;


typedef struct {
  size_t size;
  unsigned char * sApp2;
  unsigned char * pOffsStart;
} App2MpfDataT;

/* Structures for recording MP metadata.
*/
typedef struct {
  size_t   size;        /* Bytes between, and including, SOI and EOI markers. */
  size_t   FileOffset;  /* Offset of SOI from start of file. */
  uint32_t IndImgAttribute;
} MpImgIndexT;

typedef struct {
  size_t NumImages;
  MpImgIndexT * MpImgIndexes;
} MpMetaDataT;


static MpMetaDataT * InitMPMD ()
{
  MpMetaDataT * p = malloc (sizeof (MpMetaDataT));
  if (!p) return NULL;

  p->NumImages = 0;
  p->MpImgIndexes = NULL;

  return p;
}

static MpMetaDataT * SetImgsMPMD (MpMetaDataT * p, size_t numImgs)
{
  p->NumImages = numImgs;

  p->MpImgIndexes = malloc (numImgs * sizeof (MpImgIndexT));
  if (!p->MpImgIndexes) {
    free (p);
    return NULL;
  }

  size_t i;
  for (i = 0; i < numImgs; i++) {
    MpImgIndexT * pi = &p->MpImgIndexes[i];
    pi->size = 0;
    pi->FileOffset = 0;
  }

  return p;
}

static MpMetaDataT * DeInitMPMD (MpMetaDataT * p)
{
  if (!p) {
    fprintf (stderr, "BUG: DeInitMPMD: p not set\n");
    return NULL;
  }
  free (p->MpImgIndexes);
  free (p);
  return NULL;
}

static void DumpMPMD (MpMetaDataT * p)
{
  fprintf (stderr, "MpMetaData:\n");
  if (!p) {
    fprintf (stderr, "BUG: DumpMPMD: p not set\n");
    return;
  }
  fprintf (stderr, "  NumImages: %lu\n", p->NumImages);

  size_t i;
  for (i = 0; i < p->NumImages; i++) {
    MpImgIndexT * pi = &p->MpImgIndexes[i];
    fprintf (stderr, "    Image #%lu:\n", i);
    fprintf (stderr, "      size = %lu\n", pi->size);
    fprintf (stderr, "      FileOffset = %lu\n", pi->FileOffset);
    fprintf (stderr, "      IndImgAttribute = %u\n", pi->IndImgAttribute);
  }
}



signed int Get2Bytes (FILE * fhIn, FILE * fhOut)
// Get 2 bytes, MSB.
// If EOF, return -1.
{
  int c0 = getc (fhIn);
  if (c0 == EOF) return -1;
  if (fhOut) putc (c0, fhOut);
  int c1 = getc (fhIn);
  if (c1 == EOF) return -1;
  if (fhOut) putc (c1, fhOut);

  return c0 * 256 + c1;
}

// FIXME: Security risk: we need to check for exceeding array boundary.

//uint16_t Get2bytes (App2MpfDataT * pData, unsigned char * p, int IsBigEndian)
uint16_t Get2bytes (App2MpfDataT * pData, size_t offs, int IsBigEndian)
{
  unsigned char * p = &pData->pOffsStart[offs];
  fprintf (stderr, "    Get2bytes: %i %i\n", *p, *(p+1));
  if (IsBigEndian) return (uint16_t)( (*p * 256) + *(p+1) );
  else             return (uint16_t)( (*(p+1) * 256) + *p );
}

//uint32_t Get4bytes (App2MpfDataT * pData, unsigned char * p, int IsBigEndian)
uint32_t Get4bytes (App2MpfDataT * pData, size_t offs, int IsBigEndian)
{
  unsigned char * p = &pData->pOffsStart[offs];
  fprintf (stderr, "    Get4bytes: %i %i %i %i\n", *p, *(p+1), *(p+2), *(p+3));
  if (IsBigEndian) return (uint32_t)(( (*p * 256) + *(p+1) ) * 256  + *(p+2) ) * 256 + *(p+3);
  else             return (uint32_t)(( (*(p+3) * 256) + *(p+2) ) * 256  + *(p+1) ) * 256 + *p;
}

//static void RdIFDfield (App2MpfDataT * pData, unsigned char * p, int IsBigEndian, IFDfieldT * pf)
static void RdIFDfield (App2MpfDataT * pData, size_t offs, int IsBigEndian, IFDfieldT * pf)
{
//  pf->tag            = Get2bytes (pData, &p[0], IsBigEndian);
//  pf->tagType        = Get2bytes (pData, &p[2], IsBigEndian);
//  pf->tagCount       = Get4bytes (pData, &p[4], IsBigEndian);
//  pf->tagValueOffset = Get4bytes (pData, &p[8], IsBigEndian);

//  pf->tag            = Get2bytes (pData, &pData->pOffsStart[offs+0], IsBigEndian);
//  pf->tagType        = Get2bytes (pData, &pData->pOffsStart[offs+2], IsBigEndian);
//  pf->tagCount       = Get4bytes (pData, &pData->pOffsStart[offs+4], IsBigEndian);
//  pf->tagValueOffset = Get4bytes (pData, &pData->pOffsStart[offs+8], IsBigEndian);

  pf->tag            = Get2bytes (pData, offs+0, IsBigEndian);
  pf->tagType        = Get2bytes (pData, offs+2, IsBigEndian);
  pf->tagCount       = Get4bytes (pData, offs+4, IsBigEndian);
  pf->tagValueOffset = Get4bytes (pData, offs+8, IsBigEndian);
}

static void DumpIFDfield (App2MpfDataT * pData, int IsBigEndian, IFDfieldT * pf)
{
  fprintf (stderr, "    tag = %04x", pf->tag);
  fprintf (stderr, "    tagType = %04x (%s)", pf->tagType, TagTypes[pf->tagType]);
  fprintf (stderr, "    tagCount = %u", pf->tagCount);
  fprintf (stderr, "    tagValueOffset = %u\n", pf->tagValueOffset);

  switch (pf->tagType) {
    case 1: {
      break;
    }
    case 2: {
      break;
    }
    case 3: {
      break;
    }
    case 4: {
      fprintf (stderr, "    Value = %u \n", pf->tagValueOffset);
      break;
    }
    case 5: {
      uint32_t numer = Get4bytes (pData, pf->tagValueOffset, IsBigEndian);
      uint32_t denom = Get4bytes (pData, pf->tagValueOffset+4, IsBigEndian);
      fprintf (stderr, "    Value = %u / %u\n", numer, denom);
      break;
    }
    case 6: {
      break;
    }
    case 7: {
      break;
    }
    case 8: {
      break;
    }
    case 9: {
      fprintf (stderr, "    Value = %i \n", pf->tagValueOffset);
      break;
    }
    case 10: {
      int32_t numer = (int32_t)Get4bytes (pData, pf->tagValueOffset, IsBigEndian);
      int32_t denom = (int32_t)Get4bytes (pData, pf->tagValueOffset+4, IsBigEndian);
      fprintf (stderr, "    Value = %i / %i\n", numer, denom);
      break;
    }
    default:
      fprintf (stderr, "    ??\n");
  }
}

//static void RdMPentry (App2MpfDataT * pData, unsigned char * p, int IsBigEndian, MPentryT * pmpe)
static void RdMPentry (App2MpfDataT * pData, size_t offs, int IsBigEndian, MPentryT * pmpe)
{
  pmpe->IndImgAttribute  = Get4bytes (pData, offs+0,  IsBigEndian);
  pmpe->IndImgSize       = Get4bytes (pData, offs+4,  IsBigEndian);
  pmpe->IndImgDataOffset = Get4bytes (pData, offs+8,  IsBigEndian);
  pmpe->DepImg1EntNum    = Get2bytes (pData, offs+12, IsBigEndian);
  pmpe->DepImg2EntNum    = Get2bytes (pData, offs+14, IsBigEndian);
}

static void DumpMPentry (MPentryT * pmpe, size_t endianOffs)
{
  uint32_t MpTypeCode = pmpe->IndImgAttribute & 0x00FFFFFF;
  uint32_t ImageDataFormat = (pmpe->IndImgAttribute & 0x07000000) >> 24;

  fprintf (stderr, "  Individual image:\n");
  fprintf (stderr, "    IndImgAttribute = %u\n", pmpe->IndImgAttribute);
  fprintf (stderr, "      Dependent parent: %s\n", ((pmpe->IndImgAttribute & 0x80000000) == 0) ? "No" : "Yes");
  fprintf (stderr, "      Dependent child: %s\n", ((pmpe->IndImgAttribute & 0x40000000) == 0) ? "No" : "Yes");
  fprintf (stderr, "      Representative image: %s\n", ((pmpe->IndImgAttribute & 0x20000000) == 0) ? "No" : "Yes");
  fprintf (stderr, "      Image data format: %i (%s)\n", ImageDataFormat, ImageDataFormat==0 ? "JPEG" : "other");
  fprintf (stderr, "      MP type code: #%06x", MpTypeCode);
  char * desc = "??";
  switch (MpTypeCode) {
    case 0x030000: desc = "Baseline MP Primary Image"; break;
    case 0x010001: desc = "Large thumbnail Class 1 VGA"; break;
    case 0x010002: desc = "Large thumbnail Class 2 Full HD"; break;
    case 0x020001: desc = "Multi-frame Panorama"; break;
    case 0x020002: desc = "Multi-frame Disparity"; break;
    case 0x020003: desc = "Multi-frame Multi-Angle"; break;
    case 0x000000: desc = "Undefined"; break;
    default: desc = "????";
  }
  fprintf (stderr, "  %s\n", desc);
  fprintf (stderr, "    IndImgSize = %u (bytes)\n", pmpe->IndImgSize);
  fprintf (stderr, "    IndImgDataOffset = %u", pmpe->IndImgDataOffset);

  if (pmpe->IndImgDataOffset==0) {
    fprintf (stderr, "  (assumed First Individual Image)\n");
  } else {
    fprintf (stderr, "  plus endianOffs=%lu so file offset=%lu\n", endianOffs, pmpe->IndImgDataOffset+endianOffs);
  }

  fprintf (stderr, "    DepImg1EntNum = %i\n", pmpe->DepImg1EntNum);
  fprintf (stderr, "    DepImg2EntNum = %i\n", pmpe->DepImg2EntNum);
  fprintf (stderr, "\n");
}



static int ProcessJpeg (int argc, char **argv)
{
  FILE * fhIn = fopen (argv[1], "rb");
  if (!fhIn) {
    fprintf (stderr, "Can't open %s\n", argv[1]);
    return 1;
  }

  char * OutPref = NULL;
  int nOutFile = 0;

  #define PATHLEN 100
  char OutFile[PATHLEN] = "";
  FILE * fhOut = NULL;

  if (argc == 3) {
    OutPref = argv[2];
  }

  int c = getc (fhIn);
  if (c == EOF) return 1;
  // if (fhOut) putc (c, fhOut); // FIXME: delay this

  size_t ByteCount = 1;
  int NeedLen = FALSE;
  size_t SOIdepth = 0;
  size_t nApp2Mpf = 0;
  size_t nSOI = 0;
  size_t mpfNumberOfImages = 0;

  MpMetaDataT * pMPMD = InitMPMD ();
  if (!pMPMD) return 1;

  while (c != EOF) {
//    fprintf (stderr, "{%x}", c);
    NeedLen = FALSE;
    if ((char)c != MARKER_PREFIX) {
      if (fhOut) putc (c, fhOut);
    } else {
      size_t MarkerStart = ByteCount - 1;
//      fprintf (stderr, "%lu: Marker %x\n", MarkerStart, c);
      int marker = getc (fhIn);
      if (marker == EOF) return 1;
      ByteCount++;
      char cMarker = (char)marker;
      if (fhOut && cMarker != SOI_MARKER) {
        // FIXME: defer this, in case we have APP2 MPF, that should not be written.
        putc (MARKER_PREFIX, fhOut);
        putc (marker, fhOut);
      }
      //fprintf (stderr, "%lu: Marker2 %x\n", MarkerStart, marker);
      char * pMarkerName = "Unk";
      char * pMarkerDesc = "Unk";
      size_t ReadMore = 0;
      size_t nMpfThisSoi = 0;
      int iAppNum = 0;
      #define APP_NUM_LEN 6
      char sAppNum[APP_NUM_LEN] = "";
      if (cMarker >= '\xC0' && cMarker <= '\xDF') {
        pMarkerName = Markers[(char)marker-'\xC0'].Symbol;
        pMarkerDesc = Markers[(char)marker-'\xC0'].Description;
        NeedLen = Markers[(char)marker-'\xC0'].NeedLen;

        if (cMarker == SOI_MARKER) {
          // "SOI"
          //  FIXME: for non-first image, we have already written SOI.
          SOIdepth++;
          nSOI++;
          nMpfThisSoi = 0;

          if (OutPref) {
            if (snprintf (OutFile, PATHLEN, "%s_%i.jpg", OutPref, nOutFile++) >= PATHLEN) {
              fprintf (stderr, "OutPref is too long\n");
              if (fhIn) fclose (fhIn);
              return 1;
            }
            if (fhOut) fclose (fhOut);
            fprintf (stderr, "Output file: %s\n", OutFile);
            fhOut = fopen (OutFile, "wb");
            if (!fhOut) {
              fprintf (stderr, "Can't open %s\n", OutFile);
              if (fhIn) fclose (fhIn);
              return 1;
            }
            putc (MARKER_PREFIX, fhOut);
            putc (marker, fhOut);
          }

          if (nSOI > 1 && nSOI > pMPMD->NumImages) {
            fprintf (stderr, "ERROR: nSOI=%lu but pMPMD->NumImages=%lu\n", nSOI, pMPMD->NumImages);
          }

        } else if (cMarker=='\xD9') {
          // "EOI"
          if (SOIdepth == 0) {
            fprintf (stderr, "Unexpected EOI\n");
            return -1;
          }
          SOIdepth--;
        }
//        fprintf (stderr, "marker is %s %s\n", pMarkerName, pMarkerDesc);
      } else  if (cMarker >='\xE0' && cMarker <='\xEF') {
        pMarkerName = "APP";
        pMarkerDesc = "Reserved for application segments";
        NeedLen = TRUE;
        iAppNum = marker & 0x0F;
        if (snprintf (sAppNum, APP_NUM_LEN, "%i", iAppNum) >= APP_NUM_LEN) return 1;
//        fprintf (stderr, "sAppNum=%s\n", sAppNum);
      } else  if (cMarker >='\xF0' && cMarker <='\xFD') {
        pMarkerName = "JPG";
        pMarkerDesc = "Reserved for JPEG extensions";
        NeedLen = TRUE;
        if (snprintf (sAppNum, APP_NUM_LEN, "%i", (char)(cMarker & 0x0F)) >= APP_NUM_LEN) return 1;
        fprintf (stderr, "sAppNum=%s\n", sAppNum);
      } else if (cMarker == '\xFE') {
        pMarkerName = "COM";
        pMarkerDesc = "Comment";
        NeedLen = TRUE;
      } else if (cMarker != '\x0') {
        fprintf (stderr, "Null marker %x\n", marker);
      } else if ((char)marker != '\x01') {
        pMarkerName = "TEM"; // no length
      } else  if (cMarker >='\x02' && cMarker <='\xBF') {
        pMarkerName = "RES";
        pMarkerDesc = "Reserved";
        NeedLen = TRUE;
        if (snprintf (sAppNum, APP_NUM_LEN, "%i", (char)(cMarker & 0x0F)) >= APP_NUM_LEN) return 1;
        fprintf (stderr, "sAppNum=%s\n", sAppNum);
      }

      if (NeedLen) {
        signed int gm = Get2Bytes (fhIn, fhOut) - 2;
        if (gm < 0) {
          fprintf (stderr, "Unexpected EOF\n");
          break;
        }
        ReadMore = (size_t)gm;
        ByteCount += 2;
      }

      if (cMarker != '\x0') {
        if (strncmp (pMarkerName, "RST", 3) != 0) {
          fprintf (stderr, "%li: {%x} %s%s %s ReadMore=%lu\n", MarkerStart, marker, pMarkerName, sAppNum, pMarkerDesc, ReadMore);
        }
      }
      if (ReadMore > 0) {
        ByteCount += ReadMore;
        if (cMarker >='\xE0' && cMarker <='\xEF' && iAppNum == 2 && ReadMore > 4) {
          fprintf (stderr, "  This is App2\n");
          App2MpfDataT App2MpfData;
          App2MpfData.size = ReadMore;
          App2MpfData.sApp2 = malloc ((unsigned long)ReadMore * (unsigned long)sizeof (unsigned char));
          App2MpfData.pOffsStart = &App2MpfData.sApp2[4];

          if (!App2MpfData.sApp2) {
            fprintf (stderr, "malloc failed\n");
            return 1;
          }
          size_t nRead = fread (App2MpfData.sApp2, ReadMore, 1, fhIn);
          if (nRead == 0) {
            fprintf (stderr, "fread failed\n");
            return 1;
          }
          int IsApp2Mpf = App2MpfData.sApp2[0]=='M' && App2MpfData.sApp2[1]=='P' && App2MpfData.sApp2[2]=='F' && App2MpfData.sApp2[3]=='\0';
          if (fhOut) fwrite (App2MpfData.sApp2, ReadMore, 1, fhOut);  // FIXME: only if not MPF
          fprintf (stderr, "  First 3 chars: %c%c%c\n", App2MpfData.sApp2[0], App2MpfData.sApp2[1], App2MpfData.sApp2[2]);
          if (IsApp2Mpf) {
            fprintf (stderr, "  This is App2 MPF, for image #%lu\n", nSOI);
            nApp2Mpf++;
            if (nMpfThisSoi > 0) {
              fprintf (stderr, "  Error: this SOI already has %lu APP2 MPF segments\n", nMpfThisSoi);
            }
            nMpfThisSoi++;
            // The MP header is MPendian (4 bytes) plus OffsToFirstIFD (4 bytes)
            // Offsets are relative to start of MP header.
            // This is the same as a TIFF "Image File Header" at the start of TIFF files.
            size_t endianOffs = ByteCount - ReadMore + 4;

            // An image might have more than one APP2 MPF.
            // This would be an error.
            // FIXME: We should report this, and ignore any after the first.

            // MPEndian: "II" or "MM", followed by decimal 42 in 16-bit value at appropriate endian.
            int IsBigEndian = 0;
            if (App2MpfData.sApp2[4]=='\x49' && App2MpfData.sApp2[5]=='\x49' && App2MpfData.sApp2[6]=='\x2a' && App2MpfData.sApp2[7]=='\0') {
              fprintf (stderr, "  %c%c Little endian\n", App2MpfData.sApp2[4], App2MpfData.sApp2[5]);
              IsBigEndian = 0;
            } else if (App2MpfData.sApp2[4]=='\x4D' && App2MpfData.sApp2[5]=='\x4D' && App2MpfData.sApp2[6]=='\x0' && App2MpfData.sApp2[7]=='\x2A') {
              fprintf (stderr, "  %c%c Big endian\n", App2MpfData.sApp2[4], App2MpfData.sApp2[5]);
              IsBigEndian = 1;
            } else {
              fprintf (stderr, "  Unknown endian\n");
              return 1;
            }
            unsigned long OffsNextIFD = Get4bytes (&App2MpfData, 4, IsBigEndian);
            fprintf (stderr, "    OffsNextIFD = %lu\n", OffsNextIFD);

            if (nSOI == 1) {
              fprintf (stderr, "    First image, so we read index fields\n");

              // MP Index IFD is: FldCount, then each index field, then offset to next IFD (which is "MP Attributes IFD").

              // The array includes 4 bytes MP Format Identifier, but stored offsets don't include this.
              #define  OffsStart 4
              unsigned long offs = OffsNextIFD + OffsStart;
//              unsigned int FldCount = Get2bytes (&App2MpfData, &App2MpfData.sApp2[offs], IsBigEndian);
              unsigned int FldCount = Get2bytes (&App2MpfData, OffsNextIFD, IsBigEndian);
              fprintf (stderr, "    offs = %lu  FldCount (number of fields) = %u\n", offs, FldCount);

              unsigned int nFld;
              for (nFld=0; nFld < FldCount; nFld++) {
                IFDfieldT ifdFld;
                fprintf (stderr, "  IFD field %u:\n", nFld);
//                RdIFDfield (&App2MpfData, &App2MpfData.sApp2[offs + 2 + 12*nFld], IsBigEndian, &ifdFld);
                RdIFDfield (&App2MpfData, OffsNextIFD + 2 + 12*nFld, IsBigEndian, &ifdFld);
                DumpIFDfield (&App2MpfData, IsBigEndian, &ifdFld);
                switch (ifdFld.tag) {
                  case 0xB000:
                    fprintf (stderr, "      tag is B000: version\n");
                    break;
                  case 0xB001:
                    fprintf (stderr, "      tag is B001: number of images");
                    mpfNumberOfImages = ifdFld.tagValueOffset;
                    fprintf (stderr, "      mpfNumberOfImages = %lu\n", mpfNumberOfImages);
                    if (!SetImgsMPMD (pMPMD, ifdFld.tagValueOffset)) return 1;
                    break;
                  case 0xB002: {
                    fprintf (stderr, "      tag is B002: MP entry\n");

                    if (ifdFld.tagCount != 16 * mpfNumberOfImages) {
                      fprintf (stderr, "ifdFld.tagCount = %i but mpfNumberOfImages = %lu\n", ifdFld.tagCount, mpfNumberOfImages);
                    }

                    size_t mpeOffs = ifdFld.tagValueOffset;
                    size_t iImg;
                    for (iImg = 0; iImg < mpfNumberOfImages; iImg++) {
                      fprintf (stderr, "  Image #%lu\n", iImg);
                      MPentryT mpe;
                      RdMPentry (&App2MpfData, mpeOffs + iImg*16, IsBigEndian, &mpe);
                      DumpMPentry (&mpe, endianOffs);
                      if (pMPMD && pMPMD->NumImages > iImg) {
                        MpImgIndexT * pi = &pMPMD->MpImgIndexes[iImg];
                        pi->size = mpe.IndImgSize;
                        pi->IndImgAttribute = mpe.IndImgAttribute;

                        if (iImg == 0) {
                          if (mpe.IndImgDataOffset != 0) {
                            fprintf (stderr, "ERROR: First image, but mpe.IndImgDataOffset = %u\n", mpe.IndImgDataOffset);
                          }
                          pi->FileOffset = 0;
                        } else {
                          pi->FileOffset = mpe.IndImgDataOffset + endianOffs;
                        }
                      } else {
                        fprintf (stderr, "BUG: MPentry\n");
                      }
                    }
                    break;
                  }
                  case 0xB003:
                    fprintf (stderr, "      tag is B003: Image UID list\n");
                    // Count is 33 * number of images
                    break;
                  case 0xB004:
                    fprintf (stderr, "      tag is B004: Total frames = %u\n", ifdFld.tagValueOffset);
                    break;
                  default: fprintf (stderr, "Unknown tag %04x\n", ifdFld.tag);
                }
              } // end for fields

//              OffsNextIFD = Get4bytes (&App2MpfData, &App2MpfData.sApp2[offs + 2 + 12*FldCount], IsBigEndian);
              OffsNextIFD = Get4bytes (&App2MpfData, OffsNextIFD + 2 + 12*FldCount, IsBigEndian);
              fprintf (stderr, "      offs to next IFD = %lu\n", OffsNextIFD);

            } // end First image

            if (OffsNextIFD > 0) {
              // MPF Attribute IFD
              // MP Attributes IFD is: Count, then each attribute field, then offset to next IFD.
              // See spec 5.2.4. MP Attribute IFD
              unsigned long offs = OffsNextIFD + OffsStart;
              unsigned int AttrCount = Get2bytes (&App2MpfData, OffsNextIFD, IsBigEndian);
              fprintf (stderr, "    offs = %lu  AttrCount (number of individual attribute tags) = %u\n", offs, AttrCount);

// FIXME: but maybe this APP2 MPF isn't for the first image.

              unsigned int nFld;
              for (nFld=0; nFld < AttrCount; nFld++) {
                IFDfieldT ifdFld;
                fprintf (stderr, "  IFD field %u:\n", nFld);
//                RdIFDfield (&App2MpfData, &App2MpfData.sApp2[offs + 2 + 12*nFld], IsBigEndian, &ifdFld);
                RdIFDfield (&App2MpfData, OffsNextIFD + 2 + 12*nFld, IsBigEndian, &ifdFld);
                DumpIFDfield (&App2MpfData, IsBigEndian, &ifdFld);
              }
            }
          }
          free (App2MpfData.sApp2);
          App2MpfData.sApp2 = NULL;

          ReadMore = 0;
        }
        fprintf (stderr, "  ");
        int DumpThem = 1;
        size_t i;
        for (i=0; i < ReadMore; i++) {
          int d = getc (fhIn);
          if (d == EOF) return -1;
          if (fhOut) putc (d, fhOut);
          // If we are in APPn segment, it may begin with NUL-terminated string.
          // If we are splitting the files, APP2 "MPF" (Multi-Picture Format) should not be copied to output.
#define MAX_DUMP 100
          if (i < MAX_DUMP && DumpThem) {
            if (d >= 32 && d < 128) {
              fprintf (stderr, "%c", d);
            }
            else {
              fprintf (stderr, "{%i}", d);
              DumpThem = 0;
            }
          }
        }
        fprintf (stderr, "\n");
      }
    }

    c = getc (fhIn);
    if (c == EOF) break;
    ByteCount++;
  }

  if (c==EOF) {
    fprintf (stderr, "EOF\n");
  }

  fprintf (stderr, "Summary:\n");
  fprintf (stderr, "  Final ByteCount = %lu\n", ByteCount);
  fprintf (stderr, "  nApp2Mpf = %lu\n", nApp2Mpf);
  if (nApp2Mpf > nSOI) fprintf (stderr, "  (warning: more than nSOI)\n");
  if (nApp2Mpf > 0) fprintf (stderr, "  mpfNumberOfImages = %lu\n", mpfNumberOfImages);
  fprintf (stderr, "  nSOI = %lu  ", nSOI);
  if (nApp2Mpf > 0) {
    if (nSOI == mpfNumberOfImages) fprintf (stderr, "(matches mpfNumberOfImages, good)\n");
    else fprintf (stderr, "(warning: does not match mpfNumberOfImages)\n");
  }
  fprintf (stderr, "  SOIdepth = %lu (%s)\n", SOIdepth, (SOIdepth==0) ? "good" : "warning");
  fprintf (stderr, "  nOutFile = %i\n", nOutFile);

  if (fhOut) fclose (fhOut);
  fclose (fhIn);

  DumpMPMD (pMPMD);

  pMPMD = DeInitMPMD (pMPMD);

  return 0;
}

int main (int argc, char **argv)
{
  if (argc < 2) {
    fprintf (stderr, "Syntax: %s infile [outprefix]\n", argv[0]);
    fprintf (stderr, "outprefix should have no extension\n");
    fprintf (stderr, "For example, if outprefix is \"x\" files named \"x_0.jpg\", \"x_1.jpg\" etc will be created.\n");
    return 1;
  }

  return ProcessJpeg (argc, argv);
}
