#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "vsn_defines.h"

/* Updated:
     3-April-2018 for v7.0.7-28
*/

/* When an output pixel is outside the circle,
   the value will come from the "-virtual-pixel" setting.
*/

typedef enum {
  fullframe,
  circular
} FormatT;

typedef struct {
  double
    ifov,
    ofov,
    rad;
  FormatT
    format;
  MagickBooleanType
    do_verbose;
} FishSpecT;


static void usage (void)
{
  printf ("Usage: -process 'rect2eqfish [OPTION]...'\n");
  printf ("From an rectilinear image, makes equidistant (linear) fisheye image.\n");
  printf ("\n");
  printf ("  i, ifov N           input field of view, default 120\n");
  printf ("  o, ofov N           output field of view, default 180\n");
  printf ("  f, format string    fullframe [default] or circular\n");
  printf ("  r, radius N         output radius (overrides format)\n");
  printf ("  v, verbose          write text information to stdout\n");
  printf ("\n");
  printf ("Fields of view are angles in degrees from a corner to opposite corner.\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  FishSpecT * pfs
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  pfs->ifov = 120;
  pfs->ofov = 180;
  pfs->do_verbose = MagickFalse;
  pfs->format = fullframe;
  pfs->rad = 0;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "i", "ifov")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i]))
      {
        fprintf (stderr, "Non-numeric argument to 'ifov'.\n");
        status = MagickFalse;
      }
      else pfs->ifov = atof(argv[i]);
      if (pfs->ifov <= 0 || pfs->ifov >= 180)
      {
        fprintf (stderr, "'ifov' must be > 0, < 180.\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "o", "ofov")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i]))
      {
        fprintf (stderr, "Non-numeric argument to 'ofov'.\n");
        status = MagickFalse;
      }
      else pfs->ofov = atof(argv[i]);
      if (pfs->ofov <= 0 || pfs->ofov > 180)
      {
        fprintf (stderr, "'ofov' must be > 0, <= 180.\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "r", "radius")==MagickTrue) {
      i++;
      if (!isdigit ((int)*argv[i]))
      {
        fprintf (stderr, "Non-numeric argument to 'radius'.\n");
        status = MagickFalse;
      }
      else pfs->rad = atof(argv[i]);
      if (pfs->rad <= 0)
      {
        fprintf (stderr, "'radius' must be > 0.\n");
        status = MagickFalse;
      }
    } else if (IsArg (pa, "f", "format")==MagickTrue) {
      i++;
      char * pa = (char *)argv[i];
      if      (LocaleCompare(pa, "circular" )==0) pfs->format = circular;
      else if (LocaleCompare(pa, "fullframe")==0) pfs->format = fullframe;
      else {
        fprintf (stderr, "rect2eqfish: ERROR: unknown 'format' [%s]\n", pa);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pfs->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "rect2eqfish: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (pfs->do_verbose) {
    fprintf (stderr, "rect2eqfish options:");

    if (pfs->do_verbose) fprintf (stderr, "  verbose");

    // FIXME: thoughout, %g should respect "-precision"
    fprintf (stderr, "  ifov %g", pfs->ifov);

    fprintf (stderr, "  ofov %g", pfs->ofov);

    fprintf (stderr, "  format ");
    if (pfs->format == circular)
      fprintf (stderr, "circular");
    else if (pfs->format == fullframe)
      fprintf (stderr, "fullframe");
    else
      fprintf (stderr, "??");


    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *rect2eqfish(const Image *image,
  FishSpecT * pfs,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  CacheView
    *image_view,
    *out_view;

  ssize_t
    y;

  MagickBooleanType
    status;

  double
    CentX,
    CentY,
    dim,
    //dim2,
    ifoc,
    ofocinv;


  assert(image != (Image *) NULL);
  assert(image->signature == MAGICK_CORE_SIG);

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  // Make a clone of this image, same size but undefined pixel values:
  //
  new_image=CloneImage(image, image->columns, image->rows, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  image_view = AcquireVirtualCacheView(image,exception);

  out_view = AcquireAuthenticCacheView(new_image,exception);

  status = MagickTrue;

  dim = 0.0;

  if (pfs->format == circular) {
    dim = image->columns;
    if (dim > image->rows) dim = image->rows;
  } else if (pfs->format == fullframe) {
    dim = hypot (image->columns, image->rows);
  }

  if (pfs->rad > 0) dim = 2 * pfs->rad;

  if (dim == 0.0) {
    fprintf (stderr, "BUG");
    return (Image *) NULL;
  }

  //dim2 = dim / 2.0;  // But not used??

  CentX = (image->columns - 1) / 2.0;
  CentY = (image->rows    - 1) / 2.0;

  ifoc = dim / (2 * tan(pfs->ifov*M_PI/360));

  ofocinv = (pfs->ofov*M_PI)/(dim * 180);

  if (pfs->do_verbose) {
    fprintf (stderr, "CentX=%g", CentX);
    fprintf (stderr, "  CentY=%g", CentY);
    fprintf (stderr, "  dim=%g", dim);
    fprintf (stderr, "  ifoc=%g", ifoc);
    fprintf (stderr, "  ofocinv=%g  (1/%g)\n", ofocinv, 1.0 / ofocinv);

    fprintf (stderr, "\nFX equivalent:\n");
    fprintf (stderr, "  -fx \"");
    fprintf (stderr, "xd=i-%g;", CentX);
    fprintf (stderr, "yd=j-%g;", CentY);
    fprintf (stderr, "rd=hypot(xd,yd);");
    fprintf (stderr, "phiang=%g*rd;", ofocinv);
      // (phiang would have a different formula for other fisheye projections.)
    fprintf (stderr, "rr=%g*tan(phiang);", ifoc);
    fprintf (stderr, "xs=(rd?rr/rd:0)*xd+%g;", CentX);
    fprintf (stderr, "ys=(rd?rr/rd:0)*yd+%g;", CentY);
    //fprintf (stderr, "(rd>$dim2)?$bgc:u.p{xs,ys}";
    fprintf (stderr, "u.p{xs,ys}\"\n");
  }


#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) shared(status) \
    MAGICK_THREADS(new_image,new_image,new_image->rows,1)
#endif
  for (y=0; y < (ssize_t) new_image->rows; y++)
  {
    register VIEW_PIX_PTR
      *q;

    register ssize_t
      x;

    if (status == MagickFalse)
      continue;

    q=GetCacheViewAuthenticPixels(out_view,0,y,new_image->columns,1,exception);
    if (q == (const VIEW_PIX_PTR *) NULL)
    {
      status=MagickFalse;
      continue;
    }

    for (x=0; x < (ssize_t) image->columns; x++)
    {
      double
        xd,
        yd,
        phi,
        rd,
        rr,
        mult,
        xs,
        ys;

      PIX_INFO
        src_pixel;

      xd = x - CentX;
      yd = y - CentY;

      rd = hypot (xd, yd);

      phi = ofocinv * rd;
      // (phi would have a different formula for other fisheye projections.)
      rr = ifoc * tan(phi);

      if (rd == 0) {
        xs = CentX;
        ys = CentY;
      } else {
        mult = rr / rd;
        xs = CentX + mult * xd;
        ys = CentY + mult * yd;
      }

#if IMV6OR7==6
      InterpolateMagickPixelPacket (
        image, image_view,
        image->interpolate,
        xs, ys, &src_pixel, exception);
#else
     InterpolatePixelInfo (
        image, image_view,
        image->interpolate,
        xs, ys, &src_pixel, exception);
#endif

      SET_PIXEL_RED   (new_image, src_pixel.red, q);
      SET_PIXEL_GREEN (new_image, src_pixel.green, q);
      SET_PIXEL_BLUE  (new_image, src_pixel.blue, q);
#if IMV6OR7==6
      SET_PIXEL_ALPHA (new_image, QuantumRange - src_pixel.opacity, q);
#else
      SET_PIXEL_ALPHA (new_image, src_pixel.alpha, q);
#endif

      q += Inc_ViewPixPtr (new_image);
    }
    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse)
      status=MagickFalse;
  }
  out_view=DestroyCacheView(out_view);
  image_view=DestroyCacheView(image_view);

  return (new_image);
}



/*
We replace each image as we come across it.
An alternative technique is to create a new list, and add new images to it;
at the end, replace the original list with the new one.

That alternative is slightly more complex
and needs more memory as it would hold all input and output images in memory.
*/

ModuleExport size_t rect2eqfishImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  FishSpecT
    fs;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &fs);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = rect2eqfish(image, &fs, exception);
    if (new_image == (Image *) NULL)
      return(-1);

    ReplaceImageInList(&image,new_image);
    // Replace messes up the images pointer. Make it good:
    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
