#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"


typedef struct {
  char
    sSize[MaxTextExtent],
    sColor[MaxTextExtent];

  size_t
    width,
    height;

  MagickBooleanType
    do_verbose;
} replspecT;

static void usage (void)
{
  printf ("Usage: -process 'replacespec [OPTION]...'\n");
  printf ("Replace each image.\n");
  printf ("\n");
  printf ("  s, size WxH         size of new image\n");
  printf ("  c, color string     make new image this color\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType menu(
  const int argc,
  const char **argv,
  replspecT *prs
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  *prs->sSize = '\0';
  *prs->sColor = '\0';
  prs->width = prs->height = 0;
  prs->do_verbose = MagickFalse;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "s", "size")==MagickTrue) {
      i++;
      strncpy (prs->sSize, argv[i], MaxTextExtent-1);
      if (sscanf (prs->sSize, "%lux%lu", &prs->width, &prs->height) == 0) {
        if (sscanf (prs->sSize, "x%lu", &prs->height) != 1) {
          fprintf (stderr, "replacespec: invalid 'size' argument [%s]\n",
            prs->sSize);
          status = MagickFalse;
        }
      }
    } else if (IsArg (pa, "c", "color")==MagickTrue) {
      i++;
      strncpy (prs->sColor, argv[i], MaxTextExtent-1);
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      prs->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "replacespec: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (prs->do_verbose) {
    fprintf (stderr, "replacespec options:");
    if (*prs->sSize) fprintf (stderr, "  size %s ",
      prs->sSize);
    if (*prs->sColor) fprintf (stderr, "  color %s ",
      prs->sColor);
    if (prs->do_verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}



// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *CopyColor(const Image *image,
  replspecT *prs,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  PIX_INFO
    mpp;

  size_t
    width,
    height;

  assert(image != (Image *) NULL);
  assert(image->signature == MAGICK_CORE_SIG);
  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);


  if (prs->do_verbose) {
    fprintf (stderr, "  input image size %lux%lu\n", image->columns, image->rows);
    if (*prs->sSize) {
      fprintf (stderr, "  requested size %lux%lu\n", prs->width, prs->height);
    }
  }

  //width = image->columns;
  //height = image->rows;
  //if (*prs->sSize) {
  //  width = prs->width;
  //  height = prs->height;
  //}

  width  = (prs->width)  ? prs->width  : image->columns;
  height = (prs->height) ? prs->height : image->rows;


  new_image=CloneImage(image, width, height, MagickTrue, exception);
  if (new_image == (Image *) NULL)
    return(new_image);

  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

#if IMV6OR7==6
  // Make a "colour", which will initially be black.
  GetMagickPixelPacket (new_image, &mpp);

  // We set it to any colour:
  (void)QueryMagickColor(
    *prs->sColor ? prs->sColor : "brown",
    &mpp,
    exception);

  // Set all the pixels to this colour.
  (void)SetImageColor(new_image, &mpp);
#else
  GetPixelInfo (new_image, &mpp);

  QueryColorCompliance (
    *prs->sColor ? prs->sColor : "brown",
    AllCompliance, &mpp, exception);

  (void)SetImageColor(new_image, &mpp, exception);
#endif


  return (new_image);
}


/*
We replace each image as we come across it.
An alternative technique is to create a new list, and add new images to it;
at the end, replace the original list with the new one.

That alternative is slightly more complex
and needs more memory as it would hold all input and output images in memory.
*/

ModuleExport size_t replacespecImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  replspecT
    replace_spec;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &replace_spec);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = CopyColor(image, &replace_spec, exception);
    if (new_image == (Image *) NULL)
      return(-1);

    ReplaceImageInList(&image,new_image);
    // Replace messes up the images pointer. Make it good:
    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
