/* Updated:
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "vsn_defines.h"


typedef struct {
  double
    scale,
    bias,
    offsetX,
    offsetY;
  MagickBooleanType
    inverse,
    do_verbose;
} rhothetaT;


static void usage (void)
{
  printf ("Usage: -process 'rhotheta [OPTION]...'\n");
  printf ("From x and y channels, calculates the arctangent and polar distance.\n");
  printf ("\n");
  printf ("  s,   scale N          scale [1/2pi]\n");
  printf ("  b,   bias N           bias [0.5]\n");
  printf ("  o,   offset N,N       offset [0,0]\n");
  printf ("  inv, inverse          inverse (rho,theta to x,y)\n");
  printf ("  v,   verbose          write text information to stdout\n");
  printf ("[And xy of origin?]\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


static MagickBooleanType ParseCoord (
  const char * s, double * px, double * py)
{
  int n;
  char * p = (char *)s;

  sscanf (p, "%lg%n", px, &n);
  if (!n) return MagickFalse;
  p += n;
  if (*p != ',') return MagickFalse;
  p++;

  sscanf (p, "%lg%n", py, &n);
  if (!n) return MagickFalse;
  p += n;
  if (*p != '\0') return MagickFalse;

  return MagickTrue;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "baryc: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  rhothetaT * prt
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  prt->do_verbose = MagickFalse;

  prt->scale = 1 / (2.0 * M_PI);
  prt->bias = 0.5;
  prt->offsetX = 0;
  prt->offsetY = 0;
  prt->inverse = MagickFalse;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "s", "scale")==MagickTrue) {
      NEXTARG;
      prt->scale = atof(argv[i]);
    } else if (IsArg (pa, "b", "bias")==MagickTrue) {
      NEXTARG;
      prt->bias = atof(argv[i]);
    } else if (IsArg (pa, "o", "offset")==MagickTrue) {
      NEXTARG;
      if (!ParseCoord (argv[i], &prt->offsetX, &prt->offsetY)) {
        status = MagickFalse;
      }
    } else if (IsArg (pa, "inv", "inverse")==MagickTrue) {
      prt->inverse = MagickTrue;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      prt->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "rhotheta: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (prt->do_verbose) {
    fprintf (stderr, "rhotheta options:");

    // FIXME: thoughout, %g should respect "-precision"
    fprintf (stderr, "  scale %g", prt->scale);
    fprintf (stderr, "  bias %g", prt->bias);

    if (prt->offsetX != 0 || prt->offsetY != 0) {
      fprintf (stderr, "  offset %g,%g", prt->offsetX, prt->offsetY);
    }
    if (prt->inverse) fprintf (stderr, "  inverse");

    if (prt->do_verbose) fprintf (stderr, "  verbose");

    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


// The next function takes an image, and returns it.
//
static Image *rhotheta(
  Image *image,
  rhothetaT * pat,
  ExceptionInfo *exception)
{
  // This function reads and writes the R and G channels.
  // It leaves the B and alpha channels unchanged.

  ssize_t
    y;

  MagickBooleanType
    status;


  assert(image != (Image *) NULL);
  assert(image->signature == MAGICK_CORE_SIG);

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (SetNoPalette (image, exception) == MagickFalse)
    return (Image *)NULL;

  CacheView * image_view = AcquireAuthenticCacheView (image,exception);

  status = MagickTrue;

#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) shared(status) \
    MAGICK_THREADS(image,image,image->rows,1)
#endif
  for (y=0; y < (ssize_t) image->rows; y++)
  {
    register VIEW_PIX_PTR
      *p;

    ssize_t
      x;

    if (status == MagickFalse)
      continue;

    p=GetCacheViewAuthenticPixels(image_view,0,y,image->columns,1,exception);
    if (!p)
    {
      status=MagickFalse;
      continue;
    }

    double vx, vy, rho, theta;
    for (x=0; x < (ssize_t) image->columns; x++)
    {
      if (pat->inverse) {
        rho = GET_PIXEL_RED   (image, p) / (MagickRealType)QuantumRange;
        theta = (GET_PIXEL_GREEN (image, p) / (MagickRealType)QuantumRange
                 - pat->bias)
                / pat->scale;

        vx = rho * sin (theta) + pat->offsetX;
        vy = rho * cos (theta) + pat->offsetY;

        SET_PIXEL_RED   (image,
                         QuantumRange * vx,
                         p);

        SET_PIXEL_GREEN (image,
                         QuantumRange * vy,
                         p);
      } else {
        vx = GET_PIXEL_RED   (image, p) / (MagickRealType)QuantumRange - pat->offsetX;
        vy = GET_PIXEL_GREEN (image, p) / (MagickRealType)QuantumRange - pat->offsetY;

        rho = hypot (vx, vy);

        theta = atan2 (vx, vy) * pat->scale + pat->bias;

        SET_PIXEL_RED   (image,
                         QuantumRange * rho,
                         p);

        SET_PIXEL_GREEN (image,
                         QuantumRange * theta,
                         p);
      }
      p += Inc_ViewPixPtr (image);
    }
    if (SyncCacheViewAuthenticPixels(image_view,exception) == MagickFalse)
      status=MagickFalse;
  }
  image_view = DestroyCacheView (image_view);

  if (!status) return NULL;

  return (image);
}



ModuleExport size_t rhothetaImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  rhothetaT
    rt;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  MagickBooleanType status = menu (argc, argv, &rt);
  if (status == MagickFalse)
    return (-1);

  Image * image;
  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    Image * newImg = rhotheta (image, &rt, exception);
    if (!newImg) return (-1);

    if (newImg != image) {
      ReplaceImageInList (&image, newImg);
      *images=GetFirstImageInList (newImg);
    }
  }


  return (MagickImageFilterSignature);
}
