/* Updated:
     3-April-2018 for v7.0.7-28
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"

#include "rmsealpha.inc"


/* Takes two images.
   Does a subimage search of the second image within the first:
     finds the position of smallest score,
     and returns the score at that location.

   The weighting is such that a fully-transparent pixel effectively matches
   any other pixel exactly, whether that pixel is opaque or not,
   and contributes nothing to the scrore.

   At a given position:
   imageScore = sqrt (sigma(pixelScore) / sigma(pixelmAlpha) / 3)

   pixelScore = (dRed^2 + dGreen^2 + dBlue^2 ) * pixelmAlpha

   dRed = GET_PIXEL_RED(image,a) - GET_PIXEL_RED(image,b)
     etc for green and blue.

   pixelmAlpha = GET_PIXEL_ALPHA(image,a) * GET_PIXEL_ALPHA(image,b)
    (Alternative pixelmAlpha = (min (GET_PIXEL_ALPHA(image,a), GET_PIXEL_ALPHA(image,b)))^2

  Updated:
    12-August-2017 for v7.
    22-August-2017 added multi-scale and adjustLC; moved code to rmsealpha.inc
*/

// FIXME? Also option to process all images after first as searchimages?


static void usage (void)
{
  printf ("Usage: -process 'rmsealpha [OPTION]...'\n");
  printf ("Searches for lowest alpha-weighted RMSE score.\n");
  printf ("\n");
  printf ("  ai,  avoid_identical        patches with identical RGBA score 1.0\n");
  printf ("  ddo, dont_decrease_opacity  patches with decreased opacity in any pixel score 1.0\n");
  printf ("  ms,  multi_scale            multi-scale search\n");
  printf ("  adj, adjustLC number        adjust lightness and contrast\n");
  printf ("  j,   just_score             write the score only, with no \\n\n");
  printf ("  cnv, canvasOffset           add canvas offset to result\n");
  printf ("  so,  stdout                 write data to stdout\n");
  printf ("  se,  stderr                 write data to stderr (default)\n");
  printf ("  sis, saveInpScales          save input scales\n");
  printf ("  sss, saveSubScales          save subimage scales\n");
  printf ("  z,   sizeDiv number         size divider for ms\n");
  printf ("  md,  minDim integer         minimum dimension for ms\n");
  printf ("  v,   verbose                write text information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "paintpatches: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  RmseAlphaT * pra
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "ai", "avoid_identical")==MagickTrue) {
      pra->avoidIdentical = MagickTrue;
    } else if (IsArg (pa, "ddo", "dont_decrease_opacity")==MagickTrue) {
      pra->dontDecreaseOpacity = MagickTrue;
    } else if (IsArg (pa, "j", "just_score")==MagickTrue) {
      pra->outJustScore = MagickTrue;
    } else if (IsArg (pa, "cnv", "canvasOffset")==MagickTrue) {
      pra->addCanvOffs = MagickTrue;
    } else if (IsArg (pa, "ms", "multi_scale")==MagickTrue) {
      pra->multiScale = MagickTrue;
    } else if (IsArg (pa, "adj", "adjustLC")==MagickTrue) {
      NEXTARG;
      pra->adjustMeanSd = atof(argv[i]);
    } else if (IsArg (pa, "z", "sizeDiv")==MagickTrue) {
      NEXTARG;
      pra->sizeDivider = atof(argv[i]);
    } else if (IsArg (pa, "md", "minDim")==MagickTrue) {
      NEXTARG;
      pra->minDim = atoi(argv[i]);
    } else if (IsArg (pa, "so", "stdout")==MagickTrue) {
      pra->outhandle = stdout;
    } else if (IsArg (pa, "se", "stderr")==MagickTrue) {
      pra->outhandle = stderr;
    } else if (IsArg (pa, "sis", "saveInpScales")==MagickTrue) {
      pra->saveInpScales = MagickTrue;
    } else if (IsArg (pa, "sss", "saveSubScales")==MagickTrue) {
      pra->saveSubScales = MagickTrue;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pra->do_verbose = MagickTrue;
    } else {
      fprintf (stderr, "rmsealpha: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (pra->do_verbose) {
    fprintf (stderr, "rmsealpha options:");

    if (pra->avoidIdentical) fprintf (stderr, "  avoid_identical");
    if (pra->dontDecreaseOpacity) fprintf (stderr, "  dont_decrease_opacity");
    if (pra->outJustScore) fprintf (stderr, "  just_score");
    if (pra->addCanvOffs) fprintf (stderr, "  canvasOffset");
    if (pra->multiScale) fprintf (stderr, "  multi_scale");
    if (pra->adjustMeanSd > 0) fprintf (stderr, "  adjustLC %g", pra->adjustMeanSd);
    fprintf (stderr, "  sizeDiv %g", pra->sizeDivider);
    fprintf (stderr, "  minDim %i", pra->minDim);
    if (pra->outhandle == stdout) fprintf (stderr, "  stdout");
    if (pra->outhandle == stderr) fprintf (stderr, "  stderr");
    if (pra->saveInpScales) fprintf (stderr, "  saveInpScales");
    if (pra->saveSubScales) fprintf (stderr, "  saveSubScales");

    if (pra->do_verbose) fprintf (stderr, "  verbose");

    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}



// The next function corresponds in style to functions in enhance.c
// It takes one image, and returns a status.
//
static MagickBooleanType rmsealpha (
  RmseAlphaT *pra,
  Image *inp_image, Image *sub_image,
  ExceptionInfo *exception)
{
  int
    precision;

  assert(inp_image != (Image *) NULL);
  assert(inp_image->signature == MAGICK_CORE_SIG);
  assert(sub_image != (Image *) NULL);
  assert(sub_image->signature == MAGICK_CORE_SIG);
  if (inp_image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",inp_image->filename);

  precision = GetMagickPrecision();

/*==
  pra->solnX = pra->solnY = 0;
  pra->score = 0;

  nPosY = inp_image->rows - ht_sub + 1;
  nPosX = inp_image->columns - wi_sub + 1;

  for (y=0; y < nPosY; y++)
  {
    register ssize_t
      x;

    double
      score;

    if (status == MagickFalse)
      continue;

    for (x=0; x < nPosX; x++)
    {
      score = CompareWindow (
        inp_image, sub_image,
        pra, inp_view, sub_view, wi_sub, ht_sub, x, y, exception);

      if (score < 0) status = MagickFalse;

      if (pra->score > score) {
        pra->score = score;
        pra->solnX = x;
        pra->solnY = y;
      }
      if (pra->score == 0.0) break;
    }
    if (pra->score == 0.0) break;
  }
==*/

  pra->warnSubSd = MagickFalse;

  if (pra->multiScale) {
    if (!subRmseAlphaMS (pra, inp_image, sub_image, exception)) {
      fprintf (stderr, "subRmseAlphaMS failed\n");
      return MagickFalse;
    }
  } else {
    if (!subRmseAlpha (pra, inp_image, sub_image, exception)) {
      fprintf (stderr, "subRmseAlpha failed\n");
      return MagickFalse;
    }
  }

  if (pra->warnSubSd)
    fprintf (stderr, "rmsealpha: Warning: subimage SD==0, so adjustLC matches all\n");


  if (pra->outJustScore) {
    fprintf (pra->outhandle, "%.*g",
             precision, pra->score);
  } else {

    if (pra->addCanvOffs) {
      pra->solnX += inp_image->page.x;
      pra->solnY += inp_image->page.y;
    }

    fprintf (pra->outhandle, "rmsealpha: %.*g @ %lu,%lu\n",
             precision, pra->score, pra->solnX, pra->solnY);
    fprintf (pra->outhandle, "rmsealphaCrop: %lix%li+%lu+%lu\n",
             sub_image->columns, sub_image->rows, pra->solnX, pra->solnY);
  }

  return MagickTrue;
}


ModuleExport size_t rmsealphaImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *inp_image,
    *sub_image;

  MagickBooleanType
    status;

  RmseAlphaT
    ra;

  (void) argc;
  (void) argv;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  InitRmseAlpha (&ra);

  status = menu (argc, argv, &ra);
  if (status == MagickFalse)
    return (-1);

  // Compare the images in pairs.
  // If we don't have even number of images, fatal error.

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    inp_image = image;
    image=GetNextImageInList(image);
    sub_image = image;
    if (sub_image == (Image *) NULL) {
      status = MagickFalse;
    } else {
      status = rmsealpha(&ra, inp_image, sub_image, exception);

      if (ra.saveInpScales || ra.saveSubScales) {
        // Do it again, to test the saving mechanism.
        status = rmsealpha(&ra, inp_image, sub_image, exception);
        // Do it again, to test the saving mechanism.
        status = rmsealpha(&ra, inp_image, sub_image, exception);
      }
    }

    ReInitRmseAlpha (&ra);

    if (status == MagickFalse)
      continue;
  }

  DeInitRmseAlpha (&ra);

  assert (status == MagickTrue);

  return(MagickImageFilterSignature);
}
