/*
  Updated:
    17-September-2017: if a dimension difference is zero,
      exclude it as a limiter.

    19-September-2017: sub-image larger than main image is no longer fatal.
      If sub-image is larger in either dimension,
      only one search will be made in that dimension.

    22-September-2017: added doSlideX and doSlideY. When false,
      even if sub-image smaller than main, windows will not slide in
      that direction.

    3-April-2018 for v7.0.7-28

    14-April-2018 added offset processing to subRmseAlphaMS
*/

#ifndef RMSEALPHA_INC
#define RMSEALPHA_INC

#include "calcmnsd.inc"

/* Possible optimisations:

   - When searching a large image multiple times,
     use integral images to speed up mean and SD calcs.
*/

#define RMSE_ALPHA_SIGNATURE 54321

typedef struct {
  MagickBooleanType
    do_verbose,
    avoidIdentical,
    dontDecreaseOpacity,
    outJustScore,
    doSlideX,
    doSlideY,
    addCanvOffs,    // whether caller will add canvas offset to result
    multiScale,     // Not used.
    saveInpScales,  // Whether to save and reuse scaled versions of inp_image
    saveSubScales;  // Whether to save and reuse scaled versions of sub_image

  FILE * outhandle;

  double
    adjustMeanSd,
    sizeDivider;

  int
    minDim;

  int
    supSampFact;

  // Returned values:
  ssize_t
    solnX,
    solnY;

  double
    solnXf,
    solnYf;

  double
    score; // RMSE, 0.0 to 1.0.

  MagickBooleanType
    warnSubSd;

  // Used internally:
  MeanSdT
    meanSd_inp,
    meanSd_sub;

  MagickBooleanType
    inpScalesSaved, // whether inp scale have been scaled, so can be reused
    subScalesSaved, // whether sub scale have been scaled, so can be reused
    mainHasAlpha,
    subHasAlpha;

  Image
    *inp_list,
    *sub_list;

  int
    signature;

} RmseAlphaT;

typedef struct {
  double gainR;
  double gainG;
  double gainB;
  double biasR;
  double biasG;
  double biasB;
} GainBiasT;

static void InitRmseAlpha (RmseAlphaT * pra)
{
  pra->do_verbose = pra->avoidIdentical = pra->dontDecreaseOpacity = MagickFalse;
  pra->outJustScore = MagickFalse;
  pra->multiScale = MagickFalse;
  pra->saveInpScales = MagickFalse;
  pra->saveSubScales = MagickFalse;
  pra->doSlideX = pra->doSlideY = MagickTrue;
  pra->addCanvOffs = MagickFalse;
  pra->outhandle = stderr;
  pra->adjustMeanSd = 0.0;

  pra->minDim = 20;
  pra->supSampFact = 1;
  pra->sizeDivider = 2.0;

  pra->inpScalesSaved = MagickFalse;
  pra->subScalesSaved = MagickFalse;
  pra->inp_list = pra->sub_list = NULL;

  pra->warnSubSd = MagickFalse;
  pra->solnX = pra->solnY = -1;
  pra->solnXf = pra->solnYf = -1;
  pra->score = -1;

  pra->mainHasAlpha = MagickFalse;
  pra->subHasAlpha = MagickFalse;

  pra->signature = RMSE_ALPHA_SIGNATURE;
}

#define CHK_RA_SIG(pra) { \
  if ((pra)->signature != RMSE_ALPHA_SIGNATURE) \
    fprintf (stderr, "\n** bad RMSE_ALPHA_SIGNATURE **\n"); \
}

static void ReInitRmseAlpha (RmseAlphaT * pra)
// After this, the returned values are still available.
{
  // When we record resizes, this will free them,
  // allowing for fresh searches.

  CHK_RA_SIG(pra);

  if (pra->signature != RMSE_ALPHA_SIGNATURE)
    fprintf (stderr, "ReInitRmseAlpha: bad sig\n");

  if (pra->inp_list) {
    if (pra->do_verbose) fprintf (stderr, "rmsealpha: DestroyImageList\n");
    pra->inp_list = DestroyImageList (pra->inp_list);
  }

  if (pra->sub_list) {
    if (pra->do_verbose) fprintf (stderr, "rmsealpha: DestroyImageList\n");
    pra->sub_list = DestroyImageList (pra->sub_list);
  }

  pra->inpScalesSaved = MagickFalse;
  pra->subScalesSaved = MagickFalse;
}

static void DeInitRmseAlpha (RmseAlphaT * pra)
// After this, the returned values are still available.
{
  CHK_RA_SIG(pra);

  ReInitRmseAlpha (pra);

  pra->signature = 0;
}

static void inline MeanSdToGainBiasOne (
  double adjustMeanSd,
  double mnA,
  double sdA,
  double mnB,
  double sdB,
  double * gain,
  double * bias
)
{
  double gn, bs;

  if (sdA > 0) {
    gn = sdB / sdA;
  } else {
    gn = 1.0;
  }
  bs = mnB - mnA * gn;
  *gain = (1 - adjustMeanSd + adjustMeanSd*gn);
  *bias = adjustMeanSd * bs * QuantumRange;
}

static void inline MeanSdToGainBias (
  double adjustMeanSd,
  MeanSdT * mnsdA,
  MeanSdT * mnsdB,
  GainBiasT * gb)
// Sets gb to required gain and bias to make A look like B.
{

  MeanSdToGainBiasOne (adjustMeanSd,
    mnsdA->mnR, mnsdA->sdR, mnsdB->mnR, mnsdB->sdR,
    &gb->gainR, &gb->biasR);

  MeanSdToGainBiasOne (adjustMeanSd,
    mnsdA->mnG, mnsdA->sdG, mnsdB->mnG, mnsdB->sdG,
    &gb->gainG, &gb->biasG);

  MeanSdToGainBiasOne (adjustMeanSd,
    mnsdA->mnB, mnsdA->sdB, mnsdB->mnB, mnsdB->sdB,
    &gb->gainB, &gb->biasB);
}

static double CompareWindow (
  const Image *inp_image,
  const Image *sub_image,
  RmseAlphaT *pra,
  CacheView * inp_view,
  CacheView * subimage_view,
  ssize_t width, ssize_t height,
  ssize_t offsX, ssize_t offsY,
  ExceptionInfo *exception)
//
// Compares width x height of subimage with same-size window in inp,
// starting at top-left location (offsX,offsY) of inp.
// Returns number between 0.0 and 1.0 inclusive.
//   0.0 means exact match
//   (but could be because sigma(alpha) is 0.0.
//
{
  GainBiasT gb = {1.0,1.0,1.0,0.0,0.0,0.0};

  if (pra->adjustMeanSd != 0.0) {
    if (!CalcMeanSdVP (
          inp_image, inp_view, width, height, offsX, offsY,
          &pra->meanSd_inp, exception))
    {
      return MagickFalse;  // FIXME: flag error?
    }

    if (pra->meanSd_inp.isMeanSdDefined) {

      if (pra->meanSd_inp.sdR == 0
       && pra->meanSd_inp.sdG == 0
       && pra->meanSd_inp.sdB == 0)
      {
        pra->warnSubSd = MagickTrue;
      }

      // We make the window on the input look more like the subimage.
      // We set gainX relative to 1.0, but biasX relative to QuantumRange.

      MeanSdToGainBias (pra->adjustMeanSd,
        &pra->meanSd_inp, &pra->meanSd_sub, &gb);
    }
  }

  double
    sigScore = 0,
    sigmAlpha = 0;

  MagickBooleanType
    isIdentical = MagickTrue,
    okay = MagickTrue;

  ssize_t y;

#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) \
    MAGICK_THREADS(inp_image,sub_image,height,1)
#endif

  for (y = 0; y < height; y++) {
    MagickBooleanType okayY = MagickTrue;
    //if (!okay) continue;
    const VIEW_PIX_PTR *inpy = GetCacheViewVirtualPixels (
      inp_view,offsX,y+offsY,width,1,exception);
    const VIEW_PIX_PTR *suby = GetCacheViewVirtualPixels (
      subimage_view,0,y,width,1,exception);

    if (inpy == (const VIEW_PIX_PTR *) NULL || suby == (const VIEW_PIX_PTR *) NULL) {
      fprintf (stderr, "CompareWindow: badgcvvp\n");
      okayY = MagickFalse;
      continue;
    }
    double sigScoreY=0, sigmAlphaY=0;
    double inpa = 1.0, suba = 1.0;
    ssize_t x;
    for (x = 0; x < width; x++) {
      if (pra->mainHasAlpha) inpa = GET_PIXEL_ALPHA(inp_image, inpy)/(double)QuantumRange;
      if (pra->subHasAlpha)  suba = GET_PIXEL_ALPHA(sub_image, suby)/(double)QuantumRange;

      double mAlpha = inpa * suba;

      double dRed   =
         ((GET_PIXEL_RED(inp_image,inpy)*gb.gainR+gb.biasR)
        - GET_PIXEL_RED(sub_image,suby))/QuantumRange;
      double dGreen =
         ((GET_PIXEL_GREEN(inp_image,inpy)*gb.gainG+gb.biasG)
        - GET_PIXEL_GREEN(sub_image,suby))/QuantumRange;
      double dBlue  =
         ((GET_PIXEL_BLUE(inp_image,inpy)*gb.gainB+gb.biasB)
        - GET_PIXEL_BLUE(sub_image,suby))/QuantumRange;

      if (pra->dontDecreaseOpacity && inpa < suba) {
        okayY = MagickFalse;
      }

      if (dRed!=0 || dGreen!=0 || dBlue!=0 || inpa!=suba) {
        isIdentical = MagickFalse;
      }

      if (mAlpha > 0) {
        sigScoreY  += mAlpha * (dRed*dRed + dGreen*dGreen + dBlue*dBlue);
        sigmAlphaY += mAlpha;
      }

      inpy += Inc_ViewPixPtr (inp_image);
      suby += Inc_ViewPixPtr (sub_image);
    }
#if defined(MAGICKCORE_OPENMP_SUPPORT)
#pragma omp atomic
#endif
    sigScore  += sigScoreY;
#if defined(MAGICKCORE_OPENMP_SUPPORT)
#pragma omp atomic
#endif
    sigmAlpha += sigmAlphaY;
#if defined(MAGICKCORE_OPENMP_SUPPORT)
#pragma omp atomic
#endif
    okay &= okayY;
  }
  if (!okay) return 1.0;

  if (pra->avoidIdentical && isIdentical) return 1.0;

  if (sigmAlpha==0) {
    if (pra->do_verbose) fprintf (stderr, " CompareWindow: All transparent ");
    return 0.0;
  }

  return sqrt (sigScore / sigmAlpha / 3.0);
}


static MagickBooleanType subRmseAlpha (
  RmseAlphaT *pra,
  const Image *inp_image, const Image *sub_image,
  ExceptionInfo *exception)
{
  /* If sub_image is larger than inp_image in either direction,
     the extra pixels at right/bottom of sub_image are ignored,
     so there is only one search in that direction.

     Effectively, sub_image is cropped at right/bottom.
  */

  assert(inp_image != (Image *) NULL);
  assert(inp_image->signature == MAGICK_CORE_SIG);
  assert(sub_image != (Image *) NULL);
  assert(sub_image->signature == MAGICK_CORE_SIG);

  if (pra->signature != RMSE_ALPHA_SIGNATURE) return MagickFalse;

  ssize_t wi_sub = sub_image->columns;
  ssize_t ht_sub = sub_image->rows;

/*== Test replaced 19-September-2017.
  if (wi_sub > inp_image->columns || ht_sub > inp_image->rows) {
    fprintf (stderr, "subRmseAlpha: Subimage %lix%li larger than input %lix%li.\n",
       wi_sub, ht_sub,
       inp_image->columns, inp_image->rows);
    return MagickFalse;
  }
==*/

  if (wi_sub > inp_image->columns) wi_sub = inp_image->columns;
  if (ht_sub > inp_image->rows)    ht_sub = inp_image->rows;

  CacheView * inp_view = AcquireVirtualCacheView(inp_image,exception);
  CacheView * sub_view = AcquireVirtualCacheView(sub_image,exception);

  pra->solnX = pra->solnY = 0;
  pra->score = 99e9;

  if (pra->adjustMeanSd != 0.0) {
    if (!CalcMeanSdVP (
          sub_image, sub_view, wi_sub, ht_sub, 0, 0,
          &pra->meanSd_sub, exception))
    {
      return MagickFalse;
    }
    if (pra->do_verbose) WrMeanSd (&pra->meanSd_sub, stderr);
    if (!pra->meanSd_sub.isMeanSdDefined) {
      pra->adjustMeanSd = 0.0;
    }
  }

  ssize_t nPosX = inp_image->columns - wi_sub + 1;
  ssize_t nPosY = inp_image->rows    - ht_sub + 1;

  if (!pra->doSlideX) nPosX = 1;
  if (!pra->doSlideY) nPosY = 1;

  pra->mainHasAlpha = IS_ALPHA_CH(inp_image);
  pra->subHasAlpha  = IS_ALPHA_CH(sub_image);

  pra->warnSubSd = MagickFalse;

  if (pra->do_verbose) {
    fprintf (stderr, "subRmseAlpha: inp %lix%li (%s alpha)  sub %lix%li (%s alpha)  nPos %lix%li",
      inp_image->columns, inp_image->rows, (pra->mainHasAlpha) ? "has" : "no",
      wi_sub, ht_sub, (pra->subHasAlpha) ? "has" : "no",
      nPosX, nPosY);
  }

  ssize_t y;
  MagickBooleanType okay = MagickTrue;
  for (y=0; y < nPosY; y++)
  {
    ssize_t
      x;

    double
      score;

    if (okay == MagickFalse)
      continue;

    for (x=0; x < nPosX; x++)
    {
      score = CompareWindow (
        inp_image, sub_image,
        pra, inp_view, sub_view, wi_sub, ht_sub, x, y, exception);

      if (score < 0) okay = MagickFalse;
      else if (pra->score > score) {
        pra->score = score;
        pra->solnX = x;
        pra->solnY = y;

        if (pra->score == 0.0) break;
      }
    }
    if (pra->score == 0.0) break;
  }

  if (pra->do_verbose) {
    fprintf (stderr, " %g @ %li,%li\n",
      pra->score, pra->solnX, pra->solnY);
  }

  pra->solnXf = pra->solnX;
  pra->solnYf = pra->solnY;

  sub_view = DestroyCacheView (sub_view);
  inp_view = DestroyCacheView (inp_view);

  return okay;
}


static void WrList (Image * img, FILE * fh)
{
  if (!img) {
    fprintf (fh, "List empty\n");
  } else {
    while (img) {
      fprintf (fh, "%lix%li\n", img->columns, img->rows);
      img = GetNextImageInList (img);
    }
  }
}


// Each image can be in only one list.
// We don't want to mess with the main list of 5,
// so clone images for rmse lists.

static MagickBooleanType subRmseAlphaMSOne (
  RmseAlphaT *pra,
  const Image *inp_image, const Image *sub_image,
  ExceptionInfo *exception)
// Multi-scale subimage search.
// This calls itself recursively.
{
  assert(inp_image != (Image *) NULL);
  assert(inp_image->signature == MAGICK_CORE_SIG);
  assert(sub_image != (Image *) NULL);
  assert(sub_image->signature == MAGICK_CORE_SIG);

  if (pra->signature != RMSE_ALPHA_SIGNATURE) return MagickFalse;

  ssize_t smDim = inp_image->columns;
  if (smDim > inp_image->rows)    smDim = inp_image->rows;
  if (smDim > sub_image->columns) smDim = sub_image->columns;
  if (smDim > sub_image->rows)    smDim = sub_image->rows;

  // 17-September-2017: if a diff is zero, exclude it as a limiter.

  ssize_t dimDiff = inp_image->columns-sub_image->columns;
  if (dimDiff && smDim > dimDiff) smDim = dimDiff;

  dimDiff = inp_image->rows-sub_image->rows;
  if (dimDiff && smDim > dimDiff) smDim = dimDiff;

  if (pra->do_verbose) {
    fprintf (stderr, "subRmseAlphaMSOne: %lix%li+%li+%li %lix%li+%li+%li smDim %li\n",
      inp_image->columns, inp_image->rows,
      inp_image->page.x, inp_image->page.y,
      sub_image->columns, sub_image->rows,
      sub_image->page.x, sub_image->page.y,
      smDim);
  }

  if (pra->saveInpScales && !pra->inpScalesSaved) {
    // Save clone of this image to end of inp_list.
    Image * copy_img = CloneImage(inp_image, 0, 0, MagickTrue, exception);
    if (!copy_img) return MagickFalse;

    AppendImageToList (&pra->inp_list, copy_img);
    if (pra->do_verbose) {
      fprintf (stderr, "AppendImage inp %lix%li to %lix%li\n",
        copy_img->columns, copy_img->rows,
        pra->inp_list->columns, pra->inp_list->rows);
      WrList (pra->inp_list, stderr);
    }
  }

  if (pra->saveSubScales && !pra->subScalesSaved) {
    // Save clone of this image to end of sub_list.
    Image * copy_img = CloneImage(sub_image, 0, 0, MagickTrue, exception);
    if (!copy_img) return MagickFalse;

    AppendImageToList (&pra->sub_list, copy_img);
    if (pra->do_verbose) {
      fprintf (stderr, "AppendImage sub %lix%li to %lix%li\n",
        copy_img->columns, copy_img->rows,
        pra->sub_list->columns, pra->sub_list->rows);
      WrList (pra->sub_list, stderr);
    }
  }

  if (smDim < pra->minDim)
    return subRmseAlpha (pra, inp_image, sub_image, exception);

  Image *inp_sm, *sub_sm;

  if (pra->saveInpScales && pra->inpScalesSaved) {
    // Get small version from list.
    // It is the next one after inp_image.
    if (pra->do_verbose) {
      fprintf (stderr, "GetNextImage inp after %lix%li\n",
        inp_image->columns, inp_image->rows);
    }

    inp_sm = GetNextImageInList (inp_image);

    // If we don't have a small image,
    // create it. (This can happen if subimages are different sizes.)
    if (!inp_sm) {
      fprintf (stderr, "subRmseAlphaMSOne: no small inp\n");
      WrList (pra->inp_list, stderr);
      // FIXME: temp:
      inp_sm = RESIZEIMG (inp_image,
        inp_image->columns/pra->sizeDivider, inp_image->rows/pra->sizeDivider,
        exception);
      // FIXME: Append this to the list?
      Image * copy_img = CloneImage(inp_sm, 0, 0, MagickTrue, exception);
      if (!copy_img) return MagickFalse;

      AppendImageToList (&pra->inp_list, copy_img);
    }

    if (pra->do_verbose) {
      fprintf (stderr, "Got Image inp %lix%li\n",
        inp_sm->columns, inp_sm->rows);
    }
  } else {
    if (pra->do_verbose) {
      fprintf (stderr, "Resizing Image inp\n");
    }
    inp_sm = RESIZEIMG (inp_image,
      inp_image->columns/pra->sizeDivider, inp_image->rows/pra->sizeDivider,
      exception);
  }
  if (!inp_sm) return MagickFalse;

  if (pra->saveSubScales && pra->subScalesSaved) {
    // Get small version from list.
    // It is the next one after sub_image.
    if (pra->do_verbose) {
      fprintf (stderr, "GetNextImage sub after %lix%li\n",
        sub_image->columns, sub_image->rows);
    }

    sub_sm = GetNextImageInList (sub_image);

    // If we don't have a small image,
    // create it. (This can happen if subimages are different sizes.)
    if (!sub_sm) {
      fprintf (stderr, "subRmseAlphaMSOne: no small sub\n");
      WrList (pra->sub_list, stderr);
    }

    if (pra->do_verbose) {
      fprintf (stderr, "Got Image sub %lix%li\n",
        sub_sm->columns, sub_sm->rows);
    }
  } else {
    if (pra->do_verbose) {
      fprintf (stderr, "Resizing Image sub\n");
    }
    sub_sm = RESIZEIMG (sub_image,
      sub_image->columns/pra->sizeDivider, sub_image->rows/pra->sizeDivider,
      exception);
  }
  if (!sub_sm) return MagickFalse;


  if (!subRmseAlphaMSOne (pra, inp_sm, sub_sm, exception))
    return MagickFalse;

  if (!pra->subScalesSaved) sub_sm = DestroyImage (sub_sm);
  if (!pra->inpScalesSaved) inp_sm = DestroyImage (inp_sm);

  int plusMinus = floor (pra->sizeDivider + 2.5);

  RectangleInfo geom;
  geom.x = pra->solnX * pra->sizeDivider - plusMinus;
  geom.y = pra->solnY * pra->sizeDivider - plusMinus;
  if (geom.x < 0) geom.x = 0;
  if (geom.y < 0) geom.y = 0;
  geom.width  = sub_image->columns + 2*plusMinus;
  geom.height = sub_image->rows    + 2*plusMinus;

  if (geom.x + geom.width > inp_image->columns)
    geom.width = inp_image->columns - geom.x;

  if (geom.y + geom.height > inp_image->rows)
    geom.height = inp_image->rows - geom.y;

  if (pra->do_verbose) {
    fprintf (stderr,
      "subRmseAlphaMSOne: crop geom %lix%li+%li+%li out of %lix%li %lix%li+%li+%li \n",
      geom.width, geom.height, geom.x, geom.y,
      inp_image->columns, inp_image->rows,
      inp_image->page.width, inp_image->page.height,
      inp_image->page.x, inp_image->page.y);
  }

  Image * inp_crp = CropImage (inp_image, &geom, exception);
  if (!inp_crp) return MagickFalse;

  inp_crp->page.width = inp_crp->page.height
    = inp_crp->page.x = inp_crp->page.y = 0;

  if (!subRmseAlpha (pra, inp_crp, sub_image, exception))
    return MagickFalse;

  inp_crp = DestroyImage (inp_crp);

  if (pra->do_verbose) {
    WrList (pra->inp_list, stderr);
    WrList (pra->sub_list, stderr);
  }

  // Add the geometry offsets back
  pra->solnX += geom.x;
  pra->solnY += geom.y;

  pra->solnXf = pra->solnX;
  pra->solnYf = pra->solnY;

  return MagickTrue;
}

static MagickBooleanType subRmseSuper (
  RmseAlphaT *pra,
  Image *inp_image, Image *sub_image,
  ExceptionInfo *exception)
// Supersampling subimage search.
// Assumes pra->solnX and pra->solnY are the best integral solutions.
// Sets pra->solnXf and pra->solnYf.
{
  // Clone cropped inp_image, and resize it.
  // Clone sub_image, and resize it.
  // Search.
  // Adjust the numbers.
  int
    dl=0,
    dt=0,
    dr=0,
    db=0;

  if (pra->solnX > 0) dl = 1;
  if (pra->solnY > 0) dt = 1;

  if (dl + pra->solnX + sub_image->columns < inp_image->columns) dr = 1;
  if (dt + pra->solnY + sub_image->rows    < inp_image->rows   ) db = 1;

  RectangleInfo geom;
  geom.x = pra->solnX - dl;
  geom.y = pra->solnY - dt;
  geom.width = sub_image->columns + dl + dr;
  geom.height = sub_image->rows + dt + db;

  if (pra->do_verbose)
    fprintf (stderr, "subRmseSuper geom: %lix%li+%li+%li\n",
             geom.width, geom.height, geom.x, geom.y);

  Image * inp_crp = CropImage (inp_image, &geom, exception);
  if (!inp_crp) return MagickFalse;

  inp_crp->page.width = inp_crp->page.height
    = inp_crp->page.x = inp_crp->page.y = 0;

  // Resize both.

  Image * inp_res = RESIZEIMG (inp_crp,
    inp_crp->columns*pra->supSampFact, inp_crp->rows*pra->supSampFact,
        exception);

  if (!inp_res) {
    return MagickFalse;
  }

  inp_crp = DestroyImage (inp_crp);

  Image * sub_res = RESIZEIMG (sub_image,
    sub_image->columns*pra->supSampFact, sub_image->rows*pra->supSampFact,
        exception);

  if (!sub_res) {
    return MagickFalse;
  }

  if (!subRmseAlphaMSOne (pra, inp_res, sub_res, exception))
    return MagickFalse;

  sub_res = DestroyImage (sub_res);
  inp_res = DestroyImage (inp_res);

  pra->solnXf = geom.x + pra->solnX / (double)pra->supSampFact;
  pra->solnYf = geom.y + pra->solnY / (double)pra->supSampFact;

  pra->solnX = floor (pra->solnXf + 0.5);
  pra->solnY = floor (pra->solnYf + 0.5);

  return MagickTrue;
}

static MagickBooleanType subRmseAlphaMS (
  RmseAlphaT *pra,
  Image *inp_image, Image *sub_image,
  ExceptionInfo *exception)
// Multi-scale subimage search.
{
  CHK_RA_SIG(pra);

  if (pra->signature != RMSE_ALPHA_SIGNATURE) {
    fprintf (stderr, "rmsealpha: bad signature\n");
    return MagickFalse;
  }

  Image * inp_i = inp_image;
  if (pra->inpScalesSaved) {
    if (pra->do_verbose) fprintf (stderr, "subRmseAlphaMS: inpScalesSaved\n");
    inp_i = pra->inp_list;
  }

  Image * sub_i = sub_image;
  if (pra->subScalesSaved) {
    if (pra->do_verbose) fprintf (stderr, "subRmseAlphaMS: subScalesSaved\n");
    sub_i = pra->sub_list;
  }

  ssize_t ix = inp_i->page.x;
  ssize_t iy = inp_i->page.y;

  ssize_t sx = sub_i->page.x;
  ssize_t sy = sub_i->page.y;

  inp_i->page.x = sub_i->page.x =
  inp_i->page.y = sub_i->page.y = 0;

  MagickBooleanType r = subRmseAlphaMSOne (
    pra, inp_i, sub_i, exception);

  pra->solnX += ix;
  pra->solnY += iy;

  pra->solnXf = pra->solnX;
  pra->solnYf = pra->solnY;

  if (pra->supSampFact > 1) {
    if (!subRmseSuper (pra, inp_i, sub_i, exception)) {
      fprintf (stderr, "rmsealpha: bad subRmseSuper\n");
      return MagickFalse;
    }
  }

  inp_i->page.x = ix;
  inp_i->page.y = iy;
  sub_i->page.x = sx;
  sub_i->page.y = sy;

  if (pra->saveInpScales) {
    pra->inpScalesSaved = MagickTrue;
    if (pra->do_verbose) WrList (pra->inp_list, stderr);
  }

  if (pra->saveSubScales) {
    pra->subScalesSaved = MagickTrue;
    if (pra->do_verbose) WrList (pra->sub_list, stderr);
  }

  return r;
}

#endif
