/* Last updated:
     5-September-2017 use NEXTARG.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"

// CAUTION: When an image is a constant colour,
//   IM's GetImageMean() returns a standard_deviation of NaN,
//   "not a number" instead of zero.

typedef enum {
  fullframe,
  circular
} FormatT;

typedef enum {
  dIncOnly,
  dDecOnly,
  dBoth
} directionT;

typedef struct {
  double
    mid_pt,
    goal_sd,
    goal_mn,
    tolerance,
    initCon0,
    initCon2;

  int
    precision;

  directionT
    direction;

  MagickBooleanType
    has_sd_goal,
    has_mn_goal,
    pin_sd,
    pin_mn,
    mid_is_mean;

  int
    do_verbose;

  FILE *
    fh_data;

  MagickBooleanType
    sharpen;

  double
    upper_goal_sd,
    lower_goal_sd,
    mid_pt_q,
    fnd_mean,
    fnd_sd,
    error,
    pow;

  int
    nIter;
} setmnsdT;


static void usage (void)
{
  printf ("Usage: -process 'setmnsd [OPTION]...'\n");
  printf ("Adjusts mean and standard deviation.\n");
  printf ("\n");
  printf ("  sd, sdGoal N        goal for standard deviation [none]\n");
  printf ("  mn, meanGoal N      goal for mean [none]\n");
  printf ("  t, tolerance N      tolerance for results [0.00001]\n");
  printf ("  m, mid N            mid-point (percentage of quantum) [mean]\n");
  printf ("  i0, initCon0 N      lower guess for contrast [0]\n");
  printf ("  i2, initCon2 N      upper guess for contrast [1000]\n");
  printf ("  d, direction string one of 'incOnly', 'decOnly', 'both' [both]\n");
  printf ("  f, file string      write to file stream stdout or stderr\n");
  printf ("  v, verbose          write text information\n");
  printf ("  v2, verbose2        write more text information\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}

static inline MagickBooleanType EndsPc (const char *s)
{
  char c = *(s+strlen(s)-1);
  if (c == '%' || c == 'c')
    return MagickTrue;
  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "srchimg: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  setmnsdT * psss
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  psss->do_verbose = 0;

  psss->goal_sd = 0.166667;
  psss->goal_mn = 0.5;
  psss->has_sd_goal = MagickFalse;
  psss->has_mn_goal = MagickFalse;
  psss->pin_sd = MagickFalse;
  psss->pin_mn = MagickFalse;
  psss->tolerance = 0.00001;
  psss->error = psss->tolerance;
  psss->mid_pt = 50;
  psss->mid_is_mean = MagickTrue;
  psss->direction = dBoth;
  psss->fh_data = stderr;
  psss->initCon0 = 0.0000001;
  psss->initCon2 = 30.0;
  psss->pow = 1.0;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "sd", "sdGoal")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "pin")==0) {
        psss->pin_sd = MagickTrue;
      } else {
        psss->goal_sd = atof(argv[i]);
        if (EndsPc (argv[i])) psss->goal_sd /= 100.0;
      }
      psss->has_sd_goal = MagickTrue;
    } else if (IsArg (pa, "mn", "meanGoal")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "pin")==0) {
        psss->pin_mn = MagickTrue;
      } else {
        psss->goal_mn = atof(argv[i]);
        if (EndsPc (argv[i])) psss->goal_mn /= 100.0;
      }
      psss->has_mn_goal = MagickTrue;
    } else if (IsArg (pa, "t", "tolerance")==MagickTrue) {
      NEXTARG;
      psss->tolerance = atof(argv[i]);
      if (EndsPc (argv[i])) psss->tolerance /= 100.0;
    } else if (IsArg (pa, "m", "mid")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "mean")==0) {
        psss->mid_is_mean = MagickTrue;
      } else {
        psss->mid_is_mean = MagickFalse;
        psss->mid_pt = atof(argv[i]);
      }
    } else if (IsArg (pa, "i0", "initCon0")==MagickTrue) {
      NEXTARG;
      psss->initCon0 = atof(argv[i]);
    } else if (IsArg (pa, "i2", "initCon2")==MagickTrue) {
      NEXTARG;
      psss->initCon2 = atof(argv[i]);
    } else if (IsArg (pa, "d", "direction")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "incOnly")==0) {
        psss->direction = dIncOnly;
      } else if (LocaleCompare(argv[i], "decOnly")==0) {
        psss->direction = dDecOnly;
      } else if (LocaleCompare(argv[i], "both")==0) {
        psss->direction = dBoth;
      } else {
        fprintf (stderr, "setmnsd: ERROR: direction: invalid option [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      NEXTARG;
      if (strcasecmp (argv[i], "stdout")==0) psss->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) psss->fh_data = stderr;
      else status = MagickFalse;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      psss->do_verbose = 1;
    } else if (IsArg (pa, "v2", "verbose2")==MagickTrue) {
      psss->do_verbose = 2;
    } else {
      fprintf (stderr, "setmnsd: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (psss->do_verbose) {
    fprintf (stderr, "setmnsd options:");

    if (psss->do_verbose > 1) fprintf (stderr, "  verbose2");
    else fprintf (stderr, "  verbose");

    if (psss->has_mn_goal) {
      if (psss->pin_mn) fprintf (stderr, "  meanGoal pin");
      else fprintf (stderr, "  meanGoal %.*g", psss->precision, psss->goal_mn);
    }
    if (psss->has_sd_goal) {
      if (psss->pin_sd) fprintf (stderr, "  sdGoal pin");
      else fprintf (stderr, "  sdGoal %.*g", psss->precision, psss->goal_sd);
    }

    fprintf (stderr, "  tolerance %.*g", psss->precision, psss->tolerance);
    if (psss->mid_is_mean) {
      fprintf (stderr, "  mid mean");
    } else {
      fprintf (stderr, "  mid %.*g", psss->precision, psss->mid_pt);
    }

    fprintf (stderr, "  initCon0 %.*g", psss->precision, psss->initCon0);
    fprintf (stderr, "  initCon2 %.*g", psss->precision, psss->initCon2);
    fprintf (stderr, "  direction ");
    switch (psss->direction) {
      case dIncOnly: fprintf (stderr, "incOnly");
      case dDecOnly: fprintf (stderr, "decOnly");
      case dBoth:    fprintf (stderr, "both");
    }

    if (psss->fh_data == stdout) fprintf (stderr, " file stdout");

    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


static MagickBooleanType ApplySigmoid (
  Image *image,
  MagickBooleanType sharpen,
  double contrast,
  double mid,
  ExceptionInfo *exception)
{
#if IMV6OR7==6
  return SigmoidalContrastImageChannel (
    image, DefaultChannels, sharpen, contrast, mid);
#else
  return SigmoidalContrastImage (
    image, sharpen, contrast, mid, exception);
#endif

}


static Image *TryContrast (
  Image *image,
  setmnsdT * psss,
  double contrast,
  ExceptionInfo *exception)
{
  // Make a clone of this image, copied pixel values:
  //
  Image *temp_image = CloneImage(image, 0, 0, MagickTrue, exception);
  if (temp_image == (Image *) NULL)
    return(temp_image);

//  if (SetNoPalette (temp_image, exception) == MagickFalse)
//    return (Image *)NULL;

  psss->nIter++;
  if (!ApplySigmoid (temp_image, psss->sharpen, contrast, psss->mid_pt_q, exception))
    return (Image *)NULL;

  // FIXME: Beware: next doesn't treat alpha as I would like.
  if (!GetImageMean (temp_image, &psss->fnd_mean, &psss->fnd_sd, exception))
    return (Image *)NULL;
  if (isnan(psss->fnd_sd)) psss->fnd_sd = 0;

  psss->pow = 1.0;
  if (psss->has_mn_goal) {

    if (psss->do_verbose > 1) fprintf (psss->fh_data, "goal_mn=%.*g  ",
      psss->precision, psss->goal_mn);

    // We special-case 0.0 and 1.0 to avoid not-a-number, infinity etc.

    if (psss->goal_mn == 0.0) {
      psss->pow = 9e99;
      SetAllBlack (temp_image, exception);
    } else if (psss->goal_mn == 1.0) {
      psss->pow = 0.0;
      SetAllOneCol (image, "white", exception);
    } else {
      // We have a "real" goal for mean.
      // Iterate until fabs(onePow-1) < tolerance.
      int i;
      for (i=0; i < 10; i++) {
        double onePow = log(psss->goal_mn) / log(psss->fnd_mean/(double)QuantumRange);
        psss->pow *= onePow;

        if (fabs(onePow-1) < psss->error) break;

        if (psss->do_verbose > 1) fprintf (psss->fh_data, "before pow: onePow=%.*g pow=%.*g: mean=%.*g sd=%.*g  ",
          psss->precision, onePow,
          psss->precision, psss->pow,
          psss->precision, psss->fnd_mean / (double)QuantumRange,
          psss->precision, psss->fnd_sd / (double)QuantumRange);

        psss->nIter++;

        if (!EvaluateImage (temp_image, PowEvaluateOperator, onePow, exception))
          return (Image *)NULL;

        if (!GetImageMean (temp_image, &psss->fnd_mean, &psss->fnd_sd, exception))
          return (Image *)NULL;
        if (isnan(psss->fnd_sd)) psss->fnd_sd = 0;
      }
    }
  }

  if (psss->do_verbose > 1) fprintf (psss->fh_data, "contrast=%.*g: mean=%.*g sd=%.*g  ",
    psss->precision, contrast,
    psss->precision, psss->fnd_mean / (double)QuantumRange,
    psss->precision, psss->fnd_sd / (double)QuantumRange);

  double err = fabs (psss->fnd_sd/(double)QuantumRange - psss->goal_sd);
  psss->error = (err > psss->tolerance) ? err : psss->tolerance;

  return (temp_image);
}


static void AnnounceResult (
  Image *image,
  setmnsdT * psss,
  double contrast,
  ExceptionInfo *exception)
{
  if (psss->do_verbose) {
    fprintf (psss->fh_data, "nIter=%i\n", psss->nIter);

    fprintf (psss->fh_data, "result: mean=%.*g  sd=%.*g\n",
      psss->precision, psss->fnd_mean / (double)QuantumRange,
      psss->precision, psss->fnd_sd / (double)QuantumRange);
  }

  char msg[200] = "";

  if (contrast != 0) {
    sprintf (msg, "%csigmoidal-contrast %.*g,%.*g%%",
      psss->sharpen ? '-':'+',
      psss->precision, contrast,
      psss->precision, psss->mid_pt_q * 100.0 / (double)QuantumRange);
  }

  if (psss->has_mn_goal) {
    char msg2[100];
    sprintf (msg2, " -evaluate pow %.*g",
      psss->precision, psss->pow);
    strcat (msg, msg2);
  }

  // If no change, make a "no-op" operation.
  if (!*msg) strcpy (msg, "-evaluate Add 0");

  if (psss->do_verbose) {
    fprintf (psss->fh_data, "setmnsd command: %s\n", msg);
  }

#if IMV6OR7==6
  SetImageProperty (image, "filter:setmnsd", msg);
#else
  SetImageProperty (image, "filter:setmnsd", msg, exception);
#endif
}


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image (or the input image, unchanged).
//
static Image *setmnsd (
  Image *image,
  setmnsdT * psss,
  ExceptionInfo *exception)
{
  assert(image != (Image *) NULL);
  assert(image->signature == MAGICK_CORE_SIG);

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

//  if (SetNoPalette (image, exception) == MagickFalse)
//    return (Image *)NULL;

  psss->nIter = 0;

  double tol_q = psss->tolerance * QuantumRange;

  psss->mid_pt_q = psss->mid_pt/100.0 * QuantumRange;

  {
    double goal_sd_q = psss->goal_sd * QuantumRange;
    psss->upper_goal_sd = goal_sd_q + tol_q;
    psss->lower_goal_sd = goal_sd_q - tol_q;
  }

  double contrast0 = psss->initCon0;
  double contrast1;
  double contrast2 = psss->initCon2;

  if (psss->do_verbose > 1) fprintf (psss->fh_data, "setmnsd: contrasts %.*g %.*g\n",
    psss->precision, contrast0,
    psss->precision, contrast2);

  double mean, stddev;
  if (!GetImageMean (image, &mean, &stddev, exception))
    return (Image *)NULL;
  if (isnan(stddev)) stddev = 0;

  if (psss->do_verbose) fprintf (psss->fh_data, "input: mean=%.*g sd=%.*g\n",
    psss->precision, mean / (double)QuantumRange,
    psss->precision, stddev / (double)QuantumRange);

  if (stddev==0 && psss->has_sd_goal) {
    if (psss->do_verbose) fprintf (psss->fh_data, "setmnsd Warning: Input SD is zero, so ignoring SD goal\n");
    psss->has_sd_goal = MagickFalse;
    psss->pin_sd = MagickFalse;
  }

  if ((mean<=0 || mean==QuantumRange) && psss->has_mn_goal) {
    if (psss->do_verbose) fprintf (psss->fh_data, "setmnsd Warning: Input mean is <=zero or 100%%, so ignoring mean goal\n");
    psss->has_mn_goal = MagickFalse;
    psss->pin_mn = MagickFalse;
  }

  if (psss->pin_sd) {
    psss->goal_sd = stddev / (double)QuantumRange;
    double goal_sd_q = psss->goal_sd * QuantumRange;
    psss->upper_goal_sd = goal_sd_q + tol_q;
    psss->lower_goal_sd = goal_sd_q - tol_q;
  }

  if (psss->pin_mn) psss->goal_mn = mean / (double)QuantumRange;

  if (psss->mid_is_mean) {
    psss->mid_pt_q = mean;
  }

  Image * iInit = TryContrast (image, psss, 0, exception);
  if (iInit == (Image *)NULL) return (iInit);

  if (!psss->has_sd_goal) {
    // That's all we have to do.
    AnnounceResult (iInit, psss, 0, exception);
    return (iInit);
  }

  iInit = DestroyImage (iInit);

  mean = psss->fnd_mean;
  stddev = psss->fnd_sd;

  psss->sharpen = MagickTrue; // whether to increase contrast in middle

  if (stddev > psss->upper_goal_sd) {
    if (psss->do_verbose > 1) fprintf (psss->fh_data, "above upper\n");
    psss->sharpen = MagickFalse;
    if (psss->direction == dIncOnly) return (image);
    double t = contrast0;
    contrast0 = contrast2;
    contrast2 = t;
  } else if (stddev < psss->lower_goal_sd) {
    if (psss->do_verbose > 1) fprintf (psss->fh_data, "below lower\n");
    if (psss->direction == dDecOnly) return (image);
    psss->sharpen = MagickTrue;
  } else {
    if (psss->do_verbose > 1) fprintf (psss->fh_data, "within tolerance\n");
    AnnounceResult (image, psss, 0, exception);
    return (image);
  }


  Image * i0 = TryContrast (image, psss, contrast0, exception);
  if (i0 == (Image *)NULL) return (i0);

  if (psss->fnd_sd > psss->upper_goal_sd) {
    if (psss->do_verbose) fprintf (psss->fh_data, "i0 above upper -- fail\n");
    fprintf (stderr, "i0 C=%.*g stddev too high: %.*g\n",
      psss->precision, contrast0,
      psss->precision, psss->fnd_sd / (double)QuantumRange);
    i0 = DestroyImage (i0);
    return (i0);
  } else if (psss->fnd_sd < psss->lower_goal_sd) {
    if (psss->do_verbose > 1) fprintf (psss->fh_data, "i0 below lower -- good\n");
  } else {
    if (psss->do_verbose > 1) fprintf (psss->fh_data, "i0 within tolerance\n");
    AnnounceResult (i0, psss, contrast0, exception);
    return (i0);
  }

  i0 = DestroyImage (i0);


  Image * i2 = TryContrast (image, psss, contrast2, exception);
  if (i2 == (Image *)NULL) return (i2);

  if (psss->fnd_sd > psss->upper_goal_sd) {
    if (psss->do_verbose > 1) fprintf (psss->fh_data, "i2 above upper -- good\n");
  } else if (psss->fnd_sd < psss->lower_goal_sd) {
    if (psss->do_verbose) fprintf (psss->fh_data, "i2 below lower -- fail\n");
    fprintf (stderr, "i2 C=%.*g stddev too low: %.*g\n",
      psss->precision, contrast2,
      psss->precision, psss->fnd_sd / (double)QuantumRange);
    i2 = DestroyImage (i2);
    return (i2);
  } else {
    if (psss->do_verbose > 1) fprintf (psss->fh_data, "i2 within tolerance\n");
    AnnounceResult (i2, psss, contrast2, exception);
    return (i2);
  }

  i2 = DestroyImage (i2);

  // Limit the number of iterations, just in case...
  int i;
  for (i=0; i < 100; i++) {
    double prod = contrast0 * contrast2;
    if (prod > 0) contrast1 = sqrt (contrast0 * contrast2);
    else contrast1 = (contrast0 + contrast2) / 2.0;

    Image * i1 = TryContrast (image, psss, contrast1, exception);
    if (i1 == (Image *)NULL) return (i1);

    if (psss->fnd_sd > psss->upper_goal_sd) {
      if (psss->do_verbose > 1) fprintf (psss->fh_data, "i1 above upper\n");
      contrast2 = contrast1;
    } else if (psss->fnd_sd < psss->lower_goal_sd) {
      if (psss->do_verbose > 1) fprintf (psss->fh_data, "i1 below lower\n");
      contrast0 = contrast1;
    } else {
      if (psss->do_verbose > 1) fprintf (psss->fh_data, "i1 within tolerance\n");

      // If mean is a goal, do it again with tighter tolerance on mean.
      if (psss->has_mn_goal) {
        i1 = DestroyImage (i1);
        i1 = TryContrast (image, psss, contrast1, exception);
        if (i1 == (Image *)NULL) return (i1);
      }

      AnnounceResult (i1, psss, contrast1, exception);
      return (i1);
    }

    i1 = DestroyImage (i1);
  }

  if (psss->do_verbose) fprintf (psss->fh_data, "Bust. Using contrast1 = %.*g\n", psss->precision, contrast1);

  Image * i1 = TryContrast (image, psss, contrast1, exception);
  if (i1 == (Image *)NULL) return (i1);

  AnnounceResult (i1, psss, contrast1, exception);
  return (i1);
}



ModuleExport size_t setmnsdImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  setmnsdT
    sss;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  sss.precision = GetMagickPrecision();

  status = menu (argc, argv, &sss);
  if (status == MagickFalse)
    return (-1);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = setmnsd (image, &sss, exception);
    if (new_image == (Image *) NULL)
      return(-1);

    if (new_image != image) ReplaceImageInList(&image,new_image);

    *images=GetFirstImageInList(image);
  }

  return(MagickImageFilterSignature);
}
