#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"

static void inline SwapPixels (const VIEW_PIX_PTR *a, const VIEW_PIX_PTR *b, int nBytes)
{
  char * pa = (char *) a;
  char * pb = (char *) b;
  char t;

  int i;
  for (i=0; i < nBytes; i++) {
    t = *pa;
    *pa++ = *pb;
    *pb++ = t;
  }
}

// The next function corresponds in style to functions in enhance.c
// It takes two images, modifies both, and returns a status.
//
static MagickBooleanType ShadowSortPixImage(Image **images,
  ExceptionInfo *exception)
{
  CacheView
    *imageA_view, *imageB_view;

  ssize_t
    y;

  MagickBooleanType
    status = MagickTrue;


  Image * imageA = *images;

  assert(imageA != (Image *) NULL);
  assert(imageA->signature == MAGICK_CORE_SIG);
  if (imageA->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",imageA->filename);

  Image * imageB = GetNextImageInList (imageA);
  if (!imageB) {
    fprintf (stderr, "shadowsortpixels: needs two images, the same size\n");
    return MagickFalse;
  }

  if (imageA->columns != imageB->columns ||
      imageA->rows    != imageB->rows)
  {
    fprintf (stderr, "shadowsortpixels: the two images must be the same size\n");
    return MagickFalse;
  }

  if (SetNoPalette (imageA, exception) == MagickFalse)
    return MagickFalse;

  if (SetNoPalette (imageB, exception) == MagickFalse)
    return MagickFalse;

  const ssize_t incA = Inc_ViewPixPtr (imageA);
  const ssize_t incB = Inc_ViewPixPtr (imageB);

  const int nBytesA = sizeof (VIEW_PIX_PTR) * incA;
  const int nBytesB = sizeof (VIEW_PIX_PTR) * incB;

  imageA_view=AcquireAuthenticCacheView(imageA,exception);
  imageB_view=AcquireAuthenticCacheView(imageB,exception);

#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) shared(status) \
    MAGICK_THREADS(imageA,imageA,imageA->rows,1)
#endif
  for (y=0; y < (ssize_t) imageA->rows; y++)
  {
    VIEW_PIX_PTR
      *pA, *pB;

    if (status == MagickFalse)
      continue;

    pA=GetCacheViewAuthenticPixels(imageA_view,0,y,imageA->columns,1,exception);
    if (!pA)
    {
      status=MagickFalse;
      continue;
    }

    pB=GetCacheViewAuthenticPixels(imageB_view,0,y,imageB->columns,1,exception);
    if (!pB)
    {
      status=MagickFalse;
      continue;
    }

    // Shell sort.

    ssize_t x, xSwitch, xLimit;

    ssize_t xOffset = imageA->columns / 2;

    while (xOffset) {

      xLimit = imageA->columns - xOffset - 1 ;
      do {
        xSwitch = 0;     // Assume no switches at this offset.

        // Compare elements and switch ones out of order.
        for( x = 0; x <= xLimit; x++ ) {
          const VIEW_PIX_PTR * a = pA + x*incA;
          const VIEW_PIX_PTR * b = pA + (x+xOffset)*incA;

          if (GetPixelIntensity(imageA, a) > GetPixelIntensity(imageA, b)) {
            SwapPixels (a, b, nBytesA);
            SwapPixels (pB + x*incB, pB + (x+xOffset)*incB, nBytesB);
            xSwitch = x;
          }
        }

        // Sort on next pass only to where last switch was made.
        xLimit = xSwitch - xOffset;
      } while (xSwitch);

      // No switches at last offset, try one half as big.
      xOffset /= 2;
    }

    if (SyncCacheViewAuthenticPixels(imageA_view,exception) == MagickFalse)
      status=MagickFalse;

    if (SyncCacheViewAuthenticPixels(imageB_view,exception) == MagickFalse)
      status=MagickFalse;
  }
  imageA_view=DestroyCacheView(imageA_view);
  imageB_view=DestroyCacheView(imageB_view);

  return(status);
}

ModuleExport size_t shadowsortpixelsImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  (void) argc;
  (void) argv;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  int ListLen = (int)GetImageListLength(*images);
  if (ListLen != 2) {
    fprintf (stderr, "shadowsortpixels: needs 2 images\n");
    return (-1);
  }

  if (!ShadowSortPixImage (images, exception)) return (-1);

  return(MagickImageFilterSignature);
}
