/* Updated:
     27-October-2017 added version; introduced QoverLev2m1.
     3-April-2018 for v7.0.7-28
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "vsn_defines.h"

#define DEBUG 0

#define VERSION "sphaldcl v1.0  Copyright (c) 2017 Alan Gibson"

typedef enum {
  mIdentity,
  mShepards,
  mVoronoi
} methodT;

typedef struct {
  int index;
  double distance;
} distanceT;

typedef struct {
  MagickBooleanType
    do_verbose;

  methodT
    method;

  int
    haldLevel,
    nearest;

  double
    powParam;

  distanceT
    *distances;
} sphaldclT;


static void usage (void)
{
  printf ("Usage: -process 'sphaldcl [OPTION]...'\n");
  printf ("Replace last image (Nx2) with hald clut.\n");
  printf ("\n");
  printf ("  h, haldLevel N      hald level [8]\n");
  printf ("  m, method String    one of: identity, shepards, voronoi [identity]\n");
  printf ("  p, power N          power [2]\n");
  printf ("  n, nearest N        use just nearest (0=all) [0]\n");
  printf ("  v, verbose          write text information to stdout\n");
  printf ("      version          write version information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}

#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "sphaldcl: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  sphaldclT * pat
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  pat->method = mIdentity;
  pat->haldLevel = 8;
  pat->powParam = 2;
  pat->nearest = 0;
  pat->do_verbose = MagickFalse;
  pat->distances = NULL;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "h", "haldlevel")==MagickTrue) {
      NEXTARG;
      pat->haldLevel = atoi(argv[i]);
    } else if (IsArg (pa, "p", "power")==MagickTrue) {
      NEXTARG;
      pat->powParam = atof(argv[i]);
    } else if (IsArg (pa, "m", "method")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "identity")==0) pat->method = mIdentity;
      else if (LocaleCompare(argv[i], "shepards")==0) pat->method = mShepards;
      else if (LocaleCompare(argv[i], "voronoi")==0) pat->method = mVoronoi;
      else {
        fprintf (stderr, "sphaldcl: ERROR: unknown method [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "n", "nearest")==MagickTrue) {
      NEXTARG;
      pat->nearest = atoi(argv[i]);
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pat->do_verbose = MagickTrue;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "sphaldcl: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (pat->do_verbose) {
    fprintf (stderr, "sphaldcl options:");

    if (pat->do_verbose) fprintf (stderr, "  verbose");

    fprintf (stderr, "  haldLevel %i", pat->haldLevel);
    fprintf (stderr, "  method ");
    switch (pat->method) {
      case mIdentity:
        fprintf (stderr, "identity");
        break;
      case mShepards:
        fprintf (stderr, "shepards");
        fprintf (stderr, "  power %g", pat->powParam);
        break;
      case mVoronoi:
        fprintf (stderr, "voronoi");
        break;
    }
    fprintf (stderr, "  nearest %i", pat->nearest);

    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}

static double inline colDist (
  Image *image,
  const VIEW_PIX_PTR *p0,
  const VIEW_PIX_PTR *p1
)
// Returns distance between colours, scale 0.0 to 1.0.
{
  double r = GET_PIXEL_RED(image,p0)   - GET_PIXEL_RED(image,p1);
  double g = GET_PIXEL_GREEN(image,p0) - GET_PIXEL_GREEN(image,p1);
  double b = GET_PIXEL_BLUE(image,p0)  - GET_PIXEL_BLUE(image,p1);

  return sqrt ((r*r + g*g + b*b) / 3.0) / QuantumRange;
}

static double inline colDistRel (
  Image *image,
  const VIEW_PIX_PTR *p0,
  const VIEW_PIX_PTR *p1
)
// Returns relative distance between colours.
{
  double r = GET_PIXEL_RED(image,p0)   - GET_PIXEL_RED(image,p1);
  double g = GET_PIXEL_GREEN(image,p0) - GET_PIXEL_GREEN(image,p1);
  double b = GET_PIXEL_BLUE(image,p0)  - GET_PIXEL_BLUE(image,p1);

  return (r*r + g*g + b*b);
}

static void inline calcWeight (
  sphaldclT * pat,
  Image *image,
  const VIEW_PIX_PTR *p0,
  const VIEW_PIX_PTR *p1,
  double * weight,
  MagickBooleanType *DistIsZero
)
{
  double dist = colDist (image, p0, p1);
//printf ("dist=%g\n", dist);
  *DistIsZero = (dist < 1e-8);
  if (*DistIsZero) *weight = 0;
  else *weight = 1.0 / pow (dist, pat->powParam);
}


static void inline calcShepard (
  sphaldclT * pat,
  Image *image,
  const VIEW_PIX_PTR *p0,  // row of input RGB
  const VIEW_PIX_PTR *p1,  // row of differences
  ssize_t inRed,  // 0 to QuantumRange
  ssize_t inGreen,
  ssize_t inBlue,
  // Ouputs are differences, to be added to identity.
  MagickRealType * outRed,
  MagickRealType * outGreen,
  MagickRealType * outBlue
)
// Voronoi: return the colour of the sample nearest input coords.
{
  VIEW_PIX_PTR pIn;

  const VIEW_PIX_PTR *ppOut;

  SET_PIXEL_RED   (image, inRed, &pIn);
  SET_PIXEL_GREEN (image, inGreen, &pIn);
  SET_PIXEL_BLUE  (image, inBlue, &pIn);

  double weight;
  MagickBooleanType DistIsZero;

  int i;

  double sumWeight = 0.0, sumR = 0.0, sumG = 0.0, sumB = 0.0;

  for (i=0; i < (ssize_t) image->columns; i++) {

    calcWeight (
      pat, image, 
      &pIn, p0 + i * Inc_ViewPixPtr (image), 
      &weight, &DistIsZero);

    ppOut = p1 + i * Inc_ViewPixPtr (image);

    if (DistIsZero) {
      *outRed   = GET_PIXEL_RED (image, ppOut);
      *outGreen = GET_PIXEL_GREEN (image, ppOut);
      *outBlue  = GET_PIXEL_BLUE (image, ppOut);
      sumWeight = 0;
      break;
    }

    sumWeight += weight;
    sumR += weight * GET_PIXEL_RED (image, ppOut);
    sumG += weight * GET_PIXEL_GREEN (image, ppOut);
    sumB += weight * GET_PIXEL_BLUE (image, ppOut);
  }
  if (sumWeight > 0) {
    *outRed   = sumR / sumWeight;
    *outGreen = sumG / sumWeight;
    *outBlue  = sumB / sumWeight;
  }

//printf ("outRGB = %g %g %g\n", *outRed, *outGreen, *outBlue);
}


static void inline calcVoronoi (
  sphaldclT * pat,
  Image *image,
  const VIEW_PIX_PTR *p0,  // row of input RGB
  const VIEW_PIX_PTR *p1,  // row of differences
  ssize_t inRed,  // 0 to QuantumRange
  ssize_t inGreen,
  ssize_t inBlue,
  // Ouputs are differences, to be added to identity.
  MagickRealType * outRed,
  MagickRealType * outGreen,
  MagickRealType * outBlue
)
{
  VIEW_PIX_PTR pIn;

  SET_PIXEL_RED   (image, inRed, &pIn);
  SET_PIXEL_GREEN (image, inGreen, &pIn);
  SET_PIXEL_BLUE  (image, inBlue, &pIn);

  double minDist = 99e99;

  int i, iAtMin = 0;

  for (i=0; i < (ssize_t) image->columns; i++) {

    double dist = colDistRel (image, &pIn, p0 + i * Inc_ViewPixPtr (image));

    if (minDist > dist) {
      minDist = dist;
      iAtMin = i;
      if (minDist == 0) break;
    }
  }
  const VIEW_PIX_PTR *ppOut = p1 + iAtMin * Inc_ViewPixPtr (image);

  *outRed = GET_PIXEL_RED (image, ppOut);
  *outGreen = GET_PIXEL_GREEN (image, ppOut);
  *outBlue = GET_PIXEL_BLUE (image, ppOut);

//printf ("outRGB = %g %g %g\n", *outRed, *outGreen, *outBlue);
}


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *sphaldcl (
  Image *image,
  sphaldclT * pat,
  ExceptionInfo *exception)
{
  Image
    *new_image;

  CacheView
    *in_view,
    *out_view;

  ssize_t
    y;

  MagickBooleanType
    status;


  assert(image != (Image *) NULL);
  assert(image->signature == MAGICK_CORE_SIG);

  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (pat->do_verbose) {
    fprintf (stderr,
          "sphaldcl: Input image [%s]  depth %i  size %ix%i\n",
          image->filename, (int)image->depth,
          (int)image->columns, (int)image->rows);
  }

  if (image->rows != 2 ) {
    fprintf (stderr, "sphaldcl: image needs height 2\n");
    return (Image *)NULL;
  }

  if (SetNoPalette (image, exception) == MagickFalse)
    return (Image *)NULL;

  // Make a clone of this image, undefined pixel values:
  //
  int sqSize = pat->haldLevel * pat->haldLevel * pat->haldLevel;

  new_image = CloneImage (image, sqSize, sqSize, MagickTrue, exception);
  if (new_image == (Image *) NULL) {
    fprintf (stderr, "sphaldcl clone failed %i\n", sqSize);
    return(new_image);
  }

  if (SetNoPalette (new_image, exception) == MagickFalse)
    return (Image *)NULL;

  in_view = AcquireVirtualCacheView (image,exception);

  out_view = AcquireAuthenticCacheView (new_image,exception);

  status = MagickTrue;

  const VIEW_PIX_PTR *p0=GetCacheViewVirtualPixels(in_view,0,0,image->columns,1,exception);
  if (p0 == (const VIEW_PIX_PTR *) NULL) {
    return (Image *)NULL;
  }

  const VIEW_PIX_PTR *p1=GetCacheViewVirtualPixels(in_view,0,1,image->columns,1,exception);
  if (p1 == (const VIEW_PIX_PTR *) NULL) {
    return (Image *)NULL;
  }

  pat->distances = (distanceT *) AcquireQuantumMemory (image->columns, sizeof(distanceT));
  if (!pat->distances) {
    fprintf (stderr, "sphaldcl: oom\n");
    return (Image *)NULL;
  }

  const int lev = pat->haldLevel;
  const int lev2 = pat->haldLevel * pat->haldLevel;
  const int lev2m1 = lev2 - 1;

  const double QoverLev2m1 = QuantumRange / (double)lev2m1;

#if defined(MAGICKCORE_OPENMP_SUPPORT)
  #pragma omp parallel for schedule(static,4) shared(status) \
    MAGICK_THREADS(new_image,new_image,new_image->rows,1)
#endif
  for (y=0; y < (ssize_t) new_image->rows; y++)
  {
    VIEW_PIX_PTR
      *q;

    ssize_t
      x;

    MagickRealType
      fr=0, fg=0, fb=0;

    if (status == MagickFalse)
      continue;

    q=GetCacheViewAuthenticPixels(out_view,0,y,new_image->columns,1,exception);
    if (q == (const VIEW_PIX_PTR *) NULL)
    {
      status=MagickFalse;
      continue;
    }

    ssize_t
      inRed, inGreen, inBlue;

    for (x=0; x < (ssize_t) new_image->columns; x++)
    {
      inRed   = x % lev2;
      inGreen = x / lev2 + (y % lev) * lev;
      inBlue  = y / lev;

      switch (pat->method) {
        case mIdentity: {
          SET_PIXEL_RED   (new_image, QoverLev2m1 * inRed  , q);
          SET_PIXEL_GREEN (new_image, QoverLev2m1 * inGreen, q);
          SET_PIXEL_BLUE  (new_image, QoverLev2m1 * inBlue , q);

#if DEBUG==1
          ssize_t x2 = inRed % lev2 + (inGreen % lev) * lev2;
          ssize_t y2 = inBlue * lev + inGreen / lev;
          if ((x != x2) || (y != y2))
            printf ("xy=%li,%li xy2=%li,%li\n", x, y, x2, y2);
#endif
          break;
        }

        case mShepards:
        default: {
          calcShepard (pat, image, p0, p1,
            inRed   * QoverLev2m1,
            inGreen * QoverLev2m1,
            inBlue  * QoverLev2m1,
            &fr, &fg, &fb);

          SET_PIXEL_RED   (new_image, fr + QoverLev2m1 * inRed  , q);
          SET_PIXEL_GREEN (new_image, fg + QoverLev2m1 * inGreen, q);
          SET_PIXEL_BLUE  (new_image, fb + QoverLev2m1 * inBlue , q);
          break;
        }
        case mVoronoi: {
          calcVoronoi (pat, image, p0, p1,
            inRed   * QoverLev2m1,
            inGreen * QoverLev2m1,
            inBlue  * QoverLev2m1,
            &fr, &fg, &fb);

          SET_PIXEL_RED   (new_image, fr + QoverLev2m1 * inRed  , q);
          SET_PIXEL_GREEN (new_image, fg + QoverLev2m1 * inGreen, q);
          SET_PIXEL_BLUE  (new_image, fb + QoverLev2m1 * inBlue , q);
          break;
        }
      }

      q += Inc_ViewPixPtr (new_image);
    }
    if (SyncCacheViewAuthenticPixels(out_view,exception) == MagickFalse)
      status=MagickFalse;
  }
  out_view = DestroyCacheView (out_view);
  in_view = DestroyCacheView (in_view);

  if (pat->distances)
    pat->distances = RelinquishMagickMemory (pat->distances);

  if (status == MagickFalse) return NULL;

  return (new_image);
}


ModuleExport size_t sphaldclImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  Image
    *image1,
    *new_image;

  MagickBooleanType
    status;

  sphaldclT
    at;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &at);
  if (status == MagickFalse)
    return (-1);

  image1=GetLastImageInList (*images);

  new_image = sphaldcl (image1, &at, exception);
  if (new_image == (Image *) NULL)
    return(-1);

  ReplaceImageInList (images,new_image);
  // Replace messes up the images pointer. Make it good:
  *images = GetFirstImageInList (*images);

  return(MagickImageFilterSignature);
}
