rem  Searches image %1 for smaller image %2.
@rem Aggressively resizes both images down by same factors so small image is 1x1.
@rem Finds best match by difference composition.
@rem This gives crop location(s) in large image for full search(es).
@rem
@rem  Also uses:
@rem
@rem    siaSIZE_METH    sizing method: resize or scale
@rem    siaN_BEST       for the second (full-size) search,
@rem                    try this number of results from the first search.
@rem                      Default = 1 = only try the best.
@rem    siaDEBUG        if 1, creates a marked-up copy of image.
@rem    siaDEBUG_STROKE eg "-fill None -stroke #f00 -strokewidth 10" to get thick strokes
@rem
@rem    siMETRIC        defaults to RMSE
@rem    siDELTEMPSRC    if not 0, will remove temporary source (%1) files.
@rem    siDELTEMPSUB    if not 0, will remove temporary subimage (%2) files.
@rem    siDIV_DIM       Divide min dim by this, for purpose of choosing resizes. [1]
@rem    siDISSIM_THR    Dissimilarity threshold. [1]
@rem
@rem  Returns:
@rem    siCOMP_XW, siCOMP_YW coord of top-left of found area
@rem    siCOMP_FLT  RMSE difference beween %2 and found area
@rem    siCOMP_CROP=%subW%x%subH%+%COMP_XW%+%COMP_YW%  crop spec of found area
@rem    siIS_DISSIM 1 if images too dissimilar; otherwise 0
@rem
@rem Updated:
@rem   27-August-2022 for IM v7.
@rem


@if "%2"=="" findstr /B "rem @rem" %~f0 & exit /B 1

@setlocal enabledelayedexpansion

@call echoOffSave

call %PICTBAT%setInOut %1 sisc


set siERROR=0

if "%siaDEBUG%"=="1" (
  set DEBUG_OUT=%BASENAME%_sia_dbg%EXT%
  if "%siaDEBUG_STROKE%"=="" set siaDEBUG_STROKE=-fill None -stroke #f00
)

set SRCNAME=%~n1
set SRC=%INFILE%
set SUB=%2

if "%siMETRIC%"=="" set siMETRIC=RMSE

set TMPEXT=.miff
set TMPDIR=%TEMP%

set SRC_PREF=%TMPDIR%\siaSrc_%SRCNAME%
set TMPSRC=%SRC_PREF%%TMPEXT%
set TMPSUB=%TMPDIR%\siaSub%TMPEXT%
set TMPDIFF=%TMPDIR%\siaSub_diff%TMPEXT%

if "%siDISSIM_THR%"=="" set siDISSIM_THR=1

if "%siaN_BEST%"=="" set siaN_BEST=1

set subW=
for /F "usebackq" %%L in (`%IMG7%magick ^
  %SUB% ^
  -format "subW=%%w\nsubH=%%h" ^
  info:`) do set /A %%L

if "%subW%"=="" (
  echo Can't open SUB [%SUB%]
  exit /B 1
)

set srcW=
for /F "usebackq" %%L in (`%IMG7%magick ^
  %SRC% ^
  -format "srcW=%%w\nsrcH=%%h" ^
  info:`) do set /A %%L

if "%srcW%"=="" (
  echo Can't open SRC [%SRC%]
  exit /B 1
)

if not %srcW% GTR %subW% (
  echo Src width must be greater than sub width
  exit /B 1
)

if not %srcH% GTR %subH% (
  echo Src height must be greater than sub height
  exit /B 1
)

for /F "usebackq" %%L in (`%IMG7%magick identify ^
  -format "newSrcW=%%[fx:int(%srcW%/%subW%+0.5)]\nnewSrcH=%%[fx:int(%srcH%/%subH%+0.5)]" ^
  xc:`) do set /A %%L

for /F "usebackq" %%L in (`%IMG7%magick identify ^
  -format "factW=%%[fx:%srcW%/%newSrcW%]\nfactH=%%[fx:%srcH%/%newSrcH%]" ^
  xc:`) do set %%L

rem %0: echo newSrcW=%newSrcW% newSrcH=%newSrcH% factW=%factW% factH=%factH%

if "%siaN_BEST%"=="1" (
  set sSTRETCH=
) else (
  set sSTRETCH=-linear-stretch 0x%siaN_BEST%
)

%IMG7%magick ^
  %SRC% ^
  -resize "%newSrcW%x%newSrcH%^!" ^
  ( %SUB% ^
    -scale "1x1^!" ^
    -scale "%newSrcW%x%newSrcH%^!" ^
  ) ^
  -channel RGB ^
  -compose Difference -composite ^
  +channel ^
  -grayscale RMS ^
  -negate ^
  %sSTRETCH% ^
  -auto-level ^
  +depth ^
  %TMPDIFF%

if ERRORLEVEL 1 (
  echo %0: bad magick
  exit /B 1
)

set /A cropW=3*%subW%
set /A cropH=3*%subH%

:: Ensure srchImg.bat deletes resized images.
set siDELTEMPSRC=1
set siDELTEMPSUB=1

set BEST_XW=0
set BEST_YW=0
set BEST_FLT=999

for /F "usebackq skip=1 tokens=1-2 delims=, " %%X in (`%IM7DEV%magick ^
  %TMPDIFF% ^
  -process allwhite ^
  NULL: 2^>^&1`) do (

  for /F "usebackq" %%L in (`%IMG7%magick identify ^
    -format "COMP_XW=%%[fx:int(%%X*%factW%+0.5)]\nCOMP_YW=%%[fx:int(%%Y*%factH%+0.5)]" ^
    xc:`) do set /A %%L

  rem echo %0: COMP_XW=!COMP_XW! COMP_YW=!COMP_YW!

  rem Now do the crop search.

  set /A cropL=!COMP_XW!-%subW%
  set /A cropT=!COMP_YW!-%subH%
  if !cropL! LSS 0 set cropL=0
  if !cropT! LSS 0 set cropT=0

  %IMG7%magick ^
    %SRC% ^
    +repage ^
    -crop %cropW%x%cropH%+!cropL!+!cropT! ^
    +repage ^
    %TMPSRC%

  if ERRORLEVEL 1 (
    set siERROR=1
    exit /B 1
  )

  call %PICTBAT%srchImg %TMPSRC% %SUB%

  set /A COMP_XW=!siCOMP_XW!+!cropL!
  set /A COMP_YW=!siCOMP_YW!+!cropT!

echo %0: echo COMP_XW=!COMP_XW!  COMP_YW=!COMP_YW!  siCOMP_FLT=!siCOMP_FLT!

  for /F "usebackq" %%L in (`%IMG7%magick identify ^
    -format "IS_BETTER=%%[fx:!BEST_FLT!>!siCOMP_FLT!?1:0]" ^
    xc:`) do set /A %%L

  if !IS_BETTER!==1 (
    set BEST_XW=!COMP_XW!
    set BEST_YW=!COMP_YW!
    set BEST_FLT=!siCOMP_FLT!
  )

)

echo %0: echo BEST_XW=!BEST_XW!  BEST_YW=!BEST_YW!  BEST_FLT=!BEST_FLT!

if "%siaDEBUG%"=="1" (
  set /A endX=%BEST_XW%+%subW%-1
  set /A endY=%BEST_YW%+%subH%-1

  %IMG7%magick ^
    %SRC% ^
    +repage ^
    %siaDEBUG_STROKE% ^
    -draw "rectangle %BEST_XW%,%BEST_YW% !endX!,!endY!" ^
    %DEBUG_OUT%
)

call echoRestore

@endlocal & set siCOMP_XW=%BEST_XW%& set siCOMP_YW=%BEST_YW%& set siCOMP_FLT=%BEST_FLT%& set siIS_DISSIM=%siIS_DISSIM%& set siCOMP_CROP=%subW%x%subH%+%BEST_XW%+%BEST_YW%

@exit /B 0

