/*
   Reference: http://im.snibgo.com/srchimg.htm
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <ctype.h>
#include "vsn_defines.h"

#include "rmsealpha.inc"

#define VERSION "srchimg v1.0  Copyright (c) 2017 Alan Gibson"

typedef struct {
  int
    precision;

  MagickBooleanType
    doCrop,
    addCanvOffs,    // whether to add canvas offset to result
    verbose;

  double
    sizeDivider;

  int
    minDim;

  int
    supSampFact;

  FILE *
    fh_data;
} srchimgT;

static void usage (void)
{
  printf ("Usage: -process 'srchimg [OPTION]...'\n");
  printf ("Searches for second image in first.\n");
  printf ("\n");
  printf ("  n,   noCrop              don't replace images with crop\n");
  printf ("  z,   sizeDiv number      size divider for each level\n");
  printf ("  md,  minDim integer      minimum dimension for ms\n");
  printf ("  ss,  superSample number  factor for supersampling\n");
  printf ("  cnv, canvasOffset        add canvas offset to result\n");
  printf ("  f,   file string         write to file stream stdout or stderr\n");
  printf ("  v,   verbose             write text information to stdout\n");
  printf ("       version             write version information to stdout\n");
  printf ("\n");
}

/* TODO:
   Action on input offsets:
   - ignore (effectively "+repage" at start)
   - regard ("+repage" but add in at the end)
*/

static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  // LocaleCompare is not case-sensitive,
  // so we use strcmp for the short option.

  if ((strcmp(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "srchimg: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  srchimgT * psi
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status = MagickTrue;

  psi->verbose = MagickFalse;
  psi->doCrop = MagickTrue;
  psi->addCanvOffs = MagickFalse;
  psi->sizeDivider = 2.0;
  psi->minDim = 20;
  psi->supSampFact = 1;
  psi->fh_data = stderr;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];
    //printf ("Arg %i [%s]\n", i, pa);

    if (IsArg (pa, "n", "noCrop")==MagickTrue) {
      psi->doCrop = MagickFalse;
    } else if (IsArg (pa, "z", "sizeDiv")==MagickTrue) {
      NEXTARG;
      psi->sizeDivider = atof(argv[i]);
    } else if (IsArg (pa, "md", "minDim")==MagickTrue) {
      NEXTARG;
      psi->minDim = atoi(argv[i]);
    } else if (IsArg (pa, "ss", "superSample")==MagickTrue) {
      NEXTARG;
      psi->supSampFact = atoi(argv[i]);
    } else if (IsArg (pa, "cnv", "canvasOffset")==MagickTrue) {
      psi->addCanvOffs = MagickTrue;
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      NEXTARG;
      if (strcasecmp (argv[i], "stdout")==0) psi->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) psi->fh_data = stderr;
      else status = MagickFalse;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      psi->verbose = MagickTrue;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "srchimg: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (psi->verbose) {
    fprintf (stderr, "srchimg options:");

    if (!psi->doCrop) fprintf (stderr, "  noCrop");
    fprintf (stderr, "  sizeDiv %g", psi->sizeDivider);
    fprintf (stderr, "  minDim %i", psi->minDim);
    if (psi->supSampFact > 1) fprintf (stderr, "  superSample %i", psi->supSampFact);
    if (psi->addCanvOffs) fprintf (stderr, "  canvasOffset");

    if (psi->fh_data == stdout) fprintf (stderr, "  file stdout");

    if (psi->verbose)    fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


// The next function takes an image list,
// may modify it, and returns a status.
//
static MagickBooleanType srchimg (
  Image **images,
  srchimgT * psi,
  ExceptionInfo *exception)
{
  if ((*images)->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",(*images)->filename);

  int ListLen = (int)GetImageListLength(*images);
  if (ListLen !=2) {
    fprintf (stderr, "srchimg needs exactly 2 images\n");
    return MagickFalse;
  }

  Image * inp_image = *images;
  Image * sub_image = GetNextImageInList (inp_image);

  if (psi->verbose) {
    fprintf (stderr, "srchimg: Input image [%s] %ix%i depth is %i\n",
             inp_image->filename,
             (int)inp_image->columns, (int)inp_image->rows,
             (int)inp_image->depth);
  }

  RmseAlphaT ra;
  InitRmseAlpha (&ra);
  ra.do_verbose  = psi->verbose;
  ra.multiScale  = MagickTrue;
  ra.sizeDivider = psi->sizeDivider;
  ra.minDim      = psi->minDim;
  ra.supSampFact = psi->supSampFact;

  if (!subRmseAlphaMS (&ra, inp_image, sub_image, exception)) {
    fprintf (stderr, "subRmseAlphaMS failed\n");
    return MagickFalse;
  }

  DeInitRmseAlpha (&ra);

  ssize_t offsX=0, offsY=0;

  if (psi->addCanvOffs) {
    offsX = inp_image->page.x;
    offsY = inp_image->page.y;
  }

  double dx = ra.solnXf + offsX;
  double dy = ra.solnYf + offsY;

  // fprintf (psi->fh_data, "%.*g @ %li,%li\n",
  //          psi->precision, ra.score,
  //          ra.solnX + offsX, ra.solnY + offsY);

  fprintf (psi->fh_data, "%.*g @ %g,%g\n",
           psi->precision, ra.score,
           dx, dy);

  if (psi->doCrop) {
    Image * inp_crp;
    if (psi->supSampFact == 1) {
      RectangleInfo geom;
      geom.x = ra.solnX;
      geom.y = ra.solnY;
      geom.width = sub_image->columns;
      geom.height = sub_image->rows;
      inp_crp = CropImage (inp_image, &geom, exception);
      if (!inp_crp) return MagickFalse;
    } else {
      // fixme: also set viewport

      char text[MaxTextExtent];

      sprintf (text, "%lux%lu+0+0",
               sub_image->columns, sub_image->rows);

#if IMV6OR7==6
      SetImageProperty (inp_image, "viewport", text);
#else
      SetImageProperty (inp_image, "viewport", text, exception);
#endif


      MagickBooleanType bestfit = MagickTrue;
      double srtArray[6];
      srtArray[0] = dx;
      srtArray[1] = dx;
      srtArray[2] = 1;
      srtArray[3] = 0;
      srtArray[4] = 0;
      srtArray[5] = 0;

      inp_crp = DistortImage (inp_image, ScaleRotateTranslateDistortion,
          6, srtArray, bestfit, exception);
     if (!inp_crp) { fprintf (stderr, "DistortImage failed\n"); return MagickFalse; }
    }

    DeleteImageFromList (&sub_image);

    ReplaceImageInList (&inp_image, inp_crp);
    // Replace messes up the images pointer. Make it good:
    *images = GetFirstImageInList (inp_image);
  }

  if (psi->verbose) {
    fprintf (stderr, "Finished srchimg\n");
  }

  return MagickTrue;
}


ModuleExport size_t srchimgImage (
  Image **images,
  const int argc,
  const char **argv,
  ExceptionInfo *exception)
{
  srchimgT si;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  MagickBooleanType status = menu (argc, argv, &si);
  if (!status) return -1;

  si.precision = GetMagickPrecision();

  status = srchimg (images, &si, exception);
  if (!status) return (-1);

  return(MagickImageFilterSignature);
}
