#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"

#include "srchmask.inc"

#define VERSION "srchmask v1.0  Copyright (c) 2017 Alan Gibson"

typedef struct {
  int
    verbose;

  MagickBooleanType
    updateMask,
    makeComposite,
    elimClose;

  int
    precision;

  FILE *
    fh_data;

} SubSrchT;

/*
   Optionally dumps text.
   Optionally applies transformation to subimage, merging with main image.
*/

static void InitSubSrch (SubSrchT * pss)
{
  pss->verbose = 0;
  pss->updateMask = MagickFalse;
  pss->makeComposite = MagickFalse;
  pss->elimClose = MagickFalse;
  pss->precision = 6;
  pss->fh_data = stderr;
}

static void DeInitSubSrch (SubSrchT * pss)
{
  ; // Nothing.
}


static void usage (void)
{
  printf ("Usage: -process 'srchmask [OPTION]...'\n");
  printf ("Find rotation by unrolling and golden section methods.\n");
  printf ("\n");
  printf ("  u,  updateMask       update with only the 'good' positions\n");
  printf ("  c,  makeComposite    make composite output image\n");
  printf ("  e,  elimClose        eliminate close solutions\n");
  printf ("  f,  file string      write data to file stream stdout or stderr\n");
  printf ("  v,  verbose          write text information to stdout\n");
  printf ("  v2, verbose2         write more information to stdout\n");
  printf ("      version          write version information to stdout\n");
  printf ("\n");
}

// FIXME: also option to normalise mean and SD.


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "paintpatches: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  SubSrchT * pss
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "c", "makeComposite")==MagickTrue) {
      pss->makeComposite = MagickTrue;
    } else if (IsArg (pa, "e", "elimClose")==MagickTrue) {
      pss->elimClose = MagickTrue;
    } else if (IsArg (pa, "u", "updateMask")==MagickTrue) {
      pss->updateMask = MagickTrue;
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      NEXTARG;
      if (strcasecmp (argv[i], "stdout")==0) pss->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) pss->fh_data = stderr;
      else status = MagickFalse;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pss->verbose = 1;
    } else if (IsArg (pa, "v2", "verbose2")==MagickTrue) {
      pss->verbose = 2;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "srchmask: ERROR: unknown option\n");
      status = MagickFalse;
    }
  }

  if (pss->verbose) {
    fprintf (stderr, "srchmask options:");
    if (pss->updateMask) fprintf (stderr, "  updateMask");
    if (pss->makeComposite) fprintf (stderr, "  makeComposite");

    if (pss->fh_data == stdout) fprintf (stderr, " file stdout");
    if (pss->verbose==2) fprintf (stderr, "  verbose2");
    else if (pss->verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


// The next function takes an image list,
// may modify it, and returns a status.
//
static MagickBooleanType srchmask (
  Image **images,
  SubSrchT * pss,
  ExceptionInfo *exception)
{
  int ListLen = (int)GetImageListLength(*images);

  Image * img1 = *images;
  Image * img2 = GetNextImageInList (img1);
  Image * img3 = NULL;
  if (img2) img3 = GetNextImageInList (img2);

  Image *main_img=NULL, *sub_img=NULL, *mask_img=NULL;

  switch (ListLen) {
    case 1:
      mask_img = img1;
      break;

    case 2:
      sub_img = img1;
      mask_img = img2;
      break;

    case 3:
      main_img = img1;
      sub_img = img2;
      mask_img = img3;
      break;
  }

  if (main_img) fprintf (stderr,
    "main_img %lix%li  ", main_img->columns, main_img->rows);

  if (sub_img) fprintf (stderr,
    "sub_img %lix%li  ", sub_img->columns, sub_img->rows);

  if (mask_img) fprintf (stderr,
    "mask_img %lix%li", mask_img->columns, mask_img->rows);

  fprintf (stderr, "\n");


/*==
  if (sub_img->columns > main_img->columns ||
      sub_img->rows    > main_img->rows)
  {
    fprintf (stderr, "srchconcr: subimage is larger than main\n");
    return MagickFalse;
  }
==*/

//#if IMV6OR7==6
//  SetPixelCacheVirtualMethod (main_img, TransparentVirtualPixelMethod);
//#else
//  SetPixelCacheVirtualMethod (main_img,
//    TransparentVirtualPixelMethod, exception);
//#endif


  srchMaskT sm;
  InitSrchMask (&sm);
  sm.precision = GetMagickPrecision();
  sm.verbose = pss->verbose;

  if (main_img) {
//    sm.main_width = main_img->columns;
//    sm.main_height = main_img->rows;
  }

  if (sub_img) {
    sm.sub_width = sub_img->columns;
    sm.sub_height = sub_img->rows;
  }

  ReadSrchMask (mask_img, &sm, exception);
  DumpSrchMask (stderr, &sm);
  SortSrchMask (&sm);
  DumpSrchMask (stderr, &sm);
  FindGoodSrchMask (&sm);
  WrSrchMaskSrt (stderr, &sm);
  SrchMaskExclHidden (&sm);
  WrSrchMaskSrt (stderr, &sm);

  if (pss->updateMask) {
    Image * new_msk = WrGoodSrchMask (mask_img, &sm, exception);
    if (!new_msk) return MagickFalse;
    ReplaceImageInList (&mask_img, new_msk);
  }

  if (pss->elimClose) {
    SrchMaskExclHidden (&sm);
  }

  if (pss->makeComposite) {
    Image * comp_img = SrchMaskMakeComposite (main_img, sub_img, &sm, exception);
    if (!comp_img) {
      fprintf (stderr, "SrchMaskMakeComposite failed\n");
      return MagickFalse;
    }
    ReplaceImageInList (&mask_img, comp_img);
  }

  DeInitSrchMask (&sm);

  if (sub_img) DeleteImageFromList (&sub_img);
  if (main_img) DeleteImageFromList (&main_img);

  *images = GetFirstImageInList (mask_img);

  return MagickTrue;
}



ModuleExport size_t srchmaskImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  MagickBooleanType
    status;

  SubSrchT
    ss;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  InitSubSrch (&ss);

  ss.precision = GetMagickPrecision();

  status = menu (argc, argv, &ss);
  if (status == MagickFalse)
    return (-1);

  int ListLen = (int)GetImageListLength(*images);
  if (ListLen < 1 || ListLen > 3) {
    fprintf (stderr, "srchconcr: needs 1, 2 or 3 images\n");
    return (-1);
  }

  if (!srchmask (images, &ss, exception)) return (-1);

  DeInitSubSrch (&ss);

  return(MagickImageFilterSignature);
}
