/*
   Reference: http://im.snibgo.com/srt3d.htm

   Last update: 27-May-2017.
     7-September-2017 for v7.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <ctype.h>
#include "vsn_defines.h"

#include "srt3d.h"

#define VERSION "srt3d v1.0  Copyright (c) 2017 Alan Gibson"

static void usage (void)
{
  printf ("Usage: -process 'srt3d [OPTION]...'\n");
  printf ("Scale, rotate and translate in 3D.\n");
  printf ("\n");
  printf ("  t, transform string     transformation string\n");
  printf ("  a, array filename       read array from file\n");
  printf ("  A, out-array filename   write array to file\n");
  printf ("  c, coords filename      read coords from file\n");
  printf ("  C, out-coords filename  write coords to file\n");
  printf ("  n, nOutCoords integer   number of output coordinates (0-3)\n");
  printf ("  f, focal-length number  for perspective\n");
  printf ("  r, hide-reversed        hide reversed polygons\n");
  printf ("  v, verbose              write text information to stdout\n");
  printf ("     version              write version information to stdout\n");
  printf ("\n");
}


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  // LocaleCompare is not case-sensitive,
  // so we use strcmp for the short option.

  if ((strcmp(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "srt3d: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  srt3dT * ps3d
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status = MagickTrue;

  char ** pargv = (char **)argv;

  srt3dInit (ps3d);

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];
    //printf ("Arg %i [%s]\n", i, pa);

    if (IsArg (pa, "t", "transform")==MagickTrue) {
      NEXTARG;
      ps3d->transStr = pargv[i];
    } else if (IsArg (pa, "a", "array")==MagickTrue) {
      NEXTARG;
      ps3d->fInArray = pargv[i];
    } else if (IsArg (pa, "A", "out-array")==MagickTrue) {
      NEXTARG;
      ps3d->fOutArray = pargv[i];
    } else if (IsArg (pa, "c", "coords")==MagickTrue) {
      NEXTARG;
      ps3d->fInCoords = pargv[i];
    } else if (IsArg (pa, "C", "out-coords")==MagickTrue) {
      NEXTARG;
      ps3d->fOutCoords = pargv[i];
    } else if (IsArg (pa, "n", "nOutCoords")==MagickTrue) {
      NEXTARG;
      ps3d->numCoordsOut = atoi(pargv[i]);
    } else if (IsArg (pa, "f", "focal-length")==MagickTrue) {
      NEXTARG;
      ps3d->focalLen = atof (pargv[i]);
    } else if (IsArg (pa, "r", "hide-reversed")==MagickTrue) {
      ps3d->calcHidden = MagickTrue;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      ps3d->verbose = MagickTrue;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "srt3d: ERROR: unknown option [%s]\n", pa);
      status = MagickFalse;
    }
  }

  if (ps3d->verbose) {
    fprintf (stderr, "srt3d options: ");
    if (ps3d->transStr)   fprintf (stderr, " transform %s", ps3d->transStr);
    if (ps3d->fInArray)   fprintf (stderr, " array %s", ps3d->fInArray);
    if (ps3d->fOutArray)  fprintf (stderr, " out-array %s", ps3d->fOutArray);
    if (ps3d->fInCoords)  fprintf (stderr, " coords %s", ps3d->fInCoords);
    if (ps3d->fOutCoords) {
      fprintf (stderr, " out-coords %s", ps3d->fOutCoords);
      fprintf (stderr, " nOutCoords %i", ps3d->numCoordsOut);
    }
    if (ps3d->calcHidden)  fprintf (stderr, " hide-reversed");
    if (ps3d->focalLen!=0) fprintf (stderr, " focal-length %.*g", 
      ps3d->precision, ps3d->focalLen);
    if (ps3d->verbose)    fprintf (stderr, " verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


// The next function corresponds in style to functions in transform.c
// It takes one image, and returns an image.
//
static Image *srt3d (
  Image *image,
  srt3dT * s3d,
  ExceptionInfo *exception)
{
  if (image->debug != MagickFalse)
    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);

  if (s3d->verbose) {
    fprintf (stderr, "srt3d: Input image [%s] %ix%i depth is %i\n",
             image->filename,
             (int)image->columns, (int)image->rows,
             (int)image->depth);
  }

  s3d->imgWidth = image->columns;
  s3d->imgHeight = image->rows;

  if (!srt3dCalc (s3d)) {
    return NULL;
  }

  if (s3d->verbose) {
    fprintf (stderr, "nEntries = %i\n", s3d->nEntries);
  }

  double * distArray = (double *)malloc (4 * s3d->nEntries * sizeof (double));

  int i;
  double *pd = distArray;
  for (i=0; i < s3d->nEntries; i++) {
    *(pd++) = s3d->coordPrimes[i].x;
    *(pd++) = s3d->coordPrimes[i].y;
    *(pd++) = s3d->coordPrimes[i].xPrime;
    *(pd++) = s3d->coordPrimes[i].yPrime;
  }

  Image *out_image = NULL;

  if (!s3d->isHidden) {
    out_image = DistortImage (
      image, PerspectiveDistortion,
      4 * s3d->nEntries, distArray, MagickTrue, exception);
  }

  if (!out_image) {
    ClearMagickException (exception);

    // Create an image with one transparent pixel.
    if (s3d->verbose) {
      fprintf (stderr, "Creating 1x1 transparent.\n");
    }
    out_image = CloneImage (image, 1,1, MagickFalse, exception);
    if (!out_image) {
      fprintf (stderr, "srt3d: DistortImage and CloneImage failed\n");
      return NULL;
    }
#if IMV6OR7==6
    SetImageOpacity (out_image, TransparentOpacity);
#else
    SetImageAlpha (out_image, TransparentAlpha, exception);
#endif
  }

  if (s3d->verbose) {
    fprintf (stderr, "Finished srt3d\n");
  }

  return (out_image);
}


ModuleExport size_t srt3dImage (
  Image **images,
  const int argc,
  const char **argv,
  ExceptionInfo *exception)
{
  Image
    *image,
    *new_image;

  MagickBooleanType
    status;

  srt3dT s3d;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  status = menu (argc, argv, &s3d);
  if (status == MagickFalse)
    return (-1);

  s3d.precision = GetMagickPrecision();
  printf ("s3d.precision=%i\n", s3d.precision);

  for (image=(*images); image != (Image *) NULL; image=GetNextImageInList(image))
  {
    new_image = srt3d (image, &s3d, exception);
    if (!new_image) return (-1);

    ReplaceImageInList(&image,new_image);
    *images=GetFirstImageInList(image);
  }

  srt3dDeInit (&s3d);

  return(MagickImageFilterSignature);
}
