// Last update: 26-May-2017.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#ifndef BOOL
#define BOOL int
#define TRUE 1
#define FALSE 0
#endif

typedef double MatT[4][4];

typedef enum {
  opScale, opRotate, opTranslate
} OperationT;

typedef enum {
  dimX, dimY, dimZ
} DimensionT;

typedef enum {
  qNone, qPercent, qProp
} QualifierT;

typedef struct {
  double x;
  double y;
  double z;
} CoordT;

typedef struct {
  double x;
  double y;
  double z;
  double xPrime;
  double yPrime;
  double zPrime;
} CoordPrimeT;

typedef struct {
  // Inputs.
  ssize_t imgWidth;
  ssize_t imgHeight;
  int numCoordsOut;  // Typically 0, 2 or 3.
  double focalLen;   // For perspective. 0 = infinity.
  char * transStr;
  char * fInArray;   // NULL or filename or "-".
  char * fInCoords;  // NULL or filename.
  char * fOutArray;  // NULL or filename or "-".
  char * fOutCoords; // NULL or filename or "-".
  BOOL calcHidden;
  int precision;
  BOOL verbose;

  // Calculated outputs.
  MatT transMat;
  int nEntries;
  CoordT * coords;
  CoordPrimeT * coordPrimes;
  BOOL isHidden;
}  srt3dT;

#define LINE_LEN 1000


static void srt3dInit (srt3dT * psm)
{
  psm->imgWidth = 0;
  psm->imgHeight = 0;
  psm->numCoordsOut = 3;
  psm->focalLen = 0;
  psm->transStr = NULL;
  psm->fInArray = NULL;
  psm->fInCoords = NULL;
  psm->fOutArray = NULL;
  psm->fOutCoords = NULL;
  psm->calcHidden = FALSE;
  psm->precision = 6;
  psm->verbose = FALSE;
  psm->nEntries = 0;
  psm->coords = NULL;
  psm->coordPrimes = NULL;
  psm->isHidden = FALSE;
}

static void srt3dDeInit (srt3dT * psm)
{
  if (psm->coords) free (psm->coords);
  if (psm->coordPrimes) free (psm->coordPrimes);
}

static void srt3dMatId (MatT m)
{
  int x, y;
  for (y=0; y < 4; y++) {
    for (x=0; x < 4; x++) {
      m[x][y] = (x==y)? 1 : 0;
    }
  }
}

static void srt3dMatCopy (MatT r, MatT a)
// r := a
{
  int x, y;
  for (y=0; y < 4; y++) {
    for (x=0; x < 4; x++) {
      r[x][y] = a[x][y];
    }
  }
}

static void srt3dMatMult3 (MatT r, MatT a, MatT b)
// r := a * b
{
  int x, y, i;
  for (y=0; y < 4; y++) {
    for (x=0; x < 4; x++) {
      double v = 0;

      for (i=0; i < 4; i++) {
        v += a[x][i] * b[i][y];
      }
      r[x][y] = v;
    }
  }
}

static void srt3dMatMult2 (MatT r, MatT a)
// r := r * a
{
  MatT b;
  srt3dMatCopy (b, r);
  srt3dMatMult3 (r, b, a);
}

static void srt3dWrMat (FILE * fh, MatT m, int precision)
{
  printf ("precision=%i\n", precision);

  int x, y;
  for (y=0; y < 4; y++) {
    for (x=0; x < 4; x++) {
      if (x > 0) fprintf (fh, ", ");
      fprintf (fh, "%.*g", precision, m[x][y]);
    }
    fprintf (fh, "\n");
  }
}

static BOOL srt3dRdMat (FILE * fh, MatT m)
{
  int y;
  for (y=0; y < 4; y++) {
    if (fscanf (fh, "%lf,%lf,%lf,%lf", &m[0][y], &m[1][y], &m[2][y], &m[3][y]) != 4) {
      return FALSE;
    }
  }
  return TRUE;
}


static BOOL srt3dSkipSep (char ** p)
// Returns whether separator found.
{
  if (**p != ',' && **p != ' ') return FALSE;

  while (isspace((int) ((unsigned char)**p))) (*p)++;
  if (**p == ',') {
    (*p)++;
    while (isspace((int) ((unsigned char)**p))) (*p)++;
  }

  return TRUE;
}

static BOOL srt3dGetOption (char * str, OperationT * op, DimensionT * dim)
// Returns whether valid.
{
  char c = *str;

  switch (c) {
    case 's':
    case 'S':
      *op = opScale;
      break;
    case 'r':
    case 'R':
      *op = opRotate;
      break;
    case 't':
    case 'T':
      *op = opTranslate;
      break;
    default:
      return FALSE;
  }

  c = str[1];

  switch (c) {
    case 'x':
    case 'X':
      *dim = dimX;
      break;
    case 'y':
    case 'Y':
      *dim = dimY;
      break;
    case 'z':
    case 'Z':
      *dim = dimZ;
      break;
    default:
      return FALSE;
  }

  return TRUE;
}

static BOOL srt3dGetNumber (char ** p, double *pValue, QualifierT * qual)
// Returns whether valid.
{
  int len;
  if (sscanf (*p, "%lf%n", pValue, &len) != 1) return FALSE;
  *p += len;
  char c = **p;
  switch (c) {
    case 'c':
    case 'C':
    case '%':
      *qual = qPercent;
      (*p)++;
      break;
    case 'p':
    case 'P':
      *qual = qProp;
      (*p)++;
      break;
    default:
      *qual = qNone;
  }
  return TRUE;
}

static BOOL srt3fInitArray (srt3dT * psm)
{
  srt3dMatId (psm->transMat);
  if (!psm->fInArray) return TRUE;

  FILE * fh;
  BOOL isStdIn = (strcmp (psm->fInArray, "-") == 0);
  if (isStdIn) fh = stdin;
  else {
    fh = fopen (psm->fInArray, "rt");
    if (!fh) {
      fprintf (stderr, "Can't open %s\n", psm->fInArray);
      return FALSE;
    }
  }

  BOOL okay = srt3dRdMat (fh, psm->transMat);

  if (!isStdIn) fclose (fh);

  return okay;
}

static BOOL srt3dStr2Mat (srt3dT * psm)
/*
    Returns whether str is valid.
    String contains one or more occurences of:
      {option} {number}
    Separator is one or more spaces,
      or one comma with optional spaces on both sides.
    Option is two letters (any case).
    First letter is one of 's', 'r' or 't' (scale, rotate or translate).
    Second letter is one of 'x', 'y' or 'z'.
    Number is floating point.
    For 's', number may be suffixed with '%', 'c'.
    For 't', number may be suffixed with '%', 'c' or 'p'.
*/
{
  if (!srt3fInitArray (psm)) {
    fprintf (stderr, "InitArray failed\n");
    return FALSE;
  }

  MatT mOne;

  char *p = psm->transStr;

  if (!p || !*p) {
    if (psm->verbose) fprintf (stderr, "Warning: No string to process\n");
    return TRUE;
  }

  OperationT op;
  DimensionT dim;

  for (;;) {
    srt3dMatId (mOne);

    if (!srt3dGetOption (p, &op, &dim)) {
      fprintf (stderr, "Bad option at [%s]\n", p);
      return FALSE;
    }
    p += 2;
    srt3dSkipSep (&p);
    double Value;
    QualifierT qual;
    if (!srt3dGetNumber (&p, &Value, &qual)) {
      fprintf (stderr, "Bad number at [%s]\n", p);
      return FALSE;
    }

    switch (op) {
      case opScale: {
        // FIXME: also use qual.
        switch (dim) {
          case dimX: mOne[0][0] = Value; break;
          case dimY: mOne[1][1] = Value; break;
          case dimZ: mOne[2][2] = Value; break;
        }
        break;
      }
      case opRotate: {
        double rad = Value * M_PI / 180.0;
        double sv = sin (rad);
        double cv = cos (rad);
 
#define EPS 1e-15

       if (fabs(sv) < EPS) sv = 0;
       if (fabs(cv) < EPS) cv = 0;

       switch (dim) {
          case dimX:
            mOne[1][1] = cv;
            mOne[2][1] = -sv;
            mOne[1][2] = sv;
            mOne[2][2] = cv;
            break;
          case dimY:
            mOne[2][2] = cv;
            mOne[0][2] = -sv;
            mOne[2][0] = sv;
            mOne[0][0] = cv;
            break;
          case dimZ:
            mOne[0][0] = cv;
            mOne[1][0] = -sv;
            mOne[0][1] = sv;
            mOne[1][1] = cv;
            break;
        }
        break;
      }
      case opTranslate: {
        // Adjust for qual.
        if (qual==qPercent) {
          if (dim==dimX) Value *= (psm->imgWidth-1) / 100.0;
          else if (dim==dimY) Value *= (psm->imgHeight-1) / 100.0;
        } else if (qual==qProp) {
          if (dim==dimX) Value *= (psm->imgWidth-1);
          else if (dim==dimY) Value *= (psm->imgHeight-1);
        }

        switch (dim) {
          case dimX: mOne[3][0] = Value; break;
          case dimY: mOne[3][1] = Value; break;
          case dimZ: mOne[3][2] = Value; break;
        }
        break;
      }
    }

    srt3dMatMult2 (psm->transMat, mOne);

    if (!*p) break;

    if (!srt3dSkipSep (&p)) {
      fprintf (stderr, "No separator at [%s]\n", p);
      return FALSE;
    }
  }

  return TRUE;
}

static void srt3dApplyMatArr (
  CoordT * arrIn, CoordPrimeT * arrOut, int nEntries, MatT m)
// Applies matrix to array of x,y,z coordinates, making array of x,y,z,x',y',z'.
{
  int i;
  for (i = 0; i < nEntries; i++) {
    CoordT *c = &arrIn[i];
    CoordPrimeT *p = &arrOut[i];
    p->x = c->x;
    p->y = c->y;
    p->z = c->z;
    p->xPrime = p->x*m[0][0] + p->y*m[1][0] + p->z*m[2][0] + m[3][0];
    p->yPrime = p->x*m[0][1] + p->y*m[1][1] + p->z*m[2][1] + m[3][1];
    p->zPrime = p->x*m[0][2] + p->y*m[1][2] + p->z*m[2][2] + m[3][2];
  }
}

static void srt3dApplyPerspective (
  CoordPrimeT * arrOut, int nEntries, double focalLen)
{
#define SMALL_DIST 1e-3

  int i;
  for (i = 0; i < nEntries; i++) {
    CoordPrimeT *p = &arrOut[i];
    double div = focalLen - p->zPrime;
    if (fabs(div) < SMALL_DIST) {
      if (div >= 0) div = SMALL_DIST;
      else div = -SMALL_DIST;
    }
    double mult = focalLen / div;
    p->xPrime *= mult;
    p->yPrime *= mult;
    p->zPrime = 0;
  }
}

static BOOL srt3dReadCoords (srt3dT * psm)
{
  // If we don't have a file, try to use the corners,
  // clockwise, starting from top-left.
  //
  if (!psm->fInCoords || !*psm->fInCoords) {
    if (!psm->imgWidth || !psm->imgHeight) {
      fprintf (stderr, "No coords file; no width or height\n");
      return FALSE;
    }
    psm->nEntries = 4;

    psm->coords = (CoordT *)malloc (psm->nEntries * sizeof(CoordT));
    if (psm->coords==NULL) return 1;

    int i;
    for (i=0; i < 4; i++) {
      CoordT *c = &psm->coords[i];
      c->x = c->y = c->z = 0;
    }
    psm->coords[1].x = psm->imgWidth-1;
    psm->coords[2].x = psm->imgWidth-1;
    psm->coords[2].y = psm->imgHeight-1;
    psm->coords[3].y = psm->imgHeight-1;

    return TRUE;
  }

  FILE * fh = fopen (psm->fInCoords, "rt");
  if (!fh) {
    fprintf (stderr, "Can't open %s\n", psm->fInCoords);
    return FALSE;
  }

  // Count entries. Allocate arrays.
  if (fseek (fh, 0, SEEK_SET) != 0) return FALSE;

  psm->nEntries = 0;
  char sLine [LINE_LEN];

  while (fgets (sLine, LINE_LEN, fh) != NULL) psm->nEntries++;

  if (psm->coords) {
    fprintf (stderr, "coords already alloced\n");
    return FALSE;
  }

  psm->coords = (CoordT *)malloc (psm->nEntries * sizeof(CoordT));
  if (psm->coords==NULL) return 1;

  // Populate coords.
  if (fseek (fh, 0, SEEK_SET) != 0) return FALSE;
  int i = 0;
  int maxX=0, maxY=0;
  while (fgets (sLine, LINE_LEN, fh) != NULL) {
    //printf ("%i: %s", i, sLine);
    CoordT *c = &psm->coords[i];

    int len;
    c->x = c->y = c->z = 0;
    sscanf (sLine, "%lf,%lf,%lf%n", &c->x, &c->y, &c->z, &len);
    //printf ("nArgs=%i len=%i %g %g %g\n", nArgs, len, c->x, c->y, c->z);
    if (maxX < c->x) maxX = c->x;
    if (maxY < c->y) maxY = c->y;
    i++;
  }

  if (psm->imgWidth==0) psm->imgWidth = maxX+1;
  if (psm->imgHeight==0) psm->imgHeight = maxY+1;

  fclose (fh);

  return TRUE;
}


static void srt3dWrCoordPrimes (srt3dT * psm, FILE * fhout)
{
  int i;
  for (i=0; i < psm->nEntries; i++) {
    CoordPrimeT *p = &psm->coordPrimes[i];
    switch (psm->numCoordsOut) {
      case 1:
        fprintf (fhout, "%.*g, ", psm->precision, p->x);
        fprintf (fhout, "%.*g\n", psm->precision, p->xPrime);
        break;
      case 2:
        fprintf (fhout, "%.*g,%.*g, ", psm->precision, p->x, psm->precision, p->y);
        fprintf (fhout, "%.*g,%.*g\n", psm->precision, p->xPrime, psm->precision, p->yPrime);
        break;
      default:
        fprintf (fhout, "%.*g,%.*g,%.*g, ", psm->precision, p->x, psm->precision, p->y, psm->precision, p->z);
        fprintf (fhout, "%.*g,%.*g,%.*g\n", psm->precision, p->xPrime, psm->precision, p->yPrime, psm->precision, p->zPrime);
        break;
    }
  }
}


static BOOL srt3dApplyMatFile (srt3dT * psm)
// Applies matrix to file of x,y,z coordinates, making file of x,y,z,x',y',z'.
{
  if (psm->coordPrimes) {
    fprintf (stderr, "coordPrimes already alloced\n");
    return FALSE;
  }

  psm->coordPrimes = (CoordPrimeT *)malloc (psm->nEntries * sizeof(CoordPrimeT));
  if (psm->coordPrimes==NULL) return 1; 

  srt3dApplyMatArr (psm->coords, psm->coordPrimes, psm->nEntries, psm->transMat);

  if (psm->focalLen != 0) {
    srt3dApplyPerspective (psm->coordPrimes, psm->nEntries, psm->focalLen);
  }

  if (psm->fOutArray) {
    if (*psm->fOutArray) {
      FILE * fh;
      BOOL isStdOut = (strcmp (psm->fOutArray, "-") == 0);
      if (isStdOut) fh = stdout;
      else {
        fh = fopen (psm->fOutArray, "wt");
        if (!fh) {
          fprintf (stderr, "Can't open %s\n", psm->fOutArray);
          return FALSE;
        }
      }
      srt3dWrMat (fh, psm->transMat, psm->precision);
      if (!isStdOut) fclose (fh);
    }
  }

  if (psm->fOutCoords && psm->numCoordsOut > 0) {
    if (*psm->fOutCoords) {
      FILE * fh;
      BOOL isStdOut = (strcmp (psm->fOutCoords, "-") == 0);
      if (isStdOut) fh = stdout;
      else {
        fh = fopen (psm->fOutCoords, "wt");
        if (!fh) {
          fprintf (stderr, "Can't open %s\n", psm->fOutCoords);
          return FALSE;
        }
      }
      srt3dWrCoordPrimes (psm, fh);
      if (!isStdOut) fclose (fh);
    }
  }

  return TRUE;
}

static void srt3dCalcHidden (srt3dT * psm)
{
  int i, j;
  double area = 0;
  double areaPrime = 0;

  j = psm->nEntries - 1;
  for (i=0; i < psm->nEntries; i++) {
    CoordPrimeT *pi = &psm->coordPrimes[i];
    CoordPrimeT *pj = &psm->coordPrimes[j];

    area      += (pi->x      + pj->x     ) * (pi->y      - pj->y     );
    areaPrime += (pi->xPrime + pj->xPrime) * (pi->yPrime - pj->yPrime);

    j = i;
  }

  // If we wanted the true areas, we would divide them by 2.0.

#define SMALL_AREA 1e-9

  if (psm->verbose) {
    fprintf (stderr, "area=%.*g areaPrime=%.*g\n",
                     psm->precision, area,
                     psm->precision, areaPrime);
  }

  psm->isHidden = (fabs(areaPrime) < SMALL_AREA) || (area * areaPrime < 0);

  if (psm->verbose && psm->isHidden) {
    fprintf (stderr, "isHidden\n");
  }
}

static BOOL srt3dCalc (srt3dT * psm)
{
  if (!srt3dReadCoords (psm)) {
    fprintf (stderr, "ReadCoords failed\n");
    return FALSE;
  }

  if (!srt3dStr2Mat (psm)) {
    fprintf (stderr, "Str2Mat failed\n");
    return FALSE;
  }

  if (!srt3dApplyMatFile (psm)) {
    fprintf (stderr, "ApplyMatFile failed\n");
    return FALSE;
  }

  if (psm->calcHidden) srt3dCalcHidden (psm);
  else psm->isHidden = FALSE;

  return TRUE;
}

