/* Updated:
     1-September-2017 ParseXy now allows ',' or 'x' as separator.
*/

//---------------------------------------------------------------------------
// User-supplied coordinates: data structures and functions.

typedef enum {
  csPix,
  csPc,
  csProp
} CoordSpecT;

typedef struct {
  double
    Num;

  CoordSpecT
    cs;

  ssize_t
    Pix;
} UserCoordT;

typedef struct {
  UserCoordT
    x,
    y;
} UserCoordXyT;


static int ParseXy (const char * s, UserCoordXyT * uc)
// Returns 1 if okay; 0 if failure.
{
  double d;
  int n;
  char * p = (char *)s;

  sscanf (p, "%lg%n", &d, &n);

  uc->x.Num = d;
  uc->x.Pix = (int)(d+0.5);

  p += n;
  switch (*p) {
    case '%':
    case 'c':
      uc->x.cs = csPc;
      p++;
      break;
    case 'p':
      uc->x.cs = csProp;
      p++;
      break;
  }

  if (*p != ',' && *p != 'x') return 0;
  p++;

  sscanf (p, "%lg%n", &d, &n);

  uc->y.Num = d;
  uc->y.Pix = (int)(d+0.5);

  p += n;
  switch (*p) {
    case '%':
    case 'c':
      uc->y.cs = csPc;
      p++;
      break;
    case 'p':
      uc->y.cs = csProp;
      p++;
      break;
  }

  if (*p != '\0') return 0;

  return 1;
}

static int ResolveUserCoords (
  UserCoordXyT * uc,
  ssize_t width,
  ssize_t height,
  double maxFact  // Eg 1.0 or 2.0.
)
// Returns 1 if okay; 0 if failure.
{
  switch (uc->x.cs) {
    case csPc:
      uc->x.Pix = (int)(uc->x.Num * (width-1) / 100.0 + 0.5);
      break;
    case csProp:
      uc->x.Pix = (int)(uc->x.Num * (width-1) + 0.5);
      break;
    default:
      uc->x.Pix = (int)(uc->x.Num + 0.5);
  }
  switch (uc->y.cs) {
    case csPc:
      uc->y.Pix = (int)(uc->y.Num * (height-1) / 100.0 + 0.5);
      break;
    case csProp:
      uc->y.Pix = (int)(uc->y.Num * (height-1) + 0.5);
      break;
    default:
      uc->y.Pix = (int)(uc->y.Num + 0.5);
  }

  int status = 1;
  if (uc->x.Pix < 0 || uc->y.Pix < 0) {
    fprintf (stderr, "Coordinates must be positive.");
    status = 0;
  }

  if (uc->x.Pix >= width * maxFact || uc->y.Pix >= height * maxFact)
  {
    fprintf (stderr,
      "Coordinates (%li,%li) must be less than width and height (%li,%li)",
      uc->x.Pix, uc->y.Pix, width, height);

    if (maxFact != 1.0) {
      fprintf (stderr, " multipled by %g", maxFact);
    }
    fprintf (stderr, ".\n");
    status = 0;
  }

  return status;
}

static int ResolveUserDims (
  UserCoordXyT * uc,
  ssize_t width,
  ssize_t height,
  double maxFact  // Eg 1.0 or 2.0. If 0.0, no check.
)
// Returns 1 if okay; 0 if failure.
{
  switch (uc->x.cs) {
    case csPc:
      uc->x.Pix = (int)(uc->x.Num * (width) / 100.0 + 0.5);
      break;
    case csProp:
      uc->x.Pix = (int)(uc->x.Num * (width) + 0.5);
      break;
    default:
      uc->x.Pix = (int)(uc->x.Num + 0.5);
  }
  switch (uc->y.cs) {
    case csPc:
      uc->y.Pix = (int)(uc->y.Num * (height) / 100.0 + 0.5);
      break;
    case csProp:
      uc->y.Pix = (int)(uc->y.Num * (height) + 0.5);
      break;
    default:
      uc->y.Pix = (int)(uc->y.Num + 0.5);
  }

  int status = 1;
  if (uc->x.Pix < 0 || uc->y.Pix < 0) {
    fprintf (stderr, "Dimensions must be positive.");
    status = 0;
  }

  if (maxFact > 0) {
    if (uc->x.Pix > width * maxFact || uc->y.Pix > height * maxFact)
    {
      fprintf (stderr,
        "Dimensions (%li,%li) must be less than width and height (%li,%li)",
        uc->x.Pix, uc->y.Pix, width, height);

      if (maxFact != 1.0) {
        fprintf (stderr, " multipled by %g", maxFact);
      }
      fprintf (stderr, ".\n");
      status = 0;
    }
  }

  return status;
}

static void WrUserCoord (UserCoordXyT * uc)
{
  switch (uc->x.cs) {
    case csPc:
      fprintf (stderr, "%gc", uc->x.Num);
      break;
    case csProp:
      fprintf (stderr, "%gp", uc->x.Num);
      break;
    default:
      fprintf (stderr, "%li", uc->x.Pix);
  }
  fprintf (stderr, ",");
  switch (uc->y.cs) {
    case csPc:
      fprintf (stderr, "%gc", uc->y.Num);
      break;
    case csProp:
      fprintf (stderr, "%gp", uc->y.Num);
      break;
    default:
      fprintf (stderr, "%li", uc->y.Pix);
  }
}

