/* Updated:
     3-April-2018 for v7.0.7-28
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"
#include "rmsealpha.inc"

#include "absdispmap.inc"
#include "whatrot.inc"
#include "whatscale.inc"
#include "whatrotscale.inc"

#define VERSION "whatrotscale v1.0  Copyright (c) 2017 Alan Gibson"

typedef struct {
  WhatRotScaleT
    wrs;

  int
    verbose;

  Image
    *DispMap;

  MagickBooleanType
    doRotate,
    UseDispMap,
    approxOnly,
    debug;

  int
    precision;

  double
    tolerance;

  FILE *
    fh_data;

} SubSrchT;


static void InitSubSrch (SubSrchT * pss)
{
  InitWhatRotScale (&pss->wrs);
  pss->verbose = 0;
  pss->DispMap = NULL;
  pss->doRotate = MagickTrue;
  pss->UseDispMap = MagickTrue;
  pss->approxOnly = MagickFalse;
  pss->precision = 6;
  pss->tolerance = 0.01;
  pss->fh_data = stderr;
  pss->debug = MagickFalse;
}

static void DeInitSubSrch (SubSrchT * pss)
{
  if (pss->DispMap) pss->DispMap = DestroyImage (pss->DispMap);
  DeInitWhatRotScale (&pss->wrs);
}


static void usage (void)
{
  printf ("Usage: -process 'whatrot [OPTION]...'\n");
  printf ("Find scale and rotation by unrolling and golden section methods.\n");
  printf ("\n");
  printf ("  a,  approxOnly       use only the first (approximate) method\n");
  printf ("  t,  tolerance number for golden section method\n");
  printf ("  x,  noChange         don't replace images\n");
  printf ("  f,  file string      write data to file stream stdout or stderr\n");
  printf ("  d,  debug            create debugging images\n");
  printf ("  v,  verbose          write text information to stdout\n");
  printf ("  v2, verbose2         write more information to stdout\n");
  printf ("  v3, verbose3         yet more information to stdout\n");
  printf ("      version          write version information to stdout\n");
  printf ("\n");
}

// FIXME: also option to normalise mean and SD.


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "paintpatches: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  SubSrchT * pss
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "a", "approxOnly")==MagickTrue) {
      pss->approxOnly = MagickTrue;
    } else if (IsArg (pa, "x", "noChange")==MagickTrue) {
      pss->doRotate = MagickFalse;
    } else if (IsArg (pa, "t", "tolerance")==MagickTrue) {
      NEXTARG;
      pss->tolerance = atof (argv[i]);
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      NEXTARG;
      if (strcasecmp (argv[i], "stdout")==0) pss->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) pss->fh_data = stderr;
      else status = MagickFalse;
    } else if (IsArg (pa, "d", "debug")==MagickTrue) {
      pss->debug = MagickTrue;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pss->verbose = 1;
    } else if (IsArg (pa, "v2", "verbose2")==MagickTrue) {
      pss->verbose = 2;
    } else if (IsArg (pa, "v3", "verbose3")==MagickTrue) {
      pss->verbose = 3;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "whatrotscale: ERROR: unknown option\n");
      status = MagickFalse;
    }
  }

  if (pss->verbose) {
    fprintf (stderr, "whatrotscale options:");
    if (pss->approxOnly) fprintf (stderr, "  approxOnly");
    fprintf (stderr, "  tolerance %g", pss->tolerance);
    if (!pss->doRotate) fprintf (stderr, "  noChange");

    if (pss->fh_data == stdout) fprintf (stderr, " file stdout");
    if (pss->debug) fprintf (stderr, "  debug");
    if (pss->verbose > 1) fprintf (stderr, "  verbose%i", pss->verbose);
    else if (pss->verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


#if 0
// The next function takes an image list,
// may modify it, and returns a status.
//
static MagickBooleanType doWhatrotscale (
  Image **images,
  SubSrchT * pss,
  ExceptionInfo *exception)
{
  Image * main_img = *images;

  Image * sub_img = GetNextImageInList (main_img);
  if (!sub_img) return MagickFalse;

/*==
  if (sub_img->columns > main_img->columns ||
      sub_img->rows    > main_img->rows)
  {
    fprintf (stderr, "whatrotscale: subimage is larger than main\n");
    return MagickFalse;
  }
==*/

//#if IMV6OR7==6
//  SetPixelCacheVirtualMethod (main_img, TransparentVirtualPixelMethod);
//#else
//  SetPixelCacheVirtualMethod (main_img,
//    TransparentVirtualPixelMethod, exception);
//#endif

  VIRT_NONE(main_ing, exception);

  // Make displacement map, and unroll it.
  double minusOne=-1.0;
  MagickBooleanType bestfit = MagickTrue;

  Image * id_map = mIdentAbsDispMap (sub_img, exception);
#if IMV6OR7==6
  SetImageAlphaChannel (id_map, OpaqueAlphaChannel);
#else
  SetImageAlphaChannel (id_map, OpaqueAlphaChannel, exception);
#endif

  pss->DispMap = DistortImage (id_map, DePolarDistortion,
    1, &minusOne, bestfit, exception);
  if (!pss->DispMap) return MagickFalse;

  if (pss->verbose) fprintf (stderr, "srchconcr: DispMap %lix%li\n",
    pss->DispMap->columns, pss->DispMap->rows);

  pss->wrs.wr.verbose = pss->verbose;
  pss->wrs.ws.verbose = pss->verbose;
  if (pss->debug) {
    pss->wrs.debug = "wrs_dbg";
    pss->wr.debug = "wrs_dbg";
    pss->ws.debug = "wrs_dbg";
  }

  if (!CalcApproxRot (&pss->wrs.wr, main_img, sub_img, NULL, NULL, exception)) {
    fprintf (stderr, "CalcApproxRot failed\n");
    return MagickFalse;
  }

  if (pss->verbose) {
    fprintf (stderr, "Approx  ang=%.*g +/-%.*g score=%.*g\n",
      pss->precision, pss->wrs.wr.angle,
      pss->precision, pss->wrs.wr.plusOrMinus,
      pss->precision, pss->wrs.wr.score);
  }

  if (!pss->approxOnly) {
    // Save, in case GoldSect fails.
    double saveAng = pss->wrs.wr.angle;
    double savePorm = pss->wrs.wr.plusOrMinus;
    double saveScore = pss->wrs.wr.score;

    pss->wr.tolerance = pss->tolerance;

    if (GoldSectRot (&pss->wr, main_img, sub_img, exception)) {
      if (pss->verbose) fprintf (stderr,
        "From gss: ang=%.*g +/-%.*g score=%.*g\n",
        pss->precision, pss->wr.angle,
        pss->precision, pss->wr.plusOrMinus,
        pss->precision, pss->wr.score);
    } else {
      fprintf (stderr, "whatrot: GoldSectScale failed\n");
      pss->wr.angle = saveAng;
      pss->wr.plusOrMinus = savePorm;
      pss->wr.score = saveScore;
    }
  }

  fprintf (pss->fh_data, "whatrotscale:  ang=%.*g angPorm=%.*g angScore=%.*g\n",
    pss->precision, pss->wr.angle,
    pss->precision, pss->wr.plusOrMinus,
    pss->precision, pss->wr.score);

  if (pss->doRotate) {
    MagickBooleanType bestfit = MagickTrue;

    double scaleArray[2];
    scaleArray[0] = 1.0;            // Scale factor.
    scaleArray[1] = pss->wr.angle;  // Rotation angle.

//#if IMV6OR7==6
//    SetPixelCacheVirtualMethod (sub_img, TransparentVirtualPixelMethod);
//#else
//    SetPixelCacheVirtualMethod (sub_img,
//      TransparentVirtualPixelMethod, exception);
//#endif

    VIRT_NONE(sub_img, exception);

    Image *rot_img = DistortImage (sub_img, ScaleRotateTranslateDistortion,
      2, scaleArray, bestfit, exception);
    if (!rot_img) return MagickFalse;
    // We don't ResetPage (rot_img);

    DeleteImageFromList (&sub_img);

    ReplaceImageInList (&main_img, rot_img);
    // Replace messes up the images pointer. Make it good:
    *images = GetFirstImageInList (main_img);
  }

  return MagickTrue;
}
#endif


static MagickBooleanType doWhatrotscale (
  Image **images,
  SubSrchT * pss,
  ExceptionInfo *exception)
{
  int ListLen = (int)GetImageListLength(*images);
  if (ListLen != 2) {
    fprintf (stderr, "whatrotscale: needs 2 images\n");
    return (-1);
  }

  Image * main_img = *images;
  Image * sub_img = GetNextImageInList (main_img);

  pss->wrs.verbose = pss->verbose;
  if (pss->debug) {
    pss->wrs.debug = "wrs_dbg";
  }

  if (!ApproxWhatRotScale (&pss->wrs, main_img, sub_img, exception))
    return MagickFalse;

  if (!RefineWhatRotScale (&pss->wrs, main_img, sub_img, exception))
    return MagickFalse;

  fprintf (pss->fh_data, "WhatRotScale scale=%.*g angle=%.*g score=%.*g\n",
    pss->precision, pss->wrs.scale,
    pss->precision, pss->wrs.angle,
    pss->precision, pss->wrs.score);

  if (pss->doRotate) {
    MagickBooleanType bestfit = MagickTrue;
    double scaleArray[2];
    scaleArray[0] = pss->wrs.scale;  // Scale factor.
    scaleArray[1] = pss->wrs.angle;  // Rotation angle.

    VIRT_NONE(sub_img, exception);

    Image *dist_img = DistortImage (sub_img, ScaleRotateTranslateDistortion,
      2, scaleArray, bestfit, exception);
    if (!dist_img) return MagickFalse;
    // We don't ResetPage (dist_img);

    if (pss->verbose) fprintf (stderr,
      "WhatRotScale: created image %lix%li %lix%li%+li%+li\n",
      dist_img->columns, dist_img->rows,
      dist_img->page.width, dist_img->page.height,
      dist_img->page.x, dist_img->page.y);

    DeleteImageFromList (&sub_img);

    ReplaceImageInList (&main_img, dist_img);
    // Replace messes up the images pointer. Make it good:
    *images = GetFirstImageInList (main_img);
  }

  return MagickTrue;
}

ModuleExport size_t whatrotscaleImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  MagickBooleanType
    status;

  SubSrchT
    ss;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  InitSubSrch (&ss);

  ss.precision = GetMagickPrecision();

  status = menu (argc, argv, &ss);
  if (status == MagickFalse)
    return (-1);

  if (!doWhatrotscale (images, &ss, exception))
    fprintf (stderr, "doWhatrotscale failed\n");

  DeInitSubSrch (&ss);

  return(MagickImageFilterSignature);
}
