#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vsn_defines.h"
#include "rmsealpha.inc"
//#include "chklist.h"

#include "avgconcrings.inc"
#include "writeframe.inc"
#include "absdispmap.inc"
#include "whatscale.inc"

#define VERSION "whatrot v1.0  Copyright (c) 2017 Alan Gibson"

typedef struct {
  ConcRingsT
    cr;

  Image
    *AvgConcRings,
    *DispMap;

  MagickBooleanType
    doScale,
    UseDispMap,
    JustSinks,
    FindRotation,
    approxOnly,
    refineOnly,
    phApprox,
    phRefine,
    debug;

  int
    verbose,
    precision,
    scaleRadMax;

  double
    inPorm,
    tolerance;

  FILE *
    fh_data;

} SubSrchT;

static void InitSubSrch (SubSrchT * pss)
{
  InitConcRings (&pss->cr);
  pss->AvgConcRings = NULL;
  pss->DispMap = NULL;
  pss->doScale = MagickTrue;
  pss->UseDispMap = MagickTrue;
  pss->JustSinks = MagickTrue;
  pss->FindRotation = MagickFalse;
  pss->approxOnly = pss->refineOnly = MagickFalse;
  pss->phApprox = pss->phRefine = MagickTrue;
  pss->debug = MagickFalse;
  pss->verbose = 0;
  pss->precision = 6;
  pss->scaleRadMax = 0;
  pss->inPorm = 1.3;
//  pss->porm = 0.3;
  pss->tolerance = 0.01;
  pss->fh_data = stderr;
}

static void DeInitSubSrch (SubSrchT * pss)
{
  if (pss->AvgConcRings) pss->AvgConcRings = DestroyImage (pss->AvgConcRings);
  if (pss->DispMap) pss->DispMap = DestroyImage (pss->DispMap);
  DeInitConcRings (&pss->cr);
}


static void usage (void)
{
  printf ("Usage: -process 'whatscale [OPTION]...'\n");
  printf ("Find scale by unrolling and golden section methods.\n");
  printf ("\n");
  printf ("  a,  approxOnly       use only the first (approximate) method\n");
  printf ("  r,  refineOnly       use only the second (refining) method\n");
  printf ("  x,  noScale          don't replace images with rescaled\n");
  printf ("  s0, minScale number  minimum scale\n");
  printf ("  s1, maxScale number  maximum scale\n");
  printf ("  ns, nScales integer  number of scales\n");
  printf ("  p,  porm number      plus or minus for scale gss\n");
  printf ("  t,  tolerance number error tolerance for scale gss\n");
  printf ("  r,  scaleRadMax int  scale so radius is this, at max\n");
  printf ("  m,  method string    for concentric rings, 'slow' or 'quick'\n");
  printf ("  f,  file string      write data to file stream stdout or stderr\n");
  printf ("  d,  debug            create debugging images\n");
  printf ("  v,  verbose          write text information to stdout\n");
  printf ("  v2, verbose2         write more information to stdout\n");
  printf ("      version          write version information to stdout\n");
  printf ("\n");
}

// FIXME: also option to not update image list.
// FIXME: also option to normalise mean and SD.


static MagickBooleanType IsArg (char * pa, char * ShtOpt, char * LongOpt)
{
  if ((LocaleCompare(pa, ShtOpt)==0) || (LocaleCompare(pa, LongOpt)==0))
    return MagickTrue;

  return MagickFalse;
}


#define NEXTARG \
  if (i < argc-1) i++; \
  else { \
    fprintf (stderr, "paintpatches: ERROR: [%s] needs argument\n", pa); \
    status = MagickFalse; \
  }

static MagickBooleanType menu(
  const int argc,
  const char **argv,
  SubSrchT * pss
)
// Returns MagickTrue if okay.
{
  int
    i;

  MagickBooleanType
    status;

  status = MagickTrue;

  for (i=0; i < argc; i++) {
    char * pa = (char *)argv[i];

    if (IsArg (pa, "a", "approxOnly")==MagickTrue) {
      pss->approxOnly = MagickTrue;
      pss->phRefine = MagickFalse;
    } else if (IsArg (pa, "r", "refineOnly")==MagickTrue) {
      pss->refineOnly = MagickTrue;
      pss->phApprox = MagickFalse;
    } else if (IsArg (pa, "x", "noScale")==MagickTrue) {
      pss->doScale = MagickFalse;
    } else if (IsArg (pa, "s0", "minScale")==MagickTrue) {
      NEXTARG;
      pss->cr.minScale = atof(argv[i]);
    } else if (IsArg (pa, "s1", "maxScale")==MagickTrue) {
      NEXTARG;
      pss->cr.maxScale = atof(argv[i]);
    } else if (IsArg (pa, "ns", "nScales")==MagickTrue) {
      NEXTARG;
      pss->cr.nScales = atoi(argv[i]);
    } else if (IsArg (pa, "p", "porm")==MagickTrue) {
      NEXTARG;
      pss->inPorm = atof(argv[i]);
    } else if (IsArg (pa, "t", "tolerance")==MagickTrue) {
      NEXTARG;
      pss->tolerance = atof(argv[i]);
    } else if (IsArg (pa, "r", "scaleRadMax")==MagickTrue) {
      NEXTARG;
      pss->scaleRadMax = atoi(argv[i]);
    } else if (IsArg (pa, "m", "method")==MagickTrue) {
      NEXTARG;
      if (LocaleCompare(argv[i], "slow")==0) pss->cr.ConcRingsMethod = crmSlow;
      else if (LocaleCompare(argv[i], "quick")==0) pss->cr.ConcRingsMethod = crmQuick;
      else {
        fprintf (stderr, "whatscale: ERROR: unknown method [%s]\n", argv[i]);
        status = MagickFalse;
      }
    } else if (IsArg (pa, "d", "debug")==MagickTrue) {
      pss->debug = MagickTrue;
    } else if (IsArg (pa, "f", "file")==MagickTrue) {
      NEXTARG;
      if (strcasecmp (argv[i], "stdout")==0) pss->fh_data = stdout;
      else if (strcasecmp (argv[i], "stderr")==0) pss->fh_data = stderr;
      else status = MagickFalse;
    } else if (IsArg (pa, "v", "verbose")==MagickTrue) {
      pss->verbose = 1;
    } else if (IsArg (pa, "v2", "verbose2")==MagickTrue) {
      pss->verbose = 2;
    } else if (IsArg (pa, "version", "version")==MagickTrue) {
      fprintf (stdout, "%s\n", VERSION);
    } else {
      fprintf (stderr, "whatscale: ERROR: unknown option\n");
      status = MagickFalse;
    }
  }

  if (pss->verbose) {
    fprintf (stderr, "whatscale options:");
    if (pss->approxOnly) fprintf (stderr, "  approxOnly");
    if (pss->refineOnly) fprintf (stderr, "  refineOnly");
    fprintf (stderr, "  minScale %g", pss->cr.minScale);
    fprintf (stderr, "  maxScale %g", pss->cr.maxScale);
    fprintf (stderr, "  nScales %i", pss->cr.nScales);
    fprintf (stderr, "  porm %g", pss->inPorm);
    fprintf (stderr, "  tolerance %g", pss->tolerance);
    fprintf (stderr, "  method ");
    switch (pss->cr.ConcRingsMethod) {
      case crmSlow:  fprintf (stderr, "slow"); break;
      case crmQuick: fprintf (stderr, "quick"); break;
    }
    if (!pss->doScale) fprintf (stderr, "  noScale");
    if (pss->debug) fprintf (stderr, "  debug");
    if (pss->fh_data == stdout) fprintf (stderr, " file stdout");
    if (pss->verbose==2) fprintf (stderr, "  verbose2");
    else if (pss->verbose) fprintf (stderr, "  verbose");
    fprintf (stderr, "\n");
  }

  if (status == MagickFalse)
    usage ();

  return (status);
}


// The next function takes an image list,
// may modify it, and returns a status.
//
static MagickBooleanType whatscale (
  Image **images,
  SubSrchT * pss,
  ExceptionInfo *exception)
// Takes two images: can be same size, or either can be larger.
{
  Image * main_img = *images;

  Image * sub_img = GetNextImageInList (main_img);
  if (!sub_img) return MagickFalse;

  pss->cr.verbose = (pss->verbose > 1);


/*--
  // Make the concentric rings version of the subimage.
  pss->AvgConcRings = avgconcrings (sub_img, &pss->cr, exception);
  if (!pss->AvgConcRings) {
    fprintf (stderr, "whatscale: avgconcrings failed\n");
    return MagickFalse;
  }

  // Make displacement map, and unroll it.
  double minusOne=-1.0;
  MagickBooleanType bestfit = MagickTrue;

  Image * id_map = mIdentAbsDispMap (sub_img, exception);
#if IMV6OR7==6
  SetImageAlphaChannel (id_map, OpaqueAlphaChannel);
#else
  SetImageAlphaChannel (id_map, OpaqueAlphaChannel, exception);
#endif

  pss->DispMap = DistortImage (id_map, DePolarDistortion,
    1, &minusOne, bestfit, exception);
  if (!pss->DispMap) return MagickFalse;

  if (pss->verbose) fprintf (stderr, "whatscale: DispMap %lix%li\n",
    pss->DispMap->columns, pss->DispMap->rows);

  //if (!WriteFrameImage (pss->DispMap, "dispmap.png", NULL, exception))
  //  return MagickFalse;

  // Check: width of AvgConcRings should be height of DispMap.
  if (pss->AvgConcRings->columns != pss->DispMap->rows) {
    fprintf (stderr, "pss->AvgConcRings->columns != pss->DispMap->rows\n");
  }

  // Possibly shrink pss->AvgConcRings to X columns,
  // and pss->DispMap in same proportion.

  if (pss->scaleRadMax && pss->AvgConcRings->columns > pss->scaleRadMax) {

    Image * tmp_img1 = RESIZEIMG (pss->AvgConcRings, pss->scaleRadMax, pss->AvgConcRings->rows,
      exception);
    Image * tmp_img2 = RESIZEIMG (pss->DispMap, pss->DispMap->rows, pss->scaleRadMax,
      exception);

    if (!tmp_img1) return MagickFalse;
    ReplaceImageInList (&pss->AvgConcRings, tmp_img1);

    if (!tmp_img2) return MagickFalse;
    ReplaceImageInList (&pss->DispMap, tmp_img2);

  }

  pss->cr.ra.do_verbose = pss->cr.verbose;

--*/

  WhatScaleT ws;
  InitWhatScale (&ws);
  ws.scale = 1.0;
  ws.inPorm = pss->inPorm;
  ws.tolerance = pss->tolerance;
  ws.verbose = pss->verbose;
  ws.score = 1.0;
  if (pss->debug) ws.debug = "ws_dbg";

  // FIXME: next shld be a parameter, so user can prevent the loop.
#define MAX_ITER 10
  int i;
  for (i = 0; i < MAX_ITER; i++) {

    if (! PrepCalcApproxScale (
      &pss->cr, sub_img, &pss->AvgConcRings,
      NULL, exception)) return MagickFalse;

    //  pss->DispMap = DestroyImage (pss->DispMap); // FIXME: temp


    if (pss->debug) {
      WriteFrame ("wsx_main.png", main_img, 0, exception);
      WriteFrame ("wsx_sub.png", sub_img, 0, exception);
      WriteFrame ("wsx_acr.png", pss->AvgConcRings, 0, exception);
    }

    if (pss->phApprox) {
      if (!CalcApproxScale (&ws, &pss->cr, main_img, pss->AvgConcRings,
                         pss->DispMap, exception))
      {
        fprintf (stderr, "CalcApproxScale failed\n");
      }

      if (pss->verbose) {
        fprintf (stderr,
          "whatscale: approx scale=%.*g +/-=%.*g score=%.*g\n",
           pss->precision, ws.scale,
           pss->precision, ws.porm,
           pss->precision, ws.score);
      }

      if (ws.possSmaller) {
        double rangeFact = pss->cr.maxScale / pss->cr.minScale;
        pss->cr.maxScale = sqrt (pss->cr.minScale * pss->cr.maxScale);
        pss->cr.minScale = pss->cr.maxScale / rangeFact;
        if (pss->verbose) fprintf (stderr,
          "Try smaller scale, range %g to %g\n",
          pss->cr.minScale, pss->cr.maxScale);
      } else if (ws.possLarger) {
        double rangeFact = pss->cr.maxScale / pss->cr.minScale;
        pss->cr.minScale = sqrt (pss->cr.minScale * pss->cr.maxScale);
        pss->cr.maxScale = pss->cr.minScale * rangeFact;
        if (pss->verbose) fprintf (stderr,
          "Try larger scale, range %g to %g\n",
          pss->cr.minScale, pss->cr.maxScale);
      } else {
        break;
      }
    }
  }
  if (i==MAX_ITER) fprintf (stderr,
    "whatscale: scale probably wrong, between %g and %g, scale=%g?\n",
    pss->cr.minScale, pss->cr.maxScale,
    ws.scale);


  if (pss->phRefine) {
    // Save, in case GoldSect fails.
    double saveScale = ws.scale;
    double savePorm = ws.porm;
    double saveScore = ws.score;

    ws.porm = pss->tolerance;
    ws.tolerance = pss->tolerance;

    if (GoldSectScale (&ws, main_img, sub_img, exception)) {
      if (pss->verbose) {
        fprintf (stderr,
          "whatscale: gss scale=%.*g +/-=%.*g score=%.*g\n",
           pss->precision, ws.scale,
           pss->precision, ws.porm,
           pss->precision, ws.score);
      }
    } else {
      fprintf (stderr, "whatscale: GoldSectScale failed\n");
      ws.scale = saveScale;
      ws.porm = savePorm;
      ws.score = saveScore;
    }
  }

  fprintf (pss->fh_data,
    "whatscale: scale=%.*g scalePorm=%.*g scaleScore=%.*g\n",
     pss->precision, ws.scale,
     pss->precision, ws.porm,
     pss->precision, ws.score);

  DeInitWhatScale (&ws);

  if (pss->doScale) {

    MagickBooleanType bestfit = MagickTrue;
    double scaleArray[2];
    scaleArray[0] = ws.scale; // Scale factor.
    scaleArray[1] = 0.0;   // Rotation angle.

#if IMV6OR7==6
    //SetPixelCacheVirtualMethod (sub_img, TransparentVirtualPixelMethod);
    SetImageVirtualPixelMethod (sub_img, TransparentVirtualPixelMethod);
#else
    //SetPixelCacheVirtualMethod (sub_img,
    //  TransparentVirtualPixelMethod, exception);
    SetImageVirtualPixelMethod (sub_img, TransparentVirtualPixelMethod, exception);
#endif

    Image *resB = DistortImage (sub_img, ScaleRotateTranslateDistortion,
      2, scaleArray, bestfit, exception);
    if (!resB) return MagickFalse;
    // We don't ResetPage (resB);

    if (pss->verbose) fprintf (stderr,
      "WhatScale: created image %lix%li %lix%li%+li%+li\n",
      resB->columns, resB->rows,
      resB->page.width, resB->page.height,
      resB->page.x, resB->page.y);

    DeleteImageFromList (&sub_img);

    ReplaceImageInList (&main_img, resB);
    // Replace messes up the images pointer. Make it good:
    *images = GetFirstImageInList (main_img);
  }

  return MagickTrue;
}



ModuleExport size_t whatscaleImage(Image **images,const int argc,
  const char **argv,ExceptionInfo *exception)
{
  MagickBooleanType
    status;

  SubSrchT
    ss;

  assert(images != (Image **) NULL);
  assert(*images != (Image *) NULL);
  assert((*images)->signature == MAGICK_CORE_SIG);

  InitSubSrch (&ss);

  ss.precision = GetMagickPrecision();

  status = menu (argc, argv, &ss);
  if (status == MagickFalse)
    return (-1);

  int ListLen = (int)GetImageListLength(*images);
  if (ListLen != 2) {
    fprintf (stderr, "whatscale: needs 2 images\n");
    return (-1);
  }

  if (!whatscale (images, &ss, exception)) return -1;

  DeInitSubSrch (&ss);

  return(MagickImageFilterSignature);
}
