#! /bin/bash

# Distort the input so that,
# when printed and wrapped around a pipe,
# that would look like the input image.

# $1 is input file
# $2 is output file
# $3 distance of viewpoint above pipe
# $4 is either "angleMax" or "radius", describing the next argument.
# $5 is either the half the angle of the pipe we can see, or the radius of the pipe (in pixels)
#   If the angle: in degrees, more than 0, less than or equal to 90.
#   The script does NOT test for valid angles.

# See also wrapPipe.sh.

if [ $# != 5 ]; then
  echo "Usage: $0 input output VT paramStr param"
  exit 1
fi

infile=$1
outfile=$2
VT=$3
paramStr=$4
paramNum=$5

MAGICK=magick.exe

isBad=0

ww=%[w]
hh=%[h]
hmax=%[fx:h/2]
halfh="%[fx:int(h/2)]"
halfw="%[fx:int(w/2)]"

declare $( ${MAGICK} \
  $infile \
  -precision 15 \
  -format "ww=${ww}\nhh=${hh}\nhmax=${hmax}\nhalfh=${halfh}\nhalfw=${halfw}\n" \
  info: )

OVD="%[fx:atan(${hmax}/${VT})]"

declare $( ${MAGICK} \
  xc: \
  -precision 15 \
  -format "OVD=${OVD}\n" \
  info: )

echo VT=$VT ww=$ww hh=$hh hmax=$hmax halfh=$halfh OVD=$OVD

if [ "${paramStr}" = "angleMax" ]; then
  echo "Param is maxAngle: ${paramNum}"
  angleMax=%[fx:${paramNum}*pi/180]

  declare $( ${MAGICK} \
    xc: \
    -precision 15 \
    -format "angleMax=${angleMax}\n" \
    info: )

  VDO="%[fx:atan(${VT}/${hmax}) + pi/2 - ${angleMax}]"

  declare $( ${MAGICK} \
    xc: \
    -precision 15 \
    -format "VDO=${VDO}\n" \
    info: )

  rr="%[fx:sin(${OVD})*${VT} / (sin(${VDO}) - sin(${OVD}))]"

  declare $( ${MAGICK} \
    xc: \
    -precision 15 \
    -format "rr=${rr}\n" \
    info: )

  echo "Calculated angleMax=${angleMax} OVD=${OVD} VDO=${VDO} rr=${rr}"

elif [ "${paramStr}" = "radius" ]; then
  echo "Param is radius: ${paramNum}"
  rr=${paramNum}

  VDO="%[fx:pi-asin(min(1,sin(${OVD})*(${rr}+${VT})/${rr}))]"

  declare $( ${MAGICK} \
    xc: \
    -precision 15 \
    -format "VDO=${VDO}\n" \
    info: )

  angleMax="%[fx:pi/2-${VDO}+atan(${VT}/${hmax})]"

  declare $( ${MAGICK} \
    xc: \
    -precision 15 \
    -format "angleMax=${angleMax}\n" \
    info: )

  echo "Calculated angleMax=${angleMax} OVD=${OVD} VDO=${VDO} rr=${rr}"

else
  echo "Unknown paramStr [${paramStr}]"
  exit 1
fi


horizonAng=$(${MAGICK} \
  xc: \
  -precision 15 \
  -format "%[fx:acos(${rr}/(${rr}+${VT}))]\n" \
  info: )

isBad=$(${MAGICK} \
  xc: \
  -precision 15 \
  -format "%[fx:${angleMax}>=${horizonAng}||${rr}<=0||${VDO}<=pi/2?1:${isBad}]\n" \
  info: )

echo isBad=$isBad

${MAGICK} \
  xc: \
  -precision 15 \
  -format "VDOdeg=%[fx:${VDO}*180/pi]\nangleMaxdeg=%[fx:${angleMax}*180/pi]\nhorizonAngdeg=%[fx:${horizonAng}*180/pi]\n" \
  info:



amax=$angleMax

ex=$(${MAGICK} xc: -precision 15 -format "%[fx:int(($rr * ($amax) * 2 - $hh) / 2)]" info: )

dx=$(${MAGICK} xc: -precision 15 -format "%[fx:${halfh} + $ex - (${halfh}-${hh}/2)]" info: )

dx2=$(${MAGICK} xc: -precision 15 -format "%[fx:${dx}*2]" info: )

echo horizonAng=$horizonAng ww=$ww hh=$hh hmax=$hmax amax=$amax halfh=$halfh rr=$rr ex=$ex dx=$dx dx2=$dx2


exh=$(${MAGICK} xc: -precision 15 -format "%[fx:int(($ww*($VT + $rr - $rr * cos($angleMax))/$VT-$ww)/2)]" info: )

wwNewHlf=$(${MAGICK} xc: -precision 15 -format "%[fx:${halfw} + $exh - (${halfw}-${ww}/2)]" info: )

wwNew=$(${MAGICK} xc: -precision 15 -format "%[fx:$wwNewHlf * 2]" info: )

echo exh=$exh wwNewHlf=$wwNewHlf wwNew=$wwNew

${MAGICK} \
  \( $infile \
     -background None \
     -gravity Center \
     -extent ${wwNew}x${dx2} \
  \) \
  \( \
     -size ${dx2}x1 xc: \
     -fx "dr=(i-${dx})/${rr}; ($rr*sin(dr)*$VT/($VT+$rr-$rr*cos(dr))+${dx})/w" \
     -rotate 90 \
     -scale "${wwNew}x${dx2}!" \
     \( -size ${dx2}x1 xc: \
        -fx "dh=(0-$wwNewHlf)/${wwNew}; al=(i-$dx)/${rr}; ${wwNew}/(${wwNew}-2) * dh * $VT/($VT + $rr - $rr * cos(al))+0.5" \
        \( xc: \
          -fx "dh=(${wwNew}-1-$wwNewHlf)/${wwNew}; al=(i-$dx)/${rr}; ${wwNew}/(${wwNew}-2) * dh * $VT/($VT + $rr - $rr * cos(al))+0.5" \
        \) \
        -morph %[fx:${wwNew}-2] \
        -reverse -append \
        -rotate 90 \
        -alpha off \
     \) \
     +swap \
     \( -size ${wwNew}x${dx2} xc:"gray(50%)" \) \
     -combine \
     -alpha off \
  \) \
  -compose Distort -composite \
  $outfile

exit $isBad
